/*++

Copyright (c) 1991, 1992, 1993 - 1997 Microsoft Corporation

Module Name:

    initunlo.c

Abstract:

    This module contains the code that is very specific to initialization
    and unload operations in the serial driver

Author:

    Anthony V. Ercolano 26-Sep-1991

Environment:

    Kernel mode

Revision History :

   Scott Brenden            1-Jan-1997     Preliminary PnP/Power Mods
   Louis J. Giliberto, Jr.  19-May-1997    Final PnP/Power Mods

--*/

#include "precomp.h"

//
// This is the actual definition of SerialDebugLevel.
// Note that it is only defined if this is a "debug"
// build.
//
#if DBG
extern ULONG SerialDebugLevel = 0;
#endif

//
// All our global variables except DebugLevel stashed in one
// little package
//
SERIAL_GLOBALS SerialGlobals;

static const PHYSICAL_ADDRESS SerialPhysicalZero = {0};

//
// We use this to query into the registry as to whether we
// should break at driver entry.
//

SERIAL_FIRMWARE_DATA    driverDefaults;



SERIAL_MEM_COMPARES
SerialMemCompare(
                IN PHYSICAL_ADDRESS A,
                IN ULONG SpanOfA,
                IN PHYSICAL_ADDRESS B,
                IN ULONG SpanOfB
                );

//
// This is exported from the kernel.  It is used to point
// to the address that the kernel debugger is using.
//
extern PUCHAR *KdComPortInUse;

// 
// Had to change some of the INIT functions to be paged
// so they would still be around when I got a START_DEVICE
// Irp
//
#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT,DriverEntry)
#pragma alloc_text(PAGESER,SerialInitMultiPort)
#pragma alloc_text(PAGESER,SerialInitOneController)
#pragma alloc_text(PAGESER,SerialInitController)
#pragma alloc_text(PAGESER,SerialDoesPortExist)
#pragma alloc_text(PAGESER,SerialGetMappedAddress)
#pragma alloc_text(PAGESER,SerialSetupExternalNaming)
#pragma alloc_text(PAGESER,SerialMemCompare)
// #pragma alloc_text(PAGESER,SerialGetDivisorFromBaud)
#pragma alloc_text(PAGESER,SerialUnload)
#pragma alloc_text(PAGESER,SerialCleanupDevice)
#pragma alloc_text(PAGESER,SerialCleanupExternalNaming)
#pragma alloc_text(PAGESER,SerialLogError)
#pragma alloc_text(PAGESER,SerialIRPPrologue)
#pragma alloc_text(PAGESER,SerialKillPendingIrps)
#endif


NTSTATUS
DriverEntry(
           IN PDRIVER_OBJECT DriverObject,
           IN PUNICODE_STRING RegistryPath
           )

/*++

Routine Description:

    The entry point that the system point calls to initialize
    any driver.

    This routine will gather the configuration information,
    report resource usage, attempt to initialize all serial
    devices, connect to interrupts for ports.  If the above
    goes reasonably well it will fill in the dispatch points,
    reset the serial devices and then return to the system.

Arguments:

    DriverObject - Just what it says,  really of little use
    to the driver itself, it is something that the IO system
    cares more about.

    PathToRegistry - points to the entry for this driver
    in the current control set of the registry.

Return Value:

    Always STATUS_SUCCESS

--*/

{
   //
   // Lock the paged code in their frames
   //

   PVOID lockPtr = MmLockPagableCodeSection(SerialUnload);

   PAGED_CODE();


   //
   // Initialize all our globals
   //

   InitializeListHead(&SerialGlobals.AllFdos);

   //
   // Call to find out default values to use for all the devices that the 
   // driver controls, including whether or not to break on entry.
   //

   SerialGetConfigDefaults(&driverDefaults, RegistryPath);

#if DBG
   //
   // Set global debug output level
   // 

   SerialDebugLevel = driverDefaults.DebugLevel;
#endif

   //
   // Break on entry if requested via registry
   //

   if (driverDefaults.ShouldBreakOnEntry) {
      DbgBreakPoint();
   }


   //
   // Just dump out how big the extension is.
   //

   SerialDump(SERDIAG1, ("SERIAL: The number of bytes in the extension is: %d"
			 "\n", sizeof(SERIAL_DEVICE_EXTENSION)));


   //
   // Initialize the Driver Object with driver's entry points
   //

   DriverObject->DriverUnload                          = SerialUnload;
   DriverObject->DriverExtension->AddDevice            = SerialAddDevice;

   DriverObject->MajorFunction[IRP_MJ_FLUSH_BUFFERS]   = SerialFlush;
   DriverObject->MajorFunction[IRP_MJ_WRITE]           = SerialWrite;
   DriverObject->MajorFunction[IRP_MJ_READ]            = SerialRead;
   DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]  = SerialIoControl;
#if 0 // Power support is currently broken.
   DriverObject->MajorFunction[IRP_MJ_CREATE]          = SerialOpenClose;
   DriverObject->MajorFunction[IRP_MJ_CLOSE]           = SerialOpenClose;
#else
   DriverObject->MajorFunction[IRP_MJ_CREATE]          = SerialCreateOpen;
   DriverObject->MajorFunction[IRP_MJ_CLOSE]           = SerialClose;
#endif
   DriverObject->MajorFunction[IRP_MJ_CLEANUP]         = SerialCleanup;
   DriverObject->MajorFunction[IRP_MJ_PNP]             = SerialPnpDispatch;
   DriverObject->MajorFunction[IRP_MJ_POWER]           = SerialPowerDispatch;

   DriverObject->MajorFunction[IRP_MJ_QUERY_INFORMATION]
      = SerialQueryInformationFile;
   DriverObject->MajorFunction[IRP_MJ_SET_INFORMATION]
      = SerialSetInformationFile;


#define SerialDoLegacyConversion() (~0)

   //
   // Enumerate and Initialize legacy devices if necessary.  This should go away
   // and be done by setup.  Also note that driver default and other registry
   // settings are obtained here, so if this function goes away, that
   // functionality will have to be moved somewhere else.
   //

   if (SerialDoLegacyConversion()) {
      //
      // Legacy init uses some of the other text pages, so lock them here
      //

      PVOID legacyLockPtr
	 = MmLockPagableCodeSection(SerialInitController);

      (void)SerialEnumerateLegacy(DriverObject, RegistryPath, &driverDefaults);

      MmUnlockPagableImageSection(legacyLockPtr);
   }

   //
   // Unlock pageable text
   //

   MmUnlockPagableImageSection(lockPtr);

   return STATUS_SUCCESS;
}


BOOLEAN
SerialIRPPrologue(IN PSERIAL_DEVICE_EXTENSION PDevExt)
/*++

Routine Description:

   This function must be called at any IRP dispatch entry point.  It,
   with SerialIRPEpilogue(), keeps track of all pending IRP's for the given
   PDevObj.
   
Arguments:

   PDevObj - Pointer to the device object we are tracking pending IRP's for.

Return Value:

    TRUE if the device can accept IRP's.

--*/
{
   InterlockedIncrement(&PDevExt->PendingIRPCnt);

   return PDevExt->DevicePNPAccept == SERIAL_PNPACCEPT_OK ? TRUE : FALSE;
}



VOID
SerialIRPEpilogue(IN PSERIAL_DEVICE_EXTENSION PDevExt)
/*++

Routine Description:

   This function must be called at any IRP dispatch entry point.  It,
   with SerialIRPPrologue(), keeps track of all pending IRP's for the given
   PDevObj.
   
Arguments:

   PDevObj - Pointer to the device object we are tracking pending IRP's for.

Return Value:

   None.

--*/
{
   ULONG pendingCnt;

   pendingCnt = InterlockedDecrement(&PDevExt->PendingIRPCnt);

   if (pendingCnt == 0) {
      KeSetEvent(&PDevExt->PendingIRPEvent, IO_NO_INCREMENT, FALSE);
   }
}

NTSTATUS
SerialIoCallDriver(PSERIAL_DEVICE_EXTENSION PDevExt, PDEVICE_OBJECT PDevObj,
		   PIRP PIrp)
{
   NTSTATUS status;

   status = IoCallDriver(PDevObj, PIrp);
   SerialIRPEpilogue(PDevExt);
   return status;
}

NTSTATUS
SerialPoCallDriver(PSERIAL_DEVICE_EXTENSION PDevExt, PDEVICE_OBJECT PDevObj,
		   PIRP PIrp)
{
   NTSTATUS status;

   status = PoCallDriver(PDevObj, PIrp);
   SerialIRPEpilogue(PDevExt);
   return status;
}




NTSTATUS
SerialAddDevice(
               IN PDRIVER_OBJECT DriverObject,
               IN PDEVICE_OBJECT Pdo
               )

/*++

Routine Description:

    This routine creates a functional device object for com ports in the
    system and attaches them to the physical device objects for the ports


Arguments:

    DriverObject - a pointer to the object for this driver
    
    PhysicalDeviceObject - a pointer to the physical object we need to attach to

Return Value:

    status from device creation and initialization

--*/

{
   PDEVICE_OBJECT  fdo         = NULL;
   PDEVICE_OBJECT  lowerDevice = NULL;
   NTSTATUS        status;

   PSERIAL_DEVICE_EXTENSION    deviceExtension;


   PAGED_CODE();

   SerialDump(SERTRACECALLS, ("SERIAL: Enter SerialAddDevice with pdo %x\n",
			      Pdo));

   if (Pdo == NULL) {

      //
      //  Bugbug: This is where enumeration occurs.
      //           One possible use for this is to add the user defined
      //           ports from the registry
      //

      //
      // For now: just return no more devices
      //

      SerialDump(SERERRORS, ("SerialAddDevice: Enumeration request, returning"
			     " NO_MORE_ENTRIES\n"));

      return (STATUS_NO_MORE_ENTRIES);
   }



   //
   // create and initialize the new functional device object
   //

   status = SerialCreateFdo(DriverObject, Pdo, &fdo);

   if (!NT_SUCCESS(status)) {

      SerialDump(SERERRORS,
                 ("SerialAddDevice - error creating Fdo [%#08lx]\n", status));
      return status;
   }


   //
   // Layer our FDO on top of the PDO
   // The return value is a pointer to the device object to which the
   //    fdo is actually attached.
   //

   lowerDevice = IoAttachDeviceToDeviceStack(fdo, Pdo);


   //
   // No status. Do the best we can.
   //
   ASSERT(lowerDevice);


   //
   // fdo source, pdo is target
   //

   deviceExtension                     = fdo->DeviceExtension;
   deviceExtension->LowerDeviceObject  = lowerDevice;

   fdo->Flags    |= DO_POWER_PAGABLE;

   SerialDump(SERTRACECALLS,("SERIAL: Leave SerialAddDevice\n") );
   return status;
}




NTSTATUS 
SerialGetRegistryKeyValue (
                          IN HANDLE Handle,
                          IN PWCHAR KeyNameString,
                          IN ULONG KeyNameStringLength,
                          IN PVOID Data,
                          IN ULONG DataLength
                          )
/*++

Routine Description:

    Reads a registry key value from an already opened registry key.
    
Arguments:

    Handle              Handle to the opened registry key
    
    KeyNameString       ANSI string to the desired key

    KeyNameStringLength Length of the KeyNameString

    Data                Buffer to place the key value in

    DataLength          Length of the data buffer

Return Value:

    STATUS_SUCCESS if all works, otherwise status of system call that
    went wrong.

--*/
{
   UNICODE_STRING              keyName;
   ULONG                       length;
   PKEY_VALUE_FULL_INFORMATION fullInfo;

   NTSTATUS                    ntStatus = STATUS_NO_MEMORY;

   SerialDump(SERTRACECALLS, ("SERIAL: Enter SerialGetRegistryKeyValue\n"));


   RtlInitUnicodeString (&keyName, KeyNameString);

   length = sizeof(KEY_VALUE_FULL_INFORMATION) + KeyNameStringLength
      + DataLength;
   fullInfo = ExAllocatePool(PagedPool, length); 

   if (fullInfo) {
      ntStatus = ZwQueryValueKey (Handle,
                                  &keyName,
                                  KeyValueFullInformation,
                                  fullInfo,
                                  length,
                                  &length);

      if (NT_SUCCESS(ntStatus)) {
         //
         // If there is enough room in the data buffer, copy the output
         //

         if (DataLength >= fullInfo->DataLength) {
            RtlCopyMemory (Data, 
                           ((PUCHAR) fullInfo) + fullInfo->DataOffset, 
                           fullInfo->DataLength);
         }
      }

      ExFreePool(fullInfo);
   }

   return ntStatus;
}



NTSTATUS 
SerialPutRegistryKeyValue (
                          IN HANDLE Handle,
                          IN PWCHAR KeyNameString,
                          IN ULONG KeyNameStringLength,
			  IN ULONG Dtype,
                          IN PVOID Data,
                          IN ULONG DataLength
                          )
/*++

Routine Description:

    Writes a registry key value to an already opened registry key.
    
Arguments:

    Handle              Handle to the opened registry key
    
    KeyNameString       ANSI string to the desired key

    KeyNameStringLength Length of the KeyNameString
    
    Dtype		REG_XYZ value type

    Data                Buffer to place the key value in

    DataLength          Length of the data buffer

Return Value:

    STATUS_SUCCESS if all works, otherwise status of system call that
    went wrong.

--*/
{
   NTSTATUS ntStatus = STATUS_NO_MEMORY;
   UNICODE_STRING keyname;

   SerialDump(SERTRACECALLS,("SERIAL: Enter SerialPutRegistryKeyValue\n"));

   RtlInitUnicodeString(&keyname, NULL);
   keyname.MaximumLength = (USHORT)(KeyNameStringLength + sizeof(WCHAR));
   keyname.Buffer = ExAllocatePool(PagedPool, keyname.MaximumLength);

   if (keyname.Buffer == NULL) {
      return ntStatus;
   }

   RtlAppendUnicodeToString(&keyname, KeyNameString);

   ntStatus = ZwSetValueKey(Handle,
			    &keyname,
			    0,	
			    Dtype,
			    Data,
			    DataLength);

   ExFreePool(keyname.Buffer);
   
   return ntStatus;
}

#define DEVICE_OBJECT_NAME_LENGTH  128
#define SYMBOLIC_NAME_LENGTH       128

BOOLEAN
SerialCleanLists(IN PVOID Context)
{
   PSERIAL_DEVICE_EXTENSION pDevExt = Context;

   if (!IsListEmpty(&pDevExt->TopLevelSharers)) {
      RemoveEntryList(&pDevExt->TopLevelSharers);

      //
      // Others are chained on this interrupt, so we don't want to
      // disconnect it.
      //

      pDevExt->Interrupt = NULL;
   }

   if (!IsListEmpty(&pDevExt->CommonInterruptObject)) {
      RemoveEntryList(&pDevExt->CommonInterruptObject);

      //
      // Others are sharing this interrupt object so we detach ourselves
      // from it this way instead of disconnecting.
      // 

      pDevExt->Interrupt = NULL;
   }

   if (!IsListEmpty(&pDevExt->AllFdos)) {
      RemoveEntryList(&pDevExt->AllFdos);
   }

   return TRUE;

}



NTSTATUS
SerialRemoveFdo(IN PDEVICE_OBJECT pFdo)
{
   PSERIAL_DEVICE_EXTENSION pDevExt = pFdo->DeviceExtension;

   SerialDump(SERTRACECALLS,("SERIAL: Enter SerialRemoveFdo\n"));

   //
   // Only do these many things if the device has started
   //

   if (pDevExt->Flags & SERIAL_FLAGS_STARTED) {

      //
      // Remove us from any lists we may be on
      //

      KeSynchronizeExecution(pDevExt->Interrupt, SerialCleanLists, pDevExt);

      //
      // Stop servicing interrupts if we are the owner
      //

      if (pDevExt->Interrupt != NULL) {
         SerialDump(SERPNPPOWER,("SERIAL: Remove - disconnecting interrupt\n"));
         IoDisconnectInterrupt(pDevExt->Interrupt);
      }

      //
      // Detach from the lower device object
      //

      IoDetachDevice(pDevExt->LowerDeviceObject);

      //
      // Stop handling timers
      //

      KeCancelTimer(&pDevExt->ReadRequestTotalTimer);
      KeCancelTimer(&pDevExt->ReadRequestIntervalTimer);
      KeCancelTimer(&pDevExt->WriteRequestTotalTimer);
      KeCancelTimer(&pDevExt->ImmediateTotalTimer);
      KeCancelTimer(&pDevExt->XoffCountTimer);
      KeCancelTimer(&pDevExt->LowerRTSTimer);

      //
      // Stop servicing DPC's
      //

      KeRemoveQueueDpc(&pDevExt->CompleteWriteDpc);
      KeRemoveQueueDpc(&pDevExt->CompleteReadDpc);
      KeRemoveQueueDpc(&pDevExt->TotalReadTimeoutDpc);
      KeRemoveQueueDpc(&pDevExt->IntervalReadTimeoutDpc);
      KeRemoveQueueDpc(&pDevExt->TotalWriteTimeoutDpc);
      KeRemoveQueueDpc(&pDevExt->CommErrorDpc);
      KeRemoveQueueDpc(&pDevExt->CompleteImmediateDpc);
      KeRemoveQueueDpc(&pDevExt->TotalImmediateTimeoutDpc);
      KeRemoveQueueDpc(&pDevExt->CommWaitDpc);
      KeRemoveQueueDpc(&pDevExt->XoffCountTimeoutDpc);
      KeRemoveQueueDpc(&pDevExt->XoffCountCompleteDpc);
      KeRemoveQueueDpc(&pDevExt->StartTimerLowerRTSDpc);
      KeRemoveQueueDpc(&pDevExt->PerhapsLowerRTSDpc);

      //
      // Undo external names
      //
      SerialUndoExternalNaming(pDevExt);

      //
      // If necessary, unmap the device registers.
      //

      if (pDevExt->UnMapRegisters) {
         MmUnmapIoSpace(pDevExt->Controller, pDevExt->SpanOfController);
      }

      if (pDevExt->UnMapStatus) {
         MmUnmapIoSpace(pDevExt->InterruptStatus,
                        pDevExt->SpanOfInterruptStatus);
      }

   } else { // Never started
      IoDetachDevice(pDevExt->LowerDeviceObject);
   }

   //
   // Free memory allocated in the extension
   //

   if (pDevExt->NtNameForPort.Buffer != NULL) {
      ExFreePool(pDevExt->NtNameForPort.Buffer);
   }

   if (pDevExt->DeviceName.Buffer != NULL) {
      ExFreePool(pDevExt->DeviceName.Buffer);
   }

   if (pDevExt->SymbolicLinkName.Buffer != NULL) {
      ExFreePool(pDevExt->SymbolicLinkName.Buffer);
   }

   if (pDevExt->DosName.Buffer != NULL) {
      ExFreePool(pDevExt->DosName.Buffer);
   }

   if (pDevExt->ObjectDirectory.Buffer) {
      ExFreePool(pDevExt->ObjectDirectory.Buffer);
   }

#if 0
   if (pDevExt->NtNameForPort.Buffer) {
      ExFreePool(pDevExt->NtNameForPort.Buffer);
   }
#endif


   //
   // Delete the fdo
   //

   IoDeleteDevice(pFdo);


   SerialDump(SERTRACECALLS,("SERIAL: Leave SerialRemoveFdo\n"));

   return STATUS_SUCCESS;
}


VOID
SerialKillPendingIrps(PDEVICE_OBJECT DeviceObject)
/*++

Routine Description:

   This routine kills any irps pending for the passed device object.
   
Arguments:

    DeviceObject - Pointer to the device object whose irps must die.

Return Value:

    VOID

--*/
{
   PSERIAL_DEVICE_EXTENSION extension = DeviceObject->DeviceExtension;
   KIRQL oldIrql;
   
   PAGED_CODE();

   SerialDump (SERTRACECALLS,("SERIAL: Enter SerialKillPendingIrps\n"));

   //
   // First kill all the reads and writes.
   //

    SerialKillAllReadsOrWrites(
        DeviceObject,
        &extension->WriteQueue,
        &extension->CurrentWriteIrp
        );

    SerialKillAllReadsOrWrites(
        DeviceObject,
        &extension->ReadQueue,
        &extension->CurrentReadIrp
        );

    //
    // Next get rid of purges.
    //

    SerialKillAllReadsOrWrites(
        DeviceObject,
        &extension->PurgeQueue,
        &extension->CurrentPurgeIrp
        );

    //
    // Get rid of any mask operations.
    //

    SerialKillAllReadsOrWrites(
        DeviceObject,
        &extension->MaskQueue,
        &extension->CurrentMaskIrp
        );

    //
    // Now get rid a pending wait mask irp.
    //

    IoAcquireCancelSpinLock(&oldIrql);

    if (extension->CurrentWaitIrp) {

        PDRIVER_CANCEL cancelRoutine;

        cancelRoutine = extension->CurrentWaitIrp->CancelRoutine;
        extension->CurrentWaitIrp->Cancel = TRUE;

        if (cancelRoutine) {

            extension->CurrentWaitIrp->CancelIrql = oldIrql;
            extension->CurrentWaitIrp->CancelRoutine = NULL;

            cancelRoutine(
                DeviceObject,
                extension->CurrentWaitIrp
                );

        }

    } else {

        IoReleaseCancelSpinLock(oldIrql);

    }


    SerialDump (SERTRACECALLS,("SERIAL: Leave SerialKillPendingIrps\n"));
}


NTSTATUS
SerialCreateFdo(
               IN PDRIVER_OBJECT   DriverObject,
               IN PDEVICE_OBJECT   Pdo,
               OUT PDEVICE_OBJECT *NewDeviceObject
               )

/*++

Routine Description:

    This routine will create and initialize a functional device object to
    be attached to a Serial controller PDO.

Arguments:

    DriverObject - a pointer to the driver object this is created under
    DeviceObject - a location to store the pointer to the new device object

Return Value:

    STATUS_SUCCESS if everything was successful
    reason for failure otherwise

--*/

{
   //
   // BADBAD:  Sucking up too much allocation on stack; should be paged
   //          allocation
   //

   WCHAR                       deviceNameBuffer[DEVICE_OBJECT_NAME_LENGTH];
   UNICODE_STRING              deviceName;

   WCHAR                       linkNameBuffer[SYMBOLIC_NAME_LENGTH];
   UNICODE_STRING              linkName;

   WCHAR                       numberBuffer[10];
   UNICODE_STRING              numberString;

   ULONG                       resultLength;
   WCHAR                       Buffer[SYMBOLIC_NAME_LENGTH];
   HANDLE                      keyHandle;

   HANDLE                      registryHandle  = NULL;
   LONG                        deviceNumber    = 0;
   PDEVICE_OBJECT              deviceObject    = NULL;
   PSERIAL_DEVICE_EXTENSION    deviceExtension = NULL;
   NTSTATUS                    status          = STATUS_SUCCESS;
   WCHAR                       pBuffer[SYMBOLIC_NAME_LENGTH];



   //
   // bugbug:  need to rethink this whole naming scheme.  It doesn't matter what 
   //          the NT name for the device is, but the name under DosDevices
   //          should be what the user and all the applications are expecting,
   //          namely COM1, COM2, etc.  Need to find a way of mapping the Pnp
   //          names that are made up to these DosDevices names.
   //

   //
   // WORSE BUGBUG:
   // Need to clean up these names in case of failure in StartDevice or consider
   // mvoing this there.
   //


   static  ULONG currentInstance=0;

   UNICODE_STRING NtNameForPort;

   WCHAR          InstanceNumberBuffer[20];


   PAGED_CODE();

   SerialDump (SERTRACECALLS,("SERIAL: Enter SerialCreateFdo\n"));

   //
   // Zero out allocated memory pointers so we know if they must be freed
   //

   RtlZeroMemory (&deviceName, sizeof(UNICODE_STRING));
   RtlZeroMemory (&numberString, sizeof(UNICODE_STRING));
   RtlZeroMemory (&linkName, sizeof(UNICODE_STRING));
   RtlZeroMemory (deviceNameBuffer, 64);
   RtlZeroMemory (linkNameBuffer, 64);

   numberString.MaximumLength  = 10;
   numberString.Buffer         = &numberBuffer[0];

   deviceName.MaximumLength    = DEVICE_OBJECT_NAME_LENGTH*sizeof(WCHAR);
   deviceName.Buffer           = &deviceNameBuffer[0];

   linkName.MaximumLength      = SYMBOLIC_NAME_LENGTH*sizeof(WCHAR);
   linkName.Buffer             = &linkNameBuffer[0];

   RtlZeroMemory(
                deviceName.Buffer,
                deviceName.MaximumLength
                );

   RtlAppendUnicodeToString(
                           &deviceName,
                           L"\\Device\\Serial"
                           );

   RtlInitUnicodeString(
                       &NtNameForPort,
                       NULL
                       );

   NtNameForPort.MaximumLength = sizeof(InstanceNumberBuffer);
   NtNameForPort.Buffer = InstanceNumberBuffer;

   RtlIntegerToUnicodeString(
                            currentInstance++,
                            10,
                            &NtNameForPort
                            );

   RtlAppendUnicodeStringToString(
                                 &deviceName,
                                 &NtNameForPort
                                 );

   if (!deviceName.Length) {
      //
      // We didn't find a name for our devices in the registry. Error
      // out and return.
      //

      SerialDump(SERERRORS, ("SerialAddDevice: Get Nt name failed - %x \n",
			     status));
      goto SerialFdoError;    

   }

   

   //
   // We have been given a NT name, just create the device object
   //

   status = IoCreateDevice(DriverObject, sizeof(SERIAL_DEVICE_EXTENSION),
			   &deviceName, FILE_DEVICE_SERIAL_PORT, 0, TRUE,
			   &deviceObject);
   

   if (!NT_SUCCESS(status)) {
      SerialDump(SERERRORS, ("SerialAddDevice: Create device failed - %x \n",
			     status));
      goto SerialFdoError;    
   }


   //
   // The device object has a pointer to an area of non-paged
   // pool allocated for this device.  This will be the device
   // extension. Zero it out.
   //

   deviceExtension = deviceObject->DeviceExtension;
   RtlZeroMemory(deviceExtension, sizeof(SERIAL_DEVICE_EXTENSION));

   //
   // Initialize the count of IRP's pending
   //

   deviceExtension->PendingIRPCnt = 1;

   deviceExtension->NtNameForPort.Buffer = ExAllocatePool(PagedPool,
						      deviceName.MaximumLength);

   if (deviceExtension->NtNameForPort.Buffer == NULL) {
      SerialDump(SERERRORS, ("SerialAddDevice: Cannot allocate memory for "
			     "NtName\n"));
      status = STATUS_NO_MEMORY;
      goto SerialFdoError; 
   }

   deviceExtension->NtNameForPort.MaximumLength = deviceName.MaximumLength;
   RtlAppendUnicodeStringToString(&deviceExtension->NtNameForPort,
				  &deviceName);


   //
   // Open the device registry key and use the symbolic name the PnP
   // manager is recommending for us.
   //

   status = IoOpenDeviceRegistryKey(Pdo, PLUGPLAY_REGKEY_DEVICE, 
				    STANDARD_RIGHTS_READ, &keyHandle);

   if (!NT_SUCCESS(status)) {
      //
      // This is a fatal error.  If we can't get to our registry key,
      // we are sunk.
      //
      SerialDump(SERERRORS, ("SERIAL: IoOpenDeviceRegistryKey failed - %x \n",
			     status));
      goto SerialFdoError;
   }

   //
   // Fetch PortName which contains the suggested REG_SZ symbolic name.
   //

   status = SerialGetRegistryKeyValue(keyHandle, L"PortName",
				      sizeof(L"PortName"), pBuffer,
				      sizeof(pBuffer));

   if (!NT_SUCCESS(status)) {

      //
      // BUGBUG 
      // This is for PCMCIA which currently puts the name under Identifier.
      // In the future it should also use PortName.
      //

      status = SerialGetRegistryKeyValue(keyHandle, L"Identifier",
					 sizeof(L"Identifier"),
					 pBuffer, sizeof(pBuffer));

      if (!NT_SUCCESS(status)) {

	 //
	 // Hmm.  Either we have to pick a name or bail...
	 //
	 // ...we will bail.
	 //

	 SerialDump(SERERRORS, ("SERIAL: Getting PortName/Identifier failed - "
				"%x\n", status));
	 ZwClose (keyHandle);
	 goto SerialFdoError;
      }

   }

   ZwClose (keyHandle);

   //
   // Create the "\\DosDevices\\<symbolicName>" string
   //
   RtlAppendUnicodeToString(&linkName, L"\\");
   RtlAppendUnicodeToString(&linkName, DEFAULT_DIRECTORY);
   RtlAppendUnicodeToString(&linkName, L"\\");
   RtlAppendUnicodeToString(&linkName, pBuffer);


   //
   // Allocate Pool and save the nt device name in the device extension.
   //

   deviceExtension->DeviceName.Buffer =
      ExAllocatePool(PagedPool, deviceName.Length + sizeof(WCHAR));

   if (!deviceExtension->DeviceName.Buffer) {

      SerialLogError(
                    DriverObject,
                    NULL,
                    SerialPhysicalZero,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    19,
                    STATUS_SUCCESS,
                    SERIAL_INSUFFICIENT_RESOURCES,
                    0,
                    NULL,
                    0,
                    NULL
                    );
      SerialDump(
                SERERRORS,
                ("SERIAL: Couldn't allocate memory for Nt name\n"));

      return STATUS_SUCCESS;

   } else {

      deviceExtension->DeviceName.MaximumLength = deviceName.Length
	 + sizeof(WCHAR);

      //
      // Zero fill it.
      //

      RtlZeroMemory(
                   deviceExtension->DeviceName.Buffer,
                   deviceExtension->DeviceName.MaximumLength
                   );

      RtlAppendUnicodeStringToString(
                                    &deviceExtension->DeviceName,
                                    &deviceName
                                    );

   }


   //
   // Allocate Pool and save the symbolic link name in the device extension.
   //

   deviceExtension->SymbolicLinkName.Buffer =
      ExAllocatePool(PagedPool, linkName.Length + sizeof(WCHAR));

   if (!deviceExtension->SymbolicLinkName.Buffer) {

      SerialLogError(
                    DriverObject,
                    NULL,
                    SerialPhysicalZero,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    19,
                    STATUS_SUCCESS,
                    SERIAL_INSUFFICIENT_RESOURCES,
                    0,
                    NULL,
                    0,
                    NULL
                    );
      SerialDump(
                SERERRORS,
                ("SERIAL: Couldn't allocate memory for symbolic link name\n"));

      //
      // BUGBUG - why success??
      //
      return STATUS_SUCCESS;

   } else {

      deviceExtension->SymbolicLinkName.MaximumLength = linkName.Length
	 + sizeof(WCHAR);

      //
      // Zero fill it.
      //

      RtlZeroMemory(
                   deviceExtension->SymbolicLinkName.Buffer,
                   deviceExtension->SymbolicLinkName.MaximumLength
                   );

      RtlAppendUnicodeStringToString(
                                    &deviceExtension->SymbolicLinkName,
                                    &linkName
                                    );

   }


   deviceExtension->DosName.Buffer =
   ExAllocatePool (PagedPool, 64 + sizeof(WCHAR));

   if (!deviceExtension->DosName.Buffer) {

      SerialLogError(
                    DriverObject,
                    NULL,
                    SerialPhysicalZero,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    19,
                    STATUS_SUCCESS,
                    SERIAL_INSUFFICIENT_RESOURCES,
                    0,
                    NULL,
                    0,
                    NULL
                    );
      SerialDump(
                SERERRORS,
                ("SERIAL: Couldn't allocate memory for Dos name\n"));

      //
      // BUGBUG - why success??
      //
      return STATUS_SUCCESS;

   } else {

      deviceExtension->DosName.MaximumLength = 64 + sizeof(WCHAR);

      //
      // Zero fill it.
      //

      deviceExtension->DosName.Length = 0;

      RtlZeroMemory(
                   deviceExtension->DosName.Buffer,
                   deviceExtension->DosName.MaximumLength
                   );

      //
      // bugbug:  Right now I am saving the symbolic name in the form "\DosDevices\COM1".
      //          Later I am writing that name to the \HKLM\Hardware\devicemap section
      //          where the control panel ports applet looks for it.  However, the ports
      //          applet is looking for a name of the form "COM1" and doesn't work with 
      //          the way it is now.  Need to fix one or the other.
      //


      RtlAppendUnicodeToString(&deviceExtension->DosName, pBuffer);
      RtlZeroMemory(((PUCHAR)(&deviceExtension->DosName.Buffer[0]))
		    + deviceExtension->DosName.Length, sizeof(WCHAR));

      SerialDump(SERDIAG1, ("SERIAL: DosName is %wZ\n",
			    &deviceExtension->DosName));
   }

   //
   // Set up the device extension.
   //

   deviceExtension->DeviceIsOpened = FALSE;
   deviceExtension->IsStarted      = FALSE;
   deviceExtension->DeviceObject   = deviceObject;
   deviceExtension->DriverObject   = DriverObject;
   deviceExtension->DeviceObject   = deviceObject;
   deviceExtension->PowerState     = PowerDeviceD0;

   deviceExtension->TxFifoAmount           = driverDefaults.TxFIFODefault;
   deviceExtension->CreatedSymbolicLink    = TRUE;

   InitializeListHead(&deviceExtension->CommonInterruptObject);
   InitializeListHead(&deviceExtension->TopLevelSharers);
   InitializeListHead(&deviceExtension->AllFdos);
   InitializeListHead(&deviceExtension->ReadQueue);
   InitializeListHead(&deviceExtension->WriteQueue);
   InitializeListHead(&deviceExtension->MaskQueue);
   InitializeListHead(&deviceExtension->PurgeQueue);

   KeInitializeEvent(&deviceExtension->PendingIRPEvent, SynchronizationEvent,
		     FALSE);


   deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

   *NewDeviceObject = deviceObject;

   SerialDump (SERTRACECALLS,("SERIAL: Leave SerialCreateFdo\n") );
   return STATUS_SUCCESS;


   SerialFdoError:

   SerialDump (SERERRORS,("SERIAL: SerialCreateFdo Error, Cleaning up\n") );

   //
   // Free the allocated strings for the NT and symbolic names if they exist.
   //

   if (deviceObject) {

      if (deviceExtension->NtNameForPort.Buffer != NULL) {
         ExFreePool(deviceExtension->NtNameForPort.Buffer);
      }

      if (deviceExtension->DeviceName.Buffer != NULL) {
	 ExFreePool(deviceExtension->DeviceName.Buffer);
      }

      if (deviceExtension->SymbolicLinkName.Buffer != NULL) {
         ExFreePool(deviceExtension->SymbolicLinkName.Buffer);
      }

      IoDeleteDevice(deviceObject);
   }

   *NewDeviceObject = NULL;

   SerialDump (SERTRACECALLS,("SERIAL: Leave SerialCreateFdo\n") );
   return status;
}



BOOLEAN
SerialSingleToMulti(PVOID Context)
/*++

Routine Description:

    This routine converts a root device set up to be a single port
    device to a multiport device while that device is running.
    
Arguments:

    Context - Actually a pointer to the device extension of the root
              device we are turning into a multiport device.
	      
Return Value:

    Always TRUE

--*/
{
   PSERIAL_DEVICE_EXTENSION pDevExt = (PSERIAL_DEVICE_EXTENSION)Context;
   PSERIAL_MULTIPORT_DISPATCH pOurIsrContext;
   PSERIAL_MULTIPORT_DISPATCH pNewIsrContext
      = (PSERIAL_MULTIPORT_DISPATCH)pDevExt->NewExtension;


   //
   // BUGBUG
   // Propagate shared interrupt links here as needed.
   //

   //
   // Stomp OurIsrContext since we are going from one to many
   // thus our previous context was just pDevExt and doesn't
   // need to be released (i.e., no call to ExFreePool() is needed).
   //

   pOurIsrContext = pDevExt->OurIsrContext = pDevExt->TopLevelOurIsrContext
      = pNewIsrContext;

   //
   // We are now multiport
   //

   pDevExt->PortOnAMultiportCard = TRUE;

   //
   // Update our personal extensions slot
   //

   pOurIsrContext->Extensions[pDevExt->PortIndex - 1] = pDevExt;
   pOurIsrContext->InterruptStatus = pDevExt->InterruptStatus;

   pDevExt->CIsrSw.Context = pOurIsrContext;

   //
   // We have to pick a new ISR and a new context.
   // As soon as this is done, the ISR will change, so we have to
   // be ready to handle things there.
   //

   if (pDevExt->Indexed == FALSE) {
      pOurIsrContext->UsablePortMask = 1 << (pDevExt->PortIndex - 1);
      pOurIsrContext->MaskInverted = pDevExt->MaskInverted;

      pDevExt->OurIsr = SerialBitMappedMultiportIsr;
      pDevExt->CIsrSw.IsrFunc = SerialBitMappedMultiportIsr;
      pDevExt->TopLevelOurIsr = SerialBitMappedMultiportIsr;
   } else {
      pDevExt->OurIsr = SerialIndexedMultiportIsr;
      pDevExt->CIsrSw.IsrFunc = SerialIndexedMultiportIsr;
      pDevExt->TopLevelOurIsr = SerialIndexedMultiportIsr;
   }

   return TRUE;
}


BOOLEAN
SerialAddToMulti(PVOID Context)
/*++

Routine Description:

    This routine adds a new port to a multiport device while that device is
    running.
    
Arguments:

    Context - Actually a pointer to the device extension of the root
              device we are adding a port to.
	      
Return Value:

    Always TRUE

--*/
{
   PSERIAL_DEVICE_EXTENSION pDevExt = (PSERIAL_DEVICE_EXTENSION)Context;
   PSERIAL_MULTIPORT_DISPATCH pOurIsrContext
      = (PSERIAL_MULTIPORT_DISPATCH)pDevExt->OurIsrContext;

   if (pDevExt->Indexed == FALSE) {
      pOurIsrContext->UsablePortMask |= 1 << (pDevExt->NewPortIndex - 1);
      pOurIsrContext->MaskInverted |= pDevExt->NewMaskInverted;
   }

   //
   // Add us to the linked list of common interrupt objects if we are not
   // already in it. We may be if there is another device besides our
   // multiport card.
   //

   if (IsListEmpty(&((PSERIAL_DEVICE_EXTENSION)pDevExt->NewExtension)
		   ->CommonInterruptObject)) {
      InsertTailList(&pDevExt->CommonInterruptObject,
		     &((PSERIAL_DEVICE_EXTENSION)pDevExt->NewExtension)
		     ->CommonInterruptObject);
   }

   //
   // Map us in so the ISR can find us.
   //
   pOurIsrContext->Extensions[pDevExt->NewPortIndex - 1]
      = pDevExt->NewExtension;

   return TRUE;
}



NTSTATUS
SerialInitMultiPort(IN PSERIAL_DEVICE_EXTENSION pDevExt,
		    IN PCONFIG_DATA PConfigData, IN PDEVICE_OBJECT PFdo)
/*++

Routine Description:

    This routine initializes a multiport device by adding a port to an existing
    one.
    
Arguments:

    pDevExt - pointer to the device extension of the root of the multiport
              device.
	      
    PConfigData - pointer to the config data for the new port 
	     
    PFdo - pointer to the Fdo for the new port
	      
Return Value:

    STATUS_SUCCESS on success, appropriate error on failure.

--*/
{
   PSERIAL_DEVICE_EXTENSION pOurIsrContext
      = (PSERIAL_DEVICE_EXTENSION)pDevExt->OurIsrContext;
   NTSTATUS status;
   PSERIAL_MULTIPORT_DISPATCH pDispatch;

   PAGED_CODE();

   //
   // First, see if we can initialize the one we have found
   // BUGBUG should have translated data
   //
   status = SerialInitOneController(PFdo, PConfigData);

   if (!NT_SUCCESS(status)) {
      return status;
   }

   //
   // OK.  He's good to go.  Find the root controller.  He may
   // currently be a single, so we have to change him to multi.
   //

   if (pDevExt->PortOnAMultiportCard != TRUE) {

      pDispatch = pDevExt->NewExtension
	 = ExAllocatePool(NonPagedPool, sizeof(SERIAL_MULTIPORT_DISPATCH));

      if (pDispatch == NULL) {
	 // FAIL and CLEANUP
	 return STATUS_INSUFFICIENT_RESOURCES;
      }

      RtlZeroMemory(pDispatch, sizeof(*pDispatch));

      KeSynchronizeExecution(pDevExt->Interrupt, SerialSingleToMulti, pDevExt);
   }

   //
   // Update some important fields
   //

   ((PSERIAL_DEVICE_EXTENSION)PFdo->DeviceExtension)->PortOnAMultiportCard
      = TRUE;
   ((PSERIAL_DEVICE_EXTENSION)PFdo->DeviceExtension)->OurIsr = NULL;

   //
   // We can fiddle with these fields because no other
   // context besides our KeSynchronizeExecution()
   // touches them and we have a partial ordering on our execution.
   //

   pDevExt->NewPortIndex = PConfigData->PortIndex;
   pDevExt->NewMaskInverted = PConfigData->MaskInverted;
   pDevExt->NewExtension = PFdo->DeviceExtension;

   //
   // Copy the root interrupt object to the new guy.
   //

   ((PSERIAL_DEVICE_EXTENSION)PFdo->DeviceExtension)->Interrupt
      = pDevExt->Interrupt;

   //
   // Now, we can add the new guy.  He will be hooked in
   // immediately, so we need to be able to handle interrupts.
   // BUGBUG - maybe need to put a flag in extension showing
   // that we are not ready.
   //

   KeSynchronizeExecution(pDevExt->Interrupt, SerialAddToMulti, pDevExt);

   return STATUS_SUCCESS;
}



NTSTATUS
SerialInitController(
		     IN PDEVICE_OBJECT Fdo,
		     IN PCONFIG_DATA ConfigData
		     )

/*++

Routine Description:

    Really too many things to mention here.  In general initializes 
    kernel synchronization structures, allocates the typeahead buffer,
    sets up defaults, etc.

    NOTE:  The names have already been set up in an earlier routine
           (AddDevice ()).

Arguments:

    Fdo          - Functional device object for the device to be started

    ConfigData   - Pointer to a record for a single port.
    
Return Value:

    STATUS_SUCCCESS if everything went ok.  A !NT_SUCCESS status
    otherwise.

--*/

{

   PSERIAL_DEVICE_EXTENSION fdoExtension = Fdo->DeviceExtension;
   PDEVICE_OBJECT pdo = fdoExtension->LowerDeviceObject;

   //
   // This will hold the string that we need to use to describe
   // the name of the device to the IO system.
   //
   UNICODE_STRING uniNameString;

   //
   // Holds the NT Status that is returned from each call to the
   // kernel and executive.
   //
   NTSTATUS status = STATUS_SUCCESS;

   //
   // Indicates that a conflict was detected for resources
   // used by this device.
   //
   BOOLEAN conflictDetected = FALSE;

   PAGED_CODE();


   SerialDump(
             SERDIAG1,
             ("SERIAL: Initializing for configuration record of %wZ\n",
              &fdoExtension->DeviceName)
             );

   //
   // This compare is done using **untranslated** values since that is what
   // the kernel must also use to address this I/O space
   //
      if ((*KdComPortInUse) == ((PUCHAR)(ConfigData->Controller.LowPart))) {
      SerialDump(
                SERERRORS,
                ("SERIAL: Kernel debugger is using port at address %x\n"
                 "------  Serial driver will not load port %wZ\n",
                 *KdComPortInUse,&fdoExtension->SymbolicLinkName)
                );

      SerialLogError(
                    Fdo->DriverObject,
                    NULL,
                    ConfigData->TrController,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    3,
                    STATUS_SUCCESS,
                    SERIAL_KERNEL_DEBUGGER_ACTIVE,
                    fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                    fdoExtension->SymbolicLinkName.Buffer,
                    0,
                    NULL
                    );

      return STATUS_INSUFFICIENT_RESOURCES;
   }


   //
   // Propagate that it is a jensen.
   //

   fdoExtension->Jensen = ConfigData->Jensen;


   //
   // Initialize the spinlock associated with fields read (& set)
   // by IO Control functions.
   //

   KeInitializeSpinLock (&fdoExtension->ControlLock);


   //
   // Initialize the timers used to timeout operations.
   //

   KeInitializeTimer (&fdoExtension->ReadRequestTotalTimer);
   KeInitializeTimer (&fdoExtension->ReadRequestIntervalTimer);
   KeInitializeTimer (&fdoExtension->WriteRequestTotalTimer);
   KeInitializeTimer (&fdoExtension->ImmediateTotalTimer);
   KeInitializeTimer (&fdoExtension->XoffCountTimer);
   KeInitializeTimer (&fdoExtension->LowerRTSTimer);


   //
   // Intialialize the dpcs that will be used to complete
   // or timeout various IO operations.
   //

   KeInitializeDpc(
                  &fdoExtension->CompleteWriteDpc,
                  SerialCompleteWrite,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->CompleteReadDpc,
                  SerialCompleteRead,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->TotalReadTimeoutDpc,
                  SerialReadTimeout,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->IntervalReadTimeoutDpc,
                  SerialIntervalReadTimeout,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->TotalWriteTimeoutDpc,
                  SerialWriteTimeout,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->CommErrorDpc,
                  SerialCommError,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->CompleteImmediateDpc,
                  SerialCompleteImmediate,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->TotalImmediateTimeoutDpc,
                  SerialTimeoutImmediate,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->CommWaitDpc,
                  SerialCompleteWait,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->XoffCountTimeoutDpc,
                  SerialTimeoutXoff,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->XoffCountCompleteDpc,
                  SerialCompleteXoff,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->StartTimerLowerRTSDpc,
                  SerialStartTimerLowerRTS,
                  fdoExtension
                  );

   KeInitializeDpc(
                  &fdoExtension->PerhapsLowerRTSDpc,
                  SerialInvokePerhapsLowerRTS,
                  fdoExtension
                  );


   if (!((ConfigData->ClockRate == 1843200) ||
         (ConfigData->ClockRate == 3072000) ||
         (ConfigData->ClockRate == 4233600) ||
         (ConfigData->ClockRate == 8000000))) {

      SerialLogError(
                    Fdo->DriverObject,
                    Fdo,
                    ConfigData->TrController,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    6,
                    STATUS_SUCCESS,
                    SERIAL_UNSUPPORTED_CLOCK_RATE,
                    fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                    fdoExtension->SymbolicLinkName.Buffer,
                    0,
                    NULL
                    );

      SerialDump(
                SERERRORS,
                ("SERIAL: Invalid clock rate specified for %wZ\n",
                 &fdoExtension->DeviceName)
                );

      status = STATUS_SERIAL_NO_DEVICE_INITED;
      goto ExtensionCleanup;

   }


   //
   // Save the value of clock input to the part.  We use this to calculate
   // the divisor latch value.  The value is in Hertz.
   //

   fdoExtension->ClockRate = ConfigData->ClockRate;


   //
   // Specify that this driver only supports buffered IO.  This basically
   // means that the IO system copies the users data to and from
   // system supplied buffers.
   //

   Fdo->Flags |= DO_BUFFERED_IO;


   //
   // Map the memory for the control registers for the serial device
   // into virtual memory.
   //
   fdoExtension->Controller =
      SerialGetMappedAddress(ConfigData->InterfaceType,
                             ConfigData->BusNumber,
                             ConfigData->TrController,
                             ConfigData->SpanOfController, 
                             (BOOLEAN)ConfigData->AddressSpace,
                             &fdoExtension->UnMapRegisters);

   if (!fdoExtension->Controller) {

      SerialLogError(
                    Fdo->DriverObject,
                    fdoExtension->DeviceObject,
                    ConfigData->TrController,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    7,
                    STATUS_SUCCESS,
                    SERIAL_REGISTERS_NOT_MAPPED,
                    fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                    fdoExtension->SymbolicLinkName.Buffer,
                    0,
                    NULL
                    );

      SerialDump(
                SERERRORS,
                ("SERIAL: Could not map memory for device registers for %wZ\n",
                 &fdoExtension->NtNameForPort)
                );

      fdoExtension->UnMapRegisters = FALSE;
      status = STATUS_NONE_MAPPED;
      goto ExtensionCleanup;

   }

   SerialDump(SERDIAG1, ("SERIAL: NtnameForport is %wZ\n",
			 &fdoExtension->NtNameForPort));

   fdoExtension->AddressSpace          = ConfigData->AddressSpace;
   fdoExtension->OriginalController    = ConfigData->Controller;
   fdoExtension->SpanOfController      = ConfigData->SpanOfController;


   //
   // if we have an interrupt status then map it.
   //

   fdoExtension->InterruptStatus =
      (PUCHAR)ConfigData->TrInterruptStatus.QuadPart;

   if (fdoExtension->InterruptStatus) {

      fdoExtension->InterruptStatus
         = SerialGetMappedAddress(ConfigData->InterfaceType,
                                  ConfigData->BusNumber,                                                           
                                  ConfigData->TrInterruptStatus,
                                  ConfigData->SpanOfInterruptStatus,
                                  (BOOLEAN)ConfigData->AddressSpace,
                                  &fdoExtension->UnMapStatus);


      if (!fdoExtension->InterruptStatus) {

         SerialLogError(
                       Fdo->DriverObject,
                       Fdo,
                       ConfigData->TrController,
                       SerialPhysicalZero,
                       0,
                       0,
                       0,
                       8,
                       STATUS_SUCCESS,
                       SERIAL_REGISTERS_NOT_MAPPED,
                       fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                       fdoExtension->SymbolicLinkName.Buffer,
                       0,
                       NULL
                       );

         SerialDump(SERERRORS, ("SERIAL: Could not map memory for interrupt "
                                "status for %wZ\n",
                                &fdoExtension->NtNameForPort));

         fdoExtension->UnMapRegisters = FALSE;
         status = STATUS_NONE_MAPPED;
         goto ExtensionCleanup;

      }

      fdoExtension->OriginalInterruptStatus = ConfigData->InterruptStatus;
      fdoExtension->SpanOfInterruptStatus = ConfigData->SpanOfInterruptStatus;
 

   }


   if (ConfigData->PermitSystemWideShare) {

      fdoExtension->InterruptShareable = TRUE;

   } else {

      fdoExtension->InterruptShareable = FALSE;

   }


   //
   // Save off the interface type and the bus number.
   //

   fdoExtension->InterfaceType = ConfigData->InterfaceType;
   fdoExtension->BusNumber     = ConfigData->BusNumber;

   fdoExtension->PortIndex = ConfigData->PortIndex;
   fdoExtension->Indexed = (BOOLEAN)ConfigData->Indexed;
   fdoExtension->MaskInverted = ConfigData->MaskInverted;

   //
   // Get the translated interrupt vector, level, and affinity
   //

   fdoExtension->InterruptMode     = ConfigData->InterruptMode;
   fdoExtension->OriginalIrql      = ConfigData->OriginalIrql;
   fdoExtension->OriginalVector    = ConfigData->OriginalVector;


   //
   // PnP uses the passed translated values rather than calling
   // HalGetInterruptVector()
   //

   fdoExtension->Vector = ConfigData->TrVector;
   fdoExtension->Irql = (UCHAR)ConfigData->TrIrql;
   fdoExtension->ProcessorAffinity = ConfigData->Affinity;


   //
   // Set up the Isr.
   //

   fdoExtension->OurIsr        = SerialISR;
   fdoExtension->OurIsrContext = fdoExtension;


   //
   // If the user said to permit sharing within the device, propagate this
   // through.
   //

   fdoExtension->PermitShare = ConfigData->PermitShare;


   //
   // Before we test whether the port exists (which will enable the FIFO)
   // convert the rx trigger value to what should be used in the register.
   //
   // If a bogus value was given - crank them down to 1.
   //

   switch (ConfigData->RxFIFO) {

   case 1:

      fdoExtension->RxFifoTrigger = SERIAL_1_BYTE_HIGH_WATER;
      break;

   case 4:

      fdoExtension->RxFifoTrigger = SERIAL_4_BYTE_HIGH_WATER;
      break;

   case 8:

      fdoExtension->RxFifoTrigger = SERIAL_8_BYTE_HIGH_WATER;
      break;

   case 14:

      fdoExtension->RxFifoTrigger = SERIAL_14_BYTE_HIGH_WATER;
      break;

   default:

      fdoExtension->RxFifoTrigger = SERIAL_1_BYTE_HIGH_WATER;
      break;

   }


   if ((ConfigData->TxFIFO > 16) ||
       (ConfigData->TxFIFO < 1)) {

      fdoExtension->TxFifoAmount = 1;

   } else {

      fdoExtension->TxFifoAmount = ConfigData->TxFIFO;

   }


   if (!SerialDoesPortExist(
                           fdoExtension,
                           &fdoExtension->SymbolicLinkName,
                           ConfigData->ForceFifoEnable,
                           ConfigData->LogFifo
                           )) {

      //
      // We couldn't verify that there was actually a
      // port. No need to log an error as the port exist
      // code will log exactly why.
      //

      SerialDump(
                SERERRORS,
                ("SERIAL: Does Port exist test failed for %wZ\n",
                 &fdoExtension->DeviceName)
                );

      status = STATUS_NO_SUCH_DEVICE;
      goto ExtensionCleanup;

   }


   //
   // If the user requested that we disable the port, then
   // do it now.  Log the fact that the port has been disabled.
   //

   if (ConfigData->DisablePort) {

      SerialDump(
                SERERRORS,
                ("SERIAL: disabled port %wZ as requested in configuration\n",
                 &fdoExtension->NtNameForPort)
                );

      status = STATUS_NO_SUCH_DEVICE;

      SerialLogError(
                    Fdo->DriverObject,
                    Fdo,
                    ConfigData->Controller,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    57,
                    STATUS_SUCCESS,
                    SERIAL_DISABLED_PORT,
                    fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                    fdoExtension->SymbolicLinkName.Buffer,
                    0,
                    NULL
                    );

      goto ExtensionCleanup;

   }



   //
   // Set up the default device control fields.
   // Note that if the values are changed after
   // the file is open, they do NOT revert back
   // to the old value at file close.
   //

   fdoExtension->SpecialChars.XonChar      = SERIAL_DEF_XON;
   fdoExtension->SpecialChars.XoffChar     = SERIAL_DEF_XOFF;
   fdoExtension->HandFlow.ControlHandShake = SERIAL_DTR_CONTROL;
   fdoExtension->HandFlow.FlowReplace      = SERIAL_RTS_CONTROL;


   //
   // Default Line control protocol. 7E1
   //
   // Seven data bits.
   // Even parity.
   // 1 Stop bits.
   //

   fdoExtension->LineControl = SERIAL_7_DATA |
                               SERIAL_EVEN_PARITY |
                               SERIAL_NONE_PARITY;

   fdoExtension->ValidDataMask = 0x7f;
   fdoExtension->CurrentBaud   = 1200;


   //
   // We set up the default xon/xoff limits.
   //
   // BUGBUG:  This may be a bogus value.  It looks like the BufferSize
   //          is not set up until the device is actually opened.
   //

   fdoExtension->HandFlow.XoffLimit    = fdoExtension->BufferSize >> 3;
   fdoExtension->HandFlow.XonLimit     = fdoExtension->BufferSize >> 1;

   fdoExtension->BufferSizePt8 = ((3*(fdoExtension->BufferSize>>2))+
                                  (fdoExtension->BufferSize>>4));

   SerialDump(
             SERDIAG1,
             ("SERIAL: The default interrupt read buffer size is: %d\n"
              "------  The XoffLimit is                         : %d\n"
              "------  The XonLimit is                          : %d\n"
              "------  The pt 8 size is                         : %d\n",
              fdoExtension->BufferSize,
              fdoExtension->HandFlow.XoffLimit,
              fdoExtension->HandFlow.XonLimit,
              fdoExtension->BufferSizePt8)
             );


   //
   // Go through all the "named" baud rates to find out which ones
   // can be supported with this port.
   //

   fdoExtension->SupportedBauds = SERIAL_BAUD_USER;

   {

      SHORT junk;

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)75,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_075;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)110,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_110;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)135,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_134_5;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)150,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_150;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)300,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_300;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)600,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_600;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)1200,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_1200;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)1800,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_1800;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)2400,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_2400;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)4800,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_4800;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)7200,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_7200;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)9600,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_9600;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)14400,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_14400;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)19200,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_19200;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)38400,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_38400;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)56000,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_56K;

      }
      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)57600,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_57600;

      }
      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)115200,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_115200;

      }

      if (!NT_ERROR(SerialGetDivisorFromBaud(
                                            fdoExtension->ClockRate,
                                            (LONG)128000,
                                            &junk
                                            ))) {

         fdoExtension->SupportedBauds |= SERIAL_BAUD_128K;

      }

   }


   //
   // Mark this device as not being opened by anyone.  We keep a
   // variable around so that spurious interrupts are easily
   // dismissed by the ISR.
   //

   fdoExtension->DeviceIsOpened = FALSE;

   //
   // Store values into the extension for interval timing.
   //

   //
   // If the interval timer is less than a second then come
   // in with a short "polling" loop.
   //
   // For large (> then 2 seconds) use a 1 second poller.
   //

   fdoExtension->ShortIntervalAmount.QuadPart  = -1;
   fdoExtension->LongIntervalAmount.QuadPart   = -10000000;
   fdoExtension->CutOverAmount.QuadPart        = 200000000;


   //
   // Common error path cleanup.  If the status is
   // bad, get rid of the device extension, device object
   // and any memory associated with it.
   //

   ExtensionCleanup: ;

   return status;

}



NTSTATUS
SerialInitOneController(IN PDEVICE_OBJECT Fdo, IN PCONFIG_DATA ConfigData)
/*++

Routine Description:

    This routine will call the real port initialization code.  It sets
    up the ISR and Context correctly for a one port device.

Arguments:

    All args are simply passed along.

Return Value:

    Status returned from the controller initialization routine.

--*/

{

   NTSTATUS status;
   PSERIAL_DEVICE_EXTENSION pDevExt;

   PAGED_CODE();

   status = SerialInitController(Fdo, ConfigData);

   if (NT_SUCCESS(status)) {

      pDevExt = Fdo->DeviceExtension;

      //
      // We successfully initialized the single controller.
      // Stick the isr routine and the parameter for it
      // back into the extension.
      //

      pDevExt->OurIsr = SerialISR;
      pDevExt->OurIsrContext = pDevExt;
      pDevExt->TopLevelOurIsr = SerialISR;
      pDevExt->TopLevelOurIsrContext = pDevExt;
   }

   return status;

}


BOOLEAN
SerialDoesPortExist(
                   IN PSERIAL_DEVICE_EXTENSION Extension,
                   IN PUNICODE_STRING InsertString,
                   IN ULONG ForceFifo,
                   IN ULONG LogFifo
                   )

/*++

Routine Description:

    This routine examines several of what might be the serial device
    registers.  It ensures that the bits that should be zero are zero.

    In addition, this routine will determine if the device supports
    fifo's.  If it does it will enable the fifo's and turn on a boolean
    in the extension that indicates the fifo's presence.

    NOTE: If there is indeed a serial port at the address specified
          it will absolutely have interrupts inhibited upon return
          from this routine.

    NOTE: Since this routine should be called fairly early in
          the device driver initialization, the only element
          that needs to be filled in is the base register address.

    NOTE: These tests all assume that this code is the only
          code that is looking at these ports or this memory.

          This is a not to unreasonable assumption even on
          multiprocessor systems.

Arguments:

    Extension - A pointer to a serial device extension.
    InsertString - String to place in an error log entry.
    ForceFifo - !0 forces the fifo to be left on if found.
    LogFifo - !0 forces a log message if fifo found.

Return Value:

    Will return true if the port really exists, otherwise it
    will return false.

--*/

{


   UCHAR regContents;
   BOOLEAN returnValue = TRUE;
   UCHAR oldIERContents;
   UCHAR oldLCRContents;
   USHORT value1;
   USHORT value2;
   KIRQL oldIrql;

   PAGED_CODE();


   //
   // Save of the line control.
   //

   oldLCRContents = READ_LINE_CONTROL(Extension->Controller);

   //
   // Make sure that we are *aren't* accessing the divsior latch.
   //

   WRITE_LINE_CONTROL(
                     Extension->Controller,
                     (UCHAR)(oldLCRContents & ~SERIAL_LCR_DLAB)
                     );

   oldIERContents = READ_INTERRUPT_ENABLE(Extension->Controller);

   //
   // Go up to power level for a very short time to prevent
   // any interrupts from this device from coming in.
   //

   KeRaiseIrql(
              POWER_LEVEL,
              &oldIrql
              );

   WRITE_INTERRUPT_ENABLE(
                         Extension->Controller,
                         0x0f
                         );

   value1 = READ_INTERRUPT_ENABLE(Extension->Controller);
   value1 = value1 << 8;
   value1 |= READ_RECEIVE_BUFFER(Extension->Controller);

   READ_DIVISOR_LATCH(
                     Extension->Controller,
                     &value2
                     );

   WRITE_LINE_CONTROL(
                     Extension->Controller,
                     oldLCRContents
                     );

   //
   // Put the ier back to where it was before.  If we are on a
   // level sensitive port this should prevent the interrupts
   // from coming in.  If we are on a latched, we don't care
   // cause the interrupts generated will just get dropped.
   //

   WRITE_INTERRUPT_ENABLE(
                         Extension->Controller,
                         oldIERContents
                         );

   KeLowerIrql(oldIrql);

   if (value1 == value2) {

      SerialLogError(
                    Extension->DeviceObject->DriverObject,
                    Extension->DeviceObject,
                    Extension->OriginalController,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    62,
                    STATUS_SUCCESS,
                    SERIAL_DLAB_INVALID,
                    InsertString->Length+sizeof(WCHAR),
                    InsertString->Buffer,
                    0,
                    NULL
                    );
      returnValue = FALSE;
      goto AllDone;

   }

   AllDone: ;


   //
   // If we think that there is a serial device then we determine
   // if a fifo is present.
   //

   if (returnValue) {

      //
      // Well, we think it's a serial device.  Absolutely
      // positively, prevent interrupts from occuring.
      //
      // We disable all the interrupt enable bits, and
      // push down all the lines in the modem control
      // We only needed to push down OUT2 which in
      // PC's must also be enabled to get an interrupt.
      //

      DISABLE_ALL_INTERRUPTS(Extension->Controller);

      if (Extension->Jensen) {

         WRITE_MODEM_CONTROL(
                            Extension->Controller,
                            (UCHAR)SERIAL_MCR_OUT2
                            );

      } else {

         WRITE_MODEM_CONTROL(
                            Extension->Controller,
                            (UCHAR)0
                            );

      }

      //
      // See if this is a 16550.  We do this by writing to
      // what would be the fifo control register with a bit
      // pattern that tells the device to enable fifo's.
      // We then read the iterrupt Id register to see if the
      // bit pattern is present that identifies the 16550.
      //

      WRITE_FIFO_CONTROL(
                        Extension->Controller,
                        SERIAL_FCR_ENABLE
                        );

      regContents = READ_INTERRUPT_ID_REG(Extension->Controller);

      if (regContents & SERIAL_IIR_FIFOS_ENABLED) {

         //
         // Save off that the device supports fifos.
         //

         Extension->FifoPresent = TRUE;

         //
         // There is a fine new "super" IO chip out there that
         // will get stuck with a line status interrupt if you
         // attempt to clear the fifo and enable it at the same
         // time if data is present.  The best workaround seems
         // to be that you should turn off the fifo read a single
         // byte, and then re-enable the fifo.
         //

         WRITE_FIFO_CONTROL(
                           Extension->Controller,
                           (UCHAR)0
                           );

         READ_RECEIVE_BUFFER(Extension->Controller);

         //
         // There are fifos on this card.  Set the value of the
         // receive fifo to interrupt when 4 characters are present.
         //

         WRITE_FIFO_CONTROL(Extension->Controller,
			    (UCHAR)(SERIAL_FCR_ENABLE
				    | Extension->RxFifoTrigger
				    | SERIAL_FCR_RCVR_RESET
				    | SERIAL_FCR_TXMT_RESET));

      }

      //
      // The !Extension->FifoPresent is included in the test so that
      // broken chips like the WinBond will still work after we test
      // for the fifo.
      //

      if (!ForceFifo || !Extension->FifoPresent) {

         Extension->FifoPresent = FALSE;
         WRITE_FIFO_CONTROL(
                           Extension->Controller,
                           (UCHAR)0
                           );

      }

      if (Extension->FifoPresent) {

         if (LogFifo) {

            SerialLogError(
                          Extension->DeviceObject->DriverObject,
                          Extension->DeviceObject,
                          Extension->OriginalController,
                          SerialPhysicalZero,
                          0,
                          0,
                          0,
                          15,
                          STATUS_SUCCESS,
                          SERIAL_FIFO_PRESENT,
                          InsertString->Length+sizeof(WCHAR),
                          InsertString->Buffer,
                          0,
                          NULL
                          );

         }

         SerialDump(
                   SERDIAG1,
                   ("SERIAL: Fifo's detected at port address: %x\n",
                    Extension->Controller)
                   );

      }

      //
      // In case we are dealing with a bitmasked multiportcard,
      // that has the mask register enabled, enable all
      // interrupts.
      //

      if (Extension->InterruptStatus) {

         WRITE_PORT_UCHAR(
                         Extension->InterruptStatus,
                         (UCHAR)0xff
                         );

      }

   }

   return returnValue;

}


BOOLEAN
SerialReset(
           IN PVOID Context
           )

/*++

Routine Description:

    This places the hardware in a standard configuration.

    NOTE: This assumes that it is called at interrupt level.


Arguments:

    Context - The device extension for serial device
    being managed.

Return Value:

    Always FALSE.

--*/

{

   PSERIAL_DEVICE_EXTENSION extension = Context;
   UCHAR regContents;
   UCHAR oldModemControl;
   ULONG i;

   //
   // Adjust the out2 bit.
   // This will also prevent any interrupts from occuring.
   //

   oldModemControl = READ_MODEM_CONTROL(extension->Controller);

   if (extension->Jensen) {

      WRITE_MODEM_CONTROL(
                         extension->Controller,
                         (UCHAR)(oldModemControl | SERIAL_MCR_OUT2)
                         );

   } else {


      WRITE_MODEM_CONTROL(
                         extension->Controller,
                         (UCHAR)(oldModemControl & ~SERIAL_MCR_OUT2)
                         );

   }

   //
   // Reset the fifo's if there are any.
   //

   if (extension->FifoPresent) {


      //
      // There is a fine new "super" IO chip out there that
      // will get stuck with a line status interrupt if you
      // attempt to clear the fifo and enable it at the same
      // time if data is present.  The best workaround seems
      // to be that you should turn off the fifo read a single
      // byte, and then re-enable the fifo.
      //

      WRITE_FIFO_CONTROL(
                        extension->Controller,
                        (UCHAR)0
                        );

      READ_RECEIVE_BUFFER(extension->Controller);

      WRITE_FIFO_CONTROL(
                        extension->Controller,
                        (UCHAR)(SERIAL_FCR_ENABLE | extension->RxFifoTrigger |
                                SERIAL_FCR_RCVR_RESET | SERIAL_FCR_TXMT_RESET)
                        );

   }

   //
   // Make sure that the line control set up correct.
   //
   // 1) Make sure that the Divisor latch select is set
   //    up to select the transmit and receive register.
   //
   // 2) Make sure that we aren't in a break state.
   //

   regContents = READ_LINE_CONTROL(extension->Controller);
   regContents &= ~(SERIAL_LCR_DLAB | SERIAL_LCR_BREAK);

   WRITE_LINE_CONTROL(
                     extension->Controller,
                     regContents
                     );

   //
   // Read the receive buffer until the line status is
   // clear.  (Actually give up after a 5 reads.)
   //

   for (i = 0;
       i < 5;
       i++
       ) {

      READ_RECEIVE_BUFFER(extension->Controller);
      if (!(READ_LINE_STATUS(extension->Controller) & 1)) {

         break;

      }

   }

   //
   // Read the modem status until the low 4 bits are
   // clear.  (Actually give up after a 5 reads.)
   //

   for (i = 0;
       i < 1000;
       i++
       ) {

      if (!(READ_MODEM_STATUS(extension->Controller) & 0x0f)) {

         break;

      }

   }

   //
   // Now we set the line control, modem control, and the
   // baud to what they should be.
   //

   SerialSetLineControl(extension);

   SerialSetupNewHandFlow(
                         extension,
                         &extension->HandFlow
                         );

   SerialHandleModemUpdate(
                          extension,
                          FALSE
                          );

   {
      SHORT  appropriateDivisor;
      SERIAL_IOCTL_SYNC s;

      SerialGetDivisorFromBaud(
                              extension->ClockRate,
                              extension->CurrentBaud,
                              &appropriateDivisor
                              );
      s.Extension = extension;
      s.Data = (PVOID)appropriateDivisor;
      SerialSetBaud(&s);
   }

   //
   // Enable which interrupts we want to receive.
   //
   // NOTE NOTE: This does not actually let interrupts
   // occur.  We must still raise the OUT2 bit in the
   // modem control register.  We will do that on open.
   //

   ENABLE_ALL_INTERRUPTS(extension->Controller);

   //
   // Read the interrupt id register until the low bit is
   // set.  (Actually give up after a 5 reads.)
   //

   for (i = 0;
       i < 5;
       i++
       ) {

      if (READ_INTERRUPT_ID_REG(extension->Controller) & 0x01) {

         break;

      }

   }

   //
   // Now we know that nothing could be transmitting at this point
   // so we set the HoldingEmpty indicator.
   //

   extension->HoldingEmpty = TRUE;

   return FALSE;
}


NTSTATUS
SerialGetDivisorFromBaud(
                        IN ULONG ClockRate,
                        IN LONG DesiredBaud,
                        OUT PSHORT AppropriateDivisor
                        )

/*++

Routine Description:

    This routine will determine a divisor based on an unvalidated
    baud rate.

Arguments:

    ClockRate - The clock input to the controller.

    DesiredBaud - The baud rate for whose divisor we seek.

    AppropriateDivisor - Given that the DesiredBaud is valid, the
    LONG pointed to by this parameter will be set to the appropriate
    value.  NOTE: The long is undefined if the DesiredBaud is not
    supported.

Return Value:

    This function will return STATUS_SUCCESS if the baud is supported.
    If the value is not supported it will return a status such that
    NT_ERROR(Status) == FALSE.

--*/

{

   NTSTATUS status = STATUS_SUCCESS;
   SHORT calculatedDivisor;
   ULONG denominator;
   ULONG remainder;

   //
   // Allow up to a 1 percent error
   //

   ULONG maxRemain18 = 18432;
   ULONG maxRemain30 = 30720;
   ULONG maxRemain42 = 42336;
   ULONG maxRemain80 = 80000;
   ULONG maxRemain;

   //
   // Reject any non-positive bauds.
   //

   denominator = DesiredBaud*(ULONG)16;

   if (DesiredBaud <= 0) {

      *AppropriateDivisor = -1;

   } else if ((LONG)denominator < DesiredBaud) {

      //
      // If the desired baud was so huge that it cause the denominator
      // calculation to wrap, don't support it.
      //

      *AppropriateDivisor = -1;

   } else {

      if (ClockRate == 1843200) {
         maxRemain = maxRemain18;
      } else if (ClockRate == 3072000) {
         maxRemain = maxRemain30;
      } else if (ClockRate == 4233600) {
         maxRemain = maxRemain42;
      } else {
         maxRemain = maxRemain80;
      }

      calculatedDivisor = (SHORT)(ClockRate / denominator);
      remainder = ClockRate % denominator;

      //
      // Round up.
      //

      if (((remainder*2) > ClockRate) && (DesiredBaud != 110)) {

         calculatedDivisor++;
      }


      //
      // Only let the remainder calculations effect us if
      // the baud rate is > 9600.
      //

      if (DesiredBaud >= 9600) {

         //
         // If the remainder is less than the maximum remainder (wrt
         // the ClockRate) or the remainder + the maximum remainder is
         // greater than or equal to the ClockRate then assume that the
         // baud is ok.
         //

         if ((remainder >= maxRemain) && ((remainder+maxRemain) < ClockRate)) {
            calculatedDivisor = -1;
         }

      }

      //
      // Don't support a baud that causes the denominator to
      // be larger than the clock.
      //

      if (denominator > ClockRate) {

         calculatedDivisor = -1;

      }

      //
      // Ok, Now do some special casing so that things can actually continue
      // working on all platforms.
      //

      if (ClockRate == 1843200) {

         if (DesiredBaud == 56000) {
            calculatedDivisor = 2;
         }

      } else if (ClockRate == 3072000) {

         if (DesiredBaud == 14400) {
            calculatedDivisor = 13;
         }

      } else if (ClockRate == 4233600) {

         if (DesiredBaud == 9600) {
            calculatedDivisor = 28;
         } else if (DesiredBaud == 14400) {
            calculatedDivisor = 18;
         } else if (DesiredBaud == 19200) {
            calculatedDivisor = 14;
         } else if (DesiredBaud == 38400) {
            calculatedDivisor = 7;
         } else if (DesiredBaud == 56000) {
            calculatedDivisor = 5;
         }

      } else if (ClockRate == 8000000) {

         if (DesiredBaud == 14400) {
            calculatedDivisor = 35;
         } else if (DesiredBaud == 56000) {
            calculatedDivisor = 9;
         }

      }

      *AppropriateDivisor = calculatedDivisor;

   }


   if (*AppropriateDivisor == -1) {

      status = STATUS_INVALID_PARAMETER;

   }

   return status;

}


VOID
SerialUnload(
            IN PDRIVER_OBJECT DriverObject
            )

/*++

Routine Description:

    This routine cleans up all of the memory associated with
    any of the devices belonging to the driver.  It  will
    loop through the device list.

Arguments:

    DriverObject - Pointer to the driver object controling all of the
                   devices.

Return Value:

    None.

--*/

{
   PVOID lockPtr;

   lockPtr = MmLockPagableCodeSection(SerialUnload);

   PAGED_CODE();

   SerialDump(
             SERDIAG3,
             ("SERIAL: In SerialUnload\n")
             );

   MmUnlockPagableImageSection(lockPtr);

}




VOID
SerialCleanupDevice(IN PSERIAL_DEVICE_EXTENSION PDevExt)

/*++

Routine Description:

    This routine will deallocate all of the memory used for
    a particular device.  It will also disconnect any resources
    if need be.

Arguments:

    Extension - Pointer to the device extension which is getting
                rid of all it's resources.

Return Value:

    None.

--*/

{

   PAGED_CODE();

   SerialDump(SERDIAG3, ("SERIAL: in SerialCleanup for extension: %x\n",
                         PDevExt));

   if (PDevExt != NULL) {
      //
      // Stop handling timers
      //

      KeCancelTimer(&PDevExt->ReadRequestTotalTimer);
      KeCancelTimer(&PDevExt->ReadRequestIntervalTimer);
      KeCancelTimer(&PDevExt->WriteRequestTotalTimer);
      KeCancelTimer(&PDevExt->ImmediateTotalTimer);
      KeCancelTimer(&PDevExt->XoffCountTimer);
      KeCancelTimer(&PDevExt->LowerRTSTimer);

      //
      // Stop servicing DPC's
      //

      KeRemoveQueueDpc(&PDevExt->CompleteWriteDpc);
      KeRemoveQueueDpc(&PDevExt->CompleteReadDpc);
      KeRemoveQueueDpc(&PDevExt->TotalReadTimeoutDpc);
      KeRemoveQueueDpc(&PDevExt->IntervalReadTimeoutDpc);
      KeRemoveQueueDpc(&PDevExt->TotalWriteTimeoutDpc);
      KeRemoveQueueDpc(&PDevExt->CommErrorDpc);
      KeRemoveQueueDpc(&PDevExt->CompleteImmediateDpc);
      KeRemoveQueueDpc(&PDevExt->TotalImmediateTimeoutDpc);
      KeRemoveQueueDpc(&PDevExt->CommWaitDpc);
      KeRemoveQueueDpc(&PDevExt->XoffCountTimeoutDpc);
      KeRemoveQueueDpc(&PDevExt->XoffCountCompleteDpc);
      KeRemoveQueueDpc(&PDevExt->StartTimerLowerRTSDpc);
      KeRemoveQueueDpc(&PDevExt->PerhapsLowerRTSDpc);

      //
      // Get rid of all external naming as well as removing
      // the device map entry.
      //	

      SerialCleanupExternalNaming(PDevExt);
   }
}


PVOID
SerialGetMappedAddress(
                      IN INTERFACE_TYPE BusType,
                      IN ULONG BusNumber,
                      PHYSICAL_ADDRESS IoAddress,
                      ULONG NumberOfBytes,
                      ULONG AddressSpace,
                      PBOOLEAN MappedAddress
                      )

/*++

Routine Description:

    This routine maps an IO address to system address space.

Arguments:

    BusType - what type of bus - eisa, mca, isa
    IoBusNumber - which IO bus (for machines with multiple buses).
    IoAddress - base device address to be mapped.
    NumberOfBytes - number of bytes for which address is valid.
    AddressSpace - Denotes whether the address is in io space or memory.
    MappedAddress - indicates whether the address was mapped.
                    This only has meaning if the address returned
                    is non-null.

Return Value:

    Mapped address

--*/

{
   PHYSICAL_ADDRESS cardAddress;
   PVOID address;

   PAGED_CODE();

   //
   // Map the device base address into the virtual address space
   // if the address is in memory space.
   //

   if (!AddressSpace) {

      address = MmMapIoSpace(
                            IoAddress,
                            NumberOfBytes,
                            FALSE
                            );

      *MappedAddress = (BOOLEAN)((address)?(TRUE):(FALSE));


   } else {

      address = (PVOID)IoAddress.LowPart;
      *MappedAddress = FALSE;

   }

   return address;

}

VOID
SerialSetupExternalNaming(
                         IN PSERIAL_DEVICE_EXTENSION Extension
                         )

/*++

Routine Description:

    This routine will be used to create a symbolic link
    to the driver name in the given object directory.

    It will also create an entry in the device map for
    this device - IF we could create the symbolic link.

Arguments:

    Extension - Pointer to the device extension.

Return Value:

    None.

--*/

{

   UNICODE_STRING fullLinkName;
   NTSTATUS status;

   PAGED_CODE();

   //
   // Form the full symbolic link name we wish to create.
   //

   RtlInitUnicodeString(
                       &fullLinkName,
                       NULL
                       );

   //
   // Allocate some pool for the name.
   //

   fullLinkName.MaximumLength = (sizeof(L"\\")*2) +
                                Extension->ObjectDirectory.Length+
                                Extension->SymbolicLinkName.Length+
                                sizeof(WCHAR);



   fullLinkName.Buffer = ExAllocatePool(
                                       PagedPool,
                                       fullLinkName.MaximumLength
                                       );

   if (!fullLinkName.Buffer) {

      //
      // Couldn't allocate space for the name.
      //

      SerialLogError(
                    Extension->DeviceObject->DriverObject,
                    Extension->DeviceObject,
                    Extension->OriginalController,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    51,
                    STATUS_SUCCESS,
                    SERIAL_INSUFFICIENT_RESOURCES,
                    0,
                    NULL,
                    0,
                    NULL
                    );
      SerialDump(
                SERERRORS,
                ("SERIAL: Couldn't allocate space for the symbolic \n"
                 "------- name for creating the link\n"
                 "------- for port %wZ\n",
                 &Extension->DeviceName)
                );

   } else {

      RtlZeroMemory(
                   fullLinkName.Buffer,
                   fullLinkName.MaximumLength
                   );

      RtlAppendUnicodeToString(
                              &fullLinkName,
                              L"\\"
                              );

      RtlAppendUnicodeStringToString(
                                    &fullLinkName,
                                    &Extension->ObjectDirectory
                                    );

      RtlAppendUnicodeToString(
                              &fullLinkName,
                              L"\\"
                              );

      RtlAppendUnicodeStringToString(
                                    &fullLinkName,
                                    &Extension->SymbolicLinkName
                                    );


      status = IoCreateSymbolicLink(
                                   &fullLinkName,
                                   &Extension->DeviceName
                                   );
      if (!NT_SUCCESS(status)) {

         //
         // Oh well, couldn't create the symbolic link.  No point
         // in trying to create the device map entry.
         //

         SerialLogError(
                       Extension->DeviceObject->DriverObject,
                       Extension->DeviceObject,
                       Extension->OriginalController,
                       SerialPhysicalZero,
                       0,
                       0,
                       0,
                       52,
                       status,
                       SERIAL_NO_SYMLINK_CREATED,
                       Extension->SymbolicLinkName.Length+sizeof(WCHAR),
                       Extension->SymbolicLinkName.Buffer,
                       0,
                       NULL
                       );
         SerialDump(
                   SERERRORS,
                   ("SERIAL: Couldn't create the symbolic link\n"
                    "------- for port %wZ\n",
                    &Extension->DeviceName)
                   );

      } else {

         Extension->CreatedSymbolicLink = TRUE;

         status = RtlWriteRegistryValue(
                                       RTL_REGISTRY_DEVICEMAP,
                                       L"SERIALCOMM",
                                       Extension->NtNameForPort.Buffer,
                                       REG_SZ,
                                       Extension->SymbolicLinkName.Buffer,
                                       Extension->SymbolicLinkName.Length
				       +sizeof(WCHAR)
                                       );

         if (!NT_SUCCESS(status)) {

            SerialLogError(
                          Extension->DeviceObject->DriverObject,
                          Extension->DeviceObject,
                          Extension->OriginalController,
                          SerialPhysicalZero,
                          0,
                          0,
                          0,
                          53,
                          status,
                          SERIAL_NO_DEVICE_MAP_CREATED,
                          Extension->SymbolicLinkName.Length+sizeof(WCHAR),
                          Extension->SymbolicLinkName.Buffer,
                          0,
                          NULL
                          );
            SerialDump(
                      SERERRORS,
                      ("SERIAL: Couldn't create the device map entry\n"
                       "------- for port %wZ\n",
                       &Extension->DeviceName)
                      );

         }

      }

      ExFreePool(fullLinkName.Buffer);

   }

}

VOID
SerialCleanupExternalNaming(
                           IN PSERIAL_DEVICE_EXTENSION Extension
                           )

/*++

Routine Description:

    This routine will be used to delete a symbolic link
    to the driver name in the given object directory.

    It will also delete an entry in the device map for
    this device if the symbolic link had been created.

Arguments:

    Extension - Pointer to the device extension.

Return Value:

    None.

--*/

{

   UNICODE_STRING fullLinkName;

   PAGED_CODE();

   SerialDump(
             SERDIAG3,
             ("SERIAL: In SerialCleanupExternalNaming for\n"
              "------- extension: %x of port %wZ\n",
              Extension,&Extension->DeviceName)
             );

   //
   // We're cleaning up here.  One reason we're cleaning up
   // is that we couldn't allocate space for the directory
   // name or the symbolic link.
   //

   if (Extension->ObjectDirectory.Buffer &&
       Extension->SymbolicLinkName.Buffer &&
       Extension->CreatedSymbolicLink) {

      //
      // Form the full symbolic link name we wish to create.
      //

      RtlInitUnicodeString(
                          &fullLinkName,
                          NULL
                          );

      //
      // Allocate some pool for the name.
      //

      fullLinkName.MaximumLength = (sizeof(L"\\")*2) +
                                   Extension->ObjectDirectory.Length+
                                   Extension->SymbolicLinkName.Length+
                                   sizeof(WCHAR);

      fullLinkName.Buffer = ExAllocatePool(
                                          PagedPool,
                                          fullLinkName.MaximumLength
                                          );

      if (!fullLinkName.Buffer) {

         //
         // Couldn't allocate space for the name.  Just go on
         // to the device map stuff.
         //

         SerialLogError(
                       Extension->DeviceObject->DriverObject,
                       Extension->DeviceObject,
                       Extension->OriginalController,
                       SerialPhysicalZero,
                       0,
                       0,
                       0,
                       54,
                       STATUS_SUCCESS,
                       SERIAL_INSUFFICIENT_RESOURCES,
                       0,
                       NULL,
                       0,
                       NULL
                       );
         SerialDump(
                   SERERRORS,
                   ("SERIAL: Couldn't allocate space for the symbolic \n"
                    "------- name for creating the link\n"
                    "------- for port %wZ on cleanup\n",
                    &Extension->DeviceName)
                   );

      } else {

         RtlZeroMemory(
                      fullLinkName.Buffer,
                      fullLinkName.MaximumLength
                      );

         RtlAppendUnicodeToString(
                                 &fullLinkName,
                                 L"\\"
                                 );

         RtlAppendUnicodeStringToString(
                                       &fullLinkName,
                                       &Extension->ObjectDirectory
                                       );

         RtlAppendUnicodeToString(
                                 &fullLinkName,
                                 L"\\"
                                 );

         RtlAppendUnicodeStringToString(
                                       &fullLinkName,
                                       &Extension->SymbolicLinkName
                                       );

         IoDeleteSymbolicLink(&fullLinkName);

         ExFreePool(fullLinkName.Buffer);

      }

      //
      // We're cleaning up here.  One reason we're cleaning up
      // is that we couldn't allocate space for the NtNameOfPort.
      //

      if (Extension->NtNameForPort.Buffer) {

         NTSTATUS status;

         status = RtlDeleteRegistryValue(
                                        RTL_REGISTRY_DEVICEMAP,
                                        L"SERIALCOMM",
                                        Extension->NtNameForPort.Buffer
                                        );

         if (!NT_SUCCESS(status)) {

            SerialLogError(
                          Extension->DeviceObject->DriverObject,
                          Extension->DeviceObject,
                          Extension->OriginalController,
                          SerialPhysicalZero,
                          0,
                          0,
                          0,
                          55,
                          status,
                          SERIAL_NO_DEVICE_MAP_DELETED,
                          Extension->SymbolicLinkName.Length+sizeof(WCHAR),
                          Extension->SymbolicLinkName.Buffer,
                          0,
                          NULL
                          );
            SerialDump(
                      SERERRORS,
                      ("SERIAL: Couldn't delete value entry %wZ\n",
                       &Extension->DeviceName)
                      );

         }

      }

   }

}

SERIAL_MEM_COMPARES
SerialMemCompare(
                IN PHYSICAL_ADDRESS A,
                IN ULONG SpanOfA,
                IN PHYSICAL_ADDRESS B,
                IN ULONG SpanOfB
                )

/*++

Routine Description:

    Compare two phsical address.

Arguments:

    A - One half of the comparison.

    SpanOfA - In units of bytes, the span of A.

    B - One half of the comparison.

    SpanOfB - In units of bytes, the span of B.


Return Value:

    The result of the comparison.

--*/

{

   LARGE_INTEGER a;
   LARGE_INTEGER b;

   LARGE_INTEGER lower;
   ULONG lowerSpan;
   LARGE_INTEGER higher;

   PAGED_CODE();

   a = A;
   b = B;

   if (a.QuadPart == b.QuadPart) {

      return AddressesAreEqual;

   }

   if (a.QuadPart > b.QuadPart) {

      higher = a;
      lower = b;
      lowerSpan = SpanOfB;

   } else {

      higher = b;
      lower = a;
      lowerSpan = SpanOfA;

   }

   if ((higher.QuadPart - lower.QuadPart) >= lowerSpan) {

      return AddressesAreDisjoint;

   }

   return AddressesOverlap;

}



BOOLEAN
SerialBecomeSharer(PVOID Context)
{
   PSERIAL_DEVICE_EXTENSION pDevExt = (PSERIAL_DEVICE_EXTENSION)Context;
   PSERIAL_DEVICE_EXTENSION pNewExt
      = (PSERIAL_DEVICE_EXTENSION)pDevExt->NewExtension;

   //
   // Reconfigure the root (just in case this is the first sharer on the
   // interrupt) and ourselves.  The TopLevel fields are always setup in the
   // call to init
   //

   pDevExt->CIsrSw.IsrFunc = SerialSharerIsr;
   pDevExt->CIsrSw.Context = &pDevExt->TopLevelSharers;
   pNewExt->CIsrSw.IsrFunc = SerialSharerIsr;

   //
   // Add to shared interrupt chain if not already on.  We may be
   // if we are a node on a multiport.
   //

   InsertTailList(&pDevExt->TopLevelSharers, &pNewExt->TopLevelSharers);

   if (IsListEmpty(&pNewExt->CommonInterruptObject)) {
      InsertTailList(&pDevExt->CommonInterruptObject,
		     &pNewExt->CommonInterruptObject);
   }

   pNewExt->Interrupt = pDevExt->Interrupt;

   return TRUE;
}



NTSTATUS
SerialFindInitController(IN PDEVICE_OBJECT Fdo, IN PCONFIG_DATA PConfig)

{

   PSERIAL_DEVICE_EXTENSION fdoExtension    = Fdo->DeviceExtension;
   PDEVICE_OBJECT pDeviceObject;
   PSERIAL_DEVICE_EXTENSION pExtension;
   PHYSICAL_ADDRESS serialPhysicalMax;
   SERIAL_LIST_DATA listAddition;
   BOOLEAN didInit = FALSE;
   PLIST_ENTRY currentFdo;
   NTSTATUS status;


   SerialDump(SERTRACECALLS, ("SERIAL: entering SerialFindInitController\n"));

   serialPhysicalMax.LowPart = (ULONG)~0;
   serialPhysicalMax.HighPart = ~0;

   SerialDump(SERDIAG1, ("SERIAL: Attempting to init %wZ\n"
                         "------- PortAddress is %x\n"
                         "------- Interrupt Status is %x\n"
                         "------- BusNumber is %d\n"
                         "------- BusType is %d\n"
                         "------- AddressSpace is %d\n"
                         "------- Interrupt Mode is %d\n",
                         &fdoExtension->DeviceName,
                         PConfig->Controller.LowPart,
                         PConfig->InterruptStatus.LowPart,
                         PConfig->BusNumber,
                         PConfig->InterfaceType,
                         PConfig->AddressSpace,
                         PConfig->InterruptMode)
             );

   //
   // We don't support any boards whose memory wraps around
   // the physical address space.
   //

   if (SerialMemCompare(
                       PConfig->Controller,
                       PConfig->SpanOfController,
                       serialPhysicalMax,
                       (ULONG)0
                       ) != AddressesAreDisjoint) {

      SerialLogError(
                    Fdo->DriverObject,
                    NULL,
                    PConfig->Controller,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    43,
                    STATUS_SUCCESS,
                    SERIAL_DEVICE_TOO_HIGH,
                    fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                    fdoExtension->SymbolicLinkName.Buffer,
                    0,
                    NULL
                    );

      SerialDump(
                SERERRORS,
                ("SERIAL: Error in config record for %wZ\n"
                 "------  registers wrap around physical memory\n",
                 &fdoExtension->DeviceName)
                );

      return STATUS_NO_SUCH_DEVICE;

   }


   if (SerialMemCompare(
                       PConfig->InterruptStatus,
                       PConfig->SpanOfInterruptStatus,
                       serialPhysicalMax,
                       (ULONG)0
                       ) != AddressesAreDisjoint) {

      SerialLogError(
                    Fdo->DriverObject,
                    NULL,
                    PConfig->Controller,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    44,
                    STATUS_SUCCESS,
                    SERIAL_STATUS_TOO_HIGH,
                    fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                    fdoExtension->SymbolicLinkName.Buffer,
                    0,
                    NULL
                    );

      SerialDump(
                SERERRORS,
                ("SERIAL: Error in config record for %wZ\n"
                 "------  status raps around physical memory\n",
                 &fdoExtension->DeviceName)
                );

      return STATUS_NO_SUCH_DEVICE;
   }


   //
   // Make sure that the interrupt status address doesn't
   // overlap the controller registers
   //

   if (SerialMemCompare(
                       PConfig->InterruptStatus,
                       PConfig->SpanOfInterruptStatus,
                       SerialPhysicalZero,
                       (ULONG)0
                       ) != AddressesAreEqual) {

      if (SerialMemCompare(
                          PConfig->InterruptStatus,
                          PConfig->SpanOfInterruptStatus,
                          PConfig->Controller,
                          PConfig->SpanOfController
                          ) != AddressesAreDisjoint) {

         SerialLogError(
                       Fdo->DriverObject,
                       NULL,
                       PConfig->Controller,
                       PConfig->InterruptStatus,
                       0,
                       0,
                       0,
                       45,
                       STATUS_SUCCESS,
                       SERIAL_STATUS_CONTROL_CONFLICT,
                       fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                       fdoExtension->SymbolicLinkName.Buffer,
                       0,
                       NULL
                       );

         SerialDump(
                   SERERRORS,
                   ("SERIAL: Error in cofig record for %wZ\n"
                    "------- Interrupt status overlaps regular registers\n",
                    &fdoExtension->DeviceName)
                   );

         return STATUS_NO_SUCH_DEVICE;
      }
   }


   //
   // Loop through all of the driver's device objects making
   // sure that this new record doesn't overlap with any of them.
   //

   if (!IsListEmpty(&SerialGlobals.AllFdos)) {
      currentFdo = SerialGlobals.AllFdos.Flink;
      pExtension = CONTAINING_RECORD(currentFdo, SERIAL_DEVICE_EXTENSION,
                                     AllFdos);
   } else {
      currentFdo = NULL;
      pExtension = NULL;
   }


   while (currentFdo != NULL
          && currentFdo != &SerialGlobals.AllFdos) {
      //
      // We only care about this list if the elements are on the
      // same bus as this new entry.
      //

      if ((pExtension->InterfaceType  == PConfig->InterfaceType) &&
          (pExtension->AddressSpace   == PConfig->AddressSpace)  &&
          (pExtension->BusNumber      == PConfig->BusNumber)) {

         SerialDump(
                   SERDIAG1,
                   ("SERIAL: Comparing it to %wZ\n"
                    "------- already in the device list\n"
                    "------- PortAddress is %x\n"
                    "------- Interrupt Status is %x\n"
                    "------- BusNumber is %d\n"
                    "------- BusType is %d\n"
                    "------- AddressSpace is %d\n",
                    &pExtension->DeviceName,
                    pExtension->OriginalController.LowPart,
                    pExtension->OriginalInterruptStatus.LowPart,
                    pExtension->BusNumber,
                    pExtension->InterfaceType,
                    pExtension->AddressSpace
                   )
                   );

         //
         // Check to see if the controller addresses are not equal.  
         //

         if (SerialMemCompare(
                             PConfig->Controller,
                             PConfig->SpanOfController,
                             pExtension->OriginalController,
                             pExtension->SpanOfController
                             ) != AddressesAreDisjoint) {

            //
            // We don't want to log an error if the addresses
            // are the same and the name is the same and
            // the new item is from the firmware.
            //

            SerialDump(
                      SERERRORS,
                      ("SERIAL: Error in config record for %wZ\n"
                       "------- Register address overlaps with\n"
                       "------- previous serial device\n",
                       &fdoExtension->DeviceName)
                      );

            return STATUS_NO_SUCH_DEVICE;
         }


         //
         // If we have an interrupt status, make sure that
         // it doesn't overlap with the old controllers
         // registers.
         //

         if (SerialMemCompare(
                             PConfig->InterruptStatus,
                             PConfig->SpanOfInterruptStatus,
                             SerialPhysicalZero,
                             (ULONG)0
                             ) != AddressesAreEqual) {

            //
            // Check it against the existing device's controller address
            //

            if (SerialMemCompare(
                                PConfig->InterruptStatus,
                                PConfig->SpanOfInterruptStatus,
                                pExtension->OriginalController,
                                pExtension->SpanOfController
                                ) != AddressesAreDisjoint) {

               SerialLogError(
                             Fdo->DriverObject,
                             NULL,
                             PConfig->Controller,
                             pExtension->OriginalController,
                             0,
                             0,
                             0,
                             47,
                             STATUS_SUCCESS,
                             SERIAL_STATUS_OVERLAP,
                             fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                             fdoExtension->SymbolicLinkName.Buffer,
                             pExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                             pExtension->SymbolicLinkName.Buffer
                             );

               SerialDump(
                         SERERRORS,
                         ("SERIAL: Error in config record for %wZ\n"
                          "------- status address overlaps with\n"
                          "------- previous serial device registers\n",
                          &fdoExtension->DeviceName)
                         );

               return STATUS_NO_SUCH_DEVICE;
            }

            //
            // If the old configuration record has an interrupt
            // status, the addresses should not overlap.
            //

            if (SerialMemCompare(
                                PConfig->InterruptStatus,
                                PConfig->SpanOfInterruptStatus,
                                SerialPhysicalZero,
                                (ULONG)0
                                ) != AddressesAreEqual) {

               if (SerialMemCompare(
                                   PConfig->InterruptStatus,
                                   PConfig->SpanOfInterruptStatus,
                                   pExtension->OriginalInterruptStatus,
                                   pExtension->SpanOfInterruptStatus
                                   ) == AddressesOverlap) {

                  SerialLogError(
                                Fdo->DriverObject,
                                NULL,
                                PConfig->Controller,
                                pExtension->OriginalController,
                                0,
                                0,
                                0,
                                48,
                                STATUS_SUCCESS,
                                SERIAL_STATUS_STATUS_OVERLAP,
                                fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                                fdoExtension->SymbolicLinkName.Buffer,
                                pExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                                pExtension->SymbolicLinkName.Buffer
                                );

                  SerialDump(
                            SERERRORS,
                            ("SERIAL: Error in config record for %wZ\n"
                             "------- status address overlaps with\n"
                             "------- previous serial status register\n",
                             &fdoExtension->DeviceName)
                            );

                  return STATUS_NO_SUCH_DEVICE;
               }
            }
         }       // if ((pExtension->InterfaceType  == fdoExtension->InterfaceType) &&


         //
         // If the old configuration record has a status
         // address make sure that it doesn't overlap with
         // the new controllers address.  (Interrupt status
         // overlap is take care of above.
         //

         if (SerialMemCompare(
                             pExtension->OriginalInterruptStatus,
                             pExtension->SpanOfInterruptStatus,
                             SerialPhysicalZero,
                             (ULONG)0
                             ) != AddressesAreEqual) {

            if (SerialMemCompare(
                                PConfig->Controller,
                                PConfig->SpanOfController,
                                pExtension->OriginalInterruptStatus,
                                pExtension->SpanOfInterruptStatus
                                ) == AddressesOverlap) {

               SerialLogError(
                             Fdo->DriverObject,
                             NULL,
                             PConfig->Controller,
                             pExtension->OriginalController,
                             0,
                             0,
                             0,
                             49,
                             STATUS_SUCCESS,
                             SERIAL_CONTROL_STATUS_OVERLAP,
                             fdoExtension->SymbolicLinkName.Length
			     + sizeof(WCHAR),
                             fdoExtension->SymbolicLinkName.Buffer,
                             pExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                             pExtension->SymbolicLinkName.Buffer
                             );

               SerialDump(
                         SERERRORS,
                         ("SERIAL: Error in config record for %wZ\n"
                          "------- register address overlaps with\n"
                          "------- previous serial status register\n",
                          &fdoExtension->DeviceName)
                         );

               return STATUS_NO_SUCH_DEVICE;
            }
         }
      }

      currentFdo = currentFdo->Flink;

      if (currentFdo != NULL) {
         pExtension = CONTAINING_RECORD(currentFdo, SERIAL_DEVICE_EXTENSION,
                                        AllFdos);
      }
   }   // while (currentFdo != NULL)



   //
   // Now, we will check if this is a port on a multiport card.
   // The conditions are same ISR set and same IRQL/Vector
   //

   //
   // Loop through all previously attached devices
   //

   if (!IsListEmpty(&SerialGlobals.AllFdos)) {
      currentFdo = SerialGlobals.AllFdos.Flink;
      pExtension = CONTAINING_RECORD(currentFdo, SERIAL_DEVICE_EXTENSION,
                                     AllFdos);
   } else {
      currentFdo = NULL;
      pExtension = NULL;
   }


   //
   // If there is an interrupt status then we
   // loop through the config list again to look
   // for a config record with the same interrupt
   // status (on the same bus).
   //

   if ((SerialMemCompare(
                        PConfig->InterruptStatus,
                        PConfig->SpanOfInterruptStatus,
                        SerialPhysicalZero,
                        (ULONG)0
                        ) != AddressesAreEqual) &&
       (currentFdo != NULL)) {

      ASSERT(pExtension != NULL);

      //
      // We have an interrupt status.  Loop through all
      // previous records, look for an existing interrupt status
      // the same as the current interrupt status.
      //
      do {

         //
         // We only care about this list if the elements are on the
         // same bus as this new entry.  (Their interrupts must therefore
         // also be the on the same bus.  We will check that momentarily).
         //
         // We don't check here for the dissimilar interrupts since that
         // could cause us to miss the error of having the same interrupt
         // status but different interrupts - which is bizzare.
         //

         if ((pExtension->InterfaceType == PConfig->InterfaceType) &&
             (pExtension->AddressSpace == PConfig->AddressSpace) &&
             (pExtension->BusNumber == PConfig->BusNumber)) {

            //
            // If the interrupt status is the same, then same card.
            //

            if (SerialMemCompare(
                                pExtension->OriginalInterruptStatus,
                                pExtension->SpanOfInterruptStatus,
                                PConfig->InterruptStatus,
                                PConfig->SpanOfInterruptStatus
                                ) == AddressesAreEqual) {

               //
               // Same card.  Now make sure that they
               // are using the same interrupt parameters.
               //

               if ((PConfig->OriginalIrql != pExtension->OriginalIrql) ||
                   (PConfig->OriginalVector != pExtension->OriginalVector)) {

                  //
                  // We won't put this into the configuration
                  // list.
                  //
#if 0
                  SerialLogError(
                                DriverObject,
                                NULL,
                                PConfig->Controller,
                                pExtension->Controller,
                                0,
                                0,
                                0,
                                50,
                                STATUS_SUCCESS,
                                SERIAL_MULTI_INTERRUPT_CONFLICT,
                                PConfig->SymbolicLinkName.Length+sizeof(WCHAR),
                                PConfig>SymbolicLinkName.Buffer,
                                pExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                                pExtension->SymbolicLinkName.Buffer
                                );
#endif

                  SerialDump(
                            SERERRORS,
                            ("SERIAL: Configuration error for %wZ\n"
                             "------- Same multiport - different interrupts\n",
                             &PConfig->NtNameForPort)
                            );
                  return STATUS_NO_SUCH_DEVICE;

               }

               //
               // Initialize the device as part of a multiport board
               //

               status = SerialInitMultiPort(pExtension, PConfig, Fdo);

               //
               // A port can be one of three things:
               //    A standalone
               //    A non-root on a multiport
               //    A root on a multiport
               //
               // It can only share an interrupt if it is a root
               // or if it is a standalone.  Since this was a non-root
               // we don't need to check if it shares an interrupt
               // and we can return.
               //
               return status;
            }

            //
            // No match, check some more
            //

            currentFdo = currentFdo->Flink;
            if (currentFdo != NULL) {
               pExtension = CONTAINING_RECORD(currentFdo,
                                              SERIAL_DEVICE_EXTENSION, AllFdos);
            }
         }

      } while (currentFdo != NULL
               && currentFdo != &SerialGlobals.AllFdos);
   }

   status = SerialInitOneController(Fdo, PConfig);

   if (!NT_SUCCESS(status)) {
      return status;
   }

   //
   // The device is initialized.  Now we need to check if 
   // this device shares an interrupt with anyone.
   //


   //
   // Loop through all previously attached devices
   //

   if (!IsListEmpty(&SerialGlobals.AllFdos)) {
      currentFdo = SerialGlobals.AllFdos.Flink;
      pExtension = CONTAINING_RECORD(currentFdo, SERIAL_DEVICE_EXTENSION,
                                     AllFdos);
   } else {
      currentFdo = NULL;
      pExtension = NULL;
   }

   //
   // Go through the list again looking for previous devices
   // with the same interrupt.  The first one found will either be a root
   // or standalone.  Order of insertion is important here!
   //

   if (!PConfig->Jensen && (currentFdo != NULL)) {
      do {

         //
         // We only care about interrupts that are on
         // the same bus.
         //

         if ((pExtension->InterfaceType == PConfig->InterfaceType) &&
             (pExtension->BusNumber == PConfig->BusNumber)) {

            if ((pExtension->OriginalIrql == PConfig->OriginalIrql) &&
                (pExtension->OriginalVector == PConfig->OriginalVector)) {
                pExtension->NewExtension = fdoExtension;
                KeSynchronizeExecution(pExtension->Interrupt, SerialBecomeSharer,
				      pExtension);

               return STATUS_SUCCESS;

            }

         }

         //
         // No match, check some more
         //

         currentFdo = currentFdo->Flink;

         if (currentFdo != NULL) {
            pExtension = CONTAINING_RECORD(currentFdo, SERIAL_DEVICE_EXTENSION,
                                           AllFdos);
         }
      } while (currentFdo != NULL
               && currentFdo != &SerialGlobals.AllFdos);
   }

   return STATUS_SUCCESS;
}


VOID
SerialLogError(
              IN PDRIVER_OBJECT DriverObject,
              IN PDEVICE_OBJECT DeviceObject OPTIONAL,
              IN PHYSICAL_ADDRESS P1,
              IN PHYSICAL_ADDRESS P2,
              IN ULONG SequenceNumber,
              IN UCHAR MajorFunctionCode,
              IN UCHAR RetryCount,
              IN ULONG UniqueErrorValue,
              IN NTSTATUS FinalStatus,
              IN NTSTATUS SpecificIOStatus,
              IN ULONG LengthOfInsert1,
              IN PWCHAR Insert1,
              IN ULONG LengthOfInsert2,
              IN PWCHAR Insert2
              )

/*++

Routine Description:

    This routine allocates an error log entry, copies the supplied data
    to it, and requests that it be written to the error log file.

Arguments:

    DriverObject - A pointer to the driver object for the device.

    DeviceObject - A pointer to the device object associated with the
    device that had the error, early in initialization, one may not
    yet exist.

    P1,P2 - If phyical addresses for the controller ports involved
    with the error are available, put them through as dump data.

    SequenceNumber - A ulong value that is unique to an IRP over the
    life of the irp in this driver - 0 generally means an error not
    associated with an irp.

    MajorFunctionCode - If there is an error associated with the irp,
    this is the major function code of that irp.

    RetryCount - The number of times a particular operation has been
    retried.

    UniqueErrorValue - A unique long word that identifies the particular
    call to this function.

    FinalStatus - The final status given to the irp that was associated
    with this error.  If this log entry is being made during one of
    the retries this value will be STATUS_SUCCESS.

    SpecificIOStatus - The IO status for a particular error.

    LengthOfInsert1 - The length in bytes (including the terminating NULL)
                      of the first insertion string.

    Insert1 - The first insertion string.

    LengthOfInsert2 - The length in bytes (including the terminating NULL)
                      of the second insertion string.  NOTE, there must
                      be a first insertion string for their to be
                      a second insertion string.

    Insert2 - The second insertion string.

Return Value:

    None.

--*/

{
   PIO_ERROR_LOG_PACKET errorLogEntry;

   PVOID objectToUse;
   SHORT dumpToAllocate = 0;
   PUCHAR ptrToFirstInsert;
   PUCHAR ptrToSecondInsert;

   PAGED_CODE();


   if (ARGUMENT_PRESENT(DeviceObject)) {

      objectToUse = DeviceObject;

   } else {

      objectToUse = DriverObject;

   }

   if (SerialMemCompare(
                       P1,
                       (ULONG)1,
                       SerialPhysicalZero,
                       (ULONG)1
                       ) != AddressesAreEqual) {

      dumpToAllocate = (SHORT)sizeof(PHYSICAL_ADDRESS);

   }

   if (SerialMemCompare(
                       P2,
                       (ULONG)1,
                       SerialPhysicalZero,
                       (ULONG)1
                       ) != AddressesAreEqual) {

      dumpToAllocate += (SHORT)sizeof(PHYSICAL_ADDRESS);

   }

   errorLogEntry = IoAllocateErrorLogEntry(
                                          objectToUse,
                                          (UCHAR)(sizeof(IO_ERROR_LOG_PACKET) +
                                                  dumpToAllocate
						  + LengthOfInsert1 +
                                                  LengthOfInsert2)
                                          );

   if ( errorLogEntry != NULL ) {

      errorLogEntry->ErrorCode = SpecificIOStatus;
      errorLogEntry->SequenceNumber = SequenceNumber;
      errorLogEntry->MajorFunctionCode = MajorFunctionCode;
      errorLogEntry->RetryCount = RetryCount;
      errorLogEntry->UniqueErrorValue = UniqueErrorValue;
      errorLogEntry->FinalStatus = FinalStatus;
      errorLogEntry->DumpDataSize = dumpToAllocate;

      if (dumpToAllocate) {

         RtlCopyMemory(
                      &errorLogEntry->DumpData[0],
                      &P1,
                      sizeof(PHYSICAL_ADDRESS)
                      );

         if (dumpToAllocate > sizeof(PHYSICAL_ADDRESS)) {

            RtlCopyMemory(
                         ((PUCHAR)&errorLogEntry->DumpData[0])
			 +sizeof(PHYSICAL_ADDRESS),
                         &P2,
                         sizeof(PHYSICAL_ADDRESS)
                         );

            ptrToFirstInsert =
            ((PUCHAR)&errorLogEntry->DumpData[0])+(2*sizeof(PHYSICAL_ADDRESS));

         } else {

            ptrToFirstInsert =
            ((PUCHAR)&errorLogEntry->DumpData[0])+sizeof(PHYSICAL_ADDRESS);


         }

      } else {

         ptrToFirstInsert = (PUCHAR)&errorLogEntry->DumpData[0];

      }

      ptrToSecondInsert = ptrToFirstInsert + LengthOfInsert1;

      if (LengthOfInsert1) {

         errorLogEntry->NumberOfStrings = 1;
         errorLogEntry->StringOffset = (USHORT)(ptrToFirstInsert -
                                                (PUCHAR)errorLogEntry);
         RtlCopyMemory(
                      ptrToFirstInsert,
                      Insert1,
                      LengthOfInsert1
                      );

         if (LengthOfInsert2) {

            errorLogEntry->NumberOfStrings = 2;
            RtlCopyMemory(
                         ptrToSecondInsert,
                         Insert2,
                         LengthOfInsert2
                         );

         }

      }

      IoWriteErrorLogEntry(errorLogEntry);

   }

}
