/*++

Copyright (c) 1991, 1992, 1993 - 1997 Microsoft Corporation

Module Name:

    PnpPower.c

Abstract:

    This module contains the code that handles the plug and play and power
   IRPs for the serial driver.

Author:

    Scott Brenden 16-Jan-1997

Environment:

    Kernel mode

Revision History :
   Louis J. Giliberto, Jr.   19-May-1997

--*/

#include "precomp.h"

#define ALLF    0xffffffff

static const PHYSICAL_ADDRESS SerialPhysicalZero = {0};



#ifdef ALLOC_PRAGMA
#pragma alloc_text(PAGESER, SerialPnpDispatch)
#pragma alloc_text(PAGESER, SerialStartDevice)
#pragma alloc_text(PAGESER, SerialFinishStartDevice)
#pragma alloc_text(PAGESER, SerialGetPortInfo)
#pragma alloc_text(PAGESER, SerialDoExternalNaming)
#pragma alloc_text(PAGESER, SerialUndoExternalNaming)
#endif // ALLOC_PRAGMA

NTSTATUS
SerialSyncCompletion(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp,
		     IN PKEVENT SerialSyncEvent)
{
   KeSetEvent(SerialSyncEvent, IO_NO_INCREMENT, FALSE);
   return STATUS_MORE_PROCESSING_REQUIRED;
}




NTSTATUS
SerialPnpDispatch(
                    IN PDEVICE_OBJECT Fdo,
                    IN PIRP Irp
                    )

/*++

Routine Description:

    This is a dispatch routine for the IRPs that come to the driver with the
    IRP_MJ_PNP major code (plug-and-play IRPs).

Arguments:

    DeviceObject - Pointer to the device object for this device

    Irp - Pointer to the IRP for the current request

Return Value:

    The function value is the final status of the call


--*/

{

   PSERIAL_DEVICE_EXTENSION    fdoExtension    = Fdo->DeviceExtension;
   PDEVICE_OBJECT pdo = fdoExtension->LowerDeviceObject;
   PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(Irp);
   NTSTATUS                    status;
   BOOLEAN acceptingIRPs;

   PAGED_CODE();

   acceptingIRPs = SerialIRPPrologue(fdoExtension);

   if ((irpStack->MinorFunction != IRP_MN_REMOVE_DEVICE)
       && (irpStack->MinorFunction != IRP_MN_CANCEL_REMOVE_DEVICE)
       && (acceptingIRPs == FALSE)) {
      Irp->IoStatus.Information = 0;
      Irp->IoStatus.Status = STATUS_NO_SUCH_DEVICE;
      SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
      return STATUS_NO_SUCH_DEVICE;
   }

   switch (irpStack->MinorFunction) {

   case IRP_MN_QUERY_DEVICE_RELATIONS: {

         //
         // We don't do anything unless we have ejected PDO's.
         // This will only happen after serenum is reality.
         // For now, just pass the IRP down.
         //
         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_QUERY_DEVICE_RELATIONS "
                                   "Irp\n"));

         switch (irpStack->Parameters.QueryDeviceRelations.Type) {
         case BusRelations:
            SerialDump(SERPNPPOWER, ("------- BusRelations Query\n"));
            break;

         case EjectionRelations:
            SerialDump(SERPNPPOWER, ("------- EjectionRelations Query\n"));
            break;

         case PowerRelations:
            SerialDump(SERPNPPOWER, ("------- PowerRelations Query\n"));
            break;

         case RemovalRelations:
            SerialDump(SERPNPPOWER, ("------- RemovalRelations Query\n"));
            break;

         case TargetDeviceRelation:
            SerialDump(SERPNPPOWER, ("------- TargetDeviceRelation Query\n"));
            break;

         default:
            SerialDump(SERPNPPOWER, ("------- Unknown Query\n"));
            break;
         }

         IoSkipCurrentIrpStackLocation(Irp);
	 status = SerialIoCallDriver(fdoExtension, pdo, Irp);
	 return status;
      }



   case IRP_MN_QUERY_INTERFACE: {

         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_QUERY_INTERFACE Irp\n"));
         // 
         // bugbug:  Not sure that this one has to be supported.  Brian may need
         //          it for the ISA pnp modem.
         //
         break;
      }



   case IRP_MN_QUERY_CAPABILITIES: {

         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_QUERY_CAPABILITIES Irp\n"));
         // 
         // bugbug:  If legacy com ports have been reported in QUERY_DEVICE_RELATIONS
         //          one of these will come for each of those devices.
         //
         break;
      }



   case IRP_MN_QUERY_RESOURCE_REQUIREMENTS: {

         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_QUERY_RESOURCE_REQUIREMENTS Irp\n"));
         // 
         // bugbug:  If legacy com ports have been reported in QUERY_DEVICE_RELATIONS
         //          one of these will come for each of those devices.
         //
         break;
      }



   case IRP_MN_START_DEVICE: {

         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_START_DEVICE Irp\n"));

	 //
	 // SerialStartDevice will pass this Irp to the next driver,
	 // and process it as completion so just complete it here.
	 //

         status = SerialStartDevice (Fdo, Irp);

	 Irp->IoStatus.Status = status;

   	 SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
	 return status;
      }



   case IRP_MN_READ_CONFIG: {

         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_READ_CONFIG Irp\n"));
         // 
         // bugbug:  Not sure that this one has to be supported.  Brian may need
         //          it for the ISA pnp modem.
         //

         break;
      }



   case IRP_MN_WRITE_CONFIG: {

         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_WRITE_CONFIG Irp\n"));
         // 
         // bugbug:  Not sure that this one has to be supported.  
         //
         break;
      }



   case IRP_MN_EJECT: {

         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_EJECT Irp\n"));
         // 
         // bugbug:  Not sure that this one has to be supported.  
         //
         break;
      }



   case IRP_MN_SET_LOCK: {

         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_SET_LOCK Irp\n"));
         // 
         // bugbug:  Not sure that this one has to be supported.  
         //
         break;
      }



   case IRP_MN_QUERY_ID: {
   	 UNICODE_STRING pIdBuf;
	 PWCHAR pPnpIdStr;
	 ULONG pnpIdStrLen;
	 ULONG isMulti = 0;
	 HANDLE pnpKey;
	 
         SerialDump(SERPNPPOWER, ("SERIAL: Got IRP_MN_QUERY_ID Irp\n"));
	 SerialDump(SERPNPPOWER, ("------- Id'ing as Serial ID of %s\n",
				  SERIAL_PNP_ID_STR));
#if 0
	 status = IoOpenDeviceRegistryKey(pdo, PLUGPLAY_REGKEY_DEVICE,
					  STANDARD_RIGHTS_WRITE, &pnpKey);

	 if (!NT_SUCCESS(status)) {
	    Irp->IoStatus.Status = status;

	    SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
	    return status;

	 }

         status = SerialGetRegistryKeyValue (pnpKey, L"MultiportDevice",
                                             sizeof(L"MultiportDevice"),
                                             &isMulti,
                                             sizeof (ULONG));

   	 ZwClose(pnpKey);

	 if (!NT_SUCCESS(status)) {
            Irp->IoStatus.Status = status;
	    SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
	    return status;
	 }

	 pPnpIdStr = isMulti ? SERIAL_PNP_MULTI_ID_STR : SERIAL_PNP_ID_STR;
	 pnpIdStrLen = isMulti ? sizeof(SERIAL_PNP_MULTI_ID_STR)
   	    : sizeof(SERIAL_PNP_ID_STR);
#else
	 pPnpIdStr = SERIAL_PNP_ID_STR;
	 pnpIdStrLen = sizeof(SERIAL_PNP_ID_STR);
#endif

	 if (Irp->IoStatus.Information != 0) {
	    ULONG curStrLen;
	    ULONG allocLen = 0;
            PWSTR curStr = (PWSTR)Irp->IoStatus.Information;

	    //
	    // We have to walk the strings to count the amount of space to
	    // reallocate
	    //

	    while ((curStrLen = wcslen(curStr)) != 0) {
	       allocLen += curStrLen * sizeof(WCHAR) + sizeof(UNICODE_NULL);
	       curStr += curStrLen + 1;
	    }

	    allocLen += sizeof(UNICODE_NULL);

	    pIdBuf.Buffer = ExAllocatePool(PagedPool, allocLen
					   + pnpIdStrLen
					   + sizeof(WCHAR));

	    if (pIdBuf.Buffer == NULL) {
	       //
	       // Clean up after other drivers since we are
	       // sending the irp back up.
	       //

	       ExFreePool((PWSTR)Irp->IoStatus.Information);


               Irp->IoStatus.Status = STATUS_NO_MEMORY;
	       Irp->IoStatus.Information = 0;
	       SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
	       return STATUS_NO_MEMORY;
	    }

	    pIdBuf.MaximumLength = (USHORT)(allocLen + pnpIdStrLen);
	    pIdBuf.Length = (USHORT)allocLen - sizeof(UNICODE_NULL);

            RtlZeroMemory(pIdBuf.Buffer, pIdBuf.MaximumLength + sizeof(WCHAR));
	    RtlCopyMemory(pIdBuf.Buffer, (PWSTR)Irp->IoStatus.Information,
			  allocLen);
            RtlAppendUnicodeToString(&pIdBuf, pPnpIdStr);

	    //
	    // Free what the previous driver allocated
	    //

	    ExFreePool((PWSTR)Irp->IoStatus.Information);


	 } else {

	    SerialDump(SERPNPPOWER, ("SERIAL: ID is sole ID\n"));

	    pIdBuf.Buffer = ExAllocatePool(PagedPool, pnpIdStrLen
					   + sizeof(WCHAR) * 2);

	    if (pIdBuf.Buffer == NULL) {
	       Irp->IoStatus.Status = STATUS_NO_MEMORY;
	       Irp->IoStatus.Information = 0;
	       SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
	       return STATUS_NO_MEMORY;
	    }

	    pIdBuf.MaximumLength  = (USHORT)pnpIdStrLen;
	    pIdBuf.Length = 0;

	    RtlZeroMemory(pIdBuf.Buffer, pIdBuf.MaximumLength + sizeof(WCHAR)
			  * 2);

	    RtlAppendUnicodeToString(&pIdBuf, pPnpIdStr);
	 }

	 Irp->IoStatus.Information = (ULONG)pIdBuf.Buffer;
	 Irp->IoStatus.Status = STATUS_SUCCESS;

         IoCopyCurrentIrpStackLocationToNext(Irp);
	 return SerialIoCallDriver(fdoExtension, pdo, Irp);
      }

   case IRP_MN_FILTER_RESOURCE_REQUIREMENTS: {
         HANDLE pnpKey;
         KEVENT resFiltEvent;
         ULONG isMulti = 0;
         PIO_RESOURCE_REQUIREMENTS_LIST pReqList;
         PIO_RESOURCE_LIST pResList;
         PIO_RESOURCE_DESCRIPTOR pResDesc;
         ULONG i, j;
         ULONG reqCnt;

         SerialDump(SERPNPPOWER, ("SERIAL: Got "
                                  "IRP_MN_FILTER_RESOURCE_REQUIREMENTS Irp\n"));
         SerialDump(SERPNPPOWER, ("------- for device %x\n", pdo));

         

         KeInitializeEvent(&resFiltEvent, SynchronizationEvent, FALSE);

         IoCopyCurrentIrpStackLocationToNext(Irp);
         IoSetCompletionRoutine(Irp, SerialSyncCompletion, &resFiltEvent,
                                TRUE, TRUE, TRUE);

         status = IoCallDriver(pdo, Irp);


         //
         // Wait for lower drivers to be done with the Irp
         //

         if (status == STATUS_PENDING) {
            KeWaitForSingleObject (&resFiltEvent, Executive, KernelMode, FALSE,
                                   NULL);
         }

         if (Irp->IoStatus.Information == 0) {
            if (irpStack->Parameters.FilterResourceRequirements
                .IoResourceRequirementList == 0) {
               SerialDump(SERPNPPOWER, ("------- Can't filter NULL resources!"
                                        "\n"));
               status = Irp->IoStatus.Status;
               SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
               return status;
            }

            Irp->IoStatus.Information = (ULONG)irpStack->Parameters
                                        .FilterResourceRequirements
                                        .IoResourceRequirementList;

         }

         status = IoOpenDeviceRegistryKey(pdo, PLUGPLAY_REGKEY_DEVICE,
                                          STANDARD_RIGHTS_WRITE, &pnpKey);

         if (!NT_SUCCESS(status)) {
            Irp->IoStatus.Status = status;

            SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
            return status;

         }

         //
         // No matter what we add our filter if we can and return success.
         //

         status = SerialGetRegistryKeyValue (pnpKey, L"MultiportDevice",
                                             sizeof(L"MultiportDevice"),
                                             &isMulti,
                                             sizeof (ULONG));

         ZwClose(pnpKey);

         if (!NT_SUCCESS(status)) {
            Irp->IoStatus.Status = status;
            SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
            return status;
         }

         if (!isMulti) {
            status = Irp->IoStatus.Status;
            SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
            return status;
         }

         SerialDump(SERPNPPOWER, ("------- found multiport device for pdo"
                                  " %x and reslist at %x \n", pdo,
                                  Irp->IoStatus.Information));


         //
         // Force ISR ports in IO_RES_REQ_LIST to shared status
         //

         pReqList = (PIO_RESOURCE_REQUIREMENTS_LIST)Irp->IoStatus.Information;
         reqCnt = ((pReqList->ListSize - sizeof(IO_RESOURCE_REQUIREMENTS_LIST))
                   / sizeof(IO_RESOURCE_DESCRIPTOR)) + 1;

         for (i = 0; i < reqCnt; i++) {
            pResList = &pReqList->List[i];

            for (j = 0; j < pResList->Count; j++) {
               pResDesc = &pResList->Descriptors[j];
               if (pResDesc->Type == CmResourceTypePort
                   && pResDesc->u.Port.Length == SERIAL_STATUS_LENGTH) {
                  pResDesc->ShareDisposition   =   CmResourceShareShared;
                  SerialDump(SERPNPPOWER, ("------- Sharing I/O port for "
                                           "device %x\n", pdo));
                  break;
               }
            }
         }

         Irp->IoStatus.Status = STATUS_SUCCESS;
         SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
         return STATUS_SUCCESS;
      }

   case IRP_MN_STOP_DEVICE:
      SerialDump(SERPNPPOWER, ("SERIAL: Got IRP_MN_STOP_DEVICE Irp\n"));
      SerialDump(SERPNPPOWER, ("------- for device %x\n", pdo));

      Irp->IoStatus.Status = STATUS_NOT_SUPPORTED;
      SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
      return STATUS_NOT_SUPPORTED;

   case IRP_MN_QUERY_STOP_DEVICE:
      SerialDump(SERPNPPOWER, ("SERIAL: Got IRP_MN_QUERY_STOP_DEVICE Irp\n"));
      SerialDump(SERPNPPOWER, ("------- for device %x\n", pdo));

      Irp->IoStatus.Status = STATUS_NOT_SUPPORTED;
      SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
      return STATUS_NOT_SUPPORTED;

   case IRP_MN_CANCEL_REMOVE_DEVICE:
      ASSERT(fdoExtension->PNPState == SERIAL_PNP_QREMOVE);

      //
      // Restore the device state
      //

      fdoExtension->PNPState = SERIAL_PNP_STARTED;
      fdoExtension->DevicePNPAccept &= ~SERIAL_PNPACCEPT_REMOVING;

      Irp->IoStatus.Status = STATUS_SUCCESS;
      return SerialIoCallDriver(fdoExtension, pdo, Irp);

   case IRP_MN_QUERY_REMOVE_DEVICE:
      SerialDump(SERPNPPOWER, ("SERIAL: Got "
			       "IRP_MN_QUERY_REMOVE_DEVICE Irp\n"));
      SerialDump(SERPNPPOWER, ("------- for device %x\n", Fdo));

      //
      // See if we should succeed a remove query
      //

      if (fdoExtension->DeviceIsOpened) {
	 Irp->IoStatus.Status = STATUS_DEVICE_BUSY;
	 SerialDump(SERPNPPOWER, ("------- failing; device open\n"));
	 SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
	 return STATUS_DEVICE_BUSY;
      }

      if (fdoExtension->PortOnAMultiportCard) {
	 Irp->IoStatus.Status = STATUS_NOT_SUPPORTED;
	 SerialDump(SERPNPPOWER, ("------- failing; multiport node\n"));
	 SerialCompleteRequest(fdoExtension, Irp, IO_NO_INCREMENT);
	 return STATUS_NOT_SUPPORTED;
      }

 
      fdoExtension->PNPState = SERIAL_PNP_QREMOVE;
      fdoExtension->DevicePNPAccept |= SERIAL_PNPACCEPT_REMOVING;

      Irp->IoStatus.Status = STATUS_SUCCESS;
      return SerialIoCallDriver(fdoExtension, pdo, Irp);

   case IRP_MN_REMOVE_DEVICE:
      {
	 ULONG pendingIRPs;

	 SerialDump(SERPNPPOWER, ("SERIAL: Got IRP_MN_REMOVE_DEVICE Irp\n"));
	 SerialDump(SERPNPPOWER, ("------- for device %x\n", Fdo));

	 //
	 // If we get this, we have to remove
	 //
      
	 //
	 // Mark as not accepting requests
	 //

	 fdoExtension->DevicePNPAccept |= SERIAL_PNPACCEPT_REMOVING;

	 //
	 // Complete all pending requests
	 //
	 SerialKillPendingIrps(Fdo);


	 //
	 // Pass the irp down
	 //
	 
	 Irp->IoStatus.Status = STATUS_SUCCESS;

	 IoCopyCurrentIrpStackLocationToNext(Irp);

	 //
	 // We do decrement here because we incremented on entry here.
	 //
	 SerialIoCallDriver(fdoExtension, pdo, Irp);

	 //
	 // Wait for any pending requests we raced on.
	 //

	 pendingIRPs = InterlockedDecrement(&fdoExtension->PendingIRPCnt);

	 if (pendingIRPs) {
	    KeWaitForSingleObject(&fdoExtension->PendingIRPEvent, Executive,
				  KernelMode, FALSE, NULL);
	 }

	 //
	 // Remove us
	 //

	 SerialRemoveFdo(Fdo);

	 return STATUS_SUCCESS;
      }

   default:
      break;



   }   // switch (irpStack->MinorFunction)

   //
   // Pass to driver beneath us
   //

   IoSkipCurrentIrpStackLocation(Irp);
   status = SerialIoCallDriver(fdoExtension, pdo, Irp);
   return status;
}





NTSTATUS
SerialPowerDispatch(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)

/*++

Routine Description:

    This is a dispatch routine for the IRPs that come to the driver with the
    IRP_MJ_POWER major code (power IRPs).

Arguments:

    DeviceObject - Pointer to the device object for this device

    Irp - Pointer to the IRP for the current request

Return Value:

    The function value is the final status of the call


--*/

{

   PSERIAL_DEVICE_EXTENSION pdoExtension = DeviceObject->DeviceExtension;
   PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(Irp);
   NTSTATUS status = STATUS_NOT_SUPPORTED;
   BOOLEAN complete = FALSE;
   BOOLEAN delayNext = FALSE;
   PDEVICE_OBJECT pdo = pdoExtension->LowerDeviceObject;
   BOOLEAN acceptingIRPs;

   //   PAGED_CODE();


   ASSERT(pdo != NULL);

   acceptingIRPs = SerialIRPPrologue(pdoExtension);

   if (acceptingIRPs == FALSE) {
      Irp->IoStatus.Information = 0;
      Irp->IoStatus.Status = STATUS_NO_SUCH_DEVICE;
      SerialCompleteRequest(pdoExtension, Irp, IO_NO_INCREMENT);
      return STATUS_NO_SUCH_DEVICE;
   }

   switch (irpStack->MinorFunction) {


   case IRP_MN_WAIT_WAKE: {

         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_WAIT_WAKE Irp\n"));
         //
         // bugbug:  This will have to be done.  Here is where support for a
         //          serial device (like a modem) waking the system when the
         //          phone rings.
         //
         break;
      }



   case IRP_MN_POWER_SEQUENCE: {

         SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_POWER_SEQUENCE Irp\n"));
         //
         // bugbug:  This still needs to be done.  Look at \nt\public\spec\busdd.doc
         //          for an explanation.
         //          
         break;
      }



   case IRP_MN_SET_POWER:
      SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_SET_POWER Irp\n"));

      if (irpStack->Parameters.Power.Type != DevicePowerState) {
         //
         // They asked for a system power state change which we can't do.
	 // Pass it down to the lower driver.
         //

         status      = STATUS_SUCCESS;
         break;
      }

      //
      // If we are already in the requested state, just pass the IRP down
      //

      if (pdoExtension->PowerState
	  == irpStack->Parameters.Power.State.DeviceState) {
         status      = STATUS_SUCCESS;
         break;
      }


      switch (irpStack->Parameters.Power.State.DeviceState) {

      case PowerDeviceD0: 
         status = SerialSetPowerD0 (DeviceObject, Irp);
	 delayNext = TRUE;
         break;


      case PowerDeviceD1:  
      case PowerDeviceD2: 
      case PowerDeviceD3:  
         return SerialSetPowerD3(DeviceObject, Irp);
      }   // switch (IrpSp->Parameters.Power.State.DeviceState)

      break;



   case IRP_MN_QUERY_POWER:

#if 0
      complete = TRUE;

      SerialDump (SERPNPPOWER, ("SERIAL: Got IRP_MN_QUERY_POWER Irp\n"));

      if (irpStack->Parameters.Power.Type == DevicePowerState) {
         switch (irpStack->Parameters.Power.State.DeviceState) {

         case PowerDeviceD0: 
            //
            // We can always do this one
            //

            status      = STATUS_SUCCESS;
            break;


         case PowerDeviceD1:  
         case PowerDeviceD2: 
         case PowerDeviceD3:  
            //
            // If the device is not open, then we can do this.
            //

            if (!pdoExtension->DeviceIsOpened) {
               pdoExtension->ReceivedQueryD3 = TRUE;
               status = STATUS_SUCCESS;
            } else {
               status = STATUS_DEVICE_BUSY;
            }
            break;

         }   // switch (IrpSp->Parameters.Power.State.DeviceState)
      } else {
	 //
	 // Pass down the stack
	 //

         status = STATUS_SUCCESS;
	 break;
      }   // if (irpStack->Parameters.Power.Type == DevicePowerState)

#else
      Irp->IoStatus.Status = STATUS_SUCCESS;
      status = STATUS_SUCCESS;
      IoCopyCurrentIrpStackLocationToNext(Irp);
      PoStartNextPowerIrp(Irp);
      return SerialPoCallDriver(pdoExtension, pdo, Irp);

#endif      
   }   // switch (irpStack->MinorFunction)


   if (delayNext == FALSE) {
      PoStartNextPowerIrp(Irp);
   }


   if (complete) {
      // 
      // We need to complete the Irp
      //
      SerialCompleteRequest(pdoExtension, Irp, IO_NO_INCREMENT);
   } else {
      //
      // Pass to the lower driver
      //
      IoSkipCurrentIrpStackLocation(Irp);
      status = SerialPoCallDriver(pdoExtension, pdo, Irp);
   }

   return status;
}





NTSTATUS
SerialSetPowerD0(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)

/*++

Routine Description:

    This routine Decides if we need to pass the power Irp down the stack
    or not.  It then either sets up a completion handler to finish the 
    initialization or calls the completion handler directly.

Arguments:

    DeviceObject - Pointer to the device object for this device

    Irp - Pointer to the IRP for the current request

Return Value:

    Return status of either PoCallDriver of the call to the initialization
    routine.


--*/

{
   PSERIAL_DEVICE_EXTENSION    extension   = DeviceObject->DeviceExtension;
   PIO_STACK_LOCATION          irpStack    = IoGetCurrentIrpStackLocation(Irp);
   NTSTATUS status;

   SerialDump(SERTRACECALLS, ("SERIAL: In SerialSetPowerD0\n"));

   ASSERT(extension->LowerDeviceObject);

   //
   // Set up completion to init device when it is on
   //

   IoCopyCurrentIrpStackLocationToNext(Irp);
   IoSetCompletionRoutine(Irp, SerialD0Complete, NULL, TRUE, TRUE, TRUE);

   status = PoCallDriver(extension->LowerDeviceObject, Irp);

   SerialDump(SERTRACECALLS, ("SERIAL: In SerialSetPowerD0\n"));
   return status;
}



NTSTATUS
SerialD0Complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)

/*++

Routine Description:

    This routine is a completion handler that is called after the COM port
    has been powered up.  It sets the COM port in a known state by calling
    SerialReset, SerialMarkClose, SerialClrRTS, and SerialClrDTR.  The it
    does a PoCompleteRequest to finish off the power Irp.

Arguments:

    DeviceObject - Pointer to the device object for this device

    Irp - Pointer to the IRP for the current request

    Context -  None

Return Value:

    Return status in IRP when being called (if not STATUS_SUCCESS) or 
    STATUS_SUCCESS.


--*/

{
   NTSTATUS                    status      = Irp->IoStatus.Status;
   PSERIAL_DEVICE_EXTENSION    extension   = DeviceObject->DeviceExtension;
   PIO_STACK_LOCATION          IrpStack    = IoGetCurrentIrpStackLocation(Irp);

   SerialDump(SERDIAG3, ("SERIAL: In SerialD0Complete\n"));

   //
   // Check to see if the lower level driver completed successfully.
   //

   if (NT_SUCCESS(status)) {
      //
      // This should set up everything as it should be when
      // a device is to be opened.  We do need to lower the
      // modem lines, and disable the stupid fifo so that it
      // will show up if the user boots to dos.
      //

      KeSynchronizeExecution(extension->Interrupt, SerialReset, extension);
      KeSynchronizeExecution(extension->Interrupt, SerialMarkClose, extension);
      KeSynchronizeExecution(extension->Interrupt, SerialClrRTS, extension);
      KeSynchronizeExecution(extension->Interrupt, SerialClrDTR, extension);

      extension->PowerState = PowerDeviceD0;

      //
      // Now that we are powered up, call PoSetPowerState
      //

      PoSetPowerState(DeviceObject, IrpStack->Parameters.Power.Type,
		      IrpStack->Parameters.Power.State);
   } else {
      SerialDump(SERPNPPOWER, ("SerialSetPowerD0: Error status from lower "
				"driver - %x\n", status));
   }

   PoStartNextPowerIrp(Irp);

   return status;
}






NTSTATUS
SerialSetPowerD3(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)

/*++

Routine Description:

    This routine handles the SET_POWER minor function. 

Arguments:

    DeviceObject - Pointer to the device object for this device

    Irp - Pointer to the IRP for the current request

Return Value:

    The function value is the final status of the call


--*/

{
   NTSTATUS status = STATUS_SUCCESS;
   PSERIAL_DEVICE_EXTENSION extension = DeviceObject->DeviceExtension;
   PIO_STACK_LOCATION IrpStack = IoGetCurrentIrpStackLocation(Irp);

   SerialDump(SERDIAG3, ("SERIAL: In SerialSetPowerD3\n"));    

   //
   // Before we power down, call PoSetPowerState
   //
   
   PoSetPowerState(DeviceObject, IrpStack->Parameters.Power.Type,
		   IrpStack->Parameters.Power.State);

   // 
   // If the device is not closed, we will call SerialClose to stop processing
   // interrupts and allow the fifos to flush.
   //

#if 0
   if (extension->DeviceIsOpened == TRUE) {
      //
      // BUGBUG -- cannot use close routine this way!
      //

      SerialClose(DeviceObject, Irp);
   }
#endif

   //
   // Save device state here
   //

   Irp->IoStatus.Status = STATUS_SUCCESS;

   // 
   // Now call the ACPI driver (if one exists) to turn off the device
   //

   ASSERT(extension->LowerDeviceObject != NULL);

   //
   // For what we are doing, we don't need a completion routine
   // since we don't race on the power requests.
   //

   Irp->IoStatus.Status = STATUS_SUCCESS;

   IoCopyCurrentIrpStackLocationToNext(Irp);
   PoStartNextPowerIrp(Irp);

   return SerialPoCallDriver(extension, extension->LowerDeviceObject, Irp);
}





NTSTATUS
SerialD3Complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)

/*++

Routine Description:

    This routine is a completion handler that is called after the COM port
    has been powered down.  It just does a PoCompleteRequest to let the power
    manager know that this power Irp is done.

Arguments:

    DeviceObject - Pointer to the device object for this device

    Irp - Pointer to the IRP for the current request

    Context -  Not used

  Return Value:

    Return status in IRP.


--*/

{
   PSERIAL_DEVICE_EXTENSION extension = DeviceObject->DeviceExtension;
   NTSTATUS status = Irp->IoStatus.Status;

   SerialDump(SERTRACECALLS, ("SERIAL: In SerialD3Complete\n"));

   // 
   // Clear the flag saying we are waiting for a power down
   //

   extension->ReceivedQueryD3  = FALSE;

   //
   // bugbug:  The power state should be PowerDeviceD3 as far as we are 
   //          concerned now.  Brian commented it out because there was
   //          some problem with the serial driver trying to power up
   //          the comm port.
   //

   extension->PowerState = PowerDeviceD3;


   // 
   // ask for the next power Irp
   //

   PoStartNextPowerIrp(Irp);

   SerialDump(SERTRACECALLS, ("SERIAL: Leaving SerialD3Complete\n"));

   return status;
}


NTSTATUS
SerialFinishStartDevice(IN PDEVICE_OBJECT Fdo,
			IN PCM_RESOURCE_LIST resourceList,
			IN PCM_RESOURCE_LIST trResourceList,
			PSERIAL_USER_DATA PUserData)
/*++

Routine Description:

    This routine does serial-specific procedures to start a device.  It
    does this either for a legacy device detected by its registry entries,
    or for a PnP device after the start IRP has been sent down the stack.


Arguments:

   Fdo    	  -  Pointer to the Functional Device Object that is starting
   
   resourceList   -  Pointer to the untranslated resources needed by this device
   
   trResourceList -  Pointer to the translated resources needed by this device
   
   PUserData      -  Pointer to the user-specified resources/attributes 
   
   
  Return Value:

    STATUS_SUCCESS on success, something else appropriate on failure


--*/

{
   PSERIAL_DEVICE_EXTENSION fdoExtension = Fdo->DeviceExtension;
   NTSTATUS status;
   PCONFIG_DATA pConfig;

   PAGED_CODE();


   //
   // Allocate the config record.
   //

   pConfig = ExAllocatePool (PagedPool, sizeof(CONFIG_DATA));

   if (pConfig == NULL) {

      SerialLogError(
                    fdoExtension->DriverObject,
                    NULL,
                    SerialPhysicalZero,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    37,
                    STATUS_SUCCESS,
                    SERIAL_INSUFFICIENT_RESOURCES,
                    0,
                    NULL,
                    0,
                    NULL
                    );
      SerialDump(
                SERERRORS,
                ("SERIAL: Couldn't allocate memory for the\n"
                 "------  user configuration record\n")
                );

      status = STATUS_INSUFFICIENT_RESOURCES;
      goto StartDeviceCleanUp;
   }

   RtlZeroMemory(pConfig, sizeof(CONFIG_DATA));

   //
   // Get the configuration info for the device.
   //

   status = SerialGetPortInfo(Fdo, resourceList, trResourceList, pConfig,
                              PUserData);

   if (!NT_SUCCESS (status)) {
      goto StartDeviceCleanUp;
   }


#if 0
   //
   // See if we are in the proper power state.
   //
   if (fdoExtension->PowerState != PowerDeviceD0) {

      //
      // This is the same event used for starting, so we just need to clear it.
      //

      KeClearEvent(&fdoExtension->SerialSyncEvent);

      status = PoRequestPowerIrp(Fdo->LowerDevice, IRP_MN_SET_POWER,
				 PowerDeviceD0, SerialSyncCompletion,
				 &fdoExtension->SerialSyncEvent, NULL);

      if (status == STATUS_PENDING) {
	 KeWaitForSingleObject(&fdoExtension->SerialSyncEvent, Executive,
			       KernelMode, FALSE, NULL);
	 status = STATUS_SUCCESS;
      }

      if (!NT_SUCCESS(status)) {
	 goto StartDeviceCleanup;
      }


      if (fdoExtension->PowerState != PowerDeviceD0) {
	 //
	 // Setting power failed for some reason; let's assume the device
	 // isn't there (that's a good assumption if we can't power it up)
	 //

	 status = STATUS_NO_SUCH_DEVICE;
	 goto StartDeviceCleanup;
      }


   }

#endif

   //
   // Find and initialize the controller
   //


   status = SerialFindInitController(Fdo, pConfig);

   if (!NT_SUCCESS(status)) {
      goto StartDeviceCleanUp;
   }


   //
   // The hardware that is set up to NOT interrupt, connect an interrupt. 
   //

   //
   // If a device doesn't already have an interrupt and it has an isr then
   // we attempt to connect to the interrupt if it is not shareing with other
   // serial devices.  If we fail to connect to an  interrupt we will delete
   // this device.
   //

   if (fdoExtension != NULL) {
      SerialDump(SERDIAG5, ("SERIAL: fdoExtension: Interrupt %x\n"
			    "-------               OurIsr %x\n",
			    fdoExtension->Interrupt, fdoExtension->OurIsr));
   } else {
      SerialDump(SERERRORS, ("SERIAL: SerialFinishStartDevice got NULL "
			   "fdoExtension\n"));
   }

   if ((!fdoExtension->Interrupt) && (fdoExtension->OurIsr)) {

      SerialDump(
                SERDIAG5,
                ("SERIAL: About to connect to interrupt for port %wZ\n"
                 "------- address of extension is %x\n",
                 &fdoExtension->DeviceName, fdoExtension)
                );

      SerialDump(SERDIAG5, ("SERIAL: IoConnectInterrupt Args:\n"
                                "------- Interrupt           %x\n"
                                "------- OurIsr              %x\n"
                                "------- OurIsrContext       %x\n"
                                "------- NULL\n"
                                "------- Vector              %x\n"
                                "------- Irql                %x\n"
                                "------- InterruptMode       %x\n"
                                "------- InterruptShareable  %x\n"
                                "------- ProcessorAffinity   %x\n"
                                "------- FALSE\n",
                                &fdoExtension->Interrupt,
                                SerialCIsrSw,
                                &fdoExtension->CIsrSw,
                                fdoExtension->Vector,
                                fdoExtension->Irql,
                                fdoExtension->InterruptMode,
                                fdoExtension->InterruptShareable,
                                fdoExtension->ProcessorAffinity
                               ));

      //
      // Do a just in time construction of the ISR switch.
      //

      fdoExtension->CIsrSw.IsrFunc = fdoExtension->OurIsr;
      fdoExtension->CIsrSw.Context = fdoExtension->OurIsrContext;

      status = IoConnectInterrupt(&fdoExtension->Interrupt, SerialCIsrSw,
				  &fdoExtension->CIsrSw, NULL,
				  fdoExtension->Vector, fdoExtension->Irql,
				  fdoExtension->Irql,
				  fdoExtension->InterruptMode,
				  fdoExtension->InterruptShareable,
				  fdoExtension->ProcessorAffinity, FALSE);

      if (!NT_SUCCESS(status)) {

         //
         // Hmmm, how'd that happen?  Somebody either
         // didn't report their resources, or they
         // sneaked in since the last time I looked.
         //
         // Oh well,  delete this device. 
         //

         SerialDump(SERERRORS, ("SERIAL: Couldn't connect to interrupt for "
				"%wZ\n", &fdoExtension->DeviceName));

         SerialDump(SERERRORS, ("SERIAL: IoConnectInterrupt Args:\n"
                                "------- Interrupt           %x\n"
                                "------- OurIsr              %x\n"
                                "------- OurIsrContext       %x\n"
                                "------- NULL\n"
                                "------- Vector              %x\n"
                                "------- Irql                %x\n"
                                "------- InterruptMode       %x\n"
                                "------- InterruptShareable  %x\n"
                                "------- ProcessorAffinity   %x\n"
                                "------- FALSE\n",
                                &fdoExtension->Interrupt,
                                SerialCIsrSw,
                                &fdoExtension->CIsrSw,
                                fdoExtension->Vector,
                                fdoExtension->Irql,
                                fdoExtension->InterruptMode,
				fdoExtension->InterruptShareable,
                                fdoExtension->ProcessorAffinity
                               ));



         SerialLogError(Fdo->DriverObject, Fdo,
			fdoExtension->OriginalController,
			SerialPhysicalZero, 0, 0, 0, 1, status,
			SERIAL_UNREPORTED_IRQL_CONFLICT,
			fdoExtension->SymbolicLinkName.Length + sizeof(WCHAR),
			fdoExtension->SymbolicLinkName.Buffer, 0, NULL);

         status = SERIAL_UNREPORTED_IRQL_CONFLICT;
         goto StartDeviceCleanUp;

      }
   }


   //
   // Add the Fdo to the master list
   //

   InsertTailList(&SerialGlobals.AllFdos, &fdoExtension->AllFdos);


   //
   // Reset the device.
   //


   //
   // While the device isn't open, disable all interrupts.
   //

   DISABLE_ALL_INTERRUPTS (fdoExtension->Controller);


   if (fdoExtension->Jensen) {

      WRITE_MODEM_CONTROL(
                         fdoExtension->Controller,
                         (UCHAR)SERIAL_MCR_OUT2
                         );

   } else {

      WRITE_MODEM_CONTROL(
                         fdoExtension->Controller,
                         (UCHAR)0
                         );

   }

   //
   // This should set up everything as it should be when
   // a device is to be opened.  We do need to lower the
   // modem lines, and disable the stupid fifo so that it
   // will show up if the user boots to dos.
   //

   KeSynchronizeExecution(
                         fdoExtension->Interrupt,
                         SerialReset,
                         fdoExtension
                         );

   KeSynchronizeExecution( //Disables the fifo.
                           fdoExtension->Interrupt,
                           SerialMarkClose,
                           fdoExtension
                         );

   KeSynchronizeExecution(
                         fdoExtension->Interrupt,
                         SerialClrRTS,
                         fdoExtension
                         );

   KeSynchronizeExecution(
                         fdoExtension->Interrupt,
                         SerialClrDTR,
                         fdoExtension
                         );


   //
   // Do the external naming now that the device is accessible.
   //

   status = SerialDoExternalNaming(fdoExtension);


   if (!NT_SUCCESS(status)) {
      SerialDump (SERERRORS,("SERIAL: External Naming Failed - Status %x\n",
			     status));

      //
      // Allow the device to start anyhow
      //

      status = STATUS_SUCCESS;

   }


   StartDeviceCleanUp:

   //
   // Free resources created by this routine
   //

   if (pConfig) {
      ExFreePool (pConfig);
   }

   if (!NT_SUCCESS (status)) {

      SerialDump(SERDIAG1,("SERIAL: Cleaning up failed start\n"));

      //
      // Stop servicing interrupts
      //

      ASSERT(fdoExtension->Interrupt == NULL);

      if (fdoExtension->Interrupt != NULL) {
	 IoDisconnectInterrupt(fdoExtension->Interrupt);
	 fdoExtension->Interrupt = NULL;
      }

      //
      // If necessary, unmap the device registers.
      //

      if (fdoExtension->UnMapRegisters) {
	 MmUnmapIoSpace(fdoExtension->Controller,
			fdoExtension->SpanOfController);
	 fdoExtension->UnMapRegisters = FALSE;
      }

      if (fdoExtension->UnMapStatus) {
	 MmUnmapIoSpace(fdoExtension->InterruptStatus,
			fdoExtension->SpanOfInterruptStatus);
	 fdoExtension->UnMapStatus = FALSE;
      }
   } else { // SUCCESS
      fdoExtension->PNPState = SERIAL_PNP_STARTED;
      fdoExtension->Flags |= SERIAL_FLAGS_STARTED;
   }

   SerialDump (SERTRACECALLS, ("SERIAL: leaving SerialFinishStartDevice\n"));

   return status;
}



NTSTATUS
SerialStartDevice(
		  IN PDEVICE_OBJECT Fdo,
		  IN PIRP Irp
		  )

/*++

Routine Description:

    This routine first passes the start device Irp down the stack then
    it picks up the resources for the device, ititializes, puts it on any
    appropriate lists (i.e shared interrupt or interrupt status) and 
    connects the interrupt.

Arguments:

    Fdo - Pointer to the functional device object for this device

    Irp - Pointer to the IRP for the current request

Return Value:

    Return status


--*/

{


   PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(Irp);
   NTSTATUS status = STATUS_NOT_IMPLEMENTED;
   PSERIAL_DEVICE_EXTENSION fdoExtension = Fdo->DeviceExtension;
   PDEVICE_OBJECT pdo = fdoExtension->LowerDeviceObject;

   PAGED_CODE();

   SerialDump (SERTRACECALLS, ("SERIAL: entering SerialStartDevice\n"));

   //
   // Set up the external naming and create the symbolic link
   //

   

   //
   // Pass this down to the Pdo
   //

   KeInitializeEvent(&fdoExtension->SerialSyncEvent, SynchronizationEvent,
		     FALSE);

   IoCopyCurrentIrpStackLocationToNext(Irp);
   IoSetCompletionRoutine(
                         Irp,
                         SerialSyncCompletion,
                         &fdoExtension->SerialSyncEvent,
                         TRUE,
                         TRUE,
                         TRUE
                         );

   status = IoCallDriver(pdo, Irp);


   //
   // Wait for lower drivers to be done with the Irp
   //

   if (status == STATUS_PENDING) {

      KeWaitForSingleObject (&fdoExtension->SerialSyncEvent, 
                             Executive, KernelMode, FALSE, NULL);

      status = Irp->IoStatus.Status;
   }

   if (!NT_SUCCESS(status)) {
      SerialDump (SERERRORS, ("SERIAL: error with IoCallDriver %x\n", status));
      return status;
   }


   //
   // Do the serial specific items to start the device
   //
   status
      = SerialFinishStartDevice(Fdo,
				irpStack->Parameters.StartDevice
				 .AllocatedResources,
				irpStack->Parameters.StartDevice
				 .AllocatedResourcesTranslated,
				NULL);



   return status;
}



NTSTATUS
SerialItemCallBack(
                  IN PVOID Context,
                  IN PUNICODE_STRING PathName,
                  IN INTERFACE_TYPE BusType,
                  IN ULONG BusNumber,
                  IN PKEY_VALUE_FULL_INFORMATION *BusInformation,
                  IN CONFIGURATION_TYPE ControllerType,
                  IN ULONG ControllerNumber,
                  IN PKEY_VALUE_FULL_INFORMATION *ControllerInformation,
                  IN CONFIGURATION_TYPE PeripheralType,
                  IN ULONG PeripheralNumber,
                  IN PKEY_VALUE_FULL_INFORMATION *PeripheralInformation
                  )

/*++

Routine Description:

    This routine is called to check if a particular item
    is present in the registry.

Arguments:

    Context - Pointer to a boolean.

    PathName - unicode registry path.  Not Used.

    BusType - Internal, Isa, ...

    BusNumber - Which bus if we are on a multibus system.

    BusInformation - Configuration information about the bus. Not Used.

    ControllerType - Controller type.

    ControllerNumber - Which controller if there is more than one
                       controller in the system.

    ControllerInformation - Array of pointers to the three pieces of
                            registry information.

    PeripheralType - Should be a peripheral.

    PeripheralNumber - Which peripheral - not used..

    PeripheralInformation - Configuration information. Not Used.

Return Value:

    STATUS_SUCCESS

--*/

{

   *((BOOLEAN *)Context) = TRUE;
   return STATUS_SUCCESS;
}


NTSTATUS
SerialControllerCallBack(
                  IN PVOID Context,
                  IN PUNICODE_STRING PathName,
                  IN INTERFACE_TYPE BusType,
                  IN ULONG BusNumber,
                  IN PKEY_VALUE_FULL_INFORMATION *BusInformation,
                  IN CONFIGURATION_TYPE ControllerType,
                  IN ULONG ControllerNumber,
                  IN PKEY_VALUE_FULL_INFORMATION *ControllerInformation,
                  IN CONFIGURATION_TYPE PeripheralType,
                  IN ULONG PeripheralNumber,
                  IN PKEY_VALUE_FULL_INFORMATION *PeripheralInformation
                  )

/*++

Routine Description:

    This routine is called to check if a particular item
    is present in the registry.

Arguments:

    Context - Pointer to a boolean.

    PathName - unicode registry path.  Not Used.

    BusType - Internal, Isa, ...

    BusNumber - Which bus if we are on a multibus system.

    BusInformation - Configuration information about the bus. Not Used.

    ControllerType - Controller type.

    ControllerNumber - Which controller if there is more than one
                       controller in the system.

    ControllerInformation - Array of pointers to the three pieces of
                            registry information.

    PeripheralType - Should be a peripheral.

    PeripheralNumber - Which peripheral - not used..

    PeripheralInformation - Configuration information. Not Used.

Return Value:

    STATUS_SUCCESS

--*/

{
   PCM_FULL_RESOURCE_DESCRIPTOR controllerData;
   PSERIAL_PTR_CTX pContext = (PSERIAL_PTR_CTX)Context;
   ULONG i;

   if (ControllerInformation[IoQueryDeviceConfigurationData]->DataLength == 0) {
      pContext->isPointer = FALSE;
      return STATUS_SUCCESS;
   }

   controllerData =
      (PCM_FULL_RESOURCE_DESCRIPTOR)
      (((PUCHAR)ControllerInformation[IoQueryDeviceConfigurationData])
	+ ControllerInformation[IoQueryDeviceConfigurationData]->DataOffset);

   //
   // See if this is the exact port we are testing
   //
   for (i = 0; i < controllerData->PartialResourceList.Count; i++) {

      PCM_PARTIAL_RESOURCE_DESCRIPTOR partial
	 = &controllerData->PartialResourceList.PartialDescriptors[i];

      switch (partial->Type) {
      case CmResourceTypePort:
	 if (partial->u.Port.Start.QuadPart == pContext->Port.QuadPart) {
	    //
	    // Pointer on same controller. Bail out.
	    //
	    pContext->isPointer = SERIAL_FOUNDPOINTER_PORT;
	    return STATUS_SUCCESS;
	 }

      case CmResourceTypeInterrupt:
	 if (partial->u.Interrupt.Vector == pContext->Vector) {
	    //
	    // Pointer sharing this interrupt.  Bail out.
	    //
	    pContext->isPointer = SERIAL_FOUNDPOINTER_VECTOR;
	    return STATUS_SUCCESS;
	 }

      default:
	 break;
      }
   }

   pContext->isPointer = FALSE;
   return STATUS_SUCCESS;
}



NTSTATUS
SerialGetPortInfo(IN PDEVICE_OBJECT Fdo, IN PCM_RESOURCE_LIST pResourceList,
		  IN PCM_RESOURCE_LIST pTrResourceList,
		  OUT PCONFIG_DATA pConfig, IN PSERIAL_USER_DATA PUserData)

/*++

Routine Description:

    This routine will get the configuration information and put
    it and the translated values into CONFIG_DATA structures.
    It first sets up with  defaults and then queries the registry
    to see if the user has overridden these defaults; if this is a legacy
    multiport card, it uses the info in PUserData instead of groping the
    registry again.

Arguments:

    Fdo - Pointer to the functional device object.
    
    pResourceList - Pointer to the untranslated resources requested.
    
    pTrResourceList - Pointer to the translated resources requested.
    
    pConfig - Pointer to configuration info
    
    PUserData - Pointer to data discovered in the registry for
    legacy devices.

Return Value:

    STATUS_SUCCESS if consistant configuration was found - otherwise.
    returns STATUS_SERIAL_NO_DEVICE_INITED.

--*/

{
   PSERIAL_DEVICE_EXTENSION        fdoExtension    = Fdo->DeviceExtension;
   PDEVICE_OBJECT pdo = fdoExtension->LowerDeviceObject;
   NTSTATUS                        status          = STATUS_NOT_IMPLEMENTED;
   CONFIGURATION_TYPE              pointer         = PointerPeripheral;
   CONFIGURATION_TYPE              controllerType  = SerialController;

   HANDLE                          keyHandle;
   ULONG                           count;
   ULONG                           i;
   INTERFACE_TYPE                  interfaceType;

   PCM_PARTIAL_RESOURCE_LIST       pPartialResourceList, pPartialTrResourceList;
   PCM_PARTIAL_RESOURCE_DESCRIPTOR pPartialResourceDesc, pPartialTrResourceDesc;

   PCM_FULL_RESOURCE_DESCRIPTOR pFullResourceDesc = NULL,
      pFullTrResourceDesc = NULL;

   ULONG                           defaultInterruptMode;
   ULONG                           defaultAddressSpace;
   ULONG                           defaultInterfaceType;
   ULONG                           defaultClockRate;
   ULONG zero = 0;
   SERIAL_PTR_CTX foundPointerCtx;

   PAGED_CODE();



   SerialDump(SERTRACECALLS, ("SERIAL: entering SerialGetPortInfo\n"));

   SerialDump(SERPNPPOWER, ("SERIAL: resource pointer is %x\n", pResourceList));
   SerialDump(SERPNPPOWER, ("SERIAL: TR resource pointer is %x\n",
                            pTrResourceList));


   if ((pResourceList == NULL) || (pTrResourceList == NULL)) {
      // BUGBUG this shouldn't happen in theory

       ASSERT(pResourceList != NULL);
       ASSERT(pTrResourceList != NULL);

      //
      // This status is as appropriate as I can think of
      //
      return STATUS_INSUFFICIENT_RESOURCES;
   }


   pFullResourceDesc   = &pResourceList->List[0];
   pFullTrResourceDesc = &pTrResourceList->List[0];

   //
   // Ok, if we have a full resource descriptor.  Let's take it apart.
   //

   if (pFullResourceDesc) {
      pPartialResourceList    = &pFullResourceDesc->PartialResourceList;
      pPartialResourceDesc    = pPartialResourceList->PartialDescriptors;
      count                   = pPartialResourceList->Count;


      

      //
      // Pull out the stuff that is in the full descriptor.
      //

      pConfig->InterfaceType  = pFullResourceDesc->InterfaceType;
      pConfig->BusNumber      = pFullResourceDesc->BusNumber;

      //
      // Now run through the partial resource descriptors looking for the port,
      // interrupt, and clock rate.
      //

      pConfig->ClockRate = 1843200;
      pConfig->InterruptStatus = SerialPhysicalZero;
      pConfig->SpanOfInterruptStatus = SERIAL_STATUS_LENGTH;

      for (i = 0;     i < count;     i++, pPartialResourceDesc++) {

         switch (pPartialResourceDesc->Type) {
         case CmResourceTypePort: {

               if (pPartialResourceDesc->u.Port.Length
                   == SERIAL_STATUS_LENGTH) { // This is an ISR
                  pConfig->InterruptStatus = pPartialResourceDesc->u.Port.Start;
                  pConfig->SpanOfInterruptStatus
                     = pPartialResourceDesc->u.Port.Length;
                  pConfig->AddressSpace = pPartialResourceDesc->Flags;
               } else { // This is the serial register set
                  pConfig->Controller = pPartialResourceDesc->u.Port.Start;
                  pConfig->SpanOfController   = SERIAL_REGISTER_SPAN;
                  pConfig->AddressSpace       = pPartialResourceDesc->Flags;
               }

               break;
            }

         case CmResourceTypeInterrupt: {
               pConfig->OriginalIrql = pPartialResourceDesc->u.Interrupt.Level;
               pConfig->OriginalVector
                  = pPartialResourceDesc->u.Interrupt.Vector;
               pConfig->Affinity = pPartialResourceDesc->u.Interrupt.Affinity; 

               if (pPartialResourceDesc->Flags
                   & CM_RESOURCE_INTERRUPT_LATCHED) {
                  pConfig->InterruptMode  = Latched;
               } else {
                  pConfig->InterruptMode  = LevelSensitive;
               }
               break;
            }

         case CmResourceTypeDeviceSpecific: {
               PCM_SERIAL_DEVICE_DATA sDeviceData;

               sDeviceData = (PCM_SERIAL_DEVICE_DATA)(pPartialResourceDesc + 1);
               pConfig->ClockRate  = sDeviceData->BaudClock;
               break;
            }


         default: {
               break;
            }
         }   // switch (pPartialResourceDesc->Type)
      }       // for (i = 0;     i < count;     i++, pPartialResourceDesc++)
   }           // if (pFullResourceDesc)

   //
   // Call up IoQueryDeviceDescription to see if there is a pointer peripheral.
   // If there is then we ignore this controller.
   //
   // BUGBUG:
   //   This should go away when sermouse is redone.
   //

   foundPointerCtx.isPointer = 0;
   foundPointerCtx.Port = pConfig->Controller;
   foundPointerCtx.Vector = pConfig->OriginalVector;

   IoQueryDeviceDescription(&pFullResourceDesc->InterfaceType,
			    &pFullResourceDesc->BusNumber,
			    &controllerType, NULL, &pointer, NULL,
			    SerialControllerCallBack, &foundPointerCtx);

   if (foundPointerCtx.isPointer) {
      
      
      switch (foundPointerCtx.isPointer) {
      case SERIAL_FOUNDPOINTER_PORT:
         SerialLogError(Fdo->DriverObject, NULL, pConfig->Controller,
                    SerialPhysicalZero, 0, 0, 0, 56, STATUS_SUCCESS,
                    SERIAL_MOUSE_ON_PORT,
                    fdoExtension->SymbolicLinkName.Length + sizeof(WCHAR),
                    fdoExtension->SymbolicLinkName.Buffer, 0, NULL);
	 SerialDump(SERERRORS, ("SERIAL: Pointer device using port at address "
				"%x\n"
				"------- Serial driver will not load port %wZ",
				pConfig->Controller,
				&fdoExtension->SymbolicLinkName));
	 return STATUS_NO_SUCH_DEVICE;

      case SERIAL_FOUNDPOINTER_VECTOR:
         SerialLogError(Fdo->DriverObject, NULL, pConfig->Controller,
                    SerialPhysicalZero, 0, 0, 0, 57, STATUS_SUCCESS,
                    SERIAL_MOUSE_CONFLICT_IRQ,
                    fdoExtension->SymbolicLinkName.Length + sizeof(WCHAR),
                    fdoExtension->SymbolicLinkName.Buffer, 0, NULL);
	 SerialDump(SERERRORS, ("SERIAL: Pointer device using IRQ %x\n"
				"------- Serial driver will not load port %wZ",
				pConfig->OriginalVector,
				&fdoExtension->SymbolicLinkName));
	 return STATUS_NO_SUCH_DEVICE;

      default:
	 break;
      }
   }

   //
   // Do the same for the translated resources
   //

   if (pFullTrResourceDesc) {
      pPartialTrResourceList = &pFullTrResourceDesc->PartialResourceList;
      pPartialTrResourceDesc = pPartialTrResourceList->PartialDescriptors;
      count = pPartialTrResourceList->Count;

      //
      // Reload pConfig with the translated values for later use
      //

      pConfig->InterfaceType  = pFullTrResourceDesc->InterfaceType;
      pConfig->BusNumber      = pFullTrResourceDesc->BusNumber;

      pConfig->TrInterruptStatus = SerialPhysicalZero;

      for (i = 0;     i < count;     i++, pPartialTrResourceDesc++) {

         switch (pPartialTrResourceDesc->Type) {
         case CmResourceTypePort: {
               if (pPartialTrResourceDesc->u.Port.Length
		   == SERIAL_STATUS_LENGTH) { // This is an ISR
		  pConfig->TrInterruptStatus = pPartialTrResourceDesc
		     ->u.Port.Start;
               } else { // This is the serial register set
		  pConfig->TrController = pPartialTrResourceDesc->u.Port.Start;
		  pConfig->AddressSpace = pPartialTrResourceDesc->Flags;
               }

               break;
            }

         case CmResourceTypeInterrupt: {
	       pConfig->TrVector = pPartialTrResourceDesc->u.Interrupt.Vector;
	       pConfig->TrIrql = pPartialTrResourceDesc->u.Interrupt.Level;
	       pConfig->Affinity = pPartialTrResourceDesc->u.Interrupt.Affinity;
               break;
            }

         default: {
               break;
            }
         }   // switch (pPartialTrResourceDesc->Type)
      }       // for (i = 0;     i < count;     i++, pPartialTrResourceDesc++)
   }           // if (pFullTrResourceDesc)


   //
   // Initialize a config data structure with default values for those that
   // may not already be initialized.
   //

   pConfig->PortIndex = 0;
   pConfig->DisablePort = 0;
   pConfig->PermitSystemWideShare = FALSE;
   pConfig->MaskInverted = 0;
   pConfig->Indexed = 0;
   pConfig->ForceFifoEnable = driverDefaults.ForceFifoEnableDefault;
   pConfig->RxFIFO = driverDefaults.RxFIFODefault;
   pConfig->TxFIFO = driverDefaults.TxFIFODefault;
   pConfig->PermitShare = driverDefaults.PermitShareDefault;
   pConfig->LogFifo = driverDefaults.LogFifoDefault;
   pConfig->Jensen = driverDefaults.JensenDetected;

   //
   // Query the registry to look for the first bus on 
   // the system (that isn't the internal bus - we assume 
   // that the firmware code knows about those ports).  We 
   // will use that as the default bus if no bustype or bus 
   // number is specified in the "user" configuration records.
   //

   defaultInterfaceType            = (ULONG)Isa;
   defaultInterruptMode            = CM_RESOURCE_INTERRUPT_LATCHED;
   defaultAddressSpace             = CM_RESOURCE_PORT_IO;
   defaultClockRate                = 1843200;


   for (interfaceType = 0;     
       interfaceType < MaximumInterfaceType;
       interfaceType++
       ) {

      ULONG   busZero     = 0;
      BOOLEAN foundOne    = FALSE;


      if (interfaceType != Internal) {

         IoQueryDeviceDescription(
                                 &interfaceType,
                                 &busZero,
                                 NULL,
                                 NULL,
                                 NULL,
                                 NULL,
                                 SerialItemCallBack,
                                 &foundOne
                                 );

         if (foundOne) {
            defaultInterfaceType = (ULONG)interfaceType;

            if (defaultInterfaceType == MicroChannel) {

               defaultInterruptMode = CM_RESOURCE_INTERRUPT_LEVEL_SENSITIVE;
            }
            break;
         }
      }   // if (interfaceType != Internal) 
   }       // for (interfaceType = 0


   //
   // Get any user data associated with the port now and override the
   // values passed in if applicable.  If this a legacy device, this
   // is where we may actually get the parameters.
   //

   //
   // Open the "Device Parameters" section of registry for this device object.
   // If PUserData is NULL, this is PnP enumerated and we need to check,
   // otherwise we are doing a legacy device and have the info already.
   //


   if (PUserData == NULL) {
      status = IoOpenDeviceRegistryKey (pdo,
                                        PLUGPLAY_REGKEY_DEVICE,
                                        STANDARD_RIGHTS_READ,
                                        &keyHandle);

      if (!NT_SUCCESS(status)) {

         SerialDump(SERERRORS, ("SERIAL: IoOpenDeviceRegistryKey failed - %x "
				"\n", status));
         goto PortInfoCleanUp;    

      } else {
         status = SerialGetRegistryKeyValue (keyHandle,
                                             L"DisablePort",
                                             sizeof(L"DisablePort"),
                                             &pConfig->DisablePort,
                                             sizeof (ULONG));

         status = SerialGetRegistryKeyValue (keyHandle,
					     L"ForceFifoEnable",
					     sizeof(L"ForceFifoEnable"),
					     &pConfig->ForceFifoEnable,
					     sizeof (ULONG));

         status = SerialGetRegistryKeyValue (keyHandle,
					     L"RxFIFO",
					     sizeof(L"RxFIFO"),
					     &pConfig->RxFIFO,
					     sizeof (ULONG));

         status = SerialGetRegistryKeyValue (keyHandle,
					     L"TxFIFO",
					     sizeof(L"TxFIFO"),
					     &pConfig->TxFIFO,
					     sizeof (ULONG));

         status = SerialGetRegistryKeyValue (keyHandle,
					     L"MaskInverted",
					     sizeof(L"MaskInverted"),
					     &pConfig->MaskInverted,
					     sizeof (ULONG));
	 
         status = SerialGetRegistryKeyValue (keyHandle,
					     L"Share System Interrupt",
					     sizeof(L"Share System Interrupt"),
					     &pConfig->PermitShare,
					     sizeof (ULONG));

         status = SerialGetRegistryKeyValue (keyHandle,
					     L"PortIndex",
                                             sizeof(L"PortIndex"),
                                             &pConfig->PortIndex,
                                             sizeof (ULONG));

	 status = SerialGetRegistryKeyValue(keyHandle, L"Indexed",
					    sizeof(L"Indexed"),
					    &pConfig->Indexed,
					    sizeof(ULONG));

         status = SerialGetRegistryKeyValue (keyHandle,
                                             L"ClockRate",
                                             sizeof(L"ClockRate"),
                                             &pConfig->ClockRate,
                                             sizeof (ULONG));

	 if (!NT_SUCCESS(status)) {
            pConfig->ClockRate = defaultClockRate;
	 }


	 // Just default these for now
         pConfig->InterruptMode = defaultInterruptMode;
         pConfig->AddressSpace = defaultAddressSpace;

         ZwClose (keyHandle);
      }
   } else {
      //
      // This was a legacy device, either use a driver default or copy over
      // the user-specified values.
      //
      ULONG badValue = (ULONG)-1;

      pConfig->DisablePort = (PUserData->DisablePort == badValue)
	 ? 0
	 : PUserData->DisablePort;
      pConfig->ForceFifoEnable = (PUserData->ForceFIFOEnable == badValue)
	 ? PUserData->ForceFIFOEnableDefault
	 : PUserData->ForceFIFOEnable;
      pConfig->RxFIFO = (PUserData->RxFIFO == badValue)
	 ? PUserData->RxFIFODefault
	 : PUserData->RxFIFO;
      pConfig->Indexed = (PUserData->UserIndexed == badValue)
	 ? 0
	 : PUserData->UserIndexed;
      pConfig->TxFIFO = (PUserData->TxFIFO == badValue)
	 ? PUserData->TxFIFODefault
	 : PUserData->TxFIFO;
      pConfig->MaskInverted = (PUserData->MaskInverted == badValue)
	 ? 0
	 : PUserData->MaskInverted;
      pConfig->ClockRate = (PUserData->UserClockRate == badValue)
	 ? defaultClockRate
	 : PUserData->UserClockRate;
      pConfig->PermitShare = PUserData->PermitShareDefault;
      pConfig->PortIndex = PUserData->UserPortIndex;
   }

   //
   // Do some error checking on the configuration info we have.
   //
   // Make sure that the interrupt is non zero (which we defaulted
   // it to).
   //
   // Make sure that the portaddress is non zero (which we defaulted
   // it to).
   //
   // Make sure that the DosDevices is not NULL (which we defaulted
   // it to).
   //
   // We need to make sure that if an interrupt status
   // was specified, that a port index was also specfied,
   // and if so that the port index is <= maximum ports
   // on a board.
   //
   // We should also validate that the bus type and number
   // are correct.
   //
   // We will also validate that the interrupt mode makes
   // sense for the bus.
   //

   if (!pConfig->Controller.LowPart) {

      //
      // Ehhhh! Lose Game.
      //

      SerialLogError(
                    Fdo->DriverObject,
                    NULL,
                    pConfig->Controller,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    58,
                    STATUS_SUCCESS,
                    SERIAL_INVALID_USER_CONFIG,
                    0,
                    NULL,
                    sizeof(L"PortAddress") + sizeof(WCHAR),
                    L"PortAddress"
                    );

      SerialDump (SERERRORS, 
                  ("SERIAL: Bogus port address %x\n",
		   pConfig->Controller.LowPart));

      status = SERIAL_INVALID_USER_CONFIG;
      goto PortInfoCleanUp;    
   }


   if (!pConfig->OriginalVector) {

      //
      // Ehhhh! Lose Game.
      //

      SerialLogError(
                    fdoExtension->DriverObject,
                    NULL,
                    pConfig->Controller,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    59,
                    STATUS_SUCCESS,
                    SERIAL_INVALID_USER_CONFIG,
                    0,
                    NULL,
                    sizeof (L"Interrupt"),
                    L"Interrupt"
                    );

      SerialDump (SERERRORS,("SERIAL: Bogus vector %x\n",
			     pConfig->OriginalVector));

      status = SERIAL_INVALID_USER_CONFIG;
      goto PortInfoCleanUp;    
   }


   if (pConfig->InterruptStatus.LowPart != 0) {

      if (pConfig->PortIndex == MAXULONG) {

         //
         // Ehhhh! Lose Game.
         //

         SerialLogError(
                       fdoExtension->DriverObject,
                       NULL,
                       pConfig->Controller,
                       SerialPhysicalZero,
                       0,
                       0,
                       0,
                       30,
                       STATUS_SUCCESS,
                       SERIAL_INVALID_PORT_INDEX,
                       0,
                       NULL,
                       0,
                       NULL
                       );

         SerialDump (SERERRORS,
                     ("SERIAL: Bogus port index %x\n", pConfig->PortIndex));

         status = SERIAL_INVALID_PORT_INDEX;
         goto PortInfoCleanUp;    

      } else if (!pConfig->PortIndex) {

         //
         // So sorry, you must have a non-zero port index.
         //

         SerialLogError(
                       fdoExtension->DriverObject,
                       NULL,
                       pConfig->Controller,
                       SerialPhysicalZero,
                       0,
                       0,
                       0,
                       31,
                       STATUS_SUCCESS,
                       SERIAL_INVALID_PORT_INDEX,
                       0,
                       NULL,
                       0,
                       NULL
                       );

         SerialDump(
                   SERERRORS,
                   ("SERIAL: Port index must be > 0 for any\n"
                    "------- port on a multiport card: %x\n",
                    pConfig->PortIndex)
                   );

         status = SERIAL_INVALID_PORT_INDEX;
         goto PortInfoCleanUp;    

      } else {

         if (pConfig->Indexed) {

            if (pConfig->PortIndex > SERIAL_MAX_PORTS_INDEXED) {

               SerialLogError(
                             fdoExtension->DriverObject,
                             NULL,
                             pConfig->Controller,
                             SerialPhysicalZero,
                             0,
                             0,
                             0,
                             32,
                             STATUS_SUCCESS,
                             SERIAL_PORT_INDEX_TOO_HIGH,
                             0,
                             NULL,
                             0,
                             NULL
                             );

               SerialDump (SERERRORS,
                           ("SERIAL: port index to large %x\n",
			    pConfig->PortIndex));

               status = SERIAL_PORT_INDEX_TOO_HIGH;
               goto PortInfoCleanUp;    
            }

         } else {

            if (pConfig->PortIndex > SERIAL_MAX_PORTS_NONINDEXED) {

               SerialLogError(
                             fdoExtension->DriverObject,
                             NULL,
                             pConfig->Controller,
                             SerialPhysicalZero,
                             0,
                             0,
                             0,
                             33,
                             STATUS_SUCCESS,
                             SERIAL_PORT_INDEX_TOO_HIGH,
                             0,
                             NULL,
                             0,
                             NULL
                             );

               SerialDump (SERERRORS,
                           ("SERIAL: port index to large %x\n",
			    pConfig->PortIndex));

               status = SERIAL_PORT_INDEX_TOO_HIGH;
               goto PortInfoCleanUp;    
            }

         }

      }   // else  (if !pConfig->PortIndex)

   }       // if (pConfig->InterruptStatus != 0)


   //
   // We don't want to cause the hal to have a bad day,
   // so let's check the interface type and bus number.
   //
   // We only need to check the registry if they aren't
   // equal to the defaults.
   //

   if (pConfig->BusNumber != 0) {

      BOOLEAN foundIt;

      if (pConfig->InterfaceType >= MaximumInterfaceType) {

         //
         // Ehhhh! Lose Game.
         //

         SerialLogError(
                       fdoExtension->DriverObject,
                       NULL,
                       pConfig->Controller,
                       SerialPhysicalZero,
                       0,
                       0,
                       0,
                       34,
                       STATUS_SUCCESS,
                       SERIAL_UNKNOWN_BUS,
                       0,
                       NULL,
                       0,
                       NULL
                       );

         SerialDump (SERERRORS,
                     ("SERIAL: Invalid Bus type %x\n", pConfig->BusNumber));

         status = SERIAL_UNKNOWN_BUS;
         goto PortInfoCleanUp;    
      }

      IoQueryDeviceDescription(
                              (INTERFACE_TYPE *)&pConfig->InterfaceType,
                              &zero,
                              NULL,
                              NULL,
                              NULL,
                              NULL,
                              SerialItemCallBack,
                              &foundIt
                              );

      if (!foundIt) {

         SerialLogError(
                       fdoExtension->DriverObject,
                       NULL,
                       pConfig->Controller,
                       SerialPhysicalZero,
                       0,
                       0,
                       0,
                       35,
                       STATUS_SUCCESS,
                       SERIAL_BUS_NOT_PRESENT,
                       0,
                       NULL,
                       0,
                       NULL
                       );
         SerialDump(
                   SERERRORS,
                   ("SERIAL: There aren't that many of those\n"
                    "------- busses on this system,%x\n",
                    pConfig->BusNumber)
                   );

         status = SERIAL_BUS_NOT_PRESENT;
         goto PortInfoCleanUp;    

      }

   }   // if (pConfig->BusNumber != 0)


   if ((pConfig->InterfaceType == MicroChannel) &&
       (pConfig->InterruptMode == CM_RESOURCE_INTERRUPT_LATCHED)) {

      SerialLogError(
                    fdoExtension->DriverObject,
                    NULL,
                    pConfig->Controller,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    36,
                    STATUS_SUCCESS,
                    SERIAL_BUS_INTERRUPT_CONFLICT,
                    0,
                    NULL,
                    0,
                    NULL
                    );
      SerialDump(
                SERERRORS,
                ("SERIAL: Latched interrupts and MicroChannel\n"
                 "------- busses don't mix\n")
                );

      status = SERIAL_BUS_INTERRUPT_CONFLICT;
      goto PortInfoCleanUp;    
   }


   status = STATUS_SUCCESS;

   //
   // Dump out the port configuration.
   //

   SerialDump(SERDIAG1, ("SERIAL: Com Port address: %x\n",
			  pConfig->Controller.LowPart));

   SerialDump(SERDIAG1, ("SERIAL: Com Interrupt Status: %x\n",
			 pConfig->InterruptStatus));

   SerialDump(SERDIAG1, ("SERIAL: Com Port Index: %x\n",
			 pConfig->PortIndex));

   SerialDump(SERDIAG1, ("SERIAL: Com Port ClockRate: %x\n",
			 pConfig->ClockRate));

   SerialDump(SERDIAG1, ("SERIAL: Com Port BusNumber: %x\n",
			 pConfig->BusNumber));

   SerialDump(SERDIAG1, ("SERIAL: Com AddressSpace: %x\n",
			 pConfig->AddressSpace));

   SerialDump(SERDIAG1, ("SERIAL: Com InterruptMode: %x\n",
			 pConfig->InterruptMode));

   SerialDump(SERDIAG1, ("SERIAL: Com InterfaceType: %x\n",
			 pConfig->InterfaceType));

   SerialDump(SERDIAG1, ("SERIAL: Com OriginalVector: %x\n",
			 pConfig->OriginalVector));

   SerialDump(SERDIAG1, ("SERIAL: Com OriginalIrql: %x\n",
			 pConfig->OriginalIrql));

   SerialDump(SERDIAG1, ("SERIAL: Com Indexed: %x\n",
			 pConfig->Indexed));

   PortInfoCleanUp:

   //
   // Free resources created by this routine
   //

   return status;
}


#define FAR

#include <initguid.h>

DEFINE_GUID(GUID_CLASS_COMM, 0xB115ED80L, 0x46DF, 0x11D0, 0xB4, 0x65, 0x00,
            0x00, 0x1A, 0x18, 0x18, 0xE6);

#define GUID_CLASS_COMM_STR "B115ED80-46DF-11D0-B465-00001A1818E6"




NTSTATUS
SerialDoExternalNaming(IN PSERIAL_DEVICE_EXTENSION fdoExtension)

/*++

Routine Description:

    This routine will be used to create a symbolic link
    to the driver name in the given object directory.

    It will also create an entry in the device map for
    this device - IF we could create the symbolic link.

Arguments:

    Extension - Pointer to the device extension.

Return Value:

    None.

--*/

{
   NTSTATUS    status;
   PAGED_CODE();


   status = IoCreateSymbolicLink (
                                 &fdoExtension->SymbolicLinkName,
                                 &fdoExtension->DeviceName
                                 );

   if (!NT_SUCCESS(status)) {

      //
      // Oh well, couldn't create the symbolic link.  No point
      // in trying to create the device map entry.
      //

      SerialLogError(
                    fdoExtension->DeviceObject->DriverObject,
                    fdoExtension->DeviceObject,
                    SerialPhysicalZero,
                    SerialPhysicalZero,
                    0,
                    0,
                    0,
                    52,
                    status,
                    SERIAL_NO_SYMLINK_CREATED,
                    fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                    fdoExtension->SymbolicLinkName.Buffer,
                    0,
                    NULL
                    );

      SerialDump(
                SERERRORS,
                ("SERIAL: Couldn't create the symbolic link\n"
                 "------- for port %wZ\n",
                 &fdoExtension->DeviceName)
                );

   } else {

      fdoExtension->CreatedSymbolicLink = TRUE;

      status = RtlWriteRegistryValue(
                                    RTL_REGISTRY_DEVICEMAP,
                                    L"SERIALCOMM",
                                    fdoExtension->DeviceName.Buffer,
                                    REG_SZ,
                                    fdoExtension->DosName.Buffer,
                                    fdoExtension->DosName.Length
				    + sizeof(WCHAR));

      if (!NT_SUCCESS(status)) {

         SerialLogError(
                       fdoExtension->DeviceObject->DriverObject,
                       fdoExtension->DeviceObject,
                       SerialPhysicalZero,
                       SerialPhysicalZero,
                       0,
                       0,
                       0,
                       53,
                       status,
                       SERIAL_NO_DEVICE_MAP_CREATED,
                       fdoExtension->SymbolicLinkName.Length+sizeof(WCHAR),
                       fdoExtension->SymbolicLinkName.Buffer,
                       0,
                       NULL
                       );

         SerialDump(
                   SERERRORS,
                   ("SERIAL: Couldn't create the device map entry\n"
                    "------- for port %wZ\n",
                    &fdoExtension->DeviceName)
                   );
      }
	        
      //
      // Make the device visible via a device association as well.
      // The reference string is the eight digit device index
      //
        
      status = IoRegisterDeviceInterface(fdoExtension->LowerDeviceObject,
					 (LPGUID)&GUID_CLASS_COMM, NULL,
					 &fdoExtension->DeviceClassSymbolicName
					 );

      if (!NT_SUCCESS(status)) {
	 SerialDump(SERERRORS, ("SERIAL: Couldn't register class association\n"
				"------- for port %wZ\n",
				&fdoExtension->DeviceName));

	 fdoExtension->DeviceClassSymbolicName.Buffer = NULL;
        
	 return status;
      }

            
      // 
      // Now set the symbolic link for the association
      //
        
      status = IoSetDeviceInterfaceState(&fdoExtension->DeviceClassSymbolicName,
					 TRUE);

      if (!NT_SUCCESS(status)) {
            SerialDump(SERERRORS, ("SERIAL: Couldn't set class association\n"
				   "------- for port %wZ\n",
				   &fdoExtension->DeviceName));
      }
   }

   return status;
}





VOID
SerialUndoExternalNaming(IN PSERIAL_DEVICE_EXTENSION Extension)

/*++

Routine Description:

    This routine will be used to delete a symbolic link
    to the driver name in the given object directory.

    It will also delete an entry in the device map for
    this device if the symbolic link had been created.

Arguments:

    Extension - Pointer to the device extension.

Return Value:

    None.

--*/

{

   NTSTATUS status;

   PAGED_CODE();

   SerialDump(
             SERDIAG3,
             ("SERIAL: In SerialUndoExternalNaming for\n"
              "------- extension: %x of port %wZ\n",
              Extension,&Extension->DeviceName)
             );

   //
   // We're cleaning up here.  One reason we're cleaning up
   // is that we couldn't allocate space for the directory
   // name or the symbolic link.
   //

   if (Extension->SymbolicLinkName.Buffer &&
       Extension->CreatedSymbolicLink) {

	    if (Extension->DeviceClassSymbolicName.Buffer) {
	       status = IoSetDeviceInterfaceState (&Extension->
						   DeviceClassSymbolicName,
						   FALSE);
            
	       //
	       // IoRegisterDeviceClassInterface() allocated this string for us,
	       // and we no longer need it.
	       //

	       ExFreePool( Extension->DeviceClassSymbolicName.Buffer );
	    }


      IoDeleteSymbolicLink (&Extension->SymbolicLinkName);

   }


   //
   // We're cleaning up here.  One reason we're cleaning up
   // is that we couldn't allocate space for the NtNameOfPort.
   //

   if (Extension->DeviceName.Buffer) {

      status = RtlDeleteRegistryValue(
                                     RTL_REGISTRY_DEVICEMAP,
                                     L"SERIALCOMM",
                                     Extension->DeviceName.Buffer
                                     );

      if (!NT_SUCCESS(status)) {

         SerialLogError(
                       Extension->DeviceObject->DriverObject,
                       Extension->DeviceObject,
                       Extension->OriginalController,
                       SerialPhysicalZero,
                       0,
                       0,
                       0,
                       55,
                       status,
                       SERIAL_NO_DEVICE_MAP_DELETED,
                       Extension->SymbolicLinkName.Length+sizeof(WCHAR),
                       Extension->SymbolicLinkName.Buffer,
                       0,
                       NULL
                       );
         SerialDump(
                   SERERRORS,
                   ("SERIAL: Couldn't delete value entry %wZ\n",
                    &Extension->DeviceName)
                   );

      }
   }
}

