
/*++

Copyright (c) 1997    Microsoft Corporation

Module Name:

    sample.c

Abstract:

    Sample DDK PnP driver 

Environment:

    Kernel mode

Revision History:

    16-July-1997 : Incorporated KenRay's observations:
        
        - changed device extension flag names and types.
        - added SD_IoIncrement and SD_IoDecrement.
        - added a dispatch function for IRP_MJ_CLOSE.
        - added an ASSERT in SD_Unload.
        - added comments about giving up resource when
        IRP_MN_STOP_DEVICE is received.
    
    22-July-1997    :  Incorporated the latest spec:
        - HoldNewRequests will be cleared on AddDevice.
        - subsequently change the logic of SD_Pass (any request that
        attempts to touch the device before it is started will be
        be discarded).
        - changing the way CANCEL routines work (first pass the Irp,
        wait to an event, then dequeue).
        - added SD_Cleanup.
        - changed the queue to a interlocked queue.
        - added the stop event (the driver needs to wait for the
        outstanding requests to be completed when he is stopped).
        - removed the start event in the device extension (to be sure that
        different waits don't interfere, I'd rather use a variable on the
        stack).
     
    23-July-1997    :   
        - HoldNewRequests will be set on IRP_MN_START_DEVICE will be recieved,
        but before the Irp it is passed down (to store the requests received in
        between).
    
    25-July-1997    :
        - Moved the declarations in a separate header file (sample.h)
        - Moved SD_DispatchPower in a separate module    
    
    
--*/

#include "sample.h"


#if DBG

#define     TEMP_BUFFER_SIZE        1024

ULONG   g_SampleDriverDebugLevel = 3;

VOID
SampleDriverDbgPrint    (
    ULONG   DebugPrintLevel,
    PCCHAR  DebugMessage,
    ...
    )   
    
/*++

Routine Description:

    Debug print for the sample driver.

Arguments:

    DebugPrintLevel - print level between 0 and 3, with 3 the most verbose
    
Return Value:
    
    None.
  
 --*/
 {

 va_list    list;
 UCHAR      debugMessageBuffer[TEMP_BUFFER_SIZE];

 va_start(list, DebugMessage);
 if (DebugPrintLevel <= g_SampleDriverDebugLevel) {
     
     vsprintf(debugMessageBuffer, DebugMessage, list);

     DbgPrint(debugMessageBuffer);
 }

 va_end(list);

 return;
 
}

#else

VOID
SampleDriverDbgPrint    (
    ULONG   DebugPrintLevel,
    PCCHAR  DebugMessage,
    ...
    )   ;   

#endif  // DBG


//
// Private routines (used to manipulate the held IRPs queue
// or increment/decrement the requests counter)
//
NTSTATUS
pSD_QueueRequest    (
    IN OUT PSD_FDO_DATA FdoData,
    IN PIRP Irp
    );
    

VOID
pSD_ProcessQueuedRequests    (
    IN OUT PSD_FDO_DATA FdoData
    );
    

NTSTATUS
pSD_CanStopDevice    (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP           Irp
    );

NTSTATUS
pSD_CanRemoveDevice    (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP           Irp
    );



#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (PAGE, SD_DispatchPower)
#pragma alloc_text (PAGE, SD_DispatchPnp)
#pragma alloc_text (PAGE, SD_StartDevice)
#pragma alloc_text (PAGE, SD_Unload)
#endif


NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
/*++

Routine Description:

    Installable driver initialization entry point.
    This entry point is called directly by the I/O system.

Arguments:

    DriverObject - pointer to the driver object

    RegistryPath - pointer to a unicode string representing the path,
                   to driver-specific key in the registry.

Return Value:

    STATUS_SUCCESS if successful,
    STATUS_UNSUCCESSFUL otherwise.

--*/
{
    NTSTATUS            status = STATUS_SUCCESS;
    ULONG               ulIndex;
    PDRIVER_DISPATCH  * dispatch;

    UNREFERENCED_PARAMETER (RegistryPath);

    // TRAP();
    
    SD_DebugPrint ((1, "Entered the Driver Entry\n"));
    

    //
    // Create dispatch points
    //
    for (ulIndex = 0, dispatch = DriverObject->MajorFunction;
         ulIndex <= IRP_MJ_MAXIMUM_FUNCTION;
         ulIndex++, dispatch++) {

        *dispatch = SD_Pass;
    }

    DriverObject->MajorFunction[IRP_MJ_PNP]            = SD_DispatchPnp;
    DriverObject->MajorFunction[IRP_MJ_POWER]          = SD_DispatchPower;
    DriverObject->MajorFunction[IRP_MJ_CREATE]         = 
    DriverObject->MajorFunction[IRP_MJ_SHUTDOWN]       = 
    DriverObject->MajorFunction[IRP_MJ_CLOSE]          = SD_CreateClose;
    
    DriverObject->MajorFunction[IRP_MJ_CLEANUP]        = SD_Cleanup;
    
    
    DriverObject->DriverExtension->AddDevice           = SD_AddDevice;
    DriverObject->DriverUnload                         = SD_Unload;

    return status;
}


NTSTATUS
SD_AddDevice(
    IN PDRIVER_OBJECT DriverObject,
    IN PDEVICE_OBJECT PhysicalDeviceObject
    )
/*++

Routine Description:

    The PlugPlay subsystem is handing us a brand new PDO, for which we
    (by means of INF registration) have been asked to provide a driver.

    We need to determine if we need to be in the driver stack for the device.
    Create a functional device object to attach to the stack
    Initialize that device object
    Return status success.

    Remember: we can NOT actually send ANY non pnp IRPS to the given driver
    stack, UNTIL we have received an IRP_MN_START_DEVICE.

Arguments:

    DeviceObject - pointer to a device object.

    PhysicalDeviceObject -  pointer to a device object created by the
                            underlying bus driver.

Return Value:

    NT status code.

--*/
{
    NTSTATUS                status = STATUS_SUCCESS;
    PDEVICE_OBJECT          deviceObject = NULL;
    PSD_FDO_DATA            fdoData;


    SD_DebugPrint ((1, "AddDevice (0x%x)\n", PhysicalDeviceObject));

    //
    // Remember that you CANNOT send an IRP to the PDO because it has not
    // been started as of yet, but you can make PlugPlay queries to find
    // out things like hardware, compatible ID's, etc.
    //
    
    //
    // Create a functional device object.
    //

    status = IoCreateDevice (DriverObject,
                             sizeof (SD_FDO_DATA),
                             NULL,  // No Name
                             FILE_DEVICE_UNKNOWN,
                             0,
                             FALSE,
                             &deviceObject);

    
    if (!NT_SUCCESS (status)) {
        //
        // returning failure here prevents the entire stack from functioning,
        // but most likely the rest of the stack will not be able to create
        // device objects either, so it is still OK.
        //
        return status;
    }

    //
    // Initialize the device extension.
    //
    fdoData = (PSD_FDO_DATA) deviceObject->DeviceExtension;
    //
    // The device is not removed
    //
    fdoData->IsStarted = FALSE;
    fdoData->IsRemoved = FALSE;
    //
    // We will ignore all requests until start
    //
    fdoData->HoldNewRequests = FALSE;
    fdoData->Self = deviceObject;
    fdoData->PDO = PhysicalDeviceObject;
    fdoData->NextLowerDriver = NULL;

    InitializeListHead(&fdoData->NewRequestsQueue);
    KeInitializeSpinLock(&fdoData->QueueLock);
    //
    // Initalize the remove event to Not-Signaled
    //
    KeInitializeEvent(&fdoData->RemoveEvent, 
                      SynchronizationEvent, 
                      FALSE);
    //  
    // Initialize the stop event to Signaled:
    // there are no Irps that prevent the device from being 
    // stopped. This event will be set when the OutstandingIO
    // will become 0.
    //
    KeInitializeEvent(&fdoData->StopEvent, 
                      SynchronizationEvent, 
                      TRUE);



    fdoData->OutstandingIO = 1; // biased to 1.  Transition to zero during
                                // remove device means IO is finished.
                                // Transition to 1 means the device can be 
                                // stopped.

    
    deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

    //
    // Attach our driver to the device stack.
    // the return value of IoAttachDeviceToDeviceStack is the top of the
    // attachment chain.  This is where all the IRPs should be routed.
    //
    // Our driver will send IRPs to the top of the stack and use the PDO
    // for all PlugPlay functions.
    //
    fdoData->NextLowerDriver = IoAttachDeviceToDeviceStack (deviceObject,
                                                       PhysicalDeviceObject);
    //
    // if this attachment fails then top of stack will be null.
    // failure for attachment is an indication of a broken plug play system.
    //
    ASSERT (NULL != fdoData->NextLowerDriver);

    return STATUS_SUCCESS;

}

NTSTATUS
SD_Pass (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*++

Routine Description:

    The default dispatch routine.  If this driver does not recognize the
    IRP, then it should send it down, unmodified.
    If the device holds IRPs, this IRP must be queued in the device extension
    No completion routine is required.

    As we have NO idea which function we are happily passing on, we can make
    NO assumptions about whether or not it will be called at raised IRQL.
    For this reason, this function must be in put into non-paged pool
    (aka the default location).

Arguments:

   DeviceObject - pointer to a device object.

   Irp - pointer to an I/O Request Packet.

Return Value:

      NT status code

--*/
{
    PSD_FDO_DATA        fdoData;
    NTSTATUS            status;
    PIO_STACK_LOCATION  stack;

    LONG                requestCount;
    
    fdoData = (PSD_FDO_DATA) DeviceObject->DeviceExtension;
    stack = IoGetCurrentIrpStackLocation(Irp);
    
    
    //
    // We need to hold the requests that access the device when it is 
    // stopped. We are currently holding all the IRPs except the PnP, 
    // power, craete, close, shutdown.
    //
    //
    
    //
    // This IRP was sent to the function driver.
    // We need to check if we are currently holding requests
    //
    // We will count the Irp only when we're going to process
    // (dequeue) it. This is because we can't possibly count before
    // we queue it (we can receive a surprise remove in between).
    //
    if (fdoData->HoldNewRequests) {
        //
        // We are holding requests only if we are not removed
        //
        ASSERT(!fdoData->IsRemoved);

        status = pSD_QueueRequest(fdoData, Irp);
        if (NT_SUCCESS(status)) {
            //
            // The queue operation proceeded without error, so
            // we need to return STATUE_PENDING. Otherwise the specific 
            // error encountered will be returned (STATUS_CANCELLED 
            // for example).
            //
            status = STATUS_PENDING;
        }
        return status;
    }
    //
    // Since we do not know what to do with the IRP, we should pass
    // it on along down the stack. We also need to refcount it.
    //
    requestCount = SD_IoIncrement (fdoData);
    
    if (fdoData->IsRemoved || !fdoData->IsStarted) {
        //
        // The device is not active.
        // We can get here either:
        // - because a surprise removal was issued, but our request 
        // arrived after that
        // OR
        // - because we received a request before we're started.
        // Either way, we need to fail the request.
        //
        if (fdoData->IsRemoved) {
            status = STATUS_DELETE_PENDING;
        }   else    {
            status = STATUS_UNSUCCESSFUL;
        }
        Irp->IoStatus.Information = 0;
        Irp->IoStatus.Status = status;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);

    } else {
        //
        // We are in the common situation where we send the IRP
        // down on the driver stack
        //
        IoSkipCurrentIrpStackLocation (Irp);
        status = IoCallDriver (fdoData->NextLowerDriver, Irp);
    }

    requestCount = SD_IoDecrement(fdoData);
        
    return status;
}






NTSTATUS
SD_DispatchPnp (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*++

Routine Description:

    The plug and play dispatch routines.

    Most of these these the driver will completely ignore.
    In all cases it must pass on the IRP to the lower driver.

Arguments:

   DeviceObject - pointer to a device object.

   Irp - pointer to an I/O Request Packet.

Return Value:

      NT status code

--*/
{
    PSD_FDO_DATA            fdoData;
    PIO_STACK_LOCATION      stack;
    NTSTATUS                status;
    PDEVICE_CAPABILITIES    deviceCapabilities;
    
    KEVENT                  eventWaitLowerDrivers;

    LONG                    requestCount;

    PAGED_CODE ();

    fdoData = (PSD_FDO_DATA) DeviceObject->DeviceExtension;
    
    stack = IoGetCurrentIrpStackLocation (Irp);

    requestCount  = SD_IoIncrement (fdoData);
    
    SD_DebugPrint((1, "Dispatch, Device Object = 0x%x, Code = 0x%x\n",
              DeviceObject, 
              stack->MinorFunction));
    
    
    if (fdoData->IsRemoved) {

        //
        // Since the device is stopped, but we don't hold IRPs,
        // this is a surprise removal. Just fail it.
        //
        requestCount = SD_IoDecrement(fdoData);
        Irp->IoStatus.Information = 0;
        Irp->IoStatus.Status = STATUS_DELETE_PENDING;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);
        return STATUS_DELETE_PENDING;
    }

    switch (stack->MinorFunction) {
    case IRP_MN_START_DEVICE:

        //
        // The device is starting.
        //
        // We cannot touch the device (send it any non pnp irps) until a
        // start device has been passed down to the lower drivers.
        //
        SD_DebugPrint((1, "Starting Device...\n"));
        
        //
        // We can't completely rule out that we might receive a request
        // before the Irp is completed by the lower drivers. This might
        // happen if we're a filter driver above a function drivers that
        // exposes a device interface (symbolic link) and an app from user
        // mode start sending requests to us...  So, we need to queue those
        // requests (we can't touch the hardware yet).
        //
        fdoData->HoldNewRequests = TRUE;

        IoCopyCurrentIrpStackLocationToNext (Irp);
        
        //
        // Initialize the event
        // 
        KeInitializeEvent(&eventWaitLowerDrivers, 
                          SynchronizationEvent, 
                          FALSE);
        
        IoSetCompletionRoutine (Irp,
                                SD_DispatchPnpComplete,
                                &eventWaitLowerDrivers,
                                TRUE,
                                TRUE,
                                TRUE);

        status = IoCallDriver (fdoData->NextLowerDriver, Irp);
        
        if (STATUS_PENDING == status) {
            KeWaitForSingleObject(
               &eventWaitLowerDrivers,
               Executive, // Waiting for reason of a driver
               KernelMode, // Waiting in kernel mode
               FALSE, // No allert
               NULL); // No timeout

            status = Irp->IoStatus.Status;
        }

        if (NT_SUCCESS (status)) {
            //
            // Lower drivers have finished their start operation, so now
            // we can finish ours.
            //
            status = SD_StartDevice (fdoData, Irp);
        }

        //
        // We must now complete the IRP, since we stopped it in the
        // completetion routine with MORE_PROCESSING_REQUIRED.
        //
        Irp->IoStatus.Status = status;
        Irp->IoStatus.Information = 0;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);
        break;

    
    case IRP_MN_QUERY_STOP_DEVICE:
        //
        // If we can stop the device, we need to set the HoldNewRequests flag,
        // so further requests will be queued. We don't care about processing
        // some old requests (if there are any), because we expect to be 
        // started again in the future.
        //
        if (!fdoData->IsStarted) {
            //
            // If we aren't started, we'll just pass the irp down
            // A driver has no reason to veto a query when he
            // is stopped already.
            //
            IoSkipCurrentIrpStackLocation (Irp);
            status = IoCallDriver (fdoData->NextLowerDriver, Irp);
            SD_IoDecrement(fdoData);

            return status;
        }
        //
        // We can't be removed at this point
        //
        ASSERT(!fdoData->IsRemoved);
        
        status = pSD_CanStopDevice(DeviceObject, Irp);
        Irp->IoStatus.Status = status;
        if (NT_SUCCESS(status)) {
            //
            // First, don't allow any requests to be passed down
            //
            fdoData->HoldNewRequests = TRUE;
            SD_DebugPrint((1, "Query - stop holding requests...\n"));
            //
            // Then, wait for the existing ones to be finished
            // (since we expexct to give up our resources, we
            // can't allow such requests). First, we need to decrement
            // this very operation (and to keep in mind not to decrement
            // after the switch).
            //
            SD_IoDecrement(fdoData);

            KeWaitForSingleObject(
               &fdoData->StopEvent,
               Executive, // Waiting for reason of a driver
               KernelMode, // Waiting in kernel mode
               FALSE, // No allert
               NULL); // No timeout

            
            IoSkipCurrentIrpStackLocation (Irp);
            status = IoCallDriver (fdoData->NextLowerDriver, Irp);
            goto exit;
        }   else    {
            //
            // The device can't be stopped, complete the request
            //
            IoCompleteRequest(Irp, IO_NO_INCREMENT);
        }
        break;
    
   case IRP_MN_CANCEL_STOP_DEVICE:
        //
        // We need to flush the held IRPs queue, after we've passed the IRP
        // to the next driver. Note that we migth receive a spurious 
        // CANCEL_STOP even if we were never queried for stop (a driver above
        // us failed the query and we have no idea about it).
        //
        //
        // The device is still active: only after a stop
        // we'll mark the device stopped. So assert now that the
        // device is not stopped or removed.
        //
        if (!fdoData->IsStarted) {
            //
            // Nothing to do, just pass the irp down:
            // no need to start the device
            //
            IoSkipCurrentIrpStackLocation (Irp);
            status = IoCallDriver (fdoData->NextLowerDriver, Irp);
            
        } else  {
            ASSERT(!fdoData->IsRemoved);

            //
            // We need to wait for the lower drivers to do their job.
            //
            IoCopyCurrentIrpStackLocationToNext (Irp);
        
            //
            // Clear the event: it will be set in the completion
            // routine.
            //
            KeInitializeEvent(&eventWaitLowerDrivers, 
                              SynchronizationEvent, 
                              FALSE);
        
            IoSetCompletionRoutine (Irp,
                                    SD_DispatchPnpComplete,
                                    &eventWaitLowerDrivers,
                                    TRUE,
                                    TRUE,
                                    TRUE);

            status = IoCallDriver (fdoData->NextLowerDriver, Irp);
        
            SD_DebugPrint((1, "Status before waiting = 0x%x\n", status));
        
            if (STATUS_PENDING == status) {
                KeWaitForSingleObject(
                    &eventWaitLowerDrivers,
                    Executive, // Waiting for reason of a driver
                    KernelMode, // Waiting in kernel mode
                    FALSE, // No allert
                    NULL); // No timeout

                status = Irp->IoStatus.Status;
            }

        
            SD_DebugPrint((1, "Status after waiting = 0x%x\n", status));
        
            fdoData->HoldNewRequests = FALSE;
            SD_DebugPrint((1, "Cancel stop...\n"));
        
            //
            // Process the queued requests
            //
            pSD_ProcessQueuedRequests(fdoData);

            //
            // We must now complete the IRP, since we stopped it in the
            // completetion routine with MORE_PROCESSING_REQUIRED.
            //
            Irp->IoStatus.Status = status;
            Irp->IoStatus.Information = 0;
            IoCompleteRequest (Irp, IO_NO_INCREMENT);
        }   //else

        
        break;


    case IRP_MN_STOP_DEVICE:
        //
        // After the stop IRP has been sent to the lower driver object, the
        // bus may NOT send any more IRPs down that touch the device until 
        // another START has occured.  For this reason we are holding IRPs
        // and waiting for the outstanding requests to be finished when
        // QUERY_STOP is received.
        // IRP_MN_STOP_DEVICE doesn't change anything in this behavior
        // (we continue to hold IRPs until a IRP_MN_START_DEVICE is issued).
        //

        //
        // Set the current stack location to the next stack location and
        // call the next device object.
        //
        //
        // This is the right place to actually give up all the resources used
        // This might include calls to IoDisconnectInterrupt, MmUnmapIoSpace, 
        // etc.
        //
        SD_DebugPrint((1, "Stopping device...\n"));
        //
        // Mark the device not started. We don't have race conditions here, since
        // it's not possible to receive a start and a stop Irp 
        // "in the same time". 
        //
        fdoData->IsStarted = FALSE;
        IoSkipCurrentIrpStackLocation (Irp);
        status = IoCallDriver (fdoData->NextLowerDriver, Irp);
        break;

    case IRP_MN_QUERY_REMOVE_DEVICE:
         //
         // If we can stop the device, we need to set the HoldNewRequestsFlag,
         // so further requests will be queued. 
         // The difference from IRP_MN_QUERY_STOP_DFEVICE is that we will 
         // attempt to process the requests queued before 
         // (it's likely we won't have another chance to do this, since we 
         // expect that the device will be removed).
         // We then start queueing new IRPs in the event of receiving a 
         // IRP_MN_CANCEL_STOP_DEVICE
         // 
         if (!fdoData->IsStarted) {
            //
            // If we aren't started, we'll just pass the irp down
            // A driver has no reason to veto a query when he
            // is stopped already.
            //
            IoSkipCurrentIrpStackLocation (Irp);
            status = IoCallDriver (fdoData->NextLowerDriver, Irp);
            SD_IoDecrement(fdoData);

            return status;
         }
        
         status = pSD_CanRemoveDevice(DeviceObject, Irp);
         Irp->IoStatus.Status = status;
         if (NT_SUCCESS(status)) {
             //
             // First, process the old requests
             //
             SD_DebugPrint((1, "Processing requests\n"));
             
             pSD_ProcessQueuedRequests(fdoData);
             
             //
             // Now prepare to hold the new ones (eventually we might
             // get a IRP_MN_CANCEL_REMOVE_DEVICE) and we need to 
             // process the queued requests then.
             //
             fdoData->HoldNewRequests = TRUE;
             
             SD_DebugPrint((1, "Query - remove holding requests...\n"));
             
             //
             // We need to decrement this very operation (and don't
             // forget to skip the decrement after the switch).
             //
             SD_IoDecrement(fdoData);
             //
             // Wait for all the requests to be completed
             //
             KeWaitForSingleObject(
               &fdoData->StopEvent,
               Executive, // Waiting for reason of a driver
               KernelMode, // Waiting in kernel mode
               FALSE, // No allert
               NULL); // No timeout


             IoSkipCurrentIrpStackLocation (Irp);
             
             status = IoCallDriver (fdoData->NextLowerDriver, Irp);
             goto exit;
         }   else    {
             //
             // The device can't be removed, just complete the request.
             // The status returned by pSD_CanRemoveDevice is already
             // in IoStatus.Status.
             //
             IoCompleteRequest(Irp, IO_NO_INCREMENT);
         }
         break;
     
    case IRP_MN_CANCEL_REMOVE_DEVICE:
         //
         // We need to reset the HoldNewRequests flag, since the device
         // resume its normal activities.
         //
         if (!fdoData->IsStarted) {
            //
            // Nothing to do, just pass the irp down:
            // no need to start the device
            //
            IoSkipCurrentIrpStackLocation (Irp);
            status = IoCallDriver (fdoData->NextLowerDriver, Irp);
            
         } else  {
        
            //
            // Initialize the event
            // 
            KeInitializeEvent(&eventWaitLowerDrivers, 
                              SynchronizationEvent, 
                              FALSE);
        
            IoSetCompletionRoutine (Irp,
                                    SD_DispatchPnpComplete,
                                    &eventWaitLowerDrivers,
                                    TRUE,
                                    TRUE,
                                    TRUE);

            status = IoCallDriver (fdoData->NextLowerDriver, Irp);
        
            if (STATUS_PENDING == status) {
                KeWaitForSingleObject(
                    &eventWaitLowerDrivers,
                    Executive, // Waiting for reason of a driver
                    KernelMode, // Waiting in kernel mode
                    FALSE, // No allert
                    NULL); // No timeout

                status = Irp->IoStatus.Status;
            }

            
        
            fdoData->HoldNewRequests = FALSE;
            SD_DebugPrint((1, "Cancel remove...\n"));
        
            //
            // Process the queued requests
            //
            pSD_ProcessQueuedRequests(fdoData);

        
            //
            // We must now complete the IRP, since we stopped it in the
            // completetion routine with MORE_PROCESSING_REQUIRED.
            //
            Irp->IoStatus.Status = status;
            Irp->IoStatus.Information = 0;
            IoCompleteRequest (Irp, IO_NO_INCREMENT);
         } //else
        break;
        


            
   case IRP_MN_REMOVE_DEVICE:
        //
        // The PlugPlay system has dictacted the removal of this device.  We
        // have no choice but to detach and delete the device object.
        // (If we wanted to express an interest in preventing this removal,
        // we should have failed the query remove IRP).
        //
        // Note! we might receive a remove WITHOUT first receiving a stop.
        //
        // We will no longer receive requests for this device as it has been
        // removed.
        //
        SD_DebugPrint((1, "Removing device...\n"));
        
        //
        // We need to mark the fact that we don't hold requests first, since
        // we asserted earlier that we are holding requests only if
        // we're not removed.
        //
        fdoData->HoldNewRequests = FALSE;
        fdoData->IsStarted = FALSE;
        fdoData->IsRemoved = TRUE;


        //
        // Here we either have completed all the requests in a personal
        // queue when IRP_MN_QUERY_REMOVE was received, or will have to 
        // fail all of them if this is a surprise removal.
        // Note that fdoData->IsRemoved is TRUE, so pSD_ProcessQueuedRequests
        // will simply flush the queue, completing each IRP with
        // STATUS_DELETE_PENDING
        //
        pSD_ProcessQueuedRequests(fdoData);
        
        
        //
        // Send on the remove IRP.
        // We need to send the remove down the stack before we detach,
        // but we don't need to wait for the completion of this operation
        // (and to register a completion routine).
        //

        IoSkipCurrentIrpStackLocation (Irp);
        status = IoCallDriver (fdoData->NextLowerDriver, Irp);



        //
        // We need two decrements here, one for the increment in 
        // SD_PnpDispatch, the other for the 1-biased value of
        // OutstandingIO. Also, we need to wait that all the requests
        // are served.
        // 

        requestCount = SD_IoDecrement (fdoData);

        //
        // The requestCount is a least one here (is 1-biased)
        //
        ASSERT(requestCount > 0);
        
        requestCount = SD_IoDecrement (fdoData);
        
        KeWaitForSingleObject (
                &fdoData->RemoveEvent,
                Executive,
                KernelMode,
                FALSE,
                NULL);
        

        //
        // Detach the FDO from the device stack
        //
        IoDetachDevice (fdoData->NextLowerDriver);

        //
        // Clean up memory
        //
        IoDeleteDevice (fdoData->Self);
        return STATUS_SUCCESS;

    case IRP_MN_QUERY_CAPABILITIES:
        //
        // We will provide here an example of an IRP that is procesed
        // both on its way down and on its way up: there might be no need for
        // a function driver process this Irp (the bus driver will do that).
        // The driver will wait for the lower drivers (the bus driver among 
        // them) to process this IRP, then it processes it again.
        //
        //
        // We will specifically check for UINumber: even if the bus can't 
        // support such a convention, we will still supply the value we want.
        // (the UINumber is the number displayed in the user-interface dialogs:
        // usaully represents a slot number). Note that this is field
        // was chosen for demonstrative purposes only.
        //
        SD_DebugPrint((1, "Query Capabilities, way down...\n"));
        
        deviceCapabilities = stack->Parameters.DeviceCapabilities.Capabilities;
        //
        // Set some values in deviceCapabilities here...
        //
        // .............................................
        //
        deviceCapabilities->UINumber = 1;
        //
        // Prepare to pass the IRP down
        //
        IoCopyCurrentIrpStackLocationToNext (Irp);
        
        KeInitializeEvent(&eventWaitLowerDrivers, 
                          SynchronizationEvent, 
                          FALSE);
        

        IoSetCompletionRoutine (Irp,
                                SD_DispatchPnpComplete,
                                &eventWaitLowerDrivers,
                                TRUE,
                                TRUE,
                                TRUE);

        status = IoCallDriver (fdoData->NextLowerDriver, Irp);
        
        if (STATUS_PENDING == status) {
            KeWaitForSingleObject(
               &eventWaitLowerDrivers,
               Executive, // Waiting for reason of a driver
               KernelMode, // Waiting in kernel mode
               FALSE, // No allert
               NULL); // No timeout

            status = Irp->IoStatus.Status;
        }

        //
        // Lower drivers have finished their operation, so now
        // we can finish ours. We are going to check the UINumber
        // we've set on the way down and reset it if necessary. 
        // This is only an example of processing an IRP both before
        // it's sent to the lower drivers and after it was processed
        // by them.
        //
        SD_DebugPrint((1, "Query Capabilities, way up...\n"));
        if (deviceCapabilities->UINumber != 1) {
                deviceCapabilities->UINumber = 1;
        }
        
        //
        // We must now complete the IRP, since we stopped it in the
        // completetion routine with MORE_PROCESSING_REQUIRED.
        //
        Irp->IoStatus.Status = status;
        Irp->IoStatus.Information = 0;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);
        

        break;

    case IRP_MN_QUERY_DEVICE_RELATIONS:
    case IRP_MN_QUERY_INTERFACE:
    case IRP_MN_QUERY_RESOURCES:
    case IRP_MN_QUERY_RESOURCE_REQUIREMENTS:
    case IRP_MN_READ_CONFIG:
    case IRP_MN_WRITE_CONFIG:
    case IRP_MN_EJECT:
    case IRP_MN_SET_LOCK:
    case IRP_MN_QUERY_ID:
    case IRP_MN_QUERY_PNP_DEVICE_STATE:
    default:
        //
        // The driver might modify the behavior of these IRPS
        // Please see PlugPlay documentation for use of these IRPs.
        //
        IoSkipCurrentIrpStackLocation (Irp);
        status = IoCallDriver (fdoData->NextLowerDriver, Irp);
        break;
    }
    //
    // We need to decrement only in some cases.
    // For the "query" minor codes we decremented inside the switch.
    //
    requestCount = SD_IoDecrement(fdoData);
    
    exit:
    //
    // Return without decrement
    //
    return status;
}


NTSTATUS
SD_DispatchPnpComplete (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID Context
    )
/*++

Routine Description:
    The pnp IRP was completed by the lower-level drivers.
    Signal this to whoever registerd us.

Arguments:
    
   DeviceObject - pointer to a device object.

   Irp - pointer to an I/O Request Packet.
   
   Context - pointer to the event to be signaled.
Return Value:

    NT status code
    
    
--*/
{
    PIO_STACK_LOCATION  stack;
    PKEVENT             event = (PKEVENT)Context;
    NTSTATUS            status;

    UNREFERENCED_PARAMETER (DeviceObject);

    status = STATUS_SUCCESS;
    stack = IoGetCurrentIrpStackLocation (Irp);

    if (Irp->PendingReturned) {
        IoMarkIrpPending( Irp );
    }

    switch (stack->MajorFunction) {
    case IRP_MJ_PNP:

        switch (stack->MinorFunction) {
        case IRP_MN_START_DEVICE:
        case IRP_MN_QUERY_CAPABILITIES:
        case IRP_MN_CANCEL_STOP_DEVICE:
        case IRP_MN_CANCEL_REMOVE_DEVICE:
            //
            // Add other codes that need processing "on the way up" here.
            //

            KeSetEvent (event, 0, FALSE);

            //
            // Take the IRP back so that we can continue using it during
            // the dispatch routine.
            // NB: The dispatch routine will have to call IoCompleteRequest
            //
            return STATUS_MORE_PROCESSING_REQUIRED;
            
        default:
            break;
        }
        break;

    case IRP_MJ_POWER:
    default:
        break;
    }
    return status;
}





NTSTATUS
SD_CreateClose (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )

/*++

Routine Description:

    The dispatch routine for several Irps that require special treatment,
    because
    a) they shouldn't be queued if the device is stopped, but sent to the lower
    drivers.
    b) they shouldn't be failed with STATUS_DELETE_PENDING if a surprise removal
    occured.
    
    
Arguments:

   DeviceObject - pointer to a device object.

   Irp - pointer to an I/O Request Packet.

Return Value:

   NT status code

--*/
{
    PSD_FDO_DATA    fdoData;
    NTSTATUS        status;

    LONG            requestCount;

    //
    // We will just pass this IRP down, no matter what the circumstances... 
    //
    fdoData = (PSD_FDO_DATA) DeviceObject->DeviceExtension;
    
    requestCount = SD_IoIncrement (fdoData);
    
    IoSkipCurrentIrpStackLocation (Irp);
    status = IoCallDriver (fdoData->NextLowerDriver, Irp);
        
    requestCount = SD_IoDecrement(fdoData);

    SD_DebugPrint((1, "CreateClose called, reqCount = %d\n", 
                requestCount));
    
    return status;
}






NTSTATUS
SD_StartDevice (
    IN PSD_FDO_DATA     FdoData,
    IN PIRP             Irp
    )
/*++

Routine Description:
    
    Performs whatever initialization is needed for a device when it is 
    started.
    

Arguments:
    
   DeviceObject - pointer to a device object.

   Irp - pointer to an I/O Request Packet.
   
   Context - pointer to a SD_FDO_DATA structure (contains the event to be 
             signaled)

Return Value:

    NT status code
    

--*/
{
    NTSTATUS    status = STATUS_SUCCESS;
    
    PAGED_CODE();

    //
    // We need to check that we haven't received a surprise removal
    //
    if (FdoData->IsRemoved) {
        //
        // Some kind of surprise removal arrived. We will fail the IRP
        // The dispatch routine that called us will take care of 
        // completing the IRP.
        //
        status = STATUS_DELETE_PENDING;
        return status;
    }
    //
    // Mark the device as active and not holding IRPs
    //
    FdoData->IsStarted = TRUE;
    FdoData->HoldNewRequests = FALSE;
    //
    // Do whatever initialization needed when starting the device: 
    // gather information about it,  update the registry, etc.
    // At this point, the lower level drivers completed the IRP
    //
    
    //
    // The last thing to do is to process the hold IRPs queue.
    //
    pSD_ProcessQueuedRequests(FdoData);

    
    return status;

}


NTSTATUS
SD_Cleanup (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )

/*++

Routine Description:

    The dispatch routine for cleanup: we need to walk the Irp queue and
    to cancel all the requests for which the file object is the same with
    the one in the Irp.
    
    
Arguments:

   DeviceObject - pointer to a device object.

   Irp - pointer to an I/O Request Packet.

Return Value:

   NT status code

--*/
{
    PIRP currentIrp;
	PIRP requeueIrp;
 
	PIO_STACK_LOCATION cleanupStack;
    
    PIO_STACK_LOCATION currentStack;
    LONG    requestCount;


	LIST_ENTRY cancelList;
	LIST_ENTRY requeueList;

	PLIST_ENTRY headOfList;

    PSD_FDO_DATA    fdoData;
	
    fdoData = (PSD_FDO_DATA) DeviceObject->DeviceExtension;
    
    requestCount = SD_IoIncrement (fdoData);
    cleanupStack = IoGetCurrentIrpStackLocation (Irp);
        
    //
	// Prepare the two work lists
	//
	InitializeListHead( &cancelList );
	InitializeListHead( &requeueList );

	//
	// Loop until every IRP and decide where to put it
    //
    while ((headOfList = 
                ExInterlockedRemoveHeadList(&fdoData->NewRequestsQueue,
                                            &fdoData->QueueLock)) != NULL) {
        
        currentIrp = CONTAINING_RECORD(headOfList,
                                       IRP,
                                       Tail.Overlay.ListEntry);
        //
        // No need to acquire the cancel spin lock, since we manage out own
        // Irp queue.
        //
        IoSetCancelRoutine(currentIrp, 
                           NULL);
        
        currentStack = IoGetCurrentIrpStackLocation (currentIrp);
        
        if (currentStack->FileObject == cleanupStack->FileObject) {
            IoSetCancelRoutine(currentIrp,
                               NULL);
            InsertTailList(&cancelList,
                           &currentIrp->Tail.Overlay.ListEntry);
        } else  {
            InsertTailList(&requeueList,
                           &currentIrp->Tail.Overlay.ListEntry);
        
        }
    }
    //
    // Requeue the list
    //
    while (!IsListEmpty(&requeueList)) {
        headOfList = RemoveHeadList(&requeueList);
        currentIrp = CONTAINING_RECORD(headOfList,
                                       IRP,
                                       Tail.Overlay.ListEntry);
        //
        // Put it back in the list
        //
        ExInterlockedInsertTailList(&fdoData->NewRequestsQueue, 
                                    &currentIrp->Tail.Overlay.ListEntry,
                                    &fdoData->QueueLock); 
        
    }
    //
    // Complete the cancelled requests
    //
    while (!IsListEmpty(&cancelList)) {
        headOfList = RemoveHeadList(&cancelList);
        
        currentIrp = CONTAINING_RECORD(headOfList,
                                       IRP,
                                       Tail.Overlay.ListEntry);
        currentIrp->IoStatus.Status      = STATUS_CANCELLED;
        currentIrp->IoStatus.Information = 0;

        IoCompleteRequest(currentIrp, IO_NO_INCREMENT);


    }

    
    
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    requestCount = SD_IoDecrement (fdoData);
    
    SD_DebugPrint((1, "Cleanup called, reqCount = %d\n", 
                requestCount));
     
    return STATUS_SUCCESS;
}




VOID
SD_Unload(
    IN PDRIVER_OBJECT DriverObject
    )
/*++

Routine Description:

    Free all the allocated resources, etc.

Arguments:

    DriverObject - pointer to a driver object.

Return Value:

    VOID.

--*/
{
    PAGED_CODE ();

    //
    // The device object(s) should be NULL now
    // (since we unload, all the devices objects associated with this
    // driver must be deleted.
    //
    ASSERT(DriverObject->DeviceObject == NULL);
    
    //
    // We should not be unloaded until all the devices we control 
    // have been removed from our queue.  
    //
    SD_DebugPrint ((1, "unload\n"));

    return;
}



NTSTATUS
pSD_QueueRequest    (
    IN OUT PSD_FDO_DATA FdoData,
    IN PIRP Irp
    )   

/*++

Routine Description:

    Queues the Irp in the device queue. This routine will be called whenever
    the device receives IRP_MN_QUERY_STOP_DEVICE or IRP_MN_QUERY_REMOVE_DEVICE

Arguments:

    FdoData - pointer to the device's extension.
    
    Irp - the request to be queued.

Return Value:

    NT status code.

--*/
{
    
    KIRQL               oldIrql;
    PIO_STACK_LOCATION  stack;
            

    stack = IoGetCurrentIrpStackLocation(Irp);
    //
    // Check if we are allowed to queue requests.
    //
    ASSERT(FdoData->HoldNewRequests);
    //
    // Preparing for dealing with cancelling stuff.
    // We don't know how long the irp will be in the 
    // queue.  So we need to handle cancel. 
    // Since we use out own queue, we don't need to use
    // the cancel spin lock.
    // 
    if (Irp->Cancel) { 
            //
            // Already canceled
            //
            Irp->IoStatus.Status      = STATUS_CANCELLED; 
            Irp->IoStatus.Information = 0; 
            
            SD_DebugPrint((1, 
                          "Irp %x not queued because it has been cancelled\n",
                          Irp)); 
                
            IoCompleteRequest(Irp, IO_NO_INCREMENT); 
 
            return STATUS_CANCELLED; 
 
     } else { 
            
            //
            // Queue the Irp and set a cancel routine
            //
            Irp->IoStatus.Status = STATUS_PENDING; 
            
            IoMarkIrpPending(Irp); 
 
            ExInterlockedInsertTailList(&FdoData->NewRequestsQueue, 
                                        &Irp->Tail.Overlay.ListEntry,
                                        &FdoData->QueueLock); 
           
            //
            // We need to print some more info about this guy
            //
            
            SD_DebugPrint((1, "Irp %x (major 0x%x, minor 0x%x) queued : ",
                          Irp,
                          stack->MajorFunction,
                          stack->MinorFunction));
            
            IoSetCancelRoutine(Irp,
                               SD_CancelQueued); 
 
            
 
            
      }

      return STATUS_SUCCESS;
 

}
    

    
VOID
pSD_ProcessQueuedRequests    (
    IN OUT PSD_FDO_DATA FdoData
    )   

/*++

Routine Description:

    Removes an dprocesses the entries in the queue. If this routine is  called 
    when processing IRP_MN_CANCEL_STOP_DEVICE, IRP_MN_CANCEL_REMOVE_DEVICE 
    or IRP_MN_START_DEVICE, the requests are passed to the next lower driver.
    If the routine is called when IRP_MN_REMOVE_DEVICE is received, the IRPs 
    are completed with STATUS_DELETE_PENDING.
    

Arguments:

    FdoData - pointer to the device's extension (where is the held IRPs queue).
    
    
Return Value:

    VOID.

--*/
{
    
    KIRQL               oldIrql;
    PLIST_ENTRY         headOfList;
    PIRP                currentIrp;
    PIO_STACK_LOCATION  stack;
    LONG                requestCount;
    
    //
    // We need to dequeue all the entries in the queue, to reset the cancel 
    // routine for each of them and then to process then:
    // - if the device is active, we will send them down
    // - else we will complete them with STATUS_DELETE_PENDING
    // (it is a surprise removal and we need to dispose the queue)
    //
    while ((headOfList = 
                ExInterlockedRemoveHeadList(&FdoData->NewRequestsQueue,
                                            &FdoData->QueueLock)) != NULL) {
        
        currentIrp = CONTAINING_RECORD(headOfList,
                                       IRP,
                                       Tail.Overlay.ListEntry);
        IoSetCancelRoutine(currentIrp, 
                           NULL);
        
        stack = IoGetCurrentIrpStackLocation (currentIrp);
        
        requestCount = SD_IoIncrement (FdoData);
        
        if (FdoData->IsRemoved) {
            //
            // The device was removed, we need to fail the request
            //
            currentIrp->IoStatus.Information = 0;
            currentIrp->IoStatus.Status = STATUS_DELETE_PENDING;
            IoCompleteRequest (currentIrp, IO_NO_INCREMENT);

        } else {
            IoSkipCurrentIrpStackLocation (currentIrp);
            IoCallDriver (FdoData->NextLowerDriver, currentIrp);
        }
        
        requestCount = SD_IoDecrement(FdoData);
        
        
    }

    return;

}
    


VOID
SD_CancelQueued (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp
    )

/*++

Routine Description:

    The cancel routine. Will remove the IRP from the queue and will complete it.
    The cancel spin lock is already acquired when this routine is called.
    

Arguments:

    DeviceObject - pointer to the device object.
    
    Irp - pointer to the IRP to be cancelled.
    
    
Return Value:

    VOID.

--*/
{
    PSD_FDO_DATA fdoData = DeviceObject->DeviceExtension; 
    PIO_STACK_LOCATION irpSp = IoGetCurrentIrpStackLocation(Irp); 
 
    Irp->IoStatus.Status = STATUS_CANCELLED; 
    Irp->IoStatus.Information = 0; 
 
    RemoveEntryList(&Irp->Tail.Overlay.ListEntry); 
 
    
    IoReleaseCancelSpinLock(Irp->CancelIrql); 
 
    SD_DebugPrint((1, "SD_CancelQueued called for %x\n", Irp));
    
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    return;
 
} 
  

NTSTATUS
pSD_CanStopDevice    (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP           Irp
    )

/*++

Routine Description:

    This routine determines is the device can be safely stopped. In our 
    particular case, we'll assume we can always stop the device.
    A device might fail the request if it doesn't have a queue for the
    requests it might come or if it was notified that it is in the paging
    path. 
    
    

Arguments:

    DeviceObject - pointer to the device object.
    
    Irp - pointer to the current IRP.
    
    
Return Value:

    STATUS_SUCCESS if the device can be safely stopped, an appropriate 
    NT Status if not.

--*/
{
    UNREFERENCED_PARAMETER(DeviceObject);
    UNREFERENCED_PARAMETER(Irp);

    return STATUS_SUCCESS;
}

NTSTATUS
pSD_CanRemoveDevice    (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP           Irp
    )

/*++

Routine Description:

    This routine determines is the device can be safely removed. In our 
    particular case, we'll assume we can always remove the device.
    A device might be removed if, for example, it has openend handles or
    removing the device could result in losing data (plus the reasons 
    mentioned at pSD_CanStopDevice).

Arguments:

    DeviceObject - pointer to the device object.
    
    Irp - pointer to the current IRP.
    
    
Return Value:

    STATUS_SUCCESS if the device can be safely removed, an appropriate 
    NT Status if not.

--*/
{
    UNREFERENCED_PARAMETER(DeviceObject);
    UNREFERENCED_PARAMETER(Irp);

    return STATUS_SUCCESS;
}

LONG
SD_IoIncrement    (
    IN  OUT PSD_FDO_DATA   FdoData
    )   

/*++

Routine Description:

    This routine increments the number of requests the device receives
    

Arguments:

    DeviceObject - pointer to the device object.
    
Return Value:

    The value of OutstandingIO field in the device extension.


--*/

{

    LONG            result;
    
    result = InterlockedIncrement(&FdoData->OutstandingIO);
    
    ASSERT(result > 0);
    //
    // Need to clear StopEvent (when OutstandingIO bumps from 1 to 2) 
    //
    if (result == 2) {
        //
        // We need to clear the event
        //
        KeClearEvent(&FdoData->StopEvent);
    }

    return result;
}

LONG
SD_IoDecrement    (
    IN  OUT PSD_FDO_DATA  FdoData
    )   

/*++

Routine Description:

    This routine increments the number of requests the device receives
    

Arguments:

    DeviceObject - pointer to the device object.
    
Return Value:

    The value of OutstandingIO field in the device extension.


--*/
{

    LONG            result;
    
    result = InterlockedDecrement(&FdoData->OutstandingIO);
    
    ASSERT(result >= 0);

    if (result == 1) {
        //
        // Set the stop event. Note that when this happens
        // (i.e. a transition from 2 to 1), the type of requests we 
        // want to be processed are already hold instead of being 
        // passed away, so that we can't "miss" a request that
        // will appear between the decrement and the moment when
        // the value is actually used.
        //
 
        KeSetEvent (&FdoData->StopEvent, 
                    IO_NO_INCREMENT, 
                    FALSE);
        
    }
    
    if (result == 0) {
        //
        // The count is 1-biased, so it can be zero only if an 
        // extra decrement is done when a remove Irp is received 
        //
        ASSERT(FdoData->IsRemoved);
        //
        // Set the remove event, so the device object can be deleted
        //
        KeSetEvent (&FdoData->RemoveEvent, 
                    IO_NO_INCREMENT, 
                    FALSE);
        
    }


    return result;
}
