/*++

Copyright (c) 1990, 1991, 1992, 1993  Microsoft Corporation

Module Name:

    moudep.c

Abstract:

    The initialization and hardware-dependent portions of
    the Intel i8042 port driver which are specific to
    the auxiliary (PS/2 mouse) device.

Environment:

    Kernel mode only.

Notes:

    NOTES:  (Future/outstanding issues)

    - Powerfail not implemented.

    - Consolidate duplicate code, where possible and appropriate.

Revision History:

--*/

#include "stdarg.h"
#include "stdio.h"
#include "string.h"
#include "ntddk.h"
#include "i8042prt.h"
#include "i8042log.h"

//
// Use the alloc_text pragma to specify the driver initialization routines
// (they can be paged out).
//

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT,I8xMouseConfiguration)
#pragma alloc_text(INIT,I8xMousePeripheralCallout)
#pragma alloc_text(INIT,I8xInitializeMouse)
#pragma alloc_text(INIT,I8xFindWheelMouse)
#endif


BOOLEAN
I8042MouseInterruptService(
    IN PKINTERRUPT Interrupt,
    IN PVOID Context
    )

/*++

Routine Description:

    This is the interrupt service routine for the mouse device.

Arguments:

    Interrupt - A pointer to the interrupt object for this interrupt.

    Context - A pointer to the device object.

Return Value:

    Returns TRUE if the interrupt was expected (and therefore processed);
    otherwise, FALSE is returned.

--*/

{
    PDEVICE_EXTENSION deviceExtension;
    PDEVICE_OBJECT deviceObject;
    LARGE_INTEGER tickDelta, newTick;
    UCHAR previousButtons;
    UCHAR previousSignAndOverflow;
    UCHAR byte;

    UNREFERENCED_PARAMETER(Interrupt);

    I8xPrint((2, "%s-%s: %s\n", pDriverName, pFncMouseInterrupt, pEnter));

    deviceObject = (PDEVICE_OBJECT) Context;
    deviceExtension = (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

    //
    // Verify that this device really interrupted.  Check the status
    // register.  The Output Buffer Full bit should be set, and the
    // Auxiliary Device Output Buffer Full bit should be set.
    //
    if ((I8X_GET_STATUS_BYTE(deviceExtension->DeviceRegisters[CommandPort])
            & (OUTPUT_BUFFER_FULL|MOUSE_OUTPUT_BUFFER_FULL))
            != (OUTPUT_BUFFER_FULL|MOUSE_OUTPUT_BUFFER_FULL)) {

        //
        // Stall and then try again.  The Olivetti MIPS machine
        // sometimes gets a mouse interrupt before the status
        // register is set.
        //
        KeStallExecutionProcessor(10);
        if ((I8X_GET_STATUS_BYTE(deviceExtension->DeviceRegisters[CommandPort])
                & (OUTPUT_BUFFER_FULL|MOUSE_OUTPUT_BUFFER_FULL))
                != (OUTPUT_BUFFER_FULL|MOUSE_OUTPUT_BUFFER_FULL)) {

            //
            // Not our interrupt.
            //
            I8xPrint((
                1,
                "%s-%s: not our interrupt!\n",
                pDriverName, pFncMouseInterrupt
                ));
            return(FALSE);
        }
    }

    //
    // Read the byte from the i8042 data port.
    //

    I8xGetByteAsynchronous(
        (CCHAR) MouseDeviceType,
        deviceExtension,
        &byte
        );

    I8xPrint((
        2,
        "%s-%s: byte 0x%x\n",
        pDriverName, pFncMouseInterrupt,
        byte
        ));

    //
    // Watch the data stream for a reset completion (0xaa) followed by the
    // device id
    //
    // this pattern can appear as part of a normal data packet as well. This
    // code assumes that sending an enable to an already enabled mouse will:
    //  * not hang the mouse
    //  * abort the current packet and return an ACK.
    //

    if (deviceExtension->MouseExtension.LastByteReceived == MOUSE_COMPLETE &&
        byte == 0x00) {

        //
        // Tell the 8042 port to fetch the device ID of the aux device
        // We do this async so that we don't spin at IRQ1!!!
        //
        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

        //
        // This is part of the substate system for handling a (possible)
        // mouse reset.
        //
        deviceExtension->MouseExtension.InputState = MouseResetting;
        deviceExtension->MouseExtension.InputResetSubState =
            ExpectingGetDeviceIdACK;

        //
        // We don't want to execute any more of the ISR code, so lets just
        // do a few things and then return now
        //
        deviceExtension->MouseExtension.LastByteReceived = byte;

        //
        // Done
        //
        return (TRUE);
    }

    //
    // Remember what the last byte we got was
    //
    deviceExtension->MouseExtension.LastByteReceived = byte;

    //
    // Take the appropriate action, depending on the current state.
    // When the state is Idle, we expect to receive mouse button
    // data.  When the state is XMovement, we expect to receive mouse
    // motion in the X direction data.  When the state is YMovement,
    // we expect to receive mouse motion in the Y direction data.  Once
    // the Y motion data has been received, the data is queued to the
    // mouse InputData queue, the mouse ISR DPC is requested, and the
    // state returns to Idle.
    //
    KeQueryTickCount(&newTick);
    tickDelta.QuadPart =
            newTick.QuadPart -
            deviceExtension->MouseExtension.PreviousTick.QuadPart;

    if ((deviceExtension->MouseExtension.InputState != MouseIdle)
           && (deviceExtension->MouseExtension.InputState != MouseExpectingACK)
           && (deviceExtension->MouseExtension.InputState != MouseResetting)
           && ((tickDelta.LowPart >= deviceExtension->MouseExtension.SynchTickCount)
               || (tickDelta.HighPart != 0))) {

        //
        // It has been a long time since we got a byte of
        // the data packet.  Assume that we are now receiving
        // the first byte of a new packet, and discard any
        // partially received packet.
        //
        // N.B.  We assume that SynchTickCount is ULONG, and avoid
        //       a LARGE_INTEGER compare with tickDelta...
        //

        I8xPrint((
            1,
            "%s-%s: State was %d, synching\n",
            pDriverName, pFncMouseInterrupt,
            deviceExtension->MouseExtension.InputState
            ));

        //
        // Something want totally bad with the device. Lets play it safe and
        // reset the device. Note: this code is meant to handle cases where
        // some intermediate box resets the mouse and doesn't tell us about it.
        // This avoid problems with trying to detect this code since there
        // isn't a fool proof way to do it
        //
        deviceExtension->MouseExtension.InputState = MouseResetting;
        deviceExtension->MouseExtension.InputResetSubState = ExpectingReset;

        //
        // Send a reset to the device
        //
        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
        I8X_MOUSE_COMMAND( deviceExtension, MOUSE_RESET );

        //
        // We can let this go since the state machine will just eat the
        // byte that was provided
        //
    }

    deviceExtension->MouseExtension.PreviousTick = newTick;

    switch(deviceExtension->MouseExtension.InputState) {

        //
        // The mouse interrupted with a status byte.  The status byte
        // contains information on the mouse button state along with
        // the sign and overflow bits for the (yet-to-be-received)
        // X and Y motion bytes.
        //

        case MouseIdle: {

            I8xPrint((
                3,
                "%s-%s: mouse status byte\n",
                pDriverName, pFncMouseInterrupt
                ));

            //
            // This is a sanity check test. It is required because some people
            // in industry persist in their notion that you can reset a mouse
            // device any time you want and not let the OS know anything about
            // it. This results in our nice little wheel mouse (which is a
            // 4 byte packet engine) suddenly only dumping 3 byte packets.
            //
            if ((deviceExtension->HardwarePresent & WHEELMOUSE_HARDWARE_PRESENT)
                && (byte & 0xC8) != 8 ) { // Guaranteed True for Megallan

                //
                // Okay, we are getting a really bad packet for the idle
                // state. The best bet is to issue a mouse reset request
                // and hope we can recover
                //
                deviceExtension->MouseExtension.InputState = MouseResetting;
                deviceExtension->MouseExtension.InputResetSubState = ExpectingReset;

                //
                // Send a reset to the device
                //
                I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                I8X_MOUSE_COMMAND( deviceExtension, MOUSE_RESET );

                //
                // done
                //
                break;
            }

            //
            // Update CurrentInput with button transition data.
            // I.e., set a button up/down bit in the Buttons field if
            // the state of a given button has changed since we
            // received the last packet.
            //

            previousButtons =
                deviceExtension->MouseExtension.PreviousButtons;

            deviceExtension->MouseExtension.CurrentInput.ButtonFlags = 0;
            deviceExtension->MouseExtension.CurrentInput.ButtonData = 0;

            if ((!(previousButtons & LEFT_BUTTON_DOWN))
                   &&  (byte & LEFT_BUTTON_DOWN)) {
                deviceExtension->MouseExtension.CurrentInput.ButtonFlags |=
                    MOUSE_LEFT_BUTTON_DOWN;
            } else
            if ((previousButtons & LEFT_BUTTON_DOWN)
                   &&  !(byte & LEFT_BUTTON_DOWN)) {
                deviceExtension->MouseExtension.CurrentInput.ButtonFlags |=
                    MOUSE_LEFT_BUTTON_UP;
            }
            if ((!(previousButtons & RIGHT_BUTTON_DOWN))
                   &&  (byte & RIGHT_BUTTON_DOWN)) {
                deviceExtension->MouseExtension.CurrentInput.ButtonFlags |=
                    MOUSE_RIGHT_BUTTON_DOWN;
            } else
            if ((previousButtons & RIGHT_BUTTON_DOWN)
                   &&  !(byte & RIGHT_BUTTON_DOWN)) {
                deviceExtension->MouseExtension.CurrentInput.ButtonFlags |=
                    MOUSE_RIGHT_BUTTON_UP;
            }
            if ((!(previousButtons & MIDDLE_BUTTON_DOWN))
                   &&  (byte & MIDDLE_BUTTON_DOWN)) {
                deviceExtension->MouseExtension.CurrentInput.ButtonFlags |=
                    MOUSE_MIDDLE_BUTTON_DOWN;
            } else
            if ((previousButtons & MIDDLE_BUTTON_DOWN)
                   &&  !(byte & MIDDLE_BUTTON_DOWN)) {
                deviceExtension->MouseExtension.CurrentInput.ButtonFlags |=
                    MOUSE_MIDDLE_BUTTON_UP;
            }

            //
            // Save the button state for comparison the next time around.
            //

            deviceExtension->MouseExtension.PreviousButtons =
                byte & (RIGHT_BUTTON_DOWN|MIDDLE_BUTTON_DOWN|LEFT_BUTTON_DOWN);

            //
            // Save the sign and overflow information from the current byte.
            //

            deviceExtension->MouseExtension.CurrentSignAndOverflow =
                (UCHAR) (byte & MOUSE_SIGN_OVERFLOW_MASK);

            //
            // Update to the next state.
            //

            deviceExtension->MouseExtension.InputState = XMovement;

            break;
        }

        //
        // The mouse interrupted with the X motion byte.  Apply
        // the sign and overflow bits from the mouse status byte received
        // previously.  Attempt to correct for bogus changes in sign
        // that occur with large, rapid mouse movements.
        //

        case XMovement: {

            I8xPrint((
                3,
                "%s-%s: mouse LastX byte\n",
                pDriverName, pFncMouseInterrupt
                ));

            //
            // Update CurrentInput with the X motion data.
            //

            if (deviceExtension->MouseExtension.CurrentSignAndOverflow
                & X_OVERFLOW) {

                //
                // Handle overflow in the X direction.  If the previous
                // mouse movement overflowed too, ensure that the current
                // overflow is in the same direction (i.e., that the sign
                // is the same as it was for the previous event).  We do this
                // to correct for hardware problems -- it should not be possible
                // to overflow in one direction and then immediately overflow
                // in the opposite direction.
                //

                previousSignAndOverflow =
                    deviceExtension->MouseExtension.PreviousSignAndOverflow;
                if (previousSignAndOverflow & X_OVERFLOW) {
                    if ((previousSignAndOverflow & X_DATA_SIGN) !=
                        (deviceExtension->MouseExtension.CurrentSignAndOverflow
                         & X_DATA_SIGN)) {
                        deviceExtension->MouseExtension.CurrentSignAndOverflow
                            ^= X_DATA_SIGN;
                    }
                }

                if (deviceExtension->MouseExtension.CurrentSignAndOverflow &
                    X_DATA_SIGN)
                    deviceExtension->MouseExtension.CurrentInput.LastX =
                        (LONG) MOUSE_MAXIMUM_NEGATIVE_DELTA;
                else
                    deviceExtension->MouseExtension.CurrentInput.LastX =
                        (LONG) MOUSE_MAXIMUM_POSITIVE_DELTA;

            } else {

                //
                // No overflow.  Just store the data, correcting for the
                // sign if necessary.
                //

                deviceExtension->MouseExtension.CurrentInput.LastX =
                    (ULONG) byte;
                if (deviceExtension->MouseExtension.CurrentSignAndOverflow &
                    X_DATA_SIGN)
                    deviceExtension->MouseExtension.CurrentInput.LastX |=
                        MOUSE_MAXIMUM_NEGATIVE_DELTA;
            }

            //
            // Update to the next state.
            //

            deviceExtension->MouseExtension.InputState = YMovement;

            break;
        }

        //
        // The mouse interrupted with the Y motion byte.  Apply
        // the sign and overflow bits from the mouse status byte received
        // previously.  [Attempt to correct for bogus changes in sign
        // that occur with large, rapid mouse movements.]  Write the
        // data to the mouse InputData queue, and queue the mouse ISR DPC
        // to complete the interrupt processing.
        //

        case YMovement: {

            I8xPrint((
                3,
                "%s-%s: mouse LastY byte\n",
                pDriverName, pFncMouseInterrupt
                ));

            //
            // Update CurrentInput with the Y motion data.
            //

            if (deviceExtension->MouseExtension.CurrentSignAndOverflow
                & Y_OVERFLOW) {

                //
                // Handle overflow in the Y direction.  If the previous
                // mouse movement overflowed too, ensure that the current
                // overflow is in the same direction (i.e., that the sign
                // is the same as it was for the previous event).  We do this
                // to correct for hardware problems -- it should not be possible
                // to overflow in one direction and then immediately overflow
                // in the opposite direction.
                //

                previousSignAndOverflow =
                    deviceExtension->MouseExtension.PreviousSignAndOverflow;
                if (previousSignAndOverflow & Y_OVERFLOW) {
                    if ((previousSignAndOverflow & Y_DATA_SIGN) !=
                        (deviceExtension->MouseExtension.CurrentSignAndOverflow
                         & Y_DATA_SIGN)) {
                        deviceExtension->MouseExtension.CurrentSignAndOverflow
                            ^= Y_DATA_SIGN;
                    }
                }

                if (deviceExtension->MouseExtension.CurrentSignAndOverflow &
                    Y_DATA_SIGN)
                    deviceExtension->MouseExtension.CurrentInput.LastY =
                        (LONG) MOUSE_MAXIMUM_POSITIVE_DELTA;
                else
                    deviceExtension->MouseExtension.CurrentInput.LastY =
                        (LONG) MOUSE_MAXIMUM_NEGATIVE_DELTA;

            } else {

                //
                // No overflow.  Just store the data, correcting for the
                // sign if necessary.
                //

                deviceExtension->MouseExtension.CurrentInput.LastY =
                    (ULONG) byte;
                if (deviceExtension->MouseExtension.CurrentSignAndOverflow &
                    Y_DATA_SIGN)
                    deviceExtension->MouseExtension.CurrentInput.LastY |=
                        MOUSE_MAXIMUM_NEGATIVE_DELTA;

                 //
                 // Negate the LastY value (the hardware reports positive
                 // motion in the direction that we consider negative).
                 //

                 deviceExtension->MouseExtension.CurrentInput.LastY =
                     -deviceExtension->MouseExtension.CurrentInput.LastY;

            }

            //
            // Update our notion of the previous sign and overflow bits for
            // the start of the next mouse input sequence.
            //

            deviceExtension->MouseExtension.PreviousSignAndOverflow =
                deviceExtension->MouseExtension.CurrentSignAndOverflow;

            //
            // Choose the next state.  The WheelMouse has an extra byte of data
            // for us
            //

            if(deviceExtension->HardwarePresent & WHEELMOUSE_HARDWARE_PRESENT) {

                deviceExtension->MouseExtension.InputState = ZMovement;

            } else {

                I8xQueueCurrentInput(deviceObject);
                deviceExtension->MouseExtension.InputState = MouseIdle;

            }
            break;
        }

        case ZMovement: {

            I8xPrint((
                3,
                "%s-%s: mouse LastZ byte\n",
                pDriverName, pFncMouseInterrupt
                ));

            //
            // This code is here to handle the cases were mouse resets were
            // not notified to the OS. Uncomment this if you *really* want it,
            // but remember that it could possibly reset the mouse when it
            // shouldn't have...
            //
#if 1
            if ( (byte & 0xf8) != 0 && (byte & 0xf8) != 0xf8 ) {

                //
                // for some reason, the byte was not sign extanded,
                // or has a value > 7, which we assume cannot be
                // possible giving our equipment. So the packet
                // *must* be bogus...
                //

                deviceExtension->MouseExtension.InputState = MouseResetting;
                deviceExtension->MouseExtension.InputResetSubState = ExpectingReset;

                //
                // Send a reset to the device
                //
                I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                I8X_MOUSE_COMMAND( deviceExtension, MOUSE_RESET );

                //
                // Done
                //
                break;
            }
#endif
            //
            // Check to see if we got any z data
            // If there were any changes in the button state, ignore the
            // z data
            //
            if (byte) {

                //
                // The value returned should be -120 * the value reported
                //
                deviceExtension->MouseExtension.CurrentInput.ButtonData = (-120) * ((CHAR) byte);
                deviceExtension->MouseExtension.CurrentInput.ButtonFlags |= MOUSE_WHEEL;

            }

            //
            // Pack the data on to the class driver
            //

            I8xQueueCurrentInput(deviceObject);

            //
            // Reset the state
            //

            deviceExtension->MouseExtension.InputState = MouseIdle;

            break;
        }

        case MouseExpectingACK: {

            //
            // This is a special case.  We hit this on one of the very
            // first mouse interrupts following the IoConnectInterrupt.
            // The interrupt is caused when we enable mouse transmissions
            // via I8xMouseEnableTransmission() -- the hardware returns
            // an ACK.  Just toss this byte away, and set the input state
            // to coincide with the start of a new mouse data packet.
            //

            I8xPrint((
                1,
                pDumpExpectingAck,
                pDriverName, pFncMouseInterrupt,
                (ULONG) ACKNOWLEDGE,
                (ULONG) byte
                ));

            if (byte == (UCHAR) ACKNOWLEDGE) {

                deviceExtension->MouseExtension.InputState = MouseIdle;

            } else if (byte == (UCHAR) RESEND) {

                //
                // Resend the "Enable Mouse Transmission" sequence.
                //
                // NOTE: This is a hack for the Olivetti MIPS machine,
                // which sends a resend response if a key is held down
                // while we're attempting the I8xMouseEnableTransmission.
                //
                I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                I8X_MOUSE_COMMAND( deviceExtension, ENABLE_MOUSE_TRANSMISSION );

            }

            break;
        }

        case MouseResetting: {

            I8xPrint((
                1,
                "%s-%s: state (%d) substate (%2d)\n",
                pDriverName, pFncMouseInterrupt,
                deviceExtension->MouseExtension.InputState,
                deviceExtension->MouseExtension.InputResetSubState
                ));

            //
            // We enter the reset substate machine
            //
            switch (deviceExtension->MouseExtension.InputResetSubState) {
                //
                // This state (ExpectingReset) and the next one (Expecting
                // ResetId) are only called if we have to issue a reset
                // from within the substate machine.
                //
                case ExpectingReset: {

                    //
                    // Check to see if we got a reset character. If not, then
                    // we will *ignore* it. Note that this means that if we
                    // dropped this char, then we could infinite loop in
                    // this routine.
                    //
                    if (byte != MOUSE_COMPLETE) {

                        break;

                    }

                    //
                    // Now check to see how many times we have gone
                    // through this routine without exiting the
                    // MouseResetting SubState
                    //
                    if (deviceExtension->MouseExtension.ResetCount > MOUSE_NUMBER_OF_RESETS) {

                        //
                        // Let people know that the reset failed
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_MOU_RESET_RESPONSE_FAILED
                            );

                        //
                        // HACK: To prevent the reset detection code from
                        // restarting us all over again (which don't want
                        // otherwise we would not be here, we need to fool
                        // the detection into thinking that the last character
                        // wasn't MOUSE_COMPLETE
                        //
                        deviceExtension->MouseExtension.LastByteReceived = 0;

                        //
                        // Oops. Oh well, the mouse hasn't been able to reset
                        // in all these tries, so lets consider it dead.
                        // However, just in case the user yanks it out and
                        // plugs in a new one, we should reset our count
                        // back down to zero so that we will actually try to
                        // activate the thing when he plugs it back in there.
                        //
                        deviceExtension->MouseExtension.ResetCount = 0;
                        break;

                    }

                    //
                    // Because of the test for mouse resets at the start of
                    // the ISR, the following code should really have no
                    // effect
                    //
                    deviceExtension->MouseExtension.InputResetSubState =
                        ExpectingResetId;

                    break;

                }

                //
                // This state is special in that its just here as a place
                // holder because we have a detection routine at the start
                // of the ISR that automically puts us into
                // ExpectingGetDeviceIdACK. For completeness though, we
                // issue a bugcheck here since we can't normally get into
                // this state
                //
                case ExpectingResetId: {

                    //
                    // Next state
                    //
                    deviceExtension->MouseExtension.InputResetSubState =
                        ExpectingGetDeviceIdACK;

                    if (byte == WHEELMOUSE_ID_BYTE) {

                        //
                        // Start a enable command to the device. We *really* don't
                        // expect to be here.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                    } else {

                        //
                        // Log that we are in a bad state
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_INVALID_ISR_STATE
                            );

                        //
                        // This is *really* a bad place for us to be in, so
                        // bugcheck now
                        //
                        ASSERT( byte == WHEELMOUSE_ID_BYTE);

                        //
                        // For sake of completeness
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingReset;

                    }
                    break;

                }
                case ExpectingGetDeviceIdACK: {

                    I8xPrint((
                        1,
                        pDumpExpectingIdAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // Resend the "Get Mouse ID Transmission" sequence.
                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                    } else {

                        //
                        // If we got here, then we don't know what's going
                        // on with the device... The best bet is to send down
                        // a mouse reset command...
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, MOUSE_RESET );

                        //
                        // We now expect a reset to take place
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingReset;

                    }

                    break;

                }
                case ExpectingGetDeviceId: {

                    I8xPrint((
                        1,
                        pDumpExpecting,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) WHEELMOUSE_ID_BYTE,
                        (ULONG) byte
                        ));
                    //
                    // Got the device ID from the mouse and compare it with what we
                    // expect to see
                    //

                    if (byte != WHEELMOUSE_ID_BYTE) {

                        //
                        // Log an error/warning message so that we can track these
                        // problems down in the field
                        //

                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) (deviceExtension->HardwarePresent &
                                WHEELMOUSE_HARDWARE_PRESENT ?
                                I8042_UNEXPECTED_WHEEL_MOUSE_RESET :
                                I8042_UNEXPECTED_MOUSE_RESET) );


                        //
                        // For the sake of sanity, lets send a command to enable the
                        // mouse. Since we think it was a bonafide reset, we need to
                        // renable the sucker
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, ENABLE_MOUSE_TRANSMISSION );

                        //
                        // We go back to expecting an ACK
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingEnableACK;

                    } else {

                        //
                        // We have a wheel mouse present here... Log something so that
                        // we know that we got here.
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_BOGUS_MOUSE_RESET);

                        //
                        // Lets go back to the idle state
                        //
                        deviceExtension->MouseExtension.InputState = MouseIdle;

                        //
                        // Reset the number of possible mouse resets
                        //
                        deviceExtension->MouseExtension.ResetCount = 0;

                    }
                    break;

                }

                case ExpectingEnableACK: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        //
                        // Whatever hardware is present on the system, we will
                        // have to try to reset the mouse sampling rate, so
                        // do that now
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                        //
                        // Are we allowed to detect wether or not a wheel mouse
                        // is present?
                        //
                        if (deviceExtension->Configuration.EnableWheelDetection == 2) {

                            //
                            // Begin the sequence to activate the mouse wheel
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate200HzACK;

                        } else if (deviceExtension->Configuration.EnableWheelDetection == 1) {

                            //
                            // Begin the PNP sending sequence
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate20HzACK;

                        } else {

                            //
                            // Begin the sequence to set the default refresh
                            // rate
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRateDefaultACK;

                        }

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // Resend the "Enable Mouse Transmission" sequence.
                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, ENABLE_MOUSE_TRANSMISSION );

                    } else {

                        //
                        // We could not understand if we were able to reenable the
                        // mouse... Best bet here is to also reset the mouse.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, MOUSE_RESET );

                        //
                        // Expect a reset to occur
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingReset;

                    }

                    break;

                }

                case ExpectingSetSamplingRate20HzACK: {

                    I8xPrint((
                        1,
                        pDumpExpectingIdAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        //
                        // Set the mouse refresh to 20hz
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, 20 );

                        deviceExtension->MouseExtension.InputResetSubState =
                        ExpectingSetSamplingRate20Hz;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }
                case ExpectingSetSamplingRate20Hz: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE || byte == (UCHAR) RESEND) {

                        //
                        // We want to set a new sampling rate
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                        if (byte == (UCHAR) ACKNOWLEDGE) {

                            //
                            // Continue the sequence to activate PNP id
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate40HzACK;

                        } else {

                            //
                            // We have to restart a command if we get a resend
                            // on a microsoft mouse
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate20HzACK;

                        }

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }

                case ExpectingSetSamplingRate40HzACK: {
                    I8xPrint((
                        1,
                        pDumpExpectingIdAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        //
                        // Set the mouse refresh to 40hz
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, 40 );

                        deviceExtension->MouseExtension.InputResetSubState =
                        ExpectingSetSamplingRate40Hz;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }

                case ExpectingSetSamplingRate40Hz: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE || byte == (UCHAR) RESEND) {

                        //
                        // We want to set a new sampling rate
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                        if (byte == (UCHAR) ACKNOWLEDGE) {

                            //
                            // Continue the sequence to activate PNP id
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate60HzACK;

                        } else {

                            //
                            // We have to restart a command if we get a resend
                            // on a microsoft mouse
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate40HzACK;

                        }

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }
                case ExpectingSetSamplingRate60HzACK: {

                    I8xPrint((
                        1,
                        pDumpExpectingIdAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        //
                        // Set the mouse refresh to 60hz
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, 60 );

                        deviceExtension->MouseExtension.InputResetSubState =
                        ExpectingSetSamplingRate60Hz;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }

                case ExpectingSetSamplingRate60Hz: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE || byte == (UCHAR) RESEND) {

                        if (byte == (UCHAR) ACKNOWLEDGE) {

                            //
                            // Begin the sequence to receive the PNP Id
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingPnpIdByte1;

                        } else {

                            //
                            // We want to set a new sampling rate
                            //
                            I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                            I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                            //
                            // We have to restart a command if we get a resend
                            // on a microsoft mouse
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate60HzACK;

                        }

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }

                case ExpectingPnpIdByte1: {

                    I8xPrint((
                        1,
                        pDumpExpecting,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) 0x32,
                        (ULONG) byte
                        ));

                    //
                    // We expect a 0x32 here
                    //
                    if (byte == 0x32) {

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingPnpIdByte2;

                    } else {

                        //
                        // Oops, not our device, so lets stop the sequence
                        // now by sending it a GET_DEVICE_ID
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;
                }
                case ExpectingPnpIdByte2: {

                    I8xPrint((
                        1,
                        pDumpExpecting,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) 0x1F,
                        (ULONG) byte
                        ));

                    //
                    // We expect a 0x1f here
                    //
                    if (byte == 0x1F) {

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingPnpIdByte3;

                    } else {

                        //
                        // Oops, not our device, so lets stop the sequence
                        // now by sending it a GET_DEVICE_ID
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;
                }
                case ExpectingPnpIdByte3: {

                    I8xPrint((
                        1,
                        pDumpExpecting,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) 0x23,
                        (ULONG) byte
                        ));

                    //
                    // We expect a 0x23 here
                    //
                    if (byte == 0x23) {

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingPnpIdByte4;

                    } else {

                        //
                        // Oops, not our device, so lets stop the sequence
                        // now by sending it a GET_DEVICE_ID
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;
                }
                case ExpectingPnpIdByte4: {

                    I8xPrint((
                        1,
                        pDumpExpecting,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) 0x0B,
                        (ULONG) byte
                        ));

                    //
                    // We expect a 0x0B here
                    //
                    if (byte == 0x0B) {

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingPnpIdByte5;

                    } else {

                        //
                        // Oops, not our device, so lets stop the sequence
                        // now by sending it a GET_DEVICE_ID
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;
                }
                case ExpectingPnpIdByte5: {

                    I8xPrint((
                        1,
                        pDumpExpecting,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) 0x0B,
                        (ULONG) byte
                        ));

                    //
                    // We expect a 0x0B here
                    //
                    if (byte == 0x0B) {

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingPnpIdByte6;

                    } else {

                        //
                        // Oops, not our device, so lets stop the sequence
                        // now by sending it a GET_DEVICE_ID
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;
                }
                case ExpectingPnpIdByte6: {

                    I8xPrint((
                        1,
                        pDumpExpecting,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) 0x0B,
                        (ULONG) byte
                        ));

                    //
                    // We expect a 0x0B here
                    //
                    if (byte == 0x0B) {

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingPnpIdByte7;

                    } else {

                        //
                        // Oops, not our device, so lets stop the sequence
                        // now by sending it a GET_DEVICE_ID
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;
                }
                case ExpectingPnpIdByte7: {

                    I8xPrint((
                        1,
                        pDumpExpecting,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) 0x03,
                        (ULONG) byte
                        ));

                    //
                    // We expect a 0x03 here
                    //
                    if (byte == 0x03) {

                        //
                        // We are know know for sure that we have a wheel
                        // mouse on this system. However, we will update
                        // our date structures after the enable has gone
                        // through since that simplifies things a great deal
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingSetSamplingRate200HzACK;

                        //
                        // Set the new sampling rate
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                    } else {

                        //
                        // Oops, not our device, so lets stop the sequence
                        // now by sending it a GET_DEVICE_ID
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;
                }
                case ExpectingSetSamplingRate200HzACK: {

                    I8xPrint((
                        1,
                        pDumpExpectingIdAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        //
                        // Set the mouse refresh to 200hz
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, 200 );

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingSetSamplingRate200Hz;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }
                case ExpectingSetSamplingRate200Hz: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE || byte == (UCHAR) RESEND) {

                        //
                        // We want to set a new sampling rate
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                        if (byte == (UCHAR) ACKNOWLEDGE) {

                            //
                            // Begin the sequence to activate the mouse wheel
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate100HzACK;

                        } else {

                            //
                            // We have to restart a command if we get a resend
                            // on a microsoft mouse
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate200HzACK;

                        }

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }
                case ExpectingSetSamplingRate100HzACK: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        //
                        // Set the mouse refresh to 100hz
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, 100 );

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingSetSamplingRate100Hz;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }
                case ExpectingSetSamplingRate100Hz: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE || byte == (UCHAR) RESEND) {

                        //
                        // We want to set a new sampling rate
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                        if (byte == (UCHAR) ACKNOWLEDGE) {

                            //
                            // Begin the sequence to activate the mouse wheel
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate80HzACK;

                        } else {

                            //
                            // We have to restart a command if we get a resend
                            // on a microsoft mouse
                            //
                            deviceExtension->MouseExtension.InputResetSubState =
                                ExpectingSetSamplingRate100HzACK;

                        }

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }

                case ExpectingSetSamplingRate80HzACK: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        //
                        // Set the mouse refresh to 80hz
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, 80 );

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingSetSamplingRate80Hz;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }
                case ExpectingSetSamplingRate80Hz: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        //
                        // We want to set a new sampling rate
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Begin the sequence to activate the mouse wheel
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                        //
                        // We have to restart a command if we get a resend
                        // on a microsoft mouse
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingSetSamplingRate80HzACK;

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // Probably not a wheel mouse .. jump to the GetDeviceID2
                        // code
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                        //
                        // Best possible next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2ACK;

                    }

                    break;

                }
                case ExpectingGetDeviceId2ACK: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingGetDeviceId2;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, GET_DEVICE_ID );

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // We didn't get an ACK on this? Boggle. Okay, let's
                        // reset the mouse (probably AGAIN) and try to figure
                        // things out one more time
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, MOUSE_RESET );

                        //
                        // Expect a reset to occur
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingReset;

                    }

                    break;

                }
                case ExpectingGetDeviceId2: {

                    I8xPrint((
                        1,
                        pDumpExpecting,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) WHEELMOUSE_ID_BYTE,
                        (ULONG) byte
                        ));

                    if (byte == WHEELMOUSE_ID_BYTE || byte == MOUSE_ID_BYTE) {

                        //
                        // We have found a mouse, which is good
                        //
                        if (byte == WHEELMOUSE_ID_BYTE) {

                            //
                            // Update the HardwarePresent to show a Z mouse is
                            // operational and set the appropraite mouse type flags
                            //
                            deviceExtension->HardwarePresent |=
                                (WHEELMOUSE_HARDWARE_PRESENT | MOUSE_HARDWARE_PRESENT);

                            deviceExtension->Configuration.MouseAttributes.MouseIdentifier =
                                WHEELMOUSE_I8042_HARDWARE;

                            //
                            // Lets log the fact that we reenabled the wheel
                            // mouse
                            //
                            KeInsertQueueDpc(
                                &deviceExtension->ErrorLogMouseDpc,
                                (PIRP) NULL,
                                (PVOID) (ULONG) I8042_REENABLED_WHEEL_MOUSE
                                );

                        } else {

                            //
                            // Mouse Present, but no wheel
                            //
                            deviceExtension->HardwarePresent &= ~WHEELMOUSE_HARDWARE_PRESENT;
                            deviceExtension->HardwarePresent |= MOUSE_HARDWARE_PRESENT;
                            deviceExtension->Configuration.MouseAttributes.NumberOfButtons =
                                2;
                            deviceExtension->Configuration.MouseAttributes.MouseIdentifier =
                                MOUSE_I8042_HARDWARE;

                        }

                        //
                        // Send down the command to set a new sampling rate
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                        //
                        // This is our next state
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingSetSamplingRateDefaultACK;

                    } else {

                        //
                        // Make sure to log the problem
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_MOU_RESET_RESPONSE_FAILED
                            );

                        //
                        // Set up the state machine as best we can
                        //
                        deviceExtension->MouseExtension.InputState =
                            MouseExpectingACK;
                        deviceExtension->HardwarePresent &=
                            ~(WHEELMOUSE_HARDWARE_PRESENT | MOUSE_HARDWARE_PRESENT);
                        deviceExtension->Configuration.MouseAttributes.NumberOfButtons =
                            0;
                        deviceExtension->Configuration.MouseAttributes.MouseIdentifier =
                            0;

                    }

                    break;

                }
                case ExpectingSetSamplingRateDefaultACK: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        //
                        // Set the mouse refresh to the default
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, MOUSE_SAMPLE_RATE );

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingSetSamplingRateDefault;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // We didn't get an ACK on this? Boggle. Okay, let's
                        // reset the mouse (probably AGAIN) and try to figure
                        // things out one more time
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, MOUSE_RESET );

                        //
                        // Expect a reset to occur
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingReset;

                    }

                    break;

                }
                case ExpectingSetSamplingRateDefault: {

                    I8xPrint((
                        1,
                        pDumpExpectingAck,
                        pDriverName, pFncMouseInterrupt,
                        (ULONG) ACKNOWLEDGE,
                        (ULONG) byte
                        ));

                    if (byte == (UCHAR) ACKNOWLEDGE) {

                        //
                        // Done
                        //
                        deviceExtension->MouseExtension.InputState =
                            MouseIdle;

                        //
                        // Reset the number of possible mouse resets
                        //
                        deviceExtension->MouseExtension.ResetCount = 0;

                    } else if (byte == (UCHAR) RESEND) {

                        //
                        // NOTE: This is a hack for the Olivetti MIPS machine,
                        // which sends a resend response if a key is held down
                        // while we're attempting the I8xMouseEnableTransmission.
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, SET_MOUSE_SAMPLING_RATE );

                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingSetSamplingRateDefaultACK;

                    } else {

                        //
                        // Log the error
                        //
                        KeInsertQueueDpc(
                            &deviceExtension->ErrorLogMouseDpc,
                            (PIRP) NULL,
                            (PVOID) (ULONG) I8042_SET_SAMPLE_RATE_FAILED
                            );

                        //
                        // We didn't get an ACK on this? Boggle. Okay, let's
                        // reset the mouse (probably AGAIN) and try to figure
                        // things out one more time
                        //
                        I8X_WRITE_CMD_TO_MOUSE( deviceExtension );
                        I8X_MOUSE_COMMAND( deviceExtension, MOUSE_RESET );

                        //
                        // Expect a reset to occur
                        //
                        deviceExtension->MouseExtension.InputResetSubState =
                            ExpectingReset;

                    }

                    break;

                }
                default: {

                    //
                    // This is our bad state
                    //
                    I8xPrint((
                        1,
                        "%s-%s: INVALID RESET SUBSTATE %d\n",
                        pDriverName, pFncMouseInterrupt,
                        deviceExtension->MouseExtension.InputResetSubState
                        ));

                    //
                    // Queue a DPC to log an internal driver error.
                    //

                    KeInsertQueueDpc(
                        &deviceExtension->ErrorLogMouseDpc,
                        (PIRP) NULL,
                        (PVOID) (ULONG) I8042_INVALID_ISR_STATE
                        );

                    ASSERT(FALSE);

                }

            } // switch (deviceExtension->MouseExtension.InputResetSubState)

            break;
        }

        default: {

            //
            // This is our bad state
            //
            I8xPrint((
                1,
                "%s-%s: INVALID STATE %d\n",
                pDriverName, pFncMouseInterrupt,
                deviceExtension->MouseExtension.InputState
                ));

            //
            // Queue a DPC to log an internal driver error.
            //

            KeInsertQueueDpc(
                &deviceExtension->ErrorLogMouseDpc,
                (PIRP) NULL,
                (PVOID) (ULONG) I8042_INVALID_ISR_STATE
                );

            ASSERT(FALSE);
            break;
        }

    }

    I8xPrint((2, "%s-%s: %s\n", pDriverName, pFncMouseInterrupt, pExit ));

    return(TRUE);
}

NTSTATUS
I8xQueryNumberOfMouseButtons(
    IN  PDEVICE_OBJECT  DeviceObject,
    OUT PUCHAR          NumberOfMouseButtons
    )

/*++

Routine Description:

    This implements logitech's method for detecting the number of
    mouse buttons.  If anything doesn't go as expected then 0
    is returned.

    Calling this routine will set the mouse resolution to something
    really low.  The mouse resolution should be reset after this
    call.

Arguments:

    DeviceObject    - Supplies the device object.

    NumberOfMouseButtons    - Returns the number of mouse buttons or 0 if
                                the device did not support this type of
                                mouse button detection.

Return Value:

    An NTSTATUS code indicating success or failure.

--*/

{
    NTSTATUS            status;
    PDEVICE_EXTENSION   deviceExtension;
    UCHAR               byte;
    UCHAR               buttons;
    ULONG               i;

    deviceExtension = DeviceObject->DeviceExtension;

    //
    // First we need to send down a set resolution command
    //
    status = I8xPutBytePolled((CCHAR) DataPort, WAIT_FOR_ACKNOWLEDGE,
        (CCHAR) MouseDeviceType, deviceExtension,
        (UCHAR) SET_MOUSE_RESOLUTION);
    if (!NT_SUCCESS(status)) {

        return status;

    }

    //
    // This is another part of the data packet to get the info we want
    //
    status = I8xPutBytePolled((CCHAR) DataPort, WAIT_FOR_ACKNOWLEDGE,
        (CCHAR) MouseDeviceType, deviceExtension,
        (UCHAR) 0x00);

    if (!NT_SUCCESS(status)) {

        return status;

    }

    for (i = 0; i < 3; i++) {

        status = I8xPutBytePolled((CCHAR) DataPort, WAIT_FOR_ACKNOWLEDGE,
            (CCHAR) MouseDeviceType, deviceExtension,
            (UCHAR) SET_MOUSE_SCALING_1TO1);
        if (!NT_SUCCESS(status)) {

            return status;

        }

    }

    status = I8xPutBytePolled((CCHAR) DataPort, WAIT_FOR_ACKNOWLEDGE,
        (CCHAR) MouseDeviceType, deviceExtension,
        (UCHAR) READ_MOUSE_STATUS);
    if (!NT_SUCCESS(status)) {

        return status;

    }

    status = I8xGetBytePolled((CCHAR) ControllerDeviceType, deviceExtension,
        &byte);
    if (!NT_SUCCESS(status)) {

        return status;

    }

    status = I8xGetBytePolled((CCHAR) ControllerDeviceType, deviceExtension,
        &buttons);
    if (!NT_SUCCESS(status)) {

        return status;

    }

    status = I8xGetBytePolled((CCHAR) ControllerDeviceType, deviceExtension,
        &byte);
    if (!NT_SUCCESS(status)) {

        return status;

    }

    if (buttons == 2 || buttons == 3) {

        *NumberOfMouseButtons = buttons;

    } else {

        *NumberOfMouseButtons = 0;

    }
    return status;
}

NTSTATUS
I8xInitializeMouse(
    IN PDEVICE_OBJECT DeviceObject
    )

/*++

Routine Description:

    This routine initializes the i8042 mouse hardware.  It is called
    only at initialization, and does not synchronize access to the hardware.

Arguments:

    DeviceObject - Pointer to the device object.

Return Value:

    Returns status.

--*/

{

    #define DUMP_COUNT 4
    NTSTATUS                errorCode = STATUS_SUCCESS;
    NTSTATUS                status;
    PDEVICE_EXTENSION       deviceExtension;
    PIO_ERROR_LOG_PACKET    errorLogEntry;
    UCHAR                   byte;
    UCHAR                   numButtons;
    ULONG                   dumpData[DUMP_COUNT];
    ULONG                   dumpCount = 0;
    ULONG                   i;
    ULONG                   uniqueErrorValue;

    I8xPrint((2, "%s-%s: %s\n", pDriverName, pFncInitializeMouse, pEnter));

    //
    // Initialize this array
    //
    for (i = 0; i < DUMP_COUNT; i++) {

        dumpData[i] = 0;

    }

    //
    // Get the device extension.
    //
    deviceExtension = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;

    //
    // Reset the mouse.  Send a Write To Auxiliary Device command to the
    // 8042 controller.  Then send the Reset Mouse command to the mouse
    // through the 8042 data register.  Expect to get back an ACK, followed
    // by a completion code and the ID code (0x00).
    //
    status = I8xPutBytePolled(
        (CCHAR) DataPort,
        WAIT_FOR_ACKNOWLEDGE,
        (CCHAR) MouseDeviceType,
        deviceExtension,
        (UCHAR) MOUSE_RESET
        );

    if (!NT_SUCCESS(status)) {

        I8xPrint((
            1,
            "%s-%s: failed mouse reset, status 0x%x\n",
            pDriverName, pFncInitializeMouse,
            status
            ));

        //
        // Set up error log info.
        //
        errorCode = I8042_MOU_RESET_COMMAND_FAILED;
        uniqueErrorValue = I8042_ERROR_VALUE_BASE + 415;
        dumpData[0] = KBDMOU_COULD_NOT_SEND_PARAM;
        dumpData[1] = DataPort;
        dumpData[2] = I8042_WRITE_TO_AUXILIARY_DEVICE;
        dumpData[3] = MOUSE_RESET;
        dumpCount = 4;

        goto I8xInitializeMouseExit;
    }

    //
    // Get the mouse reset responses.  The first response should be a
    // MOUSE_COMPLETE.  The second response should be the mouse ID.
    // Note that it is usually necessary to stall a long time to get the
    // mouse reset/self-test to work.  The stall value was determined by
    // experimentation.
    //
    //
    for (i = 0; i < 11200; i++) {

        status = I8xGetBytePolled(
            (CCHAR) ControllerDeviceType,
            deviceExtension,
            &byte
            );

        if (NT_SUCCESS(status) && (byte == (UCHAR) MOUSE_COMPLETE)) {

            //
            // The reset completed successfully.
            //
            break;


        } else {

            //
            // Stall, and then try again to get a response from
            // the reset.
            //
            if (status == STATUS_IO_TIMEOUT) {

                //
                // Stall, and then try again to get a response from
                // the reset.
                //
                KeStallExecutionProcessor(50);

            } else {

                break;

            }

        }

    }

    if (!NT_SUCCESS(status)) {
        I8xPrint((
            1,
            "%s-%s: failed reset response 1, status 0x%x, byte 0x%x\n",
            pDriverName, pFncInitializeMouse,
            status,
            byte
            ));

        //
        // Set up error log info.
        //
        errorCode = I8042_MOU_RESET_RESPONSE_FAILED;
        uniqueErrorValue = I8042_ERROR_VALUE_BASE + 420;
        dumpData[0] = KBDMOU_INCORRECT_RESPONSE;
        dumpData[1] = ControllerDeviceType;
        dumpData[2] = MOUSE_COMPLETE;
        dumpData[3] = byte;
        dumpCount = 4;

        goto I8xInitializeMouseExit;
    }

    status = I8xGetBytePolled(
        (CCHAR) ControllerDeviceType,
        deviceExtension,
        &byte
        );

    if ((!NT_SUCCESS(status)) || (byte != MOUSE_ID_BYTE)) {

        I8xPrint((
            1,
            "%s-%s: failed reset response 2, status 0x%x, byte 0x%x\n",
            pDriverName, pFncInitializeMouse,
            status,
            byte
            ));

        //
        // Set up error log info.
        //
        errorCode = I8042_MOU_RESET_RESPONSE_FAILED;
        uniqueErrorValue = I8042_ERROR_VALUE_BASE + 425;
        dumpData[0] = KBDMOU_INCORRECT_RESPONSE;
        dumpData[1] = ControllerDeviceType;
        dumpData[2] = MOUSE_ID_BYTE;
        dumpData[3] = byte;
        dumpCount = 4;

        goto I8xInitializeMouseExit;
    }

    //
    // Check to see if this is a wheel mouse
    //
    I8xFindWheelMouse(DeviceObject);

    //
    // Try to detect the number of mouse buttons.
    //
    status = I8xQueryNumberOfMouseButtons(DeviceObject, &numButtons);

    if (!NT_SUCCESS(status)) {
        I8xPrint((
            1,
            "%s-%s: failed to get buttons, status 0x%x\n",
            pDriverName, pFncInitializeMouse,
            status
            ));

        //
        // Set up error log info.
        //
        errorCode = I8042_ERROR_DURING_BUTTONS_DETECT;
        uniqueErrorValue = I8042_ERROR_VALUE_BASE + 426;
        dumpData[0] = KBDMOU_COULD_NOT_SEND_PARAM;
        dumpData[1] = DataPort;
        dumpData[2] = I8042_WRITE_TO_AUXILIARY_DEVICE;
        dumpCount = 3;

        goto I8xInitializeMouseExit;

    } else if (numButtons) {

        deviceExtension->Configuration.MouseAttributes.NumberOfButtons =
            numButtons;

    }

    //
    // Set mouse sampling rate.  Send a Write To Auxiliary Device command
    // to the 8042 controller.  Then send the Set Mouse Sampling Rate
    // command to the mouse through the 8042 data register,
    // followed by its parameter.
    //
    status = I8xPutBytePolled(
        (CCHAR) DataPort,
        WAIT_FOR_ACKNOWLEDGE,
        (CCHAR) MouseDeviceType,
        deviceExtension,
        (UCHAR) SET_MOUSE_SAMPLING_RATE
        );

    if (!NT_SUCCESS(status)) {

        I8xPrint((
            1,
            "%s-%s: failed write set sample rate, status 0x%x\n",
            pDriverName, pFncInitializeMouse,
            status
            ));

        //
        // Set up error log info.
        //
        errorCode = I8042_SET_SAMPLE_RATE_FAILED;
        uniqueErrorValue = I8042_ERROR_VALUE_BASE + 435;
        dumpData[0] = KBDMOU_COULD_NOT_SEND_PARAM;
        dumpData[1] = DataPort;
        dumpData[2] = I8042_WRITE_TO_AUXILIARY_DEVICE;
        dumpData[3] = SET_MOUSE_SAMPLING_RATE;
        dumpCount = 4;

        goto I8xInitializeMouseExit;

    }

    status = I8xPutBytePolled(
        (CCHAR) DataPort,
        WAIT_FOR_ACKNOWLEDGE,
        (CCHAR) MouseDeviceType,
        deviceExtension,
        (UCHAR) MOUSE_SAMPLE_RATE
        );

    if (!NT_SUCCESS(status)) {

        I8xPrint((
            1,
            "%s-%s: failed write sample rate, status 0x%x\n",
            pDriverName, pFncInitializeMouse,
            status
            ));

        //
        // Set up error log info.
        //
        errorCode = I8042_SET_SAMPLE_RATE_FAILED;
        uniqueErrorValue = I8042_ERROR_VALUE_BASE + 445;
        dumpData[0] = KBDMOU_COULD_NOT_SEND_PARAM;
        dumpData[1] = DataPort;
        dumpData[2] = I8042_WRITE_TO_AUXILIARY_DEVICE;
        dumpData[3] = MOUSE_SAMPLE_RATE;
        dumpCount = 4;

        goto I8xInitializeMouseExit;

    }

    //
    // Set the mouse resolution.  Send a Write To Auxiliary Device command
    // to the 8042 controller.  Then send the Set Mouse Resolution
    // command to the mouse through the 8042 data register,
    // followed by its parameter.
    //
    status = I8xPutBytePolled(
        (CCHAR) DataPort,
        WAIT_FOR_ACKNOWLEDGE,
        (CCHAR) MouseDeviceType,
        deviceExtension,
        (UCHAR) SET_MOUSE_RESOLUTION
        );

    if (!NT_SUCCESS(status)) {

        I8xPrint((
            1,
            "%s-%s: failed write set resolution, status 0x%x\n",
            pDriverName, pFncInitializeMouse,
            status
            ));

        //
        // Set up error log info.
        //
        errorCode = I8042_SET_RESOLUTION_FAILED;
        uniqueErrorValue = I8042_ERROR_VALUE_BASE + 455;
        dumpData[0] = KBDMOU_COULD_NOT_SEND_PARAM;
        dumpData[1] = DataPort;
        dumpData[2] = I8042_WRITE_TO_AUXILIARY_DEVICE;
        dumpData[3] = SET_MOUSE_RESOLUTION;
        dumpCount = 4;

        goto I8xInitializeMouseExit;

    }

    status = I8xPutBytePolled(
        (CCHAR) DataPort,
        WAIT_FOR_ACKNOWLEDGE,
        (CCHAR) MouseDeviceType,
        deviceExtension,
        (UCHAR) deviceExtension->Configuration.MouseResolution
        );

    if (!NT_SUCCESS(status)) {

        I8xPrint((
            1,
            "%s-%s: failed set mouse resolution, status 0x%x\n",
            pDriverName, pFncInitializeMouse,
            status
            ));

        //
        // Set up error log info.
        //
        errorCode = I8042_SET_RESOLUTION_FAILED;
        uniqueErrorValue = I8042_ERROR_VALUE_BASE + 465;
        dumpData[0] = KBDMOU_COULD_NOT_SEND_PARAM;
        dumpData[1] = DataPort;
        dumpData[2] = I8042_WRITE_TO_AUXILIARY_DEVICE;
        dumpData[3] = deviceExtension->Configuration.MouseResolution;
        dumpCount = 4;

        goto I8xInitializeMouseExit;

    }

I8xInitializeMouseExit:

    if (!NT_SUCCESS(status)) {

        //
        // The mouse initialization failed.  Log an error.
        //
        if (errorCode != STATUS_SUCCESS) {

            errorLogEntry = (PIO_ERROR_LOG_PACKET)
                IoAllocateErrorLogEntry(
                    DeviceObject, (UCHAR)
                    (sizeof(IO_ERROR_LOG_PACKET) + (dumpCount * sizeof(ULONG)))
                    );

            if (errorLogEntry != NULL) {

                errorLogEntry->ErrorCode = errorCode;
                errorLogEntry->DumpDataSize = (USHORT) dumpCount * sizeof(ULONG);
                errorLogEntry->SequenceNumber = 0;
                errorLogEntry->MajorFunctionCode = 0;
                errorLogEntry->IoControlCode = 0;
                errorLogEntry->RetryCount = 0;
                errorLogEntry->UniqueErrorValue = uniqueErrorValue;
                errorLogEntry->FinalStatus = status;
                for (i = 0; i < dumpCount; i++) {

                    errorLogEntry->DumpData[i] = dumpData[i];

                }

                IoWriteErrorLogEntry(errorLogEntry);

            }

        }

    }

    //
    // Initialize current mouse input packet state.
    //
    deviceExtension->MouseExtension.PreviousSignAndOverflow = 0;
    deviceExtension->MouseExtension.InputState = MouseExpectingACK;
    deviceExtension->MouseExtension.LastByteReceived = 0;
    I8xPrint((2, "%s-%s: %s\n", pDriverName, pFncInitializeMouse, pExit));

    return(status);
}

VOID
I8xMouseConfiguration(
    IN PINIT_EXTENSION InitializationData,
    IN PUNICODE_STRING RegistryPath,
    IN PUNICODE_STRING KeyboardDeviceName,
    IN PUNICODE_STRING PointerDeviceName
    )

/*++

Routine Description:

    This routine retrieves the configuration information for the mouse.

Arguments:

    InitializationData - Pointer to the temporary device extension and some
        additional configuration information

    RegistryPath - Pointer to the null-terminated Unicode name of the
        registry path for this driver.

    KeyboardDeviceName - Pointer to the Unicode string that will receive
        the keyboard port device name.

    PointerDeviceName - Pointer to the Unicode string that will receive
        the pointer port device name.


Return Value:

    None.  As a side-effect, may set DeviceExtension->HardwarePresent.

--*/
{
    CONFIGURATION_TYPE  controllerType = PointerController;
    CONFIGURATION_TYPE  peripheralType = PointerPeripheral;
    INTERFACE_TYPE      interfaceType;
    NTSTATUS            status = STATUS_SUCCESS;
    PDEVICE_EXTENSION   deviceExtension = &(InitializationData->DeviceExtension);
    ULONG               i;

    for (i = 0; i < MaximumInterfaceType; i++) {

        //
        // Get the registry information for this device.
        //
        interfaceType = i;
        status = IoQueryDeviceDescription(&interfaceType,
            NULL,
            &controllerType,
            NULL,
            &peripheralType,
            NULL,
            I8xMousePeripheralCallout,
            (PVOID) InitializationData);

        if (InitializationData->DeviceExtension.HardwarePresent &
                MOUSE_HARDWARE_PRESENT) {

            //
            // If we didn't already get these when determining the keyboard
            // configuration, get the service parameters now (e.g.,
            // user-configurable number of resends, polling iterations, etc.).
            //
            if (!(InitializationData->DeviceExtension.HardwarePresent &
                    KEYBOARD_HARDWARE_PRESENT)) {

                I8xServiceParameters(
                    InitializationData,
                    RegistryPath,
                    KeyboardDeviceName,
                    PointerDeviceName
                    );

            }

            //
            // Initialize mouse-specific configuration parameters.
            //
            InitializationData->DeviceExtension.Configuration.MouseAttributes.
                MouseIdentifier = MOUSE_I8042_HARDWARE;

            break;

        } else {

            I8xPrint((
                1,
                "%s-I8xMouseConfiguration: "
                "IoQueryDeviceDescription for bus type %d failed\n",
                pDriverName,
                interfaceType
                ));
        }
    }
}

NTSTATUS
I8xMouseEnableTransmission(
    IN PDEVICE_OBJECT DeviceObject
    )

/*++

Routine Description:

    This routine sends an Enable command to the mouse hardware, causing
    the mouse to begin transmissions.  It is called at initialization
    time, but only after the interrupt has been connected.  This is
    necessary so the driver can keep its notion of the mouse input data
    state in sync with the hardware (i.e., for this type of mouse there is no
    way to differentiate the first byte of a packet; if the user is randomly
    moving the mouse during boot/initialization, the first mouse interrupt we
    receive following IoConnectInterrupt could be for a byte that is not the
    start of a packet, and we have no way to know that).

Arguments:

    DeviceObject - Pointer to the device object.

Return Value:

    Returns status.

--*/

{
#define DUMP_COUNT 4
    NTSTATUS                errorCode = STATUS_SUCCESS;
    NTSTATUS                status;
    PDEVICE_EXTENSION       deviceExtension;
    PIO_ERROR_LOG_PACKET    errorLogEntry;
    ULONG                   dumpCount = 0;
    ULONG                   dumpData[DUMP_COUNT];
    ULONG                   i;
    ULONG                   uniqueErrorValue;

    I8xPrint((2, "%s-%s: %s\n", pDriverName, pFncMouseEnable, pEnter));

    //
    // Initialize the dump structure
    //
    for (i = 0; i < DUMP_COUNT; i++) {

        dumpData[i] = 0;

    }

    //
    // Get the device extension.
    //
    deviceExtension = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;

    //
    // Re-enable the mouse at the mouse hardware, so that it can transmit
    // data packets in continuous mode.  Note that this is not the same
    // as enabling the mouse device at the 8042 controller.  The mouse
    // hardware is sent an Enable command here, because it was
    // Disabled as a result of the mouse reset command performed
    // in I8xInitializeMouse().
    //
    // Note that we don't wait for an ACKNOWLEDGE back.  The
    // ACKNOWLEDGE back will actually cause a mouse interrupt, which
    // then gets handled in the mouse ISR.
    //
    status = I8xPutBytePolled(
        (CCHAR) DataPort,
        NO_WAIT_FOR_ACKNOWLEDGE,
        (CCHAR) MouseDeviceType,
        deviceExtension,
        (UCHAR) ENABLE_MOUSE_TRANSMISSION
        );

    if (!NT_SUCCESS(status)) {

        I8xPrint((
            1,
            "%s-%s: "
            "failed write enable transmission, status 0x%x\n",
            pDriverName, pFncMouseEnable,
            status
            ));

        //
        // Set up error log info.
        //
        errorCode = I8042_MOU_ENABLE_XMIT;
        uniqueErrorValue = I8042_ERROR_VALUE_BASE + 475;
        dumpData[0] = KBDMOU_COULD_NOT_SEND_PARAM;
        dumpData[1] = DataPort;
        dumpData[2] = I8042_WRITE_TO_AUXILIARY_DEVICE;
        dumpData[3] = ENABLE_MOUSE_TRANSMISSION;
        dumpCount = 4;

        goto I8xEnableMouseTransmissionExit;
    }

I8xEnableMouseTransmissionExit:

    if (!NT_SUCCESS(status)) {

        //
        // The mouse initialization failed.  Log an error.
        //
        if (errorCode != STATUS_SUCCESS) {

            errorLogEntry = (PIO_ERROR_LOG_PACKET)
                IoAllocateErrorLogEntry(
                    DeviceObject, (UCHAR)
                    (sizeof(IO_ERROR_LOG_PACKET) + (dumpCount * sizeof(ULONG)))
                    );

            if (errorLogEntry != NULL) {

                errorLogEntry->ErrorCode = errorCode;
                errorLogEntry->DumpDataSize = (USHORT) dumpCount * sizeof(ULONG);
                errorLogEntry->SequenceNumber = 0;
                errorLogEntry->MajorFunctionCode = 0;
                errorLogEntry->IoControlCode = 0;
                errorLogEntry->RetryCount = 0;
                errorLogEntry->UniqueErrorValue = uniqueErrorValue;
                errorLogEntry->FinalStatus = status;
                for (i = 0; i < dumpCount; i++) {

                    errorLogEntry->DumpData[i] = dumpData[i];

                }
                IoWriteErrorLogEntry(errorLogEntry);

            }

        }

    }

    I8xPrint((2, "%s-%s: %s\n", pDriverName, pFncMouseEnable, pExit));

    return(status);
}

NTSTATUS
I8xMousePeripheralCallout(
    IN PVOID Context,
    IN PUNICODE_STRING PathName,
    IN INTERFACE_TYPE BusType,
    IN ULONG BusNumber,
    IN PKEY_VALUE_FULL_INFORMATION *BusInformation,
    IN CONFIGURATION_TYPE ControllerType,
    IN ULONG ControllerNumber,
    IN PKEY_VALUE_FULL_INFORMATION *ControllerInformation,
    IN CONFIGURATION_TYPE PeripheralType,
    IN ULONG PeripheralNumber,
    IN PKEY_VALUE_FULL_INFORMATION *PeripheralInformation
    )

/*++

Routine Description:

    This is the callout routine sent as a parameter to
    IoQueryDeviceDescription.  It grabs the pointer controller and
    peripheral configuration information.

Arguments:

    Context - Context parameter that was passed in by the routine
        that called IoQueryDeviceDescription.

    PathName - The full pathname for the registry key.

    BusType - Bus interface type (Isa, Eisa, Mca, etc.).

    BusNumber - The bus sub-key (0, 1, etc.).

    BusInformation - Pointer to the array of pointers to the full value
        information for the bus.

    ControllerType - The controller type (should be PointerController).

    ControllerNumber - The controller sub-key (0, 1, etc.).

    ControllerInformation - Pointer to the array of pointers to the full
        value information for the controller key.

    PeripheralType - The peripheral type (should be PointerPeripheral).

    PeripheralNumber - The peripheral sub-key.

    PeripheralInformation - Pointer to the array of pointers to the full
        value information for the peripheral key.


Return Value:

    None.  If successful, will have the following side-effects:

        - Sets DeviceObject->DeviceExtension->HardwarePresent.
        - Sets configuration fields in
          DeviceObject->DeviceExtension->Configuration.

--*/
{
    ANSI_STRING                         ansiString;
    BOOLEAN                             defaultInterruptShare;
    BOOLEAN                             portInfoNeeded;
    CM_PARTIAL_RESOURCE_DESCRIPTOR      tmpResourceDescriptor;
    KINTERRUPT_MODE                     defaultInterruptMode;
    NTSTATUS                            status = STATUS_SUCCESS;
    PCM_PARTIAL_RESOURCE_DESCRIPTOR     resourceDescriptor;
    PDEVICE_EXTENSION                   deviceExtension;
    PI8042_CONFIGURATION_INFORMATION    configuration;
    PINIT_EXTENSION                     initializationData;
    PUCHAR                              controllerData;
    ULONG                               i;
    ULONG                               listCount;
    UNICODE_STRING                      unicodeIdentifier;

    I8xPrint((
        1,
        "%s-%s: Path @ 0x%x\n"
        "    %s %s 0x%x\n"
        "    %s %s 0x%x\n"
        "    %s %s 0x%x\n"
        "    %s %s 0x%x\n"
        "    %s %s @ 0x%x\n"
        "    %s %s 0x%x\n"
        "    %s %s 0x%x\n"
        "    %s %s @ 0x%x\n",
        pDriverName, pFncMousePeripheral, PathName,
        pBus, pType, BusType,
        pBus, pNumber, BusNumber,
        pController, pType, ControllerType,
        pController, pNumber, ControllerNumber,
        pController, pInfo, ControllerInformation,
        pPeripheral, pType, PeripheralType,
        pPeripheral, pNumber, PeripheralNumber,
        pPeripheral, pInfo, PeripheralInformation
        ));

    //
    // Get the length of the peripheral identifier information.
    //
    unicodeIdentifier.Length = (USHORT)
        (*(PeripheralInformation + IoQueryDeviceIdentifier))->DataLength;

    //
    // If we already have the configuration information for the
    // pointer peripheral, or if the peripheral identifier is missing,
    // just return.
    //
    initializationData = (PINIT_EXTENSION) Context;
    deviceExtension = &(initializationData->DeviceExtension);
    if ((deviceExtension->HardwarePresent & MOUSE_HARDWARE_PRESENT)
         || (unicodeIdentifier.Length == 0)) {

        return (status);

    }

    configuration = &deviceExtension->Configuration;

    //
    // Get the identifier information for the peripheral.
    //
    unicodeIdentifier.MaximumLength = unicodeIdentifier.Length;
    unicodeIdentifier.Buffer = (PWSTR) (((PUCHAR)(*(PeripheralInformation +
        IoQueryDeviceIdentifier))) + (*(PeripheralInformation +
        IoQueryDeviceIdentifier))->DataOffset);

    I8xPrint((
        1,
        "%s-%s: Mouse %s %ws\n",
        pDriverName, pFncMousePeripheral, pType,
        unicodeIdentifier.Buffer));

    //
    // Verify that this is an i8042 mouse.
    //
    status = RtlUnicodeStringToAnsiString(
        &ansiString,
        &unicodeIdentifier,
        TRUE
        );

    if (!NT_SUCCESS(status)) {

        I8xPrint((
            1,
            "%s-%s: Could not convert identifier to Ansi\n",
            pDriverName, pFncMousePeripheral
            ));

        return(status);

    }

    if (strstr(ansiString.Buffer, "PS2")) {

         //
         // There is a mouse on the i8042 controller.
         //
         deviceExtension->HardwarePresent |= MOUSE_HARDWARE_PRESENT;

    }
    RtlFreeAnsiString(&ansiString);

    if (!(deviceExtension->HardwarePresent & MOUSE_HARDWARE_PRESENT)) {

        return(status);

    }

    if (!(deviceExtension->HardwarePresent & KEYBOARD_HARDWARE_PRESENT)) {

        //
        // If we don't already have the bus information as the result
        // of the keyboard configuration, grab it now.
        //
        configuration->InterfaceType = BusType;
        configuration->BusNumber = BusNumber;
        configuration->FloatingSave = I8042_FLOATING_SAVE;

    }

    if (configuration->InterfaceType == MicroChannel) {

        defaultInterruptShare = TRUE;
        defaultInterruptMode = LevelSensitive;

    } else {

        defaultInterruptShare = I8042_INTERRUPT_SHARE;
        defaultInterruptMode = I8042_INTERRUPT_MODE;

    }

    //
    // Look through the controller's resource list for interrupt
    // and (possibly) port configuration information.
    //
    if ((*(ControllerInformation + IoQueryDeviceConfigurationData))->DataLength != 0){

        controllerData = ((PUCHAR) (*(ControllerInformation +
            IoQueryDeviceConfigurationData))) + (*(ControllerInformation +
            IoQueryDeviceConfigurationData))->DataOffset;
        controllerData += FIELD_OFFSET(CM_FULL_RESOURCE_DESCRIPTOR,
            PartialResourceList);
        listCount = ((PCM_PARTIAL_RESOURCE_LIST) controllerData)->Count;

        resourceDescriptor =
            ((PCM_PARTIAL_RESOURCE_LIST) controllerData)->PartialDescriptors;
        portInfoNeeded = configuration->PortListCount? FALSE : TRUE;

        for (i = 0; i < listCount; i++, resourceDescriptor++) {
            switch(resourceDescriptor->Type) {
                case CmResourceTypePort:

                    if (portInfoNeeded) {

                        //
                        // If we don't already have the port information as
                        // a result of the keyboard configuration, copy it
                        // to the port list.
                        //
                        configuration->PortList[configuration->PortListCount] =
                            *resourceDescriptor;
                        configuration->PortList[configuration->PortListCount].ShareDisposition =
                            I8042_REGISTER_SHARE ? CmResourceShareShared:
                                CmResourceShareDriverExclusive;
                        configuration->PortListCount += 1;

                    }

                    break;

                case CmResourceTypeInterrupt:

                    //
                    // Copy the interrupt information.
                    //
                    configuration->MouseInterrupt = *resourceDescriptor;
                    configuration->MouseInterrupt.ShareDisposition =
                        defaultInterruptShare ? CmResourceShareShared :
                            CmResourceShareDeviceExclusive;

                    break;

                default:
                    break;

            }

        }

    }

    //
    // If no interrupt configuration information was found, use the
    // mouse driver defaults.
    //
    if (!(configuration->MouseInterrupt.Type & CmResourceTypeInterrupt)) {

        I8xPrint((
            1,
            "%s-%s: Using default mouse interrupt config\n",
            pDriverName, pFncMousePeripheral
            ));

        configuration->MouseInterrupt.Type = CmResourceTypeInterrupt;
        configuration->MouseInterrupt.ShareDisposition =
            defaultInterruptShare? CmResourceShareShared :
                CmResourceShareDeviceExclusive;
        configuration->MouseInterrupt.Flags =
            (defaultInterruptMode == Latched)? CM_RESOURCE_INTERRUPT_LATCHED :
                CM_RESOURCE_INTERRUPT_LEVEL_SENSITIVE;
        configuration->MouseInterrupt.u.Interrupt.Level = MOUSE_IRQL;
        configuration->MouseInterrupt.u.Interrupt.Vector = MOUSE_VECTOR;

    }

    I8xPrint((
        1,
        "%s-%s: Mouse interrupt config --\n"
        "%s, %s, Irq = %d\n",
        pDriverName, pFncMousePeripheral,
        configuration->MouseInterrupt.ShareDisposition == CmResourceShareShared?
            "Sharable" : "NonSharable",
        configuration->MouseInterrupt.Flags == CM_RESOURCE_INTERRUPT_LATCHED?
            "Latched" : "Level Sensitive",
        configuration->MouseInterrupt.u.Interrupt.Vector
        ));

    //
    // If no port configuration information was found, use the
    // mouse driver defaults.
    //
    if (configuration->PortListCount == 0) {

        //
        // No port configuration information was found, so use
        // the driver defaults.
        //
        I8xPrint((
            1,
            "%s-%s: Using default port config\n",
            pDriverName, pFncMousePeripheral
            ));

        configuration->PortList[DataPort].Type = CmResourceTypePort;
        configuration->PortList[DataPort].Flags = I8042_PORT_TYPE;
        configuration->PortList[DataPort].u.Port.Start.HighPart = 0;
        configuration->PortList[DataPort].u.Port.Length = I8042_REGISTER_LENGTH;
        configuration->PortList[DataPort].ShareDisposition =
            I8042_REGISTER_SHARE? CmResourceShareShared:
                CmResourceShareDriverExclusive;
        configuration->PortList[DataPort].u.Port.Start.LowPart =
            I8042_PHYSICAL_BASE + I8042_DATA_REGISTER_OFFSET;

        configuration->PortList[CommandPort].Type = CmResourceTypePort;
        configuration->PortList[CommandPort].Flags = I8042_PORT_TYPE;
        configuration->PortList[CommandPort].u.Port.Start.HighPart = 0;
        configuration->PortList[CommandPort].u.Port.Length = I8042_REGISTER_LENGTH;
        configuration->PortList[CommandPort].ShareDisposition =
            I8042_REGISTER_SHARE? CmResourceShareShared:
                CmResourceShareDriverExclusive;
        configuration->PortList[CommandPort].u.Port.Start.LowPart =
            I8042_PHYSICAL_BASE + I8042_COMMAND_REGISTER_OFFSET;

        configuration->PortListCount = 2;

    } else if (configuration->PortListCount == 1) {

        //
        // Kludge for Jazz machines.  Their ARC firmware neglects to
        // separate out the port addresses, so fix that up here.
        //
        configuration->PortList[CommandPort] = configuration->PortList[DataPort];
        configuration->PortList[CommandPort].u.Port.Start.LowPart +=
            I8042_COMMAND_REGISTER_OFFSET;
        configuration->PortListCount += 1;

    } else {

        //
        // Put the lowest port address range in the DataPort element of
        // the port list.
        //
        if (configuration->PortList[CommandPort].u.Port.Start.LowPart
            < configuration->PortList[DataPort].u.Port.Start.LowPart) {

            tmpResourceDescriptor = configuration->PortList[DataPort];
            configuration->PortList[DataPort] =
                 configuration->PortList[CommandPort];
            configuration->PortList[CommandPort] = tmpResourceDescriptor;

        }

    }

#ifdef PNP_IDENTIFY
    //
    // In any event we're going to use the pointer based on this data,
    // so make sure we can tell PNP that we've claimed it later on
    //
    initializationData->MouseConfig.InterfaceType = BusType;
    initializationData->MouseConfig.InterfaceNumber = BusNumber;
    initializationData->MouseConfig.ControllerType = ControllerType;
    initializationData->MouseConfig.ControllerNumber = ControllerNumber;
    initializationData->MouseConfig.PeripheralType = PeripheralType;
    initializationData->MouseConfig.PeripheralNumber = PeripheralNumber;
#endif

    for (i = 0; i < configuration->PortListCount; i++) {

        I8xPrint((
            1,
            "    %s, Ports 0x%x - 0x%x\n",
            configuration->PortList[i].ShareDisposition
                == CmResourceShareShared?  "Sharable" : "NonSharable",
            configuration->PortList[i].u.Port.Start.LowPart,
            configuration->PortList[i].u.Port.Start.LowPart +
                configuration->PortList[i].u.Port.Length - 1
            ));

    }
    return(status);
}


NTSTATUS
I8xFindWheelMouse(
    IN PDEVICE_OBJECT DeviceObject
    )

/*++

Routine Description:

    There are two methods of finding a wheel mouse on a system. The first
    method, is to send down the request to get the PNP id of the device
    and compare it with the known id for a wheel mouse. The method is
    useful since some machines hang on the second detection mechanism,
    even if no mouse is present on the system.

    The second method, which also enables a wheel mouse is set the sampling
    rate to 200hz, then 100hz, then 80hz, and then read the device id. An
    ID of 3 indicates a zoom mouse.

    If the registry entry "EnableWheelDetection" is 0 then this
    routine will just return STATUS_NO_SUCH_DEVICE. If the registry entry
    is 1 (the default), then the first and second detection mechanisms will
    be used. If the registry entry is 2, then only the second detection
    mechanism will be used.

Arguments:

    DeviceObject - Pointer to the device object

Return Value:

    Returns status

Remarks:

    As a side effect the sample rate is left at 80Hz and if a wheelmouse is
    attached it is in the wheel mode where packets are different.

--*/

{
#define DUMP_COUNT 4
    NTSTATUS                errorCode = STATUS_SUCCESS;
    NTSTATUS                status;
    PDEVICE_EXTENSION       deviceExtension;
    PIO_ERROR_LOG_PACKET    errorLogEntry;
    UCHAR                   byte;
    UCHAR                   commandCount = 0;
    UCHAR                   enableCommands[] = {
                                SET_MOUSE_SAMPLING_RATE, 200,
                                SET_MOUSE_SAMPLING_RATE, 100,
                                SET_MOUSE_SAMPLING_RATE, 80,
                                GET_DEVICE_ID, 0  // NULL terminate
                                };
    UCHAR                   pnpCommands[] = {
                                SET_MOUSE_SAMPLING_RATE, 20,
                                SET_MOUSE_SAMPLING_RATE, 40,
                                SET_MOUSE_SAMPLING_RATE, 60,
                                0  // NULL terminates
                                };
    UCHAR                   pnpID[] = {
                                0x32, 0x1f, 0x23, 0x0b, 0x0b, 0x0b, 0x03, 0x00
                                };
    ULONG                   dumpCount = 0;
    ULONG                   dumpData[DUMP_COUNT];
    ULONG                   i;
    ULONG                   idCount;
    ULONG                   uniqueErrorValue = I8042_ERROR_VALUE_BASE + 480;

    //
    // Let the world know that we have entered this routine
    //
    I8xPrint((1, "%s-%s: %s\n", pDriverName, pFncFindWheelMouse, pEnter));

    //
    // Turn the device object into a device extension
    //
    deviceExtension = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;

    //
    // Are not going to find wheel mice?
    //
    if (deviceExtension->Configuration.EnableWheelDetection == 0) {

        I8xPrint((1,
            "%s-%s: Detection disabled in registry\n",
            pDriverName, pFncFindWheelMouse
            ));
        return STATUS_NO_SUCH_DEVICE;

    }

    //
    // Initialize some variables
    //
    for(i = 0; i < DUMP_COUNT; i++) {

        dumpData[i] = 0;

    }

    //
    // What is the point of this here???
    //
    KeStallExecutionProcessor(50);

    //
    // First check to see if we will try the 'better' method of detection
    //
    if (deviceExtension->Configuration.EnableWheelDetection == 1) {

        //
        // Begin sending commands to the mouse
        //
        while (pnpCommands[commandCount] != 0) {

            status = I8xPutBytePolled(
                (CCHAR) DataPort,
                WAIT_FOR_ACKNOWLEDGE,
                (CCHAR) MouseDeviceType,
                deviceExtension,
                pnpCommands[commandCount]
                );

            if (!NT_SUCCESS(status)) {

                I8xPrint((
                    1,
                    "%s-%s: failed write set sample rate #%d, status 0x%x\n",
                    pDriverName, pFncFindWheelMouse, commandCount,
                    status
                    ));

                //
                // Set up error log info
                //
                errorCode = I8042_SET_SAMPLE_RATE_FAILED;
                dumpData[0] = KBDMOU_COULD_NOT_SEND_PARAM;
                dumpData[1] = DataPort;
                dumpData[2] = I8042_WRITE_TO_AUXILIARY_DEVICE;
                dumpData[3] = pnpCommands[commandCount];
                dumpCount = 4;

                goto I8xFindWheelMouseExit;

            }

            //
            // Next command
            //
            commandCount++;
            uniqueErrorValue += 5;
            KeStallExecutionProcessor(50);

        } // while

        //
        // We should start to see the PNP string come back our way
        //
        for (idCount = 0; pnpID[idCount] != 0; idCount++) {

            //
            // Try to get a single character
            //
            for(i = 0; i < 5; i++) {

                status = I8xGetBytePolled(
                    (CCHAR) ControllerDeviceType,
                    deviceExtension,
                    &byte
                    );

                if (NT_SUCCESS(status)) {

                    //
                    // Read was successfull. We got a byte.
                    //
                    break;

                }

                //
                // If the read timed out, stall and retry.
                // If some other error occured handle it outside the loop
                //
                if(status == STATUS_IO_TIMEOUT) {

                    KeStallExecutionProcessor(50);

                } else {

                    break;

                }

            }

            //
            // if the operation wasn't successful or the characters don't
            // match, than try to flush the buffers
            //
            if (!NT_SUCCESS(status) || byte != pnpID[idCount]) {

                //
                // Byte didn't match
                //
                do {

                    //
                    // Wait a little bit
                    //
                    KeStallExecutionProcessor( 50 );

                    //
                    // Get a byte if there is one
                    //
                    status = I8xGetBytePolled(
                        (CCHAR) ControllerDeviceType,
                        deviceExtension,
                        &byte
                        );

                } while ( status != STATUS_IO_TIMEOUT );

                //
                // We are done here
                //
                return STATUS_NO_SUCH_DEVICE;

            } // if

        } // for

    } else if (deviceExtension->Configuration.EnableWheelDetection != 2) {

        //
        // We got a bogus id. Let's just assume that they meant to disable
        // the little detection routine
        //
        I8xPrint((1,
            "%s-%s: Detection disabled in registry\n",
            pDriverName, pFncFindWheelMouse
            ));

        //
        // Done
        //
        return STATUS_NO_SUCH_DEVICE;

    } // if

    //
    // Start the second detection routine, which will also enable the
    // device if present
    //
    commandCount = 0;
    while(enableCommands[commandCount] != 0) {

        //
        // Put the byte to the mouse
        //
        status = I8xPutBytePolled(
            (CCHAR) DataPort,
            WAIT_FOR_ACKNOWLEDGE,
            (CCHAR) MouseDeviceType,
            deviceExtension,
            enableCommands[commandCount]
            );

        if(!NT_SUCCESS(status)) {

            I8xPrint((
                1,
                "%s-%s: failed write set sample rate #%d, status 0x%x\n",
                pDriverName, pFncFindWheelMouse, commandCount,
                status
                ));

            //
            // Set up error log info
            //
            errorCode = I8042_SET_SAMPLE_RATE_FAILED;
            dumpData[0] = KBDMOU_COULD_NOT_SEND_PARAM;
            dumpData[1] = DataPort;
            dumpData[2] = I8042_WRITE_TO_AUXILIARY_DEVICE;
            dumpData[3] = enableCommands[commandCount];
            dumpCount = 4;

            goto I8xFindWheelMouseExit;
        }

        commandCount++;
        uniqueErrorValue += 5;
        KeStallExecutionProcessor(50);

    }

    //
    // Get the mouse ID
    //

    for(i = 0; i < 5; i++) {

        status = I8xGetBytePolled(
            (CCHAR) ControllerDeviceType,
            deviceExtension,
            &byte
            );

        if(NT_SUCCESS(status)) {

            //
            // Read was successful - the ID has been returned
            //
            break;

        }

        //
        // If the read timed out, stall and retry.
        // If some other error occured handle it outside the loop
        //
        if(status == STATUS_IO_TIMEOUT) {

            KeStallExecutionProcessor(50);

        } else {

            break;

        }

    }

    //
    // Check to see what we got
    //
    if ((!NT_SUCCESS(status)) ||
       ((byte != MOUSE_ID_BYTE) && (byte != WHEELMOUSE_ID_BYTE))) {

        I8xPrint((
            1,
            "%s-%s: failed ID, status 0x%x, byte 0x%x\n",
            pDriverName, pFncFindWheelMouse,
            status,
            byte
            ));

        //
        // Set up error log info
        //
        errorCode = I8042_MOU_RESET_RESPONSE_FAILED;
        dumpData[0] = KBDMOU_INCORRECT_RESPONSE;
        dumpData[1] = ControllerDeviceType;
        dumpData[2] = MOUSE_ID_BYTE;
        dumpData[3] = byte;
        dumpCount = 4;
        goto I8xFindWheelMouseExit;

    } else if( byte == WHEELMOUSE_ID_BYTE) {

        //
        // Update the HardwarePresent to show a Z mouse is operational,
        // and set the appropriate mouse type flags
        //

        deviceExtension->HardwarePresent |=
            (WHEELMOUSE_HARDWARE_PRESENT | MOUSE_HARDWARE_PRESENT);

        deviceExtension->Configuration.MouseAttributes.MouseIdentifier =
            WHEELMOUSE_I8042_HARDWARE;

        I8xPrint((
            1,
            "%s-%s: wheel mouse attached - running in wheel mode.\n",
            pDriverName, pFncFindWheelMouse
            ));

    } else {

        deviceExtension->HardwarePresent |= MOUSE_HARDWARE_PRESENT;

        I8xPrint((
            1,
            "%s-%s: Mouse attached - running in mouse mode.\n",
            pDriverName, pFncFindWheelMouse
            ));

    }

I8xFindWheelMouseExit:

    if (!NT_SUCCESS(status)) {

        //
        // The mouse initialization failed. Log an error.
        //
        if(errorCode != STATUS_SUCCESS) {

            errorLogEntry = (PIO_ERROR_LOG_PACKET)
                IoAllocateErrorLogEntry(
                    DeviceObject,
                    (UCHAR) (sizeof(IO_ERROR_LOG_PACKET) +
                            (dumpCount * sizeof(ULONG)))
                    );

            if(errorLogEntry != NULL) {

                errorLogEntry->ErrorCode = errorCode;
                errorLogEntry->DumpDataSize = (USHORT) dumpCount * sizeof(ULONG);
                errorLogEntry->SequenceNumber = 0;
                errorLogEntry->MajorFunctionCode = 0;
                errorLogEntry->IoControlCode = 0;
                errorLogEntry->RetryCount = 0;
                errorLogEntry->UniqueErrorValue = uniqueErrorValue;
                errorLogEntry->FinalStatus = status;
                for(i = 0; i < dumpCount; i++) {

                    errorLogEntry->DumpData[i] = dumpData[i];

                }
                IoWriteErrorLogEntry(errorLogEntry);

            }

        }

    }

    //
    // Initialize current mouse input packet state
    //
    deviceExtension->MouseExtension.PreviousSignAndOverflow = 0;
    deviceExtension->MouseExtension.InputState = MouseExpectingACK;

    //
    // Done
    //
    I8xPrint((2, "%s-%s: %s\n", pDriverName, pFncFindWheelMouse, pExit));
    return status;
}


VOID
I8xQueueCurrentInput(
    IN PDEVICE_OBJECT DeviceObject
    )

/*++

Routine Description:

    This routine queues the current input data to be processed by a
    DPC outside the ISR

Arguments:

    DeviceObject - Pointer to the device object

Return Value:

    None

--*/

{

    PDEVICE_EXTENSION   deviceExtension;
    UCHAR               buttonsDelta;
    UCHAR               previousButtons;

    deviceExtension = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;

    //
    // If the mouse is enabled, add the data to the InputData queue
    // and queue the ISR DPC.  One might wonder why we bother to
    // do all this processing of the mouse packet, only to toss it
    // away (i.e., not queue it) at this point.  The answer is that
    // this mouse provides no data to allow the driver to determine
    // when the first byte of a packet is received -- if the driver
    // doesn't process all interrupts from the start, there is no
    // way to keep MouseExtension.InputState in synch with hardware
    // reality.
    //
    if (deviceExtension->MouseEnableCount) {

        deviceExtension->MouseExtension.CurrentInput.UnitId = deviceExtension->MouseExtension.UnitId;

        if (!I8xWriteDataToMouseQueue(
                 &deviceExtension->MouseExtension,
                 &deviceExtension->MouseExtension.CurrentInput
                 )) {

            //
            // InputData queue overflowed.
            //
            // Queue a DPC to log an overrun error.
            //
            I8xPrint((
                1,
                "%s-I8042MouseInterruptService: queue overflow\n",
                pDriverName
                ));

            if (deviceExtension->MouseExtension.OkayToLogOverflow) {

                KeInsertQueueDpc(
                    &deviceExtension->ErrorLogMouseDpc,
                    (PIRP) NULL,
                    (PVOID) (ULONG) I8042_MOU_BUFFER_OVERFLOW
                    );
                deviceExtension->MouseExtension.OkayToLogOverflow =
                    FALSE;

            }

        } else if (deviceExtension->DpcInterlockMouse >= 0) {

           //
           // The ISR DPC is already executing.  Tell the ISR DPC it has
           // more work to do by incrementing DpcInterlockMouse.
           //
           deviceExtension->DpcInterlockMouse += 1;

        } else {

           //
           // Queue the ISR DPC.
           //
           KeInsertQueueDpc(
               &deviceExtension->MouseIsrDpc,
               DeviceObject->CurrentIrp,
               NULL
               );

       }

    }

    return;

}
