/*++

Copyright (c) 1995 Microsoft Corporation

Module Name:

    busdef.h

Abstract:

    Definitions for entire 1394 Bus driver

Author:

    Shaun Pierce (shaunp) 5-Sep-95

Environment:

    Kernel mode only

Notes:


Revision History:


--*/


//
// Various definitions
//

//
// These are all of the "buckets" we keep track of.  (0-32) are "dont care"
//
#define ALLOCATED_STORAGE_SLOTS             33

//
// The Required "bucket" where most normal Required offsets will be
//
#define REQUIRED_STORAGE_SLOT               32

//
// Amount of storage we're keeping track of internally (lower 32 bits of
// 1394 48 bit address)
//
#define AVAILABLE_STORAGE_WIDTH             32

// 
// The number of DeviceExtensions that could all own a Storage Entry
//
#define MAX_STORAGE_OWNERS                  16      

//
// The number of self ids we can keep track of in our topology map
//
#define SELFID_ARRAY_SZ                     256

//
// The tag we use to denote this is a notification entry
//
#define NOTIFICATION_TAG                    0xdeedcabe
#define NOTIFICATION_QUEUE_LENGTH           32

//
// Definition of Device Information Flags field
//
#define DEVICE_INFO_NEEDS_DEVICE_OBJECT     1
#define DEVICE_INFO_VALID_DEVICE_OBJECT     2

#define BUS_INFO_BLOCK_SZ                   4       // in ulongs
#define ROOT_DIR_LENGTH                     3       // in ulongs
#define UNIT_DIR_LENGTH                     2       // in ulongs
#define SELF_ID_COUNT_OFFSET                2       // in ulongs
#define HOST_VENDOR_ID                      0x534450
#define HOST_SPEC_ID                        0x767676
#define HOST_SOFTWARE_VERSION               0x010000
#define CONFIG_ROM_SZ                       256
#define TOPOLOGY_MAP_SZ                     0x400

#define DIAG_FLAGS_ENUM_TEST_DEVICE         0x00000001
#define DIAG_FLAGS_MASK                     (DIAG_FLAGS_ENUM_TEST_DEVICE)

                                            
#define min(a,b)                    (((a) < (b)) ? (a) : (b))
#define max(a,b)                    (((a) > (b)) ? (a) : (b))


//
// Various structure definitions
//

//
// IEEE 1212 Configuration Rom header definition
//
typedef struct _CONFIG_ROM_INFO {
    union {
        USHORT          CRI_CRC_Value:16;
        struct {
            UCHAR       CRI_Saved_Info_Length;
            UCHAR       CRI_Saved_CRC_Length;
        } Saved;
    } u;
    UCHAR               CRI_CRC_Length;
    UCHAR               CRI_Info_Length;
} CONFIG_ROM_INFO, *PCONFIG_ROM_INFO;

//
// IEEE 1212 State clear register definition
//
typedef struct _STATE_CLEAR {
    ULONG               SC_Unit_Depend:16;
    ULONG               SC_Gone:1;
    ULONG               SC_Reserved1:5;
    ULONG               SC_LinkOff:1;
    ULONG               SC_Cycle_Master:1;
    ULONG               SC_Lost:1;
    ULONG               SC_Dreq:1;
    ULONG               SC_Reserved2:1;
    ULONG               SC_Elog:1;
    ULONG               SC_Atn:1;
    ULONG               SC_Off:1;
    ULONG               SC_State:2;
} STATE_CLEAR, *PSTATE_CLEAR;


//
// IEEE 1212 Capabilities entry definition
//
typedef struct _BUS_INFO_BLOCK_CAPS {
    ULONG               BIBC_Reserved3:8;
    ULONG               BIBC_Reserved2:4;
    ULONG               BIBC_Max_Async_Payload:4;
    ULONG               BIBC_Cycle_Clock_Accuracy:8;
    ULONG               BIBC_Reserved1:4;
    ULONG               BIBC_Bus_Manager_Capable:1;
    ULONG               BIBC_Isoch_Capable:1;
    ULONG               BIBC_Cycle_Master_Capable:1;
    ULONG               BIBC_Isoch_Resource_Manager_Capable:1;
} BUS_INFO_BLOCK_CAPS, *PBUS_INFO_BLOCK_CAPS;

//
// IEEE 1212 Bus Info block node/chip Id definition
//
typedef struct _BUS_INFO_BLOCK_ID {
    ULONG               BIBI_ChipIdHi:8;
    ULONG               BIBI_NodeVendorId:24;
} BUS_INFO_BLOCK_ID, *PBUS_INFO_BLOCK_ID;

//
// IEEE 1212 Directory definition
//
typedef struct _DIRECTORY_INFO {
    union {
        USHORT          DI_CRC;
        USHORT          DI_Saved_Length;
    } u;
    USHORT              DI_Length;
} DIRECTORY_INFO, *PDIRECTORY_INFO;

//
// IEEE 1212 Node Capabilities entry definition
//
typedef struct _NODE_CAPABILITES {
    ULONG               NC_Init:1;                  // These can be found
    ULONG               NC_Ded:1;                   // in the IEEE 1212 doc
    ULONG               NC_Off:1;               
    ULONG               NC_Atn:1;
    ULONG               NC_Elo:1;
    ULONG               NC_Reserved1:1;
    ULONG               NC_Drq:1;
    ULONG               NC_Lst:1;
    ULONG               NC_Fix:1;
    ULONG               NC_64:1;
    ULONG               NC_Prv:1;
    ULONG               NC_Bas:1;
    ULONG               NC_Ext:1;
    ULONG               NC_Int:1;
    ULONG               NC_Ms:1;
    ULONG               NC_Spt:1;
    ULONG               NC_Reserved2:8;
    ULONG               NC_Key:8;
} NODE_CAPABILITIES, *PNODE_CAPABILITIES;

//
// IEEE 1212 Offset entry definition
//
typedef struct _OFFSET_ENTRY {
    ULONG               OE_Offset:24;
    ULONG               OE_Key:8;
} OFFSET_ENTRY, *POFFSET_ENTRY;

//
// IEEE 1212 Immediate entry definition
//
typedef struct _IMMEDIATE_ENTRY {
    ULONG               IE_Value:24;
    ULONG               IE_Key:8;
} IMMEDIATE_ENTRY, *PIMMEDIATE_ENTRY;

//
// Holds the basic config rom that this PC will expose on the bus
//
typedef struct _HOST_CONFIG_ROM {
    CONFIG_ROM_INFO     HCR_Info;
    ULONG               HCR_Signiture;              // bus info block
    BUS_INFO_BLOCK_CAPS HCR_Capabilities;           //  "   "   "
    ULONG               HCR_UniqueId[2];            //  "   "   "
    DIRECTORY_INFO      HCR_RootInfo;
    IMMEDIATE_ENTRY     HCR_ModuleVendorId;
    NODE_CAPABILITIES   HCR_NodeCapabilities;
    OFFSET_ENTRY        HCR_UnitOffset;
    DIRECTORY_INFO      HCR_UnitInfo;
    IMMEDIATE_ENTRY     HCR_UnitSpecId;
    IMMEDIATE_ENTRY     HCR_UnitSwVersion;
} HOST_CONFIG_ROM, *PHOST_CONFIG_ROM;
    

//
// This holds information about every node on this bus
//
typedef struct _DEVICE_INFORMATION {

    //
    // Holds the flags relevant to this node
    //
    ULONG Flags;

    //
    // Holds the current generation count
    //
    ULONG Generation;

    //
    // Holds the 16 bit BusId/NodeId Node Address for this device
    //
    NODE_ADDRESS NodeAddress;

    //
    // Holds Device Object for this device
    //
    PDEVICE_OBJECT DeviceObject;

    //
    // Holds the 24 bit offset to the Unit directory for this node
    //
    ULONG UnitDirectoryInformation;

    //
    // Holds the offset within the config rom where this unit directory
    // pointer was located.  Important because the offset contained within
    // the UnitDirectoryInformation is a 24 bit relative quadlet number
    //
    PVOID UnitDirectoryLocation;

    //
    // Holds the Configuration Rom
    //
    PCONFIG_ROM ConfigRom;

} DEVICE_INFORMATION, *PDEVICE_INFORMATION;


typedef struct _TREE_INFORMATION {

    //  
    // Holds the parent topology map index 
    //
    ULONG ParentIndex;

} TREE_INFORMATION, *PTREE_INFORMATION;


typedef struct _STORAGE_ENTRY {

    //
    // Holds the list of nodes that are available (or reserved) depending
    // on what head we're on
    //
    LIST_ENTRY SE_NodeList;

    //
    // Holds the list of other storage entries that overlap with this one
    //
    LIST_ENTRY SE_OverlapList;

    //
    // Holds whether or not this entry is available 
    //
    BOOLEAN SE_bAvailable;

    //
    // Holds the 2 ^ k value of this block size (in words).  So to manage
    // a 32 bit address space, the kVal == 31.
    //
    ULONG SE_kVal;

    //
    // Holds the length of data in the storage entry (may be < (2^k)*2)
    //
    ULONG SE_Length;

    //
    // Holds the memory location of the first word
    //
    PVOID SE_StartWord;

    //
    // Holds the 1394 Offset that this storage entry represents
    //
    ADDRESS_OFFSET SE_1394Offset;

    //
    // Holds the owner of this storage entry
    //
    PVOID SE_Owner;

    //
    // Mdl that actually maps to this storage entry
    //
    PMDL SE_Mdl;

    //
    // Access type as requested by 1394 AllocateAddressRange Irb
    //
    ULONG SE_fulAccessType;

    //
    // Notification options as requested by 1394 AllocateAddressRange Irb
    //
    ULONG SE_fulNotificationOptions;

    //
    // Callback routine as requested by 1394 AllocateAddressRange Irb
    //
    PBUS_NOTIFICATION_ROUTINE SE_Callback;

    //
    // Context as requested by 1394 AllocateAddressRange Irb
    //
    PVOID SE_Context;


} STORAGE_ENTRY, *PSTORAGE_ENTRY;


typedef struct _NOTIFICATION_ENTRY {

    //
    // This holds the linked list of what this notification entry is on
    //
    LIST_ENTRY NotificationList;

    //
    // This holds the Tag for this notification entry
    //
    ULONG Tag;

    //
    // This holds the length of the operation
    //
    ULONG ulLength;

    //
    // This holds the offset within the Mdl that the access occurred at
    //
    ULONG ulOffset;

    //
    // This holds the notification option that triggered this callback
    //
    ULONG fulNotificationOption;
    
    //
    // This holds the storage entry who requested a notification
    //
    PSTORAGE_ENTRY StorageEntry;


} NOTIFICATION_ENTRY, *PNOTIFICATION_ENTRY;


typedef struct _BUS_INSTANCE_DATA {

    //
    // Driver Object of the port driver
    //
    PDRIVER_OBJECT PortDriverObject;

    //
    // Holds the port driver's device object 
    //
    PDEVICE_OBJECT PortDeviceObject;

    //
    // Holds the physical device object handled to the port driver (PDO)
    //
    PDEVICE_OBJECT PhysicalDeviceObject;

    //
    // Holds the port driver's device extension
    //
    PVOID PortDeviceExtension;

    //
    // Holds the capability bits for port driver underneath us
    //
    ULONG PortCapabilities;

    //
    // Holds the speed flags that the port driver underneath us supports
    //
    ULONG PortSpeed;    

    //
    // Holds the max Async read packet the port can accept 
    //
    ULONG PortMaxReadPacket;

    //
    // Holds the max Async write packet the port can accept 
    //
    ULONG PortMaxWritePacket;

    //
    // Holds the number of devices we know about on this bus
    //
    ULONG TotalDevices;

    //
    // Holds the spinlock that protects access to the generation count (below)
    //
    KSPIN_LOCK GenerationSpinLock;

    //
    // Holds the bus reset generation number
    //
    ULONG CurrentGeneration;

    //
    // Holds the local (host controller's node address)
    //
    NODE_ADDRESS LocalNodeAddress;

    //
    // Holds the Isoch resource managers node address
    //
    NODE_ADDRESS IsochMgrNodeAddress;

    //
    // Pointer to our own correctly formatted internal topology map
    //
    PTOPOLOGY_MAP TopologyMap;

    //
    // Pointer to our own correctly formatted internal speed map
    //
    PSPEED_MAP SpeedMap;

    //
    // Holds the length of our host configuration rom
    //
    ULONG ConfigRomLength;

    //
    // Pointer to our host configuration rom
    //
    PVOID ConfigRom;

    //
    // Holds the parent information for each node on the tree
    //
    TREE_INFORMATION TreeInfo[MAX_LOCAL_NODES];

    //
    // Holds the global current index value used in determining who a
    // nodes parent is.
    //
    ULONG TreeCurrentIndex;
    
    //
    // Holds the array of Available storage locations, grouped by size
    //
    LIST_ENTRY AvailableStorage[AVAILABLE_STORAGE_WIDTH - 1];

    //
    // Holds the list of bandwidth entries
    //
    LIST_ENTRY BandwidthList;

    //
    // Holds the slot in memory array should be used for allocation next
    //
    USHORT NextSlotToBeUsed;

    //
    // Holds spin lock that needs to be acquired before doing Indication
    // stuff.  We also acquire it if we're Isoch mgr and need to play
    // around with channels available or bandwdith available.
    //
    KSPIN_LOCK IndicationSpinLock;

    //
    // Holds the pointers to address mapping entries in this slot
    //
    LIST_ENTRY AllocatedStorage[ALLOCATED_STORAGE_SLOTS];

    //
    // Holds the bandwidth units currently available
    //
    ULONG BandwidthUnitsRemaining;

    //
    // Holds the bitmask of Isoch channels currently available
    //
    LARGE_INTEGER ChannelsAvailable;        
    
    //
    // Holds whether or not the bus driver is the Isoch resource manager
    //
    BOOLEAN bIsochResourceManager;          

    //
    // Holds whether or not the bus driver is the Bus manager
    //
    BOOLEAN bBusManager;

    //
    // Used to determine if we've been properly shutdown
    //
    BOOLEAN bShutdown;

    //
    // Holds the pointer to the port driver's PnP function
    //
    PDRIVER_DISPATCH PortPnP;

    //
    // Holds array of device information on the bus
    //
    DEVICE_INFORMATION DeviceArray[MAX_LOCAL_NODES];

    //
    // Holds the event that is to be signaled once the enumeration process
    // is complete
    //
    KEVENT EnumerationCompleteEvent;

    //
    // Holds the nodes we need to read during the enumeration process
    //
    LARGE_INTEGER NodesToEnumerate;

    //
    // Holds work area to play with lock transaction data (tLabel based)
    //
    ULONG LockWorkArea[2*64];

    //
    // Holds the notification DPC used when calling notify routines
    //
    KDPC NotificationDpc;

    //
    // Holds spin lock that needs to be acquired before playing around
    // with the available notification list
    //
    KSPIN_LOCK AvailableNotificationsSpinLock;

    //
    // Holds spin lock that needs to be acquired before playing around
    // with the notifications awaiting completion list
    //
    KSPIN_LOCK NotificationsAwaitingCompletionSpinLock;

    //
    // Holds spin lock that needs to be acquired before playing around
    // with the pending notification list
    //
    KSPIN_LOCK PendingNotificationsSpinLock;

    //
    // Holds the list of available notifications
    //
    LIST_ENTRY AvailableNotifications;

    //
    // Holds the list of notifications that haven't been completed by port yet
    //
    LIST_ENTRY NotificationsAwaitingCompletion;

    //
    // Holds the list of notifications that haven't been serviced by DPC yet
    //
    LIST_ENTRY PendingNotifications;

    //
    // Holds the pointer to the Notification queue
    //
    PNOTIFICATION_ENTRY Notifications;

    //
    // Holds who the bus manager is.  Valid when we're Isoch resource manager
    //
    ULONG BusManagerId;

    //
    // This register has valid data when we're the Isoch resource manager
    //
    ULONG IsochMgrBandwidthUnits;

    //
    // This register has valid data when we're the Isoch resource manager
    //
    LARGE_INTEGER IsochMgrChannelsAvailable;

    //
    // Holds the state clear register.  It's byte swapped.
    //
    ULONG StateClear;

    //
    // Holds the Compliance stall value we sucked out of the registry
    //
    ULONG ComplianceFlags;

    //
    // Holds the Diagnostic value we sucked out of the registry
    //
    ULONG DiagnosticFlags;


} BUS_INSTANCE_DATA, *PBUS_INSTANCE_DATA;


typedef struct _BUS_RESET_WORK_ITEM {

    //
    // Holds the BusInstanceData pointer (need an extra context)
    //
    PBUS_INSTANCE_DATA BusInstanceData;

    //
    // Holds the work item used in scheduling by the OS
    //
    WORK_QUEUE_ITEM WorkItem;

} BUS_RESET_WORK_ITEM, *PBUS_RESET_WORK_ITEM;


