/*++

Copyright (c) 1996 Microsoft Corporation

Module Name:

    sonydcr.c

Abstract:

    This is a skeleton driver for the Sony VX-1000 Camcorder.
    The main purpose of this code is to show how a driver gets loaded, and
    how it submits requests to the 1394 bus driver.

Author:
    
    Shaun Pierce (shaunp) 25-May-96

Environment:

    Kernel mode only

Revision History:


--*/

#include "wdm.h"
#include "1394.h"
#include "sonydcr.h"





NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )

/*++

Routine Description:

    This where life begins for a driver.  All that is necessary to do
    is fill in the necessary entry points (Plug & Play, Open, Read, etc)
    in the Driver Object and return.  You're not allowed to touch any 
    hardware at this point, or submit requests to the 1394 bus driver.

Arguments:

    DriverObject - Pointer to driver object created by system.

    RegistryPath - Pointer to the Unicode name of the registry path
        for this driver.

Return Value:

    The function value is the final status from the initialization operation.

--*/

{

    KdPrint(("SonyDcr DriverEntry: enter\n"));

    //
    // We've got PnP around, so all we need to do is fill in some 
    // DriverObject crap and return...
    //

    DriverObject->DriverExtension->AddDevice = SonyDcrPnPAddDevice;
    DriverObject->MajorFunction[IRP_MJ_PNP]  = SonyDcrPnP;

    //
    // NOTE: If you want to receive requests from other device drivers,
    // or from other User Mode applications - then you should also be
    // filling in the Open/Close/Read/Write/IOCtl Driver entry points 
    // so that requests can get to you.
    //

    KdPrint(("SonyDcr DriverEntry: exit\n"));

    return (STATUS_SUCCESS);

}


NTSTATUS
SonyDcrPnPAddDevice(
    IN     PDRIVER_OBJECT  DriverObject,
    IN     PDEVICE_OBJECT  PhysicalDeviceObject
    )

/*++

Routine Description:

    This routine is called by the Operating System to create a new
    instance of a Sony 1394 Desktop camera.  Still can't touch hardware
    at this point, or submit requests to the 1394 bus driver.  But at
    least at this point we get a handle to the 1394 bus driver, which
    we'll use in submitting requests in the future.

Arguments:

    DriverObject - Pointer to our driver object

    PhysicalDeviceObject - Pointer to Device Object created by parent

Return Value:

    Status is returned.

--*/

{

    NTSTATUS status;
    PDEVICE_OBJECT DeviceObject;
    PSONYDCR_EXTENSION DeviceExtension;


    KdPrint(("SonyDcrPnPAddDevice: enter\n"));

    //
    // Create our device object with a our own Sony specific device
    // extension.  
    //

    status = IoCreateDevice(
                DriverObject,
                sizeof(SONYDCR_EXTENSION),
                NULL,
                FILE_DEVICE_UNKNOWN,
                0,
                FALSE,
               &DeviceObject
                );

    if (!NT_SUCCESS(status)) {

        return (status);

    }

    //
    // Attach ourself into the driver stack on top of our parent (PCI).
    //

    DeviceExtension = DeviceObject->DeviceExtension;
    DeviceExtension->BusDeviceObject = IoAttachDeviceToDeviceStack(DeviceObject, PhysicalDeviceObject);

    if (!DeviceExtension->BusDeviceObject) {

        IoDeleteDevice(DeviceObject);
        return (STATUS_NO_SUCH_DEVICE);

    }

    //
    // If the registration went ok, then let's setup some crap in our
    // device extension.
    //

    DeviceExtension->bShutdown = FALSE;
    DeviceExtension->PhysicalDeviceObject = PhysicalDeviceObject;

    KdPrint(("SonyDcrPnPAddDevice: exit\n"));

    return (STATUS_SUCCESS);

}


NTSTATUS
SonyDcrPnP(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )

/*++

Routine Description:

    This routine will receive the various Plug N Play messages.  It is
    here that we start our device, stop it, etc.  Safe to submit requests
    to the 1394 bus driver.

Arguments:

    DeviceObject - Pointer to class device object.

    Irp - Pointer to the request packet.

Return Value:

    Status is returned.

--*/

{

    PIRB Irb;
    KEVENT Event;
    NTSTATUS status;
    PIO_STACK_LOCATION IrpStack;
    PIO_STACK_LOCATION NextIrpStack;
    PSONYDCR_EXTENSION DeviceExtension;


    //
    // Suck the minor code out of our Irp Stack so we know what
    // PnP function we're supposed to do
    //

    DeviceExtension = DeviceObject->DeviceExtension;
    IrpStack = IoGetCurrentIrpStackLocation(Irp);

    switch (IrpStack->MinorFunction) {


        case IRP_MN_START_DEVICE:

            KdPrint(("SonyDcrPnP: MN_START_DEVICE\n"));

            //
            // Before we start initializing our device, we must  
            // call to the layer below us first.
            //

            NextIrpStack = IoGetNextIrpStackLocation(Irp);
            NextIrpStack->MajorFunction = IrpStack->MajorFunction;
            NextIrpStack->MinorFunction = IrpStack->MinorFunction;
            RtlCopyMemory(&NextIrpStack->Parameters, &IrpStack->Parameters, sizeof(IrpStack->Parameters.Others));

            KeInitializeEvent(&Event, SynchronizationEvent, FALSE);

            IoSetCompletionRoutine(
                Irp,
                SonyDcrSynchCompletionRoutine,
               &Event,
                TRUE,
                TRUE,
                TRUE
                );

            //
            // Call down to the 1394 bus driver.
            //

            status = IoCallDriver(
                        DeviceExtension->BusDeviceObject,
                        Irp
                        );

            if (status == STATUS_PENDING) {

                //
                // Still pending, wait for the IRP to complete
                //

                KeWaitForSingleObject(
                   &Event,
                    Executive,
                    KernelMode,
                    FALSE,
                    NULL
                    );

            }

            if (status) {

                Irp->IoStatus.Status = status;
                Irp->IoStatus.Information = 0;

                IoCompleteRequest(Irp, IO_NO_INCREMENT);
                return (status);

            }

            //
            // Do what you need to do here in order to get the device
            // initialized.  Just for grins, we'll issue a read
            // request to see if we can read '1394' from the other side
            //

            Irb = ExAllocatePool(NonPagedPool, sizeof(IRB));

            if (!Irb) {

                Irp->IoStatus.Status = status = STATUS_INSUFFICIENT_RESOURCES;
                Irp->IoStatus.Information = 0;

                IoCompleteRequest(Irp, IO_NO_INCREMENT);
                return (status);
                
            }

            //
            // Get the current generation count first
            //

            Irb->FunctionNumber = REQUEST_GET_GENERATION_COUNT;
            Irb->Flags = 0;

            status = SonyDcrSubmitIrpSynch(DeviceExtension, Irp, Irb);

            if (status) {

                KdPrint(("SonyDcrPnP: Error %x while trying to get generation number\n", status));
                ExFreePool(Irb->u.AsyncRead.Mdl);
                ExFreePool(Irb);
                Irp->IoStatus.Status = status;
                Irp->IoStatus.Information = 0;

                IoCompleteRequest(Irp, IO_NO_INCREMENT);
                return (status);

            }


            DeviceExtension->CurrentGeneration = Irb->u.GetGenerationCount.GenerationCount;
            KdPrint(("SonyDcrPnP: Generation number = %x\n", DeviceExtension->CurrentGeneration));

            //
            // Now that we have the current generation count, do a read
            // on where the 1394 signiture resides
            //

            Irb->FunctionNumber = REQUEST_ASYNC_READ;
            Irb->Flags = 0;
            Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_High = INITIAL_REGISTER_SPACE_HI;
            Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_Low = (INITIAL_REGISTER_SPACE_LO | CONFIG_ROM_LOCATION) + sizeof(ULONG);
            Irb->u.AsyncRead.nNumberOfBytesToRead = sizeof(ULONG);
            Irb->u.AsyncRead.nBlockSize = 0;
            Irb->u.AsyncRead.fulFlags = 0;
            Irb->u.AsyncRead.ulGeneration = DeviceExtension->CurrentGeneration;
            Irb->u.AsyncRead.Mdl = MmCreateMdl(NULL, &DeviceExtension->BaseRegister, sizeof(ULONG));
            MmBuildMdlForNonPagedPool(Irb->u.AsyncRead.Mdl);

            status = SonyDcrSubmitIrpSynch(DeviceExtension, Irp, Irb);

            if (status) {

                KdPrint(("SonyDcrPnP: Error %x while trying to read rom signiture\n", status));
                ExFreePool(Irb->u.AsyncRead.Mdl);
                ExFreePool(Irb);
                Irp->IoStatus.Status = status;
                Irp->IoStatus.Information = 0;

                IoCompleteRequest(Irp, IO_NO_INCREMENT);
                return (status);

            }

            //
            // We don't really care what we read, so just say success
            //

            KdPrint(("SonyDcrPnP: Successfully read %x from my device\n", DeviceExtension->BaseRegister));
            ExFreePool(Irb->u.AsyncRead.Mdl);
            ExFreePool(Irb);
            Irp->IoStatus.Status = status = STATUS_SUCCESS;
            Irp->IoStatus.Information = 0;

            IoCompleteRequest(Irp, IO_NO_INCREMENT);
            return (status);


        case IRP_MN_STOP_DEVICE:

            //
            // Do whatever you need to do to shutdown the device.  We
            // could still be issuing requests to the 1394 bus driver
            // at this point, but we won't do anything for this example
            // driver.
            //

            KdPrint(("SonyDcrPnP: MN_STOP_DEVICE\n"));

            DeviceExtension->bShutdown = TRUE;

            NextIrpStack = IoGetNextIrpStackLocation(Irp);
            NextIrpStack->MajorFunction = IrpStack->MajorFunction;
            NextIrpStack->MinorFunction = IrpStack->MinorFunction;
            RtlCopyMemory(&NextIrpStack->Parameters, &IrpStack->Parameters, sizeof(IrpStack->Parameters.Others));

            return (IoCallDriver(DeviceExtension->BusDeviceObject, Irp));


        case IRP_MN_REMOVE_DEVICE:

            KdPrint(("SonyDcrPnP: MN_REMOVE_DEVICE\n"));

            //
            // We're being asked to remove the device.  Let's make sure
            // we've been stopped first (these might arrive out of order)
            //

            if (!DeviceExtension->bShutdown) {

                //
                // We haven't been stopped yet, so let's stop ourself first.
                //

                DeviceExtension->bShutdown = TRUE;

            }

            //
            // Clean ourself out of the driver stack layer
            //

            IoDetachDevice(DeviceExtension->BusDeviceObject);
            IoDeleteDevice(DeviceObject);

            NextIrpStack = IoGetNextIrpStackLocation(Irp);
            NextIrpStack->MajorFunction = IrpStack->MajorFunction;
            NextIrpStack->MinorFunction = IrpStack->MinorFunction;
            RtlCopyMemory(&NextIrpStack->Parameters, &IrpStack->Parameters, sizeof(IrpStack->Parameters.Others));

            return (IoCallDriver(DeviceExtension->BusDeviceObject, Irp));


        case IRP_MN_BUS_RESET:

            KdPrint(("SonyDcrPnP: IRP_MN_BUS_RESET\n"));

            //
            // The 1394 bus driver is telling us that a bus reset has
            // occurred.  At the very least, we should get the current
            // generation number as well as doing whatever else we
            // need to do after a bus reset occurs.
            //

            DeviceExtension->CurrentGeneration = *((PULONG) &IrpStack->Parameters.Others.Argument4);
            Irp->IoStatus.Status = status = STATUS_SUCCESS;
            Irp->IoStatus.Information = 0;

            IoCompleteRequest(Irp, IO_NO_INCREMENT);
            return (status);


        default:

            NextIrpStack = IoGetNextIrpStackLocation(Irp);
            NextIrpStack->MajorFunction = IrpStack->MajorFunction;
            NextIrpStack->MinorFunction = IrpStack->MinorFunction;
            RtlCopyMemory(&NextIrpStack->Parameters, &IrpStack->Parameters, sizeof(IrpStack->Parameters.Others));

            return (IoCallDriver(DeviceExtension->BusDeviceObject, Irp));

    }

}


NTSTATUS
SonyDcrSubmitIrpSynch(
    PSONYDCR_EXTENSION DeviceExtension,
    PIRP Irp,
    PIRB Irb
    )

/*++

Routine Description:

    This routine submits an Irp synchronously to the bus driver.  We'll
    wait here til the Irp comes back

Arguments:

    DeviceExtension - Pointer to my local device extension

    Irp - Pointer to Irp we're sending down to the port driver synchronously

    Irb - Pointer to Irb we're submitting to the port driver

Return Value:

    Status is returned from Irp

--*/

{


    ULONG Retries=RETRY_COUNT;
    KEVENT Event;
    NTSTATUS status;
    PIO_STACK_LOCATION NextIrpStack;


    KdPrint(("SonyDcrSubmitIrpSynch: enter\n"));

    do {

        NextIrpStack = IoGetNextIrpStackLocation(Irp);
        NextIrpStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
        NextIrpStack->Parameters.DeviceIoControl.IoControlCode = IOCTL_1394_CLASS;
        NextIrpStack->Parameters.Others.Argument1 = Irb;

        KeInitializeEvent(&Event, SynchronizationEvent, FALSE);

        IoSetCompletionRoutine(
            Irp,
            SonyDcrSynchCompletionRoutine,
           &Event,
            TRUE,
            TRUE,
            TRUE
            );

        status = IoCallDriver(
                    DeviceExtension->BusDeviceObject,
                    Irp
                    );

        if (status == STATUS_PENDING) {

            //
            // Still pending, wait for the IRP to complete
            //

            KdPrint(("SonyDcrSubmitIrpSynch: Irp is pending, waiting...\n"));
            KeWaitForSingleObject(
               &Event,
                Executive,
                KernelMode,
                FALSE,
                NULL
                );

        }

    } while ((Irp->IoStatus.Status == STATUS_TIMEOUT) && (Retries--));

    KdPrint(("SonyDcrSubmitIrpSynch: exit\n"));
    return (Irp->IoStatus.Status);

}


NTSTATUS
SonyDcrSynchCompletionRoutine(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PKEVENT Event
    )

/*++

Routine Description:

    This routine is for use with synchronous IRP processing.  
    All it does is signal an event, so the driver knows it
    can continue.

Arguments:

    DriverObject - Pointer to driver object created by system.

    Irp - Irp that just completed

    Event - Event we'll signal to say Irp is done

Return Value:

    None.

--*/

{

    KeSetEvent((PKEVENT) Event, 0, FALSE);
    return (STATUS_MORE_PROCESSING_REQUIRED);

}


