/*++

Copyright (c) 1996, 1997  Microsoft Corporation

Module Name:

   adapter.cpp
   
Abstract:

   This module contains the adapter specific code for implementing the driver for
   the Diamond Multimedia Monster Sound card.
   
Author:

   Ken Cooper (kencoope)   25-Mar-1997
   
Environment:

   Kernel Mode Only.
   
Revision History:

      
--*/
#define PUT_GUIDS_HERE

#define STR_MODULENAME "monsterAdapter"

#include "portcls.h"
#include "ksdebug.h"

#if (DBG)
#define SUCCEEDS(s) ASSERT(NT_SUCCESS(s))
#else
#define SUCCEEDS(s) (s)
#endif

// miniport init functions referenced in StartDevice.
#define MAX_MINIPORTS 2

NTSTATUS
CreateMiniportWavePciMonster
(
    OUT     PUNKNOWN *  Unknown,
    IN      REFCLSID,
    IN      PUNKNOWN    UnknownOuter    OPTIONAL,
    IN      POOL_TYPE   PoolType
);

// AddDevice is forward-referenced from DriverEntry.
NTSTATUS
AddDevice
(
    IN PVOID        Context1,
    IN PVOID        Context2
);

// StartDevice is forward-referenced from AddDevice.
NTSTATUS
StartDevice
(
    IN PVOID                    Context1,
    IN PVOID                    Context2,
    IN PRESOURCELIST            ResourceList
);

// AssignResources is forward-referenced from StartDevice.
NTSTATUS
AssignResources
(
    IN  PRESOURCELIST   ResourceList,           // All resources.
    OUT PRESOURCELIST * ResourceListMonsterWave // Monster wave resources.
);


#pragma code_seg("PAGE")

extern "C"
NTSTATUS
DriverEntry
(
    IN PVOID        Context1,
    IN PVOID        Context2
)
/*++

Routine Description:

   This function is called by the operating system when the driver is loaded.
   All adapter drivers can use this code without change.  Note that DriverEntry
   must have "C" linkage.
   
Arguments:

   Context1 - Context for the class driver.
   
   Context2 - Context for the class driver.
      
Return Value:

   STATUS_SUCCESS - success, otherwise an error code.
   
--*/        
{
   PAGED_CODE();

   // Tell the class driver to initialize the driver.
   return InitializeAdapterDriver(Context1,Context2,AddDevice);
}

NTSTATUS
AddDevice
(
    IN PVOID        Context1,
    IN PVOID        Context2
)
/*++

Routine Description:

   This function is called by the operating system when the device is added.
   All adapter drivers can use this code without change.
   
Arguments:

   Context1 - Context for the class driver.
   
   Context2 - Context for the class driver.
      
Return Value:

   STATUS_SUCCESS - success, otherwise an error code.
   
--*/        
{
   PAGED_CODE();

   // Tell the class driver to add the device.   
   return AddAdapterDevice(Context1,Context2,StartDevice,MAX_MINIPORTS);
}

NTSTATUS
InstallSubdevice
(
    IN      PVOID               Context1,
    IN      PVOID               Context2,
    IN      PWCHAR              Name,
    IN      REFCLSID            PortClassId,
    IN      PFNCREATEINSTANCE   MiniportCreate,
    IN      PRESOURCELIST       ResourceList
)
/*++

Routine Description:

   This function creates a port driver and miniport driver and calls
   RegisterSubdevice to bind them with a name and their resources.
   
Arguments:

   Context1 - Context for the class driver.
   
   Context2 - Context for the class driver.
   
   Name - The name under which to register the miniport.
   
   PortClassId - The CLSID of the port to create.
   
   MiniportCreate - A pointer to the miniport create function.
   
   ResourceList - A pointer to the resource list for the miniport.
      
Return Value:

   STATUS_SUCCESS - success, otherwise an error code.
   
--*/        
{
   PAGED_CODE();

   ASSERT(Context1);
   ASSERT(Context2);
   ASSERT(Name);
   ASSERT(MiniportCreate);
   ASSERT(ResourceList);

   PUNKNOWN     miniport;
   PPORT        port;
   NTSTATUS     ntStatus = NewPort(&port,PortClassId);

   if (NT_SUCCESS(ntStatus))
   {
        ntStatus =
            MiniportCreate
            (
                &miniport,
                PortClassId,    // Not used.
                NULL,
                NonPagedPool
            );

        if (NT_SUCCESS(ntStatus))
        {
            ntStatus =
                port->Init(Context1,Context2,miniport,NULL,ResourceList);
            miniport->Release();

            if (NT_SUCCESS(ntStatus))
            {
                ntStatus =
                    RegisterSubdevice
                    (
                        Context1,
                        Context2,
                        Name,
                        port
                    );
            }
        }

        port->Release();
    }

    return ntStatus;
}

NTSTATUS
StartDevice
(
    IN PVOID                Context1,
    IN PVOID                Context2,
    IN PRESOURCELIST        ResourceList
)
/*++

Routine Description:

   This function is called by the operating system when the device is started.
   It is responsible for starting the miniports.  This code is specific to
   the adapter because it calls out miniports for functions that are specific
   to the adapter.
   
Arguments:

   Context1 - Context for the class driver.
   
   Context2 - Context for the class driver.
   
   ResourceList - A pointer to the list of hardware resources.
      
Return Value:

   STATUS_SUCCESS - success, otherwise an error code.
   
--*/        
{
    PAGED_CODE();

    ASSERT(Context1);
    ASSERT(Context2);
    ASSERT(ResourceList);

    PRESOURCELIST       resourceListMonsterWave;
    NTSTATUS            ntStatus;

    // Process resources
    ntStatus =
        AssignResources
        (
            ResourceList,
            &resourceListMonsterWave
        );

	// Start the Monster wave miniport if it exists.
	if (resourceListMonsterWave)
	{
	    if (NT_SUCCESS(ntStatus))
	    {
	        ntStatus =
	            InstallSubdevice
	            (
	                Context1,
	                Context2,
	                L"Wave",
	                CLSID_PortWavePci,
	                CreateMiniportWavePciMonster,
	                resourceListMonsterWave
	            );
		 }
		 resourceListMonsterWave->Release();
    }

    return ntStatus;
}

NTSTATUS
AssignResources
(
    IN  PRESOURCELIST   ResourceList,           // All resources.
    OUT PRESOURCELIST * ResourceListMonsterWave // Monster wave resources.
)
/*++

Routine Description:

   This function assigns the list of resources to the various functions on
   the card.  This code is specific to the adapter.
   
   THIS FUNCTION ALLOCATES MEMORY WHICH MUST BE FREED BY THE CALLER.  Any
   non-NULL pointers returned through the out-parameters must be freed using
   ResourceList->Release().
   
Arguments:

   ResourceList - A pointer to the list of hardware resources.
   
   ResourceListMonsterWave - A pointer to the list of hardware resources for
   the Monster wave miniport.
      
Return Value:

   STATUS_SUCCESS - success, otherwise an error code.
   
--*/        
{
    PAGED_CODE();

    ASSERT(ResourceList);

    ULONG       countIO;
    ULONG       countIRQ;
    ULONG       countDMA;
    NTSTATUS    ntStatus = STATUS_SUCCESS;

    // Get counts for the types of resources.
    countIO = ResourceList->NumberOfPorts();
    countIRQ = ResourceList->NumberOfInterrupts();
    countDMA = ResourceList->NumberOfDmas();

    // Build the resource list for the Monster wave I/O.
    *ResourceListMonsterWave = NULL;
		
    // verify adequate resources (2-IO and 1-IRQ)
	if( (countIO < 2) ||
		(countIRQ < 1))
	{
		_DbgPrintF(DEBUGLVL_ERROR,("configuration error"));
		ntStatus = STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	if(NT_SUCCESS(ntStatus))
	{
		// create resource sublist for Monster wave miniport
		ntStatus = NewResourceSublist( ResourceListMonsterWave,
                                       NULL,
                                       PagedPool,
                                       ResourceList,
                                       countIO + countIRQ);	
    	if (NT_SUCCESS(ntStatus))
		{
            // Add an IO port
            SUCCEEDS((*ResourceListMonsterWave)->
                     AddPortFromParent(ResourceList,0));
				
            // Add another IO port
            SUCCEEDS((*ResourceListMonsterWave)->
                     AddPortFromParent(ResourceList,1));

            // Add the IRQ
            SUCCEEDS((*ResourceListMonsterWave)->
                     AddInterruptFromParent(ResourceList,0));
	    }
    }

    // Clean up if failure occurred.
    if (! NT_SUCCESS(ntStatus))
    {
        if (*ResourceListMonsterWave)
        {
            (*ResourceListMonsterWave)->Release();
            *ResourceListMonsterWave = NULL;
        }
    }

    return ntStatus;
}

#pragma code_seg()

int __cdecl
_purecall( void )
/*++

Routine Description:

   The C++ compiler loves me.
   TODO: Figure out how to put this into portcls.sys
   
Arguments:

   None.
      
Return Value:

   0.
   
--*/        
{
    ASSERT( !"Pure virtual function called" );
    return 0;
}

