/*++

Copyright (c) 1996, 1997  Microsoft Corporation

Module Name:

    download.cpp
   
Abstract:

    This module implements the CDspCodeDownload class for the Diamond
    Monster/Monster Wave PCI miniport.
      
Author:

    Ken Cooper (kencoope)   25-Mar-1997
   
    based heavily on code developed by Bryan Woodruff (bryanw)   
   
Environment:

    Kernel Mode Only.
   
Revision History:

    28-Mar-1997     kencoope
   
        Modified 'FREEDOM' references to 'MONSTER'.
        
    05-May-1997     kencoope
    
        Added timeouts to bit-spinning code.
      
--*/

#include "private.h"

#define NO_ZW_QIF

#ifdef UNDER_NT
extern "C"
NTSYSAPI
NTSTATUS
NTAPI
ZwQueryInformationFile(
    IN HANDLE FileHandle,
    OUT PIO_STATUS_BLOCK IoStatusBlock,
    OUT PVOID FileInformation,
    IN ULONG Length,
    IN FILE_INFORMATION_CLASS FileInformationClass
    );
#endif

#ifdef NO_ZW_FILE
extern ULONG DspBinArray[];
#endif

#pragma code_seg("PAGE")

CDspCodeDownload::
CDspCodeDownload(
    void
    )
/*++

Routine Description:

    This routine is the constructor for CDspCodeDownload objects.
   
Arguments:

    None.
   
Return Value:

    None.
   
--*/        
{
    PAGED_CODE();
}

STDMETHODIMP
CDspCodeDownload::
ADSP2181_AllocateImage(
    PADSP2181_IMAGE Image
    )
/*++

Routine Description:

    This routine allocates memory to hold the DSP image for downloading.
   
Arguments:

    PUCHAR Image - A pointer to the DSP image descriptor
   
Return Value:

    STATUS_SUCCESS if successful, otherwise an error code.
   
--*/        
{
    PAGED_CODE();

    ASSERT(Image);

    RtlZeroMemory( Image, sizeof( ADSP2181_IMAGE ) );

    if (NULL == (Image->CodeSpace = (PUCHAR) ExAllocatePool( PagedPool, 
                                                    ADSP2181_CODESPACE_SIZE ))) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    if (NULL == (Image->DataSpace = (PUCHAR) ExAllocatePool( PagedPool, 
                                                    ADSP2181_DATASPACE_SIZE ))) {
        ExFreePool( Image->CodeSpace );
        Image->CodeSpace = NULL;
        return STATUS_INSUFFICIENT_RESOURCES;
    }
   
    RtlZeroMemory( Image->CodeSpace, ADSP2181_CODESPACE_SIZE );
    RtlZeroMemory( Image->DataSpace, ADSP2181_DATASPACE_SIZE );
    
    return STATUS_SUCCESS;
}

VOID
CDspCodeDownload::
ADSP2181_FreeImage(
    PADSP2181_IMAGE Image
    )
/*++

Routine Description:

    This routine frees the allocated memory used for downloading the DSP image.
   
Arguments:

    PUCHAR Image - A pointer to the DSP image descriptor.
   
Return Value:

    None.
   
--*/        
{
    PAGED_CODE();

    ASSERT(Image);

    if (Image->CodeSpace) {
        ExFreePool( Image->CodeSpace );
        Image->CodeSpace = NULL;
    }
    if (Image->DataSpace) {
        ExFreePool( Image->DataSpace );
        Image->DataSpace = NULL;
    }
}

VOID
CDspCodeDownload::
ADSP2181_SkipCRLF(
    PUCHAR *ImagePtr
    )
/*++

Routine Description:

    This routine skips CRLF characters while processing a DSP image file.
   
Arguments:

    PUCHAR ImagePtr - A pointer to the current position in the DSP image.
   
Return Value:

    None.
   
--*/        
{
    PAGED_CODE();

    ASSERT(ImagePtr);

    char ch;

    ch = **ImagePtr;
    while ((ch == '\r') || (ch == '\n')) {
        (*ImagePtr)++;
        ch = **ImagePtr;
    }
}

ULONG
CDspCodeDownload::
ADSP2181_ReadHexValue(
    PUCHAR *ImagePtr
    )
/*++

Routine Description:

    This routine reads a hex value from the DSP image.
   
Arguments:

    PUCHAR ImagePtr - A pointer to the current position in the DSP image.
   
Return Value:

    ULONG - The hex value read.
   
--*/        
{
    PAGED_CODE();

    ASSERT(ImagePtr);

    char  ch;
    ULONG RetVal = 0;

    ch = **ImagePtr;
    while (isxdigit( ch ))
    {
        (*ImagePtr)++;
        RetVal <<= 4;
#if 1
        if((ch>='0')&&(ch<='9'))
        {
            ch -= '0';
        } else
        {
            if((ch>='a')&&(ch<='f'))
            {
                ch = (ch - 'a') + 10;
            } else
            {
                if((ch>='A')&&(ch<='F'))
                {
                    ch = (ch - 'A') + 10;
                }
            }
        }
#else
        // Note: toupper() results in importing RtlUpcaseUnicodeToMultiByteN() which is not
        // yet exported in WDM
        ch = toupper( ch );
        if ((ch -= '0') > 9) {
            ch = (ch - ('A' - '0')) + 10;
        }
#endif
        RetVal += ch;
        ch = **ImagePtr;
    }
    ADSP2181_SkipCRLF( ImagePtr );
    
    return RetVal;
}

BOOLEAN
CDspCodeDownload::
ADSP2181_ReadHeader(
    IN PUCHAR *ImagePtr
    )
/*++

Routine Description:

    This routine reads and validates the DSP image file header.
   
Arguments:

    PUCHAR ImagePtr - A pointer to the current position in the DSP image.
   
Return Value:

    BOOLEAN - TRUE is returned if the header is valid, otherwise FALSE is returned.
   
--*/        
{
    PAGED_CODE();

    ASSERT(ImagePtr);

    char    Header[] = "\x1b\x1bi\r\n";
    int     i;

    for (i = 0; i < sizeof( Header ) - 1; i++) {
        if (Header[ i ] != *(*ImagePtr)++) {
            return FALSE;
        }
    }
    return TRUE;
}

USHORT
CDspCodeDownload::
ADSP2181_ReadRegionType(
    IN PUCHAR *ImagePtr
    )
/*++

Routine Description:

    This routine reads the DSP image region type.
   
Arguments:

    PUCHAR ImagePtr - A pointer to the current position in the DSP image.
   
Return Value:

    USHORT - The region ID is returned.
   
--*/        
{
    PAGED_CODE();

    ASSERT(ImagePtr);

    char    ch;      
    USHORT  Type;

    if ((ch = *(*ImagePtr)++) == '\x1b')
        return ADSP2181_REGION_MEM_END;

    // region type has form: @cc\r\n

    if (ch != '@') {
        _DbgPrintF( DEBUGLVL_ERROR, ("expected '@' in ADSP2181_ReadRegionType, found %c", ch) );
        return ADSP2181_REGION_MEM_END;
    }

    Type = *(*((PUSHORT*)ImagePtr))++;
    ADSP2181_SkipCRLF( ImagePtr );

    switch (Type) {

    case 'AP':

        return ADSP2181_REGION_MEM_PA;

    case 'OP':

        return ADSP2181_REGION_MEM_PO;

    case 'AD':

        return ADSP2181_REGION_MEM_DA;

    case 'OD':

        return ADSP2181_REGION_MEM_DO;

    case 'OB':

        return ADSP2181_REGION_MEM_BO;

    default:
        
        _DbgPrintF( DEBUGLVL_BLAB, ("unknown memory specifier in ADSP2181_ReadRegionType") );
        return ADSP2181_REGION_MEM_END;

    }
}

VOID
CDspCodeDownload::
ADSP2181_ReadRegionData(
    IN PUCHAR *ImagePtr,
    IN OUT PADSP2181_IMAGE Image
    )
/*++

Routine Description:

    This routine reads the data from a DSP image file region.
   
Arguments:

    PUCHAR ImagePtr - A pointer to the current position in the DSP image.
   
    PADSP2181_IMAGE Image - A pointer to the DSP image descriptor.
   
Return Value:

    None.
   
--*/        
{
    PAGED_CODE();

    ASSERT(ImagePtr);
    ASSERT(Image);

    ULONG  Value;

    Image->Address = (USHORT) ADSP2181_ReadHexValue( ImagePtr );

    switch( Image->Type ) {

    case ADSP2181_REGION_MEM_PA:
    case ADSP2181_REGION_MEM_PO:

        Image->Address *= 3;
        break;

    case ADSP2181_REGION_MEM_DA:
    case ADSP2181_REGION_MEM_DO:

        Image->Address *= 2;
        break;

    default:
        _DbgPrintF( DEBUGLVL_ERROR, ("ADSP2181_ReadRegionData, unknown data type encountered, type = %d.", Image->Type) );
        break;
    }

    while ('#' != **ImagePtr) {
        Value = ADSP2181_ReadHexValue( ImagePtr );

        switch( Image->Type ) {

        case ADSP2181_REGION_MEM_PA:
        case ADSP2181_REGION_MEM_PO:
            Image->CodeSpace[ Image->Address++ ] = (UCHAR) ((Value >> 8) & 0xFF);
            Image->CodeSpace[ Image->Address++ ] = (UCHAR) ((Value >> 16) & 0xFF);
            Image->CodeSpace[ Image->Address++ ] = (UCHAR) (Value & 0xFF);
            Image->MaxCodeAddress = max( Image->Address, Image->MaxCodeAddress );
            break;

        case ADSP2181_REGION_MEM_DA:
        case ADSP2181_REGION_MEM_DO:
            Image->DataSpace[ Image->Address++ ] = (UCHAR) (Value & 0xFF);
            Image->DataSpace[ Image->Address++ ] = (UCHAR) ((Value >> 8) & 0xFF);
            Image->MaxDataAddress = max( Image->Address, Image->MaxDataAddress );
            break;

        default:
            break;
        }
        
    }
    (*ImagePtr)++;

    // read dummy pointer
    ADSP2181_ReadHexValue( ImagePtr );
}

STDMETHODIMP
CDspCodeDownload::
ADSP2181_LoadImage(
    IN OUT PADSP2181_IMAGE Image
    )
/*++

Routine Description:

    This routine reads the DSP image from the file system.
   
Arguments:

    PADSP2181_IMAGE Image - A pointer to the DSP image descriptor.
   
Return Value:

    STATUS_SUCCESS if successful, otherwise an error code.
   
--*/        
{
    PAGED_CODE();

    ASSERT(Image);

    NTSTATUS                        Status;
    PUCHAR                          FileImage, ImagePtr;
    ULONG                           FileSize;
#ifndef NO_ZW_FILE
    UNICODE_STRING                  FileName;
#endif

#ifdef NO_ZW_FILE
    // without full Zw... file support, use array from dspbin.cpp and hardcoded file size
    FileImage = (PUCHAR)DspBinArray;
    FileSize = 48171;
#else

    // Open DSP image file and read the entire contents
    // Note:  For cross O/S binary compatibility this should either reference the same
    // path or grab the path from a registry key.
//#ifdef UNDER_NT
    RtlInitUnicodeString( &FileName, 
                          L"\\SystemRoot\\system32\\drivers\\monster.bin" );
//#else
//    RtlInitUnicodeString( &FileName, 
//                          L"\\SystemRoot\\system\\monster.bin" );
//#endif

    if (!NT_SUCCESS( Status = LoadFileIntoMemory( &FileName, 
                                                  (PVOID *)&FileImage,
                                                  &FileSize ) ))
    {
        if (FileImage)
        {
            ExFreePool( FileImage );
        }
        return Status;
    }
#endif

    if (!NT_SUCCESS( Status = ADSP2181_AllocateImage( Image ) ))
    {
        _DbgPrintF( DEBUGLVL_ERROR, ("Failed to allocate DSP image.") );
        return Status;
    }

    // The image has been read into the buffer @ FileImage, ImagePtr
    // maintains the current parsing location.

    ImagePtr = FileImage;

    if (ADSP2181_ReadHeader( &ImagePtr ))
    {
        while (ADSP2181_REGION_MEM_END != 
                (Image->Type = ADSP2181_ReadRegionType( &ImagePtr )))
        {
            ADSP2181_ReadRegionData( &ImagePtr, Image );
        } 
        
    } else
    {
        _DbgPrintF( DEBUGLVL_ERROR, ("DSP image file header is invalid.") );
    }

#ifndef NO_ZW_FILE
    if (FileImage)
    {
        ExFreePool( FileImage );
    }
#endif

    if (!NT_SUCCESS( Status ))
    {
        ADSP2181_FreeImage( Image );
    }

    return Status;
}

STDMETHODIMP
CDspCodeDownload::
ADSP2181_DownloadImage(
    IN BOOLEAN DataMemory,
    IN USHORT Address,
    IN USHORT Length,
    IN PUCHAR ImagePtr
    )
/*++

Routine Description:

    This routine downloads a DSP image to the DSP hardware.
   
Arguments:

    BOOLEAN DataMemory -
   
    USHORT Address -
   
    USHORT Length -
   
    PUCHAR ImagePtr -
   
Return Value:

    STATUS_SUCCESS if successful, otherwise an error code.
   
--*/        
{
    PAGED_CODE();

    ASSERT(ImagePtr);

    int             i, Pad = 0;
    MONSTER_GPIP    gpip;
    MONSTER_FIFO    fifo;
    MONSTER_BTU     btu;
    ULONGLONG       startTime;
    BOOLEAN         timeoutFlag;

    // Initialize structures...

    // Data route is Host -> GPIP -> FIFO -> BTU -> DSP

    MONSTER_SETUP_FIFO( &fifo, 0, 0 );
    MONSTER_SETUP_GPIP( &gpip, 0, 
                        MONSTER_GPIP_CONTROL_ENABLE | 
                            MONSTER_GPIP_CONTROL_WRITE_FIFO, 
                        &fifo );
    MONSTER_SETUP_BTU( &btu, 0, 
                       (Address | 
                           (DataMemory ? MONSTER_BTU_ADDRESS_DATAMEMORY : 0)),
                       Length,
                       0,           
                       MONSTER_BTU_CONTROL_DSP_MASTER,
                       MONSTER_BTU_STATUS_ENABLE,
                       &fifo );

    Miniport->SetupFIFO( &fifo );
    Miniport->SetupGPIP( &gpip, TRUE );
    Miniport->SetupFIFO( &fifo );
    Miniport->SetupBTU( &btu, TRUE );

    for (i = 0; i < Length; i++)
    {
        WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_INDEX),
                            gpip.Address );

        // setup 2 second timeout here...
        startTime = GetTimeInterval(0);
        timeoutFlag = TRUE;
        while(GetTimeInterval(startTime) < GTI_SECONDS(2))
        {
            if(0 != (WIN95COMPAT_ReadPortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_LOWORD) ) & MONSTER_GPIP_CONTROL_AVAILABLE))
            {
                timeoutFlag = FALSE;
                break;
            }
        }
        if(timeoutFlag == TRUE)
        {
            return STATUS_IO_TIMEOUT;                                
        }
                    
        if (DataMemory) 
        {
            WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_HIWORD),
                               (USHORT) *ImagePtr++ );
        } else
        {
            if ((++Pad & 0x03) == 0)
            {
                WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_HIWORD),
                                   0 );
            } else
            {
                WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_HIWORD),
                                   (USHORT) *ImagePtr++ );
            }
        }
    }
    
    WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_INDEX),
                       (USHORT) (btu.Address + MONSTER_BTU_OFFSET_TRANSFERLENGTH ) );

    // wait for transfer to complete before disabling BTU
    // setup 2 second timeout here...
    startTime = GetTimeInterval(0);
    timeoutFlag = TRUE;
    while(GetTimeInterval(startTime) < GTI_SECONDS(2))
    {
        if(0 == WIN95COMPAT_ReadPortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_LOWORD) ))
        {
            timeoutFlag = FALSE;
            break;
        }
    }
    if(timeoutFlag == TRUE)
    {
        return STATUS_IO_TIMEOUT;                                
    }

    Miniport->StopGPIP( &gpip );
    Miniport->StopBTU( &btu );

    return STATUS_SUCCESS;   
}

#pragma code_seg()

NTSTATUS
NTAPI
ADSP2181_Handshake
(
    IN  PINTERRUPTSYNC  InterruptSync,
    IN  PVOID           StaticContext   OPTIONAL,
    IN  PVOID           DynamicContext
)
/*++

Routine Description:

    This routine handshakes with the DSP following DSP image download.
   
Arguments:

    PVOID Context - A pointer to the CMiniportWavePciMonster object.
   
Return Value:

    TRUE if the DSP responds, FALSE otherwise.
   
--*/        
{
    // get context
    CMiniportWavePciMonster *Miniport = (CMiniportWavePciMonster *)DynamicContext;

    WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_INDEX),
                        MONSTER_INDEX_MESSAGE_CONTROL_HOST );

    if (1 == WIN95COMPAT_ReadPortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_LOWORD) )) {
        WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_LOWORD), 0 );
        if (0xABCD == WIN95COMPAT_ReadPortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_HIWORD) )) {
            return TRUE;
        }
    }

    return FALSE;
}

#pragma code_seg("PAGE")

STDMETHODIMP
CDspCodeDownload::
DownloadDspCode(
    PVOID   Context,
    BOOLEAN Enabled
    )
/*++

Routine Description:

    This routine is the public method used to invoke the downloading of the DSP
    image from the file system to the DSP hardware.
   
Arguments:

    PVOID Context - A pointer to the CMiniportWavePciMonster object.
   
    BOOLEAN Enabled - If TRUE, DSP code is loaded from the file system.  If FALSE,
        the DSP is reset and restarted without a fresh code shoot.     
   
Return Value:

    STATUS_SUCCESS if successful, otherwise an error code.
   
--*/        
{
    PAGED_CODE();

    ASSERT(Context);

    ADSP2181_IMAGE   Image;
    NTSTATUS         Status;
    USHORT           Temp = 0 ;

    // get context
    Miniport = (CMiniportWavePciMonster *)Context;
    if(Miniport)
    {
       Miniport->AddRef();
    }

    // if enabled, load the DSP image from the file system.
    if (Enabled)
    {
        if (!NT_SUCCESS( Status = ADSP2181_LoadImage( &Image ) ))
        {
            return Status;
        }
    }

    // BUGBUG! With two DSPs, this should be an OR'd value...

    // Perform a software reset of the DSP
    WRITE_MONSTER_INDEXED_REG_LOWORD( Miniport->portMonster,
                                      MONSTER_INDEX_SOFTWARE_RESET, 
                                      (MONSTER_RESET_DSP1 |
                                          MONSTER_RESET_DSP2) );
    Delay( 2 ) ;

    WRITE_MONSTER_INDEXED_REG_LOWORD( Miniport->portMonster,
                                      MONSTER_INDEX_SOFTWARE_RESET, 
                                      MONSTER_RESET_DSP2 );

    // No interrupts from the DSP or ASIC will generated until the
    // code is running on the DSP which occurs on the last write to
    // the DSP code space.  All port I/O here can be unprotected until
    // we attempt a handshake with the DSP.

    if (Enabled)
    {
        int  i;

        // Load the data memory space.
        Status = 
            ADSP2181_DownloadImage( TRUE,                       // data memory
                                    0,                          // address 0
                                    ADSP2181_DATASPACE_SIZE,    // data space
                                    Image.DataSpace );

        // Write the program memory space.                                            
        Status = 
            ADSP2181_DownloadImage( FALSE,                      // program memory
                                    4,                          // address 4
                                    ADSP2181_CODESPACE_SIZE +   // program space
                                        ADSP2181_PAD_SIZE - 4,
                                    &Image.CodeSpace[ 3 ] );

        // Write the initial dispatch routine for the reset interrupt.
        Status =
            ADSP2181_DownloadImage( FALSE,              // program memory
                                    0,                  // address 0
                                    4,                  // 2 words
                                    Image.CodeSpace );

        ADSP2181_FreeImage( &Image );

        // Handshake with DSP...
        for (i = 0; i < 10; i++)
        {
            if ( Miniport->InterruptSync->CallSynchronizedRoutine( ADSP2181_Handshake,
                                                                   Miniport ) )
            {
                break;
            }
            Delay( 10 );
        }

        if (i == 10)
        {
            _DbgPrintF( DEBUGLVL_ERROR, ("DSP OS failed to respond.") );
            return STATUS_DEVICE_NOT_READY;
        }
    } else
    {
        UCHAR JmpZeroCode[] = { 0x00, 0x18, 0x0F, 0x00 };

        // Patch reset vector to jump to zero.
        Status =
            ADSP2181_DownloadImage( FALSE,              // program memory
                                    0,                  // address 0
                                    4,                  // 2 words
                                    JmpZeroCode );
    }

    return Status;
}

STDMETHODIMP
CDspCodeDownload::
LoadFileIntoMemory(
    PUNICODE_STRING FileName,
    PVOID *FileImage,
    PULONG FileSize
    )
/*++

Routine Description:
    Reads the given file into paged memory.  Allocates the memory necessary
    to satisfy the request.

Arguments:
    PUNICODE_STRING FileName - pointer to file name.

    PVOID *FileImage - pointer to pointer image memory allocated to satisfy
        the file load.

Return:
    STATUS_SUCCESS or appropriate error return.

--*/

{
    PAGED_CODE();

    ASSERT(FileName);
    ASSERT(FileImage);
    ASSERT(FileSize);

#ifndef NO_ZW_FILE
#ifndef NO_ZW_QIF
    FILE_STANDARD_INFORMATION       StandardInformation;
#endif
    HANDLE                          FileHandle;
    IO_STATUS_BLOCK                 IoStatus;
    NTSTATUS                        Status;
    OBJECT_ATTRIBUTES               ObjectAttributes;

    *FileImage = NULL;

    //
    // N.B.: For absolutely, positively secure access to this file,
    //       this should be handled in a worker item to protect the
    //       file handle, otherwise, the file handle is exposed in
    //       the user process on whatever process we are currently
    //       running on.
    //

    InitializeObjectAttributes( &ObjectAttributes, 
                                FileName,
                                OBJ_CASE_INSENSITIVE,
                                NULL,
                                NULL );

    Status = ZwCreateFile( &FileHandle, 
                           GENERIC_READ | SYNCHRONIZE, 
                           &ObjectAttributes,
                           &IoStatus,
                           NULL,
                           FILE_ATTRIBUTE_NORMAL, 
                           FILE_SHARE_READ,
                           FILE_OPEN,
                           FILE_SYNCHRONOUS_IO_NONALERT,
                           NULL,
                           0 );

    if (NT_SUCCESS( Status ))
    {
#ifdef NO_ZW_QIF
        if (NULL == (*FileImage = ExAllocatePool( PagedPool,
                                                  48171 )))
        {
            _DbgPrintF( DEBUGLVL_ERROR, ("Unable to allocate memory for file read.") );
            Status = STATUS_INSUFFICIENT_RESOURCES;
        }
#else
        if (NT_SUCCESS(Status = ZwQueryInformationFile( FileHandle,
                                                        &IoStatus, 
                                                        &StandardInformation,
                                                        sizeof( StandardInformation ), 
                                                        FileStandardInformation )))
        {
            if (NULL == (*FileImage = ExAllocatePool( PagedPool,
                                                      StandardInformation.EndOfFile.LowPart )))
            {
                _DbgPrintF( DEBUGLVL_ERROR, ("Unable to allocate memory for file read.") );
                Status = STATUS_INSUFFICIENT_RESOURCES;
            }
        }
#endif

        if (*FileImage) {
            Status = ZwReadFile( FileHandle,
                                 NULL,
                                 NULL,      // Apc
                                 NULL,      // ApcContext
                                 &IoStatus,
                                 *FileImage,
#ifdef NO_ZW_QIF
                                 48171,
#else
                                 StandardInformation.EndOfFile.LowPart,
#endif
                                 NULL,      // ByteOffset
                                 NULL );    // Key

            if (IoStatus.Information != 
#ifdef NO_ZW_QIF
                    48171)
#else
                    StandardInformation.EndOfFile.LowPart)
#endif
            {
                Status = STATUS_END_OF_FILE;
            }

            if (!NT_SUCCESS( Status )) {
                _DbgPrintF( DEBUGLVL_ERROR, ("file read error") );
            }
#ifdef NO_ZW_QIF
            *FileSize = 48171;
#else            
            *FileSize = StandardInformation.EndOfFile.LowPart;
#endif
        }

        ZwClose( FileHandle );
    }

    return Status;
#else
    return STATUS_NOT_IMPLEMENTED;
#endif
}

#pragma code_seg()



