/*++

    Copyright (c) 1996 Microsoft Corporation

Module Name:

    Hardware.c

Abstract:
    This is the main hardware interface module for Diamond Multimedia's 
    Monster/Monster PCI audio device.

    BUGLIST:

    Need to deal with non-response of DSP messages -- need a method to
    remove the message from CurrentMessage and from the queue if necessary.

    Acronymns:
        
        BTU - Burst-Transfer-Unit
            This is the ASIC's bus-master DMA controller interface.  There 
            are 48 BTUs available.  In this design 24 are designated for 
            host use and 24 are designated for DSP use.  The ASIC's BTU 
            scanner processes each BTU in round-robin fashion transferring from 
            the physical memory location on either the host or DSP to the FIFO 
            (or vice-versa).

        VBTU - Virtual BTU
            This is a structure defined in host memory which is loaded into
            the physical BTU when the previously linked BTU is completed.
            VBTUs are linked to the physical BTU via the link field.

        GPIP - General Purpose Input/Output Port

Author:

    Bryan A. Woodruff (bryanw) 16-Sep-1996
    
Environment:

    Kernel Mode Only.
    
Revision History:

    25-Mar-1997   kencoope
    
        Converted from C to C++ and merged into CMiniportWavePciMonster class.
        
--*/

#include "private.h"
#include <stdarg.h>


NTSTATUS
CMiniportWavePciMonster::
SetupFIFO
(
    IN PMONSTER_FIFO Fifo
)
/*++

Routine Description:
    This routine sets up a hardware FIFO on the Monster card.

Arguments:
    IN PMONSTER_FIFO Fifo - A pointer to a fifo structure.

Return:
    Always returns STATUS_SUCCESS.

--*/
{
    ASSERT(Fifo);

    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                      Fifo->Address,
                                      Fifo->Registers.Control );
    READ_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                     (USHORT) MONSTER_FIFO_FLAGS_BASE( Fifo->Index ) );

    return STATUS_SUCCESS;
}


NTSTATUS
CMiniportWavePciMonster::
StopGPIP
(
    IN PMONSTER_GPIP Gpip
)
/*++

Routine Description:
    This routine places a Monster GPIP in the stopped state.

Arguments:
    IN PMONSTER_GPIP Gpip - A pointer to a GPIP structure.

Return:
    Always returns STATUS_SUCCESS.

--*/
{
    ASSERT(Gpip);

    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster, 
                                      Gpip->Address,
                                      MONSTER_GPIP_CONTROL_RESET );

    // setup 2 second timeout here...
    ULONGLONG startTime = GetTimeInterval(0);
    BOOLEAN timeoutFlag = TRUE;
    while(GetTimeInterval(startTime) < GTI_SECONDS(2))
    {
        if(0 == (WIN95COMPAT_ReadPortUShort( (PUSHORT)(portMonster + MONSTER_REG_DATA_LOWORD)) & MONSTER_GPIP_CONTROL_RESET))
        {
            timeoutFlag = FALSE;
            break;
        }
    }
    if(timeoutFlag == TRUE)
    {
        return STATUS_IO_TIMEOUT;                                
    }

    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
SetupGPIP
(
    IN PMONSTER_GPIP Gpip,
    IN BOOLEAN Reset
)
/*++

Routine Description:
    This routine sets up a Monster GPIP.

Arguments:
    IN PMONSTER_GPIP Gpip - A pointer to a GPIP structure.

    IN BOOLEAN Reset - A boolean that indicates whether or not to stop the GPIP
        prior to setup.

Return:
    Always returns STATUS_SUCCESS.

--*/
{
    ASSERT(Gpip);

    if (Reset)
    {
        StopGPIP( Gpip );
    }

    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster, 
                                      Gpip->Address,
                                      Gpip->Registers.Control );

    return STATUS_SUCCESS;
}


NTSTATUS
NTAPI
GetBTUAddressSynchronized
(
    IN  PINTERRUPTSYNC  InterruptSync,
    IN  PVOID           StaticContext   OPTIONAL,
    IN  PVOID           DynamicContext
)
/*++

Routine Description:
    This routine gets the current BTU address from the Monster hardware.
    
    N.B.  This routine must be called as a synchronization routine or in
    the context of a synchronization.

Arguments:
    IN PBTU_ADDR_CONTEXT Context - A pointer to a context structure used for obtaining
        the BTU address.

Return:
    Always returns TRUE.
    

--*/
{
    // get context
    PBTU_ADDR_CONTEXT Context = PBTU_ADDR_CONTEXT(DynamicContext);

   *(Context->Address) = READ_MONSTER_INDEXED_REG_ULONG(Context->Port,
               (USHORT)(Context->Btu->Address + MONSTER_BTU_OFFSET_ADDRESS));

   return STATUS_SUCCESS;
}


NTSTATUS
CMiniportWavePciMonster::
RestartBTU
(
    IN PMONSTER_BTU Btu
)
/*++

Routine Description:
    Restarts a BTU stopped via StopBTU() if the BTU was previously enabled.

Arguments:
    IN PMONSTER_BTU Btu - pointer to the BTU structure.

Return:
    Always returns STATUS_SUCCESS.

--*/
{
    ASSERT(Btu);

    WRITE_MONSTER_INDEXED_REG_HIWORD(portMonster, 
                                     (USHORT)(Btu->Address + MONSTER_BTU_OFFSET_STATUS),
                                     Btu->Registers.Status );
    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
StopBTU
(
    IN PMONSTER_BTU Btu
)
/*++

Routine Description:
    Stops (disables) the specified BTU.

Arguments:
    IN PMONSTER_BTU Btu - pointer to the BTU structure.

Return:
    Always returns STATUS_SUCCESS.

--*/
{
    ASSERT(Btu);

    WRITE_MONSTER_INDEXED_REG_HIWORD( portMonster, 
                                      (USHORT)(Btu->Address + MONSTER_BTU_OFFSET_STATUS),
                                      0 );

    // setup 2 second timeout here...
    ULONGLONG startTime = GetTimeInterval(0);
    BOOLEAN timeoutFlag = TRUE;
    while(GetTimeInterval(startTime) < GTI_SECONDS(2))
    {
        if(0 == (WIN95COMPAT_ReadPortUShort( (PUSHORT) (portMonster + MONSTER_REG_DATA_HIWORD) ) & MONSTER_BTU_STATUS_BUSY))
        {
            timeoutFlag = FALSE;
            break;
        }
    }
    if(timeoutFlag == TRUE)
    {
        return STATUS_IO_TIMEOUT;                                
    }

    return STATUS_SUCCESS;
}

NTSTATUS
NTAPI
SetupBTUSynchronized
(
    IN  PINTERRUPTSYNC  InterruptSync,
    IN  PVOID           StaticContext   OPTIONAL,
    IN  PVOID           DynamicContext
)
/*++

Routine Description:

    Sets up a BTU for processing by the MONSTER BTU scanner.

    N.B.  This routine must be called as a synchronization routine or in
    the context of a synchronization.

Arguments:
    IN PMONSTER_BTU_CONTEXT BtuContext - A pointer to a BTU context structure.

Return:
    Always returns TRUE.
    
--*/
{
    // get context
    PMONSTER_BTU_CONTEXT BtuContext = PMONSTER_BTU_CONTEXT(DynamicContext);
    CMiniportWavePciMonster *Miniport = (CMiniportWavePciMonster *)(BtuContext->Context);
    
    if (BtuContext->Reset)
    {
        Miniport->StopBTU( BtuContext->btu );
    }

    // Note, BTUs are 32-bit registers 
    WRITE_MONSTER_INDEXED_REG_ULONG( Miniport->portMonster,
                                     (USHORT)(BtuContext->btu->Address + MONSTER_BTU_OFFSET_ADDRESS),
                                     BtuContext->btu->Registers.Address );
    
    WRITE_MONSTER_INDEXED_REG_ULONG( Miniport->portMonster, 
                                     (USHORT)(BtuContext->btu->Address + MONSTER_BTU_OFFSET_TRANSFERLENGTH),
                                     MAKELONG(BtuContext->btu->Registers.TransferLength, 
                                              BtuContext->btu->Registers.LinkDescriptor ) );

    WRITE_MONSTER_INDEXED_REG_ULONG( Miniport->portMonster, 
                                     (USHORT)(BtuContext->btu->Address + MONSTER_BTU_OFFSET_CONTROL),
                                     MAKELONG(BtuContext->btu->Registers.Control, 
                                              BtuContext->btu->Registers.Status ) );

    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
SetupBTU
(
    IN PMONSTER_BTU Btu,
    IN BOOLEAN Reset
)
/*++

Routine Description:

    Sets up a BTU for processing by the MONSTER BTU scanner.

Arguments:
    IN PMONSTER_BTU Btu - A pointer to the BTU to be set up.
    
    IN BOOLEAN Reset - A boolean flag that determines whether or not to explicitly
        stop the BTU prior to setup.

Return:
    Returns STATUS_SUCCESS if successful, otherwise STATUS_UNSUCCESSFUL.
    
--*/
{
    ASSERT(Btu);

    MONSTER_BTU_CONTEXT BtuContext;

    // setup context for synchronized routine
    BtuContext.Context = (PVOID)this;
    BtuContext.btu = Btu;
    BtuContext.Reset = Reset;

    return InterruptSync->CallSynchronizedRoutine( SetupBTUSynchronized,
                                                   &BtuContext );
}


#if (DBG)
NTSTATUS
NTAPI
DumpBTUSynchronized
(
    IN  PINTERRUPTSYNC  InterruptSync,
    IN  PVOID           StaticContext   OPTIONAL,
    IN  PVOID           DynamicContext
)
/*++

Routine Description:

    This routine is a debug routine that dumps the contents of a BTU.
    
    N.B.  This routine must be called as a synchronization routine or in
    the context of a synchronization.

Arguments:
    IN PMONSTER_BTU_CONTEXT BtuContext - A pointer to a context structure used
        to dump the BTU data.
    
Return:
    Always retruns TRUE.
    
--*/
{
    ULONG  Value;

    // get context
    PMONSTER_BTU_CONTEXT BtuContext = PMONSTER_BTU_CONTEXT(DynamicContext);
    CMiniportWavePciMonster *Miniport = (CMiniportWavePciMonster *)(BtuContext->Context);

    _DbgPrintF( DEBUGLVL_BLAB, ("Dumping BTU..."));

    Value =
        READ_MONSTER_INDEXED_REG_ULONG( Miniport->portMonster,
                                        (USHORT)(BtuContext->btu->Address + MONSTER_BTU_OFFSET_ADDRESS) );
    _DbgPrintF( DEBUGLVL_BLAB, ("  BTU(%d), Address = %08x", BtuContext->btu->Index, Value) );

    Value =
        READ_MONSTER_INDEXED_REG_ULONG( Miniport->portMonster,
                                        (USHORT)(BtuContext->btu->Address + MONSTER_BTU_OFFSET_TRANSFERLENGTH) );
    _DbgPrintF( DEBUGLVL_BLAB, ("           TransferLength = %04x, LinkDescriptor = %04x", LOWORD( Value ), HIWORD( Value )) );

    Value =
        READ_MONSTER_INDEXED_REG_ULONG( Miniport->portMonster,
                                        (USHORT)(BtuContext->btu->Address + MONSTER_BTU_OFFSET_CONTROL) );
    _DbgPrintF( DEBUGLVL_BLAB, ("           Control = %04x, Status = %04x", LOWORD( Value ), HIWORD( Value )) );

    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
DumpBTU
(
    IN PMONSTER_BTU Btu
)
/*++

Routine Description:

    This routine is a debug routine that dumps the contents of a BTU.
    
Arguments:
    IN PMONSTER_BTU Btu - A pointer to the BTU.
    
Return:
    Returns STATUS_SUCCESS if successful, otherwise STATUS_UNSUCCESSFUL.
    
--*/
{
    ASSERT(Btu);

    MONSTER_BTU_CONTEXT BtuContext;

    // setup context for synchronized routine
    BtuContext.Context = this;
    BtuContext.btu = Btu;

    return InterruptSync->CallSynchronizedRoutine( DumpBTUSynchronized,
                                                   &BtuContext );
}


NTSTATUS 
CMiniportWavePciMonster::
DumpVBTUChain
(
    PMONSTER_PHYSBTU Head
)
/*++

Routine Description:

    This routine is a debug routine that dumps an entire VBTU chain.
    
Arguments:
    IN PMONSTER_PHYSBTU Head - A pointer to the head of the VBTU chain.
    
Return:
    Always returns STATUS_SUCCESS.
    
--*/
{
    ASSERT(Head);

    PMONSTER_PHYSBTU  Link;

    Link = Head;

    _DbgPrintF( DEBUGLVL_BLAB, ("Dumping VBTU Chain..."));

    while (Link)
    {
        _DbgPrintF( DEBUGLVL_BLAB, ("  VBTU(%08x), Address = %08x", Link, Link->Address) );
        _DbgPrintF( DEBUGLVL_BLAB, ("              TransferLength = %04x", Link->TransferLength) );
        _DbgPrintF( DEBUGLVL_BLAB, ("              LinkDescriptor = %04x", Link->LinkDescriptor) );
        _DbgPrintF( DEBUGLVL_BLAB, ("              Control = %04x", Link->Control) );
        _DbgPrintF( DEBUGLVL_BLAB, ("              Status = %04x", Link->Status) );

        if (Link->Control & MONSTER_BTU_CONTROL_LINKDEFINED)
        {
            Link = (PMONSTER_PHYSBTU) ((PUCHAR) HostVBTUsVa + (MAKELONG( Link->LinkDescriptor, 
                                       HIWORD( physHostBTULinkPages.LowPart ) ) - 
                                       physHostBTULinkPages.LowPart));
        } else
        {
            Link = NULL;
        }
        _DbgPrintF( DEBUGLVL_BLAB, ("              Link = %08x", Link) );
    }

    return STATUS_SUCCESS;
}
#endif   // DBG


NTSTATUS
NTAPI
SendMessageSynchronized
(
    IN  PINTERRUPTSYNC  InterruptSync,
    IN  PVOID           StaticContext   OPTIONAL,
    IN  PVOID           DynamicContext
)
/*++

Routine Description:
    Sends the current message (DeviceInstance->CurrentMessage) to the 
    destination via the host message port.

    N.B.  This routine must be called as a synchronization routine.

Arguments:
    IN PVOID Context - pointer to the context (CMiniportWavePciMonster *)

Return:
    Always returns TRUE.

--*/
{
    int               i;
    USHORT            Control;

    // get the context
    CMiniportWavePciMonster *Miniport = (CMiniportWavePciMonster *)DynamicContext;

    // Reset the message pointer and write the message to the port.
    WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_INDEX),
                        MONSTER_INDEX_MESSAGE_CONTROL_HOST );

    Control =  WIN95COMPAT_ReadPortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_LOWORD) );
    WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_LOWORD),
                       (USHORT)(Control & MONSTER_MESSAGE_CONTROL_IRQ_MASK) );

    for (i = 0; i < (int) Miniport->CurrentMessage->Length; i++)
    {
        WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_HIWORD),
                           Miniport->CurrentMessage->Data[ i ] );
    }

    // Notify the receiver of a message pending.  This generates an
    // interrupt on the destination.
    Control = (USHORT) (Miniport->CurrentMessage->Length | 
                        Miniport->CurrentMessage->Destination);
    WIN95COMPAT_WritePortUShort( (PUSHORT)(Miniport->portMonster + MONSTER_REG_DATA_LOWORD),
                       Control );

    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
PostMessage
(
    IN PMONSTER_POSTMESSAGE PostMessage
)            
/*++

Routine Description:
    Posts a message to the DSP, if no messages are queued, then
    this routine will send the message immediately.

Arguments:
    IN PDSP_MESSAGE Message - pointer to DSP message which includes a
        pointer to the message data (including the message ID), size of
        message in USHORTs and a pointer to an Event object to signal when
        completed.

Return:
    Always returns STATUS_SUCCESS.

--*/
{
    ASSERT(PostMessage);

    //
    // The atomic unit of a message is a USHORT.  This routine
    // expects that the caller has properly aligned the data such
    // that odd bytes are not sent to the DSP in the end of the
    // message.
    //

    MessageLock->Begin();

    if (CurrentMessage)
    {
        InsertTailList( &MessageQueue,
                        &PostMessage->ListEntry );
        MessageLock->End();
    } else
    {
        //
        // There is no message pending and therefore no Dpcs are expected
        // to kick this message, so we send it to the hardware immediately.
        //
        
        CurrentMessage = PostMessage;
        MessageLock->End();

        InterruptSync->CallSynchronizedRoutine( SendMessageSynchronized,
                                                this );
    }

    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
PostMessageAndWait
(
    IN OUT PMONSTER_POSTMESSAGE Message OPTIONAL,
    IN USHORT Destination,
    IN USHORT MessageLength,
    ...
)
/*++

Routine Description:


Arguments:
    IN OUT PMONSTER_POSTMESSAGE Message OPTIONAL -
        pointer to the DSP message
        
        when specified this structure is filled by this routine to submit 
        the message to the DSP and when this message is completed, the 
        results returned from the DSP are contained in the Data element 
        of this structure.

    IN USHORT Destination -
        destination for the message

    IN USHORT MessageLength -
        length of message in USHORTS (this is also the number of 
        data parameters on the stack)

    ... -
        variable argument list of USHORTs for message data
        

Return:
    STATUS_SUCCESS or error.

--*/
{
    int                 i; 
    va_list             ArgList;
    MONSTER_POSTMESSAGE LocalMessage;
    KEVENT              Event;
    NTSTATUS            Status;

    if (!Message)
    {
        //
        // If the message parameter is not specified the caller doesn't
        // care about the return value, just use our local structure.
        //
        Message = &LocalMessage;
    }
    va_start( ArgList, MessageLength );

    for (i = 0; i < MessageLength; i++)
    {
        Message->Data[ i ] = va_arg( ArgList, USHORT );
    }

    va_end( ArgList );

    KeInitializeEvent( &Event, NotificationEvent, FALSE );

    Message->Length = MessageLength;
    Message->Event = &Event;
    Message->Destination = Destination;

    if (!NT_SUCCESS( Status = PostMessage( Message ) ))
    {
        return Status;
    }
        
    // BUGBUG - Need a message timeout and a method to remove the
    // posted message.

    //
    // Wait for the response from the DSP.
    //
    KeWaitForSingleObject( &Event,
                           Suspended,
                           KernelMode,
                           FALSE,
                           NULL );

    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
AllocateDSPPipe
( 
    OUT PULONG pipeDSP
)
/*++

Routine Description:
    Posts a message to the DSP to request a DSP pipe allocation.  
    If successful, the pipe number is returned.

Arguments:
    OUT PULONG pipeDSP -
        pointer to ULONG to receive the resultant pipe number

Return:
    STATUS_SUCCESS, *pipeDSP contains a valid pipe number, 
    otherwise an error return and *pipeDSP is undefined.

--*/
{
    ASSERT(pipeDSP);

    MONSTER_POSTMESSAGE Message;
    NTSTATUS            Status;

    //
    // Post a message to the DSP request a pipe allocation.
    //
    Status = PostMessageAndWait( &Message,  
                                 MONSTER_MESSAGE_CONTROL_IRQ_DSP1,
                                 1, 
                                 DM_ALLOC_PIPE );

    if (!NT_SUCCESS( Status ))
    {
        return Status;
    }

    if (!Message.Length)
    {
        _DbgPrintF( DEBUGLVL_BLAB, ("failed to allocate a pipe from DSP.") );
        return STATUS_DEVICE_BUSY;
    }

    *pipeDSP = (ULONG) Message.Data[ 0 ];

    _DbgPrintF( DEBUGLVL_BLAB, ("allocated pipe #%d from DSP", *pipeDSP ) );
    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
FreeDSPPipe
( 
    IN ULONG pipeDSP
)
/*++

Routine Description:
    Posts a message to the DSP to request a DSP pipe allocation.  
    If successful, the pipe number is returned.

Arguments:
    IN ULONG pipeDSP -
        pipe number to free

Return:
    STATUS_SUCCESS, otherwise appropriate error

--*/
{
    MONSTER_POSTMESSAGE Message;
    NTSTATUS            Status;

    //
    // Post a message to the DSP request a pipe allocation.
    //
    Status = PostMessageAndWait( &Message,  
                                 MONSTER_MESSAGE_CONTROL_IRQ_DSP1,
                                 2, 
                                 DM_FREE_PIPE,
                                 (USHORT) pipeDSP );

    if (!NT_SUCCESS( Status ))
    {
        return Status;
    }

    if (Message.Data[ 0 ] != TRUE)
    {
        _DbgPrintF( DEBUGLVL_BLAB, ("DSP returned FALSE when freeing pipe #%d", pipeDSP ) );
        return STATUS_UNSUCCESSFUL;
    }

    _DbgPrintF( DEBUGLVL_BLAB, ("freed pipe #%d to DSP", pipeDSP ) );
    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
GetDSPPipeInfo
( 
    IN ULONG pipeDSP,
    OUT PMONSTER_DSP_PIPE_INFO PipeInfo
)
/*++

Routine Description:
    Posts a message to the DSP to retrieves the given pipe information.

Arguments:
    IN ULONG pipeDSP -
        pipe number acquired by the MonsterAllocateDSPPipe() function.

    OUT PMONSTER_DSP_PIPE_INFO PipeInfo -
        pointer to storage for pipe information.

Return:
    STATUS_SUCCESS, *PipeInfo contains valid pipe information,
    otherwise an error return and *PipeInfo is undefined.

--*/
{
    ASSERT(PipeInfo);

    MONSTER_POSTMESSAGE Message;
    NTSTATUS            Status;

    //
    // Post a message to the DSP request the pipe information.
    //
    Status = PostMessageAndWait( &Message,  
                                 MONSTER_MESSAGE_CONTROL_IRQ_DSP1,
                                 2, 
                                 DM_GET_PIPE_INFO,
                                 LOWORD( pipeDSP ) );

    if (!NT_SUCCESS( Status ))
    {
        return Status;
    }

    if (!Message.Length)
    {
        _DbgPrintF( DEBUGLVL_BLAB, ("failed to retrieve DSP pipe information.") );
        return STATUS_DEVICE_BUSY;
    }

    PipeInfo->PingBuffer = Message.Data[ 0 ];
    PipeInfo->PongBuffer = Message.Data[ 1 ];
    PipeInfo->Size = Message.Data[ 2 ];

    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
AllocateHostPipe
(
    OUT PULONG pipeHost
)
/*++

Routine Description:
    Allocate a pipe on the host.  This function reserves a pipe
    (and therefore its associated physical BTU and FIFO) for use.  
    Additionally, this pipe is used to allocate subsequent virtual 
    BTUs via the AllocateHostVBTU() function.

Arguments:
    OUT PULONG pipeHost -
        pointer to ULONG to receive the resultant pipe number
Return:
    STATUS_SUCCESS, *pipeHost contains a valid pipe number, 
    otherwise an error return and *pipeHost is undefined.

--*/
{
    ASSERT(pipeHost);

    int     i;

    HostAllocLock->Begin();

    // If all pipes are allocated, no need to look further.
    if ((HostPipeAllocMap & MONSTER_MAXNUM_PIPES_ALLOC_MASK) == 
            MONSTER_MAXNUM_PIPES_ALLOC_MASK)
    {
        HostAllocLock->End();
        return STATUS_DEVICE_BUSY;
    }

    //
    // Scan the allocation bitmap for the first open pipe and
    // assign it, if we don't find any then return failure.
    //
    for (i = 0; i < MONSTER_MAXNUM_PIPES; i++)
    {
        if (0 == (HostPipeAllocMap & (1 << i)))
        {
            HostPipeAllocMap |= (1 << i);
            *pipeHost = i;
            break;
        }
    }

    HostAllocLock->End();

    return (i < MONSTER_MAXNUM_PIPES) ? STATUS_SUCCESS : STATUS_DEVICE_BUSY;
}


NTSTATUS 
CMiniportWavePciMonster::
FreeHostPipe
(
    IN ULONG pipeHost
)
/*++

Routine Description:
    Frees a host pipe previously allocated with MonsterAllocateHostPipe().  

Arguments:
    IN ULONG pipeHost -
        pipe to free
Return:
    STATUS_SUCCESS,  otherwise an error return

--*/
{
    HostAllocLock->Begin();

    //
    // If this pipe is not already allocated, return error status.
    //
    if (0 == (HostPipeAllocMap & (1 << pipeHost)))
    {
        HostAllocLock->End();
        return STATUS_INVALID_DEVICE_REQUEST;
    }

    HostPipeAllocMap &= ~(1 << pipeHost);

    HostAllocLock->End();

    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
AllocateHostVBTU
(
    IN ULONG pipeHost,
    OUT PMONSTER_PHYSBTU *VirtBTU
)
/*++

Routine Description:
    Allocate a host virtual BTU.  If a free VBTU is available, this routine
    hashes into the allocation bitmap and scans for the next available slot.
    The bitmap is updated and the virtual BTU is returned.

Arguments:
    IN ULONG pipeHost - Pipe number

    OUT PMONSTER_PHYSBTU VirtBTU -
        pointer to receive the pointer to the virtual BTU

Return:
    STATUS_SUCCESS, *VirtBTU contains a pointer to the allocated virtual BTU,
    otherwise an error return and *VirtBTU is undefined.

--*/
{
    ASSERT(VirtBTU);

    ULONG   Index, Offset;

    HostAllocLock->Begin();

    // If we have no more VBTUs, no need to search.
    if (!HostVBTUFree)
    {
        HostAllocLock->End();
        return STATUS_DEVICE_BUSY;
    }

    // Decrement the free count and search for the free slot.
    HostVBTUFree--;

    //
    // The allocation algorithm is as follows:
    //
    // The hash value is the offset in the bitmap which is the pipe number
    // as an index and offset into ULONG table. (Avg. 30 VBTUs per pipe, 
    // 32 bits per ULONG).
    //
    // If bit offset in ULONG is already set, move to the next bit until end
    // of ULONG continuing to next index until a free slot is located. Wrap
    // around the array if necessary during the scan.
    //

    Index = (pipeHost * HOST_AVG_LINK_DESCRIPTORS_PER_PIPE) / 32;
    Offset = (pipeHost * HOST_AVG_LINK_DESCRIPTORS_PER_PIPE) % 32;

    while (HostVBTUAllocMap[ Index ] & (1 << Offset))
    {
        if (HostVBTUAllocMap[ Index ] == (ULONG) -1)
        {
            // If this portion of the allocation map is already
            // full, skip the scan.
            Index++;
            Offset = 0;
        }
        else if (++Offset == 32)
        {
            // Scan for the empty slot.
            Offset = 0;
            if (++Index == (HOST_LINK_DESCRIPTORS_BUFFER_LENGTH / 
                           (sizeof( MONSTER_PHYSBTU ) * BITS_PER_BYTE)))
            {
                Index = 0;
            }
        }
    }

    // Set the associated bit in the allocation bitmap.
    HostVBTUAllocMap[ Index ] |= (1 << Offset);

    //
    // Return the Va to this virtual BTU -- index and offset calculation
    // required for proper access to the host virtual BTU array.
    //
    *VirtBTU = &HostVBTUsVa[ Index * 32 + Offset ];

    _DbgPrintF(DEBUGLVL_BLAB,("allocate host VBTU: %x (%d:%d)", *VirtBTU, Index, Offset));

    HostAllocLock->End();

    return STATUS_SUCCESS;
}


VOID
CMiniportWavePciMonster::
FreeHostVBTU
(
    IN PMONSTER_PHYSBTU VirtBTU
)
/*++

Routine Description:
    Free a host virtual BTU.  Zaps the VBTU memory and updates the allocation 
    bitmap.

Arguments:
    IN PMONSTER_PHYSBTU VirtBTU - virtual BTU to free

Return:
    STATUS_SUCCESS, otherwise an error return 

--*/
{
    ASSERT(VirtBTU);

    ULONG       Index, Offset;

    HostAllocLock->Begin();
    //
    // Index and offset calculation required for proper access to 
    // the host virtual BTU array.
    //

    Offset = VirtBTU - HostVBTUsVa;
    Index = Offset / 32;
    Offset = Offset % 32;

    // Clear the associated bit in the allocation bitmap.
    _DbgPrintF(DEBUGLVL_BLAB,("free host VBTU: %x (%d:%d)", VirtBTU, Index, Offset));
    ASSERT( 0 != (HostVBTUAllocMap[ Index ] & (1 << Offset)) );
    HostVBTUAllocMap[ Index ] &= ~(1 << Offset);

    // Increment the free count.
    HostVBTUFree++;

    HostAllocLock->End();
}


NTSTATUS 
CMiniportWavePciMonster::
SetupDSPPipe
(
    IN ULONG pipeDSP,
    IN PMONSTER_FIFO Fifo,
    IN BOOLEAN WriteToDevice
)
/*++

Routine Description:
    This routine sets up a DSP pipe.

Arguments:
    IN ULONG pipeDsp - The DSP pipe number.
    
    IN PMONSTER_FIFO Fifo - A pointer to a FIFO structure.
    
    IN BOOLEAN WriteToDevice - A boolean to indicate dataflow direction.

Return:
    STATUS_SUCCESS, otherwise an error return 

--*/
{
    ASSERT(Fifo);

    MONSTER_DSP_PIPE_INFO   PipeInfo;
    NTSTATUS                Status;
    USHORT                  Control;

    // First, we must retrieve the DSP pipe information including buffer
    // pointers and buffer sizes so that we can set up the VBTUs.
    if (!NT_SUCCESS( Status = GetDSPPipeInfo( pipeDSP,
                                              &PipeInfo ) ))
    {
        return Status;
    }

    _DbgPrintF( DEBUGLVL_BLAB, 
                ("pipe %d, ping buffer: %x, pong buffer: %x, size:%x",
                pipeDSP, 
                PipeInfo.PingBuffer, 
                PipeInfo.PongBuffer, 
                PipeInfo.Size) );

    // Set up the DSP VBTUs, this establishes a ping-pong buffer 
    // for this pipe in the DSP memory.

    // Note that DSP word addresses must be converted to byte addresses
    // for the ASIC.
    Control = 
        MONSTER_BTU_CONTROL_DSP_MASTER | 
        MONSTER_BTU_CONTROL_HOST_LINK |
        MONSTER_BTU_CONTROL_DSP1_IRQ;

    if (!WriteToDevice)
    {
        Control |= MONSTER_BTU_CONTROL_WRITE_FIFO;
    }

    MONSTER_SETUP_VBTU( &DSPVBTUsVa[ pipeDSP * 2 ],
                        (PipeInfo.PingBuffer << 1 ) |
                            MONSTER_BTU_ADDRESS_DATAMEMORY,
                        PipeInfo.Size,
                        MONSTER_DSP_VBTU_LINK( this, pipeDSP * 2 + 1 ),
                        Control,
                        MONSTER_BTU_STATUS_ENABLE,
                        Fifo );

    MONSTER_SETUP_VBTU( &DSPVBTUsVa[ pipeDSP * 2 + 1 ],
                        (PipeInfo.PongBuffer << 1) | 
                            MONSTER_BTU_ADDRESS_DATAMEMORY,
                        PipeInfo.Size,
                        MONSTER_DSP_VBTU_LINK( this, pipeDSP * 2 ),
                        Control,
                        MONSTER_BTU_STATUS_ENABLE,
                        Fifo );

    return STATUS_SUCCESS;
}


NTSTATUS
NTAPI
UpdateBTULinkSynchronized
(
    IN  PINTERRUPTSYNC  InterruptSync,
    IN  PVOID           StaticContext   OPTIONAL,
    IN  PVOID           DynamicContext
)
/*++

Routine Description:
    Updates the BTU link descriptor register and control register
    with the information in the shadow BTU.
    
    This routine must be called while synchronized with the ISR.

Arguments:
    PMONSTER_BTU_CONTEXT BtuContext -
        pointer to the BTU context structure

Return:
    Always returns TRUE.
    
--*/
{
   // get context
   PMONSTER_BTU_CONTEXT BtuContext = PMONSTER_BTU_CONTEXT(DynamicContext);
   CMiniportWavePciMonster *Miniport = (CMiniportWavePciMonster *)(BtuContext->Context);    
   
   WRITE_MONSTER_INDEXED_REG_HIWORD( 
        Miniport->portMonster, 
        (USHORT)(BtuContext->btu->Address + MONSTER_BTU_OFFSET_TRANSFERLENGTH),
        BtuContext->btu->Registers.LinkDescriptor );
        
    WRITE_MONSTER_INDEXED_REG_LOWORD( 
        Miniport->portMonster, 
        (USHORT)(BtuContext->btu->Address + MONSTER_BTU_OFFSET_CONTROL),
        BtuContext->btu->Registers.Control );

    return TRUE;        
}


NTSTATUS
NTAPI
RefreshStreamBTUsSynchronized
(
    IN  PINTERRUPTSYNC  InterruptSync,
    IN  PVOID           StaticContext   OPTIONAL,
    IN  PVOID           DynamicContext
)
/*++

Routine Description:
    Refresh the physical BTU if we are updating the link descriptor.
    
    This routine must be called while synchronized with the ISR.

Arguments:
    PSTREAM_REFRESH_CONTEXT RefreshContext -
        pointer to context containing pointer to the vbtu and device instance

Return:
    Always returns TRUE.

--*/
{
    ULONG   Address;
    BTU_ADDR_CONTEXT btuAddrContext;

    // get context
    PREFRESH_CONTEXT RefreshContext = PREFRESH_CONTEXT(DynamicContext);

    // get BTU address
    btuAddrContext.Port = RefreshContext->Miniport->portMonster;
    btuAddrContext.Address = &Address;
    btuAddrContext.Btu = RefreshContext->Btu;
    GetBTUAddressSynchronized( InterruptSync,
                               NULL,
                               &btuAddrContext );

    // If the current address is located in the previous VBTU, then
    // update the link.
    if ((Address >= RefreshContext->vbtu->Address) &&
        (Address <= RefreshContext->vbtu->Address + 
                        RefreshContext->vbtu->TransferLength))
    {                      
       _DbgPrintF(DEBUGLVL_BLAB, ("Refreshing BTU"));

        // Update the shadow...
        RefreshContext->Btu->Registers.LinkDescriptor = 
            RefreshContext->vbtu->LinkDescriptor;
        RefreshContext->Btu->Registers.Control |= 
            RefreshContext->vbtu->Control;

        // Update the hardware...
        MONSTER_BTU_CONTEXT  BtuContext;        
        BtuContext.Context = RefreshContext->Miniport;
        BtuContext.btu = RefreshContext->Btu;
        BtuContext.Reset = FALSE; // ignored

        UpdateBTULinkSynchronized( InterruptSync,
                                   NULL,
                                   &BtuContext );
    }

    return TRUE;
}

VOID 
NTAPI 
Delay
(
    IN ULONG Milliseconds
)
/*++

Routine Description:
    Delay the current thread of execution for the specified number of
    milliseconds in a non-alertable state. This function may only be called
    at less than DISPATCH_LEVEL.

Arguments:
    Milliseconds - Indicates the time to delay in milliseconds.

Return:
    None.

--*/
{
    ULONGLONG StartTime = GetTimeInterval(0);

    while( GetTimeInterval(StartTime) < GTI_MILLISECONDS(Milliseconds) );
}


NTSTATUS 
CMiniportWavePciMonster::
HwReset
(
    BOOLEAN Enabled
)
/*++

Routine Description:
    Performs a hardware reset including a new code shoot to the DSP
    if the Enabled parameter is specified.

Arguments:
    BOOLEAN Enabled -
        If specified, a DSP code shoot is performed, otherwise the
        reset vector is set to "jump 0"

Return:
    Return STATUS_SUCCESS if successful.

--*/
{
    int         i;
    ULONGLONG   startTime;
    BOOLEAN     timeoutFlag;

    // disable interrupts from Monster
    WIN95COMPAT_WritePortUShort( (PUSHORT)(portMonster + MONSTER_REG_IRQ_CONTROL),
                       MONSTER_IRQ_CONTROL_HOST_RESET |
                           MONSTER_IRQ_CONTROL_DSP1_RESET |
                           MONSTER_IRQ_CONTROL_DSP2_RESET );

    WIN95COMPAT_WritePortUShort( (PUSHORT)(portMonster + MONSTER_REG_IRQ_GROUP), 0 );

    // perform a software reset of the DSP
    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                      MONSTER_INDEX_SOFTWARE_RESET, 
                                      (MONSTER_RESET_DSP1 |
                                        MONSTER_RESET_DSP2) );

    // reset BTU master control
    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                      MONSTER_INDEX_BTU_CONTROL,
                                      MONSTER_BTU_CONTROL_RESET );

    // Wait...
    Delay( 10 );

    // Setup the host base link descriptor
    ASSERT( physHostBTULinkPages.HighPart == 0 );
    WRITE_MONSTER_INDEXED_REG_HIWORD( portMonster,
                                      MONSTER_INDEX_BTU_HOST_BASE_LINK,
                                      HIWORD( physHostBTULinkPages.LowPart ) );

    // Debug to verify that we've set the HBM_LINK
    USHORT Value = READ_MONSTER_INDEXED_REG_HIWORD( portMonster,
                                                    0x302 );
                                    
    // reset each individual BTU control word
    for (i = 0; i < MONSTER_MAXNUM_BTU; i++)
    {
        WRITE_MONSTER_INDEXED_REG_HIWORD( portMonster,
                                          (USHORT)(MONSTER_BTU_RAM_BASE( i ) + 
                                            MONSTER_BTU_OFFSET_STATUS),
                                          0x0000 ) ;
        
        // setup 2 second timeout here...
        startTime = GetTimeInterval(0);
        timeoutFlag = TRUE;
        while(GetTimeInterval(startTime) < GTI_SECONDS(2))
        {
            if(0 == (WIN95COMPAT_ReadPortUShort( (PUSHORT) (portMonster + MONSTER_REG_DATA_HIWORD) ) & MONSTER_BTU_STATUS_BUSY))
            {
                timeoutFlag = FALSE;
                break;
            }
        }
        if(timeoutFlag == TRUE)
        {
            return STATUS_IO_TIMEOUT;                                
        }
    }

    // reset the GPIPs
    for (i = 0; i < MONSTER_MAXNUM_GPIP; i++)
    {
        WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                          (USHORT)(MONSTER_GPIP_BASE( i ) + 
                                              MONSTER_GPIP_OFFSET_CONTROL),
                                          MONSTER_GPIP_CONTROL_RESET );

        // setup 2 second timeout here...
        startTime = GetTimeInterval(0);
        timeoutFlag = TRUE;
        while(GetTimeInterval(startTime) < GTI_SECONDS(2))
        {
            if(0 == (WIN95COMPAT_ReadPortUShort( (PUSHORT) (portMonster + MONSTER_REG_DATA_LOWORD) ) & MONSTER_GPIP_CONTROL_RESET))
            {
                timeoutFlag = FALSE;
                break;
            }
        }
        if(timeoutFlag == TRUE)
        {
            return STATUS_IO_TIMEOUT;                                
        }
    }

    // reset the FIFOs
    for (i = 0; i < MONSTER_MAXNUM_FIFO; i++)
    {
        WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                          (USHORT)(MONSTER_FIFO_CONTROL_BASE( i ) +
                                              MONSTER_FIFO_OFFSET_CONTROL),
                                          0x0000 );
    }

    // enable the BTU master control
    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                      MONSTER_INDEX_BTU_CONTROL,
                                      MONSTER_BTU_CONTROL_ENABLE );

    // Reset the message ports
    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                      MONSTER_INDEX_MESSAGE_CONTROL_HOST,
                                      MONSTER_MESSAGE_CONTROL_RESET );

    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                      MONSTER_INDEX_MESSAGE_CONTROL_DSP1,
                                      MONSTER_MESSAGE_CONTROL_RESET );

    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                      MONSTER_INDEX_MESSAGE_CONTROL_DSP2,
                                      MONSTER_MESSAGE_CONTROL_RESET );

    // Reset MIDI controller
    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                      MONSTER_INDEX_MIDI_TRANSMIT,
                                      MONSTER_MIDI_CONTROL_TX_RESET );

    WRITE_MONSTER_INDEXED_REG_LOWORD( portMonster,
                                      MONSTER_INDEX_MIDI_RECEIVE,
                                      MONSTER_MIDI_CONTROL_RX_RESET );

    // Enable interrupts
    WIN95COMPAT_WritePortUShort( (PUSHORT)(portMonster + MONSTER_REG_IRQ_GROUP),
                        MONSTER_IRQMASK_GROUP_BTU |
                        MONSTER_IRQMASK_GROUP_MESSAGE );

    // Load the DSP
    CDspCodeDownload* Loader = new(PagedPool) CDspCodeDownload;
    NTSTATUS ntStatus = Loader->DownloadDspCode( this, Enabled );
    delete Loader;

    return ntStatus;
}


NTSTATUS 
CMiniportWavePciMonster::
TestDataFormat
(
    IN PKSDATAFORMAT DataFormat,
    IN BOOLEAN Capture
)
/*++

Routine Description:
    This routine test support for a particular data format.

Arguments:
    IN PKSDATAFORMAT DataFormat - A pointer to the data format to be tested.
    
    IN BOOLEAN Capture - A boolean to indicate dataflow direction.

Return:
    Return STATUS_SUCCESS if successful, otherwise an error code.

--*/
{
    ASSERT(DataFormat);

    // check size of format structure
    if(DataFormat->FormatSize != sizeof(KSDATAFORMAT_WAVEFORMATEX))
    {
       return STATUS_INVALID_PARAMETER;
    }

    // check MajorFormat, SubFormat and Specifier GUIDs
    if (!IsEqualGUID( DataFormat->MajorFormat, 
                      KSDATAFORMAT_TYPE_AUDIO ) ||
        !IsEqualGUID( DataFormat->SubFormat,
                      KSDATAFORMAT_SUBTYPE_PCM ) ||
        !IsEqualGUID( DataFormat->Specifier,
                      KSDATAFORMAT_SPECIFIER_WAVEFORMATEX ))
    {
       return STATUS_INVALID_PARAMETER;    
    }

    _DbgPrintF( DEBUGLVL_BLAB,
                ("TestDataFormat: Frequency: %d, Channels: %d, bps: %d",
                    ((PKSDATAFORMAT_WAVEFORMATEX)DataFormat)->WaveFormatEx.nSamplesPerSec,
                    ((PKSDATAFORMAT_WAVEFORMATEX)DataFormat)->WaveFormatEx.nChannels,
                    ((PKSDATAFORMAT_WAVEFORMATEX)DataFormat)->WaveFormatEx.wBitsPerSample) );

    return STATUS_SUCCESS;
}


NTSTATUS 
CMiniportWavePciMonster::
HwRun
(
    IN PVOID Context
)
/*++

Routine Description:
    This routine places a stream in the run state.

Arguments:
    IN PVOID Context - A pointer to a context structure.  In this case, the pointer is
        a pointer to a CMiniportWavePciStreamMonster object.

Return:
    Return STATUS_SUCCESS if successful, otherwise an error code.

--*/
{
    ASSERT(Context);

    NTSTATUS            Status;

    _DbgPrintF( DEBUGLVL_BLAB, ("[CMiniportWavePciMonster::HwRun]") );

    // get the stream context
    CMiniportWavePciStreamMonster *Stream = (CMiniportWavePciStreamMonster *)Context;

    // Connect the Host memory -> Host VBTU -> BTU -> FIFO path...
    MONSTER_SETUP_BTU( &Stream->btuHost, 
                       (USHORT) Stream->pipeHost,
                       0,   // No address for link
                       0,   // 0 length to force immediate link to VBTU
                       0,   // set link below
                       ((Stream->WriteToDevice) ?
                           MONSTER_BTU_CONTROL_WRITE_FIFO : 0) |
                           MONSTER_BTU_CONTROL_HOST_LINK,
                       MONSTER_BTU_STATUS_ENABLE,
                       &Stream->fifo );

    // If there are stream mappings in the stream map queue, link the tail in
    // to the physical BTU.
    if (Stream->VbtuTail)
    {
        Stream->btuHost.Registers.LinkDescriptor =
            MONSTER_HOST_VBTU_LINK( 
                this, 
                Stream->VbtuTail );

        Stream->btuHost.Registers.Control |= 
            MONSTER_BTU_CONTROL_LINKDEFINED;
    }

    // Program the device to connect the physical data path
    SetupFIFO( &Stream->fifo );
    SetupBTU( &Stream->btuDSP, TRUE );
    SetupBTU( &Stream->btuHost, TRUE );

#if (DBG)
    DumpBTU( &Stream->btuDSP );
    DumpBTU( &Stream->btuHost );
#endif

    // Data is rolling to the DSP, start the DSP rolling...
    Status = PostMessageAndWait( NULL,
                                 MONSTER_MESSAGE_CONTROL_IRQ_DSP1,
                                 2, 
                                 DM_START_PIPE,
                                 Stream->pipeDSP );

    return Status;
}


NTSTATUS 
CMiniportWavePciMonster::
HwStop
(
    IN PVOID  Context
)
/*++

Routine Description:
    This routine places a stream in the stopped state.

Arguments:
    IN PVOID Context - A pointer to a context structure.  In this case, the pointer is
        a pointer to a CMiniportWavePciStreamMonster object.

Return:
    Return STATUS_SUCCESS if successful, otherwise an error code.

--*/
{
    ASSERT(Context);

    NTSTATUS  Status;

    _DbgPrintF( DEBUGLVL_BLAB, ("[CMiniportWavePciMonster::HwStop]") );

    // get the stream context
    CMiniportWavePciStreamMonster *Stream = (CMiniportWavePciStreamMonster *)Context;

    // Tell the DSP to stop this pipe...
    Status = PostMessageAndWait( NULL,
                                 MONSTER_MESSAGE_CONTROL_IRQ_DSP1,
                                 2, 
                                 DM_STOP_PIPE,
                                 Stream->pipeDSP );

    StopBTU( &Stream->btuDSP );
    StopBTU( &Stream->btuHost );

    return Status;
}


VOID 
CMiniportWavePciMonster::
HostAcknowledgeMsg
(
)
/*++

Routine Description:
    This routine acknowldeges a host message.

Arguments:
    None.

Return:
    None.

--*/
{
    int     i;
    USHORT  Control, Length;

    // Get the response from the destination and stuff it back
    // into the current message data member then schedule a
    // Dpc to continue processing the queue.

    // Note that this implementation does not process the message
    // queue at DIRQL allowing other routines to access the message
    // queue using a spinlock instead of requiring synchronized
    // execution with the ISR.

    ASSERT( CurrentMessage != NULL );

    // Determine message length, reset the FIFO pointer, but keep IRQ high.
    WIN95COMPAT_WritePortUShort( (PUSHORT)(portMonster + MONSTER_REG_INDEX),
                        MONSTER_INDEX_MESSAGE_CONTROL_HOST );

    Control =  WIN95COMPAT_ReadPortUShort( (PUSHORT)(portMonster + MONSTER_REG_DATA_LOWORD) );
    Length = Control & MONSTER_MESSAGE_CONTROL_PTR;
                                 
    WIN95COMPAT_WritePortUShort( (PUSHORT)(portMonster + MONSTER_REG_DATA_LOWORD),
                       (USHORT)(Control & MONSTER_MESSAGE_CONTROL_IRQ_MASK) );

    for (i = 0; i < Length; i++)
    {
        CurrentMessage->Data[ i ] = 
            WIN95COMPAT_ReadPortUShort( (PUSHORT)(portMonster +
                              MONSTER_REG_DATA_HIWORD) );
    }
    
    CurrentMessage->Length = Length;

    // Clear FIFO pointer and clear interrupt.  Acknowledgement of
    // message completed.
    WIN95COMPAT_WritePortUShort( (PUSHORT)(portMonster +
                       MONSTER_REG_DATA_LOWORD), 0 );

    // Request a Dpc so that we can process the next message in the
    // queue, if any.
    Port->Notify(ServiceGroup);
}


VOID 
CMiniportWavePciMonster::
DSPIncomingMsg
(
)
/*++

Routine Description:
    This routine retrieves and acknowledges a message from the DSP.

Arguments:
    None.

Return:
    None.

--*/
{
    USHORT  Control, Length, Message;

    // Determine message length, reset the FIFO pointer, but keep IRQ high.
    WIN95COMPAT_WritePortUShort( (PUSHORT)(portMonster + MONSTER_REG_INDEX),
                        MONSTER_INDEX_MESSAGE_CONTROL_DSP1 );

    Control =  WIN95COMPAT_ReadPortUShort( (PUSHORT)(portMonster + MONSTER_REG_DATA_LOWORD) );
    Length = Control & MONSTER_MESSAGE_CONTROL_PTR;
                                 
    WIN95COMPAT_WritePortUShort( (PUSHORT)(portMonster + MONSTER_REG_DATA_LOWORD),
                       (USHORT)(Control & MONSTER_MESSAGE_CONTROL_IRQ_MASK) );

    // Retrieve message
    Message = WIN95COMPAT_ReadPortUShort( (PUSHORT)(portMonster + MONSTER_REG_DATA_HIWORD) );

    // Clear FIFO pointer and acknowledge receipt of message
    WIN95COMPAT_WritePortUShort( (PUSHORT)(portMonster +
                       MONSTER_REG_DATA_LOWORD),
                       MONSTER_MESSAGE_CONTROL_IRQ_DSP1 );
}


VOID 
CMiniportWavePciMonster::
DSP2IncomingMsg
(
)
/*++

Routine Description:
    This routine retrieves and acknowledges a message from DSP2.  This
    is currently not implemented or supported.

Arguments:
    None.

Return:
    None.

--*/
{
}


NTSTATUS
NTAPI
DeviceIsr
(
    IN  PINTERRUPTSYNC  InterruptSync,
    IN  PVOID           StaticContext,
    IN  PVOID           DynamicContextNotUsed   OPTIONAL
)
/*++

Routine Description:
    This routine is the Monster interrupt service routine.  This routine queries
    the hardware to determine the specific hardware interrupt source, acknowledges
    (clears) the hardware interrupt, and queues a DPC to complete servicing of the
    interrupt.

Arguments:
    IN PKINTERRUPT Interrupt -

    IN PVOID Context - A pointer to a context structure.  In this case, it is a pointer
        to a CMiniportWavePciMonster object.

Return:

--*/

{
    CMiniportWavePciMonster *Miniport = (CMiniportWavePciMonster *)StaticContext;
    USHORT              Source;

    // Determine the IRQ source
    Source = 
        WIN95COMPAT_ReadPortUShort( (PUSHORT) (Miniport->portMonster + 
                                     MONSTER_REG_IRQ_GROUP) ) ;

    // Did we generate the interrupt??
    if(0 == (Source & MONSTER_IRQ_GROUP_MASK))
    {
        return STATUS_UNSUCCESSFUL;
    }
    
    while (Source & (MONSTER_IRQ_GROUP_MESSAGE | MONSTER_IRQ_GROUP_BTU))
    {
        // Dispatch each message interrupt to the appropriate
        // handler. Note that interrupts for the Host and from 
        // the Host message port are acknowledgements.

        if (Source & MONSTER_IRQ_GROUP_BTU)
        {
            // Process BTU interrupts.
            ULONG StreamId;

            // The StreamId for the the interrupting BTU is the same as
            // the BTU index for these HOST interrupts.
            StreamId = WIN95COMPAT_ReadPortUShort( (PUSHORT) (Miniport->portMonster + 
                                                    MONSTER_REG_IRQ_BTU) ) &
                                                    MONSTER_IRQ_BTU_INDEX_MASK;

            // Notify the port to queue stream service
            Miniport->Port->Notify(Miniport->Streams[StreamId]->ServiceGroup);

            // Acknowledge the IRQ
            WIN95COMPAT_WritePortUShort( (PUSHORT) (Miniport->portMonster + MONSTER_REG_IRQ_BTU),
                                         0 ) ;
            // BTU interrupt is cleared.
        }

        if (Source & MONSTER_IRQ_GROUP_MESSAGE)
        {
            // Process message port interrupts.
            USHORT  i, Port;

            for (i = 0, Port = MONSTER_INDEX_MESSAGE_CONTROL_HOST;
                 Port <= MONSTER_INDEX_MESSAGE_CONTROL_DSP2;
                 i++, Port += 4)
            {
                if (MONSTER_MESSAGE_CONTROL_IRQ_HOST == 
                    (READ_MONSTER_INDEXED_REG_LOWORD( Miniport->portMonster, Port ) &
                        MONSTER_MESSAGE_CONTROL_IRQ_MASK))
                {
                    switch (i)
                    {
                        case 0:
                            Miniport->HostAcknowledgeMsg();
                            break;

                        case 1:
                            Miniport->DSPIncomingMsg();
                            break;

                        case 2:
                            Miniport->DSP2IncomingMsg();
                            break;
                    }
                }
            }
            //  Message interrupt is cleared.
        }

        // Loop for more IRQs...
        Source = WIN95COMPAT_ReadPortUShort( (PUSHORT) (Miniport->portMonster + 
                                              MONSTER_REG_IRQ_GROUP) ) ;
    }
    return STATUS_SUCCESS;
} 

