/*++

Copyright (c) 1996, 1997  Microsoft Corporation

Module Name:

    miniport.cpp
   
Abstract:

    This module contains the implementation for the CMiniportWavePciMonster
    class of the miniport driver for the Diamond Monster Sound card.
   
Author:

    Ken Cooper (kencoope)   25-Mar-1997
   
    based heavily on code developed by Bryan Woodruff (bryanw)
   
Environment:

    Kernel Mode Only.
   
Revision History:

    26-Mar-1997     kencoope
   
        CMiniportWavePciStreamMonster implementation split off to stream.cpp.
      
    28-Mar-1997     kencoope
   
        Changed 'FREEDOM' references to 'MONSTER'.
      
--*/

#include "private.h"


static
KSDATARANGE_AUDIO PinDataRangesStream[] =
{
    {
        {
            sizeof(KSDATARANGE_AUDIO),
            0,
            0,
            0,
            STATICGUIDOF(KSDATAFORMAT_TYPE_AUDIO),
            STATICGUIDOF(KSDATAFORMAT_SUBTYPE_PCM),
            STATICGUIDOF(KSDATAFORMAT_SPECIFIER_WAVEFORMATEX)
        },
        2,      // maximum channels
        8,      // minimum bits per sample
        16,     // maximum bits per sample
        5000,   // minimum sample rate
        44100   // maximum sample rate
    }
};

static
PKSDATARANGE PinDataRangePointersStream[] =
{
    PKSDATARANGE(&PinDataRangesStream[0])
};

static
KSDATARANGE PinDataRangesBridge[] =
{
   {
      sizeof(KSDATARANGE),
      0,
      0,
      0,
      STATICGUIDOF(KSDATAFORMAT_TYPE_AUDIO),
      STATICGUIDOF(KSDATAFORMAT_SUBTYPE_ANALOG),
      STATICGUIDOF(KSDATAFORMAT_SPECIFIER_NONE)
   }
};

static
PKSDATARANGE PinDataRangePointersBridge[] =
{
    &PinDataRangesBridge[0]
};

static
MINIPORT_PIN MiniportPins[] =
{
    {
        SIZEOF_ARRAY(PinDataRangePointersStream),
        PinDataRangePointersStream,
        1,
        KSPIN_DATAFLOW_IN,
        0,
        NULL,
        0,
        &KSCATEGORY_AUDIO,
        NULL
    },
    {
        SIZEOF_ARRAY(PinDataRangePointersBridge),
        PinDataRangePointersBridge,
        0,
        KSPIN_DATAFLOW_OUT,
        0,
        NULL,
        0,
        &KSNODETYPE_SPEAKER,
        NULL
    }
};

static
GUID TopologyNodes[] =
{
    STATICGUIDOF(KSNODETYPE_DAC)
};

static
GUID TopologyNodeNames[] =
{   // TODO - fill in with correct GUID
    STATICGUIDOF(KSNODETYPE_DAC)
};

static
MINIPORT_CONNECTION MiniportConnections[] =
{
    { TOPOLOGY_FILTERNODE,  0,  0,                      1 },
    { 0,                    0,  TOPOLOGY_FILTERNODE,    1 }
};

static
MINIPORT_DESCRIPTION MiniportDescription =
{
    SIZEOF_ARRAY(MiniportPins),
    MiniportPins,
    SIZEOF_ARRAY(TopologyNodes),
    TopologyNodes,
    TopologyNodeNames,
    SIZEOF_ARRAY(MiniportConnections),
    MiniportConnections,
    0,
    NULL,
    0
};

#pragma code_seg("PAGE")

NTSTATUS
CreateMiniportWavePciMonster(
    OUT     PUNKNOWN *  Unknown,
    IN      REFCLSID,
    IN      PUNKNOWN    UnknownOuter    OPTIONAL,
    IN      POOL_TYPE   PoolType
    )
/*++

Routine Description:

    This routine creates a CMiniportWavePciMonster object and returns a handle
    to the object's IUnknown interface in Unknown.
   
Arguments:

    PUNKNOWN * Unknown - A variable to hold the IUnknown interface handle upon
    successful return.
   
    REFCLSID -
   
    PUNKNOWN UnknownOuter -
   
    POOL_TYPE PoolType - The pool type to be used when creating the object.
   
Return Value:

    NTSTATUS - STATUS_SUCCESS if successuful, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    ASSERT(Unknown);

    STD_CREATE_BODY(CMiniportWavePciMonster,Unknown,UnknownOuter,PoolType);
}

STDMETHODIMP
CMiniportWavePciMonster::
NonDelegatingQueryInterface(
    REFIID  Interface,
    PVOID * Object
    )
/*++

Routine Description:

    This routine is used to query for additional interfaces on the CMiniportWavePciMonster
    object.
   
Arguments:

    REFIID Interface - The REFIID of the desired interface.
   
    PVOID * Object - A variable to contain the interface handle upon successful return.
   
Return Value:

    STATUS_SUCCESS - success, interface handle returned in Object
   
    STATUS_INVALID_PARAMETER - interface not found
   
--*/        
{
    PAGED_CODE();

    ASSERT(Object);

    if (IsEqualGUIDAligned(Interface,IID_IUnknown))
    {
        *Object = PVOID(PUNKNOWN(this));
    }
    else
    if (IsEqualGUIDAligned(Interface,IID_IMiniport))
    {
        *Object = PVOID(PMINIPORT(this));
    }
    else
    if (IsEqualGUIDAligned(Interface,IID_IMiniportWavePci))
    {
        *Object = PVOID(PMINIPORTWAVEPCI(this));
    }
    else
    {
        *Object = NULL;
    }

    if (*Object)
    {
        PUNKNOWN(*Object)->AddRef();
        return STATUS_SUCCESS;
    }

    return STATUS_INVALID_PARAMETER;
}

CMiniportWavePciMonster::
~CMiniportWavePciMonster(
    void
    )
/*++

Routine Description:

    This routine is the destructor for CMiniportWavePciMonster objects.
   
Arguments:

    None.
   
Return Value:

    None.
   
--*/        
{
    PAGED_CODE();

    // release the port
    if (Port)
    {
        Port->Release();
    }
    
    // release the DMA channel
    if(DmaChannel)
    {
        DmaChannel->FreeBuffer();
        DmaChannel->Release();
        DmaChannel = NULL;
    }


    // release the service group
    if (ServiceGroup)
    {
        ServiceGroup->RemoveMember();
        ServiceGroup->Release();
    }

    // release the interrupt sync
    if(InterruptSync)
    {
        InterruptSync->Release();
    }

    // release the access syncs
    if(MessageLock)
    {
        MessageLock->Release();
    }
    if(HardwareLock)
    {
        HardwareLock->Release();
    }
    if(HostAllocLock)
    {
        HostAllocLock->Release();
    }
}

STDMETHODIMP
CMiniportWavePciMonster::
Init(
    IN      PUNKNOWN            UnknownAdapter,
    IN      PRESOURCELIST       ResourceList,
    IN      PPORTWAVEPCI        Port_,
    OUT     PSERVICEGROUP *     ServiceGroup_
    )
/*++

Routine Description:

    This routine initializes a CMiniportWavePciMonster object.  It saves
    a reference to the associated port driver, addref's the port, processes
    the resource list, resets the hardware, and loads the DSP binary image.
   
Arguments:

    PCM_RESOURCE_LIST ResourceList -
   
    PPORTWAVEPCI Port_ -
      
Return Value:

    STATUS_SUCCESS if successful, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    ASSERT(ResourceList);
    ASSERT(Port_);
    ASSERT(ServiceGroup_);

    NTSTATUS ntStatus;

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciMonster::Init]"));

    // save reference to our port
    Port = Port_;
    if (Port)
    {
        Port->AddRef();
    }

    // process resources
    ntStatus = ProcessResources(ResourceList);

    if(NT_SUCCESS(ntStatus))
    {
        // create the service group
        ServiceCount = 0;
        ntStatus = NewServiceGroup(&ServiceGroup,NULL);
        
        if(NT_SUCCESS(ntStatus))
        {
            ServiceGroup->AddMember();
            *ServiceGroup_ = ServiceGroup;

            // reset the hardware and load the DSP image
            ntStatus = HwReset(TRUE);

            if(!NT_SUCCESS(ntStatus))
            {
                ServiceGroup->RemoveMember();
                ServiceGroup->Release();
                ServiceGroup = NULL;
                *ServiceGroup_ = NULL;
            }
        }
    }

    return ntStatus;
}

NTSTATUS
CMiniportWavePciMonster::
ProcessResources(
    IN      PRESOURCELIST       ResourceList
    )
/*++

Routine Description:

    This routine processes the resources passed in via the ResoureList argument.  In
    addition, the routine performs a variety of initialization for the miniport including
    getting an Adapter Object for bus mastering DMA, allocating device visible host memory,
    and initializing of the host side data structures used by the hardware.
   
Arguments:

    PRESOURCELIST ResourceList - A pointer to an IResourceList interface.
   
Return Value:

    NTSTATUS - STATUS_SUCCESS if successful, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    ASSERT(ResourceList);

    NTSTATUS            ntStatus;
    ULONG               countIO,
                        countIRQ,
                        i;
    PHYSICAL_ADDRESS    AddrConstraint;

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciMonster::ProcessResources]"));

    // get resource counts
    countIO = ResourceList->NumberOfPorts();
    countIRQ = ResourceList->NumberOfInterrupts();

    // check resource counts
    if((countIO < 2) || (countIRQ < 1))
    {
        _DbgPrintF(DEBUGLVL_ERROR,("unknown configuration"));
        return STATUS_DEVICE_CONFIGURATION_ERROR;
    }

    // get translated ports
    portMonster = PUCHAR(ResourceList->FindTranslatedPort(0)->u.Port.Start.LowPart);
    portDSP = PUCHAR(ResourceList->FindTranslatedPort(1)->u.Port.Start.LowPart);

    // initialize synchronization objects
    ntStatus = NewAccessSync( &HardwareLock,
                              NULL,
                              DISPATCH_LEVEL );
    if(NT_SUCCESS(ntStatus))
    {
        ntStatus = NewAccessSync( &HostAllocLock,
                                  NULL,
                                  DISPATCH_LEVEL );
        if(NT_SUCCESS(ntStatus))
        {
            // initialize message queue
            InitializeListHead(&MessageQueue);
            ntStatus = NewAccessSync( &MessageLock,
                                      NULL,
                                      DISPATCH_LEVEL );
            if(NT_SUCCESS(ntStatus))
            {
                // create a DMACHANNEL object
                ntStatus = Port->NewMasterDmaChannel( &DmaChannel,      // OutDmaChannel
                                                      NULL,             // OuterUnknown (OPTIONAL)
                                                      NonPagedPool,     // PoolType
                                                      ResourceList,     // ResourceList (OPTIONAL)
                                                      TRUE,             // ScatterGather
                                                      TRUE,             // Dma32BitAddresses
                                                      FALSE,            // Dma64BitAddresses
                                                      FALSE,            // IgnoreCount
                                                      Width32Bits,      // DmaWidth
                                                      MaximumDmaSpeed,  // DmaSpeed
                                                      0x300000,         // MaximumLength
                                                      0 );              // DmaPort
                if(!NT_SUCCESS(ntStatus))
                {
                    _DbgPrintF(DEBUGLVL_ERROR,("failed NewDmaChannel()"));
                    return STATUS_DEVICE_CONFIGURATION_ERROR;
                }
            
                // allocate a common buffer that is non-cached, 32-bit addressable, physically
                // contiguous and does not cross a 64K boundary
                AddrConstraint.HighPart = 0;
                AddrConstraint.LowPart = 0xFFFF;
                ntStatus = DmaChannel->AllocateBuffer( HOST_LINK_DESCRIPTORS_BUFFER_LENGTH,     // BufferSize
                                                       &AddrConstraint );                       // AddressConstraint
                if(NT_SUCCESS(ntStatus))
                {
                    // get the allocated buffer virtual address
                    HostVBTUsVa = (PMONSTER_PHYSBTU)DmaChannel->SystemAddress();
                    DSPVBTUsVa = &HostVBTUsVa[DSP_LINK_DESCRIPTORS_INDEX];
            
                    // get the allocated buffer physical address
                    physHostBTULinkPages = DmaChannel->PhysicalAddress();
            
                    // 
                    // The VBTU allocation map:
                    //
                    // In the current configuration there are 768 slots described by 
                    // 24 ULONGs.  The last 48 VBTUs are reserved for the DSP pipes.
                    // 
                    // We hash into this allocation map using a crude method of 
                    // calculation:  
                    //
                    // (host pipe index * HOST_AVG_LINK_DESCRIPTORS_PER_PIPE / 32).
                    //
                    // +--------------------------+
                    // | 0|                       |
                    // |  |                       |
                    // | .| Host VBTU allocation  |
                    // | .|        map            |
                    // | .|                       |
                    // |  |                       |
                    // |23|                       |
                    // +--------------------------+
                    // |24| reserved for DSP      |
                    // +--------------------------+
                    //
                
                    //
                    // Allocate the host VBTU link list and allocation bitmap.  The link
                    // list points to the first "active" VBTU for the given pipe, this allows
                    // us to unlink and free the VBTU once the transfer of the segment is
                    // completed. 
                    //
                    // Note that we are reserving the VBTUs for the DSP pipes (2 VBTUs per 
                    // DSP pipe). We automatically assign these VBTUs without further 
                    // allocation procedures.
                    //
            
                    HostVBTUFree = (HOST_LINK_DESCRIPTORS_BUFFER_LENGTH / sizeof(MONSTER_PHYSBTU))
                                    - (2 * MONSTER_MAXNUM_PIPES);
                    i = HostVBTUFree / BITS_PER_BYTE;
                
                    // allocate host VBTU allocation map
                    HostVBTUAllocMap = (PULONG) ExAllocatePool(NonPagedPool, i);
                    if(NULL == HostVBTUAllocMap)
                    {
                        // clean up the DMA channel
                        DmaChannel->FreeBuffer();
                        DmaChannel->Release();
                        DmaChannel = NULL;
                        return STATUS_INSUFFICIENT_RESOURCES;
                    }
            
                    // clear the allocation map and the stream pointers
                    RtlZeroMemory(HostVBTUAllocMap, i);
                    RtlZeroMemory(Streams, MONSTER_MAXNUM_PIPES * sizeof(CMiniportWavePciStreamMonster *));
            
                    // reserve the DSP VBTUs from the allocation map
                    HostVBTUAllocMap[(i / 4) - 1] = (ULONG) -1;
            
                    // create a new interrupt sync
                    ASSERT(ResourceList->FindTranslatedInterrupt(0));
                    ntStatus = NewInterruptSync( &InterruptSync,
                                                 NULL,
                                                 ResourceList,
                                                 0,
                                                 DeviceIsr,
                                                 PVOID(this) );
            
                    if(!NT_SUCCESS(ntStatus))
                    {
                        ExFreePool(HostVBTUAllocMap);
                        DmaChannel->FreeBuffer();
                        DmaChannel->Release();
                        DmaChannel = NULL;
                        _DbgPrintF(DEBUGLVL_ERROR,("failed to connect interrupt"));
                    }
                }
            }        
        }
    }

    return ntStatus;
}

STDMETHODIMP
CMiniportWavePciMonster::
SetStopState(
    IN  ULONG   StopState
    )
/*++

Routine Description:

    This routine sets the stop state for the miniport.
   
Arguments:

    ULONG StopState -
   
Return Value:

    STATUS_SUCCESS if successful, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciMonster::SetStopState]"));
    return STATUS_NOT_IMPLEMENTED;
}

STDMETHODIMP
CMiniportWavePciMonster::
SetPowerState(
    IN      ULONG   PowerState
    )
/*++

Routine Description:

    This routine sets the power state for the miniport
   
Arguments:

    ULONG PowerState -
   
Return Value:

    STATUS_SUCCESS if successful, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciMonster::SetPowerState]"));
    return STATUS_NOT_IMPLEMENTED;
}

STDMETHODIMP
CMiniportWavePciMonster::
NewStream(
    OUT     PMINIPORTWAVEPCISTREAM *    Stream,
    IN      PUNKNOWN                    OuterUnknown,
    IN      POOL_TYPE                   PoolType,
    IN      PPORTWAVEPCISTREAM          PortStream,
    IN      ULONG                       Channel,
    IN      BOOLEAN                     Capture,
    IN      PKSDATAFORMAT               DataFormat,
    OUT     PDMACHANNEL *               DmaChannel_,
    OUT     PSERVICEGROUP *             ServiceGroup_
    )
/*++

Routine Description:

    This routine creates a new CMiniportWavePciStreamMonster object and returns
    a pointer to the new stream object after initializing the stream.
   
Arguments:

    PMINIPORTWAVEPCISTREAM * Stream - A pointer to the newly created stream.
   
    PUNKNOWN OuterUnknown - The outer unknown (container).
   
    POOL_TYPE PoolType - The desired pool type from which to allocate the stream.
   
    PPORTWAVEPCISTREAM PortStream - A pointer to the port's stream interface
   
    ULONG Channel - The stream channel ID.
   
    BOOLEAN Capture - The stream data flow (TRUE=input,FALSE=output).
   
    PKSDATAFORMAT DataFormat - The desired stream data format.
    
    PDMACHANNEL * DmaChannel_ - A pointer the the DMACHANNEL object used by the stream
   
Return Value:

    STATUS_SUCCESS - success, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    ASSERT(Stream);
    ASSERT(PortStream);
    ASSERT(DataFormat);
    ASSERT(DmaChannel_);
    ASSERT(ServiceGroup_);

    ULONG   StreamLocation;

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciMonster::NewStream]"));

    // check validity of Channel ID
    if(Channel >= MONSTER_MAXNUM_PIPES)
    {
       return STATUS_INVALID_PARAMETER;
    }

    // find an empty stream in miniport Streams array
    for(StreamLocation = 0; StreamLocation < MONSTER_MAXNUM_PIPES; StreamLocation++)
    {
       if(Streams[StreamLocation] == NULL)
       {
          break;
       }
    }
    if(StreamLocation == MONSTER_MAXNUM_PIPES)
    {
       // BUGBUG - return correct error code
       return STATUS_INSUFFICIENT_RESOURCES;
    }

    // validate data format
    NTSTATUS ntStatus = TestDataFormat(DataFormat, Capture);

    if(NT_SUCCESS(ntStatus))
    {
       // create a new stream object
       CMiniportWavePciStreamMonster *stream = new(PoolType) CMiniportWavePciStreamMonster(OuterUnknown);

       if(stream)
       {
          // addref the stream
          stream->AddRef();

          // initialize the stream
          ntStatus = stream->Init( this,
                                   PortStream,
                                   Channel,
                                   Capture,
                                   DataFormat,
                                   ServiceGroup_);

          if(NT_SUCCESS(ntStatus))
          {
             // save the stream pointer in the miniport
             Streams[StreamLocation] = stream;
             // set the stream return pointer
             *Stream = PMINIPORTWAVEPCISTREAM(stream);
             // set the DMA channel return pointer
             *DmaChannel_ = DmaChannel;
          } else
          {
             // release the stream
             stream->Release();
          }
       } else
       {
          ntStatus = STATUS_INSUFFICIENT_RESOURCES;
       }
    }

    return ntStatus;
}

STDMETHODIMP
CMiniportWavePciMonster::
GetDescription(
    OUT     PMINIPORT_DESCRIPTION * OutMiniportDescription
    )
/*++

Routine Description:

    This routine provides the port driver with information regarding the
    the filter that represents this miniport.
   
Arguments:

    PMINIPORT_DESCRIPTION * OutMiniportDescription - A pointer to the description pointer to be returned.
    
Return Value:

    STATUS_SUCCESS - success, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    ASSERT(OutMiniportDescription);

    _DbgPrintF(DEBUGLVL_BLAB,("[CMiniportWavePciMonster::GetDescription]"));

    *OutMiniportDescription = &MiniportDescription;

    return STATUS_SUCCESS;
}

#pragma code_seg()

STDMETHODIMP_(void)
CMiniportWavePciMonster::
Service(
    )
/*++

Routine Description:

    This routine is called by the port driver in response to a DSP message interrupt.
   
Arguments:

    
Return Value:

   
--*/        
{
    if (! ServiceGroup->QueryService(&ServiceCount))
    {
        return;
    }

    _DbgPrintF(DEBUGLVL_BLAB,("[CMiniportWavePciMonster::Service]"));

    // acquire message queue spin lock
    MessageLock->Begin();

    // grab the current message
    PMONSTER_POSTMESSAGE Message = CurrentMessage;

    // grab next message in the queue and set it to the current message
    if( !IsListEmpty(&MessageQueue))
    {
        PLIST_ENTRY ListEntry = RemoveHeadList(&MessageQueue);

        CurrentMessage = (PMONSTER_POSTMESSAGE) 
                CONTAINING_RECORD( ListEntry,
                                   MONSTER_POSTMESSAGE,
                                   ListEntry );
    } else
    {
        CurrentMessage = NULL;
    }

    // release the message queue spin lock
    MessageLock->End();

    // signal the owner that the message was acknowledged
    KeSetEvent( Message->Event, 0, FALSE);

    // if there is a new current message, send it to the DSP
    if(CurrentMessage)
    {
        InterruptSync->CallSynchronizedRoutine( SendMessageSynchronized,
                                                this );
    }
}


