/*++

Copyright (c) 1996, 1997  Microsoft Corporation

Module Name:

   private.h

Abstract:

   This module defines the CMiniportWavePciMonster and CMiniportWavePciStreamMonster
   classes, as well as defining data structures used in the Wave PCI miniport for the
   Diamond Monster/Monster card.

Author:

   Ken Cooper (kencoope)   25-Mar-1997

Environment:

   Kernel Mode Only.

Revision History:

   28-Mar-1997    kencoope

      Changed 'FREEDOM' references to 'MONSTER' references.

--*/

#ifndef _MONSTERWAVE_PRIVATE_H_
#define _MONSTERWAVE_PRIVATE_H_

#include "portcls.h"
#include "stdunk.h"
#include "ksdebug.h"
#include "ksmedia.h"

#if !defined( _NTDDK_ )
#define _NTDDK_
#include "monster.h"
#undef _NTDDK_
#else
#include "monster.h"
#endif

#define STR_MODULENAME "MonsterWaveMini: "


#define ELTS_IN_ARRAY(a) (sizeof(a) / sizeof((a)[0]))


typedef struct {
   ULONG        Vector;
   KIRQL        DIrql;
   KAFFINITY    Affinity;
   PKINTERRUPT  Interrupt;

} INTERRUPT_INFO, *PINTERRUPT_INFO;

VOID 
NTAPI 
Delay(
    IN ULONG Milliseconds
    );


//
// Constants
//
#define MAXIMUM_ALLOCATION_ATTEMPTS    0x0008
#define BITS_PER_BYTE                  8

// BUGBUG - pull these out when they get properly defined
#ifndef MAPPING_FLAG_END_OF_PACKET
#define MAPPING_FLAG_END_OF_PACKET  1
#endif

//
// Forward References
//
class CMiniportWavePciStreamMonster;
class CMiniportWavePciMonster;


//
// Data Structures
//
typedef struct
{
   USHORT         Data[8];
   ULONG          Length;
   USHORT         Destination;
   USHORT         Reserved;
   PKEVENT        Event;
   LIST_ENTRY     ListEntry;
} MONSTER_POSTMESSAGE, *PMONSTER_POSTMESSAGE;

typedef struct
{
   PVOID          Context;
   PMONSTER_BTU   btu;
   BOOLEAN        Reset;
} MONSTER_BTU_CONTEXT, *PMONSTER_BTU_CONTEXT;

typedef struct
{
   PUCHAR            Port;
   PULONG            Address;
   PMONSTER_BTU      Btu;
} BTU_ADDR_CONTEXT, *PBTU_ADDR_CONTEXT;

typedef struct
{
   CMiniportWavePciMonster *       Miniport;
   PMONSTER_BTU                    Btu;
   PMONSTER_PHYSBTU                vbtu;
} REFRESH_CONTEXT, *PREFRESH_CONTEXT;

//
// CDspCodeDownload
//
// Class for downloading DSP code image.
//
class CDspCodeDownload
{
   private:
      CMiniportWavePciMonster*  Miniport;

      STDMETHODIMP
      ADSP2181_AllocateImage
      (
          PADSP2181_IMAGE Image
      );

      VOID
      ADSP2181_FreeImage
      (
          PADSP2181_IMAGE Image
      );

      VOID
      ADSP2181_SkipCRLF
      (
          PUCHAR *ImagePtr
      );

      ULONG
      ADSP2181_ReadHexValue
      (
          PUCHAR *ImagePtr
      );

      BOOLEAN
      ADSP2181_ReadHeader
      (
          IN PUCHAR *ImagePtr
      );

      USHORT
      ADSP2181_ReadRegionType
      (
          IN PUCHAR *ImagePtr
      );

      VOID
      ADSP2181_ReadRegionData
      (
          IN PUCHAR *ImagePtr,
          IN OUT PADSP2181_IMAGE Image
      );

      STDMETHODIMP
      ADSP2181_LoadImage
      (
          IN OUT PADSP2181_IMAGE Image
      );

      STDMETHODIMP
      ADSP2181_DownloadImage
      (
          IN BOOLEAN DataMemory,
          IN USHORT Address,
          IN USHORT Length,
          IN PUCHAR ImagePtr
      );

      STDMETHODIMP
      LoadFileIntoMemory
      (
          PUNICODE_STRING FileName,
          PVOID *FileImage,
          PULONG FileSize
      );

   public:
      STDMETHODIMP
      DownloadDspCode
      (
          PVOID   Context,
          BOOLEAN Enabled
      );

      CDspCodeDownload();
};

//
// CMiniportWavePciMonster
//
// Monster wave miniport.
//
class CMiniportWavePciMonster : public IMiniportWavePci, public CUnknown
{
private:
    PPORTWAVEPCI           Port;

    //
    // Monster Driver Variables
    //

    // General
    PUCHAR                 portMonster;
    PUCHAR                 portDSP;
    PACCESSSYNC            HardwareLock;
    PINTERRUPTSYNC         InterruptSync;

    // Host message queue
    PMONSTER_POSTMESSAGE   CurrentMessage;
    PACCESSSYNC            MessageLock;
    LIST_ENTRY             MessageQueue;
    PSERVICEGROUP          ServiceGroup;
    ULONG                  ServiceCount;

    // Bus master support
    PDMACHANNEL            DmaChannel;
    PMONSTER_PHYSBTU       HostVBTUsVa;
    PMONSTER_PHYSBTU       DSPVBTUsVa;
    ULONG                  HostVBTUFree;
    PHYSICAL_ADDRESS       physHostBTULinkPages;
    ULONG                  HostPipeAllocMap;
    PACCESSSYNC            HostAllocLock;
    PULONG                 HostVBTUAllocMap;

    // Audio stream support
    CMiniportWavePciStreamMonster* Streams[MONSTER_MAXNUM_PIPES];

    //
    // CMiniportWavePciMonster methods
    //
    NTSTATUS
    ProcessResources
    (
       IN   PRESOURCELIST     ResourceList
    );

    // hardware support methods
    NTSTATUS
    PostMessage
    (
        IN PMONSTER_POSTMESSAGE PostMessage
    );

    NTSTATUS
    PostMessageAndWait
    (
        IN OUT PMONSTER_POSTMESSAGE Message OPTIONAL,
        IN USHORT Destination,
        IN USHORT MessageLength,
        ...
    );

    NTSTATUS
    AllocateDSPPipe
    (
        OUT PULONG pipeDSP
    );

    NTSTATUS
    FreeDSPPipe
    (
        IN ULONG pipeDSP
    );

    NTSTATUS
    GetDSPPipeInfo
    (
        IN ULONG pipeDSP,
        OUT PMONSTER_DSP_PIPE_INFO PipeInfo
    );

    NTSTATUS
    AllocateHostPipe
    (
        OUT PULONG pipeHost
    );

    NTSTATUS
    FreeHostPipe
    (
        IN ULONG pipeHost
    );

    NTSTATUS
    AllocateHostVBTU
    (
        IN ULONG pipeHost,
        OUT PMONSTER_PHYSBTU *VirtBTU
    );

    VOID
    FreeHostVBTU
    (
        IN PMONSTER_PHYSBTU VirtBTU
    );

    NTSTATUS
    SetupDSPPipe
    (
        IN ULONG pipeDSP,
        IN PMONSTER_FIFO Fifo,
        IN BOOLEAN WriteToDevice
    );

    NTSTATUS
    HwRun
    (
        IN PVOID Context
    );

    NTSTATUS
    HwStop
    (
        IN PVOID  Context
    );

    NTSTATUS
    HwReset
    (
        BOOLEAN Enabled
    );

    NTSTATUS
    SetupFIFO
    (
        IN PMONSTER_FIFO Fifo
    );

    NTSTATUS
    StopGPIP
    (
        IN PMONSTER_GPIP Gpip
    );

    NTSTATUS
    SetupGPIP
    (
        IN PMONSTER_GPIP Gpip,
        IN BOOLEAN Reset
    );

    NTSTATUS
    RestartBTU
    (
        IN PMONSTER_BTU Btu
    );

    NTSTATUS
    StopBTU
    (
        IN PMONSTER_BTU Btu
    );

    NTSTATUS
    SetupBTU
    (
        IN PMONSTER_BTU Btu,
        IN BOOLEAN Reset
    );

#if (DBG)
    NTSTATUS
    DumpBTU
    (
        IN PMONSTER_BTU Btu
    );

    NTSTATUS
    DumpVBTUChain
    (
        PMONSTER_PHYSBTU Head
    );
#endif

    VOID
    HostAcknowledgeMsg
    (
    );

    VOID
    DSPIncomingMsg
    (
    );

    VOID
    DSP2IncomingMsg
    (
    );

    // data format support
    NTSTATUS
    TestDataFormat
    (
        IN PKSDATAFORMAT DataFormat,
        IN BOOLEAN Capture
    );

public:
    DECLARE_STD_UNKNOWN();
    DEFINE_STD_CONSTRUCTOR(CMiniportWavePciMonster);
    ~CMiniportWavePciMonster();

    //
    // IMiniport methods
    //
    STDMETHODIMP SetStopState
    (
        IN      ULONG   StopState
    );
    STDMETHODIMP SetPowerState
    (
        IN      ULONG   PowerState
    );
    STDMETHODIMP GetDescription
    (
        OUT     PMINIPORT_DESCRIPTION * OutMiniportDescription
    );

    //
    // IMiniportWavePci methods
    //
    STDMETHODIMP Init
    (
        IN      PUNKNOWN            UnknownAdapter,
        IN      PRESOURCELIST       ResourceList,
        IN      PPORTWAVEPCI        Port,
        OUT     PSERVICEGROUP *     ServiceGroup
    );
    
    STDMETHODIMP NewStream
    (
        OUT     PMINIPORTWAVEPCISTREAM *    Stream,
        IN      PUNKNOWN                    OuterUnknown,
        IN      POOL_TYPE                   PoolType,
        IN      PPORTWAVEPCISTREAM          PortStream,
        IN      ULONG                       Channel,
        IN      BOOLEAN                     Capture,
        IN      PKSDATAFORMAT               DataFormat,
        OUT     PDMACHANNEL *               DmaChannel,
        OUT     PSERVICEGROUP *             ServiceGroup
    );
    
    STDMETHODIMP_(void) Service
    (
    );

    //
    // Friends
    //
    friend
    NTSTATUS
    NTAPI
    DeviceIsr
    (
        IN  PINTERRUPTSYNC  InterruptSync,
        IN  PVOID           StaticContext,
        IN  PVOID           DynamicContextNotUsed   OPTIONAL
    );

    friend
    NTSTATUS
    NTAPI
    SendMessageSynchronized
    (
        IN  PINTERRUPTSYNC  InterruptSync,
        IN  PVOID           StaticContext   OPTIONAL,
        IN  PVOID           DynamicContext
    );

    friend
    NTSTATUS
    NTAPI
    SetupBTUSynchronized
    (
        IN  PINTERRUPTSYNC  InterruptSync,
        IN  PVOID           StaticContext   OPTIONAL,
        IN  PVOID           DynamicContext
    );

    friend
    NTSTATUS
    UpdateBTULinkSynchronized
    (
        IN  PINTERRUPTSYNC  InterruptSync,
        IN  PVOID           StaticContext   OPTIONAL,
        IN  PVOID           DynamicContext
    );

    friend
    NTSTATUS
    RefreshStreamBTUsSynchronized
    (
        IN  PINTERRUPTSYNC  InterruptSync,
        IN  PVOID           StaticContext   OPTIONAL,
        IN  PVOID           DynamicContext
    );

    friend
    NTSTATUS
    NTAPI
    GetBTUAddressSynchronized
    (
        IN  PINTERRUPTSYNC  InterruptSync,
        IN  PVOID           StaticContext   OPTIONAL,
        IN  PVOID           DynamicContext
    );

#if (DBG)
    friend
    NTSTATUS
    NTAPI
    DumpBTUSynchronized
    (
        IN  PINTERRUPTSYNC  InterruptSync,
        IN  PVOID           StaticContext   OPTIONAL,
        IN  PVOID           DynamicContext
    );
#endif

    friend
    NTSTATUS
    NTAPI
    ADSP2181_Handshake
    (
        IN  PINTERRUPTSYNC  InterruptSync,
        IN  PVOID           StaticContext   OPTIONAL,
        IN  PVOID           DynamicContext
    );

    friend class CMiniportWavePciStreamMonster;
    friend class CDspCodeDownload;
};

//
// CMiniportWavePciStreamMonster
//
// Monster wave miniport stream.
//
class CMiniportWavePciStreamMonster : public IMiniportWavePciStream, public CUnknown
{
private:
    CMiniportWavePciMonster *  Miniport;
    PPORTWAVEPCISTREAM         PortStream;
    ULONG                      Channel;
    BOOLEAN                    Capture;
    PKSDATAFORMAT_WAVEFORMATEX DataFormat;

    // audio stream support
    ULONG                      StreamId;
    ULONG                      pipeDSP;
    ULONG                      pipeHost;
    KSSTATE                    State;
    PACCESSSYNC                ControlMutex;
    BOOLEAN                    WriteToDevice;
    MONSTER_FIFO               fifo;
    MONSTER_BTU                btuHost;
    MONSTER_BTU                btuDSP;

    PSERVICEGROUP              ServiceGroup;
    ULONG                      ServiceCount;

    // memory mapping queue
    PACCESSSYNC                MapLock;
    PMONSTER_PHYSBTU           VbtuHead;
    PMONSTER_PHYSBTU           VbtuTail;

public:
    DECLARE_STD_UNKNOWN();
    DEFINE_STD_CONSTRUCTOR(CMiniportWavePciStreamMonster);
    ~CMiniportWavePciStreamMonster();

    //
    // CMiniportWavePciStreamMonster methods
    //
    NTSTATUS Init
    (
        IN      CMiniportWavePciMonster*  Miniport,
        IN      PPORTWAVEPCISTREAM        PortStream,
        IN      ULONG                     Channel,
        IN      BOOLEAN                   Capture,
        IN      PKSDATAFORMAT             DataFormat,
        OUT     PSERVICEGROUP *           ServiceGroup
    );

    NTSTATUS ProcessNewMappings();

    NTSTATUS ProcessUnmappings();

    //
    // IMiniportWavePciStream methods
    //
    STDMETHODIMP SetFormat
    (
        IN      PKSDATAFORMAT   DataFormat
    );

    STDMETHODIMP SetState
    (
        IN      KSSTATE         State
    );

    STDMETHODIMP GetPosition
    (
        OUT     PULONGLONG      Position
    );

    STDMETHODIMP RevokeMappings
    (
        IN      PVOID           FirstTag,
        IN      PVOID           LastTag,
        OUT     PULONG          MappingsRevoked
    );

    STDMETHODIMP_(void) MappingAvailable
    (
    );
    
    STDMETHODIMP_(void) Service
    (
    );

    //
    // Friends
    //
    friend
    NTSTATUS
    NTAPI
    DeviceIsr
    (
        IN  PINTERRUPTSYNC  InterruptSync,
        IN  PVOID           StaticContext,
        IN  PVOID           DynamicContextNotUsed   OPTIONAL
    );
    
    friend
    NTSTATUS
    RefreshStreamBTUsSynchronized
    (
        IN PREFRESH_CONTEXT RefreshContext
    );

    friend
    NTSTATUS
    NTAPI
    GetBTUAddressSynchronized
    (
        IN OUT  PBTU_ADDR_CONTEXT  Context
    );

    friend class CMiniportWavePciMonster;
};


#endif   // _MONSTERWAVE_PRIVATE_H_