/*++

Copyright (c) 1996, 1997  Microsoft Corporation

Module Name:

    stream.cpp
   
Abstract:

    This module contains the implementation for the CMiniportWavePciStreamMonster
    class of the miniport driver for the Diamond Monster Sound card.
   
Author:

    Ken Cooper (kencoope)   25-Mar-1997
   
    based heavily on code developed by Bryan Woodruff (bryanw)
   
Environment:

    Kernel Mode Only.
   
Revision History:

    26-Mar-1997     kencoope
   
        Split off from miniport.cpp.
      
    28-Mar-1997     kencoope
   
        Changed 'FREEDOM' references to 'MONSTER'.
      
--*/

#include "private.h"

#pragma code_seg("PAGE")

STDMETHODIMP
CMiniportWavePciStreamMonster::
NonDelegatingQueryInterface(
    REFIID  Interface,
    PVOID * Object
    )
/*++

Routine Description:

    This routine is used to query for additional interfaces on the CMiniportWavePciStreamMonster
    object.
   
Arguments:

    REFIID Interface - The REFIID of the desired interface.
   
    PVOID * Object - A variable to contain the interface handle upon successful return.
   
Return Value:

    STATUS_SUCCESS - success, interface handle returned in Object
   
    STATUS_INVALID_PARAMETER - interface not found
   
--*/        
{
    PAGED_CODE();

    ASSERT(Object);

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciStreamMonster::NonDelagatingQueryInterface]"));

    if (IsEqualGUIDAligned(Interface,IID_IUnknown))
    {
        *Object = PVOID(PUNKNOWN(this));
    }
    else
    if (IsEqualGUIDAligned(Interface,IID_IMiniportWavePciStream))
    {
        *Object = PVOID(PMINIPORTWAVEPCISTREAM(this));
    }
    else
    {
        *Object = NULL;
    }

    if (*Object)
    {
        PUNKNOWN(*Object)->AddRef();
        return STATUS_SUCCESS;
    }

    return STATUS_INVALID_PARAMETER;
}

CMiniportWavePciStreamMonster::
~CMiniportWavePciStreamMonster(
    void
    )
/*++

Routine Description:

    This routine is the destructor for CMiniportWavePciStreamMonster objects.
   
Arguments:

    None.
   
Return Value:

    None.
   
--*/        
{
    PAGED_CODE();

    ULONG  StreamLocation;
   
    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciStreamMonster::~CMiniportWavePciStreamMonster]"));

    // remove self from miniport Streams array
    for(StreamLocation = 0; StreamLocation < MONSTER_MAXNUM_PIPES; StreamLocation++)
    {
       if(Miniport->Streams[StreamLocation] == this)
       {
          Miniport->Streams[StreamLocation] = NULL;
          break;
       }
    }

    // free the pipes
    Miniport->FreeDSPPipe(pipeDSP);
    Miniport->FreeHostPipe(pipeHost);

    // release the port stream
    if (PortStream)
    {
       PortStream->Release();
    }

    // release the miniport
    if (Miniport)
    {
        Miniport->Release();
    }

    // release the service group
    if (ServiceGroup)
    {
        ServiceGroup->RemoveMember();
        ServiceGroup->Release();
    }

    // release the sync objects
    if(MapLock)
    {
        MapLock->Release();
    }
    if(ControlMutex)
    {
        ControlMutex->Release();
    }
}

NTSTATUS
CMiniportWavePciStreamMonster::
Init
(
    IN  CMiniportWavePciMonster * Miniport_,
    IN  PPORTWAVEPCISTREAM        PortStream_,
    IN  ULONG                     Channel_,
    IN  BOOLEAN                   Capture_,
    IN  PKSDATAFORMAT             DataFormat_,
    OUT PSERVICEGROUP *           ServiceGroup_
)
/*++

Routine Description:

    This routine initializes a CMiniportWavePciStreamMonster object.  It saves
    a reference to and addrefs the miniport and the stream on the port driver.
    Additionally, the routine setups up and initializes stream specific data
    structures.
   
Arguments:

    CMiniportWavePciMonster * Miniport -
   
    PPORTWAVEPCISTREAM PortStream_ -
   
    ULONG Channel_ -
   
    BOOLEAN Capture_ -
   
    PKSDATAFORMAT DataFormat_ -
   
Return Value:

    STATUS_SUCCESS if successful, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    ASSERT(Miniport_);
    ASSERT(PortStream_);
    ASSERT(DataFormat_);
    ASSERT(ServiceGroup_);

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciStreamMonster::Init]"));
   
    // save miniport pointer and addref it
    Miniport = Miniport_;
    if(Miniport)
    {
        Miniport->AddRef();
    }

    // save portstream interface pointer and addref it
    PortStream = PortStream_;
    if(PortStream)
    {
        PortStream->AddRef();
    }

    // save channel ID and capture flag
    Channel = Channel_;
    Capture = Capture_;

    // save data format
    DataFormat = (PKSDATAFORMAT_WAVEFORMATEX)DataFormat_;

    // initialize member variables
    VbtuHead = NULL;
    VbtuTail = NULL;

    NTSTATUS ntStatus = NewAccessSync( &MapLock,
                                       NULL,
                                       DISPATCH_LEVEL );
    if(NT_SUCCESS(ntStatus))
    {
        ntStatus = NewAccessSync( &ControlMutex,
                                  NULL,
                                  PASSIVE_LEVEL );
        if(NT_SUCCESS(ntStatus))
        {
            // allocate host pipe
            ntStatus = Miniport->AllocateHostPipe( &StreamId );
            if(NT_SUCCESS(ntStatus))
            {
                // setup the audio stream support
                pipeHost = StreamId;               
                State = KSSTATE_STOP;
        
                // create the service group
                *ServiceGroup_ = NULL;
                ntStatus = NewServiceGroup(&ServiceGroup,NULL);
                if(NT_SUCCESS(ntStatus))
                {
                    // setup service group
                    ServiceCount = 0;
                    ServiceGroup->AddMember();
                    *ServiceGroup_ = ServiceGroup;

                    // allocate a DSP pipe
                    ntStatus = Miniport->AllocateDSPPipe( &pipeDSP );
                    if(NT_SUCCESS(ntStatus))
                    {
                        // post a message to the DSP to set the format of the stream
                        ntStatus = Miniport->PostMessageAndWait(NULL,
                                                                MONSTER_MESSAGE_CONTROL_IRQ_DSP1,
                                                                5,
                                                                DM_SET_FORMAT,
                                                                LOWORD(pipeDSP),
                                                                DataFormat->WaveFormatEx.nSamplesPerSec,
                                                                DataFormat->WaveFormatEx.wBitsPerSample,
                                                                DataFormat->WaveFormatEx.nChannels);
                        if(NT_SUCCESS(ntStatus))
                        {        
                            // setup FIFO
                            MONSTER_SETUP_FIFO(&fifo,
                                               (USHORT)pipeHost,
                                               (USHORT)StreamId);
                        
                            // setup stream direction
                            WriteToDevice = Capture ? FALSE : TRUE;
                        
                            // setup DSP pipe
                            Miniport->SetupDSPPipe(pipeDSP,
                                                   &fifo,
                                                   WriteToDevice);
                        
                            // connect the FIFO -> BTU -> DSP VBTU -> DSP memory path
                            MONSTER_SETUP_BTU(&btuDSP,
                                              (USHORT)(pipeDSP + MONSTER_DSP_BTU_BASE_INDEX),
                                              0,
                                              0,
                                              MONSTER_DSP_VBTU_LINK(Miniport, pipeDSP * 2),
                                              MONSTER_BTU_CONTROL_DSP_MASTER | MONSTER_BTU_CONTROL_HOST_LINK,
                                              MONSTER_BTU_STATUS_ENABLE,
                                              &fifo);
        
                        } else
                        {
                            Miniport->FreeDSPPipe( pipeDSP );
                            Miniport->FreeHostPipe( pipeHost );
                            ServiceGroup->RemoveMember();
                            ServiceGroup->Release();
                            ServiceGroup = NULL;
                        }
                    } else
                    {
                        Miniport->FreeHostPipe( pipeHost );
                        ServiceGroup->RemoveMember();
                        ServiceGroup->Release();
                        ServiceGroup = NULL;
                    }
                } else
                {
                    Miniport->FreeHostPipe( pipeHost );
                }
            }
        }
    }

    return ntStatus;
}

#pragma code_seg()

NTSTATUS
CMiniportWavePciStreamMonster::
ProcessUnmappings(
    )
/*++

Routine Description:

    This routine unmaps previously mapped memory that the hardware has completed
    processing on.  This routine is typically called at DPC level from the stream
    deferred procedure call that results from a stream interrupt.
   
Arguments:

    None.
   
Return Value:

    NTSTATUS - STATUS_SUCCESS if successful, otherwise an error code
   
--*/        
{
    ULONG            Address;
    PMONSTER_PHYSBTU vbtu;
    BTU_ADDR_CONTEXT AddrContext;

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciStreamMonster::ProcessUnmappings]"));

    // return if no maps
    if(VbtuHead == NULL)
    {
        return STATUS_SUCCESS;
    }

    // get current BTU address
    AddrContext.Port = Miniport->portMonster;
    AddrContext.Address = &Address;
    AddrContext.Btu = &btuHost;
    Miniport->InterruptSync->CallSynchronizedRoutine( GetBTUAddressSynchronized,
                                                      &AddrContext );
    _DbgPrintF(DEBUGLVL_BLAB, ("Host BTU Address: 0x%x",Address));
    
    // acquire mapping spin lock
    MapLock->Begin();

    //  unmap from the head until the list is empty or the current BTU position is found
    while(1)
    {
        // get the vbtu at the head of the list
        vbtu = VbtuHead;

        // check if hardware is in the current vbtu
        if((Address >= vbtu->Address) &&
           (Address < (vbtu->Address + vbtu->TransferLength)))
        {
            // mapping still needed
            break;
        } else
        {
            // check if vbtu link is defined
            if(vbtu->Control & MONSTER_BTU_CONTROL_LINKDEFINED)
            {
                // move VbtuHead on to next mapping
                VbtuHead = (PMONSTER_PHYSBTU) ((PUCHAR) Miniport->HostVBTUsVa + 
                                               (MAKELONG(vbtu->LinkDescriptor,
                                                         HIWORD(Miniport->physHostBTULinkPages.LowPart)) - 
                                                         Miniport->physHostBTULinkPages.LowPart));

                // free the vbtu and notify the port
                Miniport->FreeHostVBTU(vbtu);
                PortStream->ReleaseMapping((PVOID)vbtu);        
            } else
            {
                // no more mappings
                VbtuHead = VbtuTail = NULL;

                // free the vbtu, notify the port and break
                Miniport->FreeHostVBTU(vbtu);
                PortStream->ReleaseMapping((PVOID)vbtu);
                break;
            }           
        }
    }

    // release map spin lock
    MapLock->End();

    return STATUS_SUCCESS;
}

NTSTATUS
CMiniportWavePciStreamMonster::
ProcessNewMappings(
    )
/*++

Routine Description:

    This routine obtains new memory mappings from the port and links them into the hardware
    data structures.  This routine is called either as a result of the port calling the
    MappingsAvailable method or as part of the stream deferred procedure call that results
    from a stream interrrupt.
   
Arguments:

    None.
   
Return Value:

    STATUS_SUCCESS - success, otherwise an error code
   
--*/        
{
    NTSTATUS            ntStatus;
    PHYSICAL_ADDRESS    mapAddr;
    ULONG               mapLength;
    ULONG               mapFlags;
    PMONSTER_PHYSBTU    virtBTU;
    PMONSTER_PHYSBTU    lastBTU;
    PMONSTER_PHYSBTU    oldTail;
    PVOID               mapVAddr;
   
    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciStreamMonster::ProcessNewMappings]"));

    // null lastBTU and save current VbtuHead
    lastBTU = NULL;
    oldTail = VbtuTail;

    // acquire mapping spin lock
    MapLock->Begin();

    while(1)
    {
        // allocate a VBTU
        ntStatus = Miniport->AllocateHostVBTU( pipeHost,
                                               &virtBTU );
        if(!NT_SUCCESS(ntStatus))
        {
            // no VBTU available, break
            break;
        }

        // get the next mapping
        ntStatus = PortStream->GetMapping( (PVOID)virtBTU,
                                           (PPHYSICAL_ADDRESS)&mapAddr,
                                           &mapVAddr,
                                           &mapLength,
                                           &mapFlags );
        if(!NT_SUCCESS(ntStatus))
        {
            // no mapping available, free VBTU and break
            Miniport->FreeHostVBTU(virtBTU);
            break;
        }

        // setup and link the VBTU
        MONSTER_SETUP_VBTU( virtBTU,                        // Vbtu
                            mapAddr.LowPart,                // 32-bit Transfer Address
                            (USHORT) mapLength,             // Transfer Length (BUGBUG - must be < 64K)
                            0,                              // Link
                            MONSTER_BTU_CONTROL_HOST_LINK,  // Control
                            MONSTER_BTU_STATUS_ENABLE,      // Status
                            &fifo );                        // Fifo Address

        // set direction flag
        if (WriteToDevice)
        {
            virtBTU->Control |= MONSTER_BTU_CONTROL_WRITE_FIFO;
        }

        // check flags for possible notification
        if( 0 != (mapFlags & MAPPING_FLAG_END_OF_PACKET) )
        {
            virtBTU->Control |= MONSTER_BTU_CONTROL_HOST_IRQ;
        }

        // add this VBTU to the tail of the list
        if(VbtuHead)
        {
            PMONSTER_PHYSBTU Modify = VbtuTail;
            VbtuTail = virtBTU;
            Modify->LinkDescriptor = MONSTER_HOST_VBTU_LINK( Miniport, virtBTU );
            Modify->Control |= MONSTER_BTU_CONTROL_LINKDEFINED;
        } else
        {
            VbtuHead = virtBTU;
            VbtuTail = virtBTU;
        }

        // save last BTU mapping
        lastBTU = virtBTU;
    }

    // set notification on last mapping
    if(lastBTU)
    {
        lastBTU->Control |= MONSTER_BTU_CONTROL_HOST_IRQ;
    }
    
    // release the map spin lock
    MapLock->End();

    // if new mappings, see if hardware needs to be kicked
    if(lastBTU)
    {
        // see if this is not the first mapping
        if(oldTail != NULL)
        {
            // see if the BTU need refreshing
            REFRESH_CONTEXT RefreshContext;
            RefreshContext.Miniport = Miniport;
            RefreshContext.Btu = &btuHost;
            RefreshContext.vbtu = oldTail;
            Miniport->InterruptSync->CallSynchronizedRoutine( RefreshStreamBTUsSynchronized,
                                                              &RefreshContext );
        } else
        {
            // first mapping, link the BTU if running
            if(State == KSSTATE_RUN)
            {
                _DbgPrintF(DEBUGLVL_BLAB, ("Kicking H/W for first BTU"));
      
                // setup the link
                btuHost.Registers.LinkDescriptor = MONSTER_HOST_VBTU_LINK(Miniport, VbtuHead);
                btuHost.Registers.Control |= MONSTER_BTU_CONTROL_LINKDEFINED;
      
                // update the BTU link in the hardware
                MONSTER_BTU_CONTEXT btuContext;
                btuContext.Context = Miniport;
                btuContext.btu = &btuHost;             
                btuContext.Reset = FALSE;
                Miniport->InterruptSync->CallSynchronizedRoutine( UpdateBTULinkSynchronized,
                                                                  &btuContext );
            }
        }
    }

    return STATUS_SUCCESS;
}

#pragma code_seg("PAGE")

STDMETHODIMP
CMiniportWavePciStreamMonster::
SetFormat(
    IN  PKSDATAFORMAT   Format
    )
/*++

Routine Description:

    This routine sets the data format for a stream.
   
Arguments:

    PKSDATAFORMAT Format -
   
Return Value:

    STATUS_SUCCESS if successful, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    ASSERT(Format);

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciStreamMonster::SetFormat]"));
    
    return STATUS_NOT_IMPLEMENTED;
}

STDMETHODIMP
CMiniportWavePciStreamMonster::
GetPosition(
    OUT  PULONGLONG   Position
    )
/*++

Routine Description:

    This routine obtains the current stream position.
   
Arguments:

    Position -
   
Return Value:

    STATUS_SUCCESS - success, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    ASSERT(Position);

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciStreamMonster::GetPosition]"));

    *Position = 0;
    return STATUS_NOT_IMPLEMENTED;
}

STDMETHODIMP
CMiniportWavePciStreamMonster::
SetState(
    IN  KSSTATE    State_
    )
/*++

Routine Description:

    This routine sets the state for a stream.
   
Arguments:

    State_ -
   
Return Value:

    STATUS_SUCCESS - success, otherwise an error code
   
--*/        
{
    PAGED_CODE();

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciStreamMonster::SetState]"));
    
    // synchronize for control of the stream
    ControlMutex->Begin();

    // only act of a change of state (i.e. - requested != current)
    if(State != State_)
    {
        // check if we are leaving RUN state
        if(State == KSSTATE_RUN)
        {
            Miniport->HwStop(this);
        }

        // switch on requested state
        switch(State_)
        {
            case KSSTATE_RUN:
                ProcessNewMappings();
                Miniport->HwRun(this);
                break;

            case KSSTATE_STOP:
                break;
        }

        // set state
        State = State_;
    }

    // release mutex
    ControlMutex->End();

    return STATUS_SUCCESS;
}

#pragma code_seg()

STDMETHODIMP
CMiniportWavePciStreamMonster::
RevokeMappings
(
    IN    PVOID     FirstTag,
    IN    PVOID     LastTag,
    OUT   PULONG    MappingsRevoked
)
/*++

Routine Description:

    This routine is used by the port to revoke mappings previously delivered to the
    miniport stream that have not yet been unmapped.  This would typically be called
    in response to an I/O cancellation request.
   
Arguments:

    ULONG Count -
   
    PVOID * Tags -
   
Return Value:

    STATUS_SUCCESS if successful, otherwise an error code
   
--*/        
{
    ASSERT(FirstTag);
    ASSERT(LastTag);
    ASSERT(MappingsRevoked);

    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciStreamMonster::RevokeMappings]"));

    BTU_ADDR_CONTEXT    AddrContext;
    MONSTER_BTU_CONTEXT BtuContext;
    ULONG               Address;
    PMONSTER_PHYSBTU    vbtu, vbtuLast, vbtuNext;
    BOOLEAN             NeedRestart;
    BOOLEAN             FirstFound,LastFound,RevokeIt;
    ULONG               RevokeCount;

    // Synchronize with state changes
    ControlMutex->Begin();

    // acquire mapping queue spin lock
    MapLock->Begin();

    // stop the BTU
    Miniport->StopBTU(&btuHost);
    NeedRestart = TRUE;

    // get BTU address
    AddrContext.Port = Miniport->portMonster;
    AddrContext.Address = &Address;
    AddrContext.Btu = &btuHost;
    Miniport->InterruptSync->CallSynchronizedRoutine( GetBTUAddressSynchronized,
                                                      &AddrContext );

    // initialize flags and revoke count
    FirstFound = LastFound = FALSE;
    
    // get the tail vbtu
    vbtu = VbtuTail;
    vbtuLast = NULL;

    // walk the list from head to tail to find the FirstTag and LastTag entries
    while(1)
    {
        // check for FirstTag
        if(FirstTag == (PVOID)vbtu)
        {
            FirstFound = TRUE;
        }

        // check for LastTag
        if(LastTag == (PVOID)vbtu)
        {
            LastFound = TRUE;
        }

        // break if both tags found
        if((FirstFound == TRUE) && (LastFound == TRUE))
        {
            break;
        }

        // move on to next vbtu in list
        if(vbtu->Control & MONSTER_BTU_CONTROL_LINKDEFINED)
        {
            // calculate next vbtu Va
            vbtu = (PMONSTER_PHYSBTU) ((PUCHAR) Miniport->HostVBTUsVa + 
                                       (MAKELONG(vbtu->LinkDescriptor,
                                       HIWORD(Miniport->physHostBTULinkPages.LowPart)) - 
                                       Miniport->physHostBTULinkPages.LowPart));
        } else
        {
            // reached the end of the list
            break;
        }
    }

    // setup flags for revoke pass though the list
    if((FirstFound == FALSE) && (LastFound == TRUE))
    {
        // this causes us to revoke from the beginning until we find the LastTag
        FirstFound = TRUE;
        LastFound = FALSE;
    } else
    {
        // this causes us to only begin revoking when we find the FirstTag
        FirstFound = LastFound = FALSE;
    }

    // get the tail vbtu
    vbtu = VbtuTail;
    vbtuLast = NULL;
    RevokeCount = 0;

    // walk the list from head to tail, revoking and relinking along the way
    while (1)
    {
        // clear the RevokeIt flag
        RevokeIt = FALSE;

        // have we found the FirstTag
        if(FirstFound == TRUE)
        {
            // have we not found the LastTag
            if(LastFound == FALSE)
            {
                // set the RevokeIt flag
                RevokeIt = TRUE;

                // check if this is the LastTag
                if(LastTag == (PVOID)vbtu)
                {
                    LastFound = TRUE;
                }
            }
        } else
        {
            // check if this is the FirstTag
            if(FirstTag == (PVOID)vbtu)
            {
                // set the FirstFound and RevokeIt flags
                FirstFound = TRUE;
                RevokeIt = TRUE;
            }
        }

        // determine next vbtu in list
        if(vbtu->Control & MONSTER_BTU_CONTROL_LINKDEFINED)
        {
            vbtuNext = (PMONSTER_PHYSBTU) ((PUCHAR) Miniport->HostVBTUsVa + 
                                           (MAKELONG(vbtu->LinkDescriptor,
                                           HIWORD(Miniport->physHostBTULinkPages.LowPart)) - 
                                           Miniport->physHostBTULinkPages.LowPart));
        } else
        {
            vbtuNext = NULL;
        }

        // check if revoke is needed
        if(RevokeIt == TRUE)
        {
            // increment revoke count
            RevokeCount++;

            // check if hardware is in this vbtu
            if((Address >= vbtu->Address) &&
               (Address < (vbtu->Address + vbtu->TransferLength)))
            {
                // check if there is another vbtu in the list
                if(vbtuNext)
                {
                    // prepare BTU with next VBTU
                    MONSTER_SETUP_BTU( &btuHost,
                                       (USHORT)pipeHost,
                                       0,   // No address for link
                                       0,   // 0 length to force immediate link to VBTU
                                       MONSTER_HOST_VBTU_LINK( Miniport, vbtuNext ),
                                       (WriteToDevice ? MONSTER_BTU_CONTROL_WRITE_FIFO : 0) |
                                            MONSTER_BTU_CONTROL_HOST_LINK |
                                            MONSTER_BTU_CONTROL_LINKDEFINED,
                                       MONSTER_BTU_STATUS_ENABLE,
                                       &fifo );
                } else
                {
                    // prepare BTU to wait for new linkage
                    MONSTER_SETUP_BTU( &btuHost, 
                                       (USHORT)pipeHost,
                                       0,   // No address for link
                                       0,   // 0 length to force immediate link to VBTU
                                       0,   // NULL link
                                       (WriteToDevice ? MONSTER_BTU_CONTROL_WRITE_FIFO : 0) |
                                            MONSTER_BTU_CONTROL_HOST_LINK,
                                       MONSTER_BTU_STATUS_ENABLE,
                                       &fifo );
                }

                // restart the BTU hardware
                BtuContext.Context = Miniport;
                BtuContext.btu = &btuHost;
                BtuContext.Reset = FALSE;
                Miniport->InterruptSync->CallSynchronizedRoutine( SetupBTUSynchronized,
                                                                  &BtuContext );
                NeedRestart = FALSE;
            }

            // relink previous vbtu to next vbtu
            if(vbtuLast)
            {
                // check if there is a next vbtu in the list
                if(vbtuNext)
                {
                    // link the previous vbtu to the next
                    vbtuLast->LinkDescriptor = MONSTER_HOST_VBTU_LINK( Miniport, vbtuNext );
                    vbtuLast->Control |= MONSTER_BTU_CONTROL_LINKDEFINED;

                    // free the vbtu
                    Miniport->FreeHostVBTU(vbtu);

                    // move on to next vbtu in list (vbtuLast remains unchanged)
                    vbtu = vbtuNext;
                } else
                {
                    // clear the link defined bit in the previous vbtu
                    vbtuLast->Control &= ~MONSTER_BTU_CONTROL_LINKDEFINED;

                    // free the vbtu and we're done
                    Miniport->FreeHostVBTU(vbtu);
                    break;
                }
            } else
            {
                // check if there is a next vbtu in the list
                if(vbtuNext)
                {
                    // vbtuNext now becomes the head of the list
                    VbtuHead = vbtuNext;

                    // free the vbtu
                    Miniport->FreeHostVBTU(vbtu);

                    // move on to next vbtu in list
                    vbtuLast = NULL;
                    vbtu = vbtuNext;
                } else
                {
                    // list is now empty, NULL the pointers
                    VbtuHead = VbtuTail = NULL;

                    // free the vbtu and we're done
                    Miniport->FreeHostVBTU(vbtu);
                    break;
                }
            }
        } else
        {
            // move on to next vbtu
            if(vbtuNext)
            {
                vbtu = vbtuNext;
            } else
            {
                // reached end of list
                break;
            }
        }
    }

    // release map queue spin lock
    MapLock->End();
    ControlMutex->End();

    // restart the BTU
    if(NeedRestart == TRUE)
    {
        Miniport->RestartBTU( &btuHost );
    }

    // set revoke count return value
    *MappingsRevoked = RevokeCount;
    
    return STATUS_SUCCESS;
}

STDMETHODIMP_(void)
CMiniportWavePciStreamMonster::
MappingAvailable(
    )
/*++

Routine Description:

    This routine is called by the port to indicate to the stream that additional
    memory mappings are available.
   
Arguments:

    None.

Return Value:

    None.
   
--*/        
{
    _DbgPrintF(DEBUGLVL_BLAB, ("[CMiniportWavePciStreamMonster::MappingAvailable]"));

    // process new mappings
    ProcessNewMappings();
    
    return;
}

STDMETHODIMP_(void)
CMiniportWavePciStreamMonster::
Service(
    )
/*++

Routine Description:

    This routine is called by the port at DISPATCH_LEVEL in response to an interrupt.
   
Arguments:

    None.

Return Value:

    None.
   
--*/        
{
    if (! ServiceGroup->QueryService(&ServiceCount))
    {
        return;
    }

    _DbgPrintF(DEBUGLVL_BLAB,("[CMiniportWavePciStreamMonster::Service]"));

    // process unmappings
    ProcessUnmappings();

    // process new mappings
    ProcessNewMappings();
}


