/*****************************************************************************
 * common.cpp - Common code used by all the sb16 miniports.
 *****************************************************************************
 * Copyright (c) 1997 Microsoft Corporation
 *
 * Implmentation of the common code object.  This class deals with interrupts
 * for the device, and is a collection of common code used by all the
 * miniports.
 */

#include "portcls.h"
#include "stdunk.h"
#include "common.h"

#define STR_MODULENAME "sb16Adapter: "





/*****************************************************************************
 * CAdapterCommon
 *****************************************************************************
 * Adapter common object.
 */
class CAdapterCommon
:   public IAdapterCommon,
    public CUnknown
{
private:
    PINTERRUPTSYNC          m_pInterruptSync;
    PUCHAR                  m_pWaveBase;
    PUCHAR                  m_pUartBase;
    PPORTWAVECYCLIC         m_pPortWave;
    PPORTMIDI               m_pPortUart;
    PSERVICEGROUP           m_pServiceGroupWave;
    ULONGLONG               m_startTime;
    BYTE                    MixerRegisters[DSP_MIX_MAXREGS];


    void AcknowledgeIRQ
    (   void
    );
    void ResetController
    (   void
    );
    BOOLEAN AdapterISR
    (   void
    );

public:
    DECLARE_STD_UNKNOWN();
    DEFINE_STD_CONSTRUCTOR(CAdapterCommon);
    ~CAdapterCommon();

    NTSTATUS Init
    (
        IN      PRESOURCELIST   ResourceList
    );

    /*****************************************************************************
     * IAdapterCommon methods
     */
    STDMETHODIMP_(PINTERRUPTSYNC) GetInterruptSync
    (   void
    );
    STDMETHODIMP_(PUNKNOWN *) WavePortDriverDest
    (   void
    );
    STDMETHODIMP_(PUNKNOWN *) MidiPortDriverDest
    (   void
    );
    STDMETHODIMP_(void) SetWaveServiceGroup
    (
        IN      PSERVICEGROUP   ServiceGroup
    );
    STDMETHODIMP_(BYTE) ReadController
    (   void
    );
    STDMETHODIMP_(BOOLEAN) WriteController
    (
        IN      BYTE    Value
    );
    STDMETHODIMP_(void) MixerRegWrite
    (
        IN      BYTE    Index,
        IN      BYTE    Value
    );
    STDMETHODIMP_(BYTE) MixerRegRead
    (
        IN      BYTE    Index
    );
    STDMETHODIMP_(void) MixerReset
    (   void
    );

    friend
    NTSTATUS
    NewAdapterCommon
    (
        OUT     PADAPTERCOMMON *    OutAdapterCommon,
        IN      PRESOURCELIST       ResourceList
    );
    friend
    NTSTATUS
    InterruptServiceRoutine
    (
        IN      PINTERRUPTSYNC  InterruptSync,
        IN      PVOID           StaticContext,
        IN      PVOID           DynamicContextNotUsed   OPTIONAL
    );
};





#pragma code_seg("PAGE")

/*****************************************************************************
 * NewAdapterCommon()
 *****************************************************************************
 * Create a new adapter common object.
 */
NTSTATUS
NewAdapterCommon
(
    OUT     PADAPTERCOMMON *    OutAdapterCommon,
    IN      PRESOURCELIST       ResourceList
)
{
    PAGED_CODE();

    ASSERT(OutAdapterCommon);
    ASSERT(ResourceList);

    NTSTATUS ntStatus = STATUS_SUCCESS;

    //
    // Create the object
    //
    CAdapterCommon *pAdapterCommon = new(NonPagedPool) CAdapterCommon(NULL);

    if (pAdapterCommon)
    {
        pAdapterCommon->AddRef();

        ntStatus = pAdapterCommon->Init(ResourceList);

        if (NT_SUCCESS(ntStatus))
        {
            //
            // Copy the interface to the out parameter.  We don't use
            // QueryInterface() because we are intimate with the object.
            //
            pAdapterCommon->AddRef();
            *OutAdapterCommon = PADAPTERCOMMON(pAdapterCommon);
        }

        //
        // Release in any case.  If all went well, there are other references.
        // If not, the object should be deleted.
        //
        pAdapterCommon->Release();
    }
    else
    {
        ntStatus = STATUS_INSUFFICIENT_RESOURCES;
    }

    return ntStatus;
}

/*****************************************************************************
 * CAdapterCommon::Init()
 *****************************************************************************
 * Initialize an adapter common object.
 */
NTSTATUS
CAdapterCommon::
Init
(
    IN      PRESOURCELIST   ResourceList
)
{
    PAGED_CODE();

    ASSERT(ResourceList);

    //
    // Make sure we have the resources we expect
    //
    ASSERT(ResourceList->NumberOfPorts()      >= 1);
    ASSERT(ResourceList->NumberOfInterrupts() == 1);
    ASSERT(ResourceList->NumberOfDmas()       == 0);
    
    

    NTSTATUS ntStatus =
        NewInterruptSync
        (
            &m_pInterruptSync,
            NULL,
            ResourceList,
            0,
            InterruptServiceRoutine,
            PVOID(this)
        );

    if (NT_SUCCESS(ntStatus))
    {
        //
        // Get the base address for the wave device.
        //
        ASSERT(ResourceList->FindTranslatedPort(0));
        m_pWaveBase =
            PUCHAR(ResourceList->
                FindTranslatedPort(0)->u.Port.Start.LowPart);

        //
        // If there is an MPU...
        //
        if (ResourceList->NumberOfPorts() > 1)
        {
            //
            // Get the base address for the MPU.
            //
            ASSERT(ResourceList->FindTranslatedPort(1));
            m_pUartBase =
                PUCHAR(ResourceList->
                    FindTranslatedPort(1)->u.Port.Start.LowPart);
        }

        //
        // Reset the hardware.
        //
        ResetController();
        AcknowledgeIRQ();
    }

    return ntStatus;
}

/*****************************************************************************
 * CAdapterCommon::~CAdapterCommon()
 *****************************************************************************
 * Destructor.
 */
CAdapterCommon::
~CAdapterCommon
(   void
)
{
    PAGED_CODE();

    _DbgPrintF(DEBUGLVL_TERSE,("[CAdapterCommon::~CAdapterCommon]"));

    if (m_pInterruptSync)
    {
        m_pInterruptSync->Release();
    }
    if (m_pPortWave)
    {
        m_pPortWave->Release();
    }
    if (m_pPortUart)
    {
        m_pPortUart->Release();
    }
    if (m_pServiceGroupWave)
    {
        m_pServiceGroupWave->Release();
    }
}

/*****************************************************************************
 * CAdapterCommon::NonDelegatingQueryInterface()
 *****************************************************************************
 * Obtains an interface.
 */
STDMETHODIMP
CAdapterCommon::
NonDelegatingQueryInterface
(
    REFIID  Interface,
    PVOID * Object
)
{
    PAGED_CODE();

    ASSERT(Object);

    if (IsEqualGUIDAligned(Interface,IID_IUnknown))
    {
        *Object = PVOID(PUNKNOWN(this));
    }
    else
    if (IsEqualGUIDAligned(Interface,IID_IAdapterCommon))
    {
        *Object = PVOID(PADAPTERCOMMON(this));
    }
    else
    {
        *Object = NULL;
    }

    if (*Object)
    {
        PUNKNOWN(*Object)->AddRef();
        return STATUS_SUCCESS;
    }

    return STATUS_INVALID_PARAMETER;
}

/*****************************************************************************
 * CAdapterCommon::WavePortDriverDest()
 *****************************************************************************
 * Get a pointer to the wave port pointer.
 */
STDMETHODIMP_(PUNKNOWN *)
CAdapterCommon::
WavePortDriverDest
(   void
)
{
    PAGED_CODE();

    return (PUNKNOWN *) &m_pPortWave;
}

/*****************************************************************************
 * CAdapterCommon::GetInterruptSync()
 *****************************************************************************
 * Get a pointer to the interrupt synchronization object.
 */
STDMETHODIMP_(PINTERRUPTSYNC)
CAdapterCommon::
GetInterruptSync
(   void
)
{
    PAGED_CODE();

    return m_pInterruptSync;
}

/*****************************************************************************
 * CAdapterCommon::MidiPortDriverDest()
 *****************************************************************************
 * Get a pointer to the UART port pointer.
 */
STDMETHODIMP_(PUNKNOWN *)
CAdapterCommon::
MidiPortDriverDest
(   void
)
{
    PAGED_CODE();

    return (PUNKNOWN *) &m_pPortUart;
}

/*****************************************************************************
 * CAdapterCommon::SetWaveServiceGroup()
 *****************************************************************************
 * Provides a pointer to the save service group.
 */
STDMETHODIMP_(void)
CAdapterCommon::
SetWaveServiceGroup
(
    IN      PSERVICEGROUP   ServiceGroup
)
{
    ASSERT(ServiceGroup);

    if (m_pServiceGroupWave)
    {
        m_pServiceGroupWave->Release();
    }

    m_pServiceGroupWave = ServiceGroup;
    m_pServiceGroupWave->AddRef();
}

#pragma code_seg()

/*****************************************************************************
 * CAdapterCommon::ReadController()
 *****************************************************************************
 * Read a byte from the controller.
 */
STDMETHODIMP_(BYTE)
CAdapterCommon::
ReadController
(   void
)
{
    BYTE returnValue = BYTE(-1);

    ASSERT(m_pWaveBase);

    ULONGLONG m_startTime = GetTimeInterval(0);

    while
    (
        (GetTimeInterval(m_startTime) < GTI_MILLISECONDS(5))
    &&  (BYTE(-1) == returnValue)
    )
    {
        if
        (   WIN95COMPAT_ReadPortUChar
            (
                m_pWaveBase + DSP_REG_DATAAVAIL
            )
        &   0x80
        )
        {
            returnValue =
                WIN95COMPAT_ReadPortUChar
                (
                    m_pWaveBase + DSP_REG_READ
                );
        }
    }

    ASSERT((BYTE(-1) != returnValue) || !"ReadController timeout!");

    return returnValue;
}

/*****************************************************************************
 * CAdapterCommon::WriteController()
 *****************************************************************************
 * Write a byte to the controller.
 */
STDMETHODIMP_(BOOLEAN)
CAdapterCommon::
WriteController
(
    IN      BYTE    Value
)
{
    ASSERT(m_pWaveBase);

    BOOLEAN     returnValue = FALSE;
    ULONGLONG   m_startTime   = GetTimeInterval(0);

    while
    (
        (GetTimeInterval(m_startTime) < GTI_MILLISECONDS(20))
    &&  ! returnValue
    )
    {

        BYTE status =
            WIN95COMPAT_ReadPortUChar
            (
                m_pWaveBase + DSP_REG_WRITE
            );

        if ((status & 0x80) == 0)
        {
            WIN95COMPAT_WritePortUChar
            (
                m_pWaveBase + DSP_REG_WRITE,
                Value
            );

            returnValue = TRUE;
        }
    }

    ASSERT(returnValue || !"WriteController timeout");

    return returnValue;
}

/*****************************************************************************
 * CAdapterCommon::MixerRegWrite()
 *****************************************************************************
 * Writes a mixer register.
 */
STDMETHODIMP_(void)
CAdapterCommon::
MixerRegWrite
(
    IN      BYTE    Index,
    IN      BYTE    Value
)
{
    ASSERT( m_pWaveBase );
    BYTE actualIndex;

    actualIndex = (BYTE) ((Index < 0x80) ? (Index + DSP_MIX_BASEIDX) : Index);

    WIN95COMPAT_WritePortUChar
    (
        m_pWaveBase + DSP_REG_MIXREG,
        actualIndex
    );

    WIN95COMPAT_WritePortUChar
    (
        m_pWaveBase + DSP_REG_MIXDATA,
        Value
    );

    if (Index < DSP_MIX_MAXREGS)
    {
        MixerRegisters[Index] = Value;
    }
}

/*****************************************************************************
 * CAdapterCommon::MixerRegRead()
 *****************************************************************************
 * Reads a mixer register.
 */
STDMETHODIMP_(BYTE)
CAdapterCommon::
MixerRegRead
(
    IN      BYTE    Index
)
{
    if (Index < DSP_MIX_MAXREGS)
    {
        return MixerRegisters[Index];
    }
    else
    {
        return 0;
    }
}

/*****************************************************************************
 * CAdapterCommon::MixerReset()
 *****************************************************************************
 * Resets the mixer
 */
STDMETHODIMP_(void)
CAdapterCommon::
MixerReset
(   void
)
{
    ASSERT(m_pWaveBase);

    WIN95COMPAT_WritePortUChar
    (
        m_pWaveBase + DSP_REG_MIXREG,
        DSP_MIX_DATARESETIDX
    );

    WIN95COMPAT_WritePortUChar
    (
        m_pWaveBase + DSP_REG_MIXDATA,
        0
    );

    MixerRegisters[DSP_MIX_MASTERVOLIDX_L]    = 0xC0;
    MixerRegisters[DSP_MIX_MASTERVOLIDX_R]    = 0xC0;
    MixerRegisters[DSP_MIX_VOICEVOLIDX_L]     = 0xC0;
    MixerRegisters[DSP_MIX_VOICEVOLIDX_R]     = 0xC0;
    MixerRegisters[DSP_MIX_FMVOLIDX_L]        = 0xC0;
    MixerRegisters[DSP_MIX_FMVOLIDX_R]        = 0xC0;
    MixerRegisters[DSP_MIX_CDVOLIDX_L]        = 0x00;
    MixerRegisters[DSP_MIX_CDVOLIDX_R]        = 0x00;
    MixerRegisters[DSP_MIX_LINEVOLIDX_L]      = 0x00;
    MixerRegisters[DSP_MIX_LINEVOLIDX_R]      = 0x00;
    MixerRegisters[DSP_MIX_MICVOLIDX]         = 0x00;
    MixerRegisters[DSP_MIX_SPKRVOLIDX]        = 0x00;
    MixerRegisters[DSP_MIX_OUTMIXIDX]         = 0x1F;
    MixerRegisters[DSP_MIX_ADCMIXIDX_L]       = 0x15;
    MixerRegisters[DSP_MIX_ADCMIXIDX_R]       = 0x0B;
    MixerRegisters[DSP_MIX_INGAINIDX_L]       = 0x00;
    MixerRegisters[DSP_MIX_INGAINIDX_R]       = 0x00;
    MixerRegisters[DSP_MIX_OUTGAINIDX_L]      = 0x00;
    MixerRegisters[DSP_MIX_OUTGAINIDX_R]      = 0x00;
    MixerRegisters[DSP_MIX_AGCIDX]            = 0x00;
    MixerRegisters[DSP_MIX_TREBLEIDX_L]       = 0x80;
    MixerRegisters[DSP_MIX_TREBLEIDX_R]       = 0x80;
    MixerRegisters[DSP_MIX_BASSIDX_L]         = 0x80;
    MixerRegisters[DSP_MIX_BASSIDX_R]         = 0x80;
}

/*****************************************************************************
 * CAdapterCommon::AcknowledgeIRQ()
 *****************************************************************************
 * Acknowledge interrupt request.
 */
void
CAdapterCommon::
AcknowledgeIRQ
(   void
)
{
    ASSERT(m_pWaveBase);
    WIN95COMPAT_ReadPortUChar(m_pWaveBase + DSP_REG_ACK16BIT);
    WIN95COMPAT_ReadPortUChar(m_pWaveBase + DSP_REG_ACK8BIT);
}

/*****************************************************************************
 * CAdapterCommon::ResetController()
 *****************************************************************************
 * Resets the controller.
 */
VOID
CAdapterCommon::
ResetController(void)
{
    WIN95COMPAT_WritePortUChar(m_pWaveBase + DSP_REG_RESET,1);
    WIN95COMPAT_ReadPortUChar(m_pWaveBase + DSP_REG_WRITE);
    WIN95COMPAT_ReadPortUChar(m_pWaveBase + DSP_REG_WRITE);
    WIN95COMPAT_WritePortUChar(m_pWaveBase + DSP_REG_RESET,0);
    ReadController();
}

/*****************************************************************************
 * InterruptServiceRoutine()
 *****************************************************************************
 * ISR.
 */
NTSTATUS
InterruptServiceRoutine
(
    IN      PINTERRUPTSYNC  InterruptSync,
    IN      PVOID           StaticContext,
    IN      PVOID           DynamicContextNotUsed   OPTIONAL
)
{
    ASSERT(InterruptSync);
    ASSERT(StaticContext);

    CAdapterCommon *that = (CAdapterCommon *) StaticContext;

    UCHAR portStatus = 0xff;

    //
    // If we have an MPU.
    //
    if (that->m_pUartBase)
    {
        //
        // Read the MPU status byte.
        //
        portStatus =
            WIN95COMPAT_ReadPortUChar(that->m_pUartBase + MPU401_REG_STATUS);
    }

    //
    // If there is outstanding work to do and there is a port-driver for
    // the MPU miniport...
    //
    if (!(portStatus & MPU401_DSR) && that->m_pPortUart)
    {
        //
        // ...notify the MPU port driver to do work.
        //
        that->m_pPortUart->Notify(NULL);
    }
    else
    {
        //
        // Must be a wave interrupt.
        //
        ASSERT(that->m_pWaveBase);
        ASSERT(that->m_pServiceGroupWave);

        //
        // Make sure there is a wave port driver.
        //
        if (that->m_pPortWave)
        {
            //
            // Tell it it needs to do some work.
            //
            that->m_pPortWave->Notify(that->m_pServiceGroupWave);
        }

        //
        // ACK the ISR.
        //
        WIN95COMPAT_ReadPortUChar(that->m_pWaveBase + DSP_REG_ACK8BIT);
        WIN95COMPAT_ReadPortUChar(that->m_pWaveBase + DSP_REG_ACK16BIT);
    }

    return STATUS_SUCCESS;
}
