/*****************************************************************************
 * miniport.cpp - SB16 topology miniport implementation
 *****************************************************************************
 * Copyright (c) Microsoft Corporation 1997
 */

#include "mintopo.h"

#define STR_MODULENAME "sb16topo: "

#define CHAN_LEFT       0
#define CHAN_RIGHT      1
#define CHAN_MASTER     (-1)


#pragma code_seg("PAGE")

#pragma warning(disable:4200)
typedef struct
{
    ULONG               InputChannels;
    ULONG               OutputChannels;
    KSAUDIO_MIX_CAPS    Capabilities[];
} KSAUDIO_MIXCAP_TABLE, *PKSAUDIO_MIXCAP_TABLE;
#pragma warning(default:4200)

static
LONG LogVolumeTable[] = { 0x00000034,     // -62dB
                          0x00000041,     // -60dB
                          0x00000052,     // -58dB
                          0x00000067,     // -56dB
                          0x00000082,     // -54dB
                          0x000000A4,     // -52dB
                          0x000000CF,     // -50dB
                          0x00000104,     // -48dB
                          0x00000148,     // -46dB
                          0x0000019D,     // -44dB
                          0x00000208,     // -42dB
                          0x0000028F,     // -40dB
                          0x00000339,     // -38dB
                          0x0000040E,     // -36dB
                          0x0000051B,     // -34dB
                          0x0000066E,     // -32dB
                          0x00000818,     // -30dB
                          0x00000A31,     // -28dB
                          0x00000CD4,     // -26dB
                          0x00001027,     // -24dB
                          0x00001455,     // -22dB
                          0x00001999,     // -20dB
                          0x0000203A,     // -18dB
                          0x00002892,     // -16dB
                          0x00003314,     // -14dB
                          0x0000404D,     // -12dB
                          0x000050F4,     // -10dB
                          0x000065EA,     // -8dB
                          0x0000804D,     // -6dB
                          0x0000A186,     // -4dB
                          0x0000CB59,     // -2dB
                          0x00010000 };   // 0dB

static
LONG LogGainTable[] = {   0x00010000,     // 0dB
                          0x0001FEC9,     // +6dB
                          0x0003FB27,     // +12dB
                          0x0007F17A };   // +18dB

static
LONG LogToneTable[] = {   0x00003314,     // -14dB
                          0x0000404D,     // -12dB
                          0x000050F4,     // -10dB
                          0x000065EA,     // -8dB
                          0x0000804D,     // -6dB
                          0x0000A186,     // -4dB
                          0x0000CB59,     // -2dB
                          0x00010000,     // -0dB
                          0x00010000,     // -0dB
                          0x00014248,     // +2dB
                          0x000195BB,     // +4dB
                          0x0001FEC9,     // +6dB
                          0x0002830A,     // +8dB
                          0x0003298B,     // +10dB
                          0x0003FB27,     // +12dB
                          0x0005030A };   // +14dB


/*****************************************************************************
 * CreateMiniportTopologySB16()
 *****************************************************************************
 * Creates a topology miniport object for the SB16 adapter.  This uses a
 * macro from STDUNK.H to do all the work.
 */
NTSTATUS
CreateMiniportTopologySB16
(
    OUT     PUNKNOWN *  Unknown,
    IN      REFCLSID,
    IN      PUNKNOWN    UnknownOuter    OPTIONAL,
    IN      POOL_TYPE   PoolType
)
{
    PAGED_CODE();

    ASSERT(Unknown);

    STD_CREATE_BODY(CMiniportTopologySB16,Unknown,UnknownOuter,PoolType);
}

/*****************************************************************************
 * CMiniportTopologySB16::ProcessResources()
 *****************************************************************************
 * Processes the resource list.
 */
NTSTATUS
CMiniportTopologySB16::
ProcessResources
(
    IN  PRESOURCELIST   ResourceList
)
{
    PAGED_CODE();

    ASSERT(ResourceList);

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::ProcessResources]"));

    //
    // Get counts for the types of resources.
    //
    ULONG       countIO     = ResourceList->NumberOfPorts();
    ULONG       countIRQ    = ResourceList->NumberOfInterrupts();
    ULONG       countDMA    = ResourceList->NumberOfDmas();

    NTSTATUS    ntStatus    = STATUS_SUCCESS;

    //
    // Make sure we have the expected number of resources.
    //
    if  (   (countIO != 1)
        ||  (countIRQ != 0)
        ||  (countDMA != 0)
        )
    {
        _DbgPrintF(DEBUGLVL_ERROR,("unknown configuraton"));
        ntStatus = STATUS_DEVICE_CONFIGURATION_ERROR;
    }

    if (NT_SUCCESS(ntStatus))
    {
        //
        // Get the port address.
        //
        PortBase =
            PUCHAR(ResourceList->FindTranslatedPort(0)->u.Port.Start.LowPart);
    }

    return ntStatus;
}

/*****************************************************************************
 * CMiniportTopologySB16::NonDelegatingQueryInterface()
 *****************************************************************************
 * Obtains an interface.  This function works just like a COM QueryInterface
 * call and is used if the object is not being aggregated.
 */
STDMETHODIMP
CMiniportTopologySB16::
NonDelegatingQueryInterface
(
    IN      REFIID  Interface,
    OUT     PVOID * Object
)
{
    PAGED_CODE();

    ASSERT(Object);

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::NonDelegatingQueryInterface]"));

    if (IsEqualGUIDAligned(Interface,IID_IUnknown))
    {
        *Object = PVOID(PUNKNOWN(this));
    }
    else
    if (IsEqualGUIDAligned(Interface,IID_IMiniport))
    {
        *Object = PVOID(PMINIPORT(this));
    }
    else
    if (IsEqualGUIDAligned(Interface,IID_IMiniportTopology))
    {
        *Object = PVOID(PMINIPORTTOPOLOGY(this));
    }
    else
    {
        *Object = NULL;
    }

    if (*Object)
    {
        //
        // We reference the interface for the caller.
        //
        PUNKNOWN(*Object)->AddRef();
        return STATUS_SUCCESS;
    }

    return STATUS_INVALID_PARAMETER;
}

/*****************************************************************************
 * CMiniportTopologySB16::~CMiniportTopologySB16()
 *****************************************************************************
 * Destructor.
 */
CMiniportTopologySB16::
~CMiniportTopologySB16
(   void
)
{
    PAGED_CODE();

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::~CMiniportTopologySB16]"));

    if (Port)
    {
        Port->Release();
    }
    if (AdapterCommon)
    {
        AdapterCommon->Release();
    }
}

/*****************************************************************************
 * CMiniportTopologySB16::Init()
 *****************************************************************************
 * Initializes a the miniport.
 */
STDMETHODIMP
CMiniportTopologySB16::
Init
(
    IN      PUNKNOWN        UnknownAdapter,
    IN      PRESOURCELIST   ResourceList,
    IN      PPORTTOPOLOGY   Port_
)
{
    PAGED_CODE();

    ASSERT(UnknownAdapter);
    ASSERT(ResourceList);
    ASSERT(Port_);

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::Init]"));

    //
    // AddRef() is required because we are keeping this pointer.
    //
    Port = Port_;
    Port->AddRef();

    NTSTATUS ntStatus =
        UnknownAdapter->QueryInterface
        (
            IID_IAdapterCommon,
            (PVOID *) &AdapterCommon
        );

    if (NT_SUCCESS(ntStatus))
    {
        ntStatus = ProcessResources(ResourceList);

        if (NT_SUCCESS(ntStatus))
        {
            AdapterCommon->MixerReset();
        }
    }

    return ntStatus;
}

/*****************************************************************************
 * CMiniportTopologySB16::GetDescription()
 *****************************************************************************
 * Gets the topology.
 */
STDMETHODIMP
CMiniportTopologySB16::
GetDescription
(
    OUT     PMINIPORT_DESCRIPTION * OutMiniportDescription
)
{
    PAGED_CODE();

    ASSERT(OutMiniportDescription);

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::GetDescription]"));

    *OutMiniportDescription = &MiniportDescription;

    return STATUS_SUCCESS;
}

#pragma code_seg()

/*****************************************************************************
 * CMiniportTopologySB16::ReadControlChanges()
 *****************************************************************************
 * Gets control change list entries.
 */
STDMETHODIMP
CMiniportTopologySB16::
ReadControlChanges
(
    IN      ULONG                   Requested,
    OUT     PULONG                  Delivered,
    OUT     PULONG                  Buffer
)
{
    ASSERT(Delivered);
    ASSERT(Buffer);

    return STATUS_NOT_IMPLEMENTED;
}

/*****************************************************************************
 * CMiniportTopologySB16::FlushControlChanges()
 *****************************************************************************
 * Flushes control change list entries.
 */
STDMETHODIMP
CMiniportTopologySB16::
FlushControlChanges
(   void
)
{
    return STATUS_NOT_IMPLEMENTED;
}

#pragma code_seg("PAGE")

/*****************************************************************************
 * SetStopState()
 *****************************************************************************
 * Stops the miniport.
 */
STDMETHODIMP
CMiniportTopologySB16::
SetStopState
(
    IN  ULONG   StopState
)
{
    PAGED_CODE();

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::SetStopState]"));

    return STATUS_NOT_IMPLEMENTED;
}

/*****************************************************************************
 * SetPowerState()
 *****************************************************************************
 * Changes the miniport's power state.
 */
STDMETHODIMP
CMiniportTopologySB16::
SetPowerState
(
    IN  ULONG   PowerState
)
{
    PAGED_CODE();

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::SetPowerState]"));

    return STATUS_NOT_IMPLEMENTED;
}

//
// TODO:  Node properties could be done more cleverly using tables.
//

/*****************************************************************************
 * PropertyHandler_OnOff()
 *****************************************************************************
 * Accesses a KSAUDIO_ONOFF value property.
 */
static
NTSTATUS
PropertyHandler_OnOff
(
    IN      PPROPERTY_REQUEST   PropertyRequest
)
{
    PAGED_CODE();

    ASSERT(PropertyRequest);

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::PropertyHandler_OnOff]"));

    CMiniportTopologySB16 *that =
        (CMiniportTopologySB16 *) PropertyRequest->Context;

    NTSTATUS        ntStatus = STATUS_INVALID_PARAMETER;
    BYTE            data;
    LONG            channel;

    // validate node
    if (PropertyRequest->Node != ULONG(-1))
    {
        if(PropertyRequest->Verb & KSPROPERTY_TYPE_GET)
        {
            // get the instance channel parameter
            if(PropertyRequest->InstanceSize >= sizeof(LONG))
            {
                channel = *(PLONG(PropertyRequest->Instance));

                // validate and get the output parameter
                if (PropertyRequest->ValueSize >= sizeof(BOOL))
                {
                    PBOOL OnOff = PBOOL(PropertyRequest->Value);
    
                    // switch on node id
                    switch(PropertyRequest->Node)
                    {
                        case MIC_AGC:   // Microphone AGC Control (mono)
                            // check if AGC property request on mono/left channel
                            if( ( PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_AGC ) &&
                                ( channel == CHAN_LEFT ) )
                            {
                                data = that->ReadBitsFromMixer( DSP_MIX_AGCIDX,
                                                          1,
                                                          MIXBIT_MIC_AGC );
                                *OnOff = data ? TRUE : FALSE;
                                PropertyRequest->ValueSize = sizeof(BOOL);
                                ntStatus = STATUS_SUCCESS;
                            }
                            break;
    
                        case MIC_LINEOUT_MUTE:  // Microphone Lineout Mute Control (mono)
                            // check if MUTE property request on mono/left channel
                            if( ( PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_MUTE ) &&
                                ( channel == CHAN_LEFT ) )
                            {
                                data = that->ReadBitsFromMixer( DSP_MIX_OUTMIXIDX,
                                                          1,
                                                          MIXBIT_MIC_LINEOUT );
                                *OnOff = data ? TRUE : FALSE;
                                PropertyRequest->ValueSize = sizeof(BOOL);
                                ntStatus = STATUS_SUCCESS;
                            }
                            break;
                    }
                }
            }

        } else if(PropertyRequest->Verb & KSPROPERTY_TYPE_SET)
        {
            // get the instance channel parameter
            if(PropertyRequest->InstanceSize >= sizeof(LONG))
            {
                channel = *(PLONG(PropertyRequest->Instance));
                
                // validate and get the input parameter
                if (PropertyRequest->ValueSize == sizeof(BOOL))
                {
                    BYTE value = *(PBOOL(PropertyRequest->Value)) ? 1 : 0;
    
                    // switch on the node id
                    switch(PropertyRequest->Node)
                    {
                        case MIC_AGC:   // Microphone AGC Control (mono)
                            // check if AGC property request on mono/left channel
                            if( ( PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_AGC ) &&
                                ( channel == CHAN_LEFT ) )
                            {
                                that->WriteBitsToMixer( DSP_MIX_AGCIDX,
                                                  1,
                                                  MIXBIT_MIC_AGC,
                                                  value );
                                ntStatus = STATUS_SUCCESS;
                            }
                            break;
    
                        case MIC_LINEOUT_MUTE:  // Microphone Lineout Mute Control (mono)
                            // check if MUTE property request on mono/left channel
                            if( ( PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_MUTE ) &&
                                ( channel == CHAN_LEFT ) )
                            {
                                that->WriteBitsToMixer( DSP_MIX_OUTMIXIDX,
                                                  1,
                                                  MIXBIT_MIC_LINEOUT,
                                                  value );
                                ntStatus = STATUS_SUCCESS;
                            }
                            break;
                    }
                }
            }

        } else if(PropertyRequest->Verb & KSPROPERTY_TYPE_BASICSUPPORT)
        {
            // service basic support request
            switch(PropertyRequest->Node)
            {
                case MIC_AGC:   // Microphone AGC Control (mono)
                    if(PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_AGC)
                    {
                        ntStatus = STATUS_SUCCESS;
                    }
                    break;

                case MIC_LINEOUT_MUTE:  // Microphone Lineout Mute Control (mono)
                    if(PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_MUTE)
                    {
                        ntStatus = STATUS_SUCCESS;
                    }
                    break;
            }
        }
    }

    return ntStatus;
}

/*****************************************************************************
 * PropertyHandler_Level()
 *****************************************************************************
 * Accesses a KSAUDIO_LEVEL property.
 */
static
NTSTATUS
PropertyHandler_Level
(
    IN      PPROPERTY_REQUEST   PropertyRequest
)
{
    PAGED_CODE();

    ASSERT(PropertyRequest);

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::PropertyHandler_Level]"));

    CMiniportTopologySB16 *that =
        (CMiniportTopologySB16 *) PropertyRequest->Context;

    NTSTATUS        ntStatus = STATUS_INVALID_PARAMETER;
    BYTE            data;
    ULONG           count;
    LONG            channel;

    // validate node
    if(PropertyRequest->Node != ULONG(-1))
    {
        if(PropertyRequest->Verb & KSPROPERTY_TYPE_GET)
        {
            // get the instance channel parameter
            if(PropertyRequest->InstanceSize >= sizeof(LONG))
            {
                channel = *(PLONG(PropertyRequest->Instance));

                // only support get requests on either mono/left (0) or right (1) channels
                if ( (channel == CHAN_LEFT) || (channel == CHAN_RIGHT) )
                {
                    // validate and get the output parameter
                    if (PropertyRequest->ValueSize >= sizeof(LONG))
                    {
                        PLONG Level = (PLONG)PropertyRequest->Value;

                        // switch on node if
                        switch(PropertyRequest->Node)
                        {
                            case WAVEOUT_VOLUME:
                            case SYNTH_VOLUME:
                            case CD_VOLUME:
                            case LINEIN_VOLUME:
                            case MIC_VOLUME:
                            case LINEOUT_VOL:
                                // check if volume property request
                                if(PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_VOLUMELEVEL)
                                {
                                    // bail out if a right channel request on the mono mic volume
                                    if( (PropertyRequest->Node == MIC_VOLUME) &&
                                        (channel != CHAN_LEFT) )
                                    {
                                        break;
                                    }
                                    data = that->ReadBitsFromMixer( BaseRegisters[PropertyRequest->Node] + UCHAR(channel),
                                                                    5,
                                                                    3 );
                                    *Level = LogVolumeTable[data];
                                    PropertyRequest->ValueSize = sizeof(LONG);
                                    ntStatus = STATUS_SUCCESS;
                                }
                                break;
        
                            case LINEOUT_GAIN:
                            case WAVEIN_GAIN:
                                // check if volume property request
                                if(PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_VOLUMELEVEL)
                                {
                                    data = that->ReadBitsFromMixer( BaseRegisters[PropertyRequest->Node] + UCHAR(channel),
                                                                    2,
                                                                    6 );
                                    *Level = LogGainTable[data];
                                    PropertyRequest->ValueSize = sizeof(LONG);
                                    ntStatus = STATUS_SUCCESS;
                                }
                                break;
        
                            case LINEOUT_TONE:
                                if(PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_TREBLE)
                                {
                                    data = that->ReadBitsFromMixer( BaseRegisters[PropertyRequest->Node] + UCHAR(channel),
                                                                    4,
                                                                    4 );
                                    *Level = LogToneTable[data];
                                    PropertyRequest->ValueSize = sizeof(LONG);
                                    ntStatus = STATUS_SUCCESS;
                                } else if(PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_BASS)
                                {
                                    data = that->ReadBitsFromMixer( BaseRegisters[PropertyRequest->Node] + 2 + UCHAR(channel),
                                                                    4,
                                                                    4 );
                                    *Level = LogToneTable[data];
                                    PropertyRequest->ValueSize = sizeof(LONG);
                                    ntStatus = STATUS_SUCCESS;
                                }
                                break;
                        }
                    }
                }
            }

        } else if(PropertyRequest->Verb & KSPROPERTY_TYPE_SET)
        {
            // get the instance channel parameter
            if(PropertyRequest->InstanceSize >= sizeof(LONG))
            {
                channel = *(PLONG(PropertyRequest->Instance));

                // only support get requests on either mono/left (0), right (1), or master (-1) channels
                if ( (channel == CHAN_LEFT) || (channel == CHAN_RIGHT) || (channel == CHAN_MASTER))
                {
                    // validate and get the input parameter
                    if (PropertyRequest->ValueSize == sizeof(LONG))
                    {
                        PLONG Level = (PLONG)PropertyRequest->Value;

                        // switch on the node id
                        switch(PropertyRequest->Node)
                        {
                            case WAVEOUT_VOLUME:
                            case SYNTH_VOLUME:
                            case CD_VOLUME:
                            case LINEIN_VOLUME:
                            case MIC_VOLUME:
                            case LINEOUT_VOL:
                                if(PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_VOLUMELEVEL)
                                {
                                    if(*Level >= 0)
                                    {
                                        // determine register bits by checking log table index
                                        for(count=0;count<SIZEOF_ARRAY(LogVolumeTable)-1;count++)
                                        {
                                            if(*Level <= LogVolumeTable[count]) break;
                                        }

                                        // set right channel if channel requested is right or master
                                        // and node is not mic volume (mono)
                                        if ( ( (channel == CHAN_RIGHT) || (channel == CHAN_MASTER) ) &&
                                             ( PropertyRequest->Node != MIC_VOLUME ) )
                                        {
                                            that->WriteBitsToMixer( BaseRegisters[PropertyRequest->Node]+1,
                                                              5,
                                                              3,
                                                              BYTE(count) );
                                            ntStatus = STATUS_SUCCESS;
                                        }
                                        // set the left channel if channel requested is left or master
                                        if ( (channel == CHAN_LEFT) || (channel == CHAN_MASTER) )
                                        {
                                            that->WriteBitsToMixer( BaseRegisters[PropertyRequest->Node],
                                                              5,
                                                              3,
                                                              BYTE(count) );
                                            ntStatus = STATUS_SUCCESS;
                                        }
                                    }
                                }
                                break;
        
                            case LINEOUT_GAIN:
                            case WAVEIN_GAIN:
                                if(PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_VOLUMELEVEL)
                                {
                                    if(*Level >= 0)
                                    {
                                        // determine register bits by checking log table index
                                        for(count=0;count<SIZEOF_ARRAY(LogGainTable)-1;count++)
                                        {
                                            if(*Level <= LogGainTable[count]) break;
                                        }
        
                                        // set right channel if channel requested is right or master
                                        if ( (channel == CHAN_RIGHT) || (channel == CHAN_MASTER) )
                                        {
                                            that->WriteBitsToMixer( BaseRegisters[PropertyRequest->Node]+1,
                                                              2,
                                                              6,
                                                              BYTE(count) );
                                            ntStatus = STATUS_SUCCESS;
                                        }
                                        // set the left channel if channel requested is left or master
                                        if ( (channel == CHAN_LEFT) || (channel == CHAN_MASTER) )
                                        {
                                            that->WriteBitsToMixer( BaseRegisters[PropertyRequest->Node],
                                                              2,
                                                              6,
                                                              BYTE(count) );
                                            ntStatus = STATUS_SUCCESS;
                                        }
                                    }
                                }
                                break;
        
                            case LINEOUT_TONE:
                                if  (   (PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_TREBLE)
                                    ||  (PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_BASS)
                                    )
                                {
                                    BYTE RegisterOffset = 0;
        
                                    // set offset based on requested property type
                                    if(PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_BASS)
                                    {
                                        RegisterOffset = 2;
                                    }
        
                                    if(*Level >= 0)
                                    {
                                        // determine register bits by checking log table index
                                        for(count=0;count<SIZEOF_ARRAY(LogToneTable)-1;count++)
                                        {
                                            if(*Level <= LogToneTable[count]) break;
                                        }
        
                                        // set right channel if channel requested is right or master
                                        if ( (channel == CHAN_RIGHT) || (channel == CHAN_MASTER) )
                                        {
                                            that->WriteBitsToMixer( BaseRegisters[PropertyRequest->Node]+RegisterOffset+1,
                                                              4,
                                                              4,
                                                              BYTE(count) );
                                            ntStatus = STATUS_SUCCESS;
                                        }
                                        // set the left channel if channel requested is left or master
                                        if ( (channel == CHAN_LEFT) || (channel == CHAN_MASTER) )
                                        {
                                            that->WriteBitsToMixer( BaseRegisters[PropertyRequest->Node]+RegisterOffset,
                                                              4,
                                                              4,
                                                              BYTE(count) );
                                            ntStatus = STATUS_SUCCESS;
                                        }
                                    }
                                }
                                break;
                        }
                    }
                }
            }

        } else if(PropertyRequest->Verb & KSPROPERTY_TYPE_BASICSUPPORT)
        {
            // service basic support request
            switch(PropertyRequest->Node)
            {
                case WAVEOUT_VOLUME:
                case SYNTH_VOLUME:
                case CD_VOLUME:
                case LINEIN_VOLUME:
                case MIC_VOLUME:
                case LINEOUT_VOL:
                case LINEOUT_GAIN:
                case WAVEIN_GAIN:
                    if(PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_VOLUMELEVEL)
                    {
                        ntStatus = STATUS_SUCCESS;
                    }
                    break;

                case LINEOUT_TONE:
                    if  (   (PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_BASS)
                        ||  (PropertyRequest->PropertyItem->Id == KSPROPERTY_AUDIO_TREBLE)
                        )
                    {
                        ntStatus = STATUS_SUCCESS;
                    }
                    break;
            }
        }
    }

    return ntStatus;
}

/*****************************************************************************
 * PropertyHandler_SuperMixCaps()
 *****************************************************************************
 * Handles supermixer caps accesses
 */
static
NTSTATUS
PropertyHandler_SuperMixCaps
(
    IN      PPROPERTY_REQUEST   PropertyRequest
)
{
    PAGED_CODE();

    ASSERT(PropertyRequest);

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::PropertyHandler_SuperMixCaps]"));

    CMiniportTopologySB16 *that =
        (CMiniportTopologySB16 *) PropertyRequest->Context;

    NTSTATUS        ntStatus = STATUS_INVALID_PARAMETER;
    ULONG           count;

    // validate node
    if(PropertyRequest->Node != ULONG(-1))
    {
        if(PropertyRequest->Verb & KSPROPERTY_TYPE_GET)
        {
            switch(PropertyRequest->Node)
            {
                // Full 2x2 Switches
                case SYNTH_WAVEIN_SUPERMIX:
                case CD_WAVEIN_SUPERMIX:
                case LINEIN_WAVEIN_SUPERMIX:
                    if(PropertyRequest->ValueSize == sizeof(ULONG))
                    {
                        PULONG Size = (PULONG)PropertyRequest->Value;
                        *Size = 2 * sizeof(ULONG) + 4 * sizeof(KSAUDIO_MIX_CAPS);
                        ntStatus = STATUS_SUCCESS;
                    } else if(PropertyRequest->ValueSize == 2 * sizeof(ULONG))
                    {
                        PKSAUDIO_MIXCAP_TABLE MixCaps = (PKSAUDIO_MIXCAP_TABLE)PropertyRequest->Value;
                        MixCaps->InputChannels = 2;
                        MixCaps->OutputChannels = 2;
                        ntStatus = STATUS_SUCCESS;
                    } else if(PropertyRequest->ValueSize >= 2 * sizeof(ULONG) + 4 * sizeof(KSAUDIO_MIX_CAPS))
                    {
                        PropertyRequest->ValueSize = 2 * sizeof(ULONG) + 4 * sizeof(KSAUDIO_MIX_CAPS);

                        PKSAUDIO_MIXCAP_TABLE MixCaps = (PKSAUDIO_MIXCAP_TABLE)PropertyRequest->Value;
                        MixCaps->InputChannels = 2;
                        MixCaps->OutputChannels = 2;
                        for(count = 0; count < 4; count++)
                        {
                            MixCaps->Capabilities[count].Mute = TRUE;
                            MixCaps->Capabilities[count].Minimum = 0x10000;
                            MixCaps->Capabilities[count].Maximum = 0x10000;
                            MixCaps->Capabilities[count].Reset = 0x10000;
                        }
                        ntStatus = STATUS_SUCCESS;
                    }
                    break;

                // Limited 2x2 Switches
                case CD_LINEOUT_SUPERMIX:
                case LINEIN_LINEOUT_SUPERMIX:
                    if(PropertyRequest->ValueSize == sizeof(ULONG))
                    {
                        PULONG Size = (PULONG)PropertyRequest->Value;
                        *Size = 2 * sizeof(ULONG) + 4 * sizeof(KSAUDIO_MIX_CAPS);
                        ntStatus = STATUS_SUCCESS;
                    } else if(PropertyRequest->ValueSize == 2 * sizeof(ULONG))
                    {
                        PKSAUDIO_MIXCAP_TABLE MixCaps = (PKSAUDIO_MIXCAP_TABLE)PropertyRequest->Value;
                        MixCaps->InputChannels = 2;
                        MixCaps->OutputChannels = 2;
                        ntStatus = STATUS_SUCCESS;
                    } else if(PropertyRequest->ValueSize >= 2 * sizeof(ULONG) + 4 * sizeof(KSAUDIO_MIX_CAPS))
                    {
                        PropertyRequest->ValueSize = 2 * sizeof(ULONG) + 4 * sizeof(KSAUDIO_MIX_CAPS);

                        PKSAUDIO_MIXCAP_TABLE MixCaps = (PKSAUDIO_MIXCAP_TABLE)PropertyRequest->Value;
                        MixCaps->InputChannels = 2;
                        MixCaps->OutputChannels = 2;
                        for(count = 0; count < 4; count++)
                        {
                            if((count == 0) || (count == 2))
                            {
                                MixCaps->Capabilities[count].Mute = TRUE;
                                MixCaps->Capabilities[count].Minimum = 0x10000;
                                MixCaps->Capabilities[count].Maximum = 0x10000;
                                MixCaps->Capabilities[count].Reset = 0x10000;
                            } else
                            {
                                MixCaps->Capabilities[count].Mute = FALSE;
                                MixCaps->Capabilities[count].Minimum = 0;
                                MixCaps->Capabilities[count].Maximum = 0;
                                MixCaps->Capabilities[count].Reset = 0;
                            }
                        }
                        ntStatus = STATUS_SUCCESS;
                    }
                    break;


                // 1x2 Switch
                case MIC_WAVEIN_SUPERMIX:
                    if(PropertyRequest->ValueSize == sizeof(ULONG))
                    {
                        PULONG Size = (PULONG)PropertyRequest->Value;
                        *Size = 2 * sizeof(ULONG) + 2 * sizeof(KSAUDIO_MIX_CAPS);
                        ntStatus = STATUS_SUCCESS;
                    } else if(PropertyRequest->ValueSize == 2 * sizeof(ULONG))
                    {
                        PKSAUDIO_MIXCAP_TABLE MixCaps = (PKSAUDIO_MIXCAP_TABLE)PropertyRequest->Value;
                        MixCaps->InputChannels = 1;
                        MixCaps->OutputChannels = 2;
                        ntStatus = STATUS_SUCCESS;
                    } else if(PropertyRequest->ValueSize >= 2 * sizeof(ULONG) + 2 * sizeof(KSAUDIO_MIX_CAPS))
                    {
                        PropertyRequest->ValueSize = 2 * sizeof(ULONG) + 2 * sizeof(KSAUDIO_MIX_CAPS);

                        PKSAUDIO_MIXCAP_TABLE MixCaps = (PKSAUDIO_MIXCAP_TABLE)PropertyRequest->Value;
                        MixCaps->InputChannels = 1;
                        MixCaps->OutputChannels = 2;
                        for(count = 0; count < 2; count++)
                        {
                            MixCaps->Capabilities[count].Mute = TRUE;
                            MixCaps->Capabilities[count].Minimum = 0x10000;
                            MixCaps->Capabilities[count].Maximum = 0x10000;
                            MixCaps->Capabilities[count].Reset = 0x10000;
                        }
                        ntStatus = STATUS_SUCCESS;
                    }
                    break;
            }

        } else if(PropertyRequest->Verb & KSPROPERTY_TYPE_BASICSUPPORT)
        {
            // service basic support request
            switch(PropertyRequest->Node)
            {
                case SYNTH_WAVEIN_SUPERMIX:
                case CD_WAVEIN_SUPERMIX:
                case LINEIN_WAVEIN_SUPERMIX:
                case CD_LINEOUT_SUPERMIX:
                case LINEIN_LINEOUT_SUPERMIX:
                case MIC_WAVEIN_SUPERMIX:
                    ntStatus = STATUS_SUCCESS;
                    break;
            }
        }
    }

    return ntStatus;
}

/*****************************************************************************
 * PropertyHandler_SuperMixTable()
 *****************************************************************************
 * Handles supermixer level accesses
 */
static
NTSTATUS
PropertyHandler_SuperMixTable
(
    IN      PPROPERTY_REQUEST   PropertyRequest
)
{
    PAGED_CODE();

    ASSERT(PropertyRequest);

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CMiniportTopologySB16::PropertyHandler_SuperMixTable]"));

    CMiniportTopologySB16 *that =
        (CMiniportTopologySB16 *) PropertyRequest->Context;

    NTSTATUS        ntStatus = STATUS_INVALID_PARAMETER;
    BYTE            dataL,dataR;

    // validate node
    if(PropertyRequest->Node != ULONG(-1))
    {
        if(PropertyRequest->Verb & KSPROPERTY_TYPE_GET)
        {
            switch(PropertyRequest->Node)
            {
                // Full 2x2 Switches
                case SYNTH_WAVEIN_SUPERMIX:
                case CD_WAVEIN_SUPERMIX:
                case LINEIN_WAVEIN_SUPERMIX:
                    if(PropertyRequest->ValueSize == sizeof(ULONG))
                    {
                        PULONG Size = (PULONG)PropertyRequest->Value;
                        *Size = 4 * sizeof(KSAUDIO_MIXLEVEL);
                        ntStatus = STATUS_SUCCESS;
                    } else if(PropertyRequest->ValueSize >= 4 * sizeof(KSAUDIO_MIXLEVEL))
                    {
                        PropertyRequest->ValueSize = 4 * sizeof(KSAUDIO_MIXLEVEL);

                        PKSAUDIO_MIXLEVEL MixLevel = (PKSAUDIO_MIXLEVEL)PropertyRequest->Value;

                        dataL = that->ReadBitsFromMixer( DSP_MIX_ADCMIXIDX_L,
                                                  2,
                                                  BaseRegisters[PropertyRequest->Node] );
                        dataR = that->ReadBitsFromMixer( DSP_MIX_ADCMIXIDX_R,
                                                  2,
                                                  BaseRegisters[PropertyRequest->Node] );

                        MixLevel[0].Mute = dataL & 0x2 ? TRUE : FALSE;          // left to left mute
                        MixLevel[0].Level = 0x10000;

                        MixLevel[1].Mute = dataR & 0x2 ? TRUE : FALSE;          // left to right mute
                        MixLevel[1].Level = 0x10000;

                        MixLevel[2].Mute = dataL & 0x1 ? TRUE : FALSE;          // right to left mute
                        MixLevel[2].Level = 0x10000;

                        MixLevel[3].Mute = dataR & 0x1 ? TRUE : FALSE;          // right to right mute
                        MixLevel[3].Level = 0x10000;

                        ntStatus = STATUS_SUCCESS;
                    }
                    break;

                // Limited 2x2 Switches
                case CD_LINEOUT_SUPERMIX:
                case LINEIN_LINEOUT_SUPERMIX:
                    if(PropertyRequest->ValueSize == sizeof(ULONG))
                    {
                        PULONG Size = (PULONG)PropertyRequest->Value;
                        *Size = 4 * sizeof(KSAUDIO_MIXLEVEL);
                        ntStatus = STATUS_SUCCESS;
                    } else if(PropertyRequest->ValueSize >= 4 * sizeof(KSAUDIO_MIXLEVEL))
                    {
                        PropertyRequest->ValueSize = 4 * sizeof(KSAUDIO_MIXLEVEL);

                        PKSAUDIO_MIXLEVEL MixLevel = (PKSAUDIO_MIXLEVEL)PropertyRequest->Value;

                        dataL = that->ReadBitsFromMixer( DSP_MIX_OUTMIXIDX,
                                                   2,
                                                   BaseRegisters[PropertyRequest->Node] );

                        MixLevel[0].Mute = dataL & 0x2 ? TRUE : FALSE;          // left to left mute
                        MixLevel[0].Level = 0x10000;

                        MixLevel[1].Mute = FALSE;
                        MixLevel[1].Level = 0;

                        MixLevel[2].Mute = FALSE;
                        MixLevel[2].Level = 0;

                        MixLevel[3].Mute = dataL & 0x1 ? TRUE : FALSE;          // right to right mute
                        MixLevel[3].Level = 0x10000;

                        ntStatus = STATUS_SUCCESS;
                    }
                    break;


                // 1x2 Switch
                case MIC_WAVEIN_SUPERMIX:
                    if(PropertyRequest->ValueSize == sizeof(ULONG))
                    {
                        PULONG Size = (PULONG)PropertyRequest->Value;
                        *Size = 2 * sizeof(KSAUDIO_MIXLEVEL);
                        ntStatus = STATUS_SUCCESS;
                    } else if(PropertyRequest->ValueSize >= 2 * sizeof(KSAUDIO_MIXLEVEL))
                    {
                        PropertyRequest->ValueSize = 2 * sizeof(KSAUDIO_MIXLEVEL);

                        PKSAUDIO_MIXLEVEL MixLevel = (PKSAUDIO_MIXLEVEL)PropertyRequest->Value;

                        dataL = that->ReadBitsFromMixer( DSP_MIX_ADCMIXIDX_L,
                                                  1,
                                                  MIXBIT_MIC_WAVEIN );
                        dataR = that->ReadBitsFromMixer( DSP_MIX_ADCMIXIDX_R,
                                                  1,
                                                  MIXBIT_MIC_WAVEIN );

                        MixLevel[0].Mute = dataL & 0x1 ? TRUE : FALSE;          // mono to left mute
                        MixLevel[0].Level = 0x10000;

                        MixLevel[1].Mute = dataR & 0x1 ? TRUE : FALSE;          // mono to right mute
                        MixLevel[1].Level = 0x10000;

                        ntStatus = STATUS_SUCCESS;
                    }
                    break;
            }

        } else if(PropertyRequest->Verb & KSPROPERTY_TYPE_SET)
        {
            switch(PropertyRequest->Node)
            {
                // Full 2x2 Switches
                case SYNTH_WAVEIN_SUPERMIX:
                case CD_WAVEIN_SUPERMIX:
                case LINEIN_WAVEIN_SUPERMIX:
                    if(PropertyRequest->ValueSize == 4 * sizeof(KSAUDIO_MIXLEVEL))
                    {
                        PKSAUDIO_MIXLEVEL MixLevel = (PKSAUDIO_MIXLEVEL)PropertyRequest->Value;

                        dataL = MixLevel[0].Mute ? 0x2 : 0x0;
                        dataL |= MixLevel[2].Mute ? 0x1 : 0x0;

                        dataR = MixLevel[1].Mute ? 0x2 : 0x0;
                        dataR |= MixLevel[3].Mute ? 0x1 : 0x0;


                        that->WriteBitsToMixer( DSP_MIX_ADCMIXIDX_L,
                                          2,
                                          BaseRegisters[PropertyRequest->Node],
                                          dataL );

                        that->WriteBitsToMixer( DSP_MIX_ADCMIXIDX_R,
                                          2,
                                          BaseRegisters[PropertyRequest->Node],
                                          dataR );

                        ntStatus = STATUS_SUCCESS;
                    }
                    break;

                // Limited 2x2 Switches
                case CD_LINEOUT_SUPERMIX:
                case LINEIN_LINEOUT_SUPERMIX:
                    if(PropertyRequest->ValueSize == 4 * sizeof(KSAUDIO_MIXLEVEL))
                    {
                        PKSAUDIO_MIXLEVEL MixLevel = (PKSAUDIO_MIXLEVEL)PropertyRequest->Value;

                        dataL = MixLevel[0].Mute ? 0x2 : 0x0;
                        dataL |= MixLevel[3].Mute ? 0x1 : 0x0;

                        that->WriteBitsToMixer( DSP_MIX_OUTMIXIDX,
                                          2,
                                          BaseRegisters[PropertyRequest->Node],
                                          dataL );

                        ntStatus = STATUS_SUCCESS;
                    }
                    break;


                // 1x2 Switch
                case MIC_WAVEIN_SUPERMIX:
                    if(PropertyRequest->ValueSize == 2 * sizeof(KSAUDIO_MIXLEVEL))
                    {
                        PKSAUDIO_MIXLEVEL MixLevel = (PKSAUDIO_MIXLEVEL)PropertyRequest->Value;

                        dataL = MixLevel[0].Mute ? 0x1 : 0x0;
                        dataR = MixLevel[1].Mute ? 0x1 : 0x0;

                        that->WriteBitsToMixer( DSP_MIX_ADCMIXIDX_L,
                                          1,
                                          MIXBIT_MIC_WAVEIN,
                                          dataL );

                        that->WriteBitsToMixer( DSP_MIX_ADCMIXIDX_R,
                                          1,
                                          MIXBIT_MIC_WAVEIN,
                                          dataR );

                        ntStatus = STATUS_SUCCESS;
                    }
                    break;
            }

        } else if(PropertyRequest->Verb & KSPROPERTY_TYPE_BASICSUPPORT)
        {
            // service basic support request
            switch(PropertyRequest->Node)
            {
                case SYNTH_WAVEIN_SUPERMIX:
                case CD_WAVEIN_SUPERMIX:
                case LINEIN_WAVEIN_SUPERMIX:
                case CD_LINEOUT_SUPERMIX:
                case LINEIN_LINEOUT_SUPERMIX:
                case MIC_WAVEIN_SUPERMIX:
                    ntStatus = STATUS_SUCCESS;
                    break;
            }
        }
    }

    return ntStatus;
}

/*****************************************************************************
 * ThisManyOnes()
 *****************************************************************************
 * Returns a byte with the indicated number of ones in the low end.
 */
inline
BYTE
ThisManyOnes
(
    IN      BYTE Ones
)
{
    return ~(BYTE(0xff) << Ones);
}

/*****************************************************************************
 * CMiniportTopologySB16::ReadBitsFromMixer()
 *****************************************************************************
 * Reads specified bits from a mixer register.
 */
BYTE
CMiniportTopologySB16::
ReadBitsFromMixer
(
    BYTE Reg,
    BYTE Bits,
    BYTE Shift
)
{
    BYTE data = AdapterCommon->MixerRegRead(Reg);

    return( data >> Shift) & ThisManyOnes(Bits);
}

/*****************************************************************************
 * CMiniportTopologySB16::WriteBitsToMixer()
 *****************************************************************************
 * Writes specified bits to a mixer register.
 */
void
CMiniportTopologySB16::
WriteBitsToMixer
(
    BYTE Reg,
    BYTE Bits,
    BYTE Shift,
    BYTE Value
)
{
    BYTE mask = ThisManyOnes(Bits) << Shift;
    BYTE data = AdapterCommon->MixerRegRead(Reg);

    if(Reg < DSP_MIX_MAXREGS)
    {
        AdapterCommon->MixerRegWrite( Reg,
                                      (data & ~mask) | ( (Value << Shift) & mask));
    }
}

#pragma code_seg()
