/*****************************************************************************
 * private.h - SB16 wave miniport private definitions
 *****************************************************************************
 * Copyright (c) 1997 Microsoft Corporation
 */

#ifndef _SB16WAVE_PRIVATE_H_
#define _SB16WAVE_PRIVATE_H_

#include "stdunk.h"
#include "common.h"




 
/*****************************************************************************
 * Classes
 */

/*****************************************************************************
 * CMiniportWaveCyclicSB16
 *****************************************************************************
 * SB16 wave miniport.  This object is associated with the device and is
 * created when the device is started.  The class inherits IMiniportWaveCyclic
 * so it can expose this interface and CUnknown so it automatically gets
 * reference counting and aggregation support.
 */
class CMiniportWaveCyclicSB16
:   public IMiniportWaveCyclic,
    public CUnknown
{
private:
    PADAPTERCOMMON      AdapterCommon;              // Adapter common object.
    PPORTWAVECYCLIC     Port;                       // Callback interface.
    PUCHAR              PortBase;                   // Base port address.

    ULONG               NotificationInterval;       // In milliseconds.
    ULONG               SamplingFrequency;          // Frames per second.

    BOOLEAN             AllocatedCapture;           // Capture in use.
    BOOLEAN             AllocatedRender;            // Render in use.
    BOOLEAN             Allocated8Bit;              // 8-bit DMA in use.
    BOOLEAN             Allocated16Bit;             // 16-bit DMA in use.

    PDMACHANNEL         DmaChannel8;                // Abstracted channel.
    PDMACHANNEL         DmaChannel16;               // Abstracted channel.

    PSERVICEGROUP       ServiceGroup;               // For notification.

    /*************************************************************************
     * CMiniportWaveCyclicSB16 methods
     *
     * These are private member functions used internally by the object.  See
     * MINIPORT.CPP for specific descriptions.
     */
    BOOLEAN ConfigureDevice
    (
        IN      ULONG   Interrupt,
        IN      ULONG   DMA8Bit,
        IN      ULONG   DMA16Bit
    );
    NTSTATUS ProcessResources
    (
        IN      PRESOURCELIST   ResourceList
    );
    NTSTATUS ValidateFormat
    (
        IN      PKSDATAFORMAT   Format
    );

public:
    /*************************************************************************
     * The following two macros are from STDUNK.H.  DECLARE_STD_UNKNOWN()
     * defines inline IUnknown implementations that use CUnknown's aggregation
     * support.  NonDelegatingQueryInterface() is declared, but it cannot be
     * implemented generically.  Its definition appears in MINIPORT.CPP.
     * DEFINE_STD_CONSTRUCTOR() defines inline a constructor which accepts
     * only the outer unknown, which is used for aggregation.  The standard
     * create macro (in MINIPORT.CPP) uses this constructor.
     */
    DECLARE_STD_UNKNOWN();
    DEFINE_STD_CONSTRUCTOR(CMiniportWaveCyclicSB16);

    ~CMiniportWaveCyclicSB16();

    /*************************************************************************
     * IMiniport methods
     */
    STDMETHODIMP SetStopState
    (
        IN      ULONG   StopState
    );
    STDMETHODIMP SetPowerState
    (
        IN      ULONG   PowerState
    );
    STDMETHODIMP GetDescription
    (
        OUT     PMINIPORT_DESCRIPTION * OutMiniportDescription
    );

    /*************************************************************************
     * IMiniportWaveCyclic methods
     */
    STDMETHODIMP Init
    (
        IN      PUNKNOWN        UnknownAdapter,
        IN      PRESOURCELIST   ResourceList,
        IN      PPORTWAVECYCLIC Port
    );
    STDMETHODIMP NewStream
    (
        OUT     PMINIPORTWAVECYCLICSTREAM * Stream,
        IN      PUNKNOWN                    OuterUnknown,
        IN      POOL_TYPE                   PoolType,
        IN      ULONG                       Channel,
        IN      BOOLEAN                     Capture,
        IN      PKSDATAFORMAT               DataFormat,
        OUT     PDMACHANNEL *               DmaChannel,
        OUT     PSERVICEGROUP *             ServiceGroup
    );

    /*************************************************************************
     * Friends
     *
     * The miniport stream class is a friend because it needs to access the
     * private member variables of this class.
     */
    friend class CMiniportWaveCyclicStreamSB16;

    friend
    NTSTATUS
    PropertyHandler_AudioSamplingRate
    (
        IN      PPROPERTY_REQUEST   PropertyRequest
    );
};

/*****************************************************************************
 * CMiniportWaveCyclicStreamSB16
 *****************************************************************************
 * SB16 wave miniport stream.  This object is associated with a streaming pin
 * and is created when a pin is created on the filter.  The class inherits
 * IMiniportWaveCyclicStream so it can expose this interface and CUnknown so
 * it automatically gets reference counting and aggregation support.
 */
class CMiniportWaveCyclicStreamSB16
:   public IMiniportWaveCyclicStream,
    public CUnknown
{
private:
    CMiniportWaveCyclicSB16 *   Miniport;       // Miniport that created us.
    ULONG                       Channel;        // Index into channel list.
    BOOLEAN                     Capture;        // Capture or render.
    BOOLEAN                     Format16Bit;    // 16- or 8-bit samples.
    BOOLEAN                     FormatStereo;   // Two or one channel.
    KSSTATE                     State;          // Stop, pause, run.
    PDMACHANNEL                 DmaChannel;     // DMA channel to use.

public:
    /*************************************************************************
     * The following two macros are from STDUNK.H.  DECLARE_STD_UNKNOWN()
     * defines inline IUnknown implementations that use CUnknown's aggregation
     * support.  NonDelegatingQueryInterface() is declared, but it cannot be
     * implemented generically.  Its definition appears in MINIPORT.CPP.
     * DEFINE_STD_CONSTRUCTOR() defines inline a constructor which accepts
     * only the outer unknown, which is used for aggregation.  The standard
     * create macro (in MINIPORT.CPP) uses this constructor.
     */
    DECLARE_STD_UNKNOWN();
    DEFINE_STD_CONSTRUCTOR(CMiniportWaveCyclicStreamSB16);

    ~CMiniportWaveCyclicStreamSB16();

    /*************************************************************************
     * CMiniportWaveCyclicStreamSB16 methods
     */
    NTSTATUS Init
    (
        IN      CMiniportWaveCyclicSB16 *   Miniport,
        IN      ULONG                       Channel,
        IN      BOOLEAN                     Capture,
        IN      PKSDATAFORMAT               DataFormat,
        OUT     PDMACHANNEL                 DmaChannel
    );

    /*************************************************************************
     * IMiniportWaveCyclicStream methods
     */
    STDMETHODIMP SetFormat
    (
        IN      PKSDATAFORMAT   DataFormat
    );
    STDMETHODIMP_(ULONG) SetNotificationFreq
    (
        IN      ULONG           Interval
    );
    STDMETHODIMP SetState
    (
        IN      KSSTATE         State
    );
    STDMETHODIMP GetPosition
    (
        OUT     PULONG          Position
    );
};





#endif