/*++

Copyright (c) 1996    Microsoft Corporation

Module Name:

    read.c

Abstract:

    This module contains the code for translating HID reports to something
    useful.

Environment:

    Kernel & user mode

Revision History:

    Nov-96 : Created by Kenneth D. Ray

--*/

#include <stdlib.h>
#include <wtypes.h>
#include "hidsdi.h"
#include "hid.h"

BOOLEAN
UnpackReport (
   IN       PCHAR                ReportBuffer,
   IN       USHORT               ReportBufferLength,
   IN       HIDP_REPORT_TYPE     ReportType,
   IN OUT   PHID_DATA            Data,
   IN       ULONG                DataLength,
   IN       PHIDP_PREPARSED_DATA Ppd
   );

BOOLEAN
PackReport (
   OUT      PCHAR                ReportBuffer,
   IN       USHORT               ReportBufferLength,
   IN       HIDP_REPORT_TYPE     ReportType,
   IN       PHID_DATA            Data,
   IN       ULONG                DataLength,
   IN       PHIDP_PREPARSED_DATA Ppd
   );

BOOLEAN
Read (
   PHID_DEVICE    HidDevice
   )
/*++
RoutineDescription:
   Given a struct _HID_DEVICE, obtain a read report and unpack the values
   into to InputData array.
--*/
{
   DWORD    bytesRead;

   if (!ReadFile (HidDevice->HidDevice,
                  HidDevice->InputReportBuffer,
                  HidDevice->Caps.InputReportByteLength,
                  &bytesRead,
                  NULL)) { // No overlapped structure.  HidClass buffers for us.
      return FALSE;
   }

   ASSERT (bytesRead == hidDevice->Caps.InputReportByteLength);

   return UnpackReport (HidDevice->InputReportBuffer,
                        HidDevice->Caps.InputReportByteLength,
                        HidP_Input,
                        HidDevice->InputData,
                        HidDevice->InputDataLength,
                        HidDevice->Ppd);
}

BOOLEAN
Write (
   PHID_DEVICE    HidDevice
   )
/*++
RoutineDescription:
   Given a struct _HID_DEVICE, take the information in the HID_DATA array
   pack it into a write report and send it to the hid device.
--*/
{
   DWORD    bytesWritten;


   PackReport (HidDevice->OutputReportBuffer,
               HidDevice->Caps.OutputReportByteLength,
               HidP_Output,
               HidDevice->OutputData,
               HidDevice->OutputDataLength,
               HidDevice->Ppd);

   return (WriteFile (HidDevice->HidDevice,
                      HidDevice->OutputReportBuffer,
                      HidDevice->Caps.OutputReportByteLength,
                      &bytesWritten,
                      NULL)  &&
           (bytesWritten == HidDevice->Caps.OutputReportByteLength));


}

BOOLEAN
SetFeature (
   PHID_DEVICE    HidDevice
   )
/*++
RoutineDescription:
   Given a struct _HID_DEVICE, take the information in the HID_DATA array
   pack it into a write report and send it to the hid device.
--*/
{

   PackReport (HidDevice->FeatureReportBuffer,
               HidDevice->Caps.FeatureReportByteLength,
               HidP_Feature,
               HidDevice->FeatureData,
               HidDevice->FeatureDataLength,
               HidDevice->Ppd);

   return (HidD_SetFeature (HidDevice->HidDevice,
                            HidDevice->FeatureReportBuffer,
                            HidDevice->Caps.FeatureReportByteLength));

}

BOOLEAN
GetFeature (
   PHID_DEVICE    HidDevice
   )
/*++
RoutineDescription:
   Given a struct _HID_DEVICE, obtain a read report and unpack the values
   into to InputData array.
--*/
{
   DWORD    bytesRead;

   if (!HidD_GetFeature (HidDevice->HidDevice,
                         HidDevice->InputReportBuffer,
                         HidDevice->Caps.InputReportByteLength)) {
      return FALSE;
   }

   return UnpackReport (HidDevice->FeatureReportBuffer,
                        HidDevice->Caps.FeatureReportByteLength,
                        HidP_Feature,
                        HidDevice->FeatureData,
                        HidDevice->FeatureDataLength,
                        HidDevice->Ppd);
}


BOOLEAN
UnpackReport (
   IN       PCHAR                ReportBuffer,
   IN       USHORT               ReportBufferLength,
   IN       HIDP_REPORT_TYPE     ReportType,
   IN OUT   PHID_DATA            Data,
   IN       ULONG                DataLength,
   IN       PHIDP_PREPARSED_DATA Ppd
   )
/*++
Routine Description:

--*/
{
   ULONG       numUsages; // Number of usages returned from GetUsages.
   ULONG       i;

   for (i = 0; i < DataLength; i++, Data++) {
      if (Data->IsButtonData) {
         numUsages = Data->ButtonData.MaxUsageLength;
         Data->Status = HidP_GetUsages (
                           ReportType,
                           Data->UsagePage,
                           0, // All collections
                           Data->ButtonData.Usages,
                           &numUsages,
                           Ppd,
                           ReportBuffer,
                           ReportBufferLength);
         //
         // Get usages writes the list of usages into the buffer
         // Data->ButtonData.Usages newUsage is set to the number of usages
         // written into this array.
         // We assume that there will not be a usage of zero.
         // (None have been defined to date.)
         // So lets assume that a zero indicates an end of the list of usages.
         //

         if (numUsages < Data->ButtonData.MaxUsageLength) {
            Data->ButtonData.Usages[numUsages] = 0;
         }

      } else {
         Data->Status = HidP_GetUsageValue (
                              ReportType,
                              Data->UsagePage,
                              0, // All Collections.
                              Data->ValueData.Usage,
                              &Data->ValueData.Value,
                              Ppd,
                              ReportBuffer,
                              ReportBufferLength);

         Data->Status = HidP_GetScaledUsageValue (
                              ReportType,
                              Data->UsagePage,
                              0, // All Collections.
                              Data->ValueData.Usage,
                              &Data->ValueData.ScaledValue,
                              Ppd,
                              ReportBuffer,
                              ReportBufferLength);
      }
   }
   return (HIDP_STATUS_SUCCESS == Data->Status);
}


BOOLEAN
PackReport (
   OUT      PCHAR                ReportBuffer,
   IN       USHORT               ReportBufferLength,
   IN       HIDP_REPORT_TYPE     ReportType,
   IN       PHID_DATA            Data,
   IN       ULONG                DataLength,
   IN       PHIDP_PREPARSED_DATA Ppd
   )
/*++
Routine Description:
   Given a read IRP.

   See if there are any scan codes alread in the list of scan codes.
   If there are try and service this read request from that list.

   If there are none, then send a read down to HidClass to get some more.
--*/
{
   ULONG       numUsages; // Number of usages returned from GetUsages.
   ULONG       i;

   memset (ReportBuffer, (UCHAR) 0, ReportBufferLength);

   for (i = 0; i < DataLength; i++, Data++) {
      if (Data->IsButtonData) {
         numUsages = Data->ButtonData.MaxUsageLength;
         Data->Status = HidP_SetUsages (
                           ReportType,
                           Data->UsagePage,
                           0, // All collections
                           Data->ButtonData.Usages,
                           &numUsages,
                           Ppd,
                           ReportBuffer,
                           ReportBufferLength);
      } else {
         Data->Status = HidP_SetUsageValue (
                              ReportType,
                              Data->UsagePage,
                              0, // All Collections.
                              Data->ValueData.Usage,
                              Data->ValueData.Value,
                              Ppd,
                              ReportBuffer,
                              ReportBufferLength);
      }

          if (HIDP_STATUS_SUCCESS != Data->Status)
          {
                  return FALSE;
          }

   }
   return (HIDP_STATUS_SUCCESS == Data->Status);
}



