/*++

Copyright (c) 1996  Microsoft Corporation

Module Name:

    hidgame.c

Abstract: Human Input Device (HID) Gameport driver

Author:

    Eliyas Yakub

Environment:

    Kernel mode

Revision History:


--*/



#include "hidgame.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text( INIT, DriverEntry )
#pragma alloc_text( PAGE, HGAME_CreateClose)
#pragma alloc_text( PAGE, HGAME_AddDevice)
#endif // ALLOC_PRAGMA

HIDGAME_GLOBAL Global;

NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING registryPath
    )
/*++

Routine Description:

    Installable driver initialization entry point.
    This entry point is called directly by the I/O system.

Arguments:

    DriverObject - pointer to the driver object

    registryPath - pointer to a unicode string representing the path,
                   to driver-specific key in the registry.

Return Value:

    STATUS_SUCCESS if successful,
    STATUS_UNSUCCESSFUL otherwise

--*/
{
    NTSTATUS                        ntStatus = STATUS_SUCCESS;
    HID_MINIDRIVER_REGISTRATION        HidMinidriverRegistration;


    HGAME_KdPrint(("DriverEntry Enter \n"));

    DriverObject->MajorFunction[IRP_MJ_CREATE]=
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = HGAME_CreateClose;
    DriverObject->MajorFunction[IRP_MJ_INTERNAL_DEVICE_CONTROL] =
                                                    HGAME_Internal_Ioctl;
    DriverObject->MajorFunction[IRP_MJ_PNP] = HGAME_PnP;
    DriverObject->MajorFunction[IRP_MJ_POWER] = HGAME_Power;
    DriverObject->DriverUnload = HGAME_Unload;
    DriverObject->DriverExtension->AddDevice = HGAME_AddDevice;

    //
    // Register  with HID.SYS module
    //

    HidMinidriverRegistration.Revision = HID_REVISION;
    HidMinidriverRegistration.DriverObject = DriverObject;
    HidMinidriverRegistration.RegistryPath = registryPath;
    HidMinidriverRegistration.DeviceExtensionSize = sizeof(DEVICE_EXTENSION);
    HidMinidriverRegistration.DevicesArePolled = TRUE;

    HGAME_KdPrint(("DeviceExtensionSize = %x\n",
                            HidMinidriverRegistration.DeviceExtensionSize));

    HGAME_KdPrint(("Registering with HID.SYS\n"));

    ntStatus = HidRegisterMinidriver(&HidMinidriverRegistration);
    
    // 
    // Initialize the list head
    //
    InitializeListHead(&Global.DeviceListHead);

    HGAME_KdPrint(("DriverEntry Exit = %x\n", ntStatus));

    return ntStatus;
}



NTSTATUS
HGAME_CreateClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*++

Routine Description:

   Process the Create and close IRPs sent to this device.

Arguments:

   DeviceObject - pointer to a device object.

   Irp - pointer to an I/O Request Packet.

Return Value:

      NT Status code

--*/
{
    PIO_STACK_LOCATION   IrpStack;
    NTSTATUS             ntStatus = STATUS_SUCCESS;

    PAGED_CODE ();
    HGAME_KdPrint(("HGAME_CreateClose Enter\n"));
    //
    // Get a pointer to the current location in the Irp.
    //
    IrpStack = IoGetCurrentIrpStackLocation(Irp);

    switch(IrpStack->MajorFunction)
    {
        case IRP_MJ_CREATE:
            HGAME_KdPrint(("IRP_MJ_CREATE\n"));
            Irp->IoStatus.Information = 0;
            break;

        case IRP_MJ_CLOSE:
            HGAME_KdPrint(("IRP_MJ_CLOSE\n"));
            Irp->IoStatus.Information = 0;
            break;

        default:
            HGAME_KdPrint((" Invalid CreateClose Parameter\n"));
            ntStatus = STATUS_INVALID_PARAMETER;
            break;
    }

    //
    // Save Status for return and complete Irp
    //

    Irp->IoStatus.Status = ntStatus;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    HGAME_KdPrint((" HGAME_CreateClose Exit = %x\n", ntStatus));

    return ntStatus;
}


NTSTATUS
HGAME_AddDevice(
    IN PDRIVER_OBJECT DriverObject,
    IN PDEVICE_OBJECT FunctionalDeviceObject
    )
/*++

Routine Description:

    Process the IRPs sent to this device.

Arguments:

    DriverObject - pointer to a driver object.

    FunctionalDeviceObject - pointer to a functional deviceobject  created by hidclass

Return Value:

    NT Status code.

--*/
{
    NTSTATUS                ntStatus = STATUS_SUCCESS;
    PDEVICE_OBJECT          DeviceObject = NULL;
    PDEVICE_EXTENSION       pDevExt;
    ULONG                   totalExtensionSize;
    PHID_DEVICE_EXTENSION   HidDeviceExtension;

    PAGED_CODE ();

    HGAME_KdPrint(("HGAME_AddDevice Entry\n"));

    DeviceObject = FunctionalDeviceObject;
    //
    // Initialize the device extension.
    //
    pDevExt = GET_MINIDRIVER_DEVICE_EXTENSION (DeviceObject);

    pDevExt->Started =
    pDevExt->Removed = FALSE;
    pDevExt->Self = DeviceObject;
    pDevExt->TopOfStack = NULL;
    pDevExt->OutstandingIO = 1; // biassed to 1.  Transition to zero during
                                        // remove device means IO is finished.
    DeviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

    KeInitializeEvent(&pDevExt->RemoveEvent, SynchronizationEvent, FALSE);

    //
    // Attach our functional driver to the device stack.
    // the return value of IoAttachDeviceToDeviceStack is the top of the
    // attachment chain.  This is where all the IRPs should be routed.
    //
    pDevExt->TopOfStack =  GET_NEXT_DEVICE_OBJECT(DeviceObject);
    //
    // if this attachment fails then top of stack will be null.
    // failure for attachment is an indication of a broken plug play system.
    //
    ASSERT (NULL != pDevExt->TopOfStack);

    //Initialize the CustomTimerDpc object
    KeInitializeDpc( &pDevExt->PollingDpc,
                        HGAME_PollingTimerDpc,
                        (PVOID)DeviceObject );

    // Initialize the timer object
    KeInitializeTimer( &pDevExt->PollingTimer );

    // Initialize the pending Irp devicequeue
    KeInitializeDeviceQueue( &pDevExt->PendingIrpQueue );

    // 10 is multiplied because system time is specified in 100ns units
    pDevExt->PollingInterval = RtlConvertLongToLargeInteger(
                                    HGAME_RETRY_INTERVAL * -10);
    

    // Initialize the list
    InitializeListHead(&pDevExt->Link);

    // Add this device to the linked list of devices 
    InsertTailList(&Global.DeviceListHead, &pDevExt->Link);
    
    HGAME_KdPrint(("HGAME_AddDevice Exit = %x\n", ntStatus));

    return ntStatus;
}



VOID
HGAME_Unload(
    IN PDRIVER_OBJECT DriverObject
    )
/*++

Routine Description:

    Free all the allocated resources, etc.

Arguments:

    DriverObject - pointer to a driver object.

Return Value:

    VOID
--*/
{
    PAGED_CODE();
    HGAME_KdPrint(("HGAME_Unload Enter\n"));
    //
    // All the device objects should be gone
    //
    ASSERT ( NULL == DriverObject->DeviceObject);

    HGAME_KdPrint(("HGAME_Unload Exit\n"));
    return;
}

