/*++

Copyright (c) 1997  Microsoft Corporation

Module Name:

    ioctl.c

Abstract: contains routies to support HIDCLASS internal ioctl queries for game
          devices


Author:
                Eliyas Yakub (Mar, 10, 1997)

Environment:

    Kernel mode

Revision History:


--*/

#include "hidgame.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text (PAGE, HGAME_GetDeviceDescriptor)
#pragma alloc_text (PAGE, HGAME_GetReportDescriptor)
#endif

extern HIDGAME_GLOBAL Global;

NTSTATUS
HGAME_Internal_Ioctl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP pIrp
    )
/*++

Routine Description:

   Process the Control IRPs sent to this device.

Arguments:

   DeviceObject - pointer to a device object.

   pIrp - pointer to an I/O Request Packet.

Return Value:

      NT status code

--*/
{
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    PDEVICE_EXTENSION   pDevExt;
    PIO_STACK_LOCATION  pIrpStack;


    HGAME_KdPrint(("HGAME_Internal_Ioctl Enter\n"));

    //
    // Get a pointer to the current location in the pIrp
    //

    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

    //
    // Get a pointer to the device extension
    //

    pDevExt = GET_MINIDRIVER_DEVICE_EXTENSION (DeviceObject);


    switch(pIrpStack->Parameters.DeviceIoControl.IoControlCode) {

    case IOCTL_HID_GET_DEVICE_DESCRIPTOR:
        HGAME_KdPrint(("IOCTL_HID_GET_DEVICE_DESCRIPTOR\n"));
        ntStatus = HGAME_GetDeviceDescriptor (DeviceObject, pIrp);
        break;

    case IOCTL_HID_GET_REPORT_DESCRIPTOR:
        HGAME_KdPrint(("IOCTL_HID_GET_REPORT_DESCRIPTOR\n"));
        ntStatus = HGAME_GetReportDescriptor (DeviceObject, pIrp);
        break;

    case IOCTL_HID_READ_REPORT:
        HGAME_KdPrint(("IOCTL_HID_READ_REPORT\n"));
        ntStatus = HGAME_ReadReport (DeviceObject, pIrp);
        break;

    case IOCTL_HID_ACTIVATE_DEVICE:
    case IOCTL_HID_DEACTIVATE_DEVICE:
        /*
        * We don't do anything for these IOCTLs but some minidrivers might.
        */
        HGAME_KdPrint(("IOCTL_HID_(DE)ACTIVATE_DEVICE\n"));
        ntStatus = STATUS_SUCCESS;
        break;

    case IOCTL_HID_GET_STRING:
        HGAME_KdPrint(("No support for Get String\n"));
        ntStatus = STATUS_NOT_SUPPORTED;
        break;

    case IOCTL_HID_WRITE_REPORT:
        HGAME_KdPrint(("No support for Write Report\n"));
        ntStatus = STATUS_NOT_SUPPORTED;
        break;

    case IOCTL_HID_GET_FEATURE:
    case IOCTL_HID_SET_FEATURE:
        HGAME_KdPrint(("No support for Get Feature\n"));
        ntStatus = STATUS_NOT_SUPPORTED;
        break;

    case IOCTL_GET_PHYSICAL_DESCRIPTOR:
        HGAME_KdPrint(("No support for Get Physical Descriptor\n"));
        ntStatus = STATUS_NOT_SUPPORTED;
        break;

    case IOCTL_HID_GET_DEVICE_ATTRIBUTES:
        HGAME_KdPrint(("IOCTL_HID_GET_DEVICE_ATTRIBUTES\n"));
        ntStatus = HGAME_GetAttributes (DeviceObject, pIrp);
        break;

    default:
        HGAME_KdPrint(("Unknown or unsupported IOCTL (%x)\n",
                       pIrpStack->Parameters.DeviceIoControl.IoControlCode));
        ntStatus = STATUS_NOT_SUPPORTED;
        break;

    }
    //
    // Set real return status in pIrp
    //
    pIrp->IoStatus.Status = ntStatus;
    //HGAME_KdPrint(("HGAME_Internal_Ioctl IoStatus = %x\n", ntStatus));
    //
    // Complete Irp
    //
    if (STATUS_CANCELLED == ntStatus) {
        return ntStatus;

    } else if (ntStatus != STATUS_PENDING) {
        IoCompleteRequest(pIrp, IO_NO_INCREMENT);
        //
        // NOTE: Real return status set in pIrp->IoStatus.Status
        //
        ntStatus = STATUS_SUCCESS;

    } else {
        IoMarkIrpPending( pIrp );
    }

    HGAME_KdPrint(("HGAME_Internal_Ioctl Exit = %x\n", ntStatus));

    return ntStatus;
}

NTSTATUS
HGAME_GetDeviceDescriptor(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP pIrp
)
 /*++
 *   Routine Description:
 *
 *       Free all the allocated resources, etc.
 *
 *   Arguments:
 *
 *       DeviceObject - pointer to a device object.
 *               pIrp             - pointer to current Irp
 *
 *   Return Value:
 *
 *       NT status code.
 *
--*/
{
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    PDEVICE_EXTENSION   pDevExt;
    PIO_STACK_LOCATION  pIrpStack;
    ULONG               bytesToCopy;

    PAGED_CODE ();

    HGAME_KdPrint(("HGAME_GetDeviceDescriptor Enter\n"));

    //
    // Get a pointer to the current location in the Irp
    //

    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

    //
    // Get a pointer to the device extension
    //

    pDevExt = GET_MINIDRIVER_DEVICE_EXTENSION (DeviceObject);

    //
    // Copy device descriptor to HIDCLASS buffer
    //

    HGAME_KdPrint(("HIDCLASS Buffer = 0x%x, Buffer length = 0x%x\n",
                    pIrp->UserBuffer,
                    pIrpStack->Parameters.DeviceIoControl.OutputBufferLength));

    //
    // Copy MIN (OutputBufferLength, pDevExt->HidDescriptor->bLength)
    //

    bytesToCopy = pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;

    if (bytesToCopy >= pDevExt->HidDescriptor.bLength) {
        bytesToCopy = pDevExt->HidDescriptor.bLength;
    } else {
        ntStatus = STATUS_BUFFER_TOO_SMALL;
    }
    HGAME_KdPrint(("Copying %d bytes to HIDCLASS buffer\n", bytesToCopy));

    RtlCopyMemory((PUCHAR) pIrp->UserBuffer, (PUCHAR) &pDevExt->HidDescriptor, bytesToCopy);

    //
    // Report how many bytes were copied
    //
    pIrp->IoStatus.Information = bytesToCopy;



    HGAME_KdPrint(("HGAME_GetDeviceDescriptor Exit = 0x%x\n", ntStatus));

    return ntStatus;
}


NTSTATUS
HGAME_GetReportDescriptor(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP pIrp
)
/*++

Routine Description:


Arguments:

    DeviceObject - pointer to a device object.

Return Value:

    NT status code.

--*/
{
    PDEVICE_EXTENSION     pDevExt;
    PIO_STACK_LOCATION    pIrpStack;
    NTSTATUS              ntStatus = STATUS_SUCCESS;
    PVOID                 Report = NULL;
    ULONG                 ReportLength;
    ULONG                 bytesToCopy;
    UCHAR                 Game_1Stick_Report[] = GAME_1STICK_REPORT;
    UCHAR                 Game_HalfStick_Report[] = GAME_HALF_STICK_REPORT;

    PAGED_CODE ();

    HGAME_KdPrint(("HGAME_GetReportDescriptor Enter\n"));

    //
    // Get a pointer to the current location in the Irp
    //

    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

    //
    // Get a pointer to the device extension
    //

    pDevExt = GET_MINIDRIVER_DEVICE_EXTENSION (DeviceObject);


    //
    // Get report descriptor
    //

    switch (pDevExt->PortInfo.Portion) {
    case GameenumFirstHalf:
    case GameenumSecondHalf:
        ReportLength = pDevExt->HidDescriptor.DescriptorList[0].wReportLength;
        Report = &Game_HalfStick_Report;
        break;

    case GameenumWhole:
        ReportLength = pDevExt->HidDescriptor.DescriptorList[0].wReportLength;
        Report = &Game_1Stick_Report;
    }


    //
    // Copy report descriptor to HIDCLASS buffer
    //

    HGAME_KdPrint(("HIDCLASS Buffer = 0x%x, Buffer length = 0x%x\n",
        pIrp->UserBuffer,
        pIrpStack->Parameters.DeviceIoControl.OutputBufferLength));

    //
    // Copy MIN (OutputBufferLength, pHidDescriptor->bLength)
    //

    bytesToCopy =
        pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;

    if (bytesToCopy >= ReportLength) {
        bytesToCopy = ReportLength;
    }
    else {
        ntStatus = STATUS_BUFFER_TOO_SMALL;
    }

    HGAME_KdPrint(("Copying %d bytes to HIDCLASS buffer\n", bytesToCopy));

    RtlCopyMemory((PUCHAR) pIrp->UserBuffer, (PUCHAR) Report, bytesToCopy);

    //
    // Report how many bytes were copied
    //
    pIrp->IoStatus.Information = bytesToCopy;

    HGAME_KdPrint(("HGAME_GetReportDescriptor Exit = 0x%x\n", ntStatus));

    return ntStatus;
}



NTSTATUS
HGAME_ReadReport(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP pIrp
 )
 /*++
 *   Routine Description:
 *
 *       Free all the allocated resources, etc.
 *
 *   Arguments:
 *
 *       DeviceObject - pointer to a device object.
 *               pIrp             - pointer to current Irp
 *
 *   Return Value:
 *
 *       NT status code.
 *
--*/
{
    NTSTATUS            ntStatus = STATUS_PENDING;
    PDEVICE_EXTENSION   pDevExt;
    PIO_STACK_LOCATION  pIrpStack;
    KIRQL               OldIrql;


    HGAME_KdPrint(("HGAME_ReadReport Enter\n"));
    //TRAP();
    //
    // Get a pointer to the device extension.
    //
    pDevExt = GET_MINIDRIVER_DEVICE_EXTENSION (DeviceObject);

    //
    // Get Stack location
    //
    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
    //
    // Find out which device we are and read, but first make sure
    // there is enough room
    //
    if (pIrpStack->Parameters.Read.Length < sizeof(HIDGAME_INPUT_DATA))
    {
            ntStatus = STATUS_BUFFER_TOO_SMALL;
            pIrp->IoStatus.Status = ntStatus;
            pIrp->IoStatus.Information  = 0;
            return ntStatus;
    }

    KeRaiseIrql( DISPATCH_LEVEL, &OldIrql);
    //goto TEST;
    if ( !KeInsertDeviceQueue(&pDevExt->PendingIrpQueue,
                            &pIrp->Tail.Overlay.DeviceQueueEntry) )
    {
        KIRQL OldIrql;
        // Queue is empty. Lets process IRP right away.

        // don't forget to flip the queue state to NOT-BUSY
        KeRemoveDeviceQueue(&pDevExt->PendingIrpQueue);

        // Lock access to the port
        KeAcquireSpinLockAtDpcLevel(&pDevExt->SpinLock);
        //
        // Raise IRQL to prevent premption
        // This may seriously affect audio, graphics, and
        // network performance. Use this extremely high IRQL
        // only if DISPATCH_LEVEL leads to unacceptable joystick
        // performance. If you use this, be sure to uncomment the
        // KeLowerIrql below, also.
        //
        // KeRaiseIrql(PROFILE_LEVEL, &OldIrql);


        ntStatus = HGAME_AnalogPoll(DeviceObject, pIrp);

        // Lower from PROFILE_LEVEL if set above
        // KeLowerIrql(OldIrql);

        // release the lock
        KeReleaseSpinLockFromDpcLevel(&pDevExt->SpinLock);

        if(ntStatus == STATUS_PENDING)
        {
            // Polling too soon.
            // Force Insertion into the queue because
            // the Irp got to wait and set the timer
            HGAME_KdPrint(("PollingTimerDpc polls too soon:%d\n",
                    pDevExt->JoyStatistics.Polls));
          if ( ! KeInsertDeviceQueue(&pDevExt->PendingIrpQueue,
                                    &pIrp->Tail.Overlay.DeviceQueueEntry) ){
                KeInsertDeviceQueue(&pDevExt->PendingIrpQueue,
                                            &pIrp->Tail.Overlay.DeviceQueueEntry);
            }
        }
        else {
            goto END;
        }

    }

    //
    // IRP is queued, so set the cancel routine and start the timer
    //
    if(IoSetCancelRoutine (pIrp, HGAME_CancelIrp) != NULL)
        ASSERT (FALSE);
    if (pIrp->Cancel)
    {
        if (HGAME_CancelIrp == IoSetCancelRoutine (pIrp, NULL)) {
            //
            // In this case the Irp was cancelled before we set the cancel
            // routine.  We therefore need to dequeue the Irp, and return
            // the irp as status cancelled.
            //

            KeRemoveEntryDeviceQueue(
                &pDevExt->PendingIrpQueue,
                &pIrp->Tail.Overlay.DeviceQueueEntry );

            pIrp->IoStatus.Status = STATUS_CANCELLED;
            IoCompleteRequest (pIrp, IO_NO_INCREMENT);
            ntStatus = STATUS_CANCELLED;
            goto END;

        } else {
            //
            // In this case the Irp was cancelled after we set the cancel.
            // The cancel routine fired so Do not complete
            // the IRP, Do not queue, return STATUS_CANCELLED.
            //
            ntStatus = STATUS_CANCELLED;
            goto END;
        }
    }
    KeSetTimer(&pDevExt->PollingTimer, pDevExt->PollingInterval,
        &pDevExt->PollingDpc);
END:
    KeLowerIrql(OldIrql);
    return ntStatus;
}


VOID
HGAME_PollingTimerDpc(
    IN PKDPC Dpc,
    IN PVOID Context,
    IN PVOID SystemArgument1,
    IN PVOID SystemArgument2
)
 /*++
 *   Routine Description:
 *
 *       CustomTimerDpc routine. Serives Irp that have
 *       in the PendingIrpQueue
 *
 *   Arguments:
 *
 *       DeviceObject    - pointer to DPC object
 *       Context         - pointer to device object
 *       SystemArgument1 -  undefined
 *       SystemArgument2 -  undefined
 *
 *   Return Value:
 *
 *
 *
--*/
{
    PDEVICE_OBJECT          DeviceObject;
    PDEVICE_EXTENSION       pDevExt;
    PKDEVICE_QUEUE_ENTRY    QueueEntry;
    KIRQL                   OldIrql;
    NTSTATUS                ntStatus;
    PIRP                    pIrp;


    HGAME_KdPrint(("Called HGAME_PollingTimerDpc\n"));
    //TRAP();

    DeviceObject = (PDEVICE_OBJECT)Context;
    //
    // Get a pointer to the device extension.
    //
    pDevExt = GET_MINIDRIVER_DEVICE_EXTENSION (DeviceObject);
    QueueEntry = KeRemoveDeviceQueue(&pDevExt->PendingIrpQueue);
    if ( QueueEntry != NULL )
    {
        pIrp = CONTAINING_RECORD(QueueEntry, IRP,
                                    Tail.Overlay.DeviceQueueEntry);
        if (NULL == IoSetCancelRoutine (pIrp, NULL)) {
                // Irp has already been cancelled.
                // start the timer to process next packet
                goto END;
         }

        // Lock access to the port
        KeAcquireSpinLockAtDpcLevel(&pDevExt->SpinLock);

        //
        // Now it is too late to cancel the IRP.  We will complete it
        // with valid data.
        // Raise IRQL to prevent premption
        //
        KeRaiseIrql(PROFILE_LEVEL, &OldIrql);

        // poll the port
        ntStatus = HGAME_AnalogPoll(DeviceObject, pIrp);

        // Come back to the old level
        KeLowerIrql(OldIrql);

        // release the lock
        KeReleaseSpinLockFromDpcLevel(pDevExt->pSpinLock);

        if(ntStatus != STATUS_PENDING) {
            pIrp->IoStatus.Status = ntStatus;
            IoCompleteRequest( pIrp, IO_NO_INCREMENT );
            // start the timer; goto END; falls through
        }
        else
        {
            // Polling too soon.
            // Force Insertion into the queue because
            // the Irp got to wait and set the timer
            HGAME_KdPrint(("PollingTimerDpc polls too soon:%d\n",
                    pDevExt->JoyStatistics.Polls));

            if ( ! KeInsertDeviceQueue(&pDevExt->PendingIrpQueue,
                                        &pIrp->Tail.Overlay.DeviceQueueEntry) ){
                if(!KeInsertDeviceQueue(&pDevExt->PendingIrpQueue,
                                            &pIrp->Tail.Overlay.DeviceQueueEntry))
                    ASSERT(FALSE);
            }
            //
            // IRP is queued, so set the cancel routine and start the timer
            //
            if (NULL != IoSetCancelRoutine (pIrp, HGAME_CancelIrp))
                ASSERT(FALSE);
            if (pIrp->Cancel)
            {
                if (HGAME_CancelIrp == IoSetCancelRoutine (pIrp, NULL)) {
                    //
                    // In this case the Irp was cancelled before we set the cancel
                    // routine.  We therefore need to dequeue the Irp, and return
                    // the irp as status cancelled.
                    //

                    KeRemoveEntryDeviceQueue(
                        &pDevExt->PendingIrpQueue,
                        &pIrp->Tail.Overlay.DeviceQueueEntry );

                    pIrp->IoStatus.Status = STATUS_CANCELLED;
                    IoCompleteRequest (pIrp, IO_NO_INCREMENT);
                    // start the timer; goto END;                  ;

                } else {
                    //
                    // In this case the Irp was cancelled after we set the cancel.
                    // The cancel routine fired so Do not complete
                    // the IRP, Do not queue, return STATUS_CANCELLED.
                    //
                    // start the timer; goto END;;
                }
            }
        }
 END:
        KeSetTimer(&pDevExt->PollingTimer,
            pDevExt->PollingInterval,
            &pDevExt->PollingDpc);
    }
    return;

}


VOID
HGAME_CancelIrp(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP pIrp
)
 /*++
 *   Routine Description:
 *
 *       Completes the cancelled Irp
 *
 *   Arguments:
 *
 *       DeviceObject - pointer to a device object.
 *       pIrp         - pointer to a Irp
 *
 *   Return Value:
 *
 *
 *
--*/
{
    PDEVICE_EXTENSION   pDevExt;

    HGAME_KdPrint(("HGAME_CancelIrp enter\n"));

    pDevExt = GET_MINIDRIVER_DEVICE_EXTENSION (DeviceObject);
    KeRemoveEntryDeviceQueue(
            &pDevExt->PendingIrpQueue,
            &pIrp->Tail.Overlay.DeviceQueueEntry );
    IoReleaseCancelSpinLock( pIrp->CancelIrql );
    pIrp->IoStatus.Status = STATUS_CANCELLED;
    pIrp->IoStatus.Information = 0;
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);

    HGAME_KdPrint(("HGAME_CancelIrp exit\n"));

}

NTSTATUS
HGAME_GetAttributes (
    PDEVICE_OBJECT  DeviceObject,
    PIRP            Irp
    )
/*++
 *   Routine Description:
 *
 *       Fill in the given struct _HID_DEVICE_ATTRIBUTES
 *
 *   Arguments:
 *
 *       DeviceObject - pointer to a device object.
 *       pIrp         - pointer to a Irp
 *
 *   Return Value:
 *
 *
 *
--*/
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PDEVICE_EXTENSION   deviceExtension;
    PIO_STACK_LOCATION  irpStack;
    PHID_DEVICE_ATTRIBUTES deviceAttributes;

    HGAME_KdPrint(("'HIDUSB.SYS: HumGetDeviceAttributes Entry\n"));

    //
    // Get a pointer to the current location in the Irp
    //

    irpStack = IoGetCurrentIrpStackLocation(Irp);

    //
    // Get a pointer to the device extension
    //

    deviceExtension = GET_MINIDRIVER_DEVICE_EXTENSION(DeviceObject);
    deviceAttributes = (PHID_DEVICE_ATTRIBUTES) Irp->UserBuffer;

    ASSERT (sizeof (HID_DEVICE_ATTRIBUTES) ==
            irpStack->Parameters.DeviceIoControl.OutputBufferLength);

    //
    // Report how many bytes were copied
    //
    Irp->IoStatus.Information = sizeof (HID_DEVICE_ATTRIBUTES);

    deviceAttributes->Size = sizeof (HID_DEVICE_ATTRIBUTES);
    deviceAttributes->VendorID = HIDGAME_VENDOR_ID;
    deviceAttributes->ProductID = HIDGAME_PRODUCT_ID;
    deviceAttributes->VersionNumber = HIDGAME_VERSION_NUMBER;

    HGAME_KdPrint(("'HIDUSB.SYS: HumGetAttributes Exit = 0x%x\n", ntStatus));

    return ntStatus;
}
