/*++

Copyright (c) 1997  Microsoft Corporation

Module Name:

    pnp.c

Abstract: This module contains PnP Start, Stop, Remove,
                and Power dispatch routines and IRP cancel routine

  Author:

    Eliyas Yakub (Mar, 11, 1997)

Environment:

    Kernel mode

Revision History:


--*/
#include <hidgame.h>

#ifdef ALLOC_PRAGMA
#pragma alloc_text (PAGE, HGAME_Power)
#pragma alloc_text (PAGE, HGAME_PnP)
#pragma alloc_text (PAGE, HGAME_StopDevice)
#pragma alloc_text (PAGE, HGAME_GetResources)
#endif

extern HIDGAME_GLOBAL Global;

#define ARE_WE_RELATED(x, y) \
        (x->PortInfo.GamePortAddress == y->PortInfo.GamePortAddress && \
        x->PortInfo.WriteAccessor == y->PortInfo.WriteAccessor && \
        x->PortInfo.ReadAccessor == y->PortInfo.ReadAccessor) \

NTSTATUS
HGAME_PnP (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP pIrp
    )
/*++

Routine Description:

    The plug and play dispatch routines.

Arguments:

   DeviceObject - pointer to a device object.

   pIrp - pointer to an I/O Request Packet.

Return Value:

      NT Status code

--*/
{
    PDEVICE_EXTENSION       pDevExt;
    PIO_STACK_LOCATION      pIrpStack;
    NTSTATUS                ntStatus;

    PAGED_CODE ();

    HGAME_KdPrint(("HGAME_PnP Enter\n"));

    pDevExt = GET_MINIDRIVER_DEVICE_EXTENSION (DeviceObject);
    pIrpStack = IoGetCurrentIrpStackLocation (pIrp);


    InterlockedIncrement (&pDevExt->OutstandingIO);
    if (pDevExt->Removed) {

        //
        // Someone sent us another plug and play IRP after the remove IRP.
        // This should never happen.
        //
        HGAME_KdPrint(("Sporadic PnP irp after the device has been removed\n"));
        ASSERT (FALSE);

        if (0 == InterlockedDecrement (&pDevExt->OutstandingIO)) {
            KeSetEvent (&pDevExt->RemoveEvent, 0, FALSE);
        }
        pIrp->IoStatus.Information = 0;
        pIrp->IoStatus.Status = STATUS_DELETE_PENDING;
        IoCompleteRequest (pIrp, IO_NO_INCREMENT);
        return STATUS_DELETE_PENDING;
    }

    switch (pIrpStack->MinorFunction) {
    case IRP_MN_START_DEVICE:

        //
        // The device is starting.
        //
        // We cannot touch the device (send it any non pnp irps) until a
        // start device has been passed down to the lower drivers.
        //
        IoCopyCurrentIrpStackLocationToNext (pIrp);
        KeInitializeEvent(&pDevExt->StartEvent, NotificationEvent, FALSE);
        IoSetCompletionRoutine (pIrp,
                                HGAME_PnPComplete,
                                pDevExt,
                                TRUE,
                                TRUE,
                                TRUE);
        ntStatus = IoCallDriver (pDevExt->TopOfStack, pIrp);
        if (STATUS_PENDING == ntStatus) {
            KeWaitForSingleObject(
               &pDevExt->StartEvent,
               Executive, // Waiting for reason of a driver
               KernelMode, // Waiting in kernel mode
               FALSE, // No allert
               NULL); // No timeout
        }
        if (NT_SUCCESS (ntStatus)) {
            //
            // As we are now back from our start device we can do work.
            //
            ntStatus = HGAME_StartDevice ( pDevExt, pIrp );
         }

        pIrp->IoStatus.Information = 0;
        pIrp->IoStatus.Status = ntStatus;
        IoCompleteRequest (pIrp, IO_NO_INCREMENT);
        break;

    case IRP_MN_STOP_DEVICE:
        //
        // After the start IRP has been sent to the lower driver object, the
        // bus may NOT send any more IRPS down ``touch'' until another START
        // has occured.
        // What ever access is required must be done before the Irp is passed
        // on.
        //

        HGAME_StopDevice (pDevExt, TRUE);

        //
        // We don't need a completion routine so fire and forget.
        //
        // Set the current stack location to the next stack location and
        // call the next device object.
        //
        IoSkipCurrentIrpStackLocation (pIrp);
        ntStatus = IoCallDriver (pDevExt->TopOfStack, pIrp);
        break;

    case IRP_MN_REMOVE_DEVICE:
        //
        // The PlugPlay system has dictacted the removal of this device.  We
        // have no choise but to detach and delete the device object.
        // (If we wanted to express and interest in preventing this removal,
        // we should have filtered the query remove and query stop routines.)
        //
        // Note! we might receive a remove WITHOUT first receiving a stop.
        // ASSERT (!pDevExt->Removed);

        //
        // We will no longer receive requests for this device as it has been
        // removed.
        //
        pDevExt->Removed = TRUE;

        if (pDevExt->Started) {
            // Stop the device without touching the hardware.
            HGAME_StopDevice(pDevExt, FALSE);
        }

        //
        // Here if we had any outstanding requests in a personal queue we should
        // complete them all now.
        //
        // Note, the device is guarenteed stopped, so we cannot send it any non-
        // PNP IRPS.
        //

        //
        // Send on the remove IRP
        //

        IoSkipCurrentIrpStackLocation (pIrp);
        ntStatus = IoCallDriver (pDevExt->TopOfStack, pIrp);

        if (0 < InterlockedDecrement (&pDevExt->OutstandingIO)) {
            KeWaitForSingleObject (
                &pDevExt->RemoveEvent,
                Executive,
                KernelMode,
                FALSE,
                NULL);
        }
        //
        // Remove this object from the global list
        //
        RemoveEntryList(&pDevExt->Link);
        return STATUS_SUCCESS;

    case IRP_MN_QUERY_REMOVE_DEVICE:
    case IRP_MN_CANCEL_REMOVE_DEVICE:
    case IRP_MN_QUERY_STOP_DEVICE:
    case IRP_MN_CANCEL_STOP_DEVICE:
    case IRP_MN_QUERY_DEVICE_RELATIONS:
    case IRP_MN_QUERY_INTERFACE:
    case IRP_MN_QUERY_CAPABILITIES:
    case IRP_MN_QUERY_RESOURCES:
    //case IRP_MN_SET_RESOURCE_REQUIREMENTS:
    case IRP_MN_QUERY_RESOURCE_REQUIREMENTS:
    case IRP_MN_READ_CONFIG:
    case IRP_MN_WRITE_CONFIG:
    case IRP_MN_EJECT:
    case IRP_MN_SET_LOCK:
    case IRP_MN_QUERY_ID:
    default:
        //
        // Here the filter driver might modify the behavior of these IRPS
        // Please see PlugPlay documentation for use of these IRPs.
        //
        IoSkipCurrentIrpStackLocation (pIrp);
        ntStatus = IoCallDriver (pDevExt->TopOfStack, pIrp);
        break;
    }


    if (0 == InterlockedDecrement (&pDevExt->OutstandingIO)) {
        KeSetEvent (&pDevExt->RemoveEvent, 0, FALSE);
    }

    HGAME_KdPrint(("HGAME_PnP exit\n"));

    return ntStatus;
}


NTSTATUS
HGAME_PnPComplete (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP pIrp,
    IN PVOID Context
    )
/*++

Routine Description:
    The pnp IRP is in the process of completing.
    signal

Arguments:

   DeviceObject - pointer to a device object.

   pIrp - pointer to an I/O Request Packet.

   Context set to the device object in question.

Return Value:

      NT Status code


--*/
{
    PIO_STACK_LOCATION      pIrpStack;
    PDEVICE_EXTENSION       pDevExt;
    NTSTATUS                ntStatus = STATUS_SUCCESS;

    UNREFERENCED_PARAMETER (DeviceObject);

    HGAME_KdPrint(("HGAME_PnPComplete enter\n"));

    pDevExt = (PDEVICE_EXTENSION) Context;
    pIrpStack = IoGetCurrentIrpStackLocation (pIrp);

    switch (pIrpStack->MajorFunction) {
    case IRP_MJ_PNP:

        switch (pIrpStack->MinorFunction) {
        case IRP_MN_START_DEVICE:

            KeSetEvent (&pDevExt->StartEvent, 0, FALSE);

            //
            // Take the IRP back so that we can continue using it during
            // the IRP_MN_START_DEVICE dispatch routine.
            // NB: we will have to call IoCompleteRequest
            //
            return STATUS_MORE_PROCESSING_REQUIRED;

        default:
            break;
        }
        break;

    case IRP_MJ_POWER:
    default:
        break;
    }
    return ntStatus;
}

NTSTATUS
HGAME_StartDevice (
    IN PDEVICE_EXTENSION         pDevExt,
    IN PIRP                      pIrp
    )
/*++

Routine Description:

      Actual device initialization and startup takes place in this routine
      Plug and play manager produces a resource list after enumerating the
      deive.

Arguments:

     pDevExt: pointer to device extension
     pIrp - pointer to an I/O Request Packet.
Return Value:

     NT Status code


--*/
{
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    LARGE_INTEGER       LargeFrequency;
    UCHAR Game_1Stick_Report[] = GAME_1STICK_REPORT;
    UCHAR Game_HalfStick_Report[] = GAME_HALF_STICK_REPORT;
    PLIST_ENTRY  pEntry;
    PDEVICE_EXTENSION tempExt;

    PAGED_CODE ();

    HGAME_KdPrint(("HGAME_StartDevice Enter\n"));
    ASSERT (!pDevExt->Removed);
    //
    // The PlugPlay system should not have started a removed device!
    //

    if (pDevExt->Started) {
        return STATUS_SUCCESS;
    }


    ntStatus = HGAME_GetResources(pDevExt, pIrp);
    if(!NT_SUCCESS(ntStatus)){
        return ntStatus;
    }
    //
    // Initialize the HID descriptor
    //
    pDevExt->HidDescriptor.bLength = sizeof(HID_DESCRIPTOR);
    pDevExt->HidDescriptor.bDescriptorType = HID_HID_DESCRIPTOR_TYPE;
    pDevExt->HidDescriptor.bcdHID = HID_REVISION;
    pDevExt->HidDescriptor.bCountry = 0; //not localized
    pDevExt->HidDescriptor.bNumDescriptors = HGAME_NUMBER_DESCRIPTORS;
    pDevExt->HidDescriptor.DescriptorList[0].bReportType =
                                            HID_REPORT_DESCRIPTOR_TYPE ;

    // For two joysticks interface two fdos are created to service them, but
    // physicall they both share the same port. So, we have to serialize access
    // to the port. We do this by initializing  the spinlock the first time
    // the interface is added( for the first Fdo of that interface) and second
    // time when the fdo for the same interface comes down, we walk through
    // the list find out his brother fdo and assign the same spinlock data area to
    // this one.

    switch (pDevExt->PortInfo.Portion) {
    case GameenumFirstHalf:
        // This is for the first stick so initialize the spinlock
        KeInitializeSpinLock(&pDevExt->SpinLock);
        pDevExt->HidDescriptor. DescriptorList[0].wReportLength =
                                        sizeof(Game_HalfStick_Report);
        if (2 < pDevExt->PortInfo.NumberAxis) {
            pDevExt->PortInfo.NumberAxis = 2;
            HGAME_KdPrint (("Half Stick with #axis not 2\n"));
            //return STATUS_DEVICE_CONFIGURATION_ERROR;
        }
        break;
    case GameenumSecondHalf:
        // This is for the second stick. Find out his brother
        for(pEntry = Global.DeviceListHead.Flink;
            pEntry != &Global.DeviceListHead;
            pEntry = pEntry->Flink)
        {
            tempExt = CONTAINING_RECORD(pEntry, DEVICE_EXTENSION, Link);
            if(ARE_WE_RELATED(pDevExt, tempExt))
            {
                HGAME_KdPrint(("Brother found\n"));
                break;
            }
        }
        if (pEntry == &Global.DeviceListHead){
            HGAME_KdPrint(("No Brother found\n"));
            ASSERT(FALSE);
        }

        // Assign the same spinlock to this one
        pDevExt->SpinLock = tempExt->SpinLock;
        pDevExt->HidDescriptor. DescriptorList[0].wReportLength =
                                        sizeof(Game_HalfStick_Report);
        if (2 < pDevExt->PortInfo.NumberAxis) {
            pDevExt->PortInfo.NumberAxis = 2;
            HGAME_KdPrint (("Half Stick with #axis not 2\n"));
            //return STATUS_DEVICE_CONFIGURATION_ERROR;
        }
        break;

    case GameenumWhole:
        // Only one of its kind
        KeInitializeSpinLock(&pDevExt->SpinLock);
        pDevExt->HidDescriptor. DescriptorList[0].wReportLength =
                                        sizeof(Game_1Stick_Report);

        if ((pDevExt->PortInfo.NumberAxis < 2) ||
            (pDevExt->PortInfo.NumberAxis > 4)) {

            HGAME_KdPrint (("Full Stick with no axis neither 2, 3, or 4\n"));
            TRAP ();
            return  STATUS_DEVICE_CONFIGURATION_ERROR;
        }
    }

    HGAME_KdPrint(("Device address of 0x%x\n", pDevExt->PortInfo.GamePortAddress));
    HGAME_KdPrint(("Number of axes: %d\n", pDevExt->PortInfo.NumberAxis));

    // Calculate time thresholds for analog device
    // Get the system timer resolution expressed in Hertz.
    KeQueryPerformanceCounter(&LargeFrequency);

    pDevExt->Frequency = LargeFrequency.LowPart;

    HGAME_KdPrint(("Frequency: %u\n", pDevExt->Frequency));

    pDevExt->ThresholdInTicks = (ULONG) (((__int64)pDevExt->Frequency * \
                    (__int64)ANALOG_POLL_RESOLUTION) / (__int64)1000000L);
    HGAME_KdPrint(("ThresholdInTicks: %u\n", pDevExt->ThresholdInTicks));


    pDevExt->MaxTimeoutInTicks = (ULONG) (((__int64)pDevExt->Frequency * \
                    (__int64)ANALOG_POLL_TIMEOUT) / (__int64)1000000L);
    HGAME_KdPrint(("MaxTimeoutInTicks: %u\n", pDevExt->MaxTimeoutInTicks));


    // Calibrate nQuiesceLoop for spinning in read_port loops to timeout after 10ms
    {
        USHORT  i;
        PUCHAR  JoyPort;
        ULONG   ulStart, ulEnd;
        UCHAR   byteJoy;
        ULONG   LoopTimeInMicroSeconds;
        KIRQL   OldIrql;

        JoyPort = pDevExt->PortInfo.GamePortAddress;

        KeRaiseIrql(PROFILE_LEVEL, &OldIrql );
        ulStart = KeQueryPerformanceCounter(NULL).LowPart;
        for (i = 0; i < 1000; i++) {
            // READ PORT
            byteJoy = (*pDevExt->PortInfo.ReadAccessor)(JoyPort);
            if ((byteJoy & X_AXIS_BITMASK)) {
                ;
            }
        }
        ulEnd = KeQueryPerformanceCounter(NULL).LowPart;
        KeLowerIrql( OldIrql );
        LoopTimeInMicroSeconds = TimeInMicroSeconds (ulEnd - ulStart, pDevExt->Frequency);
        pDevExt->nQuiesceLoop = (ULONG) (((__int64)1000L *
                        (__int64)ANALOG_POLL_TIMEOUT) /
                        (__int64)LoopTimeInMicroSeconds);
        HGAME_KdPrint(("READ_PORT_UCHAR loop, 1000 interations: %u ticks\n",
                            ulEnd - ulStart));
        HGAME_KdPrint(("nQuiesceLoop: %u\n", pDevExt->nQuiesceLoop));
    }

    // allow max of 100 polls/s (min time  between polls 10ms),
    // which reduces time spinning in the NT kernel
    pDevExt->nMinTicksBetweenPolls = TimeInTicks (10000, pDevExt->Frequency);
    pDevExt->LastPollTime = KeQueryPerformanceCounter (NULL);

    //Initialize statistics structure
    pDevExt->JoyStatistics.Polls = 0;
    pDevExt->JoyStatistics.Timeouts = 0;
    pDevExt->JoyStatistics.PolledTooSoon = 0;
    pDevExt->JoyStatistics.Redo = 0;

    // Everything is fine so far so let's say
    // the device has started

    pDevExt->Started = TRUE;

    HGAME_KdPrint(("HGAME_StartDevice Exit\n"));
    return STATUS_SUCCESS;
}

VOID
HGAME_StopDevice (
    IN PDEVICE_EXTENSION pDevExt,
    IN BOOLEAN      TouchTheHardware
    )
/*++

Routine Description:
    The PlugPlay system has dictacted the removal of this device.  We have
    no choise but to detach and delete the device objecct.
    (If we wanted to express and interest in preventing this removal,
    we should have filtered the query remove and query stop routines.)

    Note! we might receive a remove WITHOUT first receiving a stop.

Arguments:
    pDevExt - The device extension for the device being started.
    TouchTheHardware - Can we actually send non PnP irps to this thing?

Return Value:

     NT Status code

--*/
{
    PAGED_CODE ();

    HGAME_KdPrint(("HGAME_StopDevice enter \n"));
    ASSERT (!pDevExt->Removed);
    //
    // The PlugPlay system should not have started a removed device!
    //
    if (!pDevExt->Started) {
        return;
    }
    if (TouchTheHardware) {
        //
        // Undo any value add thing required to allow this device to actually
        // stop.  If there is some shutdown procedure required, or any
        // settings required for this device before system shutdown or
        // device removal, now is the best time for it.
        //

    } else {
        //
        // The device is no longer around, so we cannot actually control it.
        // We should instead do what ever necessary in lieu of that.
        //
        ;
    }

    pDevExt->Started = FALSE;

    return;
}

NTSTATUS
HGAME_Power (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP pIrp
    )
/*++

Routine Description:

    The power dispatch routine.
    This filter does not recognize power IRPS.  It merely sends them down,
    unmodified to the next device on the attachment stack.

    As this is a POWER irp, and therefore a special irp, special power irp
    handling is required.

    No completion routine is required.

Arguments:

   DeviceObject - pointer to a device object.

   pIrp - pointer to an I/O Request Packet.

Return Value:

      NT Status code

--*/
{
    PDEVICE_EXTENSION  pDevExt;
    NTSTATUS           ntStatus;

    PAGED_CODE ();

    HGAME_KdPrint(("HGAME_Power Enter \n"));

    pDevExt = GET_MINIDRIVER_DEVICE_EXTENSION (DeviceObject);

    //
    // This IRP was sent to the filter driver.
    // Since we do not know what to do with the IRP, we should pass
    // it on along down the stack.
    //

    InterlockedIncrement (&pDevExt->OutstandingIO);

    if (pDevExt->Removed) {
        ntStatus = STATUS_DELETE_PENDING;
        pIrp->IoStatus.Information = 0;
        pIrp->IoStatus.Status = ntStatus;
        IoCompleteRequest (pIrp, IO_NO_INCREMENT);
    } else {
        IoSkipCurrentIrpStackLocation (pIrp);

        //
        // Power IRPS come synchronously; drivers must call
        // PoStartNextPowerIrp, when they are ready for the next power irp.
        // This can be called here, or in the completetion routine.
        //
        PoStartNextPowerIrp (pIrp);

        //
        // NOTE!!! PoCallDriver NOT IoCallDriver.
        //
        ntStatus =  PoCallDriver (pDevExt->TopOfStack, pIrp);
    }

    if (0 == InterlockedDecrement (&pDevExt->OutstandingIO)) {
        KeSetEvent (&pDevExt->RemoveEvent, 0, FALSE);
    }
    HGAME_KdPrint(("HGAME_Power Exit\n"));
    return ntStatus;
}


NTSTATUS
HGAME_GetResourcesComplete (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             pIrp,
    IN PVOID            Context
    )
/*+
Routine Description:

    Get the IRP back

Arguments:

    DeviceObject - pointer to the device object


Return Value:

    STATUS_MORE_PROCESSING_REQUIRED


--*/
{
    UNREFERENCED_PARAMETER (DeviceObject);
    KeSetEvent ((PKEVENT) Context, 0, FALSE);

    if (pIrp->PendingReturned) {
        IoMarkIrpPending( pIrp );
    }

    return STATUS_MORE_PROCESSING_REQUIRED;
}


NTSTATUS
HGAME_GetResources(
    IN PDEVICE_EXTENSION pDevExt,
    IN PIRP pIrp
    )
/*++

Routine Description:

    Gets gameport resource information from the GameEnum driver

Arguments:

   pDevExt - pointer to device extension.

Return Value:

      NT Status code

--*/
{
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    KEVENT              IoctlCompleteEvent;
    IO_STATUS_BLOCK     IoStatus;
    PIO_STACK_LOCATION  pIrpStack, nextStack;

    PAGED_CODE ();

    HGAME_KdPrint(("HGAME_GetResources Enter\n"));
    //
    // issue a synchronous request to get the resources info from GameEnum
    //

    KeInitializeEvent(&IoctlCompleteEvent, NotificationEvent, FALSE);

    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
    nextStack = IoGetNextIrpStackLocation(pIrp);
    ASSERT(nextStack != NULL);

    //
    // pass the Portinfo buffer of the DeviceExtension
    //

    pDevExt->PortInfo.Size = sizeof (GAMEENUM_PORT_PARAMETERS);

    nextStack->Parameters.DeviceIoControl.InputBufferLength =
                sizeof (GAMEENUM_PORT_PARAMETERS);
    nextStack->Parameters.DeviceIoControl.OutputBufferLength =
                sizeof (GAMEENUM_PORT_PARAMETERS);

    pIrp->UserBuffer = &pDevExt->PortInfo;

    nextStack->Parameters.DeviceIoControl.IoControlCode =
                IOCTL_INTERNAL_GAMEENUM_GET_PORT_PARAMETERS;

    nextStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;

    IoSetCompletionRoutine (pIrp, HGAME_GetResourcesComplete,
                &IoctlCompleteEvent, TRUE, TRUE, TRUE);

    HGAME_KdPrint (("calling GameEnum\n"));

    ntStatus = IoCallDriver(pDevExt->TopOfStack, pIrp);
    if (STATUS_PENDING == ntStatus) {
        ntStatus = KeWaitForSingleObject(
                       &IoctlCompleteEvent,
                       Suspended,
                       KernelMode,
                       FALSE,
                       NULL);
    }

    HGAME_KdPrint(("HGAME_GetResources Exit\n"));
    return pIrp->IoStatus.Status;
}

