/*++

Copyright (c) 1997  Microsoft Corporation

Module Name:

    poll.c

Abstract: This module contains  polling routine for analog gameport device

Author:

    Eliyas Yakub (Mar, 11, 1997)

Environment:

    Kernel mode

Revision History:


--*/
#include <hidgame.h>



NTSTATUS
HGAME_AnalogPoll(
    PDEVICE_OBJECT DeviceObject,
    PIRP    pIrp
    )

/*++

Routine Description:

    Polls the analog device for position and button information.
    The position  information in analog devices is coveyed by the
    duration of a pulse width. Each axis occupies one bit position.
    The read operation is started by writing a value to the joystick
    io address. Immediately thereafter we begin examing the values
    returned and the elapsed time.

    This sort of device has a few limitations:

    First, button information is not latched by the device, so if a
    button press which occurrs in between polls will be lost.
    There is really no way to prevent this short of devoting
    the entire cpu to polling.

    Secondly, although we raise IRQL to DISPATCH_LEVEL, other interrupts
    will occur during the polling routines and this will have the
    effect of lengthening the pulse width (by delaying our polling loop)
    and thus there will be some fluctuation about the actual value.
    It might be possible to try another IRQL to see if this helps,
    but ultimately, nothing short of disabling interrupts altogether
    will insure success. This is too much of a price to pay. The
    solution is a better device.

    Third, when circumstances cause a poll to last too long, we abort it
    and retry the operation. We have to do this to place an upper
    bound on the time we poll, and an upper bound on the time we
    spend at an elevated IRQL.

    But in this case both the position information and
    the button press information is lost. Note that there is an
    upper bound  on the poll duration, beyond which we conclude
    that the device is disconnected.


Arguments:

    DeviceObject     -- pointer to the device object
    pIrp             -- pointer to the requesing Irp


Return Value:

    STATUS_SUCCESS   -- if the poll succeeded,
    STATUS_TIMEOUT   -- if the poll failed (timeout),
                        or the checksum was incorrect
    STATUS_PENDING   -- if polled too soon
    STATUS_BUFFER_TOO_SMALL -- if input buffer is small

--*/
{

    PIO_STACK_LOCATION  pIrpStack;
    PHIDGAME_INPUT_DATA pInput;
    PDEVICE_EXTENSION   pDevExt;
    PUCHAR              JoyPort;
    USHORT              i;
    ULONG               xTime, yTime, zTime, rTime;
    ULONG               MaxRedos, Frequency;
    BOOLEAN             Redo, bQuiesce;
    UCHAR               Buttons, PortVal;
    UCHAR               xMask, yMask, zMask, rMask;
    
    HGAME_KdPrint(("Enter HIDG_AnalogPoll\n"));

    pIrp->IoStatus.Information  = 0;

    bQuiesce = FALSE;

    pDevExt = GET_MINIDRIVER_DEVICE_EXTENSION (DeviceObject);

    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
    pInput  = (PHIDGAME_INPUT_DATA)pIrp->UserBuffer;

    // Make sure buffer has enough space
    RtlZeroMemory( pInput, sizeof(HIDGAME_INPUT_DATA) );
    
    if (pIrpStack->Parameters.DeviceIoControl.OutputBufferLength < 
            sizeof(HIDGAME_INPUT_DATA))
                return STATUS_BUFFER_TOO_SMALL;

    // Let us assume the joystick is unplugged
    // until proven otherwise

    pDevExt->Unplugged = TRUE;

    // If currenttime is less than the lasttime polled plus
    // minimum time required for the device to settle
    // then don't poll  and return the last good packet

    if (KeQueryPerformanceCounter(NULL).QuadPart <
                (pDevExt->LastPollTime.QuadPart + pDevExt->nMinTicksBetweenPolls) )
    {
        pDevExt->JoyStatistics.PolledTooSoon++;
        return  STATUS_PENDING;
    }

    // do the analog poll

    pDevExt->JoyStatistics.Polls++;

    // Find out the current counter value
    pDevExt->LastPollTime = KeQueryPerformanceCounter (NULL);

    // Find where our port and data area are, and related parameters
    Frequency     = pDevExt->Frequency;
    JoyPort      = pDevExt->PortInfo.GamePortAddress;

    // Read port state
    // PortVal = READ_PORT_UCHAR(JoyPort);
    PortVal = (*pDevExt->PortInfo.ReadAccessor)(JoyPort);
    Buttons = 0;

    // Get current button states and build bitmasks
    // for the resistive inputs
    //

    switch (pDevExt->PortInfo.Portion) {
    case GameenumFirstHalf:
        xMask   = JOYSTICK1_X_MASK;
        yMask   = JOYSTICK1_Y_MASK;
        zMask   = 0;
        rMask   = 0;
        if (!(PortVal & JOYSTICK1_BUTTON1)) {
            pInput->Button1 = 1;
        }
        if (!(PortVal & JOYSTICK1_BUTTON2)) {
            pInput->Button2 = 1;
        }
        break;

    case GameenumSecondHalf:
        xMask   = JOYSTICK2_X_MASK;
        yMask   = JOYSTICK2_Y_MASK;
        zMask   = 0;
        rMask   = 0;
        if (!(PortVal & JOYSTICK2_BUTTON1)) {
            pInput->Button1 = 1;
        }
        if (!(PortVal & JOYSTICK2_BUTTON2)) {
            pInput->Button2 = 1;
        }
        break;

    case GameenumWhole:
        if (!(PortVal & JOYSTICK1_BUTTON1)) {
            pInput->Button1 = 1;
        }
        if (!(PortVal & JOYSTICK1_BUTTON2)) {
            pInput->Button2 = 1;
        }
        if (!(PortVal & JOYSTICK2_BUTTON1)) {
            pInput->Button3 = 1;
        }
        if (!(PortVal & JOYSTICK2_BUTTON2)) {
            pInput->Button4 = 1;
        }
        switch (pDevExt->PortInfo.NumberAxis) {
        case 2:
            xMask      = JOYSTICK1_X_MASK;
            yMask      = JOYSTICK1_Y_MASK;
            zMask      = 0;
            rMask      = 0;
            break;

        case 3:
            xMask      = JOYSTICK1_X_MASK;
            yMask      = JOYSTICK1_Y_MASK;
            zMask      = 0;
            // this is 0x08, typically the third axis on 3axis joysticks
            rMask      = JOYSTICK1_R_MASK;
            break;

        case 4:
            // Note that we read all axi because we don't know which
            // axis will be used by the joystick, and we read all the
            // buttons because no other joystick could use them

            xMask      = JOYSTICK1_X_MASK;
            yMask      = JOYSTICK1_Y_MASK;
            zMask      = JOYSTICK1_Z_MASK;
            rMask      = JOYSTICK1_R_MASK;
            break;
        default:
            HGAME_KdPrint(("Invalid value:%d\n", pDevExt->PortInfo.NumberAxis));
            ASSERT(FALSE);
            return STATUS_TIMEOUT;
            break;
        } // end of switch
        break;

    } // end of switch


    // Insure that the resistive inputs are currently reset
    // before performing the next read. If we find one or more
    // hot inputs, wait briefly for them to reset. If they
    // don't, we assume that the joystick is unplugged.

    // Wait for port to quiesce
    // HGAME_KdPrint(("Mask value:(0x%x)\n", xMask|yMask|zMask|rMask));
    for (i = 0; i < pDevExt->nQuiesceLoop; i++) {
        //read port
        PortVal = (*pDevExt->PortInfo.ReadAccessor)(JoyPort);
        if ((PortVal & (xMask|yMask|zMask|rMask)) == 0) 
            bQuiesce = TRUE;
    }

    if (!bQuiesce) {
        HGAME_KdPrint(("Failed to quiesce resistive inputs\n"));
        return STATUS_TIMEOUT;
    }

    // Note that timing is EXTREMELY critical in the loop below.
    // Avoid calling complicated arithmetic (eg TimeInMicroSeconds)
    // or we will decrease our accuracy
    // Other problems with accuracy, probably larger than
    // the delays caused by arithmetic, are the latency in calls to
    // KeQueryPerformanceCounter typically about 5 us), and delays
    // that can occur on the bus when DMA is taking place.

    // Now poll the device.  We wait until the status bit(s) set(s)
    // and note the time.  If the time since the last poll was
    // too great we ignore the answer and try again.

    // Loop until we get a decent reading or exceed the threshold

    for (Redo = TRUE, MaxRedos = 20; Redo && --MaxRedos != 0;)
    {
        ULONG StartTime;
        ULONG CurrentTime;
        ULONG PreviousTime;
        ULONG PreviousTimeButOne;
        UCHAR ResistiveInputMask;

        ResistiveInputMask = xMask | yMask | zMask | rMask;

        // Lock on to start time
        StartTime = KeQueryPerformanceCounter(NULL).LowPart;

        //WRITE_PORT_UCHAR(JoyPort, JOY_START_TIMERS);
        (*pDevExt->PortInfo.WriteAccessor)(JoyPort, JOY_START_TIMERS);

        CurrentTime = KeQueryPerformanceCounter(NULL).LowPart - StartTime;

        //PortVal = READ_PORT_UCHAR(JoyPort);
        PortVal = (*pDevExt->PortInfo.ReadAccessor)(JoyPort);

        // Now wait until our end times for each coordinate

        PreviousTimeButOne = 0;
        PreviousTime = CurrentTime;

        for (Redo = FALSE;
             ResistiveInputMask;
             PreviousTimeButOne = PreviousTime,
             PreviousTime = CurrentTime,
             PortVal = (*pDevExt->PortInfo.ReadAccessor)(JoyPort) //ReadPort
            ) {

            PortVal = ResistiveInputMask & ~PortVal;
            CurrentTime =
                KeQueryPerformanceCounter(NULL).LowPart - StartTime;

            if (CurrentTime > pDevExt->MaxTimeoutInTicks) {

                HGAME_KdPrint(("Polling failed - ResistiveInputMask = %x,  Time = %d\n",
                            (ULONG)ResistiveInputMask,
                            TimeInMicroSeconds(CurrentTime, Frequency)));

                return STATUS_TIMEOUT;
            }

            if (PortVal & xMask) {
                ResistiveInputMask &= ~xMask;
                xTime = PreviousTime;
            }
            if (PortVal & yMask) {
                ResistiveInputMask &= ~yMask;
                yTime = PreviousTime;
            }
            if (PortVal & zMask) {
                ResistiveInputMask &= ~zMask;
                zTime = PreviousTime;
            }
            if (PortVal & rMask){
                ResistiveInputMask &= ~rMask;
                rTime = PreviousTime;
            }

            if (PortVal && CurrentTime - PreviousTimeButOne >
                    pDevExt->ThresholdInTicks) {
                // Something (DMA or interrupts) delayed our read loop,
                // start again.
                HGAME_KdPrint(("Too long a gap between polls - %u us\n",
                    TimeInMicroSeconds(CurrentTime - PreviousTimeButOne, Frequency)));
                pDevExt->JoyStatistics.Redo++;
                Redo = TRUE;
                break;
            }
        }//end of inner for loop
    }//end of outer for loop

    if (MaxRedos == 0)
    {
        HGAME_KdPrint(("Overran redos to get counters\n"));
        pDevExt->Unplugged = TRUE;
        return STATUS_TIMEOUT;
    }

    pDevExt->Unplugged = FALSE;
    pInput->X = TimeInMicroSeconds(xTime, Frequency);
    pInput->Y = TimeInMicroSeconds(yTime, Frequency);
    pInput->Z = (zMask) ? TimeInMicroSeconds(zTime, Frequency) : 0;
    pInput->T = (rMask) ? TimeInMicroSeconds(rTime, Frequency) : 0;

    HGAME_KdPrint(("X = %d, Y = %d, Z = %d, R = %d,Buttons = %d,%d,%d,%d\n",
          pInput->X, pInput->Y, pInput->Z, pInput->T,
          pInput->Button1,pInput->Button2,pInput->Button3,pInput->Button4));

    pIrp->IoStatus.Information  = sizeof(HIDGAME_INPUT_DATA);
    return STATUS_SUCCESS;
}

ULONG
TimeInMicroSeconds(
    ULONG dwTime,
    ULONG Frequency
    )
{
    ULONG Remainder;

    return RtlExtendedLargeIntegerDivide(
                RtlEnlargedUnsignedMultiply( dwTime, 1000000L),
                Frequency,
                &Remainder
           ).LowPart;
}

ULONG
TimeInTicks(
    ULONG dwTimeInMicroSeconds,
    ULONG Frequency
    )
{
    return (ULONG) (((__int64)dwTimeInMicroSeconds *
                    (__int64)Frequency) /  (__int64) 1000000L);
}



