/*++

Copyright (c) 1996  Microsoft Corporation

Module Name:

   mpinit.c

Abstract:

   This is the WDM sample streaming class minidriver.  This module contains
   code related to minidriver initialization and unitialization.

Author:

Environment:

   Kernel mode only


Revision History:

--*/

#include "common.h"

/*
** DriverEntry()
**
** This routine is called when the mini driver is first loaded.  The driver
** should then call the StreamClassRegisterAdapter function to register with
** the stream class driver
**
** Arguments:
**
**  Context1:  The context arguments are private plug and play structures
**             used by the stream class driver to find the resources for this
**             adapter
**  Context2:
**
** Returns:
**
** This routine returns an NT_STATUS value indicating the result of the
** registration attempt. If a value other than STATUS_SUCCESS is returned, the
** minidriver will be unloaded.
**
** Side Effects:  none
*/

ULONG DriverEntry ( PVOID Context1, PVOID Context2 )
{

	HW_INITIALIZATION_DATA HwInitData;  // Hardware Initialization data structure
    
	DebugPrint((DebugLevelVerbose,"Stream Sample MiniDriver DriverEntry"));

    //
    // all unused fields should be zero, but in this sample driver, all fields
    // are initialised
    //

//       ZeroMemory(&HwInitData, sizeof(HwInitData));

    HwInitData.HwInitializationDataSize = sizeof(HwInitData);

/*

 Entry points for the mini Driver.  All entry routines will be called
 at High priority.  If the driver has a task that requires a large amount
 of time to accomplish (e.g. double buffer copying of large amounts of
 data) The driver should request a callback at lower priority to perform the
 desired function.  See the Syncronization comment for issues on re-entrancy

 */

	HwInitData.HwInterrupt          = HwInterrupt;      // IRQ handling routine

    //
    // data handling routines
    //

/*
 The HwReceivePacket field describes the entry point for receiving an SRB
 Request from the stream class driver that is a request to the adapter, as
 opposed to the stream based request handlers that are initialised in the
 OpenStream function
 */

	HwInitData.HwReceivePacket      = AdapterReceivePacket;

/*
 The HwCancelPacket entry is used to cancel any packet the minidriver may
 currently have outstanding.  This could be a stream, or an adapter based
 request.  This routine will only be called under extreme circumstances, when
 an upper layer is attempting to recover and packets are considered lost
 */

	HwInitData.HwCancelPacket       = AdapterCancelPacket;

/*
 The HwRequestTimeoutHandler entry is used when a packet times out.  Again,
 This could be a packet for the device, or a stream based packet.  This is
 not necessarily an error, and it is up to the minidriver to determine what
 to do with the packet that has timed out.
 */

	HwInitData.HwRequestTimeoutHandler      = AdapterTimeoutPacket;

/*
 Sizes for data structure extensions.  See mpinit.h for definitions
 */

	HwInitData.DeviceExtensionSize = sizeof(HW_DEVICE_EXTENSION);
	HwInitData.PerRequestExtensionSize = sizeof(SRB_EXTENSION);
	HwInitData.FilterInstanceExtensionSize = 0;
	HwInitData.PerStreamExtensionSize = sizeof(STREAMEX);

/*

 The BusMasterDMA field indicates to the system that the memory pointers passed
 in to the mini driver in the data packets may be used for direct Bus Master DMA
 access.  If the mini driver uses a translation buffer (to ensure minimum DMA
 size, or correct buffer alignment), this BOOLEAN should be set to FALSE.

 */

	HwInitData.BusMasterDMA = FALSE;

/*

 The Ddma24BitAddresses field indicates a device that uses DMA that only can
 access the lower 24bits of the 32 bit address space.  Again, this should only
 be set to true if the minidriver is going to be doing DMA directly to the buffer
 passed in the data packets, and not through a double buffer scenario.

 */

	HwInitData.Dma24BitAddresses = FALSE;

	HwInitData.BufferAlignment = 3;

/*

IMPORTANT NOTE:

 If the TurnOffSynchronization field is set to FALSE, the mini driver may never
 be re-entered.  

 Under these conditions, if the mini-driver uses a DISPATCH or lower priority
 callback routine, the miniDriver must disable any interrupts that it might
 receive.  If an interrupt controlled by the driver is received while code in
 the mini driver is running at DISPATCH or lower priority, the interrupt will
 be LOST.  If an interrupt is received while at HIGH priority, it will be queued
 until the code in the driver is finished.

 If the TurnOffSynchronization field is set to TRUE, the mini-driver must be
 fully capable of handling multi-processor and single processor re-entrancy

 */

    HwInitData.TurnOffSynchronization = FALSE;

/*

 The DmaBufferSize specifies the size of the single contiguous region of
 physical memory that the driver needs at hardware intialization time.  The
 memory will be returned to the driver when the driver makes the
 StreamClassGetDmaBuffer call.  It is important to use as little physical
 buffer as possible here, as this will be locked physical memory that is
 unavailable to the sytem, even when the streaming class mini driver is not
 in use

 */

	HwInitData.DmaBufferSize = 8 * 1024;
    
	DebugPrint((DebugLevelVerbose,"CSSAMPLE: call to portinitialize"));

    //
    // attempt to register with the streaming class driver.  Note, this will
    // result in calls to the HW_INITIALIZE routine.
    //

	return (StreamClassRegisterAdapter(Context1,
				       Context2,
				       &HwInitData));

}

/*
** HwInitialize()
**
**   Initializes an adapter accessed through the information provided in the
**   ConfigInfo structure
**
** Arguments:
**
**   pSRB - pointer to the request packet for the initialise command
**
**    ->ConfigInfo - provides the I/O port, memory windows, IRQ, and DMA levels
**                that should be used to access this instance of the device
**
** Returns:
**
**       STATUS_SUCCESS - if the card initializes correctly
**       STATUS_NO_SUCH_DEVICE - or other if the card is not found, or does
**                               not initialize correctly.
**
**
** Side Effects:  none
*/

NTSTATUS HwInitialize (IN PHW_STREAM_REQUEST_BLOCK pSrb)
{
    PPORT_CONFIGURATION_INFORMATION ConfigInfo = pSrb->CommandData.ConfigInfo;
	NTSTATUS Stat = STATUS_SUCCESS;

	PHW_DEVICE_EXTENSION pHwDevExt =
	      (PHW_DEVICE_EXTENSION)ConfigInfo->HwDeviceExtension;

	DebugPrint((DebugLevelVerbose,"Entry : HwInitialize()"));

	if (ConfigInfo->NumberOfAccessRanges < 1)
	{
		TRAP

		DebugPrint((DebugLevelVerbose,"ST3520: illegal config info"));

		return(STATUS_NO_SUCH_DEVICE);
	}

    //
    // pick up the I/O window for this card
    //

	pHwDevExt->ioBaseLocal = (ULONG *)ConfigInfo->AccessRanges[0].RangeStart.LowPart;

    //
    // pick up the memory window for this card.  Note: memory windows are
    // passed as logical address.  Just assign a pointer to the memory window
    // and access the address.  This is commented out for this driver, as the
    // sample does not use memory windows
    //

    // pHwDevExt->MemBase = ConfigInfo->AccessRanges[0].RangeStart.LowPart;

    //
    // pick up the interrupt level
    //

	pHwDevExt->Irq  = (USHORT)ConfigInfo->BusInterruptLevel;

    //
    // initialise some of the HwDevExtension structure
    //

	pHwDevExt->AudioDeviceExt.pCurrentSRB = NULL;
	pHwDevExt->VideoDeviceExt.pCurrentSRB = NULL;

    pHwDevExt->VideoDeviceExt.xdhl.pCurrentSRB = &
		pHwDevExt->VideoDeviceExt.pCurrentSRB;
    pHwDevExt->AudioDeviceExt.xdhl.pCurrentSRB = &
	pHwDevExt->AudioDeviceExt.pCurrentSRB;

	pHwDevExt->AudioDeviceExt.xdhl.pfnDataStub = AudioPacketStub;
    pHwDevExt->VideoDeviceExt.xdhl.pfnDataStub = VideoPacketStub;

	pHwDevExt->VideoDeviceExt.DeviceState = KSSTATE_PAUSE;
	pHwDevExt->AudioDeviceExt.DeviceState = KSSTATE_PAUSE;

/*

 attempt to initialize the requested adapter, and verify that it exists at the
 specified addresses.  Remember that this is a syncronous routine, and that if
 the TurnOffSynchronization routine is FALSE, all IRQ processing will be
 deferred until after the HwInitialize routine has completed

 */

	// Stat = MyHwInitialize (pHwDevExt);

	DebugPrint((DebugLevelVerbose,"Exit : HwInitialize()"));

    //
    // initialise the size of stream descriptor information.  For the sample
    // driver, we have three stream descriptors.
    //

	ConfigInfo->StreamDescriptorSize =
	3 * (sizeof (HW_STREAM_INFORMATION) + // 3 stream descriptors
		3 * sizeof (BOOLEAN)) +               // 3 related stream Booleans
	sizeof (HW_STREAM_HEADER);            // and 1 stream header

    //
    // if we successfully initialised, we need to allocate a DMA buffer here
    //

	if (Stat)
	{

	//
	// the size is already specified in the HwInitData structure  This
	// provides the logical address of the DMA buffer.
	//
	// Note: The GetDmaBuffer routine may be called only ONCE per instance
	// of the driver.  (This is equivalent to once for every call to
	// HwInitialize)
	//

		pHwDevExt->VideoDeviceExt.pDMABuf =
	    StreamClassGetDmaBuffer(pHwDevExt);

	//
	// pick up the physical address of the DMA buffer for programming into
	// the hardware DMA controller
	//

	pHwDevExt->VideoDeviceExt.pPhysDmaBuf =
	    StreamClassGetPhysicalAddress(
			pHwDevExt,
			NULL,
			pHwDevExt->VideoDeviceExt.pDMABuf,
			DmaBuffer,              // indicate it's for DMA buffer
			&pHwDevExt->VideoDeviceExt.cDmaBuf);
	}

    return(Stat);

}

/*
** HwUnInitialize()
**
**   Release all resources and clean up the hardware
**
** Arguments:
**
**      DeviceExtension - pointer to the deviceextension structure for the
**                       the device to be free'd
**
** Returns:
**
** Side Effects:  none
*/

BOOLEAN HwUnInitialize ( IN PVOID DeviceExtension)
{
    //
    // call the MiniDriver HwUnitialize routine 
    //

    //  return(MyHwUnInitialize (DeviceExtension));

    return(TRUE);
}

