
/*++

Copyright (c) 1996  Microsoft Corporation

Module Name:

   mpvideo.c

Abstract:

   This is the WDM sample streaming class minidriver.  This module contains
   code related to processing individual commands

Author:

Environment:

   Kernel mode only


Revision History:

--*/


#include "common.h"
#include "ks.h"
#include "ksguid.h"
#include "mpegprop.h"

/***************************************************************************

                    Data Packet Handling Routines

***************************************************************************/

/*
** StartDataPacket()
**
**   Routine to initialise the stream data packet handling
**
** Arguments:
**
**   pSrb - Pointer to the stream request block
**
** Returns:
**
** Side Effects:  none
*/

VOID StartDataPacket(PHW_STREAM_REQUEST_BLOCK pSrb)
{
	ULONG dwErrCode = NO_ERROR;

	PHW_DEVICE_EXTENSION pHwDevExt =
	 ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);

	PEXTENSION_DATA_HANDLER pxdhl = 
	 ((PSTREAMEX)pSrb->StreamObject->HwStreamExtension)->pxdhl;

     TRAP

	 //
	 // set up for initial parsing of the scatter gather packet.
	 //
 
	 pxdhl->cPacket = pxdhl->cOffs = 0;
	 pxdhl->pPacket = pSrb->CommandData.DataBufferArray;

	 pxdhl->pfnDataStub(pSrb->StreamObject);

}


/*
** VideoPacketStub()
**
**    Stub routine for changing the priority of the data copy for video
**
** Arguments:
**
** Returns:
**
** Side Effects:  none
*/

VOID VideoPacketStub(PHW_STREAM_OBJECT pstrm)
{
    //
    // since this miniport is switching to a lower priority for data copy,
    // it must disable any IRQs that it handles here
    //

	DisableIRQ(pstrm);

	StreamClassCallAtNewPriority(pstrm,
                               pstrm->HwDeviceExtension,
							   Dispatch,
							   ProcessDataPacket,
                               pstrm);
}

/*
** AudioPacketStub()
**
**    Stub routine for changing the priority of the data copy for Audio
**
** Arguments:
**
** Returns:
**
** Side Effects:  none
*/

VOID AudioPacketStub(PHW_STREAM_OBJECT pstrm)
{
    //
    // since this miniport is switching to a lower priority for data copy,
    // it must disable any IRQs that it handles here
    //
    // Note: even though this is the identical code to the VideoPacketStub,
    // we need two entry pionts to keep from cancelling video timers!
    //

	DisableIRQ(pstrm);

	StreamClassCallAtNewPriority(pstrm,
                               pstrm->HwDeviceExtension,
							   Dispatch,
							   ProcessDataPacket,
                               pstrm);
}



/*
** ProcessDataPacket()
**
**    Process the audio or video data packets.  This routine will walk through
**    the logical scatter gather packets.
**
** Arguments:
**
**    pstrm - stream data object for the packet to be processed
**
** Returns:
**
** Side Effects:  none
*/

VOID ProcessDataPacket(PHW_STREAM_OBJECT pstrm)
{
    PHW_DEVICE_EXTENSION pdevext = pstrm->HwDeviceExtension;
    PHW_STREAM_REQUEST_BLOCK pSrb;

	ULONG	uSent;

	PEXTENSION_DATA_HANDLER pxdhl = 
	 ((PSTREAMEX)pstrm->HwStreamExtension)->pxdhl;

    //
    // pick up the current request block
    //

	pSrb = *(pxdhl->pCurrentSRB);

    //
    // make sure something didn't get cancelled on us
    //

	if (!pSrb)
	{
		TRAP

		return;
	}

    //
    // loop through all the data packets, sending data to the appropriate
    // data handler for the video or audio device
    //

	do
	{

/*

 The send data routine will return the number of bytes of data that were sent
 to the appropriate device, and send the data currently pointed to by the xdhl
 to the device

 */
		uSent = pxdhl->pfnSendData(pSrb, pxdhl);

		pxdhl->cOffs += uSent;

		//
		// check if we finished this packet.  If so, go on to the
		// next packet
		//

		if (pxdhl->cOffs >=
			pxdhl->pPacket->DataUsed)
		{

            //
            // packet finished, set up for the next packet
            //

			pxdhl->pPacket++;

			//
			// reset the packet offset
			//

			pxdhl->cOffs = 0;
			pxdhl->cPacket++;

			//
			// if we have finished all the packets, then call this request back
            //
			//

			if (pxdhl->cPacket >=
				 pSrb->NumberOfBuffers)           
			{

				pSrb->Status = STATUS_SUCCESS;

				*(pxdhl->pCurrentSRB) = 0;

                //
                // call this command back successfully
                //
	
		        StreamClassStreamNotification(ReadyForNextStreamDataRequest,
				        pSrb->StreamObject);
	
		        StreamClassStreamNotification(StreamRequestComplete,
				        pSrb->StreamObject,
				        pSrb);


                //
                // re-enable the IRQs, since we are done transferring data
                //
        
	            StreamClassCallAtNewPriority(pstrm,
                               pstrm->HwDeviceExtension,
							   High,                    
							   EnableIRQ,
                               pstrm);
				return;

			}
		}

	} while (uSent);  // keep going until there is no room left in the device


	//
	// request a timer callback to complete processing the data request
    // the mini-driver could also do this on a "Buffer nearly empty" type
    // interrupt, instead of using timers.
	//

#define PACKET_TIMER 10000      // set for 10ms timer

    StreamClassScheduleTimer(pstrm,
                            pdevext,
                            PACKET_TIMER,
                            pxdhl->pfnDataStub,
                            pstrm);

    //
    // re-enable the IRQs, since we are done transferring data
    //
        
	StreamClassCallAtNewPriority(pstrm,
                    pstrm->HwDeviceExtension,
					High,                    
					EnableIRQ,
                    pstrm);
		

}


/*
** GetVideoProperty()
**
**    Routine to process video property requests
**
** Arguments:
**
**    pSrb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

VOID GetVideoProperty(PHW_STREAM_REQUEST_BLOCK pSrb)
{
	PHW_DEVICE_EXTENSION phwdevext =
		((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);

    //
    // determine which property set is being requested
    //

	switch (pSrb->CommandData.PropertyInfo->PropertySetID)
	{

    //
    // zero is the first property, which is the get buffer level
    //

	case 0:

	    GetVidLvl(pSrb);

		break;

	default:

        break;


	}
	pSrb->Status = STATUS_SUCCESS;

}

/*
** GetVidLvl()
**
**    Completes the video buffer level property
**
** Arguments:
**
**    pSrb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

void GetVidLvl(PHW_STREAM_REQUEST_BLOCK pSrb)
{
#define BUF_FULL 0;

	PVIDEO_DEVICE_EXTENSION pvidex = 
	 &(((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension)->VideoDeviceExt);

	//
	// set the max buffer to the buffer full level
	//

	((PBUF_LVL_DATA)(pSrb->CommandData.PropertyInfo->PropertyInfo))
                ->BufferCapacity = BUF_FULL;

	//
	// set current to the current video level
	//

	//((PBUF_LVL_DATA)(pSrb->CommandData.PropertyInfo->PropertyInfo))
    //            ->CurBufferLevel = GetVideoBufferLevel(pSrb);

    //
    // indicate the amount of data that is being returned
    //

	pSrb->ActualBytesTransferred = sizeof (BUF_LVL_DATA);

}

/*
** SetVideoState()
**
**    Sets the current state of the requested stream
**
** Arguments:
**
**    pSrb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

VOID SetVideoState(PHW_STREAM_REQUEST_BLOCK pSrb)
{
	PHW_DEVICE_EXTENSION phwdevext =
		((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);

    //
    // determine which new state is requested
    //

	switch (pSrb->CommandData.StreamState)  
	{
	case KSSTATE_STOP:

        //
        // stop the video
        //

        // VideoStop(pSrb, phwdevext);

		break;


	case KSSTATE_PAUSE:

        //
        // pause the video
        //

		// VideoPause(pSrb, phwdevext);

		break;

	case KSSTATE_RUN:

        //
        // play the video
        //

		// VideoPlay(pSrb, phwdevext);

        break;

	}
	pSrb->Status = STATUS_SUCCESS;

}

/*
** EnableIRQ()
**
**   routine for enabling all IRQs associated with the minidriver
**
** Arguments:
**
** Returns:
**
** Side Effects:  none
*/

void EnableIRQ(PHW_STREAM_OBJECT pstrm)
{
}

/*
** DisableIRQ()
**
**   routine for disabling all IRQs associated with the minidriver
**
** Arguments:
**
** Returns:
**
** Side Effects:  none
*/

void DisableIRQ(PHW_STREAM_OBJECT pstrm)
{
}


