
/*++

Copyright (c) 1996  Microsoft Corporation

Module Name:

   sccmd.c

Abstract:

   This is the WDM sample streaming class minidriver.  This module contains
   code related to minidriver command processing

Author:

Environment:

   Kernel mode only


Revision History:

--*/

#include "common.h"

#include <windef.h>

/*

 The following include files are needed for defining the stream type structures
 that the mini driver supports

 */

#include "strmid.h"

/******************************************************************************

                   Adapter Based Request Handling Routines

******************************************************************************/

/*
** AdapterCancelPacket()
**
**   Request to cancel a packet that is currently in process in the minidriver
**
** Arguments:
**
**   pSrb - pointer to request packet to cancel
**
** Returns:
**
** Side Effects:  none
*/

VOID AdapterCancelPacket(PHW_STREAM_REQUEST_BLOCK pSrb)
{
	PHW_DEVICE_EXTENSION  pdevex = (PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension;
	BOOL fReset = FALSE;	// indicates we need to reset the device

	//
	// need to find this packet, pull it off our queues, and cancel it
	//

	if (pdevex->VideoDeviceExt.pCurrentSRB == pSrb)
	{

        //
        // clear the video command timer
        //


		StreamClassScheduleTimer(pSrb->StreamObject,
             pdevex, 0, VideoPacketStub, pSrb->StreamObject);
		
		pdevex->VideoDeviceExt.pCurrentSRB = NULL;

		//
		// cancel the video timer
		//

		fReset = TRUE;
	}

	if (pdevex->AudioDeviceExt.pCurrentSRB == pSrb)
	{
		pdevex->AudioDeviceExt.pCurrentSRB = NULL;

		StreamClassScheduleTimer(pSrb->StreamObject,
             pdevex, 0, AudioPacketStub, pSrb->StreamObject);

        fReset = TRUE;
	}

	if (fReset)
	{
		TRAP

        //
        // call the system reset function
        //
	}

	pSrb->Status = STATUS_CANCELLED;

    //
    // it is necessary to call the request back correctly.  Determine which
    // type of command it is
    //

	switch (pSrb->Flags & (SRB_HW_FLAGS_DATA_TRANSFER |
				SRB_HW_FLAGS_STREAM_REQUEST))
	{
	//
	// find all stream commands, and do stream notifications
	//

	case SRB_HW_FLAGS_STREAM_REQUEST | SRB_HW_FLAGS_DATA_TRANSFER:

	
		StreamClassStreamNotification(ReadyForNextStreamDataRequest,
				pSrb->StreamObject);
	
		StreamClassStreamNotification(StreamRequestComplete,
				pSrb->StreamObject,
				pSrb);

		break;

	case SRB_HW_FLAGS_STREAM_REQUEST:

		StreamClassStreamNotification(ReadyForNextStreamControlRequest,
				pSrb->StreamObject);
	
		StreamClassStreamNotification(StreamRequestComplete,
				pSrb->StreamObject,
				pSrb);

		break;

	default:


		//
		// this must be a device request.  Use device notifications
		//

		StreamClassDeviceNotification(ReadyForNextDeviceRequest,
				pSrb->HwDeviceExtension);
	
		StreamClassDeviceNotification(DeviceRequestComplete,
				pSrb->HwDeviceExtension,
				pSrb);

	}

}

/*
** AdapterTimeoutPacket()
**
**   This routine is called when a packet has been in the minidriver for
**   too long.  The adapter must decide what to do with the packet
**
** Arguments:
**
**   pSrb - pointer to the request packet that timed out
**
** Returns:
**
** Side Effects:  none
*/

VOID AdapterTimeoutPacket(PHW_STREAM_REQUEST_BLOCK pSrb)
{

	//
	// if we timeout while playing, then we need to consider this
	// condition an error, and reset the hardware, and reset everything
    // as well as cancelling this and all requests
	//

	TRAP

	//
	// if we are not playing, and this is a CTRL request, we still
	// need to reset everything as well as cancelling this and all requests
	//

    //
    // if this is a data request, and the device is paused, we probably have
    // run out of data buffer, and need more time, so just reset the timer,
    // and let the packet continue
    //

    pSrb->TimeoutCounter = pSrb->TimeoutOriginal;
    
}


/*
** AdapterReceivePacket()
**
**   Main entry point for receiving adapter based request SRBs.  This routine
**   will always be called at High Priority.
**
**   Note: This is an asyncronous entry point.  The request does not complete
**         on return from this function, the request only completes when a
**         StreamClassDeviceNotification on this request block, of type
**         DeviceRequestComplete, is issued.
**
** Arguments:
**
**   pSrb - Pointer to the STREAM_REQUEST_BLOCK
**        pSrb->HwDeviceExtension - will be the hardware device extension for
**                                  as initialised in HwInitialise
**
** Returns:
**
** Side Effects:  none
*/

VOID STREAMAPI AdapterReceivePacket(IN PHW_STREAM_REQUEST_BLOCK pSrb)
{

    //
    // pick up the device extension
    //

	PHW_DEVICE_EXTENSION pdevext =
	 ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);


	//
	// determine the type of packet.
	//

	switch (pSrb->Command)
	{

	case SRB_OPEN_STREAM:

        //
        // this is a request to open a specified stream.  
        //

		AdapterOpenStream(pSrb);

		break;

	case SRB_GET_STREAM_INFO:

        //
        // this is a request for the driver to enumerate requested streams
        //

		AdapterStreamInfo(pSrb);

		break;

    case SRB_INITIALIZE_DEVICE:

        pSrb->Status = HwInitialize(pSrb);
        break;

	default:

		TRAP

        //
        // this is a request that we do not understand.  Indicate invalid
        // command and complete the request
        //

	    pSrb->Status = STATUS_NOT_IMPLEMENTED;

	}

    //
    // NOTE:
    //
    // all of the commands that we do, or do not understand can all be completed
    // syncronously at this point, so we can use a common callback routine here.
    // If any of the above commands require asyncronous processing, this will
    // have to change
    //

	StreamClassDeviceNotification(ReadyForNextDeviceRequest,
			pSrb->HwDeviceExtension);

	StreamClassDeviceNotification(DeviceRequestComplete,
			pSrb->HwDeviceExtension,
			pSrb);

}

/*
** AdapterStreamInfo()
**
**   Returns the information of all streams that are supported by the
**   mini-driver
**
** Arguments:
**
**   pSrb - Pointer to the STREAM_REQUEST_BLOCK
**        pSrb->HwDeviceExtension - will be the hardware device extension for
**                                  as initialised in HwInitialise
**
** Returns:
**
** Side Effects:  none
*/

VOID AdapterStreamInfo(PHW_STREAM_REQUEST_BLOCK pSrb)

{

	PHW_DEVICE_EXTENSION pdevext =
	 ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);

     //
     // pick up the pointer to the streaminformation data structure
     //

	 PHW_STREAM_INFORMATION pstrinfo =
            &(pSrb->CommandData.StreamBuffer->StreamInfo);

     //
     // initialise the number of streams supported
     //
     // The sample mini driver supports 3 streams:
     //
     // Mpeg2 PES video
     // Mpeg2 PES audio
     // NTSC composite video output
     //

     pSrb->CommandData.StreamBuffer->StreamHeader.NumberOfStreams = 3;

     //
     // initialise the first stream structure.  This will be the
     // MPEG2 PES VIDEO structure
     //
     // The number of instances field indicates the number of concurrent streams
     // of this type the device can support.  
     //

	 pstrinfo->NumberOfPossibleInstances = 1;

     //
     // indicates the direction of data flow for this stream, relative to the
     // driver
     //

	 pstrinfo->DataFlow = KSPIN_DATAFLOW_IN;

     //
     // dataAccessible - Indicates whether the data is "seen" by the host
     // processor.  If the data is not visible to the processor (such as
     // an NTSC output port) this boolean is set to false.
     //

	 pstrinfo->DataAccessible = TRUE;

     //
     // this stream supports only one format.
     //

	 pstrinfo->NumberOfFormatArrayEntries = 1;

     //
     // set the data format pointer in the stream which indicates that this
     // stream supports the mpeg 2 video format.
     //

	 pstrinfo->StreamFormatsArray = Mpeg2VidInfo;

	 //
	 // set the property information for the video stream
	 //

	 pstrinfo->NumStreamPropArrayEntries = 1;           // only 1 property
	 pstrinfo->StreamPropertiesArray = mpegVidPropSet;  // see strmid.h

	 pstrinfo++;

	 //
	 // set up the stream info structures for the MPEG2 audio
	 //

	 pstrinfo->NumberOfPossibleInstances = 1;
	 pstrinfo->DataFlow = KSPIN_DATAFLOW_IN;
	 pstrinfo->DataAccessible = TRUE;

     //
     // this stream supports only one format.
     //

	 pstrinfo->NumberOfFormatArrayEntries = 1;

     //
     // set the data format pointer in the stream to indicate that MPEG 2
     // audio is supported by this stream.
     //

	 pstrinfo->StreamFormatsArray = Mpeg2AudInfo;

	 pstrinfo++;

	 //
	 // set up the stream info structures for the MPEG2 NTSC stream
	 //

	 pstrinfo->NumberOfPossibleInstances = 1;
	 pstrinfo->DataFlow = KSPIN_DATAFLOW_OUT;

     //
     // since this is an NTSC output pin, the processor cannot see data flow
     // on this pin
     //

	 pstrinfo->DataAccessible = FALSE;

     //
     // this stream supports only one format.  
     //

	 pstrinfo->NumberOfFormatArrayEntries = 1;

     //
     // set the data format pointer in the stream which indicates that this
     // stream supports NTSC output.
     //

	 pstrinfo->StreamFormatsArray = NtscInfo;

     //
     // check here that the stream information provided is no larger than
     // the size specified in the HwInitData
     //

     //
     // indicate success
     //

	 pSrb->Status = STATUS_SUCCESS;
 
}


/*
** AdapterOpenStream()
**
**   This routine is called when an OpenStream SRB request is received
**
** Arguments:
**
**   pSrb - pointer to stream request block for the Open command
**
** Returns:
**
** Side Effects:  none
*/

VOID AdapterOpenStream(PHW_STREAM_REQUEST_BLOCK pSrb)
{
    //
    // the stream extension structure is allocated by the stream class driver
    //

	PSTREAMEX strm = (PSTREAMEX)pSrb->StreamObject->HwStreamExtension;

	PHW_DEVICE_EXTENSION pHwDevExt =
	 ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);

    //
    // fill in any initialization information in the stream extension structure
    //


	//
	// default to success
	//

	pSrb->Status = STATUS_SUCCESS;

    //
    // determine which stream number is being opened.  This number indicates
    // the offset into the array of streaminfo structures that was filled out
    // in the AdapterStreamInfo call.
    //
    // So:
    //   0 - MPEG2_VIDEO_PES
    //   1 - MPEG2_AUDIO_PES
    //   2 - NTSC OUTPUT
    //

	switch (pSrb->StreamObject->StreamNumber)
	{
	case 0:

		//
		// this is the video stream.  Set up the receive packet functions
        // for the video stream
        //

        pSrb->StreamObject->ReceiveDataPacket = (PVOID)VideoReceiveDataPacket;
        pSrb->StreamObject->ReceiveControlPacket = (PVOID)VideoReceiveCtrlPacket;

        //
        // initialize the stream extension data handling information
        //

        strm->pxdhl = &(pHwDevExt->VideoDeviceExt.xdhl);

		break;

	case 1:

		//
		// this is the audio stream.  Set up the receive packet functions
        // for the audio stream
		//

        pSrb->StreamObject->ReceiveDataPacket = (PVOID)AudioReceiveDataPacket;
        pSrb->StreamObject->ReceiveControlPacket = (PVOID)AudioReceiveCtrlPacket;

        //
        // initialize the stream extension data handling information
        //

        strm->pxdhl = &(pHwDevExt->AudioDeviceExt.xdhl);

        break;

	default:

		TRAP

	}

    //
    // The DMA flag must be set when the device will be performing DMA directly
    // to the data buffer addresses passed in to the ReceiceDataPacket routines.
    //

	pSrb->StreamObject->Dma = FALSE;

    //
    // The PIO flag must be set when the mini driver will be accessing the data
    // buffers passed in using logical addressing
    //

	pSrb->StreamObject->Pio = TRUE;

}

/*
** AdapterCloseStream()
**
**   Close the requested data stream
**
** Arguments:
**
**   pSrb the request block requesting to close the stream
**
** Returns:
**
** Side Effects:  none
*/

VOID AdapterCloseStream(PHW_STREAM_REQUEST_BLOCK pSrb)
{
    DEBUG_BREAKPOINT();

    //
    // the minidriver may wish to free any resources that were allocate at
    // open stream time etc.
    //
}
         

/*
** AudioReceiveDataPacket()
**
**   Receives Audio data packet commands
**
** Arguments:
**
**   pSrb - Stream request block for the Audio device
**
** Returns:
**
** Side Effects:  none
*/

VOID STREAMAPI AudioReceiveDataPacket(IN PHW_STREAM_REQUEST_BLOCK pSrb)
{
	PHW_DEVICE_EXTENSION pdevext =
	 ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);

    //
    // make sure we have a device extension
    //

/*

 This minidriver will only request one request at a time.  Thus, it does not
 have to queue commands.  However, If the mini driver  is waiting for an
 interrupt, or timer to complete a data request, the mini-driver may receive
 a control request. (Although it will not be re-entered)

 If the mini-driver wishes to receive more than one command at a time, it may
 issue another ReadyForNextRequest notification at any time

 */

	//
	// determine the type of packet.
	//

	switch (pSrb->Command)
	{
	case SRB_WRITE_DATA:

        TRAP

        //
        // set the current command
        //

        pdevext->AudioDeviceExt.pCurrentSRB = pSrb;

        //
        // handle the data packet with the correct extension data handler
        //

        StartDataPacket(pSrb);

		break;

	default:

		//
		// invalid / unsupported command. Fail it as such
		//

		TRAP

		pSrb->Status = STATUS_NOT_IMPLEMENTED;
	
		StreamClassStreamNotification(ReadyForNextStreamDataRequest,
				pSrb->StreamObject);
	
		StreamClassStreamNotification(StreamRequestComplete,
				pSrb->StreamObject,
				pSrb);


	}

/*

 NOTE: since the write data command packets are handled asyncronously, they
 will not be called back in this entry point

 */

}

/*
** VideoReceiveDataPacket()
**
**   Receives video data packets
**
** Arguments:
**
**   pSrb - Stream request block for the video device
**
** Returns:
**
** Side Effects:  none
*/

VOID STREAMAPI VideoReceiveDataPacket(IN PHW_STREAM_REQUEST_BLOCK pSrb)
{
	PHW_DEVICE_EXTENSION pdevext =
	 ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);

    //
    // make sure we have a device extension
    //

/*

 This minidriver will only request one request at a time.  Thus, it does not
 have to queue commands.  However, If the mini driver  is waiting for an
 interrupt, or timer to complete a data request, the mini-driver may receive
 a control request. (Although it will not be re-entered)

 If the mini-driver wishes to receive more than one command at a time, it may
 issue another ReadyForNextRequest notification at any time

 */

	//
	// determine the type of packet.
	//

	switch (pSrb->Command)
	{
	case SRB_WRITE_DATA:

        TRAP

        //
        // set the current command
        //

        pdevext->VideoDeviceExt.pCurrentSRB = pSrb;

        //
        // handle the data packet with the correct extension data handler
        //

        StartDataPacket(pSrb);

		break;

	default:

		//
		// invalid / unsupported command. Fail it as such
		//

		TRAP

		pSrb->Status = STATUS_NOT_IMPLEMENTED;
	
		StreamClassStreamNotification(ReadyForNextStreamDataRequest,
				pSrb->StreamObject);
	
		StreamClassStreamNotification(StreamRequestComplete,
				pSrb->StreamObject,
				pSrb);


	}

/*

 NOTE: since the write data command packets are handled asyncronously, they
 will not be called back in this entry point

 */

}


/*
** VideoReceiveCtrlPacket()
**
**   Receives packet commands that control the video stream
**
** Arguments:
**
**   pSrb - The stream request block for the video stream
**
** Returns:
**
** Side Effects:  none
*/

VOID STREAMAPI VideoReceiveCtrlPacket(IN PHW_STREAM_REQUEST_BLOCK pSrb)
{
	PHW_DEVICE_EXTENSION pdevext =
	 ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);

	//
	// determine the type of packet.
	//

	switch (pSrb->Command)
	{
		case SRB_SET_STREAM_STATE:

            SetVideoState(pSrb);

			break;

		case SRB_GET_STREAM_PROPERTY:

            GetVideoProperty(pSrb);

            break;

		default:

			//
			// invalid / unsupported command. Fail it as such
			//
	
			TRAP
	
			pSrb->Status = STATUS_NOT_IMPLEMENTED;
		
	}

	StreamClassStreamNotification(
			ReadyForNextStreamControlRequest,
			pSrb->StreamObject);

	StreamClassStreamNotification(StreamRequestComplete,
			pSrb->StreamObject,
			pSrb);
}

/*
** AudioReceiveCtrlPacket()
**
**   Receives packet commands that control the Audio stream
**
** Arguments:
**
**   pSrb - The stream request block for the Audio stream
**
** Returns:
**
** Side Effects:  none
*/

VOID STREAMAPI AudioReceiveCtrlPacket(IN PHW_STREAM_REQUEST_BLOCK pSrb)
{
	PHW_DEVICE_EXTENSION pdevext =
	 ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);

	//
	// determine the type of packet.
	//

	switch (pSrb->Command)
	{
		case SRB_SET_STREAM_STATE:

            SetAudioState(pSrb);

			break;

		case SRB_GET_STREAM_PROPERTY:

            GetAudioProperty(pSrb);

            break;

		default:

			//
			// invalid / unsupported command. Fail it as such
			//
	
			TRAP
	
			pSrb->Status = STATUS_NOT_IMPLEMENTED;
		
	}

	StreamClassStreamNotification(
			ReadyForNextStreamControlRequest,
			pSrb->StreamObject);

	StreamClassStreamNotification(StreamRequestComplete,
			pSrb->StreamObject,
			pSrb);
}


			


