/*++

Copyright (c) 1995  Microsoft Corporation

Module Name:

   ioctl.c

Abstract:

    USB device driver for Intel 82930 USB test board

Environment:

    kernel mode only

Notes:

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
  PURPOSE.

  Copyright (c) 1996 Microsoft Corporation.  All Rights Reserved.


Revision History:

    5-4-96 : created

--*/

#define DRIVER

#include "wdm.h"
#include "stdarg.h"
#include "stdio.h"

#include "usbdi.h"
#include "usbdlib.h"
#include "I82930.h"

#include "ioctl.h"
#include "usbdlib.h"


// local routine to calculate bytes per second
ULONG
I82930_BytesPerSecond(PI82930_PIPE pipe)
{
	ULONG			Remainder;
	LARGE_INTEGER	Result;
	LARGE_INTEGER	TicksPerSecond;
	
	// get ticks per second from perf counter
	KeQueryPerformanceCounter(&TicksPerSecond);

	// scale the bytes xfered
	Result = RtlExtendedIntegerMultiply(pipe->BytesXfered,
										TicksPerSecond.LowPart);
	
	// Don't divide by 0
	pipe->ElapsedTime.LowPart = (pipe->ElapsedTime.LowPart == 0L) ? 1 : 
								pipe->ElapsedTime.LowPart;

	// lets get stats here
	Result = RtlExtendedLargeIntegerDivide(Result,
										   pipe->ElapsedTime.LowPart,
										   &Remainder);

	return Result.LowPart;
}

// local routine to calculate clock cycles per byte
ULONG
I82930_ClockCyclesPerByte(PI82930_PIPE pipe)
{
#ifdef _X86_
	ULONG			Remainder;
	LARGE_INTEGER	Result;

	// calculate cycles per byte if bytes xfered is not 0
	if(pipe->BytesXfered.LowPart != 0)
		Result = RtlExtendedLargeIntegerDivide(pipe->ElapsedCycles,
											   (ULONG) pipe->BytesXfered.LowPart,
											   &Remainder);
	// bytes xfered is 0, so return 0
	else
		Result.LowPart = 0L;

	return Result.LowPart;
#else
	return 0L;
#endif
}



PUSB_CONFIGURATION_DESCRIPTOR
I82930_GetConfigDescriptor(
    IN PDEVICE_OBJECT DeviceObject
    )
/*++

Routine Description:

Arguments:

    DeviceObject - pointer to the device object for this instance of the 82930
                    devcice.


Return Value:

    NT status code

--*/
{
    PDEVICE_EXTENSION deviceExtension;
    NTSTATUS ntStatus;
    PURB urb;
    ULONG siz;
    PUSB_CONFIGURATION_DESCRIPTOR configurationDescriptor = NULL;

    I82930_KdPrint (("'I82930.SYS: enter I82930_GetConfigDescriptor\n"));

    deviceExtension = DeviceObject->DeviceExtension;

    urb = ExAllocatePool(NonPagedPool,
                         sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST));

    if (urb) {

        // BUGBUG 82930 chokes if on the next command if you don't get
        // the entire descriptor on the first try

        siz = sizeof(USB_CONFIGURATION_DESCRIPTOR)+256;

get_config_descriptor_retry2:

        configurationDescriptor = ExAllocatePool(NonPagedPool,
                                                 siz);

        if (configurationDescriptor) {

            UsbBuildGetDescriptorRequest(urb,
                                         (USHORT) sizeof (struct _URB_CONTROL_DESCRIPTOR_REQUEST),
                                         USB_CONFIGURATION_DESCRIPTOR_TYPE,
                                         0,
                                         0,
                                         configurationDescriptor,
                                         NULL,
                                         siz,
                                         NULL);

            ntStatus = I82930_CallUSBD(DeviceObject, urb);

            I82930_KdPrint (("'I82930.SYS: Configuration Descriptor = %x, len %x\n",
                            configurationDescriptor,
                            urb->UrbControlDescriptorRequest.TransferBufferLength));
        } else {
            ntStatus = STATUS_INSUFFICIENT_RESOURCES;
        }

        if (NT_SUCCESS(ntStatus) &&
            (urb->UrbControlDescriptorRequest.TransferBufferLength >= 
             sizeof(USB_CONFIGURATION_DESCRIPTOR)) &&
            (configurationDescriptor->wTotalLength >=
             sizeof(USB_CONFIGURATION_DESCRIPTOR)))
        {
            //
            // The Get Config Descriptor request did not return an error
            // AND at least enough data was transferred to fill a Config
            // Descriptor AND the Config Descriptor wLength is at least the
            // size of a Config Descriptor
            //
            if (configurationDescriptor->wTotalLength > siz)
            {
                //
                // The request buffer is not big enough to hold the
                // entire set of descriptors.  Free the current buffer
                // and retry with a buffer which should be big enough.
                //
                siz = configurationDescriptor->wTotalLength;
                ExFreePool(configurationDescriptor);
                configurationDescriptor = NULL;
                goto get_config_descriptor_retry2;
            }
            else if (configurationDescriptor->wTotalLength >
                     urb->UrbControlDescriptorRequest.TransferBufferLength)
            {
                //
                // The request buffer is greater than or equal to the
                // Config Descriptor wLength, but less data was transferred
                // than wLength.  Return NULL to indicate a device error.
                //
                ExFreePool(configurationDescriptor);
                configurationDescriptor = NULL;
            }
            //
            // else  everything is OK with the Config Descriptor, return it.
            //
        }
        else
        {
            //
            // The Get Config Descriptor request returned an error OR
            // not enough data was transferred to fill a Config Descriptor
            // OR the Config Descriptor wLength is less than the size of
            // a Config Descriptor.  Return NULL to indicate a device error.
            //
            ExFreePool(configurationDescriptor);
            configurationDescriptor = NULL;
        }

        ExFreePool(urb);

    } else {
        ntStatus = STATUS_INSUFFICIENT_RESOURCES;
    }

    I82930_KdPrint (("'I82930.SYS: exit I82930_GetConfigDescriptor\n"));

    return configurationDescriptor;
}


NTSTATUS
I82930_GetPipeInfo(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*++

Routine Description:

Arguments:

    DeviceObject - pointer to the device object for this instance of the 82930
                    devcice.


Return Value:

    NT status code

--*/
{
    PI82930_INTERFACE_INFO interfaceInfo;
    UNICODE_STRING unicodePipeName;
    ANSI_STRING ansiPipeName;
    ULONG pipeCount = 0;
    PIO_STACK_LOCATION irpStack;
    PVOID ioBuffer;
    ULONG inputBufferLength, i, length;
    ULONG outputBufferLength;
    PDEVICE_EXTENSION deviceExtension;

    irpStack = IoGetCurrentIrpStackLocation (Irp);

    deviceExtension = DeviceObject->DeviceExtension;

    ioBuffer           = Irp->AssociatedIrp.SystemBuffer;
    inputBufferLength  = irpStack->Parameters.DeviceIoControl.InputBufferLength;
    outputBufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;


    //
    // inputs  - none
    // outputs - a list of pipes available for io testing
    //

    interfaceInfo = ioBuffer;

    I82930_KdPrint (("'I82930.SYS: info buffer = 0x%x, pipelist 0x%x\n", interfaceInfo,
        &deviceExtension->PipeList[0]));

    for (i=0; i<I82930_MAX_PIPES; i++) {
        PI82930_PIPE pipe;

        pipe = &deviceExtension->PipeList[i];

        if (pipe->PipeInfo == NULL) {
            continue;
        }

        switch(pipe->PipeInfo->PipeType) {
        case UsbdPipeTypeInterrupt:
            interfaceInfo->Pipes[pipeCount].PipeType = INTERRUPT;
            break;
        case UsbdPipeTypeBulk:
            interfaceInfo->Pipes[pipeCount].PipeType = BULK;
            break;
        case UsbdPipeTypeIsochronous:
            interfaceInfo->Pipes[pipeCount].PipeType = ISO;
            break;
        case UsbdPipeTypeControl:
            interfaceInfo->Pipes[pipeCount].PipeType = CONTROL;
            break;
        default:
                TRAP();
        }

        interfaceInfo->Pipes[pipeCount].EndpointAddress =
            pipe->PipeInfo->EndpointAddress;

        interfaceInfo->Pipes[pipeCount].MaximumPacketSize =
            pipe->PipeInfo->MaximumPacketSize;

        interfaceInfo->Pipes[pipeCount].MaximumTransferSize =
            pipe->PipeInfo->MaximumTransferSize;

        interfaceInfo->Pipes[pipeCount].Interval =
            pipe->PipeInfo->Interval;

        interfaceInfo->Pipes[pipeCount].In =
            pipe->PipeInfo->EndpointAddress & 0x80;

        //
        // copy the name
        //

        RtlCopyMemory(interfaceInfo->Pipes[pipeCount].Name,
                      "\\PIPE00",
                      8);

        RtlInitUnicodeString(&unicodePipeName,
                             pipe->Name);

        RtlInitAnsiString(&ansiPipeName, interfaceInfo->Pipes[pipeCount].Name);

        RtlUnicodeStringToAnsiString(&ansiPipeName, &unicodePipeName, FALSE);

		// get bytes per second here
		interfaceInfo->Pipes[pipeCount].PerfInfo.BytesPerSecond =
					I82930_BytesPerSecond(pipe);

		// get clock cycles per byte
		interfaceInfo->Pipes[pipeCount].PerfInfo.ClockCyclesPerByte =
					I82930_ClockCyclesPerByte(pipe);
		
        pipeCount++;
    }

    interfaceInfo->PipeCount = pipeCount;

    length = sizeof(I82930_INTERFACE_INFO) + sizeof(I82930_PIPE_INFO) *
                    interfaceInfo->PipeCount;

    Irp->IoStatus.Information = length;
    Irp->IoStatus.Status = STATUS_SUCCESS;

    return STATUS_SUCCESS;
}

typedef enum _CLOCKMASTER_OP {
    TakeControl,
    FreeControl,
    ChangeClock,
    GetClock
} CLOCKMASTER_OP;

NTSTATUS
I82930_ClockMaster(
    IN PDEVICE_OBJECT DeviceObject,
    IN CLOCKMASTER_OP Op
    )
/*++

Routine Description:

    modifies the USB SOF clock

Arguments:

Return Value:

    STATUS_SUCCESS if successful,
    STATUS_UNSUCCESSFUL otherwise

--*/
{
    NTSTATUS ntStatus;
    PURB urb;
    USHORT siz;
    USHORT func;

    I82930_KdPrint (("'I82930.SYS: ClockMaster\n"));

    switch(Op) {
    case TakeControl:
        func = URB_FUNCTION_TAKE_FRAME_LENGTH_CONTROL;
        siz = sizeof(struct _URB_FRAME_LENGTH_CONTROL);
        break;
    case FreeControl:
        func = URB_FUNCTION_RELEASE_FRAME_LENGTH_CONTROL;
        siz = sizeof(struct _URB_FRAME_LENGTH_CONTROL);
        break;
    case ChangeClock:
        siz = sizeof(struct _URB_SET_FRAME_LENGTH);
        func = URB_FUNCTION_SET_FRAME_LENGTH; 
        break;
    case GetClock:
        siz = sizeof(struct _URB_GET_FRAME_LENGTH);
        func = URB_FUNCTION_GET_FRAME_LENGTH;
        break;
    }
    
    urb = ExAllocatePool(NonPagedPool,
                         sizeof(struct _URB_PIPE_REQUEST));

    if (urb) {

        switch(Op) {
        case TakeControl:
        case FreeControl:
        case GetClock:
            break;
        case ChangeClock:
            urb->UrbSetFrameLength.FrameLengthDelta = 1;
            break;
        }

        urb->UrbHeader.Length = siz;
        urb->UrbHeader.Function = func;

        ntStatus = I82930_CallUSBD(DeviceObject, urb);

        switch(Op) {
        case TakeControl:
            I82930_KdPrint (("'I82930.SYS: TakeControl\n"));
            break;
        case FreeControl:
            I82930_KdPrint (("'I82930.SYS: FreeControl\n"));
            break;
        case GetClock:
            I82930_KdPrint (("'I82930.SYS: GetClock %d\n", urb->UrbGetFrameLength.FrameLength));
            break;
        case ChangeClock:
            I82930_KdPrint (("'I82930.SYS: ChangeClock\n"));
            break;
        }

        I82930_KdPrint (("'I82930.SYS: clock op status = %x urb = %x\n", ntStatus, 
            urb->UrbHeader.Status));

        ExFreePool(urb);

    } else {
        ntStatus = STATUS_INSUFFICIENT_RESOURCES;
    }

    return ntStatus;
}


NTSTATUS
I82930_GetPortStatus(
    IN PDEVICE_OBJECT DeviceObject,
    IN PULONG PortStatus
    )
/*++

Routine Description:

    returns the port status for our device

Arguments:

Return Value:

    STATUS_SUCCESS if successful,
    STATUS_UNSUCCESSFUL otherwise

--*/
{
    NTSTATUS ntStatus, status = STATUS_SUCCESS;
    PIRP irp;
    KEVENT event;
    IO_STATUS_BLOCK ioStatus;
    PIO_STACK_LOCATION nextStack;
    PDEVICE_EXTENSION deviceExtension;

    I82930_KdPrint (("'I82930.SYS: enter I82930_GetPortStatus\n"));

    deviceExtension = DeviceObject->DeviceExtension;

    *PortStatus = 0;

    //
    // issue a synchronous request
    //

    KeInitializeEvent(&event, NotificationEvent, FALSE);

    irp = IoBuildDeviceIoControlRequest(
                IOCTL_INTERNAL_USB_GET_PORT_STATUS,
                deviceExtension->TopOfStackDeviceObject,
                NULL,
                0,
                NULL,
                0,
                TRUE, /* INTERNAL */
                &event,
                &ioStatus);

    //
    // Call the class driver to perform the operation.  If the returned status
    // is PENDING, wait for the request to complete.
    //

    nextStack = IoGetNextIrpStackLocation(irp);
    ASSERT(nextStack != NULL);

    nextStack->Parameters.Others.Argument1 = PortStatus;

    I82930_KdPrint (("'I82930.SYS: calling USBD port status api\n"));

    ntStatus = IoCallDriver(deviceExtension->TopOfStackDeviceObject,
                            irp);

    I82930_KdPrint (("'I82930.SYS: return from IoCallDriver USBD %x\n", ntStatus));

    if (ntStatus == STATUS_PENDING) {

        I82930_KdPrint (("'I82930.SYS: Wait for single object\n"));

        status = KeWaitForSingleObject(
                       &event,
                       Suspended,
                       KernelMode,
                       FALSE,
                       NULL);

        I82930_KdPrint (("'I82930.SYS: Wait for single object, returned %x\n", status));
        
    } else {
        ioStatus.Status = ntStatus;
    }

    I82930_KdPrint (("'I82930.SYS: Port status = %x\n", *PortStatus));

    //
    // USBD maps the error code for us
    //
    ntStatus = ioStatus.Status;

    I82930_KdPrint (("'I82930.SYS: I82930_GetPortStatus (%x)\n", ntStatus));

    return ntStatus;
}


NTSTATUS
I82930_ResetParentPort(
    IN IN PDEVICE_OBJECT DeviceObject
    )
/*++

Routine Description:

    Reset the our parent port

Arguments:

Return Value:

    STATUS_SUCCESS if successful,
    STATUS_UNSUCCESSFUL otherwise

--*/
{
    NTSTATUS ntStatus, status = STATUS_SUCCESS;
    PIRP irp;
    KEVENT event;
    IO_STATUS_BLOCK ioStatus;
    PIO_STACK_LOCATION nextStack;
    PDEVICE_EXTENSION deviceExtension;

    I82930_KdPrint (("'I82930.SYS: enter I82930_ResetPort\n"));

    deviceExtension = DeviceObject->DeviceExtension;

    //
    // issue a synchronous request
    //

    KeInitializeEvent(&event, NotificationEvent, FALSE);

    irp = IoBuildDeviceIoControlRequest(
                IOCTL_INTERNAL_USB_RESET_PORT,
                deviceExtension->TopOfStackDeviceObject,
                NULL,
                0,
                NULL,
                0,
                TRUE, /* INTERNAL */
                &event,
                &ioStatus);

    //
    // Call the class driver to perform the operation.  If the returned status
    // is PENDING, wait for the request to complete.
    //

    nextStack = IoGetNextIrpStackLocation(irp);
    ASSERT(nextStack != NULL);

    I82930_KdPrint (("'I82930.SYS: calling USBD enable port api\n"));

    ntStatus = IoCallDriver(deviceExtension->TopOfStackDeviceObject,
                            irp);
                            
    I82930_KdPrint (("'I82930.SYS: return from IoCallDriver USBD %x\n", ntStatus));

    if (ntStatus == STATUS_PENDING) {

        I82930_KdPrint (("'I82930.SYS: Wait for single object\n"));

        status = KeWaitForSingleObject(
                       &event,
                       Suspended,
                       KernelMode,
                       FALSE,
                       NULL);

        I82930_KdPrint (("'I82930.SYS: Wait for single object, returned %x\n", status));
        
    } else {
        ioStatus.Status = ntStatus;
    }

    //
    // USBD maps the error code for us
    //
    ntStatus = ioStatus.Status;

    I82930_KdPrint (("'I82930.SYS: I82930_ResetPort (%x)\n", ntStatus));

    return ntStatus;
}


NTSTATUS
I82930_ProcessIOCTL(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*++

Routine Description:

Arguments:

    DeviceObject - pointer to the device object for this instance of the 82930
                    devcice.


Return Value:

    NT status code

--*/
{
    PIO_STACK_LOCATION irpStack;
    PVOID ioBuffer;
    ULONG inputBufferLength;
    ULONG outputBufferLength;
    PDEVICE_EXTENSION deviceExtension;
    ULONG ioControlCode;
    NTSTATUS ntStatus;
    ULONG length;
    PUCHAR pch;
    PUSB_CONFIGURATION_DESCRIPTOR configurationDescriptor;

    I82930_KdPrint (("'I82930.SYS: IRP_MJ_DEVICE_CONTROL\n"));

    I82930_IncrementIoCount(DeviceObject);

    //
    // Get a pointer to the current location in the Irp. This is where
    //     the function codes and parameters are located.
    //

    deviceExtension = DeviceObject->DeviceExtension;
    
    if (deviceExtension->AcceptingRequests == FALSE) {
        ntStatus = STATUS_DEVICE_DATA_ERROR;
        Irp->IoStatus.Status = ntStatus;
        Irp->IoStatus.Information = 0;

        IoCompleteRequest (Irp,
                           IO_NO_INCREMENT
                          );

        I82930_DecrementIoCount(DeviceObject);                          
        return ntStatus;
    }

    irpStack = IoGetCurrentIrpStackLocation (Irp);

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    ioBuffer           = Irp->AssociatedIrp.SystemBuffer;
    inputBufferLength  = irpStack->Parameters.DeviceIoControl.InputBufferLength;
    outputBufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;

    ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;

    //
    // Handle Ioctls from User mode
    //

    switch (ioControlCode) {

    case IOCTL_I82930_RESET_PIPE:
        {
        PI82930_PIPE pipe;
	    PFILE_OBJECT fileObject;

		// get our context and see if it is a pipe
	    fileObject = irpStack->FileObject;

		pipe = (PI82930_PIPE) fileObject->FsContext;    

		if(pipe == NULL) {
			// error, this is not a pipe
	        ntStatus =
		        Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
		} else {            
            I82930_ResetPipe(DeviceObject, pipe, TRUE);
            
            ntStatus = Irp->IoStatus.Status = STATUS_SUCCESS;
        }
        }
        break;

    case IOCTL_I82930_SET_PIPE_PARAMETER:

        break;

    case IOCTL_I82930_GET_PIPE_INFO:

        ntStatus = I82930_GetPipeInfo(DeviceObject, Irp);

        break;

     case IOCTL_I82930_GET_CONFIG_DESCRIPTOR:

        //
        // This api returns a copy of the configuration descriptor
        // and all endpoint/interface descriptors.
        //

        //
        // inputs  - none
        // outputs - configuration descriptor plus interface
        //          and endpoint descriptors
        //

        pch = (PUCHAR) ioBuffer;

        configurationDescriptor =
            I82930_GetConfigDescriptor(DeviceObject);

        if (configurationDescriptor) {
            
            length = configurationDescriptor->wTotalLength;

            RtlCopyMemory(pch,
                          (PUCHAR) configurationDescriptor,
                          length);

            ExFreePool(configurationDescriptor);

            Irp->IoStatus.Information = length;

            ntStatus = Irp->IoStatus.Status = STATUS_SUCCESS;
        }
        else {

            Irp->IoStatus.Information = 0;

            ntStatus = Irp->IoStatus.Status = STATUS_DEVICE_DATA_ERROR;
        }

        break;

     case IOCTL_I82930_START_ISO_STREAM:

        {
        PI82930_PIPE pipe;

        // BUGBUG
        // hard code to first pipe
        //
        pipe = &deviceExtension->PipeList[0];

        ntStatus = I82930_StartIsoStream(DeviceObject,
                                         pipe->PipeInfo,
                                         Irp);

        }
        break;

      case IOCTL_I82930_REGISTER_NOTIFY_EVENT:
        
        {
        PHANDLE Event;
        NTSTATUS status;
            
        I82930_KdPrint (("'I82930.SYS: DeviceIoControl Event Test\n"));

        Event = ioBuffer;
        status =
              ObReferenceObjectByHandle( *Event,
                                          0, 
                                          NULL, 
                                          KernelMode, 
                                          &deviceExtension->UserEvent, 
                                          NULL ) ;

        TRAP();
        }
        break;        

    case IOCTL_I82930_STOP_ISO_STREAM:

        ntStatus = I82930_StopIsoStream(DeviceObject,
                                        *((PVOID *) ioBuffer),
                                        Irp);


        break;

    case IOCTL_I82930_START_PERF_TIMER:

        {
	        PI82930_PIPE pipe;
		    PFILE_OBJECT fileObject;

			// get our context and see if it is a pipe
		    fileObject = irpStack->FileObject;

			pipe = (PI82930_PIPE) fileObject->FsContext;    

			if(pipe == NULL)
			{
				// error, this is not a pipe
		        ntStatus =
    		        Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
			}
			else
			{
				// turn on perf timing
				pipe->bPerfTimerEnabled = TRUE;

				// reset BytesXfered, ElapsedTime and TimerStart
				pipe->BytesXfered   = RtlConvertUlongToLargeInteger(0L);
				pipe->ElapsedTime   = RtlConvertUlongToLargeInteger(0L);
				pipe->TimerStart    = RtlConvertUlongToLargeInteger(0L);
				pipe->ElapsedCycles = RtlConvertUlongToLargeInteger(0L);
				pipe->CyclesStart   = RtlConvertUlongToLargeInteger(0L);
		    }
		}
        break;

     case IOCTL_I82930_STOP_PERF_TIMER:

        {
	        PI82930_PIPE pipe;
		    PFILE_OBJECT fileObject;

			// get our context and see if it is a pipe
		    fileObject = irpStack->FileObject;

			pipe = (PI82930_PIPE) fileObject->FsContext;    

			if(pipe == NULL)
			{
				// error, this is not a pipe
		        ntStatus =
    		        Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
			}
			else
			{
				// turn off perf timing
				pipe->bPerfTimerEnabled = FALSE;
			}
    	}
        break;

     case IOCTL_I82930_RETURN_PERF_DATA:

        {
	        PI82930_PIPE pipe;
		    PFILE_OBJECT fileObject;

			// get our context and see if it is a pipe
		    fileObject = irpStack->FileObject;

			pipe = (PI82930_PIPE) fileObject->FsContext;    

			if(pipe == NULL)
			{
				// error, this is not a pipe
		        ntStatus =
    		        Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
			}
			else
			{
				// make sure we have enoug space to return perf info
				if(outputBufferLength < sizeof(I82930_PIPE_PERF_INFO))
				{
			        ntStatus =
	    		        Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
				}
				else
				{
					PI82930_PIPE_PERF_INFO	pPerf =
						(PI82930_PIPE_PERF_INFO) ioBuffer;

					pPerf->BytesPerSecond = I82930_BytesPerSecond(pipe);
					pPerf->ClockCyclesPerByte = I82930_ClockCyclesPerByte(pipe);
			        Irp->IoStatus.Information = sizeof(I82930_PIPE_PERF_INFO);
				}
			}
        }
        break;

     case IOCTL_I82930_RESET_DEVICE:
        
        {
        NTSTATUS ntStatus;
        ULONG portStatus;

        I82930_KdPrint (("'I82930.SYS: Reset Device Test\n"));
        
        TRAP(); // test this
        //
        // Check the port state, if it is disabled we will need 
        // to re-enable it
        //
        ntStatus = I82930_GetPortStatus(DeviceObject, &portStatus);

        if (NT_SUCCESS(ntStatus) && !(portStatus & USBD_PORT_ENABLED) &&
            portStatus & USBD_PORT_CONNECTED) {
            //
            // port is disabled, attempt reset
            //
            //I82930_EnableParentPort(DeviceObject);
            I82930_ResetParentPort(DeviceObject);
        }

        }
        break;               

    case IOCTL_I82930_CLOCK_MASTER_TEST:
        
        {
        NTSTATUS ntStatus = STATUS_SUCCESS;

        I82930_KdPrint (("'I82930.SYS: Clock Master Test\n"));
        
        TRAP(); // test this

        // get the clock
        I82930_ClockMaster(DeviceObject, GetClock);

        // take control to change clock
        //ntStatus = I82930_ClockMaster(DeviceObject, TakeControl);
        
        if (NT_SUCCESS(ntStatus)) {
            ntStatus = I82930_ClockMaster(DeviceObject, ChangeClock);                
        }

        I82930_ClockMaster(DeviceObject, FreeControl);                

        }
        break;               
        

    default:

        ntStatus =
            Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
    }

    IoCompleteRequest (Irp,
                       IO_NO_INCREMENT
                       );

    I82930_DecrementIoCount(DeviceObject);                       

    return ntStatus;

}





