/*++

Copyright (c) 1995  Microsoft Corporation

Module Name:

   ocrw.c

Abstract:

   read/write io test code

Environment:

    kernel mode only

Notes:

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
  PURPOSE.

  Copyright (c) 1996 Microsoft Corporation.  All Rights Reserved.


Revision History:

    5-4-96 : created

--*/

#define DRIVER

#include "wdm.h"
#include "stdarg.h"
#include "stdio.h"

#include "usbdi.h"
#include "usbdlib.h"
#include "I82930.h"

#ifdef _X86_
/************************************************************************/
/*					Debug_StartPentiumCycleCounter						*/
/************************************************************************/
/*																		*/
/* Routine Description:													*/
/*																		*/
/*  Uses pentium RDTSC instruction to get current cycle count since		*/
/*  chip was powered. This records a staring cycle count, and then		*/
/*  call the stop routine to get a cycle count.							*/
/*																		*/
/*  DANGER DANGER Will Robinson - these routines will only work when	*/
/*  used on a Pentium processor. Also, these probably should be macros	*/
/*  so subroutine calls don't skew the results. They should be close	*/
/*  enough for profiling though.										*/
/*																		*/
/* Arguments:															*/
/*																		*/
/*  cycleCount - this is where we will store the starting cycle count.	*/
/*																		*/
/* Return Value:														*/
/*																		*/
/*	VOID																*/
/*																		*/
/************************************************************************/
VOID
Debug_StartPentiumCycleCounter(PLARGE_INTEGER cycleCount)
{
	LARGE_INTEGER	startCount;

	_asm
	{
		// the RDTSC instruction is not really documented,
		// so assemblers may not support it, so use _emit to
		// create instruction
		_emit	0Fh
		_emit	31h
		mov		startCount.HighPart, edx
		mov		startCount.LowPart,  eax
	}

	// here is our starting count
	*cycleCount = startCount;
} // Debug_StartPentiumCycleCounter

/************************************************************************/
/*					Debug_StopPentiumCycleCounter						*/
/************************************************************************/
/*																		*/
/* Routine Description:													*/
/*																		*/
/*  Uses pentium RDTSC instruction to get current cycle count since		*/
/*  chip was powered. This routine returns the number of cycles			*/
/*  since the strat routine was called.									*/
/*																		*/
/*  DANGER DANGER Will Robinson - these routines will only work when	*/
/*  used on a Pentium processor. Also, these probably should be macros	*/
/*  so subroutine calls don't skew the results. They should be close	*/
/*  enough for profiling though.										*/
/*																		*/
/* Arguments:															*/
/*																		*/
/*  cycleCount - this is where we will store the cycle count since 		*/
/*               call to start routine. On entry it should contain		*/
/*               a starting cycle count obtained with a call to the		*/
/*               start routine.											*/
/*																		*/
/* Return Value:														*/
/*																		*/
/*	VOID																*/
/*																		*/
/************************************************************************/
VOID
Debug_StopPentiumCycleCounter(PLARGE_INTEGER cycleCount)
{
	LARGE_INTEGER	endCount;

	_asm
	{
		// the RDTSC instruction is not really documented,
		// so assemblers may not support it, so use _emit to
		// create instruction
		_emit	0Fh
		_emit	31h
		mov		endCount.HighPart, edx
		mov		endCount.LowPart,  eax
	}

	// figure up cycle count by subtracting starting count from ending count
	*cycleCount = RtlLargeIntegerSubtract(endCount, *cycleCount);
} // Debug_StopPentiumCycleCounter
#endif


// start up perf timing and cycle counter timing if enabled

VOID
StartPerfTimer(PI82930_PIPE pipeHandle)
{
	// set up perf stuff if perf timing enabled
	if(pipeHandle && pipeHandle->bPerfTimerEnabled)
	{
		// get current perf counter
		pipeHandle->TimerStart = KeQueryPerformanceCounter(NULL);
#ifdef _X86_
		// get clock cycle count
		Debug_StartPentiumCycleCounter(&pipeHandle->CyclesStart);
#endif
	}
} // StartPerfTimer

VOID
StopPerfTimer(PI82930_PIPE pipeHandle, ULONG bytesXfered)
{
	LARGE_INTEGER	BytesThisTransfer;
	LARGE_INTEGER	CurrentTime;
	LARGE_INTEGER	TimeThisTransfer;

	if(pipeHandle && pipeHandle->bPerfTimerEnabled)
	{
#ifdef _X86_
		// stop clock cycle counting
		Debug_StopPentiumCycleCounter(&pipeHandle->CyclesStart);
#endif
		// get updated time
		CurrentTime = KeQueryPerformanceCounter(NULL);

#ifdef _X86_
		// update cycle counter
		pipeHandle->ElapsedCycles = RtlLargeIntegerAdd(pipeHandle->ElapsedCycles,
													   pipeHandle->CyclesStart);
#endif		

		// stop perf timing with system timer
		BytesThisTransfer = RtlConvertUlongToLargeInteger(bytesXfered);

		pipeHandle->BytesXfered = RtlLargeIntegerAdd(pipeHandle->BytesXfered,
													 BytesThisTransfer);

		// now add the time it took to elapsed time
		TimeThisTransfer = RtlLargeIntegerSubtract(CurrentTime,
												   pipeHandle->TimerStart);

		pipeHandle->ElapsedTime = RtlLargeIntegerAdd(pipeHandle->ElapsedTime,
													 TimeThisTransfer);
	}

} // StopPerfTimer


//******************************************************************************
//
// I82930_CompletionStop()
//
// IO Completion Routine which just stops further completion of the Irp
//
//******************************************************************************

NTSTATUS
I82930_CompletionStop (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp,
    IN PVOID            Context
    )
{
    return STATUS_MORE_PROCESSING_REQUIRED;
}

//******************************************************************************
//
// I82930_GetCurrentFrame()
//
//******************************************************************************

ULONG
I82930_GetCurrentFrame (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP           Irp
    )
{
    PDEVICE_EXTENSION           deviceExtension;
    PIO_STACK_LOCATION          nextStack;
    NTSTATUS                    ntStatus;
    struct _URB_GET_CURRENT_FRAME_NUMBER urb;

    deviceExtension   = DeviceObject->DeviceExtension;

    // Initialize the URB
    //
    urb.Hdr.Function = URB_FUNCTION_GET_CURRENT_FRAME_NUMBER;
    urb.Hdr.Length   = sizeof(urb);
    urb.FrameNumber = (ULONG)-1;

    // Set the IRP parameters to pass the URB down the stack
    //
    nextStack = IoGetNextIrpStackLocation(Irp);

    nextStack->Parameters.Others.Argument1 = &urb;

    nextStack->Parameters.DeviceIoControl.IoControlCode = 
        IOCTL_INTERNAL_USB_SUBMIT_URB;                    

    nextStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;

    // Since this Irp is borrowed for URB_FUNCTION_GET_CURRENT_FRAME_NUMBER
    // before it is passed down later for the real URB request after this
    // routine returns, set a completion routine which stop further completion
    // of the Irp.
    //
    IoSetCompletionRoutine(
        Irp,
        I82930_CompletionStop,
        NULL,   // Context
        TRUE,   // InvokeOnSuccess
        TRUE,   // InvokeOnError
        TRUE    // InvokeOnCancel
        );

    // Now pass the Irp down the stack
    //
    ntStatus = IoCallDriver(
                   deviceExtension->TopOfStackDeviceObject, 
                   Irp
                   );

    // Don't need to wait for completion because JD guarantees that
    // URB_FUNCTION_GET_CURRENT_FRAME_NUMBER will never return STATUS_PENDING

    return urb.FrameNumber;
}


//******************************************************************************
//
// I82930_BuildIsoRequest()
//
//******************************************************************************

PURB
I82930_BuildIsoRequest(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PI82930_PIPE PipeHandle,
    IN BOOLEAN Read
    )
/*++

Routine Description:

Arguments:

    DeviceObject - pointer to the device extension for this instance of the
                     82930 device.

    Irp -

    PipeHandle -

Return Value:

    initialized async urb.

--*/
{
    ULONG siz;
    ULONG length, packetSize, numPackets, i;
    PURB urb = NULL;
    PIO_STACK_LOCATION irpSp;
    LARGE_INTEGER byteOffset;

    irpSp = IoGetCurrentIrpStackLocation(Irp);

    I82930_KdPrint (("'I82930.SYS: handle = 0x%x\n", PipeHandle));

    length = MmGetMdlByteCount(Irp->MdlAddress);

    I82930_KdPrint (("'I82930.SYS: length = 0x%x\n", length));

    byteOffset = irpSp->Parameters.Read.ByteOffset;

    I82930_KdPrint (("'I82930.SYS: offset = 0x%08X.%08X\n",
                     byteOffset.HighPart,
                     byteOffset.LowPart));

    packetSize = PipeHandle->PipeInfo->MaximumPacketSize;
    numPackets = length/packetSize;
    if (numPackets*packetSize < length) {
        numPackets++;
    }

    siz = GET_ISO_URB_SIZE(numPackets);
    urb = ExAllocatePool(NonPagedPool, siz);

    I82930_KdPrint (("'I82930.SYS: siz = 0x%x urb 0x%x\n", siz, urb));

    if (urb) {
        RtlZeroMemory(urb, siz);

        urb->UrbIsochronousTransfer.Hdr.Length = (USHORT) siz;
        urb->UrbIsochronousTransfer.Hdr.Function =
                    URB_FUNCTION_ISOCH_TRANSFER;
        urb->UrbIsochronousTransfer.PipeHandle =
                   PipeHandle->PipeInfo->PipeHandle;
        urb->UrbIsochronousTransfer.TransferFlags =
            Read ? USBD_TRANSFER_DIRECTION_IN : 0;

        urb->UrbIsochronousTransfer.TransferBufferMDL =
            Irp->MdlAddress;
        urb->UrbIsochronousTransfer.TransferBufferLength =
            length;

        if (byteOffset.HighPart)
        {
            urb->UrbIsochronousTransfer.StartFrame =
                I82930_GetCurrentFrame(DeviceObject, Irp) +
                byteOffset.LowPart;
        }
        else
        {
            // start sending/receiving right away
            urb->UrbIsochronousTransfer.TransferFlags |=
                USBD_START_ISO_TRANSFER_ASAP;
        }

        urb->UrbIsochronousTransfer.NumberOfPackets = numPackets;
        urb->UrbIsochronousTransfer.UrbLink = NULL;

        for (i=0; i< urb->UrbIsochronousTransfer.NumberOfPackets; i++) {
            urb->UrbIsochronousTransfer.IsoPacket[i].Offset
                        = i * packetSize;
        }

        I82930_KdPrint (("'I82930.SYS: Init iso urb Length = 0x%x buf = 0x%x\n",
            urb->UrbIsochronousTransfer.TransferBufferLength,
            urb->UrbIsochronousTransfer.TransferBuffer));
    }

    I82930_KdPrint (("'I82930.SYS: exit I82930_BuildIsoRequest\n"));

    return urb;
}


PURB
I82930_BuildAsyncRequest(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PI82930_PIPE PipeHandle,
    IN BOOLEAN Read
    )
/*++

Routine Description:

Arguments:

    DeviceObject - pointer to the device extension for this instance of the
                     82930 device.

    Irp -

    PipeHandle -

Return Value:

    initialized async urb.

--*/
{
    ULONG siz;
    ULONG length;
    PURB urb = NULL;

    I82930_KdPrint (("'I82930.SYS: handle = 0x%x\n", PipeHandle));

    length = MmGetMdlByteCount(Irp->MdlAddress);

    I82930_KdPrint (("'I82930.SYS: length = 0x%x\n", length));

    siz = sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER);
    urb = ExAllocatePool(NonPagedPool, siz);

    I82930_KdPrint (("'I82930.SYS: siz = 0x%x urb 0x%x\n", siz, urb));

    if (urb) {
        RtlZeroMemory(urb, siz);

        urb->UrbBulkOrInterruptTransfer.Hdr.Length = (USHORT) siz;
        urb->UrbBulkOrInterruptTransfer.Hdr.Function =
                    URB_FUNCTION_BULK_OR_INTERRUPT_TRANSFER;
        urb->UrbBulkOrInterruptTransfer.PipeHandle =
                   PipeHandle->PipeInfo->PipeHandle;
        urb->UrbBulkOrInterruptTransfer.TransferFlags =
            Read ? USBD_TRANSFER_DIRECTION_IN : 0;

        // short packet is not treated as an error.
        urb->UrbBulkOrInterruptTransfer.TransferFlags |= 
            USBD_SHORT_TRANSFER_OK;            
                
        //
        // no linkage for now
        //

        urb->UrbBulkOrInterruptTransfer.UrbLink = NULL;

        urb->UrbBulkOrInterruptTransfer.TransferBufferMDL =
            Irp->MdlAddress;
        urb->UrbBulkOrInterruptTransfer.TransferBufferLength =
            length;

        I82930_KdPrint (("'I82930.SYS: Init async urb Length = 0x%x buf = 0x%x\n",
            urb->UrbBulkOrInterruptTransfer.TransferBufferLength,
            urb->UrbBulkOrInterruptTransfer.TransferBuffer));
    }

    I82930_KdPrint (("'I82930.SYS: exit I82930_BuildAsyncRequest\n"));

    return urb;
}


NTSTATUS
I82930_AsyncReadWrite_Complete(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID Context
    )
/*++

Routine Description:


Arguments:

    DeviceObject - Pointer to the device object for the i82930 device.

    Irp - Irp completed.

    Context - Driver defined context.

Return Value:

    The function value is the final status from the operation.

--*/
{
    NTSTATUS			ntStatus = STATUS_SUCCESS;
    PURB				urb;
    PI82930_RW_CONTEXT  context = Context;
    PI82930_PIPE		pipeHandle;
    PFILE_OBJECT		fileObject;
    PIO_STACK_LOCATION	irpStack;
    PDEVICE_OBJECT      deviceObject;

    if (Irp->PendingReturned) {
        IoMarkIrpPending(Irp);
    }

    urb = context->Urb;
    deviceObject = context->DeviceObject;
    
    I82930_KdPrint (("'I82930.SYS: Async Completion: Length 0x%08X, Status 0x%08X\n",
                     urb->UrbBulkOrInterruptTransfer.TransferBufferLength,
                     urb->UrbHeader.Status));

    //
    // set the length based on the TransferBufferLength
    // value in the URB
    //
    Irp->IoStatus.Information =
        urb->UrbBulkOrInterruptTransfer.TransferBufferLength;

    irpStack = IoGetCurrentIrpStackLocation (Irp);
    fileObject = irpStack->FileObject;

	// get pipe handle
	pipeHandle = fileObject->FsContext;

	// stop perf timer
	StopPerfTimer(pipeHandle,
				  urb->UrbBulkOrInterruptTransfer.TransferBufferLength);

    I82930_DecrementIoCount(deviceObject);                       

    ExFreePool(context);
    ExFreePool(urb);

    return ntStatus;
}


NTSTATUS
I82930_IsoReadWrite_Complete(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID Context
    )
/*++

Routine Description:


Arguments:

    DeviceObject - Pointer to the device object for the i82930 device.

    Irp - Irp completed.

    Context - Driver defined context.

Return Value:

    The function value is the final status from the operation.

--*/
{
    NTSTATUS			ntStatus = STATUS_SUCCESS;
    PURB				urb = Context;
    PI82930_PIPE		pipeHandle;
    PFILE_OBJECT		fileObject;
    PIO_STACK_LOCATION	irpStack;
    ULONG               i;

    if (Irp->PendingReturned) {
        IoMarkIrpPending(Irp);
    }

    //
    // BUGBUG check here for interesting iso error conditions
    //
    I82930_KdPrint (("'I82930.SYS: Iso Completion: StartFrame 0x%08X, Status 0x%08X\n",
                     urb->UrbIsochronousTransfer.StartFrame,
                     urb->UrbHeader.Status));

    for (i=0; i< urb->UrbIsochronousTransfer.NumberOfPackets; i++)
    {
        I82930_KdPrint (("'I82930.SYS: [%02d] Length 0x%08X, Status 0x%08X\n",
                         i,
                         urb->UrbIsochronousTransfer.IsoPacket[i].Length,
                         urb->UrbIsochronousTransfer.IsoPacket[i].Status
                        ));
    }

    //
    // set the length based on the TransferBufferLength
    // value in the URB
    //
    Irp->IoStatus.Information =
        urb->UrbBulkOrInterruptTransfer.TransferBufferLength;

    irpStack = IoGetCurrentIrpStackLocation (Irp);
    fileObject = irpStack->FileObject;

	// get pipe handle
	pipeHandle = fileObject->FsContext;

	// stop perf timer
	StopPerfTimer(pipeHandle,
				  urb->UrbBulkOrInterruptTransfer.TransferBufferLength);

    I82930_DecrementIoCount(DeviceObject);                       

    ExFreePool(urb);

    return ntStatus;
}


NTSTATUS
I82930_Read(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*++

Routine Description:

Arguments:

    DeviceObject - pointer to the device object for this instance of the 82930
                    devcice.


Return Value:

    NT status code

--*/
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PI82930_PIPE pipeHandle = NULL;
    PFILE_OBJECT fileObject;
    PIO_STACK_LOCATION irpStack, nextStack;
    PDEVICE_EXTENSION deviceExtension;
    PURB urb;
    PI82930_RW_CONTEXT context = NULL;

    I82930_KdPrint (("'I82930.SYS: enter I82930_Read\n"));

    I82930_IncrementIoCount(DeviceObject);

    deviceExtension = DeviceObject->DeviceExtension;

    if (deviceExtension->AcceptingRequests == FALSE) {
        ntStatus = STATUS_DELETE_PENDING;
        Irp->IoStatus.Status = ntStatus;
        Irp->IoStatus.Information = 0;

        IoCompleteRequest (Irp,
                           IO_NO_INCREMENT
                          );

        I82930_DecrementIoCount(DeviceObject);                          
        return ntStatus;
    }
    
    irpStack = IoGetCurrentIrpStackLocation (Irp);
    fileObject = irpStack->FileObject;

    pipeHandle =  fileObject->FsContext;

    if (!pipeHandle) {
       ntStatus = STATUS_INVALID_HANDLE;
       goto I82930_Read_Reject;
    }

    //
    // submit the write request to USB
    //

    switch (pipeHandle->PipeInfo->PipeType) {
    case UsbdPipeTypeIsochronous:
        I82930_ResetPipe(DeviceObject, pipeHandle, FALSE);
        
        urb = I82930_BuildIsoRequest(DeviceObject,
                                     Irp,
                                     pipeHandle,
                                     TRUE);
        if (urb) {

            nextStack = IoGetNextIrpStackLocation(Irp);
            ASSERT(nextStack != NULL);
            ASSERT(DeviceObject->StackSize>1);

            nextStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
            nextStack->Parameters.Others.Argument1 = urb;
            nextStack->Parameters.DeviceIoControl.IoControlCode =
                IOCTL_INTERNAL_USB_SUBMIT_URB;

            IoSetCompletionRoutine(Irp,
                                   I82930_IsoReadWrite_Complete,
                                   urb,
                                   TRUE,
                                   TRUE,
                                   TRUE);

            I82930_KdPrint (("'I82930.SYS: IRP = 0x%x current = 0x%x next = 0x%x\n",
                Irp, irpStack, nextStack));

			// start perf timer here if needed
			StartPerfTimer(pipeHandle);

            ntStatus = IoCallDriver(deviceExtension->TopOfStackDeviceObject,
                                    Irp);
            goto I82930_Read_Done;
        } else {
            ntStatus = STATUS_INSUFFICIENT_RESOURCES;
        }
        break;
    case UsbdPipeTypeInterrupt:
    case UsbdPipeTypeBulk:
        urb = I82930_BuildAsyncRequest(DeviceObject,
                                       Irp,
                                       pipeHandle,
                                       TRUE);
        if (urb) {
            context = ExAllocatePool(NonPagedPool, sizeof(I82930_RW_CONTEXT));
        }
        
        if (urb && context) {
            context->Urb = urb;
            context->DeviceObject = DeviceObject;
            
            nextStack = IoGetNextIrpStackLocation(Irp);
            ASSERT(nextStack != NULL);
            ASSERT(DeviceObject->StackSize>1);

            nextStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
            nextStack->Parameters.Others.Argument1 = urb;
            nextStack->Parameters.DeviceIoControl.IoControlCode =
                IOCTL_INTERNAL_USB_SUBMIT_URB;

            IoSetCompletionRoutine(Irp,
                                   I82930_AsyncReadWrite_Complete,
                                   context,
                                   TRUE,
                                   TRUE,
                                   TRUE);

            I82930_KdPrint (("'I82930.SYS: IRP = 0x%x current = 0x%x next = 0x%x\n",
                Irp, irpStack, nextStack));

			// start perf timer here if needed
			StartPerfTimer(pipeHandle);

            ntStatus = IoCallDriver(deviceExtension->TopOfStackDeviceObject,
                                    Irp);
            goto I82930_Read_Done;
        } else {
            ntStatus = STATUS_INSUFFICIENT_RESOURCES;
        }

        break;
    default:
        ntStatus = STATUS_INVALID_PARAMETER;
        TRAP();
    }

I82930_Read_Reject:

    Irp->IoStatus.Status = ntStatus;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest (Irp,
                       IO_NO_INCREMENT
                       );

I82930_Read_Done:

    return ntStatus;
}


NTSTATUS
I82930_Write(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*++

Routine Description:

Arguments:

    DeviceObject - pointer to the device object for this instance of the 82930
                    device.


Return Value:

    NT status code

--*/
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PI82930_PIPE pipeHandle = NULL;
    PFILE_OBJECT fileObject;
    PIO_STACK_LOCATION irpStack, nextStack;
    PDEVICE_EXTENSION deviceExtension;
    PURB urb;
    PI82930_RW_CONTEXT context = NULL;

    I82930_KdPrint (("'I82930.SYS: enter I82930_Write\n"));

    I82930_IncrementIoCount(DeviceObject);

    deviceExtension = DeviceObject->DeviceExtension;

    if (deviceExtension->AcceptingRequests == FALSE) {
        ntStatus = STATUS_DELETE_PENDING;
        Irp->IoStatus.Status = ntStatus;
        Irp->IoStatus.Information = 0;

        I82930_DecrementIoCount(DeviceObject);

        IoCompleteRequest (Irp,
                           IO_NO_INCREMENT
                          );
        return ntStatus;
    }
    
    irpStack = IoGetCurrentIrpStackLocation (Irp);
    fileObject = irpStack->FileObject;

//    MmProbeAndLockPages(Irp->MdlAddress,
//                        KernelMode,
//                        IoReadAccess);

    pipeHandle =  fileObject->FsContext;
    if (!pipeHandle)
    {
       ntStatus = STATUS_INVALID_HANDLE;
       goto I82930_Write_Reject;
    }

    //
    // submit the write request to USB
    //

    switch (pipeHandle->PipeInfo->PipeType) {
    case UsbdPipeTypeIsochronous:
        I82930_ResetPipe(DeviceObject, pipeHandle, FALSE);
        
        urb = I82930_BuildIsoRequest(DeviceObject,
                                     Irp,
                                     pipeHandle,
                                     FALSE);
        if (urb) {

            nextStack = IoGetNextIrpStackLocation(Irp);
            ASSERT(nextStack != NULL);
            ASSERT(DeviceObject->StackSize>1);

            nextStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
            nextStack->Parameters.Others.Argument1 = urb;
            nextStack->Parameters.DeviceIoControl.IoControlCode =
                IOCTL_INTERNAL_USB_SUBMIT_URB;

            IoSetCompletionRoutine(Irp,
                                   I82930_IsoReadWrite_Complete,
                                   urb,
                                   TRUE,
                                   TRUE,
                                   TRUE);

            I82930_KdPrint (("'I82930.SYS: IRP = 0x%x current = 0x%x next = 0x%x\n",
                Irp, irpStack, nextStack));

			// start perf timer here if needed
			StartPerfTimer(pipeHandle);

            ntStatus = IoCallDriver(deviceExtension->TopOfStackDeviceObject,
                                    Irp);
            goto I82930_Write_Done;
        } else {
            ntStatus = STATUS_INSUFFICIENT_RESOURCES;
        }
        break;
    case UsbdPipeTypeInterrupt:
    case UsbdPipeTypeBulk:
        urb = I82930_BuildAsyncRequest(DeviceObject,
                                       Irp,
                                       pipeHandle,
                                       FALSE);

        if (urb) {
            context = ExAllocatePool(NonPagedPool, sizeof(I82930_RW_CONTEXT));
        }
        
        if (urb && context) {
            context->Urb = urb;
            context->DeviceObject = DeviceObject;                                       
            
            nextStack = IoGetNextIrpStackLocation(Irp);
            ASSERT(nextStack != NULL);
            ASSERT(DeviceObject->StackSize>1);

            nextStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
            nextStack->Parameters.Others.Argument1 = urb;
            nextStack->Parameters.DeviceIoControl.IoControlCode =
                IOCTL_INTERNAL_USB_SUBMIT_URB;

            IoSetCompletionRoutine(Irp,
                                   I82930_AsyncReadWrite_Complete,
                                   context,
                                   TRUE,
                                   TRUE,
                                   TRUE);

            I82930_KdPrint (("'I82930.SYS: IRP = 0x%x current = 0x%x next = 0x%x\n",
                Irp, irpStack, nextStack));

			// start perf timer here if needed
			StartPerfTimer(pipeHandle);

            ntStatus = IoCallDriver(deviceExtension->TopOfStackDeviceObject,
                                    Irp);
            goto I82930_Write_Done;
        } else {
            ntStatus = STATUS_INSUFFICIENT_RESOURCES;
        }

        break;
    default:
        ntStatus = STATUS_INVALID_PARAMETER;
        TRAP();
    }

I82930_Write_Reject:

    Irp->IoStatus.Status = ntStatus;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest (Irp,
                       IO_NO_INCREMENT
                       );

I82930_Write_Done:

    return ntStatus;
}


NTSTATUS
I82930_Close(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*++

Routine Description:

Arguments:

    DeviceObject - pointer to the device object for this instance of the 82930
                    devcice.


Return Value:

    NT status code

--*/
{
    NTSTATUS ntStatus;
    PFILE_OBJECT fileObject;
    PIO_STACK_LOCATION irpStack;
    PDEVICE_EXTENSION deviceExtension;
    PI82930_PIPE pipeHandle = NULL;

    I82930_KdPrint (("'I82930.SYS: entering I82930_Close\n"));

    I82930_IncrementIoCount(DeviceObject);

    deviceExtension = DeviceObject->DeviceExtension;
    irpStack = IoGetCurrentIrpStackLocation (Irp);
    fileObject = irpStack->FileObject;

    if (fileObject->FsContext) {
        // closing pipe handle
        pipeHandle =  fileObject->FsContext;
        I82930_KdPrint (("'I82930.SYS: closing pipe %x\n", pipeHandle));

        pipeHandle->Opened = FALSE;
    }

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;


    ntStatus = Irp->IoStatus.Status;

    IoCompleteRequest (Irp,
                       IO_NO_INCREMENT
                       );
                       
    I82930_DecrementIoCount(DeviceObject);

    return ntStatus;
}


NTSTATUS
I82930_Create(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*++

Routine Description:

    //
    // Entry point for CreateFile calls
    // user mode apps may open "\\.\I82930-x\yy"
    // where yy is the internal pipe id
    //

Arguments:

    DeviceObject - pointer to the device object for this instance of the 82930
                    devcice.


Return Value:

    NT status code

--*/
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PI82930_PIPE pipeHandle = NULL;
    PFILE_OBJECT fileObject;
    PIO_STACK_LOCATION irpStack;
    PDEVICE_EXTENSION deviceExtension;
    ULONG i;

    I82930_KdPrint (("'I82930.SYS: entering I82930_Create\n"));

    I82930_IncrementIoCount(DeviceObject);

    deviceExtension = DeviceObject->DeviceExtension;

    if (deviceExtension->AcceptingRequests == FALSE) {
        ntStatus = STATUS_DELETE_PENDING;
        Irp->IoStatus.Status = ntStatus;
        Irp->IoStatus.Information = 0;

        IoCompleteRequest (Irp,
                           IO_NO_INCREMENT
                          );

        I82930_DecrementIoCount(DeviceObject);                          
        
        return ntStatus;
    }
    
    irpStack = IoGetCurrentIrpStackLocation (Irp);
    fileObject = irpStack->FileObject;

    // fscontext is null for device
    fileObject->FsContext = NULL;

    if (fileObject->FileName.Length != 0) {

        ntStatus = STATUS_INSUFFICIENT_RESOURCES;

        //
        // a name was specified, convert it to a pipe id
        //

        for (i=0; i<I82930_MAX_PIPES; i++) {
            if (RtlCompareMemory (fileObject->FileName.Buffer,
                                  deviceExtension->PipeList[i].Name,
                                  fileObject->FileName.Length)
                    == fileObject->FileName.Length &&
                !deviceExtension->PipeList[i].Opened) {
                //
                // found a match
                //
                pipeHandle = &deviceExtension->PipeList[i];
                //I82930_ResetPipe(DeviceObject, pipeHandle);
                break;
            }
        }
    }

	// if we are opening a pipe set stuff up and set FsContext
    if (pipeHandle) {
        I82930_KdPrint (("'I82930.SYS: open pipe %x\n", pipeHandle));
        fileObject->FsContext = pipeHandle;
        pipeHandle->Opened = TRUE;
		pipeHandle->bPerfTimerEnabled = FALSE;
        ntStatus = STATUS_SUCCESS;
    }

    Irp->IoStatus.Status = ntStatus;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest (Irp,
                       IO_NO_INCREMENT
                       );

    I82930_DecrementIoCount(DeviceObject);                               

    I82930_KdPrint (("'I82930.SYS: exit I82930_Create %x\n", ntStatus));

    return ntStatus;
}


