/*++

Copyright (c) 1996 Microsoft Corporation

Module Name:

    CapProp.c

Abstract:

    This is a driver for the Sony Desktop Camera (CCM-DS250 V1.0x).
    This file contains code to handle the video and camera control properties.

Author:
    
    Yee J. Wu (ezuwu) 9-Sep-97

Environment:

    Kernel mode only

Revision History:


--*/

// E-Zu
NTSTATUS
SonyDCamReadRegister(
    IN PIRB Irb,
    PSONYDCAM_EXTENSION DeviceExtension,
	ULONG ulFieldOffset,
	ULONG * pulValue
)
{
	NTSTATUS status;
    LARGE_INTEGER deltaTime;


    //
    // Delay for camera before next request
    //
    deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
    deltaTime.HighPart = -1;
    KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);

    Irb->FunctionNumber = REQUEST_ASYNC_READ;
    Irb->Flags = 0;

    Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_High = 
			INITIAL_REGISTER_SPACE_HI;

	Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_Low = 		
			DeviceExtension->BaseRegister + ulFieldOffset;

	Irb->u.AsyncRead.nNumberOfBytesToRead = sizeof(ULONG);
	Irb->u.AsyncRead.nBlockSize = 0;
	Irb->u.AsyncRead.fulFlags = 0;
	Irb->u.AsyncRead.ulGeneration = DeviceExtension->CurrentGeneration;
	DeviceExtension->RegisterWorkArea.AsULONG = 0;    // Initilize the return buffer.
	Irb->u.AsyncRead.Mdl = MmCreateMdl(NULL, &DeviceExtension->RegisterWorkArea, sizeof(ULONG));
    MmBuildMdlForNonPagedPool(Irb->u.AsyncRead.Mdl);

	KdPrint(("\'SonyDCamReadRegister: Read from address (%x, %x)\n",
		Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_High,
		Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_Low));	

	status = SonyDCamSubmitIrpSynch(DeviceExtension, DeviceExtension->Irp, Irb);

    if (status) {

        KdPrint(("\'SonyDCamRange: Error %x while trying to read from register\n", status));       		
    } else {

		*pulValue = DeviceExtension->RegisterWorkArea.AsULONG;
        KdPrint(("\'SonyDCamReadRegister: status=0x%x, value=0x%x\n", status, *pulValue));       		
	}
	
	ExFreePool(Irb->u.AsyncWrite.Mdl);
	return status;
}

// E-Zu
NTSTATUS
SonyDCamWriteRegister(
    IN PIRB Irb,
    PSONYDCAM_EXTENSION DeviceExtension,
	ULONG ulFieldOffset,
	ULONG ulValue
)
{
	NTSTATUS status;
    LARGE_INTEGER deltaTime;


    //
    // Delay for camera before next request
    //
    deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
    deltaTime.HighPart = -1;
    KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);

    Irb->FunctionNumber = REQUEST_ASYNC_WRITE;
    Irb->Flags = 0;

    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_High = 
			INITIAL_REGISTER_SPACE_HI;

	Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low = 		
			DeviceExtension->BaseRegister + ulFieldOffset;

	Irb->u.AsyncWrite.nNumberOfBytesToWrite = sizeof(ULONG);
	Irb->u.AsyncWrite.nBlockSize = 0;
	Irb->u.AsyncWrite.fulFlags = 0;
	Irb->u.AsyncWrite.ulGeneration = DeviceExtension->CurrentGeneration;
	DeviceExtension->RegisterWorkArea.AsULONG = ulValue;    // Initilize the return buffer.
	Irb->u.AsyncWrite.Mdl = MmCreateMdl(NULL, &DeviceExtension->RegisterWorkArea, sizeof(ULONG));
    MmBuildMdlForNonPagedPool(Irb->u.AsyncWrite.Mdl);

	KdPrint(("\'SonyDCamWriteRegister: Read from address (%x, %x)\n",
		Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_High,
		Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low));	

	status = SonyDCamSubmitIrpSynch(DeviceExtension, DeviceExtension->Irp, Irb);

    if (status) {

        KdPrint(("\'SonyDCamRange: Error %x while trying to read from register\n", status));       		
    } 
	
	ExFreePool(Irb->u.AsyncWrite.Mdl);
	return status;
}

// E-Zu
NTSTATUS
SonyDCamGetProperty(
    IN PIRB Irb,
    PSONYDCAM_EXTENSION DeviceExtension, 
	ULONG ulFieldOffset,
	LONG * plValue
)
{

	NTSTATUS status;
	DCamRegArea RegArea;

	status = SonyDCamReadRegister(Irb, DeviceExtension, ulFieldOffset, &(RegArea.AsULONG));
    if (!status) {    

		RegArea.AsULONG = bswap(RegArea.AsULONG);

		KdPrint(("\'SonyDCamGetProperty: 0x%x: PresenceIng=%d, ONOff=%d, AutoMode=%d, Value=%d\n", 
			RegArea.AsULONG, 
			RegArea.Brightness.PresenceInq,			
			RegArea.Brightness.OnOff,
			RegArea.Brightness.AutoMode,
			RegArea.Brightness.Value
			));

		if (RegArea.Brightness.PresenceInq) {

			*plValue = (LONG) RegArea.Brightness.Value;
			KdPrint(("\'SonyDCamGetProperty: Currrent value is %d.\n", *plValue));

		} else {

		    KdPrint(("\'SonyDCamGetProperty: PresenceInq not set, feature not supported.\n"));
		}
	}

	return status;
}


// E-Zu
NTSTATUS
SonyDCamSetProperty(
    IN PIRB Irb,
    PSONYDCAM_EXTENSION DeviceExtension, 
	ULONG ulFieldOffset,
	LONG  lValue
)
{
	NTSTATUS status;
	DCamRegArea RegArea;

	status = SonyDCamReadRegister(Irb, DeviceExtension, ulFieldOffset, &(RegArea.AsULONG));
    if (!status) {    

		RegArea.AsULONG = bswap(RegArea.AsULONG);

		KdPrint(("\'SonyDCamSetProperty: 0x%x: PresenceIng=%d, ONOff=%d, AutoMode=%d, Value=%d\n", 
			RegArea.AsULONG, 
			RegArea.Brightness.PresenceInq,			
			RegArea.Brightness.OnOff,
			RegArea.Brightness.AutoMode,
			RegArea.Brightness.Value
			));

		// Feature supported and in Manual mode
		if (RegArea.Brightness.PresenceInq && 
			!RegArea.Brightness.AutoMode) {

			RegArea.Brightness.Value = lValue;

			RegArea.AsULONG = bswap(RegArea.AsULONG);

			status = SonyDCamWriteRegister(Irb, DeviceExtension, ulFieldOffset, RegArea.AsULONG);

			if (status) { 

				KdPrint(("\'SonyDCamSetRange: failed with status=0x%x\n", status));

			}
		}
	}

	return status;

}

// E-Zu
NTSTATUS
SonyDCamGetRange(
    IN PIRB Irb,
    PSONYDCAM_EXTENSION DeviceExtension,
	ULONG ulFieldOffset,
	LONG * pMinValue,
	LONG * pMaxValue
)
{
	NTSTATUS status;
	DCamRegArea RegArea;


	status = SonyDCamReadRegister(Irb, DeviceExtension, ulFieldOffset, &(RegArea.AsULONG));

    if (!status) {    
		
		RegArea.AsULONG = bswap(RegArea.AsULONG);

		KdPrint(("\'SonyDCamGetRange: 0x%x: PresenceIng=%d, ReadOut=%d, ONOff=%d, AutoMode=%d, Manual=%d, min=%d, max=%d\n", 
			RegArea.AsULONG, 
			RegArea.Feature.PresenceInq,			
			RegArea.Feature.ReadOut_Inq,
			RegArea.Feature.OnOff,
			RegArea.Feature.AutoMode,
			RegArea.Feature.Manualode,
			RegArea.Feature.MIN_Value,
			RegArea.Feature.MAX_Value
			));

		if (RegArea.Feature.PresenceInq) {

			*pMinValue = RegArea.Feature.MIN_Value;		
			*pMaxValue = RegArea.Feature.MAX_Value; 
 
			KdPrint(("\'SonyDCamGetRange: status=0x%x, (min=%d, max=%d)\n", status, *pMinValue, *pMaxValue));

		} else {	    
		    KdPrint(("\'SonyDCamFeature: PresenceInq not set, feature not supported.\n"));
			*pMinValue = *pMinValue = 0;
		}
	}

	return status;
	
}


/*
** AdapterGetVideoProcAmpProperty ()
**
**    Handles Set operations on the VideoProcAmp property set.
**      Testcap uses this for demo purposes only.
**
** Arguments:
**
**      pSRB -
**          Pointer to the HW_STREAM_REQUEST_BLOCK 
**
** Returns:
**
** Side Effects:  none
*/

VOID 
AdapterGetVideoProcAmpProperty(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
	LONG CurValue;

    PSONYDCAM_EXTENSION 
		DeviceExtension = (PSONYDCAM_EXTENSION) pSrb->HwDeviceExtension;

    PSTREAM_PROPERTY_DESCRIPTOR 
		pSPD = pSrb->CommandData.PropertyInfo;

    PKSPROPERTY_VIDEOPROCAMP_S 
		pS = (PKSPROPERTY_VIDEOPROCAMP_S) pSPD->PropertyInfo;    // pointer to the data

    ASSERT (pSPD->PropertyOutputSize >= sizeof (KSPROPERTY_VIDEOPROCAMP_S));
    
	pSrb->Status = STATUS_SUCCESS;

    switch (pSPD->Property->Id) {

    case KSPROPERTY_VIDEOPROCAMP_BRIGHTNESS:
		
		if (!SonyDCamGetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Brightness), &CurValue))        
			pS->Value = DeviceExtension->Brightness = CurValue;
        break;

    case KSPROPERTY_VIDEOPROCAMP_SHARPNESS:
		
		if (!SonyDCamGetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Sharpness), &CurValue))        
			pS->Value = DeviceExtension->Sharpness = CurValue; 
        break;

    case KSPROPERTY_VIDEOPROCAMP_HUE:
		
		if (!SonyDCamGetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Hue), &CurValue))        
			pS->Value = DeviceExtension->Hue = CurValue; 
        break;
        
    case KSPROPERTY_VIDEOPROCAMP_SATURATION:

		if (!SonyDCamGetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Saturation), &CurValue))
	        pS->Value = DeviceExtension->Saturation = CurValue; 
        break;

    case KSPROPERTY_VIDEOPROCAMP_WHITEBALANCE:

		if (!SonyDCamGetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, WhiteBalance), &CurValue))
	        pS->Value = DeviceExtension->Saturation = CurValue; 
        break;


    default:
	    KdPrint(("\'AdapterGetVideoProcAmpProperty, Id (%x)not supported.\n", pSPD->Property->Id));
        TRAP();
        break;
    }

    pSrb->ActualBytesTransferred = sizeof (KSPROPERTY_VIDEOPROCAMP_S);

}

/*
** AdapterGetCameraControlProperty ()
**
**    Handles Set operations on the VideoProcAmp property set.
**      Testcap uses this for demo purposes only.
**
** Arguments:
**
**      pSRB -
**          Pointer to the HW_STREAM_REQUEST_BLOCK 
**
** Returns:
**
** Side Effects:  none
*/

VOID 
AdapterGetCameraControlProperty(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
	LONG CurValue;

    PSONYDCAM_EXTENSION 
		DeviceExtension = (PSONYDCAM_EXTENSION) pSrb->HwDeviceExtension;

    PSTREAM_PROPERTY_DESCRIPTOR 
		pSPD = pSrb->CommandData.PropertyInfo;

    PKSPROPERTY_CAMERACONTROL_S 
		pS = (PKSPROPERTY_CAMERACONTROL_S) pSPD->PropertyInfo;    // pointer to the data

    ASSERT (pSPD->PropertyOutputSize >= sizeof (KSPROPERTY_CAMERACONTROL_S));
    
	pSrb->Status = STATUS_SUCCESS;

    switch (pSPD->Property->Id) {

    case KSPROPERTY_CAMERACONTROL_FOCUS:

		if (!SonyDCamGetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Focus), &CurValue)) {
			pS->Value = DeviceExtension->Focus = CurValue; 

			KdPrint(("\'@@@@-- Get Focus current value = %d\n",CurValue));
		}
        break;       

    case KSPROPERTY_CAMERACONTROL_ZOOM:

		if (!SonyDCamGetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Zoom), &CurValue)) {
			pS->Value = DeviceExtension->Zoom = CurValue; 

			KdPrint(("\'@@@@-- Get Zoom current value = %d\n",CurValue));
		}

        break;       


    default:	    
		KdPrint(("\'AdapterGetCameraControlProperty, Id (%x)not supported.\n", pSPD->Property->Id));
        TRAP();
        break;
    }

    pSrb->ActualBytesTransferred = sizeof (KSPROPERTY_CAMERACONTROL_S);

}


/*
** AdapterGetProperty ()
**
**    Handles Get operations for all adapter properties.
**
** Arguments:
**
**      pSRB -
**          Pointer to the HW_STREAM_REQUEST_BLOCK 
**
** Returns:
**
** Side Effects:  none
*/

VOID
STREAMAPI 
AdapterGetProperty(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )

{
    PSTREAM_PROPERTY_DESCRIPTOR pSPD = pSrb->CommandData.PropertyInfo;

	if (IsEqualGUID(&PROPSETID_VIDCAP_VIDEOPROCAMP, &pSPD->Property->Set)) {
        AdapterGetVideoProcAmpProperty (pSrb);
    } else 	if (IsEqualGUID(&PROPSETID_VIDCAP_CAMERACONTROL, &pSPD->Property->Set)) {
        AdapterGetCameraControlProperty (pSrb);
    } else {
        //
        // We should never get here
        //

        TRAP();
    }
}

/*
** AdapterSetVideoProcAmpProperty ()
**
**    Handles Set operations on the VideoProcAmp property set.
**      Testcap uses this for demo purposes only.
**
** Arguments:
**
**      pSRB -
**          Pointer to the HW_STREAM_REQUEST_BLOCK 
**
** Returns:
**
** Side Effects:  none
*/

VOID 
AdapterSetVideoProcAmpProperty(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
    PSONYDCAM_EXTENSION 
		DeviceExtension = (PSONYDCAM_EXTENSION) pSrb->HwDeviceExtension;

    PSTREAM_PROPERTY_DESCRIPTOR 
		pSPD = pSrb->CommandData.PropertyInfo;

    PKSPROPERTY_VIDEOPROCAMP_S 
		pS = (PKSPROPERTY_VIDEOPROCAMP_S) pSPD->PropertyInfo;    // pointer to the data

    ASSERT (pSPD->PropertyOutputSize >= sizeof (KSPROPERTY_VIDEOPROCAMP_S));
    
	pSrb->Status = STATUS_SUCCESS;

    switch (pSPD->Property->Id) {

    case KSPROPERTY_VIDEOPROCAMP_BRIGHTNESS:				

		if (!SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Brightness), pS->Value))
			DeviceExtension->Brightness = pS->Value;
        break;
        
    case KSPROPERTY_VIDEOPROCAMP_SHARPNESS:

		if (!SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Sharpness), pS->Value))
			DeviceExtension->Sharpness = pS->Value;        
        break;

    case KSPROPERTY_VIDEOPROCAMP_HUE:

		if (!SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Hue), pS->Value))
			DeviceExtension->Hue = pS->Value;        
        break;

    case KSPROPERTY_VIDEOPROCAMP_SATURATION:

		if (!SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Saturation), pS->Value))
			DeviceExtension->Saturation = pS->Value;        
        break;

    case KSPROPERTY_VIDEOPROCAMP_WHITEBALANCE:

		if (!SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, WhiteBalance), pS->Value))
			DeviceExtension->WhiteBalance = pS->Value;        
        break;

    default:
        TRAP();
        break;
    }

    pSrb->ActualBytesTransferred = sizeof (KSPROPERTY_VIDEOPROCAMP_S);

}
/*
** AdapterSetCameraControlProperty ()
**
**    Handles Set operations on the CameraControl property set.
**
** Arguments:
**
**      pSRB -
**          Pointer to the HW_STREAM_REQUEST_BLOCK 
**
** Returns:
**
** Side Effects:  none
*/

VOID 
AdapterSetCameraControlProperty(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
    PSONYDCAM_EXTENSION 
		DeviceExtension = (PSONYDCAM_EXTENSION) pSrb->HwDeviceExtension;

    PSTREAM_PROPERTY_DESCRIPTOR 
		pSPD = pSrb->CommandData.PropertyInfo;

    PKSPROPERTY_CAMERACONTROL_S 
		pS = (PKSPROPERTY_CAMERACONTROL_S) pSPD->PropertyInfo;    // pointer to the data

    ASSERT (pSPD->PropertyOutputSize >= sizeof (KSPROPERTY_CAMERACONTROL_S));
    
	pSrb->Status = STATUS_SUCCESS;

    switch (pSPD->Property->Id) {

    case KSPROPERTY_CAMERACONTROL_FOCUS:

		if (!SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Focus), pS->Value)) {

			DeviceExtension->Focus = pS->Value; 
			KdPrint(("\'@@@@--Set Focus current value = %d\n",pS->Value));

		} else {
			TRAP();		
		}
        break;

    case KSPROPERTY_CAMERACONTROL_ZOOM:

		if (!SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Zoom), pS->Value)) {

			DeviceExtension->Zoom = pS->Value;        
			KdPrint(("\'@@@@-- Set Zoom current value = %d\n",pS->Value));

		} else {
			TRAP();
		}
        break;

    default:
        TRAP();
        break;
    }

    pSrb->ActualBytesTransferred = sizeof (KSPROPERTY_CAMERACONTROL_S);

}


/*
** AdapterSetProperty ()
**
**    Handles Get operations for all adapter properties.
**
** Arguments:
**
**      pSRB -
**          Pointer to the HW_STREAM_REQUEST_BLOCK 
**
** Returns:
**
** Side Effects:  none
*/

VOID
STREAMAPI 
AdapterSetProperty(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )

{
    PSTREAM_PROPERTY_DESCRIPTOR pSPD = pSrb->CommandData.PropertyInfo;

	if (IsEqualGUID(&PROPSETID_VIDCAP_VIDEOPROCAMP, &pSPD->Property->Set)) {
        AdapterSetVideoProcAmpProperty (pSrb);
    } else 	if (IsEqualGUID(&PROPSETID_VIDCAP_CAMERACONTROL, &pSPD->Property->Set)) {
        AdapterSetCameraControlProperty (pSrb);
    } else {
        //
        // We should never get here
        //

        TRAP();
    }
}


void
InitilizePropertyArray(
    IN PHW_STREAM_REQUEST_BLOCK pSrb
	)
{
	NTSTATUS status;
    PHW_STREAM_HEADER StreamHeader = &(pSrb->CommandData.StreamBuffer->StreamHeader);        
    PSONYDCAM_EXTENSION DeviceExtension = (PSONYDCAM_EXTENSION) pSrb->HwDeviceExtension;

    KdPrint(("\'@@@@--------------begin------------\n"));
    KdPrint(("\'There are %d Property Set\n", NUMBER_OF_ADAPTER_PROPERTY_SETS));

	status = SonyDCamReadRegister((PIRB) pSrb->SRBExtension, DeviceExtension, 
		FIELDOFFSET(CAMERA_REGISTER_MAP, FeaturePresent1), 
		&(DeviceExtension->CameraCapArea1.AsULONG));

	status = SonyDCamReadRegister((PIRB) pSrb->SRBExtension, DeviceExtension, 
		FIELDOFFSET(CAMERA_REGISTER_MAP, FeaturePresent2), 
		&(DeviceExtension->CameraCapArea2.AsULONG));

	//
	// To do: retrieve default value from registry. E-Zu
	//
	// To do: need to save them when device is uninitialized.
	//


	//
	// 1. Get ranges of many camera features, 
	// 2. Set default as the average of min and max
	// 3. Set current value if exist or use default value
	//

	// Brightness
	SonyDCamGetRange((PIRB) pSrb->SRBExtension, DeviceExtension, 
		FIELDOFFSET(CAMERA_REGISTER_MAP, Brightness_Inq), 
		&BrightnessRangeAndStep[0].Bounds.SignedMinimum, 		
		&BrightnessRangeAndStep[0].Bounds.SignedMaximum); 
	BrightnessDefault = (BrightnessRangeAndStep[0].Bounds.SignedMinimum + BrightnessRangeAndStep[0].Bounds.SignedMaximum) / 2;	
	SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Brightness), 
		DeviceExtension->Brightness == 0 ? BrightnessDefault : DeviceExtension->Brightness);

	// Sharpness
	SonyDCamGetRange((PIRB) pSrb->SRBExtension, DeviceExtension, 
		FIELDOFFSET(CAMERA_REGISTER_MAP, Sharpness_Inq), 
		&SharpnessRangeAndStep[0].Bounds.SignedMinimum, 		
		&SharpnessRangeAndStep[0].Bounds.SignedMaximum); 
	SharpnessDefault = (SharpnessRangeAndStep[0].Bounds.SignedMinimum + SharpnessRangeAndStep[0].Bounds.SignedMaximum) / 2;
	SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Sharpness), 
		DeviceExtension->Sharpness == 0 ? SharpnessDefault : DeviceExtension->Sharpness);

	// Hue
	SonyDCamGetRange((PIRB) pSrb->SRBExtension, DeviceExtension, 
		FIELDOFFSET(CAMERA_REGISTER_MAP, Hue_Inq), 
		&HueRangeAndStep[0].Bounds.SignedMinimum, 		
		&HueRangeAndStep[0].Bounds.SignedMaximum); 
	HueDefault = (HueRangeAndStep[0].Bounds.SignedMinimum + HueRangeAndStep[0].Bounds.SignedMaximum) / 2;
	SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Hue), 
		DeviceExtension->Hue == 0 ? HueDefault : DeviceExtension->Hue);

    KdPrint(("\'@@@@--------------Saturation----------\n"));

	// Saturation
	SonyDCamGetRange((PIRB) pSrb->SRBExtension, DeviceExtension, 
		FIELDOFFSET(CAMERA_REGISTER_MAP, Saturation_Inq), 
		&SaturationRangeAndStep[0].Bounds.SignedMinimum, 		
		&SaturationRangeAndStep[0].Bounds.SignedMaximum); 
	SaturationDefault = (SaturationRangeAndStep[0].Bounds.SignedMinimum + SaturationRangeAndStep[0].Bounds.SignedMaximum) / 2;
	SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Saturation), 
		DeviceExtension->Saturation == 0 ? SaturationDefault : DeviceExtension->Saturation);

    KdPrint(("\'@@@@--------------Saturation---end-------\n"));

	// WhiteBalance
	SonyDCamGetRange((PIRB) pSrb->SRBExtension, DeviceExtension, 
		FIELDOFFSET(CAMERA_REGISTER_MAP, WhiteBalance_Inq), 
		&WhiteBalanceRangeAndStep[0].Bounds.SignedMinimum, 		
		&WhiteBalanceRangeAndStep[0].Bounds.SignedMaximum); 
	WhiteBalanceDefault = (WhiteBalanceRangeAndStep[0].Bounds.SignedMinimum + WhiteBalanceRangeAndStep[0].Bounds.SignedMaximum) / 2;
	SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, WhiteBalance), 
		DeviceExtension->WhiteBalance == 0 ? WhiteBalanceDefault : DeviceExtension->WhiteBalance);

    KdPrint(("\'@@@@--------------Camera control----------\n"));

	// Focus
	SonyDCamGetRange((PIRB) pSrb->SRBExtension, DeviceExtension, 
		FIELDOFFSET(CAMERA_REGISTER_MAP, Focus_Inq), 
		&FocusRangeAndStep[0].Bounds.SignedMinimum, 		
		&FocusRangeAndStep[0].Bounds.SignedMaximum);
	FocusDefault = (FocusRangeAndStep[0].Bounds.SignedMinimum + FocusRangeAndStep[0].Bounds.SignedMaximum) / 2;
	SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Focus), 
		DeviceExtension->Focus == 0 ? FocusDefault : DeviceExtension->Focus);

    KdPrint(("\'@@@@--Zoom (%d, %d), default=%d\n",
		FocusRangeAndStep[0].Bounds.SignedMinimum, 		
		FocusRangeAndStep[0].Bounds.SignedMaximum,
		FocusDefault
		));

	// Zoom
	SonyDCamGetRange((PIRB) pSrb->SRBExtension, DeviceExtension, 
		FIELDOFFSET(CAMERA_REGISTER_MAP, Zoom_Inq), 
		&ZoomRangeAndStep[0].Bounds.SignedMinimum, 		
		&ZoomRangeAndStep[0].Bounds.SignedMaximum);
	ZoomDefault = (ZoomRangeAndStep[0].Bounds.SignedMinimum + ZoomRangeAndStep[0].Bounds.SignedMaximum) / 2;
	SonyDCamSetProperty((PIRB) pSrb->SRBExtension, DeviceExtension, FIELDOFFSET(CAMERA_REGISTER_MAP, Zoom), 
		DeviceExtension->Zoom == 0 ? ZoomDefault : DeviceExtension->Zoom);

    KdPrint(("\'@@@@--Focus (%d, %d), default=%d\n",
		ZoomRangeAndStep[0].Bounds.SignedMinimum, 		
		ZoomRangeAndStep[0].Bounds.SignedMaximum,
		ZoomDefault
		));

    KdPrint(("\'@@@@--------------end------------\n"));

    StreamHeader->NumDevPropArrayEntries = NUMBER_OF_ADAPTER_PROPERTY_SETS;
    StreamHeader->DevicePropertiesArray = (PKSPROPERTY_SET) AdapterPropertyTable; 
}