/*++

Copyright (c) 1996 Microsoft Corporation

Module Name:

    dcampkt.c

Abstract:

    This is a driver for the Sony Desktop Camera (CCM-DS250 V1.0x).
    This file contains code to handle the stream class packets.

Author:
    
    Shaun Pierce (shaunp) 25-May-96

Environment:

    Kernel mode only

Revision History:


--*/


#include "strmini.h"
#include "ksmedia.h"
#include "1394.h"
#include "wdm.h"      // for DbgBreakPoint() defined in dbg.h
#include "dbg.h"
#include "dcamdef.h"
#include "dcampkt.h"

#include "capprop.h"   // Video and camera property function prototype
#include "PropData.h"  // Define once property data definition
#include "CapProp.c"   // Source code to process VideoProcAmp and CameraControl





VOID
SonyDCamReceivePacket(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    This is where most of the interesting Stream requests come to us

Arguments:

    Srb - Pointer to Stream request block

Return Value:

    Nothing

--*/

{

    PIO_STACK_LOCATION IrpStack;
    PSONYDCAM_EXTENSION DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;


    //
    // Switch on the command within the Srb itself
    //

    KdPrint(("\'SonyDCamReceivePacket: entry with command = %x\n", Srb->Command));

    switch (Srb->Command) {

		// VideoProcAmp and CameraControl requests
        case SRB_GET_DEVICE_PROPERTY:

			KdPrint(("\'SonyDCamReceivePacket: SRB_GET_DEVICE_PROPERTY"));
			AdapterGetProperty(Srb);
            break;
		
        case SRB_SET_DEVICE_PROPERTY:

			KdPrint(("\'SonyDCamReceivePacket: SRB_SET_DEVICE_PROPERTY"));
			AdapterSetProperty(Srb);
            break;

        case SRB_GET_STREAM_INFO:

            //
            // this is a request for the driver to enumerate requested streams
            //

            SonyDCamGetStreamInfo(Srb);
            break;

        case SRB_GET_DATA_INTERSECTION:

            //
            // Return a format, given a range
            //

            AdapterFormatFromRange(Srb);
            break;

        case SRB_OPEN_STREAM:

            //
            // this is a request to open a specified stream.  
            //

            SonyDCamOpenStream(Srb);
            break;

        case SRB_CLOSE_STREAM:

            //
            // this is a request to close a specified stream
            //

            SonyDCamCloseStream(Srb);
            break;

        case SRB_INITIALIZE_DEVICE:

            Srb->Status = SonyDCamHwInitialize(Srb);
            break;

        case SRB_CHANGE_POWER_STATE:

            Srb->Status = STATUS_SUCCESS;
            break;

        case SRB_UNINITIALIZE_DEVICE:

            DeviceExtension->bShutdown = TRUE;
            SonyDCamCloseStream(Srb);
            if (DeviceExtension->Irb) 
                ExFreePool(DeviceExtension->Irb);
            if (DeviceExtension->FormatBlocks)
                ExFreePool(DeviceExtension->FormatBlocks);
            Srb->Status = STATUS_SUCCESS;
            break;
            
        case SRB_UNKNOWN_DEVICE_COMMAND:

            //
            // We might be interested in unknown commands if they pertain
            // to bus resets.  Bus resets are important cuz we need to know
            // what the current generation count is.
            //

            IrpStack = IoGetCurrentIrpStackLocation(Srb->Irp);

            if ((IrpStack->MajorFunction == IRP_MJ_PNP) &&
                (IrpStack->MinorFunction == IRP_MN_BUS_RESET)) {

                KdPrint(("\'SonyDCamReceivePacket: Bus Reset has occurred, new generation = %x\n", *((PULONG) &IrpStack->Parameters.Others.Argument4)));
                DeviceExtension->CurrentGeneration = \
                    *((PULONG) &IrpStack->Parameters.Others.Argument4);

                //
                // We're going to shut down the stream at this point
                //
                
                SonyDCamStopStream(Srb);
                Srb->Status = STATUS_SUCCESS;
                break;

            }

            Srb->Status = STATUS_NOT_IMPLEMENTED;
            break;

        default:

            //
            // this is a request that we do not understand.  Indicate invalid
            // command and complete the request
            //

            Srb->Status = STATUS_NOT_IMPLEMENTED;

    }

    //
    // NOTE:
    //
    // all of the commands that we do, or do not understand can all be completed
    // synchronously at this point, so we can use a common callback routine here.
    // If any of the above commands require asynchronous processing, this will
    // have to change
    //

    StreamClassDeviceNotification(DeviceRequestComplete, Srb->HwDeviceExtension, Srb);

}


VOID
SonyDCamGetStreamInfo(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    Returns the information of all streams that are supported by the driver

Arguments:

    Srb - Pointer to Stream request block

Return Value:

    Nothing

--*/

{
     
    //
    // pick up the pointer to the stream information data structure
    //

    PHW_STREAM_HEADER StreamHeader = &(Srb->CommandData.StreamBuffer->StreamHeader);        
    PSONYDCAM_EXTENSION DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;
    PHW_STREAM_INFORMATION StreamInfo = &(Srb->CommandData.StreamBuffer->StreamInfo);


    KdPrint(("\'SonyDCamGetStreamInfo: enter\n"));

    //
    // set number of streams
    //

    ASSERT (Srb->NumberOfBytesToTransfer >= 
            sizeof (HW_STREAM_HEADER) +
            sizeof (HW_STREAM_INFORMATION));

    //
    // initialize stream header
    //

    RtlZeroMemory(StreamHeader, 
                sizeof (HW_STREAM_HEADER) +
                sizeof (HW_STREAM_INFORMATION));

    //
    // initialize the number of streams supported
    //

    StreamHeader->NumberOfStreams = 1;
    StreamHeader->SizeOfHwStreamInformation = sizeof(HW_STREAM_INFORMATION);


    //
    // set the VideoProcAmp and CameraControl property information in a static array
	// for the video stream.  
	// It also queries the property value range, set default value.
    // 
	InitilizePropertyArray(Srb);


    //
    // Initialize the stream structure.
    //
    // Number of instances field indicates the number of concurrent streams
    // of this type the device can support.  
    //

    StreamInfo->NumberOfPossibleInstances = 1;

    //
    // indicates the direction of data flow for this stream, relative to 
    // the driver
    //

    StreamInfo->DataFlow = KSPIN_DATAFLOW_OUT;

    //
    // dataAccessible - Indicates whether the data is "seen" by the host
    // processor.
    //

    StreamInfo->DataAccessible = TRUE;

    //
    // Return how many formats this stream supports 
    //

    StreamInfo->NumberOfFormatArrayEntries = DeviceExtension->ModesSupported;
    StreamInfo->StreamFormatsArray = &SONYDCAM_StreamFormats[0];

    //
    // set the property information for the video stream
    //

    StreamInfo->NumStreamPropArrayEntries = NUMBER_VIDEO_STREAM_PROPERTIES;
    StreamInfo->StreamPropertiesArray = (PKSPROPERTY_SET) VideoStreamProperties;

    //
    // set the pin name and category
    //

    StreamInfo->Name = (GUID *) &PINNAME_VIDEO_CAPTURE;
    StreamInfo->Category = (GUID *) &PINNAME_VIDEO_CAPTURE;


    //
    // store a pointer to the topology for the device
    //
        
    Srb->CommandData.StreamBuffer->StreamHeader.Topology = &Topology;

    //
    // BUGBUG: check here that the stream information provided is no larger
    // than the size specified in the HwInitData
    //

    //
    // indicate success
    //

    Srb->Status = STATUS_SUCCESS;

}


VOID
SonyDCamOpenStream(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    Called when an OpenStream Srb request is received

Arguments:

    Srb - Pointer to Stream request block

Return Value:

    Nothing

--*/

{

    PIRB Irb;
    PIRP Irp;
    CCHAR StackSize;
    ULONG nSize;
    ULONG ModeIndex;
    PSONYDCAM_EXTENSION DeviceExtension;
    PKS_DATAFORMAT_VIDEOINFOHEADER  pKSDataFormat = 
                (PKS_DATAFORMAT_VIDEOINFOHEADER) Srb->CommandData.OpenFormat;
    PKS_VIDEOINFOHEADER     pVideoInfoHdrRequested = 
                &pKSDataFormat->VideoInfoHeader;


    KdPrint(("\'SonyDCamOpenStream: enter\n"));
    Irb = (PIRB) Srb->SRBExtension;
    DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;

    //
    // default to success
    //

    Srb->Status = STATUS_SUCCESS;

    //
    // determine which stream number is being opened.  This number indicates
    // the offset into the array of streaminfo structures that was filled out
    // in the AdapterStreamInfo call.
    //
    // So:
    //   0 - Video data from camera
    //

    switch (Srb->StreamObject->StreamNumber) {

        case 0:

            //
            // Figure out what format they're trying to open first
            //

            if (!AdapterVerifyFormat (pKSDataFormat, Srb->StreamObject->StreamNumber)) {

                Srb->Status = STATUS_INVALID_PARAMETER;
                return;

            }

            ASSERT(DeviceExtension->pVideoInfoHeader == NULL);

            nSize = sizeof (KS_VIDEOINFOHEADER) + 
                        pVideoInfoHdrRequested->bmiHeader.biSize;

            DebugPrint((DebugLevelVerbose,"TestCap: pVideoInfoHdrRequested=%x\n", 
                        pVideoInfoHdrRequested));
            DebugPrint((DebugLevelVerbose,"TestCap: KS_VIDEOINFOHEADER size=%d\n", 
                        nSize));
            DebugPrint((DebugLevelVerbose,"TestCap: Width=%d  Height=%d  BitCount=%d\n", 
                        pVideoInfoHdrRequested->bmiHeader.biWidth,
                        pVideoInfoHdrRequested->bmiHeader.biHeight,
                        pVideoInfoHdrRequested->bmiHeader.biBitCount));
            DebugPrint((DebugLevelVerbose,"TestCap: biSizeImage=%d\n", 
                        pVideoInfoHdrRequested->bmiHeader.biSizeImage));

            DeviceExtension->pVideoInfoHeader = ExAllocatePool(NonPagedPool, nSize);

            if (DeviceExtension->pVideoInfoHeader == NULL) {

                DebugPrint((DebugLevelVerbose,"TestCap: ExAllocatePool failed\n"));
                Srb->Status = STATUS_INSUFFICIENT_RESOURCES;
                return;

            }

            // Copy the VIDEOINFOHEADER requested to our storage

            RtlCopyMemory(
                    DeviceExtension->pVideoInfoHeader,
                    pVideoInfoHdrRequested,
                    nSize);

            ModeIndex = 1;

            //
            // Now if they're on a YUV4:2:2 format, we've gotta check what
            // resolution they want it at, since we support this format
            // but in two different resolutions (modes on the camera).
            //
                            
            if ((ModeIndex == 1) &&
                (DeviceExtension->Resolution == 3)) {

                ModeIndex = 3;

            }

            //
            // Get an Irp so we can send some allocation commands down
            //

            StackSize = DeviceExtension->BusDeviceObject->StackSize;
            Irp = IoAllocateIrp(StackSize, FALSE);

            if (!Irp) {
        
                SONYDCAM_ASSERT(FALSE);
                Srb->Status = STATUS_INSUFFICIENT_RESOURCES;
                return;

            }

            Irb->FunctionNumber = REQUEST_ISOCH_ALLOCATE_RESOURCES;
            Irb->Flags = 0;
            Irb->u.IsochAllocateResources.fulSpeed = IsochInfoTable[ModeIndex*NUM_POSSIBLE_RATES+DeviceExtension->FrameRate].SpeedRequired;
            Irb->u.IsochAllocateResources.fulFlags = RESOURCE_USED_IN_LISTENING;
            Irb->u.IsochAllocateResources.nMaxBytesPerFrame = IsochInfoTable[ModeIndex*NUM_POSSIBLE_RATES+DeviceExtension->FrameRate].QuadletPayloadPerPacket << 2;
            Irb->u.IsochAllocateResources.nNumberOfBuffers = MAX_BUFFERS_SUPPLIED;
            Irb->u.IsochAllocateResources.nMaxBufferSize = IsochInfoTable[ModeIndex*NUM_POSSIBLE_RATES+DeviceExtension->FrameRate].CompletePictureSize;

            Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

            if (Srb->Status) {

                KdPrint(("\'SonyDCamOpenStream: Error %x while trying to allocate Isoch resources\n", Srb->Status));
                IoFreeIrp(Irp);
                return;

            }

            DeviceExtension->hResource = Irb->u.IsochAllocateResources.hResource;

            Irb->FunctionNumber = REQUEST_ISOCH_ALLOCATE_BANDWIDTH;
            Irb->Flags = 0;
            Irb->u.IsochAllocateBandwidth.nMaxBytesPerFrameRequested = IsochInfoTable[ModeIndex*NUM_POSSIBLE_RATES+DeviceExtension->FrameRate].QuadletPayloadPerPacket << 2;
            Irb->u.IsochAllocateBandwidth.fulSpeed = IsochInfoTable[ModeIndex*NUM_POSSIBLE_RATES+DeviceExtension->FrameRate].SpeedRequired;

            Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

            if (Srb->Status) {

                KdPrint(("\'SonyDCamOpenStream: Error %x while trying to allocate Isoch bandwidth\n", Srb->Status));
                IoFreeIrp(Irp);
                return;

            }

            DeviceExtension->hBandwidth = Irb->u.IsochAllocateBandwidth.hBandwidth;

            Irb->FunctionNumber = REQUEST_ISOCH_ALLOCATE_CHANNEL;
            Irb->Flags = 0;
            Irb->u.IsochAllocateChannel.nRequestedChannel = 0;

            Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

            if (Srb->Status) {

                KdPrint(("\'SonyDCamOpenStream: Error %x while trying to allocate Isoch channel\n", Srb->Status));
                IoFreeIrp(Irp);
                return;

            }

            DeviceExtension->IsochChannel = Irb->u.IsochAllocateChannel.Channel;
            DeviceExtension->CurrentModeIndex = ModeIndex;
            DeviceExtension->bCameraStopped = FALSE;
            IoFreeIrp(Irp);

            Srb->StreamObject->ReceiveDataPacket = (PVOID) SonyDCamReceiveDataPacket;
            Srb->StreamObject->ReceiveControlPacket = (PVOID) SonyDCamReceiveCtrlPacket;

            //
            // initialize the stream extension data handling information
            //

            break;

        default:

            break;
    }

    // Fill in clock information

    Srb->StreamObject->HwClockObject.HwClockFunction = StreamClockRtn;
    Srb->StreamObject->HwClockObject.ClockSupportFlags = \
        CLOCK_SUPPORT_CAN_READ_ONBOARD_CLOCK |
        CLOCK_SUPPORT_CAN_RETURN_STREAM_TIME;

    // We don't use DMA.

    Srb->StreamObject->Dma = FALSE;
    Srb->StreamObject->StreamHeaderMediaSpecific = sizeof(KS_FRAME_INFO);

    //
    // The PIO flag must be set when the mini driver will be accessing the data
    // buffers passed in using logical addressing
    //

    Srb->StreamObject->Pio = FALSE;

}


VOID
SonyDCamCloseStream(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    Called when an CloseStream Srb request is received

Arguments:

    Srb - Pointer to Stream request block

Return Value:

    Nothing

--*/

{

    PIRB Irb;
    PIRP Irp;
    CCHAR StackSize;
    PSONYDCAM_EXTENSION DeviceExtension;
    KS_VIDEOINFOHEADER *pVideoInfoHdr; 


    KdPrint(("\'SonyDCamCloseStream: enter\n"));
    Irb = (PIRB) Srb->SRBExtension;
    DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;
    pVideoInfoHdr = DeviceExtension->pVideoInfoHeader;

    //
    // Get an Irp so we can send some free commands down
    //

    StackSize = DeviceExtension->BusDeviceObject->StackSize;
    Irp = IoAllocateIrp(StackSize, FALSE);

    if (!Irp) {
    
        SONYDCAM_ASSERT(FALSE);
        Srb->Status = STATUS_INSUFFICIENT_RESOURCES;
        return;

    }

    if (DeviceExtension->IsochChannel != (ULONG) -1) {

        Irb->FunctionNumber = REQUEST_ISOCH_FREE_CHANNEL;
        Irb->Flags = 0;
        Irb->u.IsochFreeChannel.nChannel = DeviceExtension->IsochChannel;
        Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

        DeviceExtension->IsochChannel = (ULONG) -1;

        if (Srb->Status) {

            KdPrint(("\'SonyDCamOpenStream: Error %x while trying to free Isoch channel\n", Srb->Status));
            IoFreeIrp(Irp);
            return;

        }

    }

    if (DeviceExtension->hBandwidth) {

        Irb->FunctionNumber = REQUEST_ISOCH_FREE_BANDWIDTH;
        Irb->Flags = 0;
        Irb->u.IsochFreeBandwidth.hBandwidth = DeviceExtension->hBandwidth;
        Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

        DeviceExtension->hBandwidth = NULL;

        if (Srb->Status) {

            KdPrint(("\'SonyDCamOpenStream: Error %x while trying to free Isoch bandwidth\n", Srb->Status));
            IoFreeIrp(Irp);
            return;

        }

    }

    if (DeviceExtension->hResource) {

        Irb->FunctionNumber = REQUEST_ISOCH_FREE_RESOURCES;
        Irb->Flags = 0;
        Irb->u.IsochFreeResources.hResource = DeviceExtension->hResource;
        Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

        DeviceExtension->hResource = NULL;

        if (Srb->Status) {

            KdPrint(("\'SonyDCamOpenStream: Error %x while trying to free Isoch resources\n", Srb->Status));
            IoFreeIrp(Irp);
            return;

        }

    }

    IoFreeIrp(Irp);

    if (pVideoInfoHdr) {

        ExFreePool(pVideoInfoHdr);
        DeviceExtension->pVideoInfoHeader = NULL;

    }

}


VOID
SonyDCamReceiveDataPacket(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    Called with video data packet commands

Arguments:

    Srb - Pointer to Stream request block

Return Value:

    Nothing

--*/

{

    //
    // determine the type of packet.
    //

    switch (Srb->Command) {

        case SRB_READ_DATA:

            KdPrint(("\'SonyDCamReceiveDataPacket: SRB_READ_DATA\n"));
            SonyDCamReadStream(Srb);

            // This request will be completed asynchronously...
        
            break;
            
        case SRB_WRITE_DATA:

            KdPrint(("\'SonyDCamReceiveDataPacket: SRB_WRITE_DATA\n"));
            Srb->Status = STATUS_NOT_IMPLEMENTED;

            // This request will be completed asynchronously...
            
            break;

        default:

            //
            // invalid / unsupported command. Fail it as such
            //

            Srb->Status = STATUS_NOT_IMPLEMENTED;
    
            StreamClassStreamNotification(StreamRequestComplete, Srb->StreamObject, Srb);

    }

}


VOID
SonyDCamReceiveCtrlPacket(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    Called with packet commands that control the video stream

Arguments:

    Srb - Pointer to Stream request block

Return Value:

    Nothing

--*/

{

    PIRB Irb;
    PIRP Irp;
    PMDL Mdl;
    CCHAR StackSize;
    LARGE_INTEGER deltaTime;
    PSONYDCAM_EXTENSION DeviceExtension;


    //
    // determine the type of packet.
    //

    Irb = (PIRB) Srb->SRBExtension;
    Srb->Status = STATUS_SUCCESS;
    DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;

    switch (Srb->Command) {

        case SRB_GET_STREAM_STATE:
            VideoGetState (Srb);
            break;

        case SRB_SET_STREAM_STATE:

            switch (Srb->CommandData.StreamState) {

                case KSSTATE_STOP:

                    DeviceExtension->QST_Now = GetSystemTime();
                    DeviceExtension->QST_Start = DeviceExtension->QST_Now;

                    SonyDCamStopStream(Srb);
                    break;

                case KSSTATE_PAUSE:

                    KdPrint(("\'SonyDCamReceiveCtrlPacket: setting PAUSE stream state\n"));
                    Srb->Status = STATUS_SUCCESS;
#if 1
                    if ((DeviceExtension->KSState == KSSTATE_ACQUIRE) ||
                        (DeviceExtension->KSState == KSSTATE_STOP)) {  
            
                        // Remember the time at which the clock was started

                        DeviceExtension->QST_Start = GetSystemTime();

                        // And initialize the last frame timestamp

                        DeviceExtension->QST_Now = DeviceExtension->QST_Start;

                    }
#endif

                    if (DeviceExtension->bCameraSpewing) {

                        StackSize = DeviceExtension->BusDeviceObject->StackSize;
                        Irp = IoAllocateIrp(StackSize, FALSE);

                        if (!Irp) {
    
                            SONYDCAM_ASSERT(Irp);
                            Srb->Status = STATUS_INSUFFICIENT_RESOURCES;
                            break;

                        }

                        //
                        // stop the stream internally inside 1394 stack
                        //

                        if (DeviceExtension->hResource) {

                            Irb->FunctionNumber = REQUEST_ISOCH_STOP;
                            Irb->Flags = 0;
                            Irb->u.IsochStop.nChannel = DeviceExtension->IsochChannel;
                            Irb->u.IsochStop.hResource = DeviceExtension->hResource;
                            Irb->u.IsochStop.fulFlags = 0;

                            Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

                            if (Srb->Status) {

                                KdPrint(("\'SonyDCamReceiveCtrlPacket: Error %x while trying to Stop Isoch stream\n", Srb->Status));
                                IoFreeIrp(Irp);
                                break;

                            }

                        }

                    }

                    DeviceExtension->bCameraStopped = FALSE;
                    break;

                case KSSTATE_RUN:

                    KdPrint(("\'SonyDCamReceiveCtrlPacket: setting RUN stream state\n"));

                    // Remember the time at which the stream was started

                    DeviceExtension->QST_Now = GetSystemTime();
                    DeviceExtension->QST_Start = DeviceExtension->QST_Now;

                    StackSize = DeviceExtension->BusDeviceObject->StackSize;
                    Irp = IoAllocateIrp(StackSize, FALSE);

                    if (!Irp) {
    
                        SONYDCAM_ASSERT(Irp);
                        Srb->Status = STATUS_INSUFFICIENT_RESOURCES;
                        break;

                    }

                    DeviceExtension->bCameraStopped = FALSE;
                    Irb->FunctionNumber = REQUEST_ISOCH_LISTEN;
                    Irb->Flags = 0;
                    Irb->u.IsochListen.nChannel = DeviceExtension->IsochChannel;
                    Irb->u.IsochListen.hResource = DeviceExtension->hResource;
                    Irb->u.IsochListen.fulFlags = 0;
                    Irb->u.IsochListen.ulSynch = 0;
                    Irb->u.IsochListen.ulTag = 0;

                    Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

                    //
                    // If we get an insufficient resources error that means
                    // we don't have any Reads down yet.  Set flag to TRUE
                    // indicating that when we do get a Read down we'll 
                    // actually need to begin the listening process.
                    //

                    if (Srb->Status == STATUS_INSUFFICIENT_RESOURCES) {

                        DeviceExtension->bNeedToListen = TRUE;

                    }
                        
                    if ((Srb->Status) &&
                        (Srb->Status != STATUS_INSUFFICIENT_RESOURCES)) {

                        KdPrint(("\'SonyDCamReceiveCtrlPacket: Error %x while trying to start listening\n", Srb->Status));
                        IoFreeIrp(Irp);
                        break;

                    }

                    //
                    // Four writes to the registers on the camera will 
                    // get it going
                    //

                    deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
                    deltaTime.HighPart = -1;
                    KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);
                    DeviceExtension->RegisterWorkArea.AsULONG = DeviceExtension->FrameRate << FRAME_RATE_SHIFT;
                    Irb->FunctionNumber = REQUEST_ASYNC_WRITE;
                    Irb->Flags = 0;
                    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_High = INITIAL_REGISTER_SPACE_HI;
                    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low = DeviceExtension->BaseRegister + FIELDOFFSET(CAMERA_REGISTER_MAP, CurrentVFrmRate);
                    Irb->u.AsyncWrite.nNumberOfBytesToWrite = sizeof(ULONG);
                    Irb->u.AsyncWrite.nBlockSize = 0;
                    Irb->u.AsyncWrite.fulFlags = 0;
                    Irb->u.AsyncWrite.ulGeneration = DeviceExtension->CurrentGeneration;
                    Irb->u.AsyncWrite.Mdl = Mdl = MmCreateMdl(NULL, &DeviceExtension->RegisterWorkArea, sizeof(ULONG));
                    MmBuildMdlForNonPagedPool(Irb->u.AsyncWrite.Mdl);
                    KdPrint(("\'SonyDCamReceiveCtrlPacket: current generation = %d\n", DeviceExtension->CurrentGeneration));

                    Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

                    if (Srb->Status) {

                        KdPrint(("\'SonyDCamReceiveCtrlPacket: Error %x while writing to frame rate register\n", Srb->Status));
                        ExFreePool(Mdl);
                        IoFreeIrp(Irp);
                        break;

                    }

                    deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
                    deltaTime.HighPart = -1;
                    KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);
                    DeviceExtension->RegisterWorkArea.AsULONG = DeviceExtension->CurrentModeIndex << VIDEO_MODE_SHIFT;
                    Irb->FunctionNumber = REQUEST_ASYNC_WRITE;
                    Irb->Flags = 0;
                    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_High = INITIAL_REGISTER_SPACE_HI;
                    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low = DeviceExtension->BaseRegister + FIELDOFFSET(CAMERA_REGISTER_MAP, CurrentVMode);
                    Irb->u.AsyncWrite.nNumberOfBytesToWrite = sizeof(ULONG);
                    Irb->u.AsyncWrite.nBlockSize = 0;
                    Irb->u.AsyncWrite.fulFlags = 0;
                    Irb->u.AsyncWrite.ulGeneration = DeviceExtension->CurrentGeneration;
                    KdPrint(("\'SonyDCamReceiveCtrlPacket: current generation = %d\n", DeviceExtension->CurrentGeneration));

                    Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

                    if (Srb->Status) {

                        KdPrint(("\'SonyDCamReceiveCtrlPacket: Error %x while writing to video mode register\n", Srb->Status));
                        ExFreePool(Mdl);
                        IoFreeIrp(Irp);
                        break;

                    }

                    deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
                    deltaTime.HighPart = -1;
                    KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);
                    DeviceExtension->RegisterWorkArea.AsULONG = FORMAT_VGA_NON_COMPRESSED;
                    Irb->FunctionNumber = REQUEST_ASYNC_WRITE;
                    Irb->Flags = 0;
                    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_High = INITIAL_REGISTER_SPACE_HI;
                    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low = DeviceExtension->BaseRegister + FIELDOFFSET(CAMERA_REGISTER_MAP, CurrentVFormat);
                    Irb->u.AsyncWrite.nNumberOfBytesToWrite = sizeof(ULONG);
                    Irb->u.AsyncWrite.nBlockSize = 0;
                    Irb->u.AsyncWrite.fulFlags = 0;
                    Irb->u.AsyncWrite.ulGeneration = DeviceExtension->CurrentGeneration;
                    KdPrint(("\'SonyDCamReceiveCtrlPacket: current generation = %d\n", DeviceExtension->CurrentGeneration));

                    Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

                    if (Srb->Status) {

                        KdPrint(("\'SonyDCamReceiveCtrlPacket: Error %x while writing to Iso Channel register\n", Srb->Status));
                        ExFreePool(Mdl);
                        IoFreeIrp(Irp);
                        break;

                    }

                    deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
                    deltaTime.HighPart = -1;
                    KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);
                    DeviceExtension->RegisterWorkArea.AsULONG = DeviceExtension->IsochChannel << ISOCH_CHANNEL_SHIFT;
                    Irb->FunctionNumber = REQUEST_ASYNC_WRITE;
                    Irb->Flags = 0;
                    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_High = INITIAL_REGISTER_SPACE_HI;
                    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low = DeviceExtension->BaseRegister + FIELDOFFSET(CAMERA_REGISTER_MAP, IsoChannel);
                    Irb->u.AsyncWrite.nNumberOfBytesToWrite = sizeof(ULONG);
                    Irb->u.AsyncWrite.nBlockSize = 0;
                    Irb->u.AsyncWrite.fulFlags = 0;
                    Irb->u.AsyncWrite.ulGeneration = DeviceExtension->CurrentGeneration;
                    KdPrint(("\'SonyDCamReceiveCtrlPacket: current generation = %d\n", DeviceExtension->CurrentGeneration));

                    Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

                    if (Srb->Status) {

                        KdPrint(("\'SonyDCamReceiveCtrlPacket: Error %x while writing to Iso Channel register\n", Srb->Status));
                        ExFreePool(Mdl);
                        IoFreeIrp(Irp);
                        break;

                    }

                    deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
                    deltaTime.HighPart = -1;
                    KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);
                    DeviceExtension->RegisterWorkArea.AsULONG = START_ISOCH_TRANSMISSION;
                    Irb->FunctionNumber = REQUEST_ASYNC_WRITE;
                    Irb->Flags = 0;
                    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_High = INITIAL_REGISTER_SPACE_HI;
                    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low = DeviceExtension->BaseRegister + FIELDOFFSET(CAMERA_REGISTER_MAP, IsoEnable);
                    Irb->u.AsyncWrite.nNumberOfBytesToWrite = sizeof(ULONG);
                    Irb->u.AsyncWrite.nBlockSize = 0;
                    Irb->u.AsyncWrite.fulFlags = 0;
                    Irb->u.AsyncWrite.ulGeneration = DeviceExtension->CurrentGeneration;
                    KdPrint(("\'SonyDCamReceiveCtrlPacket: current generation = %d\n", DeviceExtension->CurrentGeneration));

                    Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

                    if (Srb->Status) {

                        KdPrint(("\'SonyDCamReceiveCtrlPacket: Error %x while writing to Iso Enable register\n", Srb->Status));
                        ExFreePool(Mdl);
                        IoFreeIrp(Irp);
                        break;

                    }

                    ExFreePool(Mdl);
                    IoFreeIrp(Irp);
                    DeviceExtension->bCameraSpewing = TRUE;
                    break;

                case KSSTATE_ACQUIRE:

                    Srb->Status = STATUS_SUCCESS;
                    break;

                default:

                    KdPrint(("\'SonyDCamReceiveCtrlPacket: Error unknown state\n"));
                    Srb->Status = STATUS_NOT_IMPLEMENTED;
                    break;

            }

            DeviceExtension->KSState = Srb->CommandData.StreamState;
            break;

        case SRB_GET_STREAM_PROPERTY:

            KdPrint(("\'SonyDCamReceiveDataPacket: SRB_GET_STREAM_PROPERTY\n"));
            VideoGetProperty(Srb);
            break;

        case SRB_OPEN_MASTER_CLOCK:
        case SRB_CLOSE_MASTER_CLOCK:

            //
            // This stream is being selected to provide a Master clock
            //

            VideoOpenCloseMasterClock (Srb);

            break;

        case SRB_INDICATE_MASTER_CLOCK:

            //
            // Assigns a clock to a stream
            //

            VideoIndicateMasterClock (Srb);

            break;

        default:

            //
            // invalid / unsupported command. Fail it as such
            //
    
            Srb->Status = STATUS_NOT_IMPLEMENTED;

    }

    StreamClassStreamNotification(StreamRequestComplete, Srb->StreamObject, Srb);

}


VOID
SonyDCamReadStream(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    Called when an Read Data Srb request is received

Arguments:

    Srb - Pointer to Stream request block

Return Value:

    Nothing

--*/

{

    PIRB Irb;
    PIRP Irp;
    KIRQL oldIrql;
    PSONYDCAM_EXTENSION DeviceExtension;
    PISOCH_DESCRIPTOR IsochDescriptor;
    PISOCH_DESCRIPTOR_RESERVED IsochDescriptorReserved;


    KdPrint(("\'SonyDCamReadStream: enter with Srb = %x\n", Srb));
    Irb = (PIRB) Srb->SRBExtension;
    Irp = Srb->Irp;
    DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;

    //
    // If the camera is stopped right now, return an error.
    //

    if (DeviceExtension->bCameraStopped) {

        Srb->Status = STATUS_UNSUCCESSFUL;
        Srb->CommandData.DataBufferArray->DataUsed = 0;
        StreamClassStreamNotification(StreamRequestComplete, Srb->StreamObject, Srb);
        return;
        
    }

    IsochDescriptor = ExAllocatePool(NonPagedPool, sizeof(ISOCH_DESCRIPTOR));
    if (!IsochDescriptor) {

        SONYDCAM_ASSERT(FALSE);
        Srb->Status = STATUS_INSUFFICIENT_RESOURCES;
        StreamClassStreamNotification(StreamRequestComplete, Srb->StreamObject, Srb);
        return;

    }

    KdPrint(("\'SonyDCamReadStream: IsochDescriptor = %x\n", IsochDescriptor));
    IsochDescriptor->fulFlags = SYNCH_ON_SY;
    KdPrint(("\'SonyDCamReadStream: Incoming Mdl = %x\n", Irp->MdlAddress));

    IsochDescriptor->Mdl = Irp->MdlAddress;
    IsochDescriptor->ulLength = Srb->CommandData.DataBufferArray->FrameExtent;
    IsochDescriptor->nMaxBytesPerFrame = IsochInfoTable[DeviceExtension->CurrentModeIndex*NUM_POSSIBLE_RATES+DeviceExtension->FrameRate].QuadletPayloadPerPacket << 2;
    IsochDescriptor->ulSynch = START_OF_PICTURE;
    IsochDescriptor->ulTag = 0;
    IsochDescriptor->Callback = SonyDCamIsochCallback;
    IsochDescriptor->Context1 = DeviceExtension;
    IsochDescriptor->Context2 = IsochDescriptor;
    IsochDescriptorReserved = (PISOCH_DESCRIPTOR_RESERVED) &IsochDescriptor->DeviceReserved[0];
    IsochDescriptorReserved->Srb = Srb;
    IsochDescriptorReserved->Flags = 0;

    //
    // Checking here to see if we have enuff resources to put this read
    // down right away.  Since we only allocated N amount of resources
    // from the 1394 stack beneath us, we'll have to stay within that 
    // limit and do some of the throttling ourself.
    //

    KeAcquireSpinLock(&DeviceExtension->IsochWaitingLock, &oldIrql);
    if (InterlockedIncrement(&DeviceExtension->PendingReadCount) > MAX_BUFFERS_SUPPLIED) {

        //
        // don't have enuff resources to do an attach buffers right now.
        // we'll queue this request and pull it off later when another
        // read completes.
        //

        KdPrint(("\'SonyDCamReadStream: Queueing request - Read Count = %x\n", DeviceExtension->PendingReadCount));
        InsertTailList(
           &DeviceExtension->IsochWaitingList,
           &IsochDescriptorReserved->DescriptorList
            );

        KeReleaseSpinLock(&DeviceExtension->IsochWaitingLock, oldIrql);
        return;

    }

    //
    // Do actual read work here via our Read worker function
    //

    KeReleaseSpinLock(&DeviceExtension->IsochWaitingLock, oldIrql);
    SonyDCamReadStreamWorker(Srb, IsochDescriptor);

}


VOID
SonyDCamReadStreamWorker(
    IN PHW_STREAM_REQUEST_BLOCK Srb,
    IN PISOCH_DESCRIPTOR IsochDescriptor
    )

/*++

Routine Description:

    Does most of the work for handling reads via Attach buffers

Arguments:

    Srb - Pointer to Stream request block

    IsochDescriptor - Pointer to IsochDescriptor to be used

Return Value:

    Nothing

--*/

{

    PIRB Irb;
    PIRP Irp;
    KIRQL oldIrql;
    NTSTATUS status;
    PIO_STACK_LOCATION NextIrpStack;
    PSONYDCAM_EXTENSION DeviceExtension;
    PISOCH_DESCRIPTOR_RESERVED IsochDescriptorReserved;


    Irb = (PIRB) Srb->SRBExtension;
    Irp = Srb->Irp;
    Srb->Status = STATUS_PENDING;
    DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;
    IsochDescriptorReserved = (PISOCH_DESCRIPTOR_RESERVED) &IsochDescriptor->DeviceReserved[0];

    //
    // Attach descriptor onto our pending descriptor list
    //

    ExInterlockedInsertTailList(
       &DeviceExtension->IsochDescriptorList,
       &IsochDescriptorReserved->DescriptorList,
       &DeviceExtension->IsochDescriptorLock
        );

    Irb->FunctionNumber = REQUEST_ISOCH_ATTACH_BUFFERS;
    Irb->Flags = 0;
    Irb->u.IsochAttachBuffers.hResource = DeviceExtension->hResource;
    Irb->u.IsochAttachBuffers.nNumberOfDescriptors = 1;
    Irb->u.IsochAttachBuffers.pIsochDescriptor = IsochDescriptor;

    status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

    if ((status) &&
        (status != STATUS_PENDING)) {

        KdPrint(("\'SonyDCamReadStreamWorker: Error %x while trying to attach Isoch buffers\n", status));

        if (!(IsochDescriptorReserved->Flags & STATE_SRB_IS_COMPLETE)) {

            IsochDescriptorReserved->Flags |= STATE_SRB_IS_COMPLETE;
            Srb->Status = status;
            KeAcquireSpinLock(&DeviceExtension->IsochDescriptorLock, &oldIrql);
            RemoveEntryList(&IsochDescriptorReserved->DescriptorList);
            KeReleaseSpinLock(&DeviceExtension->IsochDescriptorLock, oldIrql);
            ExFreePool(IsochDescriptor);
            StreamClassStreamNotification(StreamRequestComplete, Srb->StreamObject, Srb);

            if (InterlockedDecrement(&DeviceExtension->PendingReadCount) >= MAX_BUFFERS_SUPPLIED) {

                //
                // We had someone blocked waiting for us to complete.  Pull
                // them off the waiting list and get them running
                //

                KdPrint(("\'SonyDCamReadStreamWorker: Dequeueing request - Read Count = %x\n", DeviceExtension->PendingReadCount));
                SONYDCAM_ASSERT ( !IsListEmpty(&DeviceExtension->IsochWaitingList) );
                IsochDescriptorReserved = \
                    (PISOCH_DESCRIPTOR_RESERVED) ExInterlockedRemoveHeadList(
                       &DeviceExtension->IsochWaitingList,
                       &DeviceExtension->IsochWaitingLock
                        );

                IsochDescriptor = \
                    (PISOCH_DESCRIPTOR) (((PUCHAR) IsochDescriptorReserved) - 
                        FIELDOFFSET(ISOCH_DESCRIPTOR, DeviceReserved[0]));

                SonyDCamReadStreamWorker(IsochDescriptorReserved->Srb, IsochDescriptor);

            }

        }

        return;

    }

    if (DeviceExtension->bNeedToListen) {

        DeviceExtension->bNeedToListen = FALSE;
        DeviceExtension->Irb->FunctionNumber = REQUEST_ISOCH_LISTEN;
        DeviceExtension->Irb->Flags = 0;
        DeviceExtension->Irb->u.IsochListen.nChannel = DeviceExtension->IsochChannel;
        DeviceExtension->Irb->u.IsochListen.hResource = DeviceExtension->hResource;
        DeviceExtension->Irb->u.IsochListen.fulFlags = 0;
        DeviceExtension->Irb->u.IsochListen.ulSynch = 0;
        DeviceExtension->Irb->u.IsochListen.ulTag = 0;

        NextIrpStack = IoGetNextIrpStackLocation(DeviceExtension->Irp);
        NextIrpStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
        NextIrpStack->Parameters.DeviceIoControl.IoControlCode = IOCTL_1394_CLASS;
        NextIrpStack->Parameters.Others.Argument1 = DeviceExtension->Irb;

        IoSetCompletionRoutine(
            DeviceExtension->Irp,
            SonyDCamListenCompletionRoutine,
            NULL,
            TRUE,
            TRUE,
            TRUE
            );

        IoCallDriver(DeviceExtension->BusDeviceObject, DeviceExtension->Irp);

    }

}


VOID
SonyDCamIsochCallback(
    IN PSONYDCAM_EXTENSION DeviceExtension,
    IN PISOCH_DESCRIPTOR IsochDescriptor
    )

/*++

Routine Description:

    Called when an Isoch Descriptor completes

Arguments:

    DeviceExtension - Pointer to our DeviceExtension

    IsochDescriptor - IsochDescriptor that completed

Return Value:

    Nothing

--*/

{

    PIRB Irb;
    PIRP Irp;
    KIRQL oldIrql;
    PIO_STACK_LOCATION NextIrpStack;
    PISOCH_DESCRIPTOR_RESERVED IsochDescriptorReserved;
    PKSSTREAM_HEADER pDataPacket;
    REFERENCE_TIME              WallTime100ns;       // time since capture started based on PerfCounter


    //
    // Debug check to make sure we're dealing with a real IsochDescriptor
    //

    SONYDCAM_ASSERT ( IsochDescriptor );

    IsochDescriptorReserved = (PISOCH_DESCRIPTOR_RESERVED) &IsochDescriptor->DeviceReserved[0];
    DeviceExtension->Frames++;

    //
    // Make sure this IsochDescriptor is not already marked as completed
    //

    if (!(IsochDescriptorReserved->Flags & STATE_SRB_IS_COMPLETE)) {

        //
        // Return the timestamp for the frame
        //

        pDataPacket = IsochDescriptorReserved->Srb->CommandData.DataBufferArray;
        DeviceExtension->QST_Now = GetSystemTime();
        WallTime100ns = DeviceExtension->QST_Now - DeviceExtension->QST_Start;
        
        if (DeviceExtension->hMasterClock) {

            pDataPacket->PresentationTime.Time = WallTime100ns;
//            pDataPacket->PresentationTime.Time = DeviceExtension->pVideoInfoHeader->AvgTimePerFrame * DeviceExtension->Frames;
            pDataPacket->PresentationTime.Numerator = 1;
            pDataPacket->PresentationTime.Denominator = 1;
            pDataPacket->Duration = DeviceExtension->pVideoInfoHeader->AvgTimePerFrame;

        }

        else {

#ifdef TODO
// BUGBUG If we're not the master clock, we need to query the master clock
// to get a timestamp
            StreamClassQueryMasterClock(
                IN PHW_STREAM_OBJECT HwStreamObject,
                IN HANDLE MasterClockHandle,
                IN TIME_FUNCTION TimeFunction,
                IN PHW_QUERY_CLOCK_ROUTINE ClockCallbackRoutine);
#endif

        }

        pDataPacket->OptionsFlags |= \
            KSSTREAM_HEADER_OPTIONSF_TIMEVALID |
            KSSTREAM_HEADER_OPTIONSF_DURATIONVALID;

        // Every frame we generate is a key frame (aka SplicePoint)
        pDataPacket->OptionsFlags |= KSSTREAM_HEADER_OPTIONSF_SPLICEPOINT;

        //
        // Indicate that the Srb should be complete
        //

        IsochDescriptorReserved->Flags |= STATE_SRB_IS_COMPLETE;
        IsochDescriptorReserved->Srb->Status = STATUS_SUCCESS;
        IsochDescriptorReserved->Srb->CommandData.DataBufferArray->DataUsed = IsochDescriptor->ulLength;
        IsochDescriptorReserved->Srb->ActualBytesTransferred = IsochDescriptor->ulLength;
        KdPrint(("\'SonyDCamIsochCallback: Completing Srb %x\n", IsochDescriptorReserved->Srb));

        KeAcquireSpinLock(&DeviceExtension->IsochDescriptorLock, &oldIrql);
        RemoveEntryList(&IsochDescriptorReserved->DescriptorList);
        KeReleaseSpinLock(&DeviceExtension->IsochDescriptorLock, oldIrql);
        StreamClassStreamNotification(StreamRequestComplete, IsochDescriptorReserved->Srb->StreamObject, IsochDescriptorReserved->Srb);

        //
        // Indicate that we're in the Detachment phase
        //

        IsochDescriptorReserved->Flags |= STATE_DETACHING_BUFFERS;

        Irb = DeviceExtension->Irb;
        Irp = DeviceExtension->Irp;
        Irb->FunctionNumber = REQUEST_ISOCH_DETACH_BUFFERS;
        Irb->u.IsochDetachBuffers.hResource = DeviceExtension->hResource;
        Irb->u.IsochDetachBuffers.nNumberOfDescriptors = 1;
        Irb->u.IsochDetachBuffers.pIsochDescriptor = IsochDescriptor;

        NextIrpStack = IoGetNextIrpStackLocation(Irp);
        NextIrpStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
        NextIrpStack->Parameters.DeviceIoControl.IoControlCode = IOCTL_1394_CLASS;
        NextIrpStack->Parameters.Others.Argument1 = Irb;

        IoSetCompletionRoutine(
            Irp,
            SonyDCamCompletionRoutine,
            IsochDescriptor,
            TRUE,
            TRUE,
            TRUE
            );
    
        IoCallDriver(DeviceExtension->BusDeviceObject, Irp);

        if (InterlockedDecrement(&DeviceExtension->PendingReadCount) >= MAX_BUFFERS_SUPPLIED) {

            //
            // We had someone blocked waiting for us to complete.  Pull
            // them off the waiting list and get them running
            //

            KdPrint(("\'SonyDCamIsochCallback: Dequeueing request - Read Count = %x\n", DeviceExtension->PendingReadCount));
            SONYDCAM_ASSERT ( !IsListEmpty(&DeviceExtension->IsochWaitingList) );
            IsochDescriptorReserved = \
                (PISOCH_DESCRIPTOR_RESERVED) ExInterlockedRemoveHeadList(
                   &DeviceExtension->IsochWaitingList,
                   &DeviceExtension->IsochWaitingLock
                    );

            IsochDescriptor = \
                (PISOCH_DESCRIPTOR) (((PUCHAR) IsochDescriptorReserved) - 
                    FIELDOFFSET(ISOCH_DESCRIPTOR, DeviceReserved[0]));

            SonyDCamReadStreamWorker(IsochDescriptorReserved->Srb, IsochDescriptor);

        }

    }
    
}


NTSTATUS
SonyDCamCompletionRoutine(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PISOCH_DESCRIPTOR IsochDescriptor
    )

/*++

Routine Description:

    Returns more processing required so the IO Manager will leave us alone

Arguments:

    DriverObject - Pointer to driver object created by system.

    Irp - Irp that just completed

    IsochDescriptor - This Isoch descriptor will be freed

Return Value:

    None.

--*/

{

    ExFreePool(IsochDescriptor);
    return (STATUS_MORE_PROCESSING_REQUIRED);

}


VOID
SonyDCamCancelPacket(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    Cancel routine for Srbs (not implemented yet)

Arguments:

    Srb - Pointer to the stream request block

Return Value:

    None.

--*/

{

    return;

}


VOID
SonyDCamTimeoutHandler(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    This routine is called when a packet has been in the minidriver too long

Arguments:

    Srb - Pointer to Stream request block

Return Value:

    Nothing

--*/

{

    KdPrint(("\'SonyDCamTimeoutHandler: TIMEOUT!  Srb %x hasn't completed!\n", Srb));

}


NTSTATUS
SonyDCamListenCompletionRoutine(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID SystemArgument
    )

/*++

Routine Description:

    Returns more processing required so the IO Manager will leave us alone

Arguments:

    DriverObject - Pointer to driver object created by system.

    Irp - Irp that just completed

    SystemArgument - Pointer to NULL

Return Value:

    None.

--*/

{

    return (STATUS_MORE_PROCESSING_REQUIRED);

}


VOID
SonyDCamStopStream(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    Called to stop the stream (if running)

Arguments:

    Srb - Pointer to Stream request block

Return Value:

    Nothing

--*/

{

    PIRB Irb;
    PIRP Irp;
    PMDL Mdl;
    CCHAR StackSize;
    PSONYDCAM_EXTENSION DeviceExtension;
    PISOCH_DESCRIPTOR IsochDescriptor;
    PISOCH_DESCRIPTOR_RESERVED IsochDescriptorReserved;
    

    //
    // Get the Irb and DeviceExtension out of the Srb
    //

    Irb = (PIRB) Srb->SRBExtension;
    DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;

    StackSize = DeviceExtension->BusDeviceObject->StackSize;
    Irp = IoAllocateIrp(StackSize, FALSE);

    if (!Irp) {
    
        SONYDCAM_ASSERT(Irp);
        Srb->Status = STATUS_INSUFFICIENT_RESOURCES;
        return;

    }

    //
    // First stop the stream internally inside the PC's 1394
    // stack
    //

    if (DeviceExtension->hResource) {

        Irb->FunctionNumber = REQUEST_ISOCH_STOP;
        Irb->Flags = 0;
        Irb->u.IsochStop.nChannel = DeviceExtension->IsochChannel;
        Irb->u.IsochStop.hResource = DeviceExtension->hResource;
        Irb->u.IsochStop.fulFlags = 0;

        Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

        if (Srb->Status) {

            KdPrint(("\'SonyDCamStopStream: Error %x while trying to Stop Isoch stream\n", Srb->Status));
            IoFreeIrp(Irp);
            return;

        }

    }

    //
    // Now stop the stream at the device itself
    //

    DeviceExtension->RegisterWorkArea.AsULONG = STOP_ISOCH_TRANSMISSION;
    Irb->FunctionNumber = REQUEST_ASYNC_WRITE;
    Irb->Flags = 0;
    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_High = \
        INITIAL_REGISTER_SPACE_HI;
    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low = \
        DeviceExtension->BaseRegister +
        FIELDOFFSET(CAMERA_REGISTER_MAP, IsoEnable);
    Irb->u.AsyncWrite.nNumberOfBytesToWrite = sizeof(ULONG);
    Irb->u.AsyncWrite.nBlockSize = 0;
    Irb->u.AsyncWrite.fulFlags = 0;
    Irb->u.AsyncWrite.ulGeneration = DeviceExtension->CurrentGeneration;
    Irb->u.AsyncWrite.Mdl = Mdl = \
        MmCreateMdl(
            NULL,
           &DeviceExtension->RegisterWorkArea,
            sizeof(ULONG)
            );
    MmBuildMdlForNonPagedPool(Irb->u.AsyncWrite.Mdl);

    Srb->Status = SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);
    
    ExFreePool(Irb->u.AsyncWrite.Mdl);

    if (Srb->Status) {
    
        //
        // An error is acceptable here because our camera might have been
        // removed from the bus
        //

        KdPrint(("\'SonyDCamStopStream: Error %x while stopping at Iso Enable register\n", Srb->Status));

    }

    //
    // Detach all buffers that might still be attached.
    //

    while (!IsListEmpty(&DeviceExtension->IsochDescriptorList)) {

        IsochDescriptorReserved = \
            (PISOCH_DESCRIPTOR_RESERVED) ExInterlockedRemoveHeadList(
               &DeviceExtension->IsochDescriptorList,
               &DeviceExtension->IsochDescriptorLock
                );

        IsochDescriptor = \
            (PISOCH_DESCRIPTOR) (((PUCHAR) IsochDescriptorReserved) -
                FIELDOFFSET(ISOCH_DESCRIPTOR, DeviceReserved[0]));

        if (!(IsochDescriptorReserved->Flags & STATE_SRB_IS_COMPLETE)) {

            KdPrint(("\'SonyDCamStopStream: stopping IsochDescriptor %x Srb %x\n", IsochDescriptor, IsochDescriptorReserved->Srb));

            IsochDescriptorReserved->Flags |= STATE_SRB_IS_COMPLETE;
            IsochDescriptorReserved->Srb->Status = STATUS_UNSUCCESSFUL;
            IsochDescriptorReserved->Srb->CommandData.DataBufferArray->DataUsed = 0;
            StreamClassStreamNotification(
                StreamRequestComplete,
                IsochDescriptorReserved->Srb->StreamObject,
                IsochDescriptorReserved->Srb
                );

            IsochDescriptorReserved->Flags |= STATE_DETACHING_BUFFERS;
            Irb->FunctionNumber = REQUEST_ISOCH_DETACH_BUFFERS;
            Irb->Flags = 0;
            Irb->u.IsochDetachBuffers.hResource = DeviceExtension->hResource;
            Irb->u.IsochDetachBuffers.nNumberOfDescriptors = 1;
            Irb->u.IsochDetachBuffers.pIsochDescriptor = IsochDescriptor;

            SonyDCamSubmitIrpSynch(DeviceExtension, Irp, Irb);

            ExFreePool(IsochDescriptor);

            if (InterlockedDecrement(&DeviceExtension->PendingReadCount) >= MAX_BUFFERS_SUPPLIED) {

                //
                // We had someone blocked waiting for us to complete.  Pull
                // them off the waiting list and get them running
                //

                KdPrint(("\'SonyDCamStopStream: Dequeueing request - Read Count = %x\n", DeviceExtension->PendingReadCount));
                SONYDCAM_ASSERT ( !IsListEmpty(&DeviceExtension->IsochWaitingList) );
                IsochDescriptorReserved = \
                    (PISOCH_DESCRIPTOR_RESERVED) ExInterlockedRemoveHeadList(
                       &DeviceExtension->IsochWaitingList,
                       &DeviceExtension->IsochWaitingLock
                        );

                IsochDescriptor = \
                    (PISOCH_DESCRIPTOR) (((PUCHAR) IsochDescriptorReserved) - 
                        FIELDOFFSET(ISOCH_DESCRIPTOR, DeviceReserved[0]));

                SonyDCamReadStreamWorker(IsochDescriptorReserved->Srb, IsochDescriptor);

            }

        }

    }

    IoFreeIrp(Irp);
    DeviceExtension->bCameraStopped = TRUE;

}


/*
** AdapterCompareGUIDsAndFormatSize()
**
**   Checks for a match on the three GUIDs and FormatSize
**
** Arguments:
**
**         IN DataRange1
**         IN DataRange2
**
** Returns:
** 
**   TRUE if all elements match
**   FALSE if any are different
**
** Side Effects:  none
*/

BOOL 
AdapterCompareGUIDsAndFormatSize(
    IN PKSDATARANGE DataRange1,
    IN PKSDATARANGE DataRange2)
{
    return (
        IsEqualGUID (
            &DataRange1->MajorFormat, 
            &DataRange2->MajorFormat) &&
        IsEqualGUID (
            &DataRange1->SubFormat, 
            &DataRange2->SubFormat) &&
        IsEqualGUID (
            &DataRange1->Specifier, 
            &DataRange2->Specifier) &&
        (DataRange1->FormatSize == DataRange2->FormatSize));
}

/*
** AdapterVerifyFormat()
**
**   Checks the validity of a format request by walking through the
**       array of supported KSDATA_RANGEs for a given stream.
**
** Arguments:
**
**   pKSDataFormat - pointer of a KS_DATAFORMAT_VIDEOINFOHEADER structure.
**   StreamNumber - index of the stream being queried / opened.
**
** Returns:
** 
**   TRUE if the format is supported
**   FALSE if the format cannot be suppored
**
** Side Effects:  none
*/

BOOL 
AdapterVerifyFormat(
        PKS_DATAFORMAT_VIDEOINFOHEADER pKSDataFormatToVerify, 
        int StreamNumber)
{
    PKS_VIDEOINFOHEADER                 pVideoInfoHdrToVerify = 
                &pKSDataFormatToVerify->VideoInfoHeader;
    PKS_VIDEOINFOHEADER                 pVideoInfoHdr;
    PKSDATAFORMAT                       *pAvailableFormats;
    PKS_DATARANGE_VIDEO                 pKSDataRange;
    KS_VIDEO_STREAM_CONFIG_CAPS         *pConfigCaps;
    int                                 NumberOfFormatArrayEntries;
    int                                 nSize;
    int                                 j;
    RECT                                rcImage;
     
    //
    // Make sure the stream index is valid
    //
    if (StreamNumber >= 1) {
        return FALSE;
    }

    //
    // How many formats does this stream support?
    //
    NumberOfFormatArrayEntries = 1;

    nSize = sizeof (KS_VIDEOINFOHEADER) + 
                pVideoInfoHdrToVerify->bmiHeader.biSize;

    DebugPrint((DebugLevelVerbose,"TestCap: AdapterVerifyFormat, Stream=%d\n",
            StreamNumber));
    DebugPrint((DebugLevelVerbose,"TestCap: FormatSize=%d\n", 
            pKSDataFormatToVerify->DataFormat.FormatSize));
    DebugPrint((DebugLevelVerbose,"TestCap: MajorFormat=%x\n", 
            pKSDataFormatToVerify->DataFormat.MajorFormat));

    
    DebugPrint((DebugLevelVerbose,"TestCap: pVideoInfoHdrToVerify=%x\n", 
                pVideoInfoHdrToVerify));
    DebugPrint((DebugLevelVerbose,"TestCap: KS_VIDEOINFOHEADER size =%d\n", 
                nSize));
    DebugPrint((DebugLevelVerbose,"TestCap: Width=%d Height=%d  biBitCount=%d\n", 
                pVideoInfoHdrToVerify->bmiHeader.biWidth, 
                pVideoInfoHdrToVerify->bmiHeader.biHeight,
                pVideoInfoHdrToVerify->bmiHeader.biBitCount));

    DebugPrint((DebugLevelVerbose,"TestCap: biSizeImage =%d\n", 
                pVideoInfoHdrToVerify->bmiHeader.biSizeImage));

    //
    // Get the pointer to the array of available formats
    //
    pAvailableFormats = &SONYDCAM_StreamFormats[0];

    //
    // Walk the array, searching for a match
    //
    for (j = 0; j < NumberOfFormatArrayEntries; j++, pAvailableFormats++) {
        pKSDataRange = (PKS_DATARANGE_VIDEO) *pAvailableFormats;
        pVideoInfoHdr = &pKSDataRange->VideoInfoHeader;
        pConfigCaps = &pKSDataRange->ConfigCaps;
        
        //
        // Check for matching size, Major Type, Sub Type, and Specifier
        //

        if (!IsEqualGUID (&pKSDataRange->DataRange.MajorFormat, 
            &pKSDataFormatToVerify->DataFormat.MajorFormat)) {
            break;
        }

        if (!IsEqualGUID (&pKSDataRange->DataRange.SubFormat, 
            &pKSDataFormatToVerify->DataFormat.SubFormat)) {
            break;
        }

        if (!IsEqualGUID (&pKSDataRange->DataRange.Specifier,
            &pKSDataFormatToVerify->DataFormat.Specifier)) {
            break;
        }

        /* 
        **  HOW BIG IS THE IMAGE REQUESTED (pseudocode follows)
        **
        **  if (IsRectEmpty (&rcTarget) {
        **      SetRect (&rcImage, 0, 0, 
        **              BITMAPINFOHEADER.biWidth,
                        BITMAPINFOHEADER.biHeight);
        **  }
        **  else {
        **      // Probably rendering to a DirectDraw surface,
        **      // where biWidth is used to expressed the "stride" 
        **      // in units of pixels (not bytes) of the destination surface.
        **      // Therefore, use rcTarget to get the actual image size 
        **      
        **      rcImage = rcTarget;
        **  }
        */
        
        if ((pVideoInfoHdrToVerify->rcTarget.right - 
             pVideoInfoHdrToVerify->rcTarget.left <= 0) ||
            (pVideoInfoHdrToVerify->rcTarget.bottom - 
             pVideoInfoHdrToVerify->rcTarget.top <= 0)) {
            
             rcImage.left = rcImage.top = 0;
             rcImage.right = pVideoInfoHdrToVerify->bmiHeader.biWidth;
             rcImage.bottom = pVideoInfoHdrToVerify->bmiHeader.biHeight;
        }
        else {
             rcImage = pVideoInfoHdrToVerify->rcTarget;
        }

        //
        // BUGBUG, perform all other verification tests here!!!
        //




        //
        // HOORAY, the format passed all of the tests, so we support it
        //

        return TRUE;
    } 

    //
    // The format requested didn't match any of our listed ranges,
    // so refuse the connection.
    //

    return FALSE;
}

/*
** AdapterFormatFromRange()
**
**   Returns a DATAFORMAT from a DATARANGE
**
** Arguments:
**
**         IN PHW_STREAM_REQUEST_BLOCK Srb 
**
** Returns:
** 
**   TRUE if the format is supported
**   FALSE if the format cannot be suppored
**
** Side Effects:  none
*/

BOOL 
AdapterFormatFromRange(
        IN PHW_STREAM_REQUEST_BLOCK Srb)
{
    PSONYDCAM_EXTENSION DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;
    PSTREAM_DATA_INTERSECT_INFO IntersectInfo;
    PKSDATARANGE                DataRange;
    BOOL                        OnlyWantsSize;
    BOOL                        MatchFound = FALSE;
    ULONG                       FormatSize;
    ULONG                       StreamNumber;
    ULONG                       j;
    ULONG                       NumberOfFormatArrayEntries;
    PKSDATAFORMAT               *pAvailableFormats;

    IntersectInfo = Srb->CommandData.IntersectInfo;
    StreamNumber = IntersectInfo->StreamNumber;
    DataRange = IntersectInfo->DataRange;

    //
    // Check that the stream number is valid
    //

    if (StreamNumber >= 1) {
        Srb->Status = STATUS_NOT_IMPLEMENTED;
        TRAP();
        return FALSE;
    }
    
    NumberOfFormatArrayEntries = 1;
    Srb->Status = STATUS_SUCCESS;

    //
    // Get the pointer to the array of available formats
    //

    pAvailableFormats = &SONYDCAM_StreamFormats[0];

    //
    // Is the caller trying to get the format, or the size of the format?
    //

    OnlyWantsSize = (IntersectInfo->SizeOfDataFormatBuffer == sizeof(ULONG));

    //
    // Walk the formats supported by the stream searching for a match
    // of the three GUIDs which together define a DATARANGE
    //

    for (j = 0; j < NumberOfFormatArrayEntries; j++, pAvailableFormats++) {

        if (!AdapterCompareGUIDsAndFormatSize(DataRange, *pAvailableFormats)) {
            continue;
        }

        //
        // Now that the three GUIDs match, switch on the Specifier
        // to do a further type specific check
        //

        // -------------------------------------------------------------------
        // Specifier FORMAT_VideoInfo for VIDEOINFOHEADER
        // -------------------------------------------------------------------

        if (IsEqualGUID (&DataRange->Specifier, 
                &KSDATAFORMAT_SPECIFIER_VIDEOINFO)) {
                
            PKS_DATARANGE_VIDEO DataRangeVideoToVerify = 
                    (PKS_DATARANGE_VIDEO) DataRange;
            PKS_DATARANGE_VIDEO DataRangeVideo = 
                    (PKS_DATARANGE_VIDEO) *pAvailableFormats;

            //
            // Check that the other fields match
            //
            if ((DataRangeVideoToVerify->bFixedSizeSamples != DataRangeVideo->bFixedSizeSamples) ||
                (DataRangeVideoToVerify->bTemporalCompression != DataRangeVideo->bTemporalCompression) ||
                (DataRangeVideoToVerify->StreamDescriptionFlags != DataRangeVideo->StreamDescriptionFlags) ||
                (DataRangeVideoToVerify->MemoryAllocationFlags != DataRangeVideo->MemoryAllocationFlags) ||
                (RtlCompareMemory (&DataRangeVideoToVerify->ConfigCaps,
                        &DataRangeVideo->ConfigCaps,
                        sizeof (KS_VIDEO_STREAM_CONFIG_CAPS)) != 
                        sizeof (KS_VIDEO_STREAM_CONFIG_CAPS))) {
                continue;
            }
            
            // MATCH FOUND!
            MatchFound = TRUE;            
            FormatSize = sizeof (KSDATAFORMAT) + 
                KS_SIZE_VIDEOHEADER (&DataRangeVideoToVerify->VideoInfoHeader);

            if (OnlyWantsSize) {
                break;
            }
            
            // Caller wants the full data format
            if (IntersectInfo->SizeOfDataFormatBuffer < FormatSize) {
                Srb->Status = STATUS_BUFFER_TOO_SMALL;
                return FALSE;
            }

            // Copy over the KSDATAFORMAT, followed by the 
            // actual VideoInfoHeader
                
            RtlCopyMemory(
                &((PKS_DATAFORMAT_VIDEOINFOHEADER)IntersectInfo->DataFormatBuffer)->DataFormat,
                &DataRangeVideoToVerify->DataRange,
                sizeof (KSDATARANGE));

            ((PKSDATAFORMAT)IntersectInfo->DataFormatBuffer)->FormatSize = FormatSize;

            RtlCopyMemory(
                &((PKS_DATAFORMAT_VIDEOINFOHEADER) IntersectInfo->DataFormatBuffer)->VideoInfoHeader,
                &DataRangeVideoToVerify->VideoInfoHeader,
                KS_SIZE_VIDEOHEADER (&DataRangeVideoToVerify->VideoInfoHeader));

            //
            // BUGBUG Validate format here, and calculate biSizeImage
            // 

            break;

        } // End of VIDEOINFOHEADER specifier

        else {
            Srb->Status = STATUS_NO_MATCH;
            return FALSE;
        }

    } // End of loop on all formats for this stream
    
    if (OnlyWantsSize) {
        *(PULONG) IntersectInfo->DataFormatBuffer = FormatSize;
        Srb->ActualBytesTransferred = sizeof(ULONG);
        return TRUE;
    }        

    Srb->ActualBytesTransferred = FormatSize;
    return TRUE;
}


/*
** VideoGetProperty()
**
**    Routine to process video property requests
**
** Arguments:
**
**    Srb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

VOID VideoGetProperty(PHW_STREAM_REQUEST_BLOCK Srb)
{
    PSTREAM_PROPERTY_DESCRIPTOR pSPD = Srb->CommandData.PropertyInfo;


    // preset to success

    Srb->Status = STATUS_SUCCESS;

    if (IsEqualGUID (&KSPROPSETID_Connection, &pSPD->Property->Set)) {

        VideoStreamGetConnectionProperty (Srb);

    }

    else {

       Srb->Status = STATUS_NOT_IMPLEMENTED;

    }

}


/*
** VideoGetState()
**
**    Gets the current state of the requested stream
**
** Arguments:
**
**    Srb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

VOID VideoGetState(PHW_STREAM_REQUEST_BLOCK Srb)
{
    PSONYDCAM_EXTENSION DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;

    Srb->CommandData.StreamState = DeviceExtension->KSState;
    Srb->ActualBytesTransferred = sizeof (KSSTATE);

    // A very odd rule:
    // When transitioning from stop to pause, DShow tries to preroll
    // the graph.  Capture sources can't preroll, and indicate this
    // by returning VFW_S_CANT_CUE in user mode.  To indicate this
    // condition from drivers, they must return ERROR_NO_DATA_DETECTED

    Srb->Status = STATUS_SUCCESS;

    if (DeviceExtension->KSState == KSSTATE_PAUSE) {
       Srb->Status = STATUS_NO_DATA_DETECTED;
    }
}


VOID  VideoStreamGetConnectionProperty (PHW_STREAM_REQUEST_BLOCK Srb)
{
    PSONYDCAM_EXTENSION DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;
    PSTREAM_PROPERTY_DESCRIPTOR pSPD = Srb->CommandData.PropertyInfo;
    ULONG Id = pSPD->Property->Id;              // index of the property

    switch (Id) {

    case KSPROPERTY_CONNECTION_ALLOCATORFRAMING:
        if (DeviceExtension->pVideoInfoHeader) {
            PKSALLOCATOR_FRAMING Framing = 
                (PKSALLOCATOR_FRAMING) pSPD->PropertyInfo;
            Framing->RequirementsFlags =
                KSALLOCATOR_REQUIREMENTF_SYSTEM_MEMORY |
                KSALLOCATOR_REQUIREMENTF_INPLACE_MODIFIER |
                KSALLOCATOR_REQUIREMENTF_PREFERENCES_ONLY;
            Framing->PoolType = PagedPool;
            Framing->Frames = 1;
            Framing->FrameSize = DeviceExtension->pVideoInfoHeader->bmiHeader.biSizeImage;
            Framing->FileAlignment = 0; // FILE_LONG_ALIGNMENT;
            Framing->Reserved = 0;
            Srb->ActualBytesTransferred = sizeof (KSALLOCATOR_FRAMING);
        }
        else {
            Srb->Status = STATUS_INVALID_PARAMETER;
        }
        break;

    default:
        TRAP();
        break;
    }
}

//==========================================================================;
//                   Clock Handling Routines
//==========================================================================;

//
// We are being opened or closed as a Master clock
// hMasterClock will be NULL if we're being closed
//

VOID VideoOpenCloseMasterClock (PHW_STREAM_REQUEST_BLOCK Srb)
{
    PSONYDCAM_EXTENSION DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;

    DeviceExtension->hMasterClock = Srb->CommandData.MasterClockHandle;
}

//
// Another clock is being assigned as the Master clock
//

VOID VideoIndicateMasterClock (PHW_STREAM_REQUEST_BLOCK Srb)
{
    PSONYDCAM_EXTENSION DeviceExtension = (PSONYDCAM_EXTENSION) Srb->HwDeviceExtension;

    DeviceExtension->hClock = Srb->CommandData.MasterClockHandle;
}

//
// GetSystemTime in 100 nS units
//

ULONGLONG GetSystemTime()
{
    ULONGLONG ticks;
    ULONGLONG rate;

    ticks = (ULONGLONG)KeQueryPerformanceCounter((PLARGE_INTEGER)&rate).QuadPart;

    //
    // convert from ticks to 100ns clock
    //

    ticks = (ticks & 0xFFFFFFFF00000000) / rate * 10000000 +
    		(ticks & 0xFFFFFFFF) * 10000000 / rate;

    return(ticks);

}

//
// This routine is called whenever someone in the graph wants to know
// what time it is, and we are the Master Clock.
//

VOID STREAMAPI StreamClockRtn(IN PHW_TIME_CONTEXT TimeContext)
{
    PSONYDCAM_EXTENSION DeviceExtension = (PSONYDCAM_EXTENSION) TimeContext->HwDeviceExtension;

    DeviceExtension->QST_Now = GetSystemTime();

    switch (TimeContext->Function) {
    
    case TIME_GET_STREAM_TIME:

        //
        // How long since the stream was first set into the run state?
        //

        TimeContext->Time = DeviceExtension->QST_Now - DeviceExtension->QST_Start;
        TimeContext->SystemTime = DeviceExtension->QST_Now;
        break;

    case TIME_READ_ONBOARD_CLOCK:

        //
        // Devices which provide a VSYNC interrupt would create a clock based on
        // VSYNC interrupts here.  Testcap just uses the system clock.
        //
        TimeContext->Time = DeviceExtension->QST_Now;
        TimeContext->SystemTime = DeviceExtension->QST_Now;
        break;
    
    default:
        TRAP();
        break;
    } // switch TimeContext->Function

}




