/*++

Copyright (c) 1996 Microsoft Corporation

Module Name:

    sonydcam.c

Abstract:

    This is a driver for the Sony Desktop Camera (CCM-DS250 V1.0x).
    This driver fits under the WDM stream class.

Author:
    
    Shaun Pierce (shaunp) 25-May-96

Environment:

    Kernel mode only

Revision History:


--*/

#include "strmini.h"
#include "1394.h"
#include "dbg.h"
#include "ksmedia.h"
#include "dcamdef.h"
#include "sonydcam.h"

#include "capprop.h"   // Video and camera property function prototype



NTSTATUS
DriverEntry(
    IN PVOID Context1,
    IN PVOID Context2
    )

/*++

Routine Description:

    This where life begins for a driver.  The stream class takes care
    of alot of stuff for us, but we still need to fill in an initialization
    structure for the stream class and call it.

Arguments:

    Context1 - Pointer to operating system context value (don't need to know)

    Context2 - Pointer to operating system context value (don't need to know)

Return Value:

    The function value is the final status from the initialization operation.

--*/

{

    HW_INITIALIZATION_DATA HwInitData;


    KdPrint(("\'SonyDCam DriverEntry: enter\n"));

    //
    // Fill in the HwInitData structure
    //

    HwInitData.HwInitializationDataSize = sizeof(HwInitData);
    HwInitData.HwInterrupt = NULL;
    HwInitData.HwReceivePacket = SonyDCamReceivePacket;
    HwInitData.HwCancelPacket = SonyDCamCancelPacket;
    HwInitData.HwRequestTimeoutHandler = SonyDCamTimeoutHandler;
    HwInitData.DeviceExtensionSize = sizeof(SONYDCAM_EXTENSION);
    HwInitData.PerRequestExtensionSize = sizeof(IRB);
    HwInitData.PerStreamExtensionSize = 0;
    HwInitData.FilterInstanceExtensionSize = 0;
    HwInitData.BusMasterDMA = FALSE;
    HwInitData.Dma24BitAddresses = FALSE;
    HwInitData.BufferAlignment = sizeof(ULONG) - 1;
    HwInitData.TurnOffSynchronization = TRUE;
    HwInitData.DmaBufferSize = 0;

    return (StreamClassRegisterAdapter(Context1, Context2, &HwInitData));

}


NTSTATUS
SonyDCamHwInitialize(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )

/*++

Routine Description:

    This where we perform the necessary initialization tasks.

Arguments:

    Srb - Pointer to stream request block

Return Value:

    Nothing

--*/

{

    PIRB Irb;
    CCHAR StackSize;
    ULONG i;
    ULONG DirectoryLength;
    NTSTATUS status;
    LARGE_INTEGER deltaTime;
    PSONYDCAM_EXTENSION DeviceExtension;
    PPORT_CONFIGURATION_INFORMATION ConfigInfo;
    

    KdPrint(("\'SonyDCamHwInitialize: enter\n"));
    ConfigInfo = Srb->CommandData.ConfigInfo;
    Irb = (PIRB) Srb->SRBExtension;
    DeviceExtension = (PSONYDCAM_EXTENSION) ConfigInfo->HwDeviceExtension;

    DeviceExtension->SharedDeviceObject = ConfigInfo->ClassDeviceObject;
    DeviceExtension->BusDeviceObject = ConfigInfo->PhysicalDeviceObject;
    DeviceExtension->BaseRegister = 0;
    DeviceExtension->FrameRate = DEFAULT_FRAME_RATE;
    InitializeListHead(&DeviceExtension->IsochDescriptorList);
    KeInitializeSpinLock(&DeviceExtension->IsochDescriptorLock);
    DeviceExtension->bCameraSpewing = FALSE;
    DeviceExtension->bShutdown = FALSE;
    DeviceExtension->bNeedToListen = FALSE;
    DeviceExtension->hResource = NULL;
    DeviceExtension->hBandwidth = NULL;
    DeviceExtension->IsochChannel = (ULONG) -1;
    DeviceExtension->PendingReadCount = 0;
    DeviceExtension->Frames = 0;
    InitializeListHead(&DeviceExtension->IsochWaitingList);
    KeInitializeSpinLock(&DeviceExtension->IsochWaitingLock);


    StackSize = DeviceExtension->BusDeviceObject->StackSize;
    DeviceExtension->Irp = IoAllocateIrp(StackSize, FALSE);

    if (!DeviceExtension->Irp) {

        SONYDCAM_ASSERT(FALSE);
        return (STATUS_INSUFFICIENT_RESOURCES);

    }

    DeviceExtension->Irb = ExAllocatePool(NonPagedPool, sizeof(IRB));

    if (!DeviceExtension->Irb) {

        SONYDCAM_ASSERT(FALSE);
        return (STATUS_INSUFFICIENT_RESOURCES);

    }
        
    //
    // Make a call to determine what the generation # is on the bus,
    // followed by a call to find out about ourself (config rom info)
    //
    //
    // Get the current generation count first
    //

    Irb->FunctionNumber = REQUEST_GET_GENERATION_COUNT;
    Irb->Flags = 0;

    status = SonyDCamSubmitIrpSynch(DeviceExtension, DeviceExtension->Irp, Irb);

    if (status) {

        KdPrint(("\'SonyDCamHwInitialize: Error %x while trying to get generation number\n", status));
        return (STATUS_UNSUCCESSFUL);

    }

    DeviceExtension->CurrentGeneration = Irb->u.GetGenerationCount.GenerationCount;
    KdPrint(("\'SonyDCamHwInitialize: Generation number = %x\n", DeviceExtension->CurrentGeneration));

    //
    // Now that we have the current generation count, find out how much
    // configuration space we need by setting lengths to zero.
    //

    Irb->FunctionNumber = REQUEST_GET_CONFIGURATION_INFO;
    Irb->Flags = 0;
    Irb->u.GetConfigurationInformation.UnitDirectoryBufferSize = 0;
    Irb->u.GetConfigurationInformation.UnitDependentDirectoryBufferSize = 0;
    Irb->u.GetConfigurationInformation.VendorLeafBufferSize = 0;
    Irb->u.GetConfigurationInformation.ModelLeafBufferSize = 0;

    status = SonyDCamSubmitIrpSynch(DeviceExtension, DeviceExtension->Irp, Irb);

    if (status) {

        KdPrint(("\'SonyDCamHwInitialize: Error %x while trying to get configuration info (1)\n", status));
        return (STATUS_UNSUCCESSFUL);

    }

    //
    // Now go thru and allocate what we need to so we can get our info.
    //

    DeviceExtension->ConfigRom = ExAllocatePool(PagedPool, sizeof(CONFIG_ROM));
    if (!DeviceExtension->ConfigRom) {

        KdPrint(("\'SonyDCamHwInitialize: Couldn't allocate memory for the Config Rom\n"));
        return (STATUS_INSUFFICIENT_RESOURCES);

    }

    DeviceExtension->UnitDirectory = ExAllocatePool(PagedPool, Irb->u.GetConfigurationInformation.UnitDirectoryBufferSize);
    if (!DeviceExtension->UnitDirectory) {

        KdPrint(("\'SonyDCamHwInitialize: Couldn't allocate memory for the UnitDirectory\n"));
        return (STATUS_INSUFFICIENT_RESOURCES);

    }

    if (Irb->u.GetConfigurationInformation.UnitDependentDirectoryBufferSize) {

        DeviceExtension->UnitDependentDirectory = ExAllocatePool(PagedPool, Irb->u.GetConfigurationInformation.UnitDependentDirectoryBufferSize);
        if (!DeviceExtension->UnitDependentDirectory) {

            KdPrint(("\'SonyDCamHwInitialize: Couldn't allocate memory for the UnitDependentDirectory\n"));
            return (STATUS_INSUFFICIENT_RESOURCES);

        }

    }

    if (Irb->u.GetConfigurationInformation.VendorLeafBufferSize) {

        DeviceExtension->VendorLeaf = ExAllocatePool(PagedPool, Irb->u.GetConfigurationInformation.VendorLeafBufferSize);
        if (!DeviceExtension->VendorLeaf) {

            KdPrint(("\'SonyDCamHwInitialize: Couldn't allocate memory for the VendorLeaf\n"));
            return (STATUS_INSUFFICIENT_RESOURCES);

        }

    }

    if (Irb->u.GetConfigurationInformation.ModelLeafBufferSize) {

        DeviceExtension->ModelLeaf = ExAllocatePool(PagedPool, Irb->u.GetConfigurationInformation.ModelLeafBufferSize);
        if (!DeviceExtension->ModelLeaf) {

            KdPrint(("\'SonyDCamHwInitialize: Couldn't allocate memory for the ModelLeaf\n"));
            return (STATUS_INSUFFICIENT_RESOURCES);

        }

    }

    //
    // Now resubmit the Irb with the appropriate pointers inside
    //

    Irb->FunctionNumber = REQUEST_GET_CONFIGURATION_INFO;
    Irb->Flags = 0;
    Irb->u.GetConfigurationInformation.ConfigRom = DeviceExtension->ConfigRom;
    Irb->u.GetConfigurationInformation.UnitDirectory = DeviceExtension->UnitDirectory;
    Irb->u.GetConfigurationInformation.UnitDependentDirectory = DeviceExtension->UnitDependentDirectory;
    Irb->u.GetConfigurationInformation.VendorLeaf = DeviceExtension->VendorLeaf;
    Irb->u.GetConfigurationInformation.ModelLeaf = DeviceExtension->ModelLeaf;

    status = SonyDCamSubmitIrpSynch(DeviceExtension, DeviceExtension->Irp, Irb);

    if (status) {

        KdPrint(("\'SonyDCamHwInitialize: Error %x while trying to get configuration info (2)\n", status));
        return (STATUS_UNSUCCESSFUL);

    }

    //
    // Now we chew thru the Unit Dependent Directory looking for our command
    // base register key.
    //

    DirectoryLength = Irb->u.GetConfigurationInformation.UnitDependentDirectoryBufferSize >> 2;
    for (i=1; i < DirectoryLength; i++) {

        if ((*(((PULONG) DeviceExtension->UnitDependentDirectory)+i) & CONFIG_ROM_KEY_MASK) == COMMAND_BASE_KEY_SIGNITURE) {

            //
            // Found the command base offset.  This is a quadlet offset from
            // the initial register space.  (Should come out to 0xf0f00000)
            //

            DeviceExtension->BaseRegister = bswap(*(((PULONG) DeviceExtension->UnitDependentDirectory)+i) & CONFIG_ROM_OFFSET_MASK);
            DeviceExtension->BaseRegister <<= 2;
            DeviceExtension->BaseRegister |= INITIAL_REGISTER_SPACE_LO;
            break;

        }
        
    }

    SONYDCAM_ASSERT ( DeviceExtension->BaseRegister );

    Irb->FunctionNumber = REQUEST_ASYNC_READ;
    Irb->Flags = 0;
    Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_High = INITIAL_REGISTER_SPACE_HI;
    Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_Low = DeviceExtension->BaseRegister + FIELDOFFSET(CAMERA_REGISTER_MAP, VModeInq[0]);
    Irb->u.AsyncRead.nNumberOfBytesToRead = sizeof(ULONG);
    Irb->u.AsyncRead.nBlockSize = 0;
    Irb->u.AsyncRead.fulFlags = 0;
    Irb->u.AsyncRead.ulGeneration = DeviceExtension->CurrentGeneration;
    Irb->u.AsyncRead.Mdl = MmCreateMdl(NULL, &DeviceExtension->RegisterWorkArea, sizeof(ULONG));
    MmBuildMdlForNonPagedPool(Irb->u.AsyncRead.Mdl);

    status = SonyDCamSubmitIrpSynch(DeviceExtension, DeviceExtension->Irp, Irb);

    KdPrint(("\'SonyDCamHwInitialize: VModes before bswap = %x\n", DeviceExtension->RegisterWorkArea.AsULONG));
    DeviceExtension->RegisterWorkArea.AsULONG = bswap(DeviceExtension->RegisterWorkArea.AsULONG);
    KdPrint(("\'SonyDCamHwInitialize: VModes = %x\n", DeviceExtension->RegisterWorkArea.AsULONG));
    ExFreePool(Irb->u.AsyncRead.Mdl);

    //
    // For each mode supported, we setup a format block in our format array
    // for our stream.  We just copy all the supported stream formats into
    // one place in our DeviceExtension so when GetStreamInfo is called
    // we can return the array correctly.
    //

    // DEBUG

    DeviceExtension->RegisterWorkArea.AsULONG = 0;
    DeviceExtension->RegisterWorkArea.VMode.Mode1 = TRUE;

    // DEBUG 

    DeviceExtension->ModesSupported = bitcount(DeviceExtension->RegisterWorkArea.AsULONG);
    DeviceExtension->FormatBlocks = ExAllocatePool(NonPagedPool, sizeof(KSDATAFORMAT) * DeviceExtension->ModesSupported);

    if (!DeviceExtension->FormatBlocks) {

        KdPrint(("\'SonyDCamHwInitialize: Couldn't allocate memory for the FormatBlocks\n"));
        return (STATUS_INSUFFICIENT_RESOURCES);

    }

    //
    // Now let's actually do a write request to initialize the device
    //

    DeviceExtension->RegisterWorkArea.AsULONG = 0;
    DeviceExtension->RegisterWorkArea.Initialize.Initialize = TRUE;
    DeviceExtension->RegisterWorkArea.AsULONG = bswap(DeviceExtension->RegisterWorkArea.AsULONG);

    //
    // Delay for camera before next request
    //

    deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
    deltaTime.HighPart = -1;
    KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);

    Irb->FunctionNumber = REQUEST_ASYNC_WRITE;
    Irb->Flags = 0;
    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_High = INITIAL_REGISTER_SPACE_HI;
    Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low = DeviceExtension->BaseRegister + FIELDOFFSET(CAMERA_REGISTER_MAP, Initialize);
    Irb->u.AsyncWrite.nNumberOfBytesToWrite = sizeof(ULONG);
    Irb->u.AsyncWrite.nBlockSize = 0;
    Irb->u.AsyncWrite.fulFlags = 0;
    Irb->u.AsyncWrite.ulGeneration = DeviceExtension->CurrentGeneration;
    Irb->u.AsyncWrite.Mdl = MmCreateMdl(NULL, &DeviceExtension->RegisterWorkArea, sizeof(ULONG));
    MmBuildMdlForNonPagedPool(Irb->u.AsyncWrite.Mdl);

    status = SonyDCamSubmitIrpSynch(DeviceExtension, DeviceExtension->Irp, Irb);

    if (status) {

        KdPrint(("\'SonyDCamHwInitialize: Error %x while trying to write to Initialize register\n", status));
        return (STATUS_UNSUCCESSFUL);

    }


// Most of these auto mode are implemented as a video ProcAmp or camera control properties.
// Therefore, this section is not used.  Delete it after everything is stabalized.
#if 0
    //
    // Setup Automatic mode on several control registers
    //

    for (i=0; i < REGISTERS_TO_SET_TO_AUTO; i++) {

        //
        // Delay for camera before next request
        //

        deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
        deltaTime.HighPart = -1;
        KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);

        Irb->FunctionNumber = REQUEST_ASYNC_READ;
        Irb->Flags = 0;
        Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_High = INITIAL_REGISTER_SPACE_HI;
        Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_Low = DeviceExtension->BaseRegister + FIELDOFFSET(CAMERA_REGISTER_MAP, Brightness);
        Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_Low += i * sizeof(ULONG);
        Irb->u.AsyncRead.nNumberOfBytesToRead = sizeof(ULONG);
        Irb->u.AsyncRead.nBlockSize = 0;
        Irb->u.AsyncRead.fulFlags = 0;
        Irb->u.AsyncRead.ulGeneration = DeviceExtension->CurrentGeneration;

        status = SonyDCamSubmitIrpSynch(DeviceExtension, DeviceExtension->Irp, Irb);

        KdPrint(("\'SonyDCamHwInitialize: Brightness before bswap = %x\n", DeviceExtension->RegisterWorkArea.AsULONG));
        DeviceExtension->RegisterWorkArea.AsULONG = bswap(DeviceExtension->RegisterWorkArea.AsULONG);
        KdPrint(("\'SonyDCamHwInitialize: Brightness = %x\n", DeviceExtension->RegisterWorkArea.AsULONG));

        if ((DeviceExtension->RegisterWorkArea.Brightness.PresenceInq) &&
            (Irb->u.AsyncRead.DestinationAddress.IA_Destination_Offset.Off_Low != 0xf0f0080c)) {

            DeviceExtension->RegisterWorkArea.Brightness.AutoMode = TRUE;
            DeviceExtension->RegisterWorkArea.AsULONG = bswap(DeviceExtension->RegisterWorkArea.AsULONG);

            //
            // Delay for camera before next request
            //

            deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
            deltaTime.HighPart = -1;
            KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);

            Irb->FunctionNumber = REQUEST_ASYNC_WRITE;
            Irb->Flags = 0;
            Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_High = INITIAL_REGISTER_SPACE_HI;
            Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low = DeviceExtension->BaseRegister + FIELDOFFSET(CAMERA_REGISTER_MAP, Brightness);
            Irb->u.AsyncWrite.DestinationAddress.IA_Destination_Offset.Off_Low += i * sizeof(ULONG);
            Irb->u.AsyncWrite.nNumberOfBytesToWrite = sizeof(ULONG);
            Irb->u.AsyncWrite.nBlockSize = 0;
            Irb->u.AsyncWrite.fulFlags = 0;
            Irb->u.AsyncWrite.ulGeneration = DeviceExtension->CurrentGeneration;

            status = SonyDCamSubmitIrpSynch(DeviceExtension, DeviceExtension->Irp, Irb);

        }

    }
#endif

    ExFreePool(Irb->u.AsyncWrite.Mdl);
    ExFreePool(DeviceExtension->ConfigRom);
    ExFreePool(DeviceExtension->UnitDirectory);
    if (DeviceExtension->UnitDependentDirectory)
        ExFreePool(DeviceExtension->UnitDependentDirectory);
    if (DeviceExtension->VendorLeaf)
        ExFreePool(DeviceExtension->VendorLeaf);
    if (DeviceExtension->ModelLeaf)
        ExFreePool(DeviceExtension->ModelLeaf);

    //
    // Now we initialize the size of stream descriptor information.
    // We have one stream descriptor, and we attempt to dword align the
    // structure.
    //

    ConfigInfo->StreamDescriptorSize = 
        1 * (sizeof (HW_STREAM_INFORMATION)) +      // 1 stream descriptor
        sizeof(HW_STREAM_HEADER);                   // and 1 stream header

    return (STATUS_SUCCESS);

}


NTSTATUS
SonyDCamSubmitIrpSynch(
    PSONYDCAM_EXTENSION DeviceExtension,
    PIRP Irp,
    PIRB Irb
    )

/*++

Routine Description:

    This routine submits an Irp synchronously to the bus driver.  We'll
    wait here til the Irp comes back

Arguments:

    DeviceExtension - Pointer to my local device extension

    Irp - Pointer to Irp we're sending down to the port driver synchronously

    Irb - Pointer to Irb we're submitting to the port driver

Return Value:

    Status is returned from Irp

--*/

{


    ULONG Retries=RETRY_COUNT;
    KEVENT Event;
    NTSTATUS status;
    LARGE_INTEGER deltaTime;
    PIO_STACK_LOCATION NextIrpStack;


    KdPrint(("\'SonyDCamSubmitIrpSynch: enter\n"));

    do {

        NextIrpStack = IoGetNextIrpStackLocation(Irp);
        NextIrpStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
        NextIrpStack->Parameters.DeviceIoControl.IoControlCode = IOCTL_1394_CLASS;
        NextIrpStack->Parameters.Others.Argument1 = Irb;

        KeInitializeEvent(&Event, SynchronizationEvent, FALSE);

        IoSetCompletionRoutine(
            Irp,
            SonyDCamSynchCompletionRoutine,
           &Event,
            TRUE,
            TRUE,
            TRUE
            );

        status = IoCallDriver(
                    DeviceExtension->BusDeviceObject,
                    Irp
                    );

        if (status == STATUS_PENDING) {

            //
            // Still pending, wait for the IRP to complete
            //

            KdPrint(("\'SonyDCamSubmitIrpSynch: Irp is pending, waiting...\n"));
            KeWaitForSingleObject(
               &Event,
                Executive,
                KernelMode,
                FALSE,
                NULL
                );

        }

        if (Irp->IoStatus.Status == STATUS_TIMEOUT) {

            //
            // Camera isn't fast enough to respond to us sometimes, so
            // we'll delay a little bit in order for it to recover
            //

            deltaTime.LowPart = SONYDCAM_DELAY_VALUE;
            deltaTime.HighPart = -1;
            KeDelayExecutionThread(KernelMode, TRUE, &deltaTime);

        }

    } while ((Irp->IoStatus.Status == STATUS_TIMEOUT) && (Retries--));

    KdPrint(("\'SonyDCamSubmitIrpSynch: exit\n"));
    return (Irp->IoStatus.Status);

}


NTSTATUS
SonyDCamSynchCompletionRoutine(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PKEVENT Event
    )

/*++

Routine Description:

    This routine is for use with synchronous IRP processing.  
    All it does is signal an event, so the driver knows it
    can continue.

Arguments:

    DriverObject - Pointer to driver object created by system.

    Irp - Irp that just completed

    Event - Event we'll signal to say Irp is done

Return Value:

    None.

--*/

{

    KeSetEvent((PKEVENT) Event, 0, FALSE);
    return (STATUS_MORE_PROCESSING_REQUIRED);

}


ULONG
bitcount(
    IN ULONG x
    )

{

    ULONG b;

    for (b=0; x != 0; x >>= 1) {

        if (x & 01) {

            b++;

        }

    }

    return b;

}

