//==========================================================================;
//
//  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
//  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
//  PURPOSE.
//
//  Copyright (c) 1992 - 1996  Microsoft Corporation.  All Rights Reserved.
//
//==========================================================================;

#include "strmini.h"
#include "ksmedia.h"
#include "capmain.h"
#include "capdebug.h"
#include "capxfer.h"
#include "ntstatus.h"

//==========================================================================;
// Routines for processing video streams
//==========================================================================;

/*
** VideoReceiveDataPacket()
**
**   Receives Video data packet commands
**
** Arguments:
**
**   pSrb - Stream request block for the Video stream
**
** Returns: nothing
**
** Side Effects:  none
*/

VOID 
STREAMAPI 
VideoReceiveDataPacket(
    IN PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
    PHW_DEVICE_EXTENSION    pHwDevExt = ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);
    PSTREAMEX               pStrmEx = (PSTREAMEX)pSrb->StreamObject->HwStreamExtension;

    //
    // make sure we have a device extension
    //

    DEBUG_ASSERT((ULONG)pHwDevExt);

    KdPrint(("TestCap: Receiving Stream Data    SRB %8x, %x\n", pSrb, pSrb->Command));

    // 
    // Default to success
    //

    pSrb->Status = STATUS_SUCCESS;

    //
    // determine the type of packet.
    //

    switch (pSrb->Command){

    case SRB_READ_DATA:

        DEBUG_ASSERT (pStrmEx->pCurrentSrb == NULL);

        // Rule: 
        // Only accept read requests when in either the Pause or Run
        // States.  If Stopped, immediately return the SRB.

        if (pStrmEx->KSState == KSSTATE_STOP) {

            CompleteStreamSRB (         
                    pSrb, 
                    StreamRequestComplete,
                    TRUE,
                    ReadyForNextStreamDataRequest);

            break;
        } 
        
        // Keep a single SRB locally which is completed on a timer
        // or VSYNC.  Other drivers may wish to maintain a queue of
        // pending reads.

        pStrmEx->pCurrentSrb = pSrb;

#ifdef DRIVER_DEBUGGING_TEST
        //
        // When initially bringing up a driver, it is useful to just
        // try immediately completing the SRB, thus verifying
        // the streaming process independent of really accessing
        // your hardware.  

        pStrmEx->pCurrentSrb = NULL;

        CompleteStreamSRB (         
            pSrb, 
            StreamRequestComplete,
            TRUE,
            ReadyForNextStreamDataRequest);

#endif // DRIVER_DEBUGGING_TEST

        break;

    case SRB_WRITE_DATA:

        // Rule: 
        // Only accept read requests when in either the Pause or Run
        // States.  If Stopped, immediately return the SRB.

        if (pStrmEx->KSState != KSSTATE_STOP) {
            PKSSTREAM_HEADER pDataPacket = pSrb->CommandData.DataBufferArray;
        
            // 
            // This data packet contains the channel change information
            // passed on the AnalogVideoIn
            //
        
            if (pDataPacket->FrameExtent == sizeof (KS_TVTUNER_CHANGE_INFO)) {

                RtlCopyMemory(
                        &pHwDevExt->TVTunerChangeInfo,
		        pDataPacket->Data,
		        sizeof (KS_TVTUNER_CHANGE_INFO));
            }

        } // If we're not stopped

        CompleteStreamSRB (         
            pSrb, 
            StreamRequestComplete,
            TRUE,
            ReadyForNextStreamDataRequest);


        break;

    default:

        //
        // invalid / unsupported command. Fail it as such
        //

        TRAP

        pSrb->Status = STATUS_NOT_IMPLEMENTED;

        CompleteStreamSRB (         
            pSrb, 
            StreamRequestComplete,
            TRUE,
            ReadyForNextStreamDataRequest);

    }  // switch (pSrb->Command)
}


/*
** VideoReceiveCtrlPacket()
**
**   Receives packet commands that control the Video stream
**
** Arguments:
**
**   pSrb - The stream request block for the Video stream
**
** Returns: nothing
**
** Side Effects:  none
*/

VOID 
STREAMAPI 
VideoReceiveCtrlPacket(
    IN PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
    PHW_DEVICE_EXTENSION pHwDevExt = ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);
    PSTREAMEX pStrmEx = (PSTREAMEX)pSrb->StreamObject->HwStreamExtension;

    DEBUG_ASSERT((ULONG)pHwDevExt);

    KdPrint(("TestCap: Receiving Stream Control SRB %8x, %x\n", pSrb, pSrb->Command));

    // 
    // Default to success
    //

    pSrb->Status = STATUS_SUCCESS;

    //
    // determine the type of packet.
    //

    switch (pSrb->Command)
    {

    case SRB_PROPOSE_DATA_FORMAT:

        if (!(AdapterVerifyFormat (
                pSrb->CommandData.OpenFormat, 
                pSrb->StreamObject->StreamNumber))) {
            pSrb->Status = STATUS_NO_MATCH;
        }
        break;

    case SRB_SET_STREAM_STATE:

        VideoSetState(pSrb);
        break;

    case SRB_GET_STREAM_STATE:

        VideoGetState(pSrb);
        break;

    case SRB_GET_STREAM_PROPERTY:

        VideoGetProperty(pSrb);
        break;

    case SRB_OPEN_MASTER_CLOCK:
    case SRB_CLOSE_MASTER_CLOCK:

        //
        // This stream is being selected to provide a Master clock
        //

        VideoOpenCloseMasterClock (pSrb);

        break;

    case SRB_INDICATE_MASTER_CLOCK:

        //
        // Assigns a clock to a stream
        //

        VideoIndicateMasterClock (pSrb);

        break;

    default:

        //
        // invalid / unsupported command. Fail it as such
        //

        TRAP

        pSrb->Status = STATUS_NOT_IMPLEMENTED;
    }

    CompleteStreamSRB (         
            pSrb, 
            StreamRequestComplete,
            TRUE,
            ReadyForNextStreamControlRequest);    
}


/*
** CompleteStreamSRB ()
**
**   This routine is called when a packet is being completed.
**   The optional second notification type is used to indicate ReadyForNext
**
** Arguments:
**
**   pSrb - pointer to the request packet that timed out
**
**   NotificationType1 - what kind of notification to return
**
**   NotificationType2 - what kind of notification to return (may be 0)
**
**
** Returns:
**
** Side Effects:  none
*/

VOID 
CompleteStreamSRB (
     IN PHW_STREAM_REQUEST_BLOCK pSrb, 
     STREAM_MINIDRIVER_STREAM_NOTIFICATION_TYPE NotificationType1,
     BOOL fUseNotification2,
     STREAM_MINIDRIVER_STREAM_NOTIFICATION_TYPE NotificationType2
    )
{
    KdPrint(("TestCap: Completing Stream        SRB %8x\n", pSrb));

    StreamClassStreamNotification(
            NotificationType1,
            pSrb->StreamObject,
            pSrb);

    if (fUseNotification2) {            // ReadyForNext

            StreamClassStreamNotification(
                  NotificationType2,
                  pSrb->StreamObject);
      }
}


/*
** VideoGetProperty()
**
**    Routine to process video property requests
**
** Arguments:
**
**    pSrb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

VOID 
VideoGetProperty(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
    PSTREAM_PROPERTY_DESCRIPTOR pSPD = pSrb->CommandData.PropertyInfo;

    if (IsEqualGUID (&KSPROPSETID_Connection, &pSPD->Property->Set)) {
        VideoStreamGetConnectionProperty (pSrb);
    }
    else {
       pSrb->Status = STATUS_NOT_IMPLEMENTED;
    }
}



/*
** VideoTimerRoutine()
**
**    Routine to capture video frames based on a timer.
**
**    Note:  Devices capable of using interrupts should always
**           trigger capture on a VSYNC interrupt, and not use a timer.
**
** Arguments:
**
**    Context - pointer to the stream extension
**
** Returns: nothing
**
** Side Effects:  none
*/

VOID 
STREAMAPI 
VideoTimerRoutine(
    PVOID Context
    )
{
    PSTREAMEX                   pStrmEx = ((PSTREAMEX)Context);
    PHW_STREAM_REQUEST_BLOCK    pSrb = pStrmEx->pCurrentSrb;
    PHW_DEVICE_EXTENSION        pHwDevExt = pStrmEx->pHwDevExt;
    PKSSTREAM_HEADER            pDataPacket;
    PKS_FRAME_INFO              pFrameInfo;
    

    // Remember the current time, so we can generate a timestamp

    pHwDevExt->QST_Now = VideoGetSystemTime();

    // If we're stopped and the timer is still running, just return.
    // This will stop the timer.

    if (pStrmEx->KSState == KSSTATE_STOP) {  
        return;
    }

    // Only capture in the RUN state

    if (pStrmEx->KSState == KSSTATE_RUN && pSrb) {  
    
        pDataPacket = pSrb->CommandData.DataBufferArray;
        pFrameInfo = (PKS_FRAME_INFO) (pDataPacket + 1);
        
        //
        // Determine if it is time to capture a frame based on 
        // how much time has elapsed since capture started
        //

        pHwDevExt->WallTime100ns = pHwDevExt->QST_Now - pHwDevExt->QST_Start;

        pStrmEx->FrameTime100ns = (pStrmEx->pVideoInfoHeader->AvgTimePerFrame * pStrmEx->FrameInfo.PictureNumber);
        
        if (pHwDevExt->WallTime100ns > pStrmEx->FrameTime100ns) {

            // Increment the picture count (usually this is VSYNC count)

            pStrmEx->FrameInfo.PictureNumber++;

            if (pSrb) {        // If we have an outstanding read request

                //
                // Call the routine which synthesizes images
                //
                
                ImageSynth (pSrb, pHwDevExt->InputConnected);

                // Clear pCurrentSrb to indicate request has been completed

                InterlockedExchange ((PLONG) &pStrmEx->pCurrentSrb, 0);

                // Set additional info fields about the data captured such as:
                //   Frames Captured
                //   Frames Dropped
                //   Field Polarity
                
                pStrmEx->FrameInfo.ExtendedHeaderSize = pFrameInfo->ExtendedHeaderSize;

                *pFrameInfo = pStrmEx->FrameInfo;

                // Set the discontinuity flag if frames have been previously
                // dropped, and then reset our internal flag

                if (pStrmEx->fDiscontinuity) {
                    pDataPacket->OptionsFlags |= KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY;
                    pStrmEx->fDiscontinuity = FALSE;
                }

                //
                // Return the timestamp for the frame
                //
                pDataPacket->PresentationTime.Numerator = 1;
                pDataPacket->PresentationTime.Denominator = 1;
                pDataPacket->Duration = 
                pStrmEx->pVideoInfoHeader->AvgTimePerFrame;

                if (pStrmEx->hMasterClock) {
                    pDataPacket->PresentationTime.Time = pStrmEx->FrameTime100ns;
                }
                else {
                    pDataPacket->PresentationTime.Time = pHwDevExt->WallTime100ns;
                    
#ifdef TODO
// BUGBUG If we're not the master clock, we need to query the master clock
// to get a timestamp
                    StreamClassQueryMasterClock(
                            pSrb->StreamObject,
                            pStrmEx->hClock,            // MasterClockHandle,
                            TIME_GET_STREAM_TIME,       // TimeFunction,
                            IN PHW_QUERY_CLOCK_ROUTINE ClockCallbackRoutine);
#endif
                }

                pDataPacket->OptionsFlags |= 
                        KSSTREAM_HEADER_OPTIONSF_TIMEVALID |
                        KSSTREAM_HEADER_OPTIONSF_DURATIONVALID;

                // Every frame we generate is a key frame (aka SplicePoint)
                pDataPacket->OptionsFlags |= KSSTREAM_HEADER_OPTIONSF_SPLICEPOINT;

                CompleteStreamSRB (
                        pSrb, 
                        StreamRequestComplete,
                        TRUE,
                        ReadyForNextStreamDataRequest);
                
            } // if we have an SRB

            else {

                //
                // No buffer was available when we should have captured one

                // Increment the counter which keeps track of
                // dropped frames

                pStrmEx->FrameInfo.DropCount++;

                // Set the (local) discontinuity flag
                // This will cause the next packet processed to have the
                //   KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY flag set.

                pStrmEx->fDiscontinuity = TRUE;

            }
            
        } // endif time to capture a frame
    } // endif we're running

    // Schedule the next timer event
    StreamClassScheduleTimer (
            pStrmEx->pStreamObject,     // StreamObject
            pHwDevExt,                  // HwDeviceExtension
            33000,                      // NumberOfMicroseconds
            VideoTimerRoutine,          // TimerRoutine
            pStrmEx);                   // Context

}


/*
** VideoSetState()
**
**    Sets the current state of the requested stream
**
** Arguments:
**
**    pSrb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

VOID 
VideoSetState(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
    PHW_DEVICE_EXTENSION    pHwDevExt = ((PHW_DEVICE_EXTENSION)pSrb->HwDeviceExtension);
    PSTREAMEX               pStrmEx = (PSTREAMEX)pSrb->StreamObject->HwStreamExtension;

    //
    // For each stream, the following states are used:
    // 
    // Stop:    Absolute minimum resources are used.  No outstanding IRPs.
    // Acquire: KS only state that has no DirectShow correpondence
    //          Acquire needed resources.
    // Pause:   Getting ready to run.  Allocate needed resources so that 
    //          the eventual transition to Run is as fast as possible.
    //          Read SRBs will be queued at either the Stream class 
    //          or in your driver (depending on when you send "ReadyForNext")
    // Run:     Streaming. 
    //
    // Moving to Stop to Run always transitions through Pause.
    //
    // But since a client app could crash unexpectedly, drivers should handle
    // the situation of having outstanding IRPs cancelled and open streams
    // being closed WHILE THEY ARE STREAMING!
    //
    // Note that it is quite possible to transition repeatedly between states:
    // Stop -> Pause -> Stop -> Pause -> Run -> Pause -> Run -> Pause -> Stop
    //

    switch (pSrb->CommandData.StreamState)  

    {
    case KSSTATE_STOP:

        //
        // If transitioning to the stopped state, then complete any outstanding IRPs
        //

        if (pStrmEx->pCurrentSrb) {

            KdPrint(("TestCap: Cancelling %X\n", 
                    pStrmEx->pCurrentSrb));

            // Indicate the buffer is cancelled
            pStrmEx->pCurrentSrb->Status = STATUS_CANCELLED;

            // Set the size of data read to zero
            pStrmEx->pCurrentSrb->CommandData.DataBufferArray->DataUsed = 0;

            // Complete the IRP
            CompleteStreamSRB (
                    pStrmEx->pCurrentSrb,
                    StreamRequestComplete,
                    TRUE,
                    ReadyForNextStreamDataRequest);

            // Clear pCurrentSrb to indicate request has been completed
            InterlockedExchange ((PLONG) &pStrmEx->pCurrentSrb, 0);                
        }
        break;

    case KSSTATE_ACQUIRE:

        //
        // This is a KS only state, that has no correspondence in DirectShow
        // 
        break;

    case KSSTATE_PAUSE:

        //
        // On a transition to pause from acquire, start our timer running.
        //

        if (pStrmEx->KSState == KSSTATE_ACQUIRE || pStrmEx->KSState == KSSTATE_STOP) {  
            
            // Remember the time at which the clock was started

            pHwDevExt->QST_Start = VideoGetSystemTime();

            // And initialize the last frame timestamp

            pHwDevExt->QST_Now = pHwDevExt->QST_Start;

            StreamClassScheduleTimer (
                    pSrb->StreamObject,         // StreamObject
                    pHwDevExt,                  // HwDeviceExtension
                    33000,                      // NumberOfMicroseconds
                    VideoTimerRoutine,          // TimerRoutine
                    pStrmEx);                   // Context
        }
        break;

    case KSSTATE_RUN:

        // 
        // Begin Streaming.
        //

        // Remember the time at which the clock was started

        pHwDevExt->QST_Start = VideoGetSystemTime();

        // Zero the frame counters

        RtlZeroMemory (&pStrmEx->FrameInfo, sizeof (KS_FRAME_INFO));

        // Reset the discontinuity flag

        pStrmEx->fDiscontinuity = FALSE;

        break;

    } // end switch (pSrb->CommandData.StreamState)  
    
    //
    // Remember the state of this stream
    //

    pStrmEx->KSState = pSrb->CommandData.StreamState;
}

/*
** VideoGetState()
**
**    Gets the current state of the requested stream
**
** Arguments:
**
**    pSrb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

VOID 
VideoGetState(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
    PSTREAMEX               pStrmEx = (PSTREAMEX)pSrb->StreamObject->HwStreamExtension;

    pSrb->CommandData.StreamState = pStrmEx->KSState;
    pSrb->ActualBytesTransferred = sizeof (KSSTATE);

    // A very odd rule:
    // When transitioning from stop to pause, DShow tries to preroll
    // the graph.  Capture sources can't preroll, and indicate this
    // by returning VFW_S_CANT_CUE in user mode.  To indicate this
    // condition from drivers, they must return ERROR_NO_DATA_DETECTED

    if (pStrmEx->KSState == KSSTATE_PAUSE) {
       pSrb->Status = STATUS_NO_DATA_DETECTED;
    }
}


/*
** VideoStreamGetConnectionProperty()
**
**    Gets the current state of the requested stream
**
** Arguments:
**
**    pSrb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

VOID
VideoStreamGetConnectionProperty(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
    PSTREAMEX pStrmEx = (PSTREAMEX)pSrb->StreamObject->HwStreamExtension;
    PSTREAM_PROPERTY_DESCRIPTOR pSPD = pSrb->CommandData.PropertyInfo;
    ULONG Id = pSPD->Property->Id;              // index of the property

    switch (Id) {

    case KSPROPERTY_CONNECTION_ALLOCATORFRAMING:
        if (pStrmEx->pVideoInfoHeader) {
            PKSALLOCATOR_FRAMING Framing = 
                (PKSALLOCATOR_FRAMING) pSPD->PropertyInfo;
            Framing->RequirementsFlags =
                KSALLOCATOR_REQUIREMENTF_SYSTEM_MEMORY |
                KSALLOCATOR_REQUIREMENTF_INPLACE_MODIFIER |
                KSALLOCATOR_REQUIREMENTF_PREFERENCES_ONLY;
            Framing->PoolType = PagedPool;
            Framing->Frames = 1;
            Framing->FrameSize = pStrmEx->pVideoInfoHeader->bmiHeader.biSizeImage;
            Framing->FileAlignment = 0; // FILE_LONG_ALIGNMENT???;
            Framing->Reserved = 0;
            pSrb->ActualBytesTransferred = sizeof (KSALLOCATOR_FRAMING);
        }
        else {
            pSrb->Status = STATUS_INVALID_PARAMETER;
        }
        break;

    default:
        TRAP
        break;
    }
}

//==========================================================================;
//                   Clock Handling Routines
//==========================================================================;

/*
** VideoOpenCloseMasterClock ()
**
**    Opens and Closes the clock on this stream as the master clock.
**        On open, we are passed a handle which is used to represent our clock.
**        On close, this handle will be set to NULL.
**
** Arguments:
**
**    pSrb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

VOID 
VideoOpenCloseMasterClock(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
    PSTREAMEX pStrmEx = (PSTREAMEX)pSrb->StreamObject->HwStreamExtension;

    pStrmEx->hMasterClock = pSrb->CommandData.MasterClockHandle;
}


/*
** VideoIndicateMasterClock ()
**
**    If this stream is not being used as the master clock, this function
**      is used to provide us with a handle to the clock to use.
**
** Arguments:
**
**    pSrb - pointer to the stream request block for properties
**
** Returns:
**
** Side Effects:  none
*/

VOID 
VideoIndicateMasterClock(
    PHW_STREAM_REQUEST_BLOCK pSrb
    )
{
    PSTREAMEX pStrmEx = (PSTREAMEX)pSrb->StreamObject->HwStreamExtension;

    pStrmEx->hClock = pSrb->CommandData.MasterClockHandle;
}

/*
** GetSystemTime ()
**
**    Returns the system time in 100 nS units
**
** Arguments:
**
** Returns:
**
** Side Effects:  none
*/

ULONGLONG 
VideoGetSystemTime(
    )
{
    ULONGLONG ticks;
    ULONGLONG rate;

    ticks = (ULONGLONG)KeQueryPerformanceCounter((PLARGE_INTEGER)&rate).QuadPart;

    //
    // convert from ticks to 100ns clock
    //

    ticks = (ticks & 0xFFFFFFFF00000000) / rate * 10000000 +
            (ticks & 0x00000000FFFFFFFF) * 10000000 / rate;

    return(ticks);

}

/*
** StreamClockRtn ()
**
**    When this stream is providing the master clock, this function is called
**      whenever another filter wants to know the time.
**
** Arguments:
**
**      TimeContext -
**          Pointer to the time context 
**
** Returns:
**
** Side Effects:  none
*/

VOID
STREAMAPI 
StreamClockRtn(
    IN PHW_TIME_CONTEXT TimeContext
    )
{
    PHW_DEVICE_EXTENSION pHwDevExt = ((PHW_DEVICE_EXTENSION) TimeContext->HwDeviceExtension);
    PSTREAMEX pStrmEx = (PSTREAMEX) (TimeContext->HwStreamObject->HwStreamExtension);

    pHwDevExt->QST_Now = VideoGetSystemTime();

    switch (TimeContext->Function) {
    
    case TIME_GET_STREAM_TIME:

        //
        // How long since the stream was first set into the run state?
        //

        TimeContext->Time = pHwDevExt->QST_Now - pHwDevExt->QST_Start;
        TimeContext->SystemTime = pHwDevExt->QST_Now;

        break;

    case TIME_READ_ONBOARD_CLOCK:

        //
        // Devices which provide a VSYNC interrupt would create a clock based on
        // VSYNC interrupts here.  Testcap just uses the system clock.
        //
        TimeContext->Time = pHwDevExt->QST_Now;
        TimeContext->SystemTime = pHwDevExt->QST_Now;
        
        break;
    
    default:
        TRAP
        break;
    } // switch TimeContext->Function

    KdPrint(("TestCap:    Time=%16lx\n", TimeContext->Time));
    KdPrint(("TestCap: SysTime=%16lx\n", TimeContext->SystemTime));
}


