/*++

Copyright (c) 1996, 1997  Microsoft Corporation

Module Name:

   intelcam.c

Abstract:

   driver for the intel camera.

Author:

    jdunn

Environment:

   Kernel mode only


Revision History:

--*/

#include "warn.h"
#include "wdm.h"


#include "intelcam.h"
#include "strmini.h"

#include "prpobj.h"
#include "prpftn.h"

// maps fps to an interface number
UCHAR InterfaceMap[] = {7, 7, 7, 7, 7, 7,       // 0..5
                        1, 1, 1, 1,             // 6.. 9
                        0, 0,                   // 10..11
                        3, 3, 3,                // 12..14
                        2, 2, 2,                // 15..17
                        4, 4,                   // 18..19
                        5, 5, 5, 5, 5,          // 20..24
                        6};                     // 25


ULONG INTELCAM_DebugTraceLevel
#ifdef MAX_DEBUG
    = MAX_TRACE;
#else
    = MIN_TRACE;
#endif


// **
// format for each type of video stream we support
// **

#ifndef mmioFOURCC    
#define mmioFOURCC( ch0, ch1, ch2, ch3 )                \
        ( (DWORD)(BYTE)(ch0) | ( (DWORD)(BYTE)(ch1) << 8 ) |    \
        ( (DWORD)(BYTE)(ch2) << 16 ) | ( (DWORD)(BYTE)(ch3) << 24 ) )
#endif  

#define D_X 176
#define D_Y 144


KS_DATARANGE_VIDEO INTELCAM_StreamFormat_0 = 
{
    // KSDATARANGE
    {   
        sizeof (KS_DATARANGE_VIDEO),
        0,                      // Flags
        (D_X * D_Y * 12)/8,     // SampleSize
        0,                      // Reserved
//MEDIATYPE_Video
        STATIC_KSDATAFORMAT_TYPE_VIDEO,      
//MEDIASUBTYPE_Y411        
        0x31313459, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71,
//FORMAT_VideoInfo        
        STATIC_KSDATAFORMAT_SPECIFIER_VIDEOINFO
    },

    TRUE,                       // BOOL,  bFixedSizeSamples (all samples same size?)
    TRUE,                       // BOOL,  bTemporalCompression (all I frames?)
    KS_VIDEOSTREAM_CAPTURE,     // StreamDescriptionFlags
    0,                          // MemoryAllocationFlags   (KS_VIDEO_ALLOC_*)

    // _KS_VIDEO_STREAM_CONFIG_CAPS  
    {
        STATIC_KSDATAFORMAT_SPECIFIER_VIDEOINFO,
        KS_AnalogVideo_None,    // VideoStandard
        D_X, D_Y,               // InputSize, (the inherent size of the incoming signal
                                //             with every digitized pixel unique)
        D_X, D_Y,               // MinCroppingSize, smallest rcSrc cropping rect allowed
        D_X, D_Y,               // MaxCroppingSize, largest  rcSrc cropping rect allowed
        1,                      // CropGranularityX, granularity of cropping size
        1,                      // CropGranularityY
        1,                      // CropAlignX, alignment of cropping rect 
        1,                      // CropAlignY;
        D_X, D_Y,               // MinOutputSize, smallest bitmap stream can produce
        D_X, D_Y,               // MaxOutputSize, largest  bitmap stream can produce
        1,                      // OutputGranularityX, granularity of output bitmap size
        1,                      // OutputGranularityY;
        0,                      // StretchTapsX
    	0,                      // StretchTapsY
        0,                      // ShrinkTapsX
    	0,                      // ShrinkTapsY
        333667,                 // MinFrameInterval, 100 nS units
        333667,                 // MaxFrameInterval, 100 nS units
        3 * 30 * 160 * 120,     // MinBitsPerSecond;
        3 * 30 * 320 * 240      // MaxBitsPerSecond;
    }, 
        
    // KS_VIDEOINFOHEADER (default format)
    {
        0,0,0,0,                        // RECT  rcSource; 
        0,0,0,0,                        // RECT  rcTarget; 
        D_X * D_Y * 3 * 30,             // DWORD dwBitRate;
        0L,                             // DWORD dwBitErrorRate; 
        333667,                         // REFERENCE_TIME  AvgTimePerFrame;   
        sizeof (KS_BITMAPINFOHEADER),   // DWORD biSize;
        D_X,                            // LONG  biWidth;
        D_Y,                            // LONG  biHeight;
        1,                              // WORD  biPlanes;
        12,                             // WORD  biBitCount;
        FCC_FORMAT_YUV12N,              // DWORD biCompression;
        (D_X * D_Y * 12)/8,             // DWORD biSizeImage;
        0,                              // LONG  biXPelsPerMeter;
        0,                              // LONG  biYPelsPerMeter;
        0,                              // DWORD biClrUsed;
        0                               // DWORD biClrImportant;

    }
}; 
    
#undef D_X
#undef D_Y
    

#define D_X 320
#define D_Y 240


KS_DATARANGE_VIDEO INTELCAM_StreamFormat_1 = 
{
    // KSDATARANGE
    {   
        sizeof (KS_DATARANGE_VIDEO),
        0,                      // Flags
        (D_X * D_Y * 12)/8,     // SampleSize
        0,                      // Reserved
//MEDIATYPE_Video
        STATIC_KSDATAFORMAT_TYPE_VIDEO,      
//MEDIASUBTYPE_Y411        
        0x31313459, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71,
//FORMAT_VideoInfo        
        STATIC_KSDATAFORMAT_SPECIFIER_VIDEOINFO
    },

    TRUE,                       // BOOL,  bFixedSizeSamples (all samples same size?)
    TRUE,                       // BOOL,  bTemporalCompression (all I frames?)
    KS_VIDEOSTREAM_CAPTURE,     // StreamDescriptionFlags
    0,                          // MemoryAllocationFlags   (KS_VIDEO_ALLOC_*)

    // _KS_VIDEO_STREAM_CONFIG_CAPS  
    {
        STATIC_KSDATAFORMAT_SPECIFIER_VIDEOINFO,
        KS_AnalogVideo_None,    // VideoStandard
        D_X, D_Y,               // InputSize, (the inherent size of the incoming signal
                                //             with every digitized pixel unique)
        D_X, D_Y,               // MinCroppingSize, smallest rcSrc cropping rect allowed
        D_X, D_Y,               // MaxCroppingSize, largest  rcSrc cropping rect allowed
        1,                      // CropGranularityX, granularity of cropping size
        1,                      // CropGranularityY
        1,                      // CropAlignX, alignment of cropping rect 
        1,                      // CropAlignY;
        D_X, D_Y,               // MinOutputSize, smallest bitmap stream can produce
        D_X, D_Y,               // MaxOutputSize, largest  bitmap stream can produce
        1,                      // OutputGranularityX, granularity of output bitmap size
        1,                      // OutputGranularityY;
        0,                      // StretchTapsX
    	0,                      // StretchTapsY
        0,                      // ShrinkTapsX
    	0,                      // ShrinkTapsY
        333667,                 // MinFrameInterval, 100 nS units
        333667,                 // MaxFrameInterval, 100 nS units
        3 * 30 * 160 * 120,     // MinBitsPerSecond;
        3 * 30 * 320 * 240      // MaxBitsPerSecond;
    }, 
        
    // KS_VIDEOINFOHEADER (default format)
    {
        0,0,0,0,                            // RECT  rcSource; 
        0,0,0,0,                            // RECT  rcTarget; 
        D_X * D_Y * 3 * 30,                 // DWORD dwBitRate;
        0L,                                 // DWORD dwBitErrorRate; 
        333667,                             // REFERENCE_TIME  AvgTimePerFrame;   

        sizeof (KS_BITMAPINFOHEADER),       // DWORD biSize;
        D_X,                                // LONG  biWidth;
        D_Y,                                // LONG  biHeight;
        1,                                  // WORD  biPlanes;
        12,                                 // WORD  biBitCount;
        FCC_FORMAT_YUV12N,                  // DWORD biCompression;
        (D_X * D_Y * 12)/8,                 // DWORD biSizeImage;
        0,                                  // LONG  biXPelsPerMeter;
        0,                                  // LONG  biYPelsPerMeter;
        0,                                  // DWORD biClrUsed;
        0                                   // DWORD biClrImportant;

    }
}; 
    
#undef D_X
#undef D_Y

PKSDATAFORMAT INTELCAM_StreamFormats[] = 
{
    (PKSDATAFORMAT) &INTELCAM_StreamFormat_0,
    (PKSDATAFORMAT) &INTELCAM_StreamFormat_1
};

#define NUM_INTELCAM_STREAM_FORMATS (SIZEOF_ARRAY(INTELCAM_StreamFormats))

typedef struct _ALL_STREAM_INFO {
    HW_STREAM_INFORMATION   hwStreamInfo;
} ALL_STREAM_INFO, *PALL_STREAM_INFO;
    
    
ALL_STREAM_INFO Streams [] = 
{
    // -----------------------------------------------------------------
    // INTELCAM_Stream
    // -----------------------------------------------------------------
  {
    // HW_STREAM_INFORMATION -------------------------------------------
    1,                                      // NumberOfPossibleInstances
    KSPIN_DATAFLOW_OUT,                     // DataFlow
    TRUE,                                   // DataAccessible
    NUM_INTELCAM_STREAM_FORMATS,            // NumberOfFormatArrayEntries
    INTELCAM_StreamFormats,                 // StreamFormatsArray
    0,                                      // ClassReserved[0]
    0,                                      // ClassReserved[1]
    0,                                      // ClassReserved[2]
    0,                                      // ClassReserved[3]
    0,                                      // NumStreamPropArrayEntries
    0,                                      // StreamPropertiesArray
    0,                                      // NumStreamEventArrayEntries;
    0,                                      // StreamEventsArray;
    NULL,                                   // Category;
    (GUID *)&PINNAME_VIDEO_CAPTURE,         // Name;
    0,                                      // MediumsCount
    NULL,                                   // Mediums
  }
};


/*
** AdapterCompareGUIDsAndFormatSize()
**
**   Checks for a match on the three GUIDs and FormatSize
**
** Arguments:
**
**         IN DataRange1
**         IN DataRange2
**
** Returns:
** 
**   TRUE if all elements match
**   FALSE if any are different
**
** Side Effects:  none
*/

BOOLEAN 
AdapterCompareGUIDsAndFormatSize(
    IN PKSDATARANGE DataRange1,
    IN PKSDATARANGE DataRange2)
{
    return (
        IsEqualGUID (
            &DataRange1->MajorFormat, 
            &DataRange2->MajorFormat) &&
        IsEqualGUID (
            &DataRange1->SubFormat, 
            &DataRange2->SubFormat) &&
        IsEqualGUID (
            &DataRange1->Specifier, 
            &DataRange2->Specifier) &&
        (DataRange1->FormatSize == DataRange2->FormatSize));
}

/*
** AdapterFormatFromRange()
**
**   Returns a DATAFORMAT from a DATARANGE
**
** Arguments:
**
**         IN PHW_STREAM_REQUEST_BLOCK pSrb 
**
** Returns:
** 
**  STATUS_SUCCESS if format is supported
**
** Side Effects:  none
*/

NTSTATUS 
AdapterFormatFromRange(
        IN PHW_STREAM_REQUEST_BLOCK Srb)
{
    PSTREAM_DATA_INTERSECT_INFO intersectInfo;
    PKSDATARANGE  dataRange;
    BOOL onlyWantsSize;
    ULONG formatSize;
    ULONG streamNumber;
    ULONG j;
    ULONG numberOfFormatArrayEntries;
    PKSDATAFORMAT *availableFormats;
    NTSTATUS ntStatus = STATUS_SUCCESS;
    
    intersectInfo = Srb->CommandData.IntersectInfo;
    streamNumber = intersectInfo->StreamNumber;
    dataRange = intersectInfo->DataRange;

    //
    // Check that the stream number is valid
    //

    INTELCAM_KdPrint (MAX_TRACE, ("'AdapterFormatFromRange: Stream=%d\n", 
            streamNumber));

    ASSERT(streamNumber == 0);
    
    numberOfFormatArrayEntries = 
            Streams[0].hwStreamInfo.NumberOfFormatArrayEntries;

    //
    // Get the pointer to the array of available formats
    //

    availableFormats =
		Streams[0].hwStreamInfo.StreamFormatsArray;

    //
    // Is the caller trying to get the format, or the size of the format?
    //

    onlyWantsSize =
	    (intersectInfo->SizeOfDataFormatBuffer == sizeof(ULONG));

    //
    // Walk the formats supported by the stream searching for a match
    // of the three GUIDs which together define a DATARANGE
    //

    for (j = 0; j < numberOfFormatArrayEntries; j++, availableFormats++) {

        INTELCAM_KdPrint (MAX_TRACE, ("'checking format %d\n", 
            j));


        if (!AdapterCompareGUIDsAndFormatSize(dataRange,
		                                      *availableFormats)) {
            // not the format we want		                                      
            continue;
        }

        //
        // Now that the three GUIDs match, switch on the Specifier
        // to do a further type specific check
        //

        // -------------------------------------------------------------------
        // Specifier FORMAT_VideoInfo for VIDEOINFOHEADER
        // -------------------------------------------------------------------

        if (IsEqualGUID (&dataRange->Specifier, 
                         &KSDATAFORMAT_SPECIFIER_VIDEOINFO)) {

            PKS_DATARANGE_VIDEO dataRangeVideoToVerify = 
                    (PKS_DATARANGE_VIDEO) dataRange;
            PKS_DATARANGE_VIDEO dataRangeVideo = 
                    (PKS_DATARANGE_VIDEO) *availableFormats;

            
            INTELCAM_KdPrint (MAX_TRACE, ("'guid matched\n")); 
            
            //
            // Check that the other fields match
            //
            if ((dataRangeVideoToVerify->bFixedSizeSamples !=
			                 dataRangeVideo->bFixedSizeSamples) ||
                (dataRangeVideoToVerify->bTemporalCompression !=
				             dataRangeVideo->bTemporalCompression) ||
                (dataRangeVideoToVerify->StreamDescriptionFlags !=
				             dataRangeVideo->StreamDescriptionFlags) ||
                (dataRangeVideoToVerify->MemoryAllocationFlags !=
				             dataRangeVideo->MemoryAllocationFlags) ||
                (RtlCompareMemory (&dataRangeVideoToVerify->ConfigCaps,
                        &dataRangeVideo->ConfigCaps,
                        sizeof (KS_VIDEO_STREAM_CONFIG_CAPS)) != 
                        sizeof (KS_VIDEO_STREAM_CONFIG_CAPS))) {
                // not the format want                        
                INTELCAM_KdPrint (MAX_TRACE, ("'format does not match\n")); 
                continue;
            }

            INTELCAM_KdPrint (MAX_TRACE, ("'match found\n"));                 
            formatSize = sizeof (KSDATAFORMAT) + 
                KS_SIZE_VIDEOHEADER (&dataRangeVideoToVerify->
    				                        VideoInfoHeader);

            if (onlyWantsSize) {
                break;
            }
            
            // Caller wants the full data format, make sure we have room
            if (intersectInfo->SizeOfDataFormatBuffer < formatSize) {
                ntStatus = STATUS_BUFFER_TOO_SMALL;
                INTELCAM_KdPrint (MAX_TRACE, ("'formatsize too small\n")); 
                break;
            }

            // Copy over the KSDATAFORMAT, followed by the 
            // actual VideoInfoHeader
                
            RtlCopyMemory(
                &((PKS_DATAFORMAT_VIDEOINFOHEADER)intersectInfo->
				                          DataFormatBuffer)->DataFormat,
                &dataRangeVideoToVerify->DataRange,
                sizeof (KSDATARANGE));

            ((PKSDATAFORMAT)intersectInfo->
			                 DataFormatBuffer)->FormatSize = formatSize;

            RtlCopyMemory(
                &((PKS_DATAFORMAT_VIDEOINFOHEADER) intersectInfo->
				                      DataFormatBuffer)->VideoInfoHeader,
                &dataRangeVideoToVerify->VideoInfoHeader,
                KS_SIZE_VIDEOHEADER (&dataRangeVideoToVerify->
				                                VideoInfoHeader));

            //
            // BUGBUG Validate format here, and calculate biSizeImage
            // 

            break;

        } else {
            ntStatus = STATUS_NO_MATCH;
            INTELCAM_KdPrint (MAX_TRACE, ("'no match\n"));
            break;
        }

    } // End of loop on all formats for this stream

    if (NT_SUCCESS(ntStatus)) {
        if (onlyWantsSize) {
            *(PULONG) intersectInfo->DataFormatBuffer = formatSize;
            Srb->ActualBytesTransferred = sizeof(ULONG);
        } else {       
            Srb->ActualBytesTransferred = formatSize;
        }
    }    

    INTELCAM_KdPrint (MAX_TRACE, ("'AdapterFormatFromRange ret %x\n", ntStatus));
    
    return ntStatus;
}

/*
** AdapterVerifyFormat()
**
**   Checks the validity of a format request by walking through the
**       array of supported KSDATA_RANGEs for a given stream.
**
** Arguments:
**
**   pKSDataFormat - pointer of a KS_DATAFORMAT_VIDEOINFOHEADER structure.
**   StreamNumber - index of the stream being queried / opened.
**
** Returns:
** 
**   TRUE if the format is supported
**   FALSE if the format cannot be suppored
**
** Side Effects:  none
*/

BOOL 
AdapterVerifyFormat(
    PKS_DATAFORMAT_VIDEOINFOHEADER pKSDataFormatToVerify, 
    int StreamNumber
    )
{
    PKS_VIDEOINFOHEADER                 pVideoInfoHdrToVerify = 
                &pKSDataFormatToVerify->VideoInfoHeader;
    PKS_VIDEOINFOHEADER                 pVideoInfoHdr;
    PKSDATAFORMAT                       *pAvailableFormats;
    PKS_DATARANGE_VIDEO                 pKSDataRange;
    KS_VIDEO_STREAM_CONFIG_CAPS         *pConfigCaps;
    int                                 NumberOfFormatArrayEntries;
    int                                 nSize;
    int                                 j;
    RECT                                rcImage;
     
    //
    // Make sure the stream index is valid
    //
    if (StreamNumber >= 2 || StreamNumber < 0) {
        return FALSE;
    }

    //
    // How many formats does this stream support?
    //
    NumberOfFormatArrayEntries = 
                Streams[StreamNumber].hwStreamInfo.NumberOfFormatArrayEntries;

    nSize = sizeof (KS_VIDEOINFOHEADER) + 
                pVideoInfoHdrToVerify->bmiHeader.biSize;

    INTELCAM_KdPrint (MAX_TRACE, ("AdapterVerifyFormat: Stream=%d\n", 
            StreamNumber));

    INTELCAM_KdPrint (MAX_TRACE, ("AdapterVerifyFormat: FormatSize=%d\n", 
            pKSDataFormatToVerify->DataFormat.FormatSize));

    INTELCAM_KdPrint (MAX_TRACE, ("AdapterVerifyFormat: MajorFormat=%x\n", 
            pKSDataFormatToVerify->DataFormat.MajorFormat));

    INTELCAM_KdPrint (MAX_TRACE, ("AdapterVerifyFormat: pVideoInfoHdrToVerify=%x\n", 
                pVideoInfoHdrToVerify));

    INTELCAM_KdPrint (MAX_TRACE, ("AdapterVerifyFormat: KS_VIDEOINFOHEADER size =%d\n", 
                nSize));

    INTELCAM_KdPrint (MAX_TRACE, ("AdapterVerifyFormat: Width=%d Height=%d  biBitCount=%d\n", 
                pVideoInfoHdrToVerify->bmiHeader.biWidth, 
                pVideoInfoHdrToVerify->bmiHeader.biHeight,
                pVideoInfoHdrToVerify->bmiHeader.biBitCount));

    INTELCAM_KdPrint (MAX_TRACE, ("AdapterVerifyFormat: biSizeImage =%d\n", 
                pVideoInfoHdrToVerify->bmiHeader.biSizeImage));

    //
    // Get the pointer to the array of available formats
    //
    pAvailableFormats = Streams[StreamNumber].hwStreamInfo.StreamFormatsArray;

    //
    // Walk the array, searching for a match
    //
    for (j = 0; j < NumberOfFormatArrayEntries; j++, pAvailableFormats++)
	{
        pKSDataRange = (PKS_DATARANGE_VIDEO) *pAvailableFormats;
        pVideoInfoHdr = &pKSDataRange->VideoInfoHeader;
        pConfigCaps = &pKSDataRange->ConfigCaps;
        
        //
        // Check for matching size, Major Type, Sub Type, and Specifier
        //

        if (!IsEqualGUID (&pKSDataRange->DataRange.MajorFormat, 
            &pKSDataFormatToVerify->DataFormat.MajorFormat))
		{
            break;
        }

        if (!IsEqualGUID (&pKSDataRange->DataRange.SubFormat, 
            &pKSDataFormatToVerify->DataFormat.SubFormat))
		{
            break;
        }

        if (!IsEqualGUID (&pKSDataRange->DataRange.Specifier,
            &pKSDataFormatToVerify->DataFormat.Specifier))
		{
            break;
        }

        /* 
        **  HOW BIG IS THE IMAGE REQUESTED (pseudocode follows)
        **
        **  if (IsRectEmpty (&rcTarget) {
        **      SetRect (&rcImage, 0, 0, 
        **              BITMAPINFOHEADER.biWidth,
                        BITMAPINFOHEADER.biHeight);
        **  }
        **  else {
        **      // Probably rendering to a DirectDraw surface,
        **      // where biWidth is used to expressed the "stride" 
        **      // in units of pixels (not bytes) of the destination surface.
        **      // Therefore, use rcTarget to get the actual image size 
        **      
        **      rcImage = rcTarget;
        **  }
        */
        
        if ((pVideoInfoHdrToVerify->rcTarget.right - 
             pVideoInfoHdrToVerify->rcTarget.left <= 0) ||
            (pVideoInfoHdrToVerify->rcTarget.bottom - 
             pVideoInfoHdrToVerify->rcTarget.top <= 0))
		{
            
             rcImage.left = rcImage.top = 0;
             rcImage.right = pVideoInfoHdrToVerify->bmiHeader.biWidth;
             rcImage.bottom = pVideoInfoHdrToVerify->bmiHeader.biHeight;
        }
        else
		{
             rcImage = pVideoInfoHdrToVerify->rcTarget;
        }

        //
        // BUGBUG, perform all other verification tests here!!!
        //




        //
        // HOORAY, the format passed all of the tests, so we support it
        //

        return TRUE;
    } 

    //
    // The format requested didn't match any of our listed ranges,
    // so refuse the connection.
    //

    return FALSE;
}



//
// hooks for stream SRBs
//

VOID STREAMAPI
INTELCAM_ReceiveDataPacket(
    IN PHW_STREAM_REQUEST_BLOCK Srb,
    IN PVOID DeviceContext,
    IN PBOOLEAN Completed
    )
{
     INTELCAM_KdPrint (MAX_TRACE, ("'INTELCAM_ReceiveDataPacket\n"));
} 


VOID STREAMAPI
INTELCAM_ReceiveCtrlPacket(
    IN PHW_STREAM_REQUEST_BLOCK Srb,
    IN PVOID DeviceContext,
    IN PBOOLEAN Completed
    )
{
     INTELCAM_KdPrint (MAX_TRACE, ("'INTELCAM_ReceiveCtrlPacket\n"));
} 

// ** 
// Describe the camera
//

USBCAMD_DEVICE_DATA INTELCAM_DeviceData = {
    0,
    INTELCAM_Initialize,
    INTELCAM_UnInitialize,
    INTELCAM_ProcessUSBPacket,
    INTELCAM_NewFrame,
    INTELCAM_ProcessRawVideoFrame,
    INTELCAM_StartVideoCapture,
    INTELCAM_StopVideoCapture,
    INTELCAM_Configure,
    INTELCAM_SaveState,
    INTELCAM_RestoreState,
    INTELCAM_AllocateBandwidth,
    INTELCAM_FreeBandwidth
    };

VOID 
INTELCAM_AdapterReceivePacket(
    IN PHW_STREAM_REQUEST_BLOCK Srb
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext;
    PHW_STREAM_INFORMATION streamInformation =
        &(Srb->CommandData.StreamBuffer->StreamInfo);
    PHW_STREAM_HEADER streamHeader =
        &(Srb->CommandData.StreamBuffer->StreamHeader);        
    PDEVICE_OBJECT deviceObject;        
    
    INTELCAM_KdPrint (MAX_TRACE, ("'INTELCAM_AdapterReceivePacket\n"));
    
    switch (Srb->Command) {
    case SRB_GET_STREAM_INFO:
        //
        // this is a request for the driver to enumerate requested streams
        //
        INTELCAM_KdPrint (MAX_TRACE, ("'SRB_GET_STREAM_INFO\n"));

        //
        // we support one stream
        //
        
        streamHeader->NumberOfStreams = 1;
        
        streamInformation->StreamFormatsArray = &INTELCAM_StreamFormats[0];
        streamInformation->NumberOfFormatArrayEntries = NUM_INTELCAM_STREAM_FORMATS;

        //
        // set the property information for the video stream
        //

        streamHeader->DevicePropertiesArray =
            INTELCAM_GetAdapterPropertyTable(&streamHeader->NumDevPropArrayEntries); 

        // pass to usbcamd to finish the job
        deviceContext =     
            USBCAMD_AdapterReceivePacket(Srb, &INTELCAM_DeviceData, NULL, TRUE);   
            
        ASSERT_DEVICE_CONTEXT(deviceContext); 
        break;

    case SRB_GET_DEVICE_PROPERTY:
        //
        // we handle all the property stuff 
        //
        INTELCAM_KdPrint (MAX_TRACE, ("'SRB_GET_DEVICE_PROPERTY\n"));
        
        deviceContext =     
            USBCAMD_AdapterReceivePacket(Srb, &INTELCAM_DeviceData, &deviceObject, FALSE);   
        ASSERT_DEVICE_CONTEXT(deviceContext);  
        
        INTELCAM_KdPrint (MAX_TRACE, ("'SRB_GET_STREAM_INFO\n"));
        INTELCAM_PropertyRequest(
            FALSE,
            deviceObject,
            deviceContext,
            Srb);

        StreamClassDeviceNotification(DeviceRequestComplete,
                                      Srb->HwDeviceExtension,
                                      Srb);
        break;            
            
    case SRB_SET_DEVICE_PROPERTY:
        //
        // we handle all the property stuff 
        //
        INTELCAM_KdPrint (MAX_TRACE, ("'SRB_SET_DEVICE_PROPERTY\n"));
        
        deviceContext =     
            USBCAMD_AdapterReceivePacket(Srb, &INTELCAM_DeviceData, &deviceObject, FALSE);   
        ASSERT_DEVICE_CONTEXT(deviceContext);  
        
        INTELCAM_KdPrint (MAX_TRACE, ("'SRB_GET_STREAM_INFO\n"));
        INTELCAM_PropertyRequest(
            TRUE,
            deviceObject,
            deviceContext,
            Srb);

        StreamClassDeviceNotification(DeviceRequestComplete,
                                      Srb->HwDeviceExtension,
                                      Srb);
        break;

    case SRB_OPEN_STREAM:        
        {
        PKS_DATAFORMAT_VIDEOINFOHEADER  pKSDataFormat = 
                (PKS_DATAFORMAT_VIDEOINFOHEADER) Srb->CommandData.OpenFormat;
        PKS_VIDEOINFOHEADER  pVideoInfoHdrRequested = 
                &pKSDataFormat->VideoInfoHeader;

        // pass to usbcamd to finish the job
        Srb->StreamObject->ReceiveDataPacket = (PVOID) INTELCAM_ReceiveDataPacket;
        Srb->StreamObject->ReceiveControlPacket = (PVOID) INTELCAM_ReceiveCtrlPacket;

  
        if (AdapterVerifyFormat(pKSDataFormat, 
                                Srb->StreamObject->StreamNumber)) {
            deviceContext =     
                USBCAMD_AdapterReceivePacket(Srb, &INTELCAM_DeviceData, NULL, TRUE);   
        } else {
            Srb->Status = STATUS_INVALID_PARAMETER;
            StreamClassDeviceNotification(DeviceRequestComplete,
                                      Srb->HwDeviceExtension,
                                      Srb);
        }
        }
        break;                    


    case SRB_GET_DATA_INTERSECTION:
		//
		// Return a format, given a range
		//
        //deviceContext =     
        //    USBCAMD_AdapterReceivePacket(Srb,
		//	                             &INTELCAM_DeviceData,
		//	                             &deviceObject,
		//								 FALSE);   
										 
		Srb->Status = AdapterFormatFromRange(Srb);
		
		StreamClassDeviceNotification(DeviceRequestComplete,
                                      Srb->HwDeviceExtension,
                                      Srb);        
        break;
        
    default:
        //
        // let usbcamd handle it
        //
    
        deviceContext =     
            USBCAMD_AdapterReceivePacket(Srb, &INTELCAM_DeviceData, NULL, TRUE);            
        ASSERT_DEVICE_CONTEXT(deviceContext);  
    }
    
}


/*
** DriverEntry()
**
** This routine is called when the mini driver is first loaded.  The driver
** should then call the StreamClassRegisterAdapter function to register with
** the stream class driver
**
** Arguments:
**
**  Context1:  The context arguments are private plug and play structures
**             used by the stream class driver to find the resources for this
**             adapter
**  Context2:
**
** Returns:
**
** This routine returns an NT_STATUS value indicating the result of the
** registration attempt. If a value other than STATUS_SUCCESS is returned, the
** minidriver will be unloaded.
**
** Side Effects:  none
*/

ULONG
DriverEntry(
    PVOID Context1,
    PVOID Context2
    )
{

    INTELCAM_KdPrint (MAX_TRACE, ("'Driver Entry\n")); 

    return USBCAMD_DriverEntry(Context1, 
                               Context2, 
                               sizeof(INTELCAM_DEVICE_CONTEXT),
                               sizeof(INTELCAM_FRAME_CONTEXT),
                               INTELCAM_AdapterReceivePacket);
}


/*
** INTELCAM_Initialize()
**
** On entry the device has been configured and the initial alt
** interface selected -- this is where we may send additional 
** vendor commands to enable the device.
**
** Arguments:
**
** BusDeviceObject - pdo associated with this device
**
** DeviceContext - driver specific context
**
** Returns:
**
** NTSTATUS code
**
** Side Effects:  none
*/

NTSTATUS
INTELCAM_Initialize(
    PDEVICE_OBJECT BusDeviceObject,
    PVOID DeviceContext
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext;
    LARGE_INTEGER dueTime;
    ULONG wait = 0;
    NTSTATUS ntStatus = STATUS_SUCCESS;
    USHORT powerState;
    ULONG length;

    deviceContext = DeviceContext;

    ASSERT_DEVICE_CONTEXT(deviceContext);
    //
    // perform any hardware specific 
    // initialization
    //

    INTELCAM_KdPrint (MAX_TRACE, ("'power up the camera\n"));

#ifdef WAIT_CAMERA_INIT

    // wait 10 seconds for the beta camera

    dueTime.QuadPart = -10000 * 10000;

    (VOID) KeDelayExecutionThread(KernelMode,
                                  FALSE,
                                  &dueTime);
#else

    //
    // NOTE
    // This only works on the production level cameras.
    //

    do {

        dueTime.QuadPart = -10000 * 1000;

        (VOID) KeDelayExecutionThread(KernelMode,
                                      FALSE,
                                      &dueTime);
        length = sizeof(powerState);
        
        ntStatus = USBCAMD_ControlVendorCommand(DeviceContext,
                                         REQUEST_GET_PREFERENCE,
                                         0,
                                         INDEX_PREF_POWER,
                                         &powerState,
                                         &length,
                                         TRUE,
                                         NULL,
                                         NULL);

        ILOGENTRY("PWck", wait, powerState, ntStatus);

        INTELCAM_KdPrint (MAX_TRACE, ("'get power state power = 0x%x, ntStatus %x\n",
            powerState, ntStatus));

        // timeout ater ~10 seconds
        wait++;
        if (wait > 9 ) {
            ILOGENTRY("HWto", wait, 0, ntStatus);
            TEST_TRAP();
            ntStatus = STATUS_UNSUCCESSFUL;
        }

    } while (NT_SUCCESS(ntStatus) && powerState < 3);

#endif /* WAIT_CAMERA_INIT */


    //
    // initialize the camera to default settings
    //

    if (NT_SUCCESS(ntStatus)) {
        ntStatus = USBCAMD_ControlVendorCommand(
                                DeviceContext,
                                REQUEST_SET_PREFERENCE,
                                0,
                                INDEX_PREF_INITIALIZE,
                                NULL,
                                NULL,
                                FALSE,
                                NULL,
                                NULL);
    }

    INTELCAM_KdPrint (MAX_TRACE, ("'InitializeHardware 0x%x\n", ntStatus));

    ILOGENTRY("inHW", 0, 0, ntStatus);
    
    return ntStatus;
}


/*
** INTELCAM_UnInitialize()
**
** Assume the device hardware is gone -- all that needs to be done is to 
** free any allocated resources (like memory).
**
** Arguments:
**
** BusDeviceObject - pdo associated with this device
**
** DeviceContext - driver specific context
**
** Returns:
**
** NTSTATUS code
**
** Side Effects:  none
*/

NTSTATUS
INTELCAM_UnInitialize(
    PDEVICE_OBJECT BusDeviceObject,
    PVOID DeviceContext
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext;
    NTSTATUS ntStatus = STATUS_SUCCESS;

    deviceContext = DeviceContext;
    ASSERT_DEVICE_CONTEXT(deviceContext);

    ExFreePool(deviceContext->Interface);

    INTELCAM_KdPrint (MAX_TRACE, ("'UnInitialize 0x%x\n", ntStatus));
    
    return ntStatus;
}


/*
** INTELCAM_Configure()
**
** Configure the iso streaming Interface:
**
** Called just before the device is configured, this is where we tell
** usbcamd which interface and alternate setting to use for the idle state.
**
** NOTE: The assumption here is that the device will have a single interface 
**  with multiple alt settings and each alt setting has the same number of 
**  pipes.
**
** Arguments:
**
**  BusDeviceObject - device object created by the hub whe can submit
**                  urbs to our device through this deviceObject
**
**  DeviceContext - minidriver device  context
**
**  Interface - USBD interface structure initialized with the proper values
**              for select_configuration. This Interface structure corresponds 
**              a single iso interafce on the device.  This is the drivers 
**              chance to pick a particular alternate setting and pipe 
**              parameters.
**
**
**  ConfigurationDescriptor - USB configuration Descriptor for
**      this device.
**
** Returns:
**
**  NTSTATUS code
**
** Side Effects:  none
*/

NTSTATUS
INTELCAM_Configure(
    IN PDEVICE_OBJECT BusDeviceObject,
    IN PVOID DeviceContext,
    IN OUT PUSBD_INTERFACE_INFORMATION Interface,
    IN PUSB_CONFIGURATION_DESCRIPTOR ConfigurationDescriptor,
    IN OUT PLONG DataPipeIndex,
    IN OUT PLONG SyncPipeIndex
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext;
    NTSTATUS ntStatus = STATUS_SUCCESS;

    deviceContext = DeviceContext;

    deviceContext->Sig = INTELCAM_DEVICE_SIG;
    deviceContext->DefaultFrameRate = INTELCAM_DEFAULT_FRAME_RATE;

    //
    // initilialize any other context stuff
    //

    deviceContext->Interface = ExAllocatePool(NonPagedPool,
                                              Interface->Length);

    *DataPipeIndex = 1;
    *SyncPipeIndex = 0;
    
    if (deviceContext->Interface) {
    
        Interface->AlternateSetting = INTELCAM_IDLE_ALT_SETTING;

        // This interface has two pipes,
        // initialize input parameters to USBD for both pipes.
        // The MaximumTransferSize is the size of the largest
        // buffer we want to submit for an aingle iso urb
        // request.
        //
        Interface->Pipes[INTELCAM_SYNC_PIPE].MaximumTransferSize =
            USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE;
        Interface->Pipes[INTELCAM_DATA_PIPE].MaximumTransferSize =
            1024*32;       // 32k transfer per urb

        RtlCopyMemory(deviceContext->Interface,
                      Interface,
                      Interface->Length);                 

        INTELCAM_KdPrint (MAX_TRACE, ("''size of interface request = %d\n", Interface->Length));

                
    } else {
        ntStatus = STATUS_INSUFFICIENT_RESOURCES;
    }
    

    //
    // return interface number and alternate setting
    //
   
    INTELCAM_KdPrint (MAX_TRACE, ("''Configure 0x%x\n", ntStatus));

    return ntStatus;
}



/*
** INTELCAM_StartVideoCapture()
**
** Arguments:
**
**  BusDeviceObject - device object created by the hub we can submit
**                  urbs to our device through this deviceObject
**
**  DeviceContext - minidriver device  context
**
** Returns:
**
** NTSTATUS code
**
** Side Effects:  none
*/

NTSTATUS
INTELCAM_StartVideoCapture(
    IN PDEVICE_OBJECT BusDeviceObject,
    IN PVOID DeviceContext
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext = DeviceContext;
    NTSTATUS ntStatus;
    
    ASSERT_DEVICE_CONTEXT(deviceContext);
    
    //
    // This is where we select the interface we need and send
    // commands to start capturing
    //
    INTELCAM_KdPrint (MAX_TRACE, ("'starting capture \n"));

    ntStatus = USBCAMD_ControlVendorCommand(DeviceContext,
                                            REQUEST_SET_PREFERENCE,
                                            1,
                                            INDEX_PREF_CAPMOTION,
                                            NULL,
                                            NULL,
                                            FALSE,
                                            NULL,
                                            NULL);
    return ntStatus;        
}

/*
** INTELCAM_AllocateBandwidth()
**
** Called just before the iso video capture stream is
** started, here is where we select the appropriate 
** alternate interface and set up the device to stream.
**
**  Called in connection with the stream class RUN command
**
** Arguments:
**
**  BusDeviceObject - device object created by the hub we can submit
**                  urbs to our device through this deviceObject
**
**  DeviceContext - minidriver device  context
**
**  RawFrameLength - pointer to be filled in with size of buffer needed to 
**                  receive the raw frame data from the packet stream.
**
**  Format - pointer to PKS_DATAFORMAT_VIDEOINFOHEADER associated with this 
**          stream.          
**
** Returns:
**
** NTSTATUS code
**
** Side Effects:  none
*/

NTSTATUS
INTELCAM_AllocateBandwidth(
    IN PDEVICE_OBJECT BusDeviceObject,
    IN PVOID DeviceContext,
    OUT PULONG RawFrameLength,
    IN PVOID Format            // BUGBUG
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext = DeviceContext;
    ULONG PixelCount, bytesInImage, maxDataPacketSize;
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PKS_DATAFORMAT_VIDEOINFOHEADER dataFormatHeader;
    PKS_BITMAPINFOHEADER bmInfoHeader;
    
    USHORT hwSetting;
    
    ASSERT_DEVICE_CONTEXT(deviceContext);
    
    //
    // This is where we select the interface we need and send
    // commands to start capturing
    //
    INTELCAM_KdPrint (MAX_TRACE, ("'starting capture %x\n", Format));

    ntStatus = INTELCAM_ReadRegistry(BusDeviceObject,
                                     DeviceContext);

    *RawFrameLength = 0;
    dataFormatHeader = Format;
    bmInfoHeader = &dataFormatHeader->VideoInfoHeader.bmiHeader;

    //
    // BUGBUG sanity check the format
    //      NOTE: we could look at the haeader here to determine
    //            frame rate to use.

    if (NT_SUCCESS(ntStatus)) {

        INTELCAM_KdPrint (MAX_TRACE, ("'format %d x %d\n", bmInfoHeader->biWidth,
                            bmInfoHeader->biHeight));
    
        PixelCount = bmInfoHeader->biWidth * bmInfoHeader->biHeight;

        bytesInImage = (bmInfoHeader->biWidth * bmInfoHeader->biHeight * 12)/8;
        
        deviceContext->StartOffset_y = 0;
        deviceContext->StartOffset_u = PixelCount;
        deviceContext->StartOffset_v = PixelCount + PixelCount/4;        

        if (bmInfoHeader->biWidth == 176) {
            hwSetting = 4;
        } else {
            hwSetting = 5;
        }

        ntStatus = USBCAMD_ControlVendorCommand(DeviceContext,
                                                REQUEST_SET_PREFERENCE,
                                                hwSetting,
                                                INDEX_PREF_SCALING,
                                                NULL,
                                                NULL,
                                                FALSE,
                                                NULL,
                                                NULL);
    }        

    if (NT_SUCCESS(ntStatus)) {
        //
        // bugbug this is where we select the interface
        // based on requested frame rate.
        //
        deviceContext->Interface->AlternateSetting = 
            InterfaceMap[deviceContext->DefaultFrameRate];
        
        ntStatus = USBCAMD_SelectAlternateInterface(
                        deviceContext,
                        deviceContext->Interface);
    }        

    if (NT_SUCCESS(ntStatus)) {
    
        maxDataPacketSize = 
            deviceContext->Interface->Pipes[INTELCAM_DATA_PIPE].MaximumPacketSize;            
        
        *RawFrameLength = ((bytesInImage / maxDataPacketSize) 
                +1) * maxDataPacketSize;
                
    }

    return ntStatus;        
}

/*
** INTELCAM_FreeBandwidth()
**
** Called after the iso video stream is stopped, this is where we 
** select an alternate interface that uses no bandwidth.
**
** Arguments:
**
**  BusDeviceObject - device object created by the hub we can submit
**                  urbs to our device through this deviceObject
**
**  DeviceContext - minidriver device  context
**
** Returns:
**
**  NTSTATUS code
**
** Side Effects:  none
*/

NTSTATUS
INTELCAM_FreeBandwidth(
    PDEVICE_OBJECT BusDeviceObject,
    PVOID DeviceContext
    )
{
    NTSTATUS ntStatus;
    PINTELCAM_DEVICE_CONTEXT deviceContext = DeviceContext;

    // turn off streaming on the device

    ASSERT_DEVICE_CONTEXT(deviceContext);

    deviceContext->Interface->AlternateSetting = 
        INTELCAM_IDLE_ALT_SETTING;

    ntStatus = USBCAMD_SelectAlternateInterface(
                    deviceContext,
                    deviceContext->Interface);

    return ntStatus;
}


/*
** INTELCAM_StopVideoCapture()
**
** Called after the iso video stream is stopped, this is where we 
** select an alternate interface that uses no bandwidth.
**
** Arguments:
**
**  BusDeviceObject - device object created by the hub we can submit
**                  urbs to our device through this deviceObject
**
**  DeviceContext - minidriver device  context
**
** Returns:
**
**  NTSTATUS code
**
** Side Effects:  none
*/

NTSTATUS
INTELCAM_StopVideoCapture(
    PDEVICE_OBJECT BusDeviceObject,
    PVOID DeviceContext
    )
{
    NTSTATUS ntStatus;
    PINTELCAM_DEVICE_CONTEXT deviceContext = DeviceContext;

    // turn off streaming on the device

    ASSERT_DEVICE_CONTEXT(deviceContext);

    ntStatus = USBCAMD_ControlVendorCommand(DeviceContext,
                                            REQUEST_SET_PREFERENCE,
                                            0,
                                            INDEX_PREF_CAPMOTION,
                                            NULL,
                                            NULL,
                                            FALSE,
                                            NULL,
                                            NULL);

    return ntStatus;
}


/*
** INTELCAM_NewFrame()
**
**  called at DPC level to allow driver to initialize a new video frame
**  context structure
**
** Arguments:
**
**  DeviceContext - minidriver device  context
**
**  FrameContext - frame context to be initialized
**
** Returns:
**
**  NTSTATUS code
**  
** Side Effects:  none
*/

VOID
INTELCAM_NewFrame(
    PVOID DeviceContext,
    PVOID FrameContext
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext = DeviceContext;
     
    INTELCAM_KdPrint (MAX_TRACE, ("'INTELCAM_NewFrame\n"));
    ASSERT_DEVICE_CONTEXT(deviceContext); 
}



/*
** INTELCAM_ProcessUSBPacket()
**
**  called at DPC level to allow driver to determine if this packet is part
**  of the current video frame or a new video frame.
**
**  This function should complete as quickly as possible, any image processing 
**  should be deferred to the ProcessRawFrame routine.
**
** Arguments:
**
**  BusDeviceObject - device object created by the hub we can submit
**                  urbs to our device through this deviceObject
**
**  DeviceContext - minidriver device  context
**
**  CurrentFrameContext - some context for this particular frame
**
**  SyncPacket - iso packet descriptor from sync pipe, not used if the interface
**              has only one pipe.
**
**  SyncBuffer - pointer to data for the sync packet
**
**  DataPacket - iso packet descriptor from data pipe
**
**  DataBuffer - pointer to data for the data packet
**
**  FrameComplete - indicates to usbcamd that this is the first data packet 
**          for a new video frame 
**
** Returns:
**  
** number of bytes that should be copied in to the rawFrameBuffer of FrameBuffer.
**
** Side Effects:  none
*/


ULONG
INTELCAM_ProcessUSBPacket(
    PDEVICE_OBJECT BusDeviceObject,
    PVOID DeviceContext,
    PVOID CurrentFrameContext,
    PUSBD_ISO_PACKET_DESCRIPTOR SyncPacket,
    PVOID SyncBuffer,
    PUSBD_ISO_PACKET_DESCRIPTOR DataPacket,
    PVOID DataBuffer,
    PBOOLEAN FrameComplete,
    PBOOLEAN NextFrameIsStill
    )
{
    UCHAR syncByte;
    ULONG length = 0, maxDataPacketSize;
    PINTELCAM_DEVICE_CONTEXT deviceContext = DeviceContext;

    INTELCAM_KdPrint (MAX_TRACE, ("'INTELCAM_ProcessUSBPacket\n"));

    *NextFrameIsStill = FALSE;
    ASSERT_DEVICE_CONTEXT(deviceContext);

    maxDataPacketSize = 
            deviceContext->Interface->Pipes[INTELCAM_DATA_PIPE].MaximumPacketSize;            

    syncByte = * ((PUCHAR) SyncBuffer);

    if (!USBD_SUCCESS(
        SyncPacket->Status)) {

        //
        // if we get an error on the sync frame process the data packet 
        // anyway so that our buffer pointers will be updated
        //
        
        syncByte = 0;
        
    }                
    
    if (syncByte & USBCAMD_SOV) {
        
        //                       
        // start of new video frame
        //

        *FrameComplete = TRUE;

        //
        // copy the packet to the raw frame buffer
        //
        
        length = maxDataPacketSize;

     } else {
     
        if (syncByte & USBCAMD_ERROR) {
        
            //                    
            // error reported for this packet, I beleive we are 
            // supposed to ignore the packet.
            //
#if DBG
            deviceContext->IgnorePacketCount++;  
#endif                
//            ILOGENTRY("igPe", FrameContext, 0, 0);
            
        } else if (syncByte & USBCAMD_I) {
        
            //
            // interfield data reported for this packet, I beleive we are 
            // supposed to ignore the packet.
            //
#if DBG
            deviceContext->IgnorePacketCount++;  
#endif                
//            ILOGENTRY("igPi", FrameContext, 0, 0);

        } else {
        
            //
            // copy the packet to the raw frame buffer
            //
            
            length = maxDataPacketSize;
        }
        
    }   

    return length;
}



/*
** INTELCAM_ProcessRawVideoFrame()
**
**  Called at PASSIVE level to allow driver to perform any decoding of the 
**  raw video frame.
**
**    This routine will convert the packetized data in to the fromat 
**    the CODEC expects, ie y,u,v
**
**    data is always of the form 256y 64u 64v (384 byte chunks) regardless of USB
**    packet size.
**
**
** Arguments:
**
**  DeviceContext - driver context
**
**  FrameContext - driver context for this frame
**
**  FrameBuffer - pointer to the buffer that should receive the final 
**              processed video frame.
**
**  FrameLength - length of the Frame buffer (from the original read 
**                  request)
**
**  RawFrameBuffer - pointer to buffer containing the received USB packets
** 
**  RawFrameLength - length of the raw frame.
**
**  NumberOfPackets - number of USB packets received in to the RawFrameBuffer
**
**  BytesReturned - pointer to value to return for number of bytes read.
**              
** Returns:
**
**  NT status completion code for the read irp
**  
** Side Effects:  none
*/

NTSTATUS
INTELCAM_ProcessRawVideoFrame(
    PDEVICE_OBJECT BusDeviceObject,
    PVOID DeviceContext,
    PVOID FrameContext,
    PVOID FrameBuffer,
    ULONG FrameLength,
    PVOID RawFrameBuffer,
    ULONG RawFrameLength,
    ULONG NumberOfPackets,
    PULONG BytesReturned
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext = DeviceContext;
    PINTELCAM_FRAME_CONTEXT frameContext = FrameContext;
    ULONG i, rawDataLength, processedDataLength, maxDataPacketSize;
    PUCHAR frameBuffer, rawFrameBuffer, frameBufferEnd;

    //TEST_TRAP();
    ASSERT_DEVICE_CONTEXT(deviceContext);

    frameBuffer = FrameBuffer;
    rawFrameBuffer = RawFrameBuffer;
    frameBufferEnd = frameBuffer+FrameLength;
    processedDataLength = 0;        
    
    ASSERT(RawFrameLength != 0);
    
    //
    // raw frame is in the format 256 bytes y 64 bytes u 64 bytes v
    //
    // we need to convert this to y u v in the frame buffer
    //

    INTELCAM_KdPrint (MAX_TRACE, ("'frameBuffer = 0x%x  max = %d numpackets = %d\n", frameBuffer,
        FrameLength, NumberOfPackets)); 

    maxDataPacketSize = 
            deviceContext->Interface->Pipes[INTELCAM_DATA_PIPE].MaximumPacketSize;            

    rawDataLength = NumberOfPackets * maxDataPacketSize;        

    INTELCAM_KdPrint (MAX_TRACE, ("'maxDataPacketSize = 0x%x  rawDataLength = 0x%x\n",
        maxDataPacketSize, rawDataLength)); 
    ILOGENTRY("prfr", 0, 0, 0);        

    INTELCAM_KdPrint (MAX_TRACE, ("'so_u = 0x%x so_v 0x%x\n",
        deviceContext->StartOffset_u, deviceContext->StartOffset_v)); 
    
    for (i=0; i<rawDataLength/384; i++) {

        //
        // copy the y
        //
        
        // check for buffer overrun
        //ASSERT(frameBuffer + (i*256) + 256 <= frameBufferEnd);

        if (i*256+256 <= deviceContext->StartOffset_u) {
            RtlCopyMemory(frameBuffer + (i*256), 
                          rawFrameBuffer,
                          256);
            processedDataLength+=256;                          
        }                          

        //
        // copy the u
        //

        // check for buffer overrun
        //ASSERT(frameBuffer + deviceContext->StartOffset_u + 
        //     (i*64) + 64 <= frameBufferEnd);

        if (i*64+deviceContext->StartOffset_u+64 <= deviceContext->StartOffset_v) {
            RtlCopyMemory(frameBuffer +
                          deviceContext->StartOffset_u + (i*64), 
                          rawFrameBuffer+256,
                          64);                        
            processedDataLength+=64;              
        }                          

        //
        // copy the v
        //

        // check for buffer overrun
        //ASSERT(frameBuffer + deviceContext->StartOffset_v + 
        //    (i*64) + 64 <= frameBufferEnd);

        if (deviceContext->StartOffset_v+(i*64)+64 <= FrameLength) {
            RtlCopyMemory(frameBuffer+
                          deviceContext->StartOffset_v + (i*64), 
                          rawFrameBuffer+256+64,
                          64);      
            processedDataLength+=64;                          
        }                          

        rawFrameBuffer+=384;
        //processedDataLength+=384;

    }

    *BytesReturned = processedDataLength;

    INTELCAM_KdPrint (MAX_TRACE, ("'INTELCAM_ProcessRawVideoFrame return length = %d\n", 
        processedDataLength));

    return STATUS_SUCCESS;
    
}

/*
** INTELCAM_PropertyRequest()
**
** Arguments:
**
**  DeviceContext - driver context
**
** Returns:
**
**  NT status completion code for the read irp
**  
** Side Effects:  none
*/

NTSTATUS
INTELCAM_PropertyRequest(
    BOOLEAN SetProperty,
    PDEVICE_OBJECT BusDeviceObject,
    PVOID DeviceContext,
    PVOID PropertyContext
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PHW_STREAM_REQUEST_BLOCK srb = (PHW_STREAM_REQUEST_BLOCK)PropertyContext;
    PSTREAM_PROPERTY_DESCRIPTOR propertyDescriptor;

    propertyDescriptor = srb->CommandData.PropertyInfo;


    //
    // identify the property to set
    //
    
    if (IsEqualGUID(&PROPSETID_VIDCAP_VIDEOPROCAMP,
                    &propertyDescriptor->Property->Set)) {
        if (SetProperty) {
            ntStatus = INTELCAM_SetCameraProperty(DeviceContext, srb);
        } else {
            ntStatus = INTELCAM_GetCameraProperty(DeviceContext, srb);
        }
    } else if (IsEqualGUID(&PROPSETID_INTEL_CUSTOM_PROP,
               &propertyDescriptor->Property->Set)) {
        if (SetProperty) {
            ntStatus = INTELCAM_SetCustomProperty(DeviceContext, srb);
        } else {
            ntStatus = INTELCAM_GetCustomProperty(DeviceContext, srb);
        }           
    } else {
        ntStatus = STATUS_NOT_SUPPORTED;
    }

    return ntStatus;
}

/*
** INTELCAM_SaveState()
**
** Arguments:
**
** Returns:
**
** Side Effects:  none
*/

NTSTATUS
INTELCAM_SaveState(
    PDEVICE_OBJECT BusDeviceObject,
    PVOID DeviceContext
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext = DeviceContext;


    return STATUS_SUCCESS;
}    


/*
** INTELCAM_RestoreState()
**
** Arguments:
**
** Returns:
**
** Side Effects:  none
*/

NTSTATUS
INTELCAM_RestoreState(
    PDEVICE_OBJECT BusDeviceObject,
    PVOID DeviceContext
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext = DeviceContext;


    return STATUS_SUCCESS;
}    


/*
** INTELCAM_ReadRegistry()
**
** Arguments:
**
** Returns:
**
** Side Effects:  none
*/

NTSTATUS
INTELCAM_ReadRegistry(
    PDEVICE_OBJECT BusDeviceObject,
    PVOID DeviceContext
    )
{
    PINTELCAM_DEVICE_CONTEXT deviceContext = DeviceContext;
    NTSTATUS ntStatus;
    HANDLE handle;

    //keys we have
    WCHAR defaultFrameRateKey[] = L"FrameRate";

    INTELCAM_KdPrint (MAX_TRACE, ("'read registry parameters\n")); 
    
    ntStatus=IoOpenDeviceRegistryKey(BusDeviceObject,
                                     PLUGPLAY_REGKEY_DRIVER,
                                     STANDARD_RIGHTS_ALL,
                                     &handle);
                                     
    if (NT_SUCCESS(ntStatus)) {

        USBCAMD_GetRegistryKeyValue(handle,
                                 defaultFrameRateKey,
                                 sizeof(defaultFrameRateKey),
                                 &deviceContext->DefaultFrameRate,
                                 sizeof(deviceContext->DefaultFrameRate));

        ZwClose(handle);
    }

    if (deviceContext->DefaultFrameRate > INTELCAM_MAX_FRAME_RATE) {
        deviceContext->DefaultFrameRate = INTELCAM_MAX_FRAME_RATE;
    }        

    INTELCAM_KdPrint (MAX_TRACE, ("'REG: DefaultFrameRate = %d %x\n", 
        deviceContext->DefaultFrameRate, ntStatus)); 

    return ntStatus;
}    





